/*****************************************************************************
 * bitstream.h: bitstream writing
 *****************************************************************************
 * Copyright (C) 2003-2014 x264 project
 *
 * Authors: Loren Merritt <lorenm@u.washington.edu>
 *          Fiona Glaser <fiona@x264.com>
 *          Laurent Aimar <fenrir@via.ecp.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 * This program is also available under a commercial proprietary license.
 * For more information, contact us at licensing@x264.com.
 *****************************************************************************/



#define LEVEL_TABLE_SIZE 128

struct _vlc_t {
    uint8_t i_bits;
    uint8_t i_size;
} ;

typedef struct _vlc_t vlc_t;

struct  _vlc_large_t{
    uint16_t i_bits;
    uint8_t i_size;
    /* Next level table to use */
    uint8_t i_next;
} ;
typedef struct  _vlc_large_t vlc_large_t;


 struct bs_s {
    uint8_t *p_start;
    uint8_t *p;
    uint8_t *p_end;

    uintptr_t cur_bits;
    int i_left; /* i_count number of available bits */
    int i_bits_encoded; /* RD only */
};

typedef struct bs_s bs_t;

struct _x264_run_level_t{
    int32_t last;
    int32_t mask;
    dctcoef level[18];
};

typedef struct _x264_run_level_t x264_run_level_t;

extern const vlc_t x264_coeff0_token[6];
extern const vlc_t x264_coeff_token[6][16][4];
extern const vlc_t x264_total_zeros[15][16];
extern const vlc_t x264_total_zeros_2x2_dc[3][4];
extern const vlc_t x264_total_zeros_2x4_dc[7][8];

struct _x264_bitstream_function_t {
    uint8_t *(*nal_escape) (uint8_t *dst, uint8_t *src, uint8_t *end);
    void (*cabac_block_residual_internal)(dctcoef *l, int b_interlaced,
            intptr_t ctx_block_cat, x264_cabac_t *cb);
    void (*cabac_block_residual_rd_internal)(dctcoef *l, int b_interlaced,
            intptr_t ctx_block_cat, x264_cabac_t *cb);
    void (*cabac_block_residual_8x8_rd_internal)(dctcoef *l, int b_interlaced,
            intptr_t ctx_block_cat, x264_cabac_t *cb);
};

typedef struct _x264_bitstream_function_t x264_bitstream_function_t;


void x264_bitstream_init(int cpu, x264_bitstream_function_t *pf);

/* A larger level table size theoretically could help a bit at extremely
 * high bitrates, but the cost in cache is usually too high for it to be
 * useful.
 * This size appears to be optimal for QP18 encoding on a Nehalem CPU.
 * FIXME: Do further testing? */

extern vlc_large_t x264_level_token[7][LEVEL_TABLE_SIZE];

/* The longest possible set of zero run codes sums to 25 bits.  This leaves
 * plenty of room for both the code (25 bits) and size (5 bits) in a uint32_t. */

extern uint32_t x264_run_before[1 << 16];

void bs_init(bs_t *s, void *p_data, int i_data);
int bs_pos(bs_t *s);
/* Write the rest of cur_bits to the bitstream; results in a bitstream no longer 32-bit aligned. */
void bs_flush(bs_t *s);

/* The inverse of bs_flush: prepare the bitstream to be written to again. */
void bs_realign(bs_t *s) ;

void bs_write(bs_t *s, int i_count, uint32_t i_bits);

/* Special case to eliminate branch in normal bs_write. */

/* Golomb never writes an even-size code, so this is only used in slice headers. */
void bs_write32(bs_t *s, uint32_t i_bits) ;

void bs_write1(bs_t *s, uint32_t i_bit);

void bs_align_0(bs_t *s);

void bs_align_1(bs_t *s);

void bs_align_10(bs_t *s);

/* golomb functions */

extern const uint8_t x264_ue_size_tab[256];











void bs_write_ue_big(bs_t *s, unsigned int val);
/* Only works on values under 255. */
void bs_write_ue(bs_t *s, int val);

void bs_write_se(bs_t *s, int val);

void bs_write_te(bs_t *s, int x, int val);

void bs_rbsp_trailing(bs_t *s);













 int bs_size_ue(unsigned int val);

 int bs_size_ue_big(unsigned int val);

 int bs_size_se(int val);
 int bs_size_te(int x, int val);
