
#include "common.h"

#include "cabacConstHeader.h"
const uint16_t x264_cabac_entropy [128] = {FIX8 (0.0273), FIX8 (5.7370), FIX8 (0.0288), FIX8 (5.6618), FIX8 (0.0303), FIX8 (5.5866), FIX8 (0.0320), FIX8 (5.5114), FIX8 (0.0337), FIX8 (5.4362), FIX8 (0.0355), FIX8 (5.3610), FIX8 (0.0375), FIX8 (5.2859), FIX8 (0.0395), FIX8 (5.2106), FIX8 (0.0416), FIX8 (5.1354), FIX8 (0.0439), FIX8 (5.0602), FIX8 (0.0463), FIX8 (4.9851), FIX8 (0.0488), FIX8 (4.9099), FIX8 (0.0515), FIX8 (4.8347), FIX8 (0.0543), FIX8 (4.7595), FIX8 (0.0572), FIX8 (4.6843), FIX8 (0.0604), FIX8 (4.6091), FIX8 (0.0637), FIX8 (4.5339), FIX8 (0.0671), FIX8 (4.4588), FIX8 (0.0708), FIX8 (4.3836), FIX8 (0.0747), FIX8 (4.3083), FIX8 (0.0788), FIX8 (4.2332), FIX8 (0.0832), FIX8 (4.1580), FIX8 (0.0878), FIX8 (4.0828), FIX8 (0.0926), FIX8 (4.0076), FIX8 (0.0977), FIX8 (3.9324), FIX8 (0.1032), FIX8 (3.8572), FIX8 (0.1089), FIX8 (3.7820), FIX8 (0.1149), FIX8 (3.7068), FIX8 (0.1214), FIX8 (3.6316), FIX8 (0.1282), FIX8 (3.5565), FIX8 (0.1353), FIX8 (3.4813), FIX8 (0.1429), FIX8 (3.4061), FIX8 (0.1510), FIX8 (3.3309), FIX8 (0.1596), FIX8 (3.2557), FIX8 (0.1686), FIX8 (3.1805), FIX8 (0.1782), FIX8 (3.1053), FIX8 (0.1884), FIX8 (3.0301), FIX8 (0.1992), FIX8 (2.9549), FIX8 (0.2107), FIX8 (2.8797), FIX8 (0.2229), FIX8 (2.8046), FIX8 (0.2358), FIX8 (2.7294), FIX8 (0.2496), FIX8 (2.6542), FIX8 (0.2642), FIX8 (2.5790), FIX8 (0.2798), FIX8 (2.5038), FIX8 (0.2964), FIX8 (2.4286), FIX8 (0.3142), FIX8 (2.3534), FIX8 (0.3331), FIX8 (2.2782), FIX8 (0.3532), FIX8 (2.2030), FIX8 (0.3748), FIX8 (2.1278), FIX8 (0.3979), FIX8 (2.0527), FIX8 (0.4226), FIX8 (1.9775), FIX8 (0.4491), FIX8 (1.9023), FIX8 (0.4776), FIX8 (1.8271), FIX8 (0.5082), FIX8 (1.7519), FIX8 (0.5412), FIX8 (1.6767), FIX8 (0.5768), FIX8 (1.6015), FIX8 (0.6152), FIX8 (1.5263), FIX8 (0.6568), FIX8 (1.4511), FIX8 (0.7020), FIX8 (1.3759), FIX8 (0.7513), FIX8 (1.3008), FIX8 (0.8050), FIX8 (1.2256), FIX8 (0.8638), FIX8 (1.1504), FIX8 (0.9285), FIX8 (1.0752), FIX8 (1.0000), FIX8 (1.0000)};
uint8_t x264_cabac_contexts [4] [QP_MAX_SPEC + 1] [1024];

void x264_cabac_init (x264_t *h) {
    int ctx_count = CHROMA444 ? 1024 : 460;
    for (int i = 0; i < 4; i++) {
        const int8_t (*cabac_context_init) [1024] [2] = i == 0 ? &x264_cabac_context_init_I : &x264_cabac_context_init_PB[i - 1];
        for (int qp = 0; qp <= QP_MAX_SPEC; qp++)
            for (int j = 0; j < ctx_count; j++) {
                int state = x264_clip3 ((((*cabac_context_init)[j][0] * qp) >> 4) + (*cabac_context_init)[j][1], 1, 126);
                x264_cabac_contexts[i][qp][j] = (X264_MIN (state, 127 - state) << 1) | (state >> 6);
            }
    }
}

void x264_cabac_context_init (x264_t *h, x264_cabac_t *cb, int i_slice_type, int i_qp, int i_model) {
    memcpy (cb -> state, x264_cabac_contexts [i_slice_type == SLICE_TYPE_I ? 0 : i_model + 1] [i_qp], CHROMA444 ? 1024 : 460);
}

void x264_cabac_encode_init_core (x264_cabac_t *cb) {
    cb->i_low = 0;
    cb->i_range = 0x01FE;
    cb->i_queue = -9;
    cb->i_bytes_outstanding = 0;
}

void x264_cabac_encode_init (x264_cabac_t *cb, uint8_t *p_data, uint8_t *p_end) {
    x264_cabac_encode_init_core (cb);
    cb->p_start = p_data;
    cb->p = p_data;
    cb->p_end = p_end;
}

void x264_cabac_putbyte (x264_cabac_t *cb) {
    if (cb->i_queue >= 0) {
        int out = cb->i_low >> (cb->i_queue + 10);
        cb->i_low &= (0x400 << cb->i_queue) - 1;
        cb->i_queue -= 8;
        if ((out & 0xff) == 0xff)
            cb->i_bytes_outstanding++;
        else {
            int carry = out >> 8;
            int bytes_outstanding = cb->i_bytes_outstanding;
            cb->p[-1] += carry;
            while (bytes_outstanding > 0) {
                *(cb->p++) = carry - 1;
                bytes_outstanding--;
            }
            *(cb->p++) = out;
            cb->i_bytes_outstanding = 0;
        }
    }
}

void x264_cabac_encode_renorm (x264_cabac_t *cb) {
    int shift = x264_cabac_renorm_shift[cb->i_range >> 3];
    cb->i_range <<= shift;
    cb->i_low <<= shift;
    cb->i_queue += shift;
    x264_cabac_putbyte (cb);
}

void x264_cabac_encode_decision_c (x264_cabac_t *cb, int i_ctx, int b) {
    int i_state = cb->state[i_ctx];
    int i_range_lps = x264_cabac_range_lps[i_state >> 1][(cb->i_range >> 6) - 4];
    cb->i_range -= i_range_lps;
    if (b != (i_state & 1)) {
        cb->i_low += cb->i_range;
        cb->i_range = i_range_lps;
    }
    cb->state[i_ctx] = x264_cabac_transition[i_state][b];
    x264_cabac_encode_renorm (cb);
}

void x264_cabac_encode_bypass_c (x264_cabac_t *cb, int b) {
    cb->i_low <<= 1;
    cb->i_low += b & cb->i_range;
    cb->i_queue += 1;
    x264_cabac_putbyte (cb);
}

const int bypass_lut [16] = {-1, 0x2, 0x14, 0x68, 0x1d0, 0x7a0, 0x1f40, 0x7e80, 0x1fd00, 0x7fa00, 0x1ff400, 0x7fe800, 0x1ffd000, 0x7ffa000, 0x1fff4000, 0x7ffe8000};

void x264_cabac_encode_ue_bypass (x264_cabac_t *cb, int exp_bits, int val) {
    uint32_t v = val + (1 << exp_bits);
    int k = 31 - x264_clz (v);
    uint32_t x = (bypass_lut[k - exp_bits] << exp_bits) + v;
    k = 2 * k + 1 - exp_bits;
    int i = ((k - 1) & 7) + 1;
    do {
        k -= i;
        cb->i_low <<= i;
        cb->i_low += ((x >> k) & 0xff) * cb->i_range;
        cb->i_queue += i;
        x264_cabac_putbyte (cb);
        i = 8;
    }
    while (k > 0);
}

void x264_cabac_encode_terminal_c (x264_cabac_t *cb) {
    cb->i_range -= 2;
    x264_cabac_encode_renorm (cb);
}

void x264_cabac_encode_flush (x264_t *h, x264_cabac_t *cb) {
    cb->i_low += cb->i_range - 2;
    cb->i_low |= 1;
    cb->i_low <<= 9;
    cb->i_queue += 9;
    x264_cabac_putbyte (cb);
    x264_cabac_putbyte (cb);
    cb->i_low <<= -cb->i_queue;
    cb->i_low |= (0x35a4e4f5 >> (h->i_frame & 31) & 1) << 10;
    cb->i_queue = 0;
    x264_cabac_putbyte (cb);
    while (cb->i_bytes_outstanding > 0) {
        *(cb->p++) = 0xff;
        cb->i_bytes_outstanding--;
    }
}

int x264_cabac_pos (x264_cabac_t *cb) {
    return (cb->p - cb->p_start + cb->i_bytes_outstanding) * 8 + cb->i_queue;
}

void x264_cabac_size_decision (x264_cabac_t *cb, long  i_ctx, long  b) {
    int i_state = cb->state[i_ctx];
    cb->state[i_ctx] = x264_cabac_transition[i_state][b];
    cb->f8_bits_encoded += x264_cabac_entropy[i_state ^ b];
}

int x264_cabac_size_decision2 (uint8_t *state, long  b) {
    int i_state = *state;
    *state = x264_cabac_transition[i_state][b];
    return x264_cabac_entropy[i_state ^ b];
}

void x264_cabac_size_decision_noup (x264_cabac_t *cb, long  i_ctx, long  b) {
    int i_state = cb->state[i_ctx];
    cb->f8_bits_encoded += x264_cabac_entropy[i_state ^ b];
}

int x264_cabac_size_decision_noup2 (uint8_t *state, long  b) {
    return x264_cabac_entropy[*state ^ b];
}

