
#include "common.h"

#include "macroblockENCODER.h"

void x264_cabac_mb_type_intra (x264_t *h, x264_cabac_t *cb, int i_mb_type, int ctx0, int ctx1, int ctx2, int ctx3, int ctx4, int ctx5) {
    if (i_mb_type == I_4x4 || i_mb_type == I_8x8) {
        x264_cabac_encode_decision_c (cb, ctx0, 0);
    }
    else if (i_mb_type == I_PCM) {
        x264_cabac_encode_decision_c (cb, ctx0, 1);
        x264_cabac_encode_flush (h, cb);
    }
    else {
        int i_pred = x264_mb_pred_mode16x16_fix[h->mb.i_intra16x16_pred_mode];
        x264_cabac_encode_decision_c (cb, ctx0, 1);
        x264_cabac_encode_terminal_c (cb);
        x264_cabac_encode_decision_c (cb, ctx1, ! ! h -> mb.i_cbp_luma);
        if (h->mb.i_cbp_chroma == 0)
            x264_cabac_encode_decision_c (cb, ctx2, 0);
        else {
            x264_cabac_encode_decision_c (cb, ctx2, 1);
            x264_cabac_encode_decision_c (cb, ctx3, h -> mb.i_cbp_chroma >> 1);
        }
        x264_cabac_encode_decision_c (cb, ctx4, i_pred >> 1);
        x264_cabac_encode_decision_c (cb, ctx5, i_pred & 1);
    }
}

void x264_cabac_field_decoding_flag (x264_t *h, x264_cabac_t *cb) {
    int ctx = 0;
    ctx += h->mb.field_decoding_flag & !!h->mb.i_mb_x;
    ctx += (h->mb.i_mb_top_mbpair_xy >= 0 && h->mb.slice_table[h->mb.i_mb_top_mbpair_xy] == h->sh.i_first_mb && h->mb.field[h->mb.i_mb_top_mbpair_xy]);
    x264_cabac_encode_decision_c (cb, 70 + ctx, 0);
    h->mb.field_decoding_flag = 0;
}

void x264_cabac_intra4x4_pred_mode (x264_cabac_t *cb, int i_pred, int i_mode) {
    if (i_pred == i_mode)
        x264_cabac_encode_decision_c (cb, 68, 1);
    else {
        x264_cabac_encode_decision_c (cb, 68, 0);
        if (i_mode > i_pred)
            i_mode--;
        x264_cabac_encode_decision_c (cb, 69, (i_mode) & 0x01);
        x264_cabac_encode_decision_c (cb, 69, (i_mode >> 1) & 0x01);
        x264_cabac_encode_decision_c (cb, 69, (i_mode >> 2));
    }
}

void x264_cabac_intra_chroma_pred_mode (x264_t *h, x264_cabac_t *cb) {
    int i_mode = x264_mb_chroma_pred_mode_fix[h->mb.i_chroma_pred_mode];
    int ctx = 0;
    if ((h->mb.i_neighbour & MB_LEFT) && h->mb.chroma_pred_mode[h->mb.i_mb_left_xy[0]] != 0)
        ctx++;
    if ((h->mb.i_neighbour & MB_TOP) && h->mb.chroma_pred_mode[h->mb.i_mb_top_xy] != 0)
        ctx++;
    x264_cabac_encode_decision_c (cb, 64 + ctx, i_mode > 0);
    if (i_mode > 0) {
        x264_cabac_encode_decision_c (cb, 64 + 3, i_mode > 1);
        if (i_mode > 1)
            x264_cabac_encode_decision_c (cb, 64 + 3, i_mode > 2);
    }
}

void x264_cabac_cbp_luma (x264_t *h, x264_cabac_t *cb) {
    int cbp = h->mb.i_cbp_luma;
    int cbp_l = h->mb.cache.i_cbp_left;
    int cbp_t = h->mb.cache.i_cbp_top;
    x264_cabac_encode_decision_c (cb, 76 - ((cbp_l >> 1) & 1) - ((cbp_t >> 1) & 2), (cbp >> 0) & 1);
    x264_cabac_encode_decision_c (cb, 76 - ((cbp >> 0) & 1) - ((cbp_t >> 2) & 2), (cbp >> 1) & 1);
    x264_cabac_encode_decision_c (cb, 76 - ((cbp_l >> 3) & 1) - ((cbp << 1) & 2), (cbp >> 2) & 1);
    x264_cabac_encode_decision_c (cb, 76 - ((cbp >> 2) & 1) - ((cbp >> 0) & 2), (cbp >> 3) & 1);
}

void x264_cabac_cbp_chroma (x264_t *h, x264_cabac_t *cb) {
    int cbp_a = h->mb.cache.i_cbp_left & 0x30;
    int cbp_b = h->mb.cache.i_cbp_top & 0x30;
    int ctx = 0;
    if (cbp_a && h->mb.cache.i_cbp_left != -1)
        ctx++;
    if (cbp_b && h->mb.cache.i_cbp_top != -1)
        ctx += 2;
    if (h->mb.i_cbp_chroma == 0)
        x264_cabac_encode_decision_c (cb, 77 + ctx, 0);
    else {
        x264_cabac_encode_decision_c (cb, 77 + ctx, 1);
        ctx = 4;
        if (cbp_a == 0x20)
            ctx++;
        if (cbp_b == 0x20)
            ctx += 2;
        x264_cabac_encode_decision_c (cb, 77 + ctx, h -> mb.i_cbp_chroma >> 1);
    }
}

void x264_cabac_qp_delta (x264_t *h, x264_cabac_t *cb) {
    int i_dqp = h->mb.i_qp - h->mb.i_last_qp;
    int ctx;
    if (h->mb.i_type == I_16x16 && !h->mb.cbp[h->mb.i_mb_xy] && h->mb.i_qp > h->mb.i_last_qp) {
        h->mb.i_qp = h->mb.i_last_qp;
        i_dqp = 0;
    }
    ctx = h->mb.i_last_dqp && (h->mb.type[h->mb.i_mb_prev_xy] == I_16x16 || (h->mb.cbp[h->mb.i_mb_prev_xy] & 0x3f));
    if (i_dqp != 0) {
        i_dqp *= 2;
        int val = 1 - i_dqp;
        if (val < 0)
            val = i_dqp;
        val--;
        if (val >= (51 + (6 * (8 - 8))) && val != (51 + (6 * (8 - 8))) + 1)
            val = 2 * (51 + (6 * (8 - 8))) + 1 - val;
        do {
            x264_cabac_encode_decision_c (cb, 60 + ctx, 1);
            ctx = 2 + (ctx >> 1);
        }
        while (--val);
    }
    x264_cabac_encode_decision_c (cb, 60 + ctx, 0);
}

void x264_cabac_mb_skip (x264_t *h, int b_skip) {
    int ctx = h->mb.cache.i_neighbour_skip + 11;
    if (h->sh.i_type != SLICE_TYPE_P)
        ctx += 13;
    x264_cabac_encode_decision_c (& h -> cabac, ctx, b_skip);
}

void x264_cabac_subpartition_p (x264_cabac_t *cb, int i_sub) {
    if (i_sub == D_L0_8x8) {
        x264_cabac_encode_decision_c (cb, 21, 1);
        return;
    }
    x264_cabac_encode_decision_c (cb, 21, 0);
    if (i_sub == D_L0_8x4)
        x264_cabac_encode_decision_c (cb, 22, 0);
    else {
        x264_cabac_encode_decision_c (cb, 22, 1);
        x264_cabac_encode_decision_c (cb, 23, i_sub == D_L0_4x8);
    }
}

void x264_cabac_subpartition_b (x264_cabac_t *cb, int i_sub) {
    if (i_sub == D_DIRECT_8x8) {
        x264_cabac_encode_decision_c (cb, 36, 0);
        return;
    }
    x264_cabac_encode_decision_c (cb, 36, 1);
    if (i_sub == D_BI_8x8) {
        x264_cabac_encode_decision_c (cb, 37, 1);
        x264_cabac_encode_decision_c (cb, 38, 0);
        x264_cabac_encode_decision_c (cb, 39, 0);
        x264_cabac_encode_decision_c (cb, 39, 0);
        return;
    }
    x264_cabac_encode_decision_c (cb, 37, 0);
    x264_cabac_encode_decision_c (cb, 39, i_sub == D_L1_8x8);
}

void x264_cabac_transform_size (x264_t *h, x264_cabac_t *cb) {
    int ctx = 399 + h->mb.cache.i_neighbour_transform_size;
    x264_cabac_encode_decision_c (cb, ctx, h -> mb.b_transform_8x8);
}

void x264_cabac_ref_internal (x264_t *h, x264_cabac_t *cb, int i_list, int idx, int bframe) {
    const int i8 = x264_scan8[idx];
    const int i_refa = h->mb.cache.ref[i_list][i8 - 1];
    const int i_refb = h->mb.cache.ref[i_list][i8 - 8];
    int ctx = 0;
    if (i_refa > 0 && (!bframe || !h->mb.cache.skip[i8 - 1]))
        ctx++;
    if (i_refb > 0 && (!bframe || !h->mb.cache.skip[i8 - 8]))
        ctx += 2;
    for (int i_ref = h->mb.cache.ref[i_list][i8]; i_ref > 0; i_ref--) {
        x264_cabac_encode_decision_c (cb, 54 + ctx, 1);
        ctx = (ctx >> 2) + 4;
    }
    x264_cabac_encode_decision_c (cb, 54 + ctx, 0);
}

void x264_cabac_ref_p (x264_t *h, x264_cabac_t *cb, int idx) {
    x264_cabac_ref_internal (h, cb, 0, idx, 0);
}

void x264_cabac_ref_b (x264_t *h, x264_cabac_t *cb, int i_list, int idx) {
    x264_cabac_ref_internal (h, cb, i_list, idx, 1);
}

int x264_cabac_mvd_cpn (x264_t *h, x264_cabac_t *cb, int i_list, int idx, int l, int mvd, int ctx) {
    int ctxbase = l ? 47 : 40;
    if (mvd == 0) {
        x264_cabac_encode_decision_c (cb, ctxbase + ctx, 0);
        return 0;
    }
    int i_abs = abs (mvd);
    x264_cabac_encode_decision_c (cb, ctxbase + ctx, 1);
    const uint8_t ctxes [8] = {3, 4, 5, 6, 6, 6, 6, 6};
    if (i_abs < 9) {
        for (int i = 1; i < i_abs; i++)
            x264_cabac_encode_decision_c (cb, ctxbase +ctxes[i - 1], 1);
        x264_cabac_encode_decision_c (cb, ctxbase + ctxes [i_abs - 1], 0);
    }
    else {
        for (int i = 1; i < 9; i++)
            x264_cabac_encode_decision_c (cb, ctxbase +ctxes[i - 1], 1);
        x264_cabac_encode_ue_bypass (cb, 3, i_abs - 9);
    }
    x264_cabac_encode_bypass_c (cb, mvd >> 31);
    return ((i_abs) < (66) ? (i_abs) : (66));
}

uint16_t x264_cabac_mvd (x264_t *h, x264_cabac_t *cb, int i_list, int idx, int width) {
    int16_t mvp [2];
    int mdx, mdy;
    x264_mb_predict_mv (h, i_list, idx, width, mvp);
    mdx = h->mb.cache.mv[i_list][x264_scan8[idx]][0] - mvp[0];
    mdy = h->mb.cache.mv[i_list][x264_scan8[idx]][1] - mvp[1];
    uint16_t amvd = x264_cabac_mvd_sum (h->mb.cache.mvd[i_list][x264_scan8[idx] - 1], h->mb.cache.mvd[i_list][x264_scan8[idx] - 8]);
    mdx = x264_cabac_mvd_cpn (h, cb, i_list, idx, 0, mdx, amvd &0xFF);
    mdy = x264_cabac_mvd_cpn (h, cb, i_list, idx, 1, mdy, amvd >> 8);
    return pack8to16 (mdx, mdy);
}

void x264_cabac_8x8_mvd (x264_t *h, x264_cabac_t *cb, int i) {
    switch (h->mb.i_sub_partition[i]) {
    case D_L0_8x8 :
        do {
            uint16_t mvd = x264_cabac_mvd (h, cb, 0, 4 * i, 2);
            x264_macroblock_cache_mvd (h, block_idx_x [4 * i], block_idx_y [4 * i], 2, 2, 0, mvd);
        }
        while (0);
        break;
    case D_L0_8x4 :
        do {
            uint16_t mvd = x264_cabac_mvd (h, cb, 0, 4 * i + 0, 2);
            x264_macroblock_cache_mvd (h, block_idx_x [4 * i + 0], block_idx_y [4 * i + 0], 2, 1, 0, mvd);
        }
        while (0);
        do {
            uint16_t mvd = x264_cabac_mvd (h, cb, 0, 4 * i + 2, 2);
            x264_macroblock_cache_mvd (h, block_idx_x [4 * i + 2], block_idx_y [4 * i + 2], 2, 1, 0, mvd);
        }
        while (0);
        break;
    case D_L0_4x8 :
        do {
            uint16_t mvd = x264_cabac_mvd (h, cb, 0, 4 * i + 0, 1);
            x264_macroblock_cache_mvd (h, block_idx_x [4 * i + 0], block_idx_y [4 * i + 0], 1, 2, 0, mvd);
        }
        while (0);
        do {
            uint16_t mvd = x264_cabac_mvd (h, cb, 0, 4 * i + 1, 1);
            x264_macroblock_cache_mvd (h, block_idx_x [4 * i + 1], block_idx_y [4 * i + 1], 1, 2, 0, mvd);
        }
        while (0);
        break;
    case D_L0_4x4 :
        do {
            uint16_t mvd = x264_cabac_mvd (h, cb, 0, 4 * i + 0, 1);
            x264_macroblock_cache_mvd (h, block_idx_x [4 * i + 0], block_idx_y [4 * i + 0], 1, 1, 0, mvd);
        }
        while (0);
        do {
            uint16_t mvd = x264_cabac_mvd (h, cb, 0, 4 * i + 1, 1);
            x264_macroblock_cache_mvd (h, block_idx_x [4 * i + 1], block_idx_y [4 * i + 1], 1, 1, 0, mvd);
        }
        while (0);
        do {
            uint16_t mvd = x264_cabac_mvd (h, cb, 0, 4 * i + 2, 1);
            x264_macroblock_cache_mvd (h, block_idx_x [4 * i + 2], block_idx_y [4 * i + 2], 1, 1, 0, mvd);
        }
        while (0);
        do {
            uint16_t mvd = x264_cabac_mvd (h, cb, 0, 4 * i + 3, 1);
            x264_macroblock_cache_mvd (h, block_idx_x [4 * i + 3], block_idx_y [4 * i + 3], 1, 1, 0, mvd);
        }
        while (0);
        break;
    default :
        ((0) ? (void) (0) : __assert_fail ("0", "DonorSourceCode/cabacENCODER.c", 377, __PRETTY_FUNCTION__));
    }
}

void x264_cabac_mb_header_i (x264_t *h, x264_cabac_t *cb, int i_mb_type, int slice_type, int chroma) {
    if (slice_type == SLICE_TYPE_I) {
        int ctx = 0;
        if ((h->mb.i_neighbour & MB_LEFT) && h->mb.i_mb_type_left[0] != I_4x4)
            ctx++;
        if ((h->mb.i_neighbour & MB_TOP) && h->mb.i_mb_type_top != I_4x4)
            ctx++;
        x264_cabac_mb_type_intra (h, cb, i_mb_type, 3 + ctx, 3 + 3, 3 + 4, 3 + 5, 3 + 6, 3 + 7);
    }
    else if (slice_type == SLICE_TYPE_P) {
        x264_cabac_encode_decision_c (cb, 14, 1);
        x264_cabac_mb_type_intra (h, cb, i_mb_type, 17 + 0, 17 + 1, 17 + 2, 17 + 2, 17 + 3, 17 + 3);
    }
    else if (slice_type == SLICE_TYPE_B) {
        x264_cabac_encode_decision_c (cb, 27 + 3, 1);
        x264_cabac_encode_decision_c (cb, 27 + 4, 1);
        x264_cabac_encode_decision_c (cb, 27 + 5, 1);
        x264_cabac_encode_decision_c (cb, 27 + 5, 0);
        x264_cabac_encode_decision_c (cb, 27 + 5, 1);
        x264_cabac_mb_type_intra (h, cb, i_mb_type, 32 + 0, 32 + 1, 32 + 2, 32 + 2, 32 + 3, 32 + 3);
    }
    if (i_mb_type == I_PCM)
        return;
    if (i_mb_type != I_16x16) {
        if (h->pps->b_transform_8x8_mode)
            x264_cabac_transform_size (h, cb);
        int di = h->mb.b_transform_8x8 ? 4 : 1;
        for (int i = 0; i < 16; i += di) {
            const int i_pred = x264_mb_predict_intra4x4_mode (h, i);
            const int i_mode = x264_mb_pred_mode4x4_fix[(h->mb.cache.intra4x4_pred_mode[x264_scan8[i]]) + 1];
            x264_cabac_intra4x4_pred_mode (cb, i_pred, i_mode);
        }
    }
    if (chroma)
        x264_cabac_intra_chroma_pred_mode (h, cb);
}

void x264_cabac_mb_header_p (x264_t *h, x264_cabac_t *cb, int i_mb_type, int chroma) {
    if (i_mb_type == P_L0) {
        x264_cabac_encode_decision_c (cb, 14, 0);
        if (h->mb.i_partition == D_16x16) {
            x264_cabac_encode_decision_c (cb, 15, 0);
            x264_cabac_encode_decision_c (cb, 16, 0);
            if (h->mb.pic.i_fref[0] > 1)
                x264_cabac_ref_p (h, cb, 0);
            do {
                uint16_t mvd = x264_cabac_mvd (h, cb, 0, 0, 4);
                x264_macroblock_cache_mvd (h, block_idx_x [0], block_idx_y [0], 4, 4, 0, mvd);
            }
            while (0);
        }
        else if (h->mb.i_partition == D_16x8) {
            x264_cabac_encode_decision_c (cb, 15, 1);
            x264_cabac_encode_decision_c (cb, 17, 1);
            if (h->mb.pic.i_fref[0] > 1) {
                x264_cabac_ref_p (h, cb, 0);
                x264_cabac_ref_p (h, cb, 8);
            }
            do {
                uint16_t mvd = x264_cabac_mvd (h, cb, 0, 0, 4);
                x264_macroblock_cache_mvd (h, block_idx_x [0], block_idx_y [0], 4, 2, 0, mvd);
            }
            while (0);
            do {
                uint16_t mvd = x264_cabac_mvd (h, cb, 0, 8, 4);
                x264_macroblock_cache_mvd (h, block_idx_x [8], block_idx_y [8], 4, 2, 0, mvd);
            }
            while (0);
        }
        else {
            x264_cabac_encode_decision_c (cb, 15, 1);
            x264_cabac_encode_decision_c (cb, 17, 0);
            if (h->mb.pic.i_fref[0] > 1) {
                x264_cabac_ref_p (h, cb, 0);
                x264_cabac_ref_p (h, cb, 4);
            }
            do {
                uint16_t mvd = x264_cabac_mvd (h, cb, 0, 0, 2);
                x264_macroblock_cache_mvd (h, block_idx_x [0], block_idx_y [0], 2, 4, 0, mvd);
            }
            while (0);
            do {
                uint16_t mvd = x264_cabac_mvd (h, cb, 0, 4, 2);
                x264_macroblock_cache_mvd (h, block_idx_x [4], block_idx_y [4], 2, 4, 0, mvd);
            }
            while (0);
        }
    }
    else if (i_mb_type == P_8x8) {
        x264_cabac_encode_decision_c (cb, 14, 0);
        x264_cabac_encode_decision_c (cb, 15, 0);
        x264_cabac_encode_decision_c (cb, 16, 1);
        for (int i = 0; i < 4; i++)
            x264_cabac_subpartition_p (cb, h->mb.i_sub_partition[i]);
        if (h->mb.pic.i_fref[0] > 1) {
            x264_cabac_ref_p (h, cb, 0);
            x264_cabac_ref_p (h, cb, 4);
            x264_cabac_ref_p (h, cb, 8);
            x264_cabac_ref_p (h, cb, 12);
        }
        for (int i = 0; i < 4; i++)
            x264_cabac_8x8_mvd (h, cb, i);
    }
    else
        x264_cabac_mb_header_i (h, cb, i_mb_type, SLICE_TYPE_P, chroma);
}

void x264_cabac_mb_header_b (x264_t *h, x264_cabac_t *cb, int i_mb_type, int chroma) {
    int ctx = 0;
    if ((h->mb.i_neighbour & MB_LEFT) && h->mb.i_mb_type_left[0] != B_SKIP && h->mb.i_mb_type_left[0] != B_DIRECT)
        ctx++;
    if ((h->mb.i_neighbour & MB_TOP) && h->mb.i_mb_type_top != B_SKIP && h->mb.i_mb_type_top != B_DIRECT)
        ctx++;
    if (i_mb_type == B_DIRECT) {
        x264_cabac_encode_decision_c (cb, 27 + ctx, 0);
        return;
    }
    x264_cabac_encode_decision_c (cb, 27 + ctx, 1);
    if (i_mb_type == B_8x8) {
        x264_cabac_encode_decision_c (cb, 27 + 3, 1);
        x264_cabac_encode_decision_c (cb, 27 + 4, 1);
        x264_cabac_encode_decision_c (cb, 27 + 5, 1);
        x264_cabac_encode_decision_c (cb, 27 + 5, 1);
        x264_cabac_encode_decision_c (cb, 27 + 5, 1);
        for (int i = 0; i < 4; i++)
            x264_cabac_subpartition_b (cb, h->mb.i_sub_partition[i]);
        if (h->mb.pic.i_fref[0] > 1)
            for (int i = 0; i < 4; i++)
                if (x264_mb_partition_listX_table[0][h->mb.i_sub_partition[i]])
                    x264_cabac_ref_b (h, cb, 0, 4 * i);
        if (h->mb.pic.i_fref[1] > 1)
            for (int i = 0; i < 4; i++)
                if (x264_mb_partition_listX_table[1][h->mb.i_sub_partition[i]])
                    x264_cabac_ref_b (h, cb, 1, 4 * i);
        for (int i = 0; i < 4; i++)
            if (x264_mb_partition_listX_table[0][h->mb.i_sub_partition[i]])
                do {
                    uint16_t mvd = x264_cabac_mvd (h, cb, 0, 4 * i, 2);
                    x264_macroblock_cache_mvd (h, block_idx_x [4 * i], block_idx_y [4 * i], 2, 2, 0, mvd);
                }
                while (0);
        for (int i = 0; i < 4; i++)
            if (x264_mb_partition_listX_table[1][h->mb.i_sub_partition[i]])
                do {
                    uint16_t mvd = x264_cabac_mvd (h, cb, 1, 4 * i, 2);
                    x264_macroblock_cache_mvd (h, block_idx_x [4 * i], block_idx_y [4 * i], 2, 2, 1, mvd);
                }
                while (0);
    }
    else if (i_mb_type >= B_L0_L0 && i_mb_type <= B_BI_BI) {
        const uint8_t i_mb_bits [9 * 3] = {0x31, 0x29, 0x4, 0x35, 0x2d, 0, 0x43, 0x63, 0, 0x3d, 0x2f, 0, 0x39, 0x25, 0x6, 0x53, 0x73, 0, 0x4b, 0x6b, 0, 0x5b, 0x7b, 0, 0x47, 0x67, 0x21};
        const int idx = (i_mb_type - B_L0_L0) * 3 + (h->mb.i_partition - D_16x8);
        int bits = i_mb_bits[idx];
        x264_cabac_encode_decision_c (cb, 27 + 3, bits & 1);
        x264_cabac_encode_decision_c (cb, 27 + 5 - (bits & 1), (bits >> 1) & 1);
        bits >>= 2;
        if (bits != 1) {
            x264_cabac_encode_decision_c (cb, 27 + 5, bits & 1);
            bits >>= 1;
            x264_cabac_encode_decision_c (cb, 27 + 5, bits & 1);
            bits >>= 1;
            x264_cabac_encode_decision_c (cb, 27 + 5, bits & 1);
            bits >>= 1;
            if (bits != 1)
                x264_cabac_encode_decision_c (cb, 27 + 5, bits &1);
        }
        const uint8_t (*b_list) [2] = x264_mb_type_list_table[i_mb_type];
        if (h->mb.pic.i_fref[0] > 1) {
            if (b_list[0][0])
                x264_cabac_ref_b (h, cb, 0, 0);
            if (b_list[0][1] && h->mb.i_partition != D_16x16)
                x264_cabac_ref_b (h, cb, 0, 8 >> (h->mb.i_partition == D_8x16));
        }
        if (h->mb.pic.i_fref[1] > 1) {
            if (b_list[1][0])
                x264_cabac_ref_b (h, cb, 1, 0);
            if (b_list[1][1] && h->mb.i_partition != D_16x16)
                x264_cabac_ref_b (h, cb, 1, 8 >> (h->mb.i_partition == D_8x16));
        }
        for (int i_list = 0; i_list < 2; i_list++) {
            if (h->mb.i_partition == D_16x16) {
                if (b_list[i_list][0])
                    do {
                        uint16_t mvd = x264_cabac_mvd (h, cb, i_list, 0, 4);
                        x264_macroblock_cache_mvd (h, block_idx_x [0], block_idx_y [0], 4, 4, i_list, mvd);
                    }
                    while (0);
            }
            else if (h->mb.i_partition == D_16x8) {
                if (b_list[i_list][0])
                    do {
                        uint16_t mvd = x264_cabac_mvd (h, cb, i_list, 0, 4);
                        x264_macroblock_cache_mvd (h, block_idx_x [0], block_idx_y [0], 4, 2, i_list, mvd);
                    }
                    while (0);
                if (b_list[i_list][1])
                    do {
                        uint16_t mvd = x264_cabac_mvd (h, cb, i_list, 8, 4);
                        x264_macroblock_cache_mvd (h, block_idx_x [8], block_idx_y [8], 4, 2, i_list, mvd);
                    }
                    while (0);
            }
            else {
                if (b_list[i_list][0])
                    do {
                        uint16_t mvd = x264_cabac_mvd (h, cb, i_list, 0, 2);
                        x264_macroblock_cache_mvd (h, block_idx_x [0], block_idx_y [0], 2, 4, i_list, mvd);
                    }
                    while (0);
                if (b_list[i_list][1])
                    do {
                        uint16_t mvd = x264_cabac_mvd (h, cb, i_list, 4, 2);
                        x264_macroblock_cache_mvd (h, block_idx_x [4], block_idx_y [4], 2, 4, i_list, mvd);
                    }
                    while (0);
            }
        }
    }
    else
        x264_cabac_mb_header_i (h, cb, i_mb_type, SLICE_TYPE_B, chroma);
}

int x264_cabac_cbf_ctxidxinc (x264_t *h, int i_cat, int i_idx, int b_intra, int b_dc) {
    const uint16_t base_ctx [14] = {85, 89, 93, 97, 101, 1012, 460, 464, 468, 1016, 472, 476, 480, 1020};
    if (b_dc) {
        i_idx -= 48;
        if (i_cat == DCT_CHROMA_DC) {
            int i_nza = h->mb.cache.i_cbp_left != -1 ? (h->mb.cache.i_cbp_left >> (8 + i_idx)) & 1 : b_intra;
            int i_nzb = h->mb.cache.i_cbp_top != -1 ? (h->mb.cache.i_cbp_top >> (8 + i_idx)) & 1 : b_intra;
            return base_ctx[i_cat] + 2 * i_nzb + i_nza;
        }
        else {
            int i_nza = (h->mb.cache.i_cbp_left >> (8 + i_idx)) & 1;
            int i_nzb = (h->mb.cache.i_cbp_top >> (8 + i_idx)) & 1;
            return base_ctx[i_cat] + 2 * i_nzb + i_nza;
        }
    }
    else {
        int i_nza = h->mb.cache.non_zero_count[x264_scan8[i_idx] - 1];
        int i_nzb = h->mb.cache.non_zero_count[x264_scan8[i_idx] - 8];
        if (__builtin_constant_p (b_intra) && !b_intra)
            return base_ctx[i_cat] + ((2 * i_nzb + i_nza) & 0x7f);
        else {
            i_nza &= 0x7f + (b_intra << 7);
            i_nzb &= 0x7f + (b_intra << 7);
            return base_ctx[i_cat] + 2 * !!i_nzb + !!i_nza;
        }
    }
}

extern const uint8_t x264_significant_coeff_flag_offset_8x8 [2] [64];
extern const uint8_t x264_last_coeff_flag_offset_8x8 [63];
extern const uint8_t x264_coeff_flag_offset_chroma_422_dc [7];
extern const uint16_t x264_significant_coeff_flag_offset [2] [16];
extern const uint16_t x264_last_coeff_flag_offset [2] [16];
extern const uint16_t x264_coeff_abs_level_m1_offset [16];
extern const uint8_t x264_count_cat_m1 [14];
const uint8_t coeff_abs_level1_ctx [8] = {1, 2, 3, 4, 0, 0, 0, 0};
const uint8_t coeff_abs_levelgt1_ctx [8] = {5, 5, 5, 5, 6, 7, 8, 9};
const uint8_t coeff_abs_levelgt1_ctx_chroma_dc [8] = {5, 5, 5, 5, 6, 7, 8, 8};
const uint8_t coeff_abs_level_transition [2] [8] = {{1, 2, 3, 3, 4, 5, 6, 7}, {4, 4, 4, 4, 5, 6, 7, 7}};

void x264_cabac_block_residual_internal (x264_t *h, x264_cabac_t *cb, int ctx_block_cat, dctcoef *l, int chroma422dc) {
    int ctx_sig = x264_significant_coeff_flag_offset[0][ctx_block_cat];
    int ctx_last = x264_last_coeff_flag_offset[0][ctx_block_cat];
    int ctx_level = x264_coeff_abs_level_m1_offset[ctx_block_cat];
    int coeff_idx = -1, node_ctx = 0;
    int last = h->quantf.coeff_last[ctx_block_cat] (l);
    const uint8_t *levelgt1_ctx = chroma422dc ? coeff_abs_levelgt1_ctx_chroma_dc : coeff_abs_levelgt1_ctx;
    dctcoef coeffs [64];
    if (chroma422dc) {
        int count_m1 = 7;
        {
            int i = 0;
            while (1) {
                if (l[i]) {
                    coeffs[++coeff_idx] = l[i];
                    x264_cabac_encode_decision_c (cb, ctx_sig + x264_coeff_flag_offset_chroma_422_dc [i], 1);
                    if (i == last) {
                        x264_cabac_encode_decision_c (cb, ctx_last + x264_coeff_flag_offset_chroma_422_dc [i], 1);
                        break;
                    }
                    else
                        x264_cabac_encode_decision_c (cb, ctx_last +x264_coeff_flag_offset_chroma_422_dc[i], 0);
                }
                else
                    x264_cabac_encode_decision_c (cb, ctx_sig +x264_coeff_flag_offset_chroma_422_dc[i], 0);
                if (++i == count_m1) {
                    coeffs[++coeff_idx] = l[i];
                    break;
                }
            }
        }
    }
    else {
        int count_m1 = x264_count_cat_m1[ctx_block_cat];
        if (count_m1 == 63) {
            const uint8_t *sig_offset = x264_significant_coeff_flag_offset_8x8[0];
            {
                int i = 0;
                while (1) {
                    if (l[i]) {
                        coeffs[++coeff_idx] = l[i];
                        x264_cabac_encode_decision_c (cb, ctx_sig + sig_offset [i], 1);
                        if (i == last) {
                            x264_cabac_encode_decision_c (cb, ctx_last + x264_last_coeff_flag_offset_8x8 [i], 1);
                            break;
                        }
                        else
                            x264_cabac_encode_decision_c (cb, ctx_last +x264_last_coeff_flag_offset_8x8[i], 0);
                    }
                    else
                        x264_cabac_encode_decision_c (cb, ctx_sig +sig_offset[i], 0);
                    if (++i == count_m1) {
                        coeffs[++coeff_idx] = l[i];
                        break;
                    }
                }
            }
        }
        else {
            int i = 0;
            while (1) {
                if (l[i]) {
                    coeffs[++coeff_idx] = l[i];
                    x264_cabac_encode_decision_c (cb, ctx_sig + i, 1);
                    if (i == last) {
                        x264_cabac_encode_decision_c (cb, ctx_last + i, 1);
                        break;
                    }
                    else
                        x264_cabac_encode_decision_c (cb, ctx_last +i, 0);
                }
                else
                    x264_cabac_encode_decision_c (cb, ctx_sig +i, 0);
                if (++i == count_m1) {
                    coeffs[++coeff_idx] = l[i];
                    break;
                }
            }
        }
    }
    do {
        int coeff = coeffs[coeff_idx];
        int abs_coeff = abs (coeff);
        int coeff_sign = coeff >> 31;
        int ctx = coeff_abs_level1_ctx[node_ctx] + ctx_level;
        if (abs_coeff > 1) {
            x264_cabac_encode_decision_c (cb, ctx, 1);
            ctx = levelgt1_ctx[node_ctx] + ctx_level;
            for (int i = ((abs_coeff) < (15) ? (abs_coeff) : (15)) - 2; i > 0; i--)
                x264_cabac_encode_decision_c (cb, ctx, 1);
            if (abs_coeff < 15)
                x264_cabac_encode_decision_c (cb, ctx, 0);
            else
                x264_cabac_encode_ue_bypass (cb, 0, abs_coeff -15);
            node_ctx = coeff_abs_level_transition[1][node_ctx];
        }
        else {
            x264_cabac_encode_decision_c (cb, ctx, 0);
            node_ctx = coeff_abs_level_transition[0][node_ctx];
        }
        x264_cabac_encode_bypass_c (cb, coeff_sign);
    }
    while (--coeff_idx >= 0);
}

void x264_cabac_block_residual_c (x264_t *h, x264_cabac_t *cb, int ctx_block_cat, dctcoef *l) {
    x264_cabac_block_residual_internal (h, cb, ctx_block_cat, l, 0);
}

void x264_cabac_block_residual (x264_t *h, x264_cabac_t *cb, int ctx_block_cat, dctcoef *l) {
    x264_cabac_block_residual_c (h, cb, ctx_block_cat, l);
}

void x264_cabac_block_residual_422_dc (x264_t *h, x264_cabac_t *cb, int ctx_block_cat, dctcoef *l) {
    x264_cabac_block_residual_internal (h, cb, DCT_CHROMA_DC, l, 1);
}

void x264_macroblock_write_cabac_internal (x264_t *h, x264_cabac_t *cb, int plane_count, int chroma) {
    const int i_mb_type = h->mb.i_type;
    const int i_mb_pos_start = x264_cabac_pos (cb);
    int i_mb_pos_tex;
    if (0 && (!(h->mb.i_mb_y & 1) || ((h->mb.type[h->mb.i_mb_xy - h->mb.i_mb_stride]) == P_SKIP || (h->mb.type[h->mb.i_mb_xy - h->mb.i_mb_stride]) == B_SKIP))) {
        x264_cabac_field_decoding_flag (h, cb);
    }
    if (h->sh.i_type == SLICE_TYPE_P)
        x264_cabac_mb_header_p (h, cb, i_mb_type, chroma);
    else if (h->sh.i_type == SLICE_TYPE_B)
        x264_cabac_mb_header_b (h, cb, i_mb_type, chroma);
    else
        x264_cabac_mb_header_i (h, cb, i_mb_type, SLICE_TYPE_I, chroma);
    i_mb_pos_tex = x264_cabac_pos (cb);
    h->stat.frame.i_mv_bits += i_mb_pos_tex - i_mb_pos_start;
    if (i_mb_type == I_PCM) {
        bs_t s;
        bs_init (& s, cb -> p, cb -> p_end - cb -> p);
        for (int p = 0; p < plane_count; p++)
            for (int i = 0; i < 256; i++)
                bs_write (&s, 8, h->mb.pic.p_fenc[p][i]);
        if (chroma)
            for (int ch = 1; ch < 3; ch++)
                for (int i = 0; i < 16 >> h->mb.chroma_v_shift; i++)
                    for (int j = 0; j < 8; j++)
                        bs_write (&s, 8, h->mb.pic.p_fenc[ch][i * 16 + j]);
        bs_flush (& s);
        cb->p = s.p;
        x264_cabac_encode_init_core (cb);
        h->stat.frame.i_tex_bits += x264_cabac_pos (cb) - i_mb_pos_tex;
        return;
    }
    if (i_mb_type != I_16x16) {
        x264_cabac_cbp_luma (h, cb);
        if (chroma)
            x264_cabac_cbp_chroma (h, cb);
    }
    if (x264_mb_transform_8x8_allowed (h) && h->mb.i_cbp_luma) {
        x264_cabac_transform_size (h, cb);
    }
    if (h->mb.i_cbp_luma || (chroma && h->mb.i_cbp_chroma) || i_mb_type == I_16x16) {
        const int b_intra = ((i_mb_type) == I_4x4 || (i_mb_type) == I_8x8 || (i_mb_type) == I_16x16 || (i_mb_type) == I_PCM);
        x264_cabac_qp_delta (h, cb);
        if (i_mb_type == I_16x16) {
            for (int p = 0; p < plane_count; p++) {
                do {
                    int ctxidxinc = x264_cabac_cbf_ctxidxinc (h, ctx_cat_plane[DCT_LUMA_DC][p], 48 + p, 1, 1);
                    if (h->mb.cache.non_zero_count[x264_scan8[48 + p]]) {
                        x264_cabac_encode_decision_c (cb, ctxidxinc, 1);
                        x264_cabac_block_residual (h, cb, ctx_cat_plane [DCT_LUMA_DC] [p], h -> dct.luma16x16_dc [p]);
                    }
                    else
                        x264_cabac_encode_decision_c (cb, ctxidxinc, 0);
                }
                while (0);
                if (h->mb.i_cbp_luma)
                    for (int i = p * 16; i < p * 16 + 16; i++)
                        do {
                            int ctxidxinc = x264_cabac_cbf_ctxidxinc (h, ctx_cat_plane[DCT_LUMA_AC][p], i, 1, 0);
                            if (h->mb.cache.non_zero_count[x264_scan8[i]]) {
                                x264_cabac_encode_decision_c (cb, ctxidxinc, 1);
                                x264_cabac_block_residual (h, cb, ctx_cat_plane [DCT_LUMA_AC] [p], h -> dct.luma4x4 [i] + 1);
                            }
                            else
                                x264_cabac_encode_decision_c (cb, ctxidxinc, 0);
                        }
                        while (0);
            }
        }
        else if (h->mb.b_transform_8x8) {
            if (plane_count == 3) {
                uint8_t nnzbak [3] [8];
                if ((h->mb.i_neighbour & MB_LEFT) && !h->mb.mb_transform_size[h->mb.i_mb_left_xy[0]]) {
                    nnzbak[0][0] = h->mb.cache.non_zero_count[x264_scan8[16 * 0 + 0] - 1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 0 + 0] - 1] = 0x80;
                    nnzbak[0][1] = h->mb.cache.non_zero_count[x264_scan8[16 * 0 + 2] - 1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 0 + 2] - 1] = 0x80;
                    nnzbak[1][0] = h->mb.cache.non_zero_count[x264_scan8[16 * 1 + 0] - 1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 1 + 0] - 1] = 0x80;
                    nnzbak[1][1] = h->mb.cache.non_zero_count[x264_scan8[16 * 1 + 2] - 1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 1 + 2] - 1] = 0x80;
                    nnzbak[2][0] = h->mb.cache.non_zero_count[x264_scan8[16 * 2 + 0] - 1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 2 + 0] - 1] = 0x80;
                    nnzbak[2][1] = h->mb.cache.non_zero_count[x264_scan8[16 * 2 + 2] - 1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 2 + 2] - 1] = 0x80;
                }
                if ((h->mb.i_neighbour & MB_LEFT) && !h->mb.mb_transform_size[h->mb.i_mb_left_xy[1]]) {
                    nnzbak[0][2] = h->mb.cache.non_zero_count[x264_scan8[16 * 0 + 8] - 1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 0 + 8] - 1] = 0x80;
                    nnzbak[0][3] = h->mb.cache.non_zero_count[x264_scan8[16 * 0 + 10] - 1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 0 + 10] - 1] = 0x80;
                    nnzbak[1][2] = h->mb.cache.non_zero_count[x264_scan8[16 * 1 + 8] - 1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 1 + 8] - 1] = 0x80;
                    nnzbak[1][3] = h->mb.cache.non_zero_count[x264_scan8[16 * 1 + 10] - 1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 1 + 10] - 1] = 0x80;
                    nnzbak[2][2] = h->mb.cache.non_zero_count[x264_scan8[16 * 2 + 8] - 1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 2 + 8] - 1] = 0x80;
                    nnzbak[2][3] = h->mb.cache.non_zero_count[x264_scan8[16 * 2 + 10] - 1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 2 + 10] - 1] = 0x80;
                }
                if ((h->mb.i_neighbour & MB_TOP) && !h->mb.mb_transform_size[h->mb.i_mb_top_xy]) {
                    (((x264_union32_t *) (&nnzbak[0][4]))->i) = (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * 0] - 8]))->i);
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * 0] - 8]))->i) = 0x80808080U;
                    (((x264_union32_t *) (&nnzbak[1][4]))->i) = (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * 1] - 8]))->i);
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * 1] - 8]))->i) = 0x80808080U;
                    (((x264_union32_t *) (&nnzbak[2][4]))->i) = (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * 2] - 8]))->i);
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * 2] - 8]))->i) = 0x80808080U;
                }
                for (int p = 0; p < 3; p++)
                    for (int i = 0, msk = h->mb.i_cbp_luma, skip; msk && (skip = x264_ctz_4bit (msk), i += skip, msk >>= skip + 1, 1); i++)
                        do {
                            int ctxidxinc = x264_cabac_cbf_ctxidxinc (h, ctx_cat_plane[DCT_LUMA_8x8][p], i *4 + p * 16, b_intra, 0);
                            if (h->mb.cache.non_zero_count[x264_scan8[i * 4 + p * 16]]) {
                                x264_cabac_encode_decision_c (cb, ctxidxinc, 1);
                                x264_cabac_block_residual (h, cb, ctx_cat_plane [DCT_LUMA_8x8] [p], h -> dct.luma8x8 [i + p * 4]);
                            }
                            else
                                x264_cabac_encode_decision_c (cb, ctxidxinc, 0);
                        }
                        while (0);
                if ((h->mb.i_neighbour & MB_LEFT) && !h->mb.mb_transform_size[h->mb.i_mb_left_xy[0]]) {
                    h->mb.cache.non_zero_count[x264_scan8[16 * 0 + 0] - 1] = nnzbak[0][0];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 0 + 2] - 1] = nnzbak[0][1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 1 + 0] - 1] = nnzbak[1][0];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 1 + 2] - 1] = nnzbak[1][1];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 2 + 0] - 1] = nnzbak[2][0];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 2 + 2] - 1] = nnzbak[2][1];
                }
                if ((h->mb.i_neighbour & MB_LEFT) && !h->mb.mb_transform_size[h->mb.i_mb_left_xy[1]]) {
                    h->mb.cache.non_zero_count[x264_scan8[16 * 0 + 8] - 1] = nnzbak[0][2];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 0 + 10] - 1] = nnzbak[0][3];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 1 + 8] - 1] = nnzbak[1][2];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 1 + 10] - 1] = nnzbak[1][3];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 2 + 8] - 1] = nnzbak[2][2];
                    h->mb.cache.non_zero_count[x264_scan8[16 * 2 + 10] - 1] = nnzbak[2][3];
                }
                if ((h->mb.i_neighbour & MB_TOP) && !h->mb.mb_transform_size[h->mb.i_mb_top_xy]) {
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * 0] - 8]))->i) = (((x264_union32_t *) (&nnzbak[0][4]))->i);
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * 1] - 8]))->i) = (((x264_union32_t *) (&nnzbak[1][4]))->i);
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * 2] - 8]))->i) = (((x264_union32_t *) (&nnzbak[2][4]))->i);
                }
            }
            else {
                for (int i = 0, msk = h->mb.i_cbp_luma, skip; msk && (skip = x264_ctz_4bit (msk), i += skip, msk >>= skip + 1, 1); i++)
                    x264_cabac_block_residual (h, cb, DCT_LUMA_8x8, h->dct.luma8x8[i]);
            }
        }
        else {
            for (int p = 0; p < plane_count; p++)
                for (int i8x8 = 0, msk = h->mb.i_cbp_luma, skip; msk && (skip = x264_ctz_4bit (msk), i8x8 += skip, msk >>= skip + 1, 1); i8x8++)
                    for (int i = 0; i < 4; i++)
                        do {
                            int ctxidxinc = x264_cabac_cbf_ctxidxinc (h, ctx_cat_plane[DCT_LUMA_4x4][p], i +i8x8 * 4 + p * 16, b_intra, 0);
                            if (h->mb.cache.non_zero_count[x264_scan8[i + i8x8 * 4 + p * 16]]) {
                                x264_cabac_encode_decision_c (cb, ctxidxinc, 1);
                                x264_cabac_block_residual (h, cb, ctx_cat_plane [DCT_LUMA_4x4] [p], h -> dct.luma4x4 [i + i8x8 * 4 + p * 16]);
                            }
                            else
                                x264_cabac_encode_decision_c (cb, ctxidxinc, 0);
                        }
                        while (0);
        }
        if (chroma && h->mb.i_cbp_chroma) {
            if (h->sps->i_chroma_format_idc == CHROMA_422) {
                do {
                    int ctxidxinc = x264_cabac_cbf_ctxidxinc (h, DCT_CHROMA_DC, 49 + (0), b_intra, 1);
                    if (h->mb.cache.non_zero_count[x264_scan8[49 + (0)]]) {
                        x264_cabac_encode_decision_c (cb, ctxidxinc, 1);
                        x264_cabac_block_residual_422_dc (h, cb, DCT_CHROMA_DC, h -> dct.chroma_dc [0]);
                    }
                    else
                        x264_cabac_encode_decision_c (cb, ctxidxinc, 0);
                }
                while (0);
                do {
                    int ctxidxinc = x264_cabac_cbf_ctxidxinc (h, DCT_CHROMA_DC, 49 + (1), b_intra, 1);
                    if (h->mb.cache.non_zero_count[x264_scan8[49 + (1)]]) {
                        x264_cabac_encode_decision_c (cb, ctxidxinc, 1);
                        x264_cabac_block_residual_422_dc (h, cb, DCT_CHROMA_DC, h -> dct.chroma_dc [1]);
                    }
                    else
                        x264_cabac_encode_decision_c (cb, ctxidxinc, 0);
                }
                while (0);
            }
            else {
                do {
                    int ctxidxinc = x264_cabac_cbf_ctxidxinc (h, DCT_CHROMA_DC, 49 + 0, b_intra, 1);
                    if (h->mb.cache.non_zero_count[x264_scan8[49 + 0]]) {
                        x264_cabac_encode_decision_c (cb, ctxidxinc, 1);
                        x264_cabac_block_residual (h, cb, DCT_CHROMA_DC, h -> dct.chroma_dc [0]);
                    }
                    else
                        x264_cabac_encode_decision_c (cb, ctxidxinc, 0);
                }
                while (0);
                do {
                    int ctxidxinc = x264_cabac_cbf_ctxidxinc (h, DCT_CHROMA_DC, 49 + 1, b_intra, 1);
                    if (h->mb.cache.non_zero_count[x264_scan8[49 + 1]]) {
                        x264_cabac_encode_decision_c (cb, ctxidxinc, 1);
                        x264_cabac_block_residual (h, cb, DCT_CHROMA_DC, h -> dct.chroma_dc [1]);
                    }
                    else
                        x264_cabac_encode_decision_c (cb, ctxidxinc, 0);
                }
                while (0);
            }
            if (h->mb.i_cbp_chroma == 2) {
                int step = 8 << h->mb.chroma_v_shift;
                for (int i = 16; i < 3 * 16; i += step)
                    for (int j = i; j < i + 4; j++)
                        do {
                            int ctxidxinc = x264_cabac_cbf_ctxidxinc (h, DCT_CHROMA_AC, j, b_intra, 0);
                            if (h->mb.cache.non_zero_count[x264_scan8[j]]) {
                                x264_cabac_encode_decision_c (cb, ctxidxinc, 1);
                                x264_cabac_block_residual (h, cb, DCT_CHROMA_AC, h -> dct.luma4x4 [j] + 1);
                            }
                            else
                                x264_cabac_encode_decision_c (cb, ctxidxinc, 0);
                        }
                        while (0);
            }
        }
    }
    h->stat.frame.i_tex_bits += x264_cabac_pos (cb) - i_mb_pos_tex;
}

void x264_macroblock_write_cabac (x264_t *h, x264_cabac_t *cb) {
    if ((h->sps->i_chroma_format_idc == CHROMA_444))
        x264_macroblock_write_cabac_internal (h, cb, 3, 0);
    else
        x264_macroblock_write_cabac_internal (h, cb, 1, 1);
}

