
#include "common.h"

#include <stdarg.h>

#include <ctype.h>

#include <sys/mman.h>

#include <malloc.h>
const int x264_bit_depth = BIT_DEPTH;
const int x264_chroma_format = X264_CHROMA_FORMAT;
void x264_log_default (void *, int, const char *, va_list);

void x264_param_default (x264_param_t *param) {
    memset (param, 0, sizeof (x264_param_t));
    param->cpu = x264_cpu_detect ();
    param->i_threads = 0;
    param->i_lookahead_threads = 0;
    param->b_deterministic = 1;
    param->i_sync_lookahead = (-1);
    param->i_csp = 0 ? 0 : 0x0001;
    param->i_width = 0;
    param->i_height = 0;
    param->vui.i_sar_width = 0;
    param->vui.i_sar_height = 0;
    param->vui.i_overscan = 0;
    param->vui.i_vidformat = 5;
    param->vui.b_fullrange = -1;
    param->vui.i_colorprim = 2;
    param->vui.i_transfer = 2;
    param->vui.i_colmatrix = -1;
    param->vui.i_chroma_loc = 0;
    param->i_fps_num = 25;
    param->i_fps_den = 1;
    param->i_level_idc = -1;
    param->i_slice_max_size = 0;
    param->i_slice_max_mbs = 0;
    param->i_slice_count = 0;
    param->i_frame_reference = 3;
    param->i_keyint_max = 250;
    param->i_keyint_min = 0;
    param->i_bframe = 3;
    param->i_scenecut_threshold = 40;
    param->i_bframe_adaptive = 1;
    param->i_bframe_bias = 0;
    param->i_bframe_pyramid = 2;
    param->b_interlaced = 0;
    param->b_constrained_intra = 0;
    param->b_deblocking_filter = 1;
    param->i_deblocking_filter_alphac0 = 0;
    param->i_deblocking_filter_beta = 0;
    param->b_cabac = 1;
    param->i_cabac_init_idc = 0;
    param->rc.i_rc_method = 1;
    param->rc.i_bitrate = 0;
    param->rc.f_rate_tolerance = 1.0;
    param->rc.i_vbv_max_bitrate = 0;
    param->rc.i_vbv_buffer_size = 0;
    param->rc.f_vbv_buffer_init = 0.9;
    param->rc.i_qp_constant = 23 + (6 * (8 - 8));
    param->rc.f_rf_constant = 23;
    param->rc.i_qp_min = 0;
    param->rc.i_qp_max = ((51 + (6 * (8 - 8))) + 18);
    param->rc.i_qp_step = 4;
    param->rc.f_ip_factor = 1.4;
    param->rc.f_pb_factor = 1.3;
    param->rc.i_aq_mode = 1;
    param->rc.f_aq_strength = 1.0;
    param->rc.i_lookahead = 40;
    param->rc.b_stat_write = 0;
    param->rc.psz_stat_out = "x264_2pass.log";
    param->rc.b_stat_read = 0;
    param->rc.psz_stat_in = "x264_2pass.log";
    param->rc.f_qcompress = 0.6;
    param->rc.f_qblur = 0.5;
    param->rc.f_complexity_blur = 20;
    param->rc.i_zones = 0;
    param->rc.b_mb_tree = 1;
    param->pf_log = x264_log_default;
    param->p_log_private = ((void *) 0);
    param->i_log_level = 2;
    param->analyse.intra = 0x0001 | 0x0002;
    param->analyse.inter = 0x0001 | 0x0002 | 0x0010 | 0x0100;
    param->analyse.i_direct_mv_pred = 1;
    param->analyse.i_me_method = 1;
    param->analyse.f_psy_rd = 1.0;
    param->analyse.b_psy = 1;
    param->analyse.f_psy_trellis = 0;
    param->analyse.i_me_range = 16;
    param->analyse.i_subpel_refine = 7;
    param->analyse.b_mixed_references = 1;
    param->analyse.b_chroma_me = 1;
    param->analyse.i_mv_range_thread = -1;
    param->analyse.i_mv_range = -1;
    param->analyse.i_chroma_qp_offset = 0;
    param->analyse.b_fast_pskip = 1;
    param->analyse.b_weighted_bipred = 1;
    param->analyse.i_weighted_pred = 2;
    param->analyse.b_dct_decimate = 1;
    param->analyse.b_transform_8x8 = 1;
    param->analyse.i_trellis = 1;
    param->analyse.i_luma_deadzone[0] = 21;
    param->analyse.i_luma_deadzone[1] = 11;
    param->analyse.b_psnr = 0;
    param->analyse.b_ssim = 0;
    param->i_cqm_preset = 0;
    memset (param -> cqm_4iy, 16, sizeof (param -> cqm_4iy));
    memset (param -> cqm_4py, 16, sizeof (param -> cqm_4py));
    memset (param -> cqm_4ic, 16, sizeof (param -> cqm_4ic));
    memset (param -> cqm_4pc, 16, sizeof (param -> cqm_4pc));
    memset (param -> cqm_8iy, 16, sizeof (param -> cqm_8iy));
    memset (param -> cqm_8py, 16, sizeof (param -> cqm_8py));
    memset (param -> cqm_8ic, 16, sizeof (param -> cqm_8ic));
    memset (param -> cqm_8pc, 16, sizeof (param -> cqm_8pc));
    param->b_repeat_headers = 1;
    param->b_annexb = 1;
    param->b_aud = 0;
    param->b_vfr_input = 1;
    param->i_nal_hrd = 0;
    param->b_tff = 1;
    param->b_pic_struct = 0;
    param->b_fake_interlaced = 0;
    param->i_frame_packing = -1;
    param->b_opencl = 0;
    param->i_opencl_device = 0;
    param->opencl_device_id = ((void *) 0);
    param->psz_clbin_file = ((void *) 0);
}

int x264_param_apply_preset (x264_param_t *param, const char *preset) {
    char *end;
    int i = strtol (preset, &end, 10);
    if (*end == 0 && i >= 0 && i < sizeof (x264_preset_names) / sizeof (*x264_preset_names) - 1)
        preset = x264_preset_names[i];
    if (!strcasecmp (preset, "ultrafast")) {
        param->i_frame_reference = 1;
        param->i_scenecut_threshold = 0;
        param->b_deblocking_filter = 0;
        param->b_cabac = 0;
        param->i_bframe = 0;
        param->analyse.intra = 0;
        param->analyse.inter = 0;
        param->analyse.b_transform_8x8 = 0;
        param->analyse.i_me_method = 0;
        param->analyse.i_subpel_refine = 0;
        param->rc.i_aq_mode = 0;
        param->analyse.b_mixed_references = 0;
        param->analyse.i_trellis = 0;
        param->i_bframe_adaptive = 0;
        param->rc.b_mb_tree = 0;
        param->analyse.i_weighted_pred = 0;
        param->analyse.b_weighted_bipred = 0;
        param->rc.i_lookahead = 0;
    }
    else if (!strcasecmp (preset, "superfast")) {
        param->analyse.inter = 0x0002 | 0x0001;
        param->analyse.i_me_method = 0;
        param->analyse.i_subpel_refine = 1;
        param->i_frame_reference = 1;
        param->analyse.b_mixed_references = 0;
        param->analyse.i_trellis = 0;
        param->rc.b_mb_tree = 0;
        param->analyse.i_weighted_pred = 1;
        param->rc.i_lookahead = 0;
    }
    else if (!strcasecmp (preset, "veryfast")) {
        param->analyse.i_me_method = 1;
        param->analyse.i_subpel_refine = 2;
        param->i_frame_reference = 1;
        param->analyse.b_mixed_references = 0;
        param->analyse.i_trellis = 0;
        param->analyse.i_weighted_pred = 1;
        param->rc.i_lookahead = 10;
    }
    else if (!strcasecmp (preset, "faster")) {
        param->analyse.b_mixed_references = 0;
        param->i_frame_reference = 2;
        param->analyse.i_subpel_refine = 4;
        param->analyse.i_weighted_pred = 1;
        param->rc.i_lookahead = 20;
    }
    else if (!strcasecmp (preset, "fast")) {
        param->i_frame_reference = 2;
        param->analyse.i_subpel_refine = 6;
        param->analyse.i_weighted_pred = 1;
        param->rc.i_lookahead = 30;
    }
    else if (!strcasecmp (preset, "medium")) {
    }
    else if (!strcasecmp (preset, "slow")) {
        param->analyse.i_me_method = 2;
        param->analyse.i_subpel_refine = 8;
        param->i_frame_reference = 5;
        param->i_bframe_adaptive = 2;
        param->analyse.i_direct_mv_pred = 3;
        param->rc.i_lookahead = 50;
    }
    else if (!strcasecmp (preset, "slower")) {
        param->analyse.i_me_method = 2;
        param->analyse.i_subpel_refine = 9;
        param->i_frame_reference = 8;
        param->i_bframe_adaptive = 2;
        param->analyse.i_direct_mv_pred = 3;
        param->analyse.inter |= 0x0020;
        param->analyse.i_trellis = 2;
        param->rc.i_lookahead = 60;
    }
    else if (!strcasecmp (preset, "veryslow")) {
        param->analyse.i_me_method = 2;
        param->analyse.i_subpel_refine = 10;
        param->analyse.i_me_range = 24;
        param->i_frame_reference = 16;
        param->i_bframe_adaptive = 2;
        param->analyse.i_direct_mv_pred = 3;
        param->analyse.inter |= 0x0020;
        param->analyse.i_trellis = 2;
        param->i_bframe = 8;
        param->rc.i_lookahead = 60;
    }
    else if (!strcasecmp (preset, "placebo")) {
        param->analyse.i_me_method = 4;
        param->analyse.i_subpel_refine = 11;
        param->analyse.i_me_range = 24;
        param->i_frame_reference = 16;
        param->i_bframe_adaptive = 2;
        param->analyse.i_direct_mv_pred = 3;
        param->analyse.inter |= 0x0020;
        param->analyse.b_fast_pskip = 0;
        param->analyse.i_trellis = 2;
        param->i_bframe = 16;
        param->rc.i_lookahead = 60;
    }
    else {
        x264_log (((void *) 0), 0, "invalid preset '%s'\n", preset);
        return -1;
    }
    return 0;
}

int x264_param_apply_tune (x264_param_t *param, const char *tune) {
    char *tmp = x264_malloc (strlen (tune) +1);
    if (!tmp)
        return -1;
    tmp = strcpy (tmp, tune);
    char *s = strtok (tmp, ",./-+");
    int psy_tuning_used = 0;
    while (s) {
        if (!strncasecmp (s, "film", 4)) {
            if (psy_tuning_used++)
                goto psy_failure;
            param->i_deblocking_filter_alphac0 = -1;
            param->i_deblocking_filter_beta = -1;
            param->analyse.f_psy_trellis = 0.15;
        }
        else if (!strncasecmp (s, "animation", 9)) {
            if (psy_tuning_used++)
                goto psy_failure;
            param->i_frame_reference = param->i_frame_reference > 1 ? param->i_frame_reference * 2 : 1;
            param->i_deblocking_filter_alphac0 = 1;
            param->i_deblocking_filter_beta = 1;
            param->analyse.f_psy_rd = 0.4;
            param->rc.f_aq_strength = 0.6;
            param->i_bframe += 2;
        }
        else if (!strncasecmp (s, "grain", 5)) {
            if (psy_tuning_used++)
                goto psy_failure;
            param->i_deblocking_filter_alphac0 = -2;
            param->i_deblocking_filter_beta = -2;
            param->analyse.f_psy_trellis = 0.25;
            param->analyse.b_dct_decimate = 0;
            param->rc.f_pb_factor = 1.1;
            param->rc.f_ip_factor = 1.1;
            param->rc.f_aq_strength = 0.5;
            param->analyse.i_luma_deadzone[0] = 6;
            param->analyse.i_luma_deadzone[1] = 6;
            param->rc.f_qcompress = 0.8;
        }
        else if (!strncasecmp (s, "stillimage", 10)) {
            if (psy_tuning_used++)
                goto psy_failure;
            param->i_deblocking_filter_alphac0 = -3;
            param->i_deblocking_filter_beta = -3;
            param->analyse.f_psy_rd = 2.0;
            param->analyse.f_psy_trellis = 0.7;
            param->rc.f_aq_strength = 1.2;
        }
        else if (!strncasecmp (s, "psnr", 4)) {
            if (psy_tuning_used++)
                goto psy_failure;
            param->rc.i_aq_mode = 0;
            param->analyse.b_psy = 0;
        }
        else if (!strncasecmp (s, "ssim", 4)) {
            if (psy_tuning_used++)
                goto psy_failure;
            param->rc.i_aq_mode = 2;
            param->analyse.b_psy = 0;
        }
        else if (!strncasecmp (s, "fastdecode", 10)) {
            param->b_deblocking_filter = 0;
            param->b_cabac = 0;
            param->analyse.b_weighted_bipred = 0;
            param->analyse.i_weighted_pred = 0;
        }
        else if (!strncasecmp (s, "zerolatency", 11)) {
            param->rc.i_lookahead = 0;
            param->i_sync_lookahead = 0;
            param->i_bframe = 0;
            param->b_sliced_threads = 1;
            param->b_vfr_input = 0;
            param->rc.b_mb_tree = 0;
        }
        else if (!strncasecmp (s, "touhou", 6)) {
            if (psy_tuning_used++)
                goto psy_failure;
            param->i_frame_reference = param->i_frame_reference > 1 ? param->i_frame_reference * 2 : 1;
            param->i_deblocking_filter_alphac0 = -1;
            param->i_deblocking_filter_beta = -1;
            param->analyse.f_psy_trellis = 0.2;
            param->rc.f_aq_strength = 1.3;
            if (param->analyse.inter & 0x0010)
                param->analyse.inter |= 0x0020;
        }
        else {
            x264_log (((void *) 0), 0, "invalid tune '%s'\n", s);
            x264_free (tmp);
            return -1;
        }
        if (0) {
        psy_failure :
            x264_log (((void *) 0), 1, "only 1 psy tuning can be used: ignoring tune %s\n", s);
        }
        s = strtok (((void *) 0), ",./-+");
    }
    x264_free (tmp);
    return 0;
}

int x264_param_default_preset (x264_param_t *param, const char *preset, const char *tune) {
    x264_param_default (param);
    if (preset && x264_param_apply_preset (param, preset) < 0)
        return -1;
    if (tune && x264_param_apply_tune (param, tune) < 0)
        return -1;
    return 0;
}

void x264_param_apply_fastfirstpass (x264_param_t *param) {
    if (param->rc.b_stat_write && !param->rc.b_stat_read) {
        param->i_frame_reference = 1;
        param->analyse.b_transform_8x8 = 0;
        param->analyse.inter = 0;
        param->analyse.i_me_method = 0;
        param->analyse.i_subpel_refine = ((2) < (param->analyse.i_subpel_refine) ? (2) : (param->analyse.i_subpel_refine));
        param->analyse.i_trellis = 0;
        param->analyse.b_fast_pskip = 1;
    }
}

int profile_string_to_int (const char *str) {
    if (!strcasecmp (str, "baseline"))
        return PROFILE_BASELINE;
    if (!strcasecmp (str, "main"))
        return PROFILE_MAIN;
    if (!strcasecmp (str, "high"))
        return PROFILE_HIGH;
    if (!strcasecmp (str, "high10"))
        return PROFILE_HIGH10;
    if (!strcasecmp (str, "high422"))
        return PROFILE_HIGH422;
    if (!strcasecmp (str, "high444"))
        return PROFILE_HIGH444_PREDICTIVE;
    return -1;
}

int x264_param_apply_profile (x264_param_t *param, const char *profile) {
    if (!profile)
        return 0;
    int p = profile_string_to_int (profile);
    if (p < 0) {
        x264_log (((void *) 0), 0, "invalid profile: %s\n", profile);
        return -1;
    }
    if (p < PROFILE_HIGH444_PREDICTIVE && ((param->rc.i_rc_method == 0 && param->rc.i_qp_constant <= 0) || (param->rc.i_rc_method == 1 && (int) (param->rc.f_rf_constant + (6 * (8 - 8))) <= 0))) {
        x264_log (((void *) 0), 0, "%s profile doesn't support lossless\n", profile);
        return -1;
    }
    if (p < PROFILE_HIGH444_PREDICTIVE && (param->i_csp & 0x00ff) >= 0x0008) {
        x264_log (((void *) 0), 0, "%s profile doesn't support 4:4:4\n", profile);
        return -1;
    }
    if (p < PROFILE_HIGH422 && (param->i_csp & 0x00ff) >= 0x0004) {
        x264_log (((void *) 0), 0, "%s profile doesn't support 4:2:2\n", profile);
        return -1;
    }
    if (p < PROFILE_HIGH10 && 8 > 8) {
        x264_log (((void *) 0), 0, "%s profile doesn't support a bit depth of %d\n", profile, 8);
        return -1;
    }
    if (p == PROFILE_BASELINE) {
        param->analyse.b_transform_8x8 = 0;
        param->b_cabac = 0;
        param->i_cqm_preset = 0;
        param->psz_cqm_file = ((void *) 0);
        param->i_bframe = 0;
        param->analyse.i_weighted_pred = 0;
        if (param->b_interlaced) {
            x264_log (((void *) 0), 0, "baseline profile doesn't support interlacing\n");
            return -1;
        }
        if (param->b_fake_interlaced) {
            x264_log (((void *) 0), 0, "baseline profile doesn't support fake interlacing\n");
            return -1;
        }
    }
    else if (p == PROFILE_MAIN) {
        param->analyse.b_transform_8x8 = 0;
        param->i_cqm_preset = 0;
        param->psz_cqm_file = ((void *) 0);
    }
    return 0;
}

int parse_enum (const char *arg, const char * const *names, int *dst) {
    for (int i = 0; names[i]; i++)
        if (!strcmp (arg, names[i])) {
            *dst = i;
            return 0;
        }
    return -1;
}

int parse_cqm (const char *str, uint8_t *cqm, int length) {
    int i = 0;
    do {
        int coef;
        if (!sscanf (str, "%d", &coef) || coef < 1 || coef > 255)
            return -1;
        cqm[i++] = coef;
    }
    while (i < length && (str = strchr (str, ',')) && str++);
    return (i == length) ? 0 : -1;
}

int x264_atobool (const char *str, int *b_error) {
    if (!strcmp (str, "1") || !strcmp (str, "true") || !strcmp (str, "yes"))
        return 1;
    if (!strcmp (str, "0") || !strcmp (str, "false") || !strcmp (str, "no"))
        return 0;
    *b_error = 1;
    return 0;
}

int x264_atoi (const char *str, int *b_error) {
    char *end;
    int v = strtol (str, &end, 0);
    if (end == str || *end != '\0')
        *b_error = 1;
    return v;
}

double x264_atof (const char *str, int *b_error) {
    char *end;
    double v = strtod (str, &end);
    if (end == str || *end != '\0')
        *b_error = 1;
    return v;
}

int x264_param_parse (x264_param_t *p, const char *name, const char *value) {
    char *name_buf = ((void *) 0);
    int b_error = 0;
    int name_was_bool;
    int value_was_null = !value;
    int i;
    if (!name)
        return (-1);
    if (!value)
        value = "true";
    if (value[0] == '=')
        value++;
    if (strchr (name, '_')) {
        char *c;
        name_buf = strdup (name);
        while ((c = strchr (name_buf, '_')))
            *c = '-';
        name = name_buf;
    }
    if ((!strncmp (name, "no-", 3) && (i = 3)) || (!strncmp (name, "no", 2) && (i = 2))) {
        name += i;
        value = (name_was_bool = 1, x264_atobool (value, &b_error)) ? "false" : "true";
    }
    name_was_bool = 0;
    if (0)
        ;
    else if (!strcmp (name, "asm")) {
        p->cpu = ((*__ctype_b_loc ())[(int) ((value[0]))] & (unsigned  short  int) _ISdigit) ? x264_atoi (value, &b_error) : !strcmp (value, "auto") || (name_was_bool = 1, x264_atobool (value, &b_error)) ? x264_cpu_detect () : 0;
        if (b_error) {
            char *buf = strdup (value);
            char *tok, *saveptr = ((void *) 0), *init;
            b_error = 0;
            p->cpu = 0;
            for (init = buf; (tok = strtok_r (init, ",", &saveptr)); init = ((void *) 0)) {
                for (i = 0; x264_cpu_names[i].flags && strcasecmp (tok, x264_cpu_names[i].name); i++)
                    ;
                p->cpu |= x264_cpu_names[i].flags;
                if (!x264_cpu_names[i].flags)
                    b_error = 1;
            }
            free (buf);
            if ((p->cpu & 0x0000040) && !(p->cpu & 0x0080000))
                p->cpu |= 0x0100000;
        }
    }
    else if (!strcmp (name, "threads")) {
        if (!strcmp (value, "auto"))
            p->i_threads = 0;
        else
            p->i_threads = x264_atoi (value, &b_error);
    }
    else if (!strcmp (name, "lookahead-threads")) {
        if (!strcmp (value, "auto"))
            p->i_lookahead_threads = 0;
        else
            p->i_lookahead_threads = x264_atoi (value, &b_error);
    }
    else if (!strcmp (name, "sliced-threads"))
        p->b_sliced_threads = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "sync-lookahead")) {
        if (!strcmp (value, "auto"))
            p->i_sync_lookahead = (-1);
        else
            p->i_sync_lookahead = x264_atoi (value, &b_error);
    }
    else if (!strcmp (name, "deterministic") || !strcmp (name, "n-deterministic"))
        p->b_deterministic = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "cpu-independent"))
        p->b_cpu_independent = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "level") || !strcmp (name, "level-idc")) {
        if (!strcmp (value, "1b"))
            p->i_level_idc = 9;
        else if (x264_atof (value, &b_error) < 6)
            p->i_level_idc = (int) (10 * x264_atof (value, &b_error) + .5);
        else
            p->i_level_idc = x264_atoi (value, &b_error);
    }
    else if (!strcmp (name, "bluray-compat"))
        p->b_bluray_compat = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "avcintra-class"))
        p->i_avcintra_class = x264_atoi (value, &b_error);
    else if (!strcmp (name, "sar")) {
        b_error = (2 != sscanf (value, "%d:%d", &p->vui.i_sar_width, &p->vui.i_sar_height) && 2 != sscanf (value, "%d/%d", &p->vui.i_sar_width, &p->vui.i_sar_height));
    }
    else if (!strcmp (name, "overscan"))
        b_error |= parse_enum (value, x264_overscan_names, &p->vui.i_overscan);
    else if (!strcmp (name, "videoformat"))
        b_error |= parse_enum (value, x264_vidformat_names, &p->vui.i_vidformat);
    else if (!strcmp (name, "fullrange"))
        b_error |= parse_enum (value, x264_fullrange_names, &p->vui.b_fullrange);
    else if (!strcmp (name, "colorprim"))
        b_error |= parse_enum (value, x264_colorprim_names, &p->vui.i_colorprim);
    else if (!strcmp (name, "transfer"))
        b_error |= parse_enum (value, x264_transfer_names, &p->vui.i_transfer);
    else if (!strcmp (name, "colormatrix"))
        b_error |= parse_enum (value, x264_colmatrix_names, &p->vui.i_colmatrix);
    else if (!strcmp (name, "chromaloc")) {
        p->vui.i_chroma_loc = x264_atoi (value, &b_error);
        b_error = (p->vui.i_chroma_loc < 0 || p->vui.i_chroma_loc > 5);
    }
    else if (!strcmp (name, "fps")) {
        if (sscanf (value, "%u/%u", &p->i_fps_num, &p->i_fps_den) == 2)
            ;
        else {
            float fps = x264_atof (value, &b_error);
            if (fps > 0 && fps <= 2147483647 / 1000) {
                p->i_fps_num = (int) (fps * 1000 + .5);
                p->i_fps_den = 1000;
            }
            else {
                p->i_fps_num = x264_atoi (value, &b_error);
                p->i_fps_den = 1;
            }
        }
    }
    else if (!strcmp (name, "ref") || !strcmp (name, "frameref"))
        p->i_frame_reference = x264_atoi (value, &b_error);
    else if (!strcmp (name, "dpb-size"))
        p->i_dpb_size = x264_atoi (value, &b_error);
    else if (!strcmp (name, "keyint")) {
        if (strstr (value, "infinite"))
            p->i_keyint_max = (1 << 30);
        else
            p->i_keyint_max = x264_atoi (value, &b_error);
    }
    else if (!strcmp (name, "min-keyint") || !strcmp (name, "keyint-min")) {
        p->i_keyint_min = x264_atoi (value, &b_error);
        if (p->i_keyint_max < p->i_keyint_min)
            p->i_keyint_max = p->i_keyint_min;
    }
    else if (!strcmp (name, "scenecut")) {
        p->i_scenecut_threshold = (name_was_bool = 1, x264_atobool (value, &b_error));
        if (b_error || p->i_scenecut_threshold) {
            b_error = 0;
            p->i_scenecut_threshold = x264_atoi (value, &b_error);
        }
    }
    else if (!strcmp (name, "intra-refresh"))
        p->b_intra_refresh = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "bframes"))
        p->i_bframe = x264_atoi (value, &b_error);
    else if (!strcmp (name, "b-adapt")) {
        p->i_bframe_adaptive = (name_was_bool = 1, x264_atobool (value, &b_error));
        if (b_error) {
            b_error = 0;
            p->i_bframe_adaptive = x264_atoi (value, &b_error);
        }
    }
    else if (!strcmp (name, "b-bias"))
        p->i_bframe_bias = x264_atoi (value, &b_error);
    else if (!strcmp (name, "b-pyramid")) {
        b_error |= parse_enum (value, x264_b_pyramid_names, &p->i_bframe_pyramid);
        if (b_error) {
            b_error = 0;
            p->i_bframe_pyramid = x264_atoi (value, &b_error);
        }
    }
    else if (!strcmp (name, "open-gop"))
        p->b_open_gop = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "nf"))
        p->b_deblocking_filter = !(name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "filter") || !strcmp (name, "deblock")) {
        if (2 == sscanf (value, "%d:%d", &p->i_deblocking_filter_alphac0, &p->i_deblocking_filter_beta) || 2 == sscanf (value, "%d,%d", &p->i_deblocking_filter_alphac0, &p->i_deblocking_filter_beta)) {
            p->b_deblocking_filter = 1;
        }
        else if (sscanf (value, "%d", &p->i_deblocking_filter_alphac0)) {
            p->b_deblocking_filter = 1;
            p->i_deblocking_filter_beta = p->i_deblocking_filter_alphac0;
        }
        else
            p->b_deblocking_filter = (name_was_bool = 1, x264_atobool (value, &b_error));
    }
    else if (!strcmp (name, "slice-max-size"))
        p->i_slice_max_size = x264_atoi (value, &b_error);
    else if (!strcmp (name, "slice-max-mbs"))
        p->i_slice_max_mbs = x264_atoi (value, &b_error);
    else if (!strcmp (name, "slice-min-mbs"))
        p->i_slice_min_mbs = x264_atoi (value, &b_error);
    else if (!strcmp (name, "slices"))
        p->i_slice_count = x264_atoi (value, &b_error);
    else if (!strcmp (name, "slices-max"))
        p->i_slice_count_max = x264_atoi (value, &b_error);
    else if (!strcmp (name, "cabac"))
        p->b_cabac = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "cabac-idc"))
        p->i_cabac_init_idc = x264_atoi (value, &b_error);
    else if (!strcmp (name, "interlaced"))
        p->b_interlaced = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "tff"))
        p->b_interlaced = p->b_tff = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "bff")) {
        p->b_interlaced = (name_was_bool = 1, x264_atobool (value, &b_error));
        p->b_tff = !p->b_interlaced;
    }
    else if (!strcmp (name, "constrained-intra"))
        p->b_constrained_intra = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "cqm")) {
        if (strstr (value, "flat"))
            p->i_cqm_preset = 0;
        else if (strstr (value, "jvt"))
            p->i_cqm_preset = 1;
        else
            p->psz_cqm_file = strdup (value);
    }
    else if (!strcmp (name, "cqmfile"))
        p->psz_cqm_file = strdup (value);
    else if (!strcmp (name, "cqm4")) {
        p->i_cqm_preset = 2;
        b_error |= parse_cqm (value, p->cqm_4iy, 16);
        b_error |= parse_cqm (value, p->cqm_4py, 16);
        b_error |= parse_cqm (value, p->cqm_4ic, 16);
        b_error |= parse_cqm (value, p->cqm_4pc, 16);
    }
    else if (!strcmp (name, "cqm8")) {
        p->i_cqm_preset = 2;
        b_error |= parse_cqm (value, p->cqm_8iy, 64);
        b_error |= parse_cqm (value, p->cqm_8py, 64);
        b_error |= parse_cqm (value, p->cqm_8ic, 64);
        b_error |= parse_cqm (value, p->cqm_8pc, 64);
    }
    else if (!strcmp (name, "cqm4i")) {
        p->i_cqm_preset = 2;
        b_error |= parse_cqm (value, p->cqm_4iy, 16);
        b_error |= parse_cqm (value, p->cqm_4ic, 16);
    }
    else if (!strcmp (name, "cqm4p")) {
        p->i_cqm_preset = 2;
        b_error |= parse_cqm (value, p->cqm_4py, 16);
        b_error |= parse_cqm (value, p->cqm_4pc, 16);
    }
    else if (!strcmp (name, "cqm4iy")) {
        p->i_cqm_preset = 2;
        b_error |= parse_cqm (value, p->cqm_4iy, 16);
    }
    else if (!strcmp (name, "cqm4ic")) {
        p->i_cqm_preset = 2;
        b_error |= parse_cqm (value, p->cqm_4ic, 16);
    }
    else if (!strcmp (name, "cqm4py")) {
        p->i_cqm_preset = 2;
        b_error |= parse_cqm (value, p->cqm_4py, 16);
    }
    else if (!strcmp (name, "cqm4pc")) {
        p->i_cqm_preset = 2;
        b_error |= parse_cqm (value, p->cqm_4pc, 16);
    }
    else if (!strcmp (name, "cqm8i")) {
        p->i_cqm_preset = 2;
        b_error |= parse_cqm (value, p->cqm_8iy, 64);
        b_error |= parse_cqm (value, p->cqm_8ic, 64);
    }
    else if (!strcmp (name, "cqm8p")) {
        p->i_cqm_preset = 2;
        b_error |= parse_cqm (value, p->cqm_8py, 64);
        b_error |= parse_cqm (value, p->cqm_8pc, 64);
    }
    else if (!strcmp (name, "log"))
        p->i_log_level = x264_atoi (value, &b_error);
    else if (!strcmp (name, "dump-yuv"))
        p->psz_dump_yuv = strdup (value);
    else if (!strcmp (name, "analyse") || !strcmp (name, "partitions")) {
        p->analyse.inter = 0;
        if (strstr (value, "none"))
            p->analyse.inter = 0;
        if (strstr (value, "all"))
            p->analyse.inter = ~0;
        if (strstr (value, "i4x4"))
            p->analyse.inter |= 0x0001;
        if (strstr (value, "i8x8"))
            p->analyse.inter |= 0x0002;
        if (strstr (value, "p8x8"))
            p->analyse.inter |= 0x0010;
        if (strstr (value, "p4x4"))
            p->analyse.inter |= 0x0020;
        if (strstr (value, "b8x8"))
            p->analyse.inter |= 0x0100;
    }
    else if (!strcmp (name, "8x8dct"))
        p->analyse.b_transform_8x8 = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "weightb") || !strcmp (name, "weight-b"))
        p->analyse.b_weighted_bipred = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "weightp"))
        p->analyse.i_weighted_pred = x264_atoi (value, &b_error);
    else if (!strcmp (name, "direct") || !strcmp (name, "direct-pred"))
        b_error |= parse_enum (value, x264_direct_pred_names, &p->analyse.i_direct_mv_pred);
    else if (!strcmp (name, "chroma-qp-offset"))
        p->analyse.i_chroma_qp_offset = x264_atoi (value, &b_error);
    else if (!strcmp (name, "me"))
        b_error |= parse_enum (value, x264_motion_est_names, &p->analyse.i_me_method);
    else if (!strcmp (name, "merange") || !strcmp (name, "me-range"))
        p->analyse.i_me_range = x264_atoi (value, &b_error);
    else if (!strcmp (name, "mvrange") || !strcmp (name, "mv-range"))
        p->analyse.i_mv_range = x264_atoi (value, &b_error);
    else if (!strcmp (name, "mvrange-thread") || !strcmp (name, "mv-range-thread"))
        p->analyse.i_mv_range_thread = x264_atoi (value, &b_error);
    else if (!strcmp (name, "subme") || !strcmp (name, "subq"))
        p->analyse.i_subpel_refine = x264_atoi (value, &b_error);
    else if (!strcmp (name, "psy-rd")) {
        if (2 == sscanf (value, "%f:%f", &p->analyse.f_psy_rd, &p->analyse.f_psy_trellis) || 2 == sscanf (value, "%f,%f", &p->analyse.f_psy_rd, &p->analyse.f_psy_trellis) || 2 == sscanf (value, "%f|%f", &p->analyse.f_psy_rd, &p->analyse.f_psy_trellis)) {
        }
        else if (sscanf (value, "%f", &p->analyse.f_psy_rd)) {
            p->analyse.f_psy_trellis = 0;
        }
        else {
            p->analyse.f_psy_rd = 0;
            p->analyse.f_psy_trellis = 0;
        }
    }
    else if (!strcmp (name, "psy"))
        p->analyse.b_psy = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "chroma-me"))
        p->analyse.b_chroma_me = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "mixed-refs"))
        p->analyse.b_mixed_references = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "trellis"))
        p->analyse.i_trellis = x264_atoi (value, &b_error);
    else if (!strcmp (name, "fast-pskip"))
        p->analyse.b_fast_pskip = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "dct-decimate"))
        p->analyse.b_dct_decimate = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "deadzone-inter"))
        p->analyse.i_luma_deadzone[0] = x264_atoi (value, &b_error);
    else if (!strcmp (name, "deadzone-intra"))
        p->analyse.i_luma_deadzone[1] = x264_atoi (value, &b_error);
    else if (!strcmp (name, "nr"))
        p->analyse.i_noise_reduction = x264_atoi (value, &b_error);
    else if (!strcmp (name, "bitrate")) {
        p->rc.i_bitrate = x264_atoi (value, &b_error);
        p->rc.i_rc_method = 2;
    }
    else if (!strcmp (name, "qp") || !strcmp (name, "qp_constant")) {
        p->rc.i_qp_constant = x264_atoi (value, &b_error);
        p->rc.i_rc_method = 0;
    }
    else if (!strcmp (name, "crf")) {
        p->rc.f_rf_constant = x264_atof (value, &b_error);
        p->rc.i_rc_method = 1;
    }
    else if (!strcmp (name, "crf-max"))
        p->rc.f_rf_constant_max = x264_atof (value, &b_error);
    else if (!strcmp (name, "rc-lookahead"))
        p->rc.i_lookahead = x264_atoi (value, &b_error);
    else if (!strcmp (name, "qpmin") || !strcmp (name, "qp-min"))
        p->rc.i_qp_min = x264_atoi (value, &b_error);
    else if (!strcmp (name, "qpmax") || !strcmp (name, "qp-max"))
        p->rc.i_qp_max = x264_atoi (value, &b_error);
    else if (!strcmp (name, "qpstep") || !strcmp (name, "qp-step"))
        p->rc.i_qp_step = x264_atoi (value, &b_error);
    else if (!strcmp (name, "ratetol"))
        p->rc.f_rate_tolerance = !strncmp ("inf", value, 3) ? 1e9 : x264_atof (value, &b_error);
    else if (!strcmp (name, "vbv-maxrate"))
        p->rc.i_vbv_max_bitrate = x264_atoi (value, &b_error);
    else if (!strcmp (name, "vbv-bufsize"))
        p->rc.i_vbv_buffer_size = x264_atoi (value, &b_error);
    else if (!strcmp (name, "vbv-init"))
        p->rc.f_vbv_buffer_init = x264_atof (value, &b_error);
    else if (!strcmp (name, "ipratio") || !strcmp (name, "ip-factor"))
        p->rc.f_ip_factor = x264_atof (value, &b_error);
    else if (!strcmp (name, "pbratio") || !strcmp (name, "pb-factor"))
        p->rc.f_pb_factor = x264_atof (value, &b_error);
    else if (!strcmp (name, "aq-mode"))
        p->rc.i_aq_mode = x264_atoi (value, &b_error);
    else if (!strcmp (name, "aq-strength"))
        p->rc.f_aq_strength = x264_atof (value, &b_error);
    else if (!strcmp (name, "pass")) {
        int pass = x264_clip3 (x264_atoi (value, &b_error), 0, 3);
        p->rc.b_stat_write = pass & 1;
        p->rc.b_stat_read = pass & 2;
    }
    else if (!strcmp (name, "stats")) {
        p->rc.psz_stat_in = strdup (value);
        p->rc.psz_stat_out = strdup (value);
    }
    else if (!strcmp (name, "qcomp"))
        p->rc.f_qcompress = x264_atof (value, &b_error);
    else if (!strcmp (name, "mbtree"))
        p->rc.b_mb_tree = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "qblur"))
        p->rc.f_qblur = x264_atof (value, &b_error);
    else if (!strcmp (name, "cplxblur") || !strcmp (name, "cplx-blur"))
        p->rc.f_complexity_blur = x264_atof (value, &b_error);
    else if (!strcmp (name, "zones"))
        p->rc.psz_zones = strdup (value);
    else if (!strcmp (name, "crop-rect"))
        b_error |= sscanf (value, "%u,%u,%u,%u", &p->crop_rect.i_left, &p->crop_rect.i_top, &p->crop_rect.i_right, &p->crop_rect.i_bottom) != 4;
    else if (!strcmp (name, "psnr"))
        p->analyse.b_psnr = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "ssim"))
        p->analyse.b_ssim = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "aud"))
        p->b_aud = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "sps-id"))
        p->i_sps_id = x264_atoi (value, &b_error);
    else if (!strcmp (name, "global-header"))
        p->b_repeat_headers = !(name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "repeat-headers"))
        p->b_repeat_headers = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "annexb"))
        p->b_annexb = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "force-cfr"))
        p->b_vfr_input = !(name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "nal-hrd"))
        b_error |= parse_enum (value, x264_nal_hrd_names, &p->i_nal_hrd);
    else if (!strcmp (name, "filler"))
        p->rc.b_filler = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "pic-struct"))
        p->b_pic_struct = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "fake-interlaced"))
        p->b_fake_interlaced = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "frame-packing"))
        p->i_frame_packing = x264_atoi (value, &b_error);
    else if (!strcmp (name, "stitchable"))
        p->b_stitchable = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "opencl"))
        p->b_opencl = (name_was_bool = 1, x264_atobool (value, &b_error));
    else if (!strcmp (name, "opencl-clbin"))
        p->psz_clbin_file = strdup (value);
    else if (!strcmp (name, "opencl-device"))
        p->i_opencl_device = x264_atoi (value, &b_error);
    else
        return (-1);
    if (name_buf)
        free (name_buf);
    b_error |= value_was_null && !name_was_bool;
    return b_error ? (-2) : 0;
}

void x264_log (x264_t *h, int i_level, const char *psz_fmt, ... ) {
    if (!h || i_level <= h->param.i_log_level) {
        va_list arg;
        __builtin_va_start (arg, psz_fmt);
        if (!h)
            x264_log_default (((void *) 0), i_level, psz_fmt, arg);
        else
            h->param.pf_log (h->param.p_log_private, i_level, psz_fmt, arg);
        __builtin_va_end (arg);
    }
}

void x264_log_default (void *p_unused, int i_level, const char *psz_fmt, va_list arg) {
    char *psz_prefix;
    switch (i_level) {
    case 0 :
        psz_prefix = "error";
        break;
    case 1 :
        psz_prefix = "warning";
        break;
    case 2 :
        psz_prefix = "info";
        break;
    case 3 :
        psz_prefix = "debug";
        break;
    default :
        psz_prefix = "unknown";
        break;
    }
    fprintf (stderr, "x264 [%s]: ", psz_prefix);
    vfprintf (stderr, psz_fmt, arg);
}

void x264_picture_init (x264_picture_t *pic) {
    memset (pic, 0, sizeof (x264_picture_t));
    pic->i_type = 0x0000;
    pic->i_qpplus1 = 0;
    pic->i_pic_struct = PIC_STRUCT_AUTO;
}

int x264_picture_alloc (x264_picture_t *pic, int i_csp, int i_width, int i_height) {
    typedef struct {
        int planes;
        int width_fix8 [3];
        int height_fix8 [3];
    } x264_csp_tab_t;
    const x264_csp_tab_t x264_csp_tab [] = {
        [0x0001] = {3, {256 * 1, 256 / 2, 256 / 2}, {256 * 1, 256 / 2, 256 / 2}},
        [0x0002] = {3, {256 * 1, 256 / 2, 256 / 2}, {256 * 1, 256 / 2, 256 / 2}},
        [0x0003] = {2, {256 * 1, 256 * 1}, {256 * 1, 256 / 2},},
        [0x0004] = {3, {256 * 1, 256 / 2, 256 / 2}, {256 * 1, 256 * 1, 256 * 1}},
        [0x0005] = {3, {256 * 1, 256 / 2, 256 / 2}, {256 * 1, 256 * 1, 256 * 1}},
        [0x0006] = {2, {256 * 1, 256 * 1}, {256 * 1, 256 * 1},},
        [0x0008] = {3, {256 * 1, 256 * 1, 256 * 1}, {256 * 1, 256 * 1, 256 * 1}},
        [0x0009] = {3, {256 * 1, 256 * 1, 256 * 1}, {256 * 1, 256 * 1, 256 * 1}},
        [0x000a] = {1, {256 * 3}, {256 * 1},},
        [0x000b] = {1, {256 * 4}, {256 * 1},},
        [0x000c] = {1, {256 * 3}, {256 * 1},},
    };
    int csp = i_csp & 0x00ff;
    if (csp <= 0x0000 || csp >= 0x000d || csp == 0x0007)
        return -1;
    x264_picture_init (pic);
    pic->img.i_csp = i_csp;
    pic->img.i_plane = x264_csp_tab[csp].planes;
    int depth_factor = i_csp & 0x2000 ? 2 : 1;
    int plane_offset [3] = {0};
    int frame_size = 0;
    for (int i = 0; i < pic->img.i_plane; i++) {
        int stride = (((int64_t) i_width * x264_csp_tab[csp].width_fix8[i]) >> 8) * depth_factor;
        int plane_size = (((int64_t) i_height * x264_csp_tab[csp].height_fix8[i]) >> 8) * stride;
        pic->img.i_stride[i] = stride;
        plane_offset[i] = frame_size;
        frame_size += plane_size;
    }
    pic->img.plane[0] = x264_malloc (frame_size);
    if (!pic->img.plane[0])
        return -1;
    for (int i = 1; i < pic->img.i_plane; i++)
        pic->img.plane[i] = pic->img.plane[0] + plane_offset[i];
    return 0;
}

void x264_picture_clean (x264_picture_t *pic) {
    x264_free (pic -> img.plane [0]);
    memset (pic, 0, sizeof (x264_picture_t));
}

void *x264_malloc (int i_size) {
    uint8_t *align_buf = ((void *) 0);
    if (i_size >= 2 * 1024 * 1024 * 7 / 8) {
        align_buf = memalign (2 * 1024 * 1024, i_size);
        if (align_buf) {
            size_t madv_size = (i_size + 2 * 1024 * 1024 - 2 * 1024 * 1024 * 7 / 8) & ~(2 * 1024 * 1024 - 1);
            madvise (align_buf, madv_size, 14);
        }
    }
    else
        align_buf = memalign (32, i_size);
    if (!align_buf)
        x264_log (((void *) 0), 0, "malloc of size %d failed\n", i_size);
    return align_buf;
}

void x264_free (void *p) {
    if (p) {
        free (p);
    }
}

void x264_reduce_fraction (uint32_t *n, uint32_t *d) {
    uint32_t a = *n;
    uint32_t b = *d;
    uint32_t c;
    if (!a || !b)
        return;
    c = a % b;
    while (c) {
        a = b;
        b = c;
        c = a % b;
    }
    *n /= b;
    *d /= b;
}

void x264_reduce_fraction64 (uint64_t *n, uint64_t *d) {
    uint64_t a = *n;
    uint64_t b = *d;
    uint64_t c;
    if (!a || !b)
        return;
    c = a % b;
    while (c) {
        a = b;
        b = c;
        c = a % b;
    }
    *n /= b;
    *d /= b;
}

char *x264_slurp_file (const char *filename) {
    int b_error = 0;
    size_t i_size;
    char *buf;
    FILE *fh = fopen (filename, "rb");
    if (!fh)
        return ((void *) 0);
    b_error |= fseeko (fh, 0, 2) < 0;
    b_error |= (i_size = ftello (fh)) <= 0;
    b_error |= fseeko (fh, 0, 0) < 0;
    if (b_error)
        goto error;
    buf = x264_malloc (i_size +2);
    if (!buf)
        goto error;
    b_error |= fread (buf, 1, i_size, fh) != i_size;
    if (buf[i_size - 1] != '\n')
        buf[i_size++] = '\n';
    buf[i_size] = 0;
    fclose (fh);
    if (b_error) {
        x264_free (buf);
        return ((void *) 0);
    }
    return buf;
error :
    fclose (fh);
    return ((void *) 0);
}

char *x264_param2string (x264_param_t *p, int b_res) {
    int len = 1000;
    char *buf, *s;
    if (p->rc.psz_zones)
        len += strlen (p->rc.psz_zones);
    buf = s = x264_malloc (len);
    if (!buf)
        return ((void *) 0);
    if (b_res) {
        s += sprintf (s, "%dx%d ", p->i_width, p->i_height);
        s += sprintf (s, "fps=%u/%u ", p->i_fps_num, p->i_fps_den);
        s += sprintf (s, "timebase=%u/%u ", p->i_timebase_num, p->i_timebase_den);
        s += sprintf (s, "bitdepth=%d ", 8);
    }
    if (p->b_opencl)
        s += sprintf (s, "opencl=%d ", p->b_opencl);
    s += sprintf (s, "cabac=%d", p->b_cabac);
    s += sprintf (s, " ref=%d", p->i_frame_reference);
    s += sprintf (s, " deblock=%d:%d:%d", p->b_deblocking_filter, p->i_deblocking_filter_alphac0, p->i_deblocking_filter_beta);
    s += sprintf (s, " analyse=%#x:%#x", p->analyse.intra, p->analyse.inter);
    s += sprintf (s, " me=%s", x264_motion_est_names[p->analyse.i_me_method]);
    s += sprintf (s, " subme=%d", p->analyse.i_subpel_refine);
    s += sprintf (s, " psy=%d", p->analyse.b_psy);
    if (p->analyse.b_psy)
        s += sprintf (s, " psy_rd=%.2f:%.2f", p->analyse.f_psy_rd, p->analyse.f_psy_trellis);
    s += sprintf (s, " mixed_ref=%d", p->analyse.b_mixed_references);
    s += sprintf (s, " me_range=%d", p->analyse.i_me_range);
    s += sprintf (s, " chroma_me=%d", p->analyse.b_chroma_me);
    s += sprintf (s, " trellis=%d", p->analyse.i_trellis);
    s += sprintf (s, " 8x8dct=%d", p->analyse.b_transform_8x8);
    s += sprintf (s, " cqm=%d", p->i_cqm_preset);
    s += sprintf (s, " deadzone=%d,%d", p->analyse.i_luma_deadzone[0], p->analyse.i_luma_deadzone[1]);
    s += sprintf (s, " fast_pskip=%d", p->analyse.b_fast_pskip);
    s += sprintf (s, " chroma_qp_offset=%d", p->analyse.i_chroma_qp_offset);
    s += sprintf (s, " threads=%d", p->i_threads);
    s += sprintf (s, " lookahead_threads=%d", p->i_lookahead_threads);
    s += sprintf (s, " sliced_threads=%d", p->b_sliced_threads);
    if (p->i_slice_count)
        s += sprintf (s, " slices=%d", p->i_slice_count);
    if (p->i_slice_count_max)
        s += sprintf (s, " slices_max=%d", p->i_slice_count_max);
    if (p->i_slice_max_size)
        s += sprintf (s, " slice_max_size=%d", p->i_slice_max_size);
    if (p->i_slice_max_mbs)
        s += sprintf (s, " slice_max_mbs=%d", p->i_slice_max_mbs);
    if (p->i_slice_min_mbs)
        s += sprintf (s, " slice_min_mbs=%d", p->i_slice_min_mbs);
    s += sprintf (s, " nr=%d", p->analyse.i_noise_reduction);
    s += sprintf (s, " decimate=%d", p->analyse.b_dct_decimate);
    s += sprintf (s, " interlaced=%s", p->b_interlaced ? p->b_tff ? "tff" : "bff" : p->b_fake_interlaced ? "fake" : "0");
    s += sprintf (s, " bluray_compat=%d", p->b_bluray_compat);
    if (p->b_stitchable)
        s += sprintf (s, " stitchable=%d", p->b_stitchable);
    s += sprintf (s, " constrained_intra=%d", p->b_constrained_intra);
    s += sprintf (s, " bframes=%d", p->i_bframe);
    if (p->i_bframe) {
        s += sprintf (s, " b_pyramid=%d b_adapt=%d b_bias=%d direct=%d weightb=%d open_gop=%d", p->i_bframe_pyramid, p->i_bframe_adaptive, p->i_bframe_bias, p->analyse.i_direct_mv_pred, p->analyse.b_weighted_bipred, p->b_open_gop);
    }
    s += sprintf (s, " weightp=%d", p->analyse.i_weighted_pred > 0 ? p->analyse.i_weighted_pred : 0);
    if (p->i_keyint_max == (1 << 30))
        s += sprintf (s, " keyint=infinite");
    else
        s += sprintf (s, " keyint=%d", p->i_keyint_max);
    s += sprintf (s, " keyint_min=%d scenecut=%d intra_refresh=%d", p->i_keyint_min, p->i_scenecut_threshold, p->b_intra_refresh);
    if (p->rc.b_mb_tree || p->rc.i_vbv_buffer_size)
        s += sprintf (s, " rc_lookahead=%d", p->rc.i_lookahead);
    s += sprintf (s, " rc=%s mbtree=%d", p->rc.i_rc_method == 2 ? (p->rc.b_stat_read ? "2pass" : p->rc.i_vbv_max_bitrate == p->rc.i_bitrate ? "cbr" : "abr") : p->rc.i_rc_method == 1 ? "crf" : "cqp", p->rc.b_mb_tree);
    if (p->rc.i_rc_method == 2 || p->rc.i_rc_method == 1) {
        if (p->rc.i_rc_method == 1)
            s += sprintf (s, " crf=%.1f", p->rc.f_rf_constant);
        else
            s += sprintf (s, " bitrate=%d ratetol=%.1f", p->rc.i_bitrate, p->rc.f_rate_tolerance);
        s += sprintf (s, " qcomp=%.2f qpmin=%d qpmax=%d qpstep=%d", p->rc.f_qcompress, p->rc.i_qp_min, p->rc.i_qp_max, p->rc.i_qp_step);
        if (p->rc.b_stat_read)
            s += sprintf (s, " cplxblur=%.1f qblur=%.1f", p->rc.f_complexity_blur, p->rc.f_qblur);
        if (p->rc.i_vbv_buffer_size) {
            s += sprintf (s, " vbv_maxrate=%d vbv_bufsize=%d", p->rc.i_vbv_max_bitrate, p->rc.i_vbv_buffer_size);
            if (p->rc.i_rc_method == 1)
                s += sprintf (s, " crf_max=%.1f", p->rc.f_rf_constant_max);
        }
    }
    else if (p->rc.i_rc_method == 0)
        s += sprintf (s, " qp=%d", p->rc.i_qp_constant);
    if (p->rc.i_vbv_buffer_size)
        s += sprintf (s, " nal_hrd=%s filler=%d", x264_nal_hrd_names[p->i_nal_hrd], p->rc.b_filler);
    if (p->crop_rect.i_left | p->crop_rect.i_top | p->crop_rect.i_right | p->crop_rect.i_bottom)
        s += sprintf (s, " crop_rect=%u,%u,%u,%u", p->crop_rect.i_left, p->crop_rect.i_top, p->crop_rect.i_right, p->crop_rect.i_bottom);
    if (p->i_frame_packing >= 0)
        s += sprintf (s, " frame-packing=%d", p->i_frame_packing);
    if (!(p->rc.i_rc_method == 0 && p->rc.i_qp_constant == 0)) {
        s += sprintf (s, " ip_ratio=%.2f", p->rc.f_ip_factor);
        if (p->i_bframe && !p->rc.b_mb_tree)
            s += sprintf (s, " pb_ratio=%.2f", p->rc.f_pb_factor);
        s += sprintf (s, " aq=%d", p->rc.i_aq_mode);
        if (p->rc.i_aq_mode)
            s += sprintf (s, ":%.2f", p->rc.f_aq_strength);
        if (p->rc.psz_zones)
            s += sprintf (s, " zones=%s", p->rc.psz_zones);
        else if (p->rc.i_zones)
            s += sprintf (s, " zones");
    }
    return buf;
}

pixel x264_clip_pixel (int x) {
    return ((x & ~((1 << 8) - 1)) ? (-x) >> 31 & ((1 << 8) - 1) : x);
}

int x264_clip3 (int v, int i_min, int i_max) {
    return ((v < i_min) ? i_min : (v > i_max) ? i_max : v);
}

double x264_clip3f (double v, double f_min, double f_max) {
    return ((v < f_min) ? f_min : (v > f_max) ? f_max : v);
}

int x264_median (int a, int b, int c) {
    int t = (a - b) & ((a - b) >> 31);
    a -= t;
    b += t;
    b -= (b - c) & ((b - c) >> 31);
    b += (a - b) & ((a - b) >> 31);
    return b;
}

void x264_median_mv (int16_t *dst, int16_t *a, int16_t *b, int16_t *c) {
    dst[0] = x264_median (a[0], b[0], c[0]);
    dst[1] = x264_median (a[1], b[1], c[1]);
}

int x264_predictor_difference (int16_t (*mvc) [2], intptr_t i_mvc) {
    int sum = 0;
    for (int i = 0; i < i_mvc - 1; i++) {
        sum += abs (mvc[i][0] - mvc[i + 1][0]) + abs (mvc[i][1] - mvc[i + 1][1]);
    }
    return sum;
}

uint16_t x264_cabac_mvd_sum (uint8_t *mvdleft, uint8_t *mvdtop) {
    int amvd0 = mvdleft[0] + mvdtop[0];
    int amvd1 = mvdleft[1] + mvdtop[1];
    amvd0 = (amvd0 > 2) + (amvd0 > 32);
    amvd1 = (amvd1 > 2) + (amvd1 > 32);
    return amvd0 + (amvd1 << 8);
}

int x264_exp2fix8 (float x) {
    int i = x * (-64.f / 6.f) + 512.5f;
    if (i < 0)
        return 0;
    if (i > 1023)
        return 0xffff;
    return (x264_exp2_lut[i & 63] + 256) << (i >> 6) >> 8;
}

float x264_log2 (uint32_t x) {
    int lz = __builtin_clz (x);
    return x264_log2_lut[(x << lz >> 24) & 0x7f] + x264_log2_lz_lut[lz];
}

const char slice_type_to_char [3] = {'P', 'B', 'I'};

int x264_predictor_roundclip (int16_t (*dst) [2], int16_t (*mvc) [2], int i_mvc, int16_t mv_limit [2] [2], uint32_t pmv) {
    int cnt = 0;
    for (int i = 0; i < i_mvc; i++) {
        int mx = (mvc[i][0] + 2) >> 2;
        int my = (mvc[i][1] + 2) >> 2;
        uint32_t mv = pack16to32_mask (mx, my);
        if (!mv || mv == pmv)
            continue;
        dst[cnt][0] = x264_clip3 (mx, mv_limit[0][0], mv_limit[1][0]);
        dst[cnt][1] = x264_clip3 (my, mv_limit[0][1], mv_limit[1][1]);
        cnt++;
    }
    return cnt;
}

int x264_predictor_clip (int16_t (*dst) [2], int16_t (*mvc) [2], int i_mvc, int16_t mv_limit [2] [2], uint32_t pmv) {
    int cnt = 0;
    int qpel_limit [4] = {mv_limit[0][0] << 2, mv_limit[0][1] << 2, mv_limit[1][0] << 2, mv_limit[1][1] << 2};
    for (int i = 0; i < i_mvc; i++) {
        uint32_t mv = (((x264_union32_t *) (mvc[i]))->i);
        int mx = mvc[i][0];
        int my = mvc[i][1];
        if (!mv || mv == pmv)
            continue;
        dst[cnt][0] = x264_clip3 (mx, qpel_limit[0], qpel_limit[2]);
        dst[cnt][1] = x264_clip3 (my, qpel_limit[1], qpel_limit[3]);
        cnt++;
    }
    return cnt;
}

const uint8_t x264_scan8 [16 * 3 + 3] = {4 + 1 * 8, 5 + 1 * 8, 4 + 2 * 8, 5 + 2 * 8, 6 + 1 * 8, 7 + 1 * 8, 6 + 2 * 8, 7 + 2 * 8, 4 + 3 * 8, 5 + 3 * 8, 4 + 4 * 8, 5 + 4 * 8, 6 + 3 * 8, 7 + 3 * 8, 6 + 4 * 8, 7 + 4 * 8, 4 + 6 * 8, 5 + 6 * 8, 4 + 7 * 8, 5 + 7 * 8, 6 + 6 * 8, 7 + 6 * 8, 6 + 7 * 8, 7 + 7 * 8, 4 + 8 * 8, 5 + 8 * 8, 4 + 9 * 8, 5 + 9 * 8, 6 + 8 * 8, 7 + 8 * 8, 6 + 9 * 8, 7 + 9 * 8, 4 + 11 * 8, 5 + 11 * 8, 4 + 12 * 8, 5 + 12 * 8, 6 + 11 * 8, 7 + 11 * 8, 6 + 12 * 8, 7 + 12 * 8, 4 + 13 * 8, 5 + 13 * 8, 4 + 14 * 8, 5 + 14 * 8, 6 + 13 * 8, 7 + 13 * 8, 6 + 14 * 8, 7 + 14 * 8, 0 + 0 * 8, 0 + 5 * 8, 0 + 10 * 8};
