
#include "video.h"

#define NAME "depth"

#define FAIL_IF_ERROR( cond, ... ) FAIL_IF_ERR( cond, NAME, __VA_ARGS__ )
cli_vid_filter_t depth_filter;
struct _depth_hnd_t {
    hnd_t prev_hnd;
    cli_vid_filter_t prev_filter;
    int bit_depth;
    int dst_csp;
    cli_pic_t buffer;
    int16_t *error_buf;
};
typedef struct _depth_hnd_t depth_hnd_t;

int depth_filter_csp_is_supported (int csp) {
    int csp_mask = csp & 0x00ff;
    return csp_mask == 0x0001 || csp_mask == 0x0004 || csp_mask == 0x0008 || csp_mask == 0x0002 || csp_mask == 0x0005 || csp_mask == 0x0009 || csp_mask == 0x0003 || csp_mask == 0x0006 || csp_mask == 0x000a || csp_mask == 0x000c || csp_mask == 0x000b;
}

int csp_num_interleaved (int csp, int plane) {
    int csp_mask = csp & 0x00ff;
    return (csp_mask == 0x0003 || csp_mask == 0x0006) && plane == 1 ? 2 : csp_mask == 0x000a || csp_mask == 0x000c ? 3 : csp_mask == 0x000b ? 4 : 1;
}

void dither_plane_1 (pixel *dst, int dst_stride, uint16_t *src, int src_stride, int width, int height, int16_t *errors) {
    const int lshift = 16 - 8;
    const int rshift = 16 - 8 + 2;
    const int half = 1 << (16 - 8 + 1);
    const int pixel_max = (1 << 8) - 1;
    memset (errors, 0, (width + 1) * sizeof (int16_t));
    for (int y = 0; y < height; y++, src += src_stride, dst += dst_stride) {
        int err = 0;
        for (int x = 0; x < width; x++) {
            err = err * 2 + errors[x] + errors[x + 1];
            dst[x * 1] = x264_clip3 (((src[x * 1] << 2) + err + half) >> rshift, 0, pixel_max);
            errors[x] = err = src[x * 1] - (dst[x * 1] << lshift);
        }
    }
}

void dither_plane_2 (pixel *dst, int dst_stride, uint16_t *src, int src_stride, int width, int height, int16_t *errors) {
    const int lshift = 16 - 8;
    const int rshift = 16 - 8 + 2;
    const int half = 1 << (16 - 8 + 1);
    const int pixel_max = (1 << 8) - 1;
    memset (errors, 0, (width + 1) * sizeof (int16_t));
    for (int y = 0; y < height; y++, src += src_stride, dst += dst_stride) {
        int err = 0;
        for (int x = 0; x < width; x++) {
            err = err * 2 + errors[x] + errors[x + 1];
            dst[x * 2] = x264_clip3 (((src[x * 2] << 2) + err + half) >> rshift, 0, pixel_max);
            errors[x] = err = src[x * 2] - (dst[x * 2] << lshift);
        }
    }
}

void dither_plane_3 (pixel *dst, int dst_stride, uint16_t *src, int src_stride, int width, int height, int16_t *errors) {
    const int lshift = 16 - 8;
    const int rshift = 16 - 8 + 2;
    const int half = 1 << (16 - 8 + 1);
    const int pixel_max = (1 << 8) - 1;
    memset (errors, 0, (width + 1) * sizeof (int16_t));
    for (int y = 0; y < height; y++, src += src_stride, dst += dst_stride) {
        int err = 0;
        for (int x = 0; x < width; x++) {
            err = err * 2 + errors[x] + errors[x + 1];
            dst[x * 3] = x264_clip3 (((src[x * 3] << 2) + err + half) >> rshift, 0, pixel_max);
            errors[x] = err = src[x * 3] - (dst[x * 3] << lshift);
        }
    }
}

void dither_plane_4 (pixel *dst, int dst_stride, uint16_t *src, int src_stride, int width, int height, int16_t *errors) {
    const int lshift = 16 - 8;
    const int rshift = 16 - 8 + 2;
    const int half = 1 << (16 - 8 + 1);
    const int pixel_max = (1 << 8) - 1;
    memset (errors, 0, (width + 1) * sizeof (int16_t));
    for (int y = 0; y < height; y++, src += src_stride, dst += dst_stride) {
        int err = 0;
        for (int x = 0; x < width; x++) {
            err = err * 2 + errors[x] + errors[x + 1];
            dst[x * 4] = x264_clip3 (((src[x * 4] << 2) + err + half) >> rshift, 0, pixel_max);
            errors[x] = err = src[x * 4] - (dst[x * 4] << lshift);
        }
    }
}

void dither_image (cli_image_t *out, cli_image_t *img, int16_t *error_buf) {
    int csp_mask = img->csp & 0x00ff;
    for (int i = 0; i < img->planes; i++) {
        int num_interleaved = csp_num_interleaved (img->csp, i);
        int height = x264_cli_csps[csp_mask].height[i] * img->height;
        int width = x264_cli_csps[csp_mask].width[i] * img->width / num_interleaved;
        if (num_interleaved == 4) {
            dither_plane_4 (((pixel *) out -> plane [i]) + 0, out -> stride [i] / sizeof (pixel), ((uint16_t *) img -> plane [i]) + 0, img -> stride [i] / 2, width, height, error_buf);
            dither_plane_4 (((pixel *) out -> plane [i]) + 1, out -> stride [i] / sizeof (pixel), ((uint16_t *) img -> plane [i]) + 1, img -> stride [i] / 2, width, height, error_buf);
            dither_plane_4 (((pixel *) out -> plane [i]) + 2, out -> stride [i] / sizeof (pixel), ((uint16_t *) img -> plane [i]) + 2, img -> stride [i] / 2, width, height, error_buf);
            dither_plane_4 (((pixel *) out -> plane [i]) + 3, out -> stride [i] / sizeof (pixel), ((uint16_t *) img -> plane [i]) + 3, img -> stride [i] / 2, width, height, error_buf);
        }
        else if (num_interleaved == 3) {
            dither_plane_3 (((pixel *) out -> plane [i]) + 0, out -> stride [i] / sizeof (pixel), ((uint16_t *) img -> plane [i]) + 0, img -> stride [i] / 2, width, height, error_buf);
            dither_plane_3 (((pixel *) out -> plane [i]) + 1, out -> stride [i] / sizeof (pixel), ((uint16_t *) img -> plane [i]) + 1, img -> stride [i] / 2, width, height, error_buf);
            dither_plane_3 (((pixel *) out -> plane [i]) + 2, out -> stride [i] / sizeof (pixel), ((uint16_t *) img -> plane [i]) + 2, img -> stride [i] / 2, width, height, error_buf);
        }
        else if (num_interleaved == 2) {
            dither_plane_2 (((pixel *) out -> plane [i]) + 0, out -> stride [i] / sizeof (pixel), ((uint16_t *) img -> plane [i]) + 0, img -> stride [i] / 2, width, height, error_buf);
            dither_plane_2 (((pixel *) out -> plane [i]) + 1, out -> stride [i] / sizeof (pixel), ((uint16_t *) img -> plane [i]) + 1, img -> stride [i] / 2, width, height, error_buf);
        }
        else {
            dither_plane_1 (((pixel *) out -> plane [i]) + 0, out -> stride [i] / sizeof (pixel), ((uint16_t *) img -> plane [i]) + 0, img -> stride [i] / 2, width, height, error_buf);
        }
    }
}

void scale_image (cli_image_t *output, cli_image_t *img) {
    int csp_mask = img->csp & 0x00ff;
    const int shift = 8 - 8;
    for (int i = 0; i < img->planes; i++) {
        uint8_t *src = img->plane[i];
        uint16_t *dst = (uint16_t *) output->plane[i];
        int height = x264_cli_csps[csp_mask].height[i] * img->height;
        int width = x264_cli_csps[csp_mask].width[i] * img->width;
        for (int j = 0; j < height; j++) {
            for (int k = 0; k < width; k++)
                dst[k] = src[k] << shift;
            src += img->stride[i];
            dst += output->stride[i] / 2;
        }
    }
}

int get_frame (hnd_t handle, cli_pic_t *output, int frame) {
    depth_hnd_t *h = handle;
    if (h->prev_filter.get_frame (h->prev_hnd, output, frame))
        return -1;
    if (h->bit_depth < 16 && output->img.csp & 0x2000) {
        dither_image (& h -> buffer.img, & output -> img, h -> error_buf);
        output->img = h->buffer.img;
    }
    else if (h->bit_depth > 8 && !(output->img.csp & 0x2000)) {
        scale_image (& h -> buffer.img, & output -> img);
        output->img = h->buffer.img;
    }
    return 0;
}

int release_frame (hnd_t handle, cli_pic_t *pic, int frame) {
    depth_hnd_t *h = handle;
    return h->prev_filter.release_frame (h->prev_hnd, pic, frame);
}

void free_filter (hnd_t handle) {
    depth_hnd_t *h = handle;
    h->prev_filter.free (h->prev_hnd);
    x264_cli_pic_clean (& h -> buffer);
    x264_free (h);
}

int init (hnd_t *handle, cli_vid_filter_t *filter, video_info_t *info, x264_param_t *param, char *opt_string) {
    int ret = 0;
    int change_fmt = (info->csp ^ param->i_csp) & 0x2000;
    int csp = ~(~info->csp ^ change_fmt);
    int bit_depth = 8 * x264_cli_csp_depth_factor (csp);
    if (opt_string) {
        const char *optlist [] = {"bit_depth", ((void *) 0)};
        char **opts = x264_split_options (opt_string, optlist);
        if (opts) {
            char *str_bit_depth = x264_get_option ("bit_depth", opts);
            bit_depth = x264_otoi (str_bit_depth, -1);
            ret = bit_depth < 8 || bit_depth > 16;
            csp = bit_depth > 8 ? csp | 0x2000 : csp & ~0x2000;
            change_fmt = (info->csp ^ csp) & 0x2000;
            x264_free_string_array (opts);
        }
        else
            ret = 1;
    }
    if (bit_depth != 8) {
        x264_cli_log ("depth", 0, "this build supports only bit depth %d\n", 8);
        return -1;
    }
    if (ret) {
        x264_cli_log ("depth", 0, "unsupported bit depth conversion.\n");
        return -1;
    }
    if (change_fmt || bit_depth != 8 * x264_cli_csp_depth_factor (csp)) {
        if (!depth_filter_csp_is_supported (csp)) {
            x264_cli_log ("depth", 0, "unsupported colorspace.\n");
            return -1;
        }
        depth_hnd_t *h = x264_malloc (sizeof (depth_hnd_t) + (info->width + 1) * sizeof (int16_t));
        if (!h)
            return -1;
        h->error_buf = (int16_t *) (h + 1);
        h->dst_csp = csp;
        h->bit_depth = bit_depth;
        h->prev_hnd = *handle;
        h->prev_filter = *filter;
        if (x264_cli_pic_alloc (&h->buffer, h->dst_csp, info->width, info->height)) {
            x264_free (h);
            return -1;
        }
        *handle = h;
        *filter = depth_filter;
        info->csp = h->dst_csp;
    }
    return 0;
}

cli_vid_filter_t depth_filter = {"depth", ((void *) 0), init, get_frame, release_frame, free_filter, ((void *) 0)};
