
#include "video.h"

#include "internal.h"
struct _fix_vfr_pts_hnd_t {
    hnd_t prev_hnd;
    cli_vid_filter_t prev_filter;
    cli_pic_t buffer;
    cli_pic_t holder;
    int buffer_allocated;
    int holder_frame;
    int holder_ret;
    int64_t pts;
    int64_t last_duration;
};
typedef struct _fix_vfr_pts_hnd_t fix_vfr_pts_hnd_t;
cli_vid_filter_t fix_vfr_pts_filter;

int init_FIX (hnd_t *handle, cli_vid_filter_t *filter, video_info_t *info, x264_param_t *param, char *opt_string) {
    if (!info->vfr)
        return 0;
    fix_vfr_pts_hnd_t *h = calloc (1, sizeof (fix_vfr_pts_hnd_t));
    if (!h)
        return -1;
    h->holder_frame = -1;
    h->prev_hnd = *handle;
    h->prev_filter = *filter;
    *handle = h;
    *filter = fix_vfr_pts_filter;
    return 0;
}

int get_frame_FIX (hnd_t handle, cli_pic_t *output, int frame) {
    fix_vfr_pts_hnd_t *h = handle;
    if (frame == h->holder_frame) {
        if (h->holder_ret)
            return h->holder_ret;
    }
    else {
        if (h->holder_frame > 0 && h->holder_frame < frame && h->prev_filter.release_frame (h->prev_hnd, &h->holder, h->holder_frame))
            return -1;
        h->holder_frame = -1;
        if (h->prev_filter.get_frame (h->prev_hnd, &h->holder, frame))
            return -1;
    }
    if (!h->holder.duration) {
        if (!h->buffer_allocated) {
            if (x264_cli_pic_alloc (&h->buffer, h->holder.img.csp, h->holder.img.width, h->holder.img.height))
                return -1;
            h->buffer_allocated = 1;
        }
        h->holder_frame = frame + 1;
        if (x264_cli_pic_copy (&h->buffer, &h->holder) || h->prev_filter.release_frame (h->prev_hnd, &h->holder, frame))
            return -1;
        h->holder_ret = h->prev_filter.get_frame (h->prev_hnd, &h->holder, h->holder_frame);
        if (!h->holder_ret)
            h->last_duration = X264_MAX (h->holder.pts - h->buffer.pts, 1);
        h->buffer.duration = h->last_duration;
        *output = h->buffer;
    }
    else
        *output = h->holder;
    output->pts = h->pts;
    h->pts += output->duration;
    return 0;
}

int release_frame_FIX (hnd_t handle, cli_pic_t *pic, int frame) {
    fix_vfr_pts_hnd_t *h = handle;
    if (frame == (h->holder_frame - 1))
        return 0;
    return h->prev_filter.release_frame (h->prev_hnd, pic, frame);
}

void free_filter_FIX (hnd_t handle) {
    fix_vfr_pts_hnd_t *h = handle;
    h->prev_filter.free (h->prev_hnd);
    if (h->buffer_allocated)
        x264_cli_pic_clean (&h->buffer);
    free (h);
}

cli_vid_filter_t fix_vfr_pts_filter = {"fix_vfr_pts", NULL, init_FIX, get_frame_FIX, release_frame_FIX, free_filter_FIX, NULL};
