
#include "common.h"

int align_stride (int x, int align, int disalign) {
    x = (((x) +((align) -1)) & ~((align) -1));
    if (!(x & (disalign - 1)))
        x += align;
    return x;
}

int align_plane_size (int x, int disalign) {
    if (!(x & (disalign - 1)))
        x += 128;
    return x;
}

int x264_frame_internal_csp (int external_csp) {
    switch (external_csp & 0x00ff) {
    case 0x0003 :
    case 0x0001 :
    case 0x0002 :
        return 0x0003;
    case 0x0006 :
    case 0x0004 :
    case 0x0005 :
    case 0x0007 :
        return 0x0006;
    case 0x0008 :
    case 0x0009 :
    case 0x000a :
    case 0x000b :
    case 0x000c :
        return 0x0008;
    default :
        return 0x0000;
    }
}

x264_frame_t *x264_frame_new (x264_t *h, int b_fdec) {
    x264_frame_t *frame;
    int i_csp = x264_frame_internal_csp (h->param.i_csp);
    int i_mb_count = h->mb.i_mb_count;
    int i_stride, i_width, i_lines, luma_plane_count;
    int i_padv = 32 << 0;
    int align = 16;
    if (h->param.cpu & 0x0040000)
        align = 64;
    else if (h->param.cpu & 0x0020000 || h->param.cpu & 0x0002000)
        align = 32;
    int disalign = 1 << 10;
    do {
        do {
            frame = x264_malloc (sizeof (x264_frame_t));
            if (!frame)
                goto fail;
        }
        while (0);
        memset (frame, 0, sizeof (x264_frame_t));
    }
    while (0);
    int prealloc_idx = 0;
    size_t prealloc_size = 0;
    uint8_t **preallocs [1024];
    i_width = h->mb.i_mb_width * 16;
    i_lines = h->mb.i_mb_height * 16;
    i_stride = align_stride (i_width +2 * 32, align, disalign);
    if (i_csp == 0x0003 || i_csp == 0x0006) {
        luma_plane_count = 1;
        frame->i_plane = 2;
        for (int i = 0; i < 2; i++) {
            frame->i_width[i] = i_width >> i;
            frame->i_lines[i] = i_lines >> (i && i_csp == 0x0003);
            frame->i_stride[i] = i_stride;
        }
    }
    else if (i_csp == 0x0008) {
        luma_plane_count = 3;
        frame->i_plane = 3;
        for (int i = 0; i < 3; i++) {
            frame->i_width[i] = i_width;
            frame->i_lines[i] = i_lines;
            frame->i_stride[i] = i_stride;
        }
    }
    else
        goto fail;
    frame->i_csp = i_csp;
    frame->i_width_lowres = frame->i_width[0] / 2;
    frame->i_lines_lowres = frame->i_lines[0] / 2;
    frame->i_stride_lowres = align_stride (frame->i_width_lowres + 2 * 32, align, disalign << 1);
    for (int i = 0; i < h->param.i_bframe + 2; i++)
        for (int j = 0; j < h->param.i_bframe + 2; j++)
            do {
                frame->i_row_satds[i][j] = (void *) prealloc_size;
                preallocs[prealloc_idx++] = (uint8_t **) &frame->i_row_satds[i][j];
                prealloc_size += (((i_lines / 16 * sizeof (int)) + ((32) - 1)) & ~((32) - 1));
            }
            while (0);
    frame->i_poc = -1;
    frame->i_type = 0x0000;
    frame->i_qpplus1 = 0;
    frame->i_pts = -1;
    frame->i_frame = -1;
    frame->i_frame_num = -1;
    frame->i_lines_completed = -1;
    frame->b_fdec = b_fdec;
    frame->i_pic_struct = PIC_STRUCT_AUTO;
    frame->i_field_cnt = -1;
    frame->i_duration = frame->i_cpb_duration = frame->i_dpb_output_delay = frame->i_cpb_delay = 0;
    frame->i_coded_fields_lookahead = frame->i_cpb_delay_lookahead = -1;
    frame->orig = frame;
    if (i_csp == 0x0003 || i_csp == 0x0006) {
        int chroma_padv = i_padv >> (i_csp == 0x0003);
        int chroma_plane_size = (frame->i_stride[1] * (frame->i_lines[1] + 2 * chroma_padv));
        do {
            frame->buffer[1] = (void *) prealloc_size;
            preallocs[prealloc_idx++] = (uint8_t **) &frame->buffer[1];
            prealloc_size += (((chroma_plane_size * sizeof (pixel)) + ((32) - 1)) & ~((32) - 1));
        }
        while (0);
        if (0)
            do {
                frame->buffer_fld[1] = (void *) prealloc_size;
                preallocs[prealloc_idx++] = (uint8_t **) &frame->buffer_fld[1];
                prealloc_size += (((chroma_plane_size * sizeof (pixel)) + ((32) - 1)) & ~((32) - 1));
            }
            while (0);
    }
    for (int p = 0; p < luma_plane_count; p++) {
        int luma_plane_size = align_plane_size (frame->i_stride[p] * (frame->i_lines[p] + 2 * i_padv), disalign);
        if (h->param.analyse.i_subpel_refine && b_fdec) {
            do {
                frame->buffer[p] = (void *) prealloc_size;
                preallocs[prealloc_idx++] = (uint8_t **) &frame->buffer[p];
                prealloc_size += (((4 * luma_plane_size * sizeof (pixel)) + ((32) - 1)) & ~((32) - 1));
            }
            while (0);
            if (0)
                do {
                    frame->buffer_fld[p] = (void *) prealloc_size;
                    preallocs[prealloc_idx++] = (uint8_t **) &frame->buffer_fld[p];
                    prealloc_size += (((4 * luma_plane_size * sizeof (pixel)) + ((32) - 1)) & ~((32) - 1));
                }
                while (0);
        }
        else {
            do {
                frame->buffer[p] = (void *) prealloc_size;
                preallocs[prealloc_idx++] = (uint8_t **) &frame->buffer[p];
                prealloc_size += (((luma_plane_size * sizeof (pixel)) + ((32) - 1)) & ~((32) - 1));
            }
            while (0);
            if (0)
                do {
                    frame->buffer_fld[p] = (void *) prealloc_size;
                    preallocs[prealloc_idx++] = (uint8_t **) &frame->buffer_fld[p];
                    prealloc_size += (((luma_plane_size * sizeof (pixel)) + ((32) - 1)) & ~((32) - 1));
                }
                while (0);
        }
    }
    frame->b_duplicate = 0;
    if (b_fdec) {
        do {
            frame->mb_type = (void *) prealloc_size;
            preallocs[prealloc_idx++] = (uint8_t **) &frame->mb_type;
            prealloc_size += (((i_mb_count * sizeof (int8_t)) + ((32) - 1)) & ~((32) - 1));
        }
        while (0);
        do {
            frame->mb_partition = (void *) prealloc_size;
            preallocs[prealloc_idx++] = (uint8_t **) &frame->mb_partition;
            prealloc_size += (((i_mb_count * sizeof (uint8_t)) + ((32) - 1)) & ~((32) - 1));
        }
        while (0);
        do {
            frame->mv[0] = (void *) prealloc_size;
            preallocs[prealloc_idx++] = (uint8_t **) &frame->mv[0];
            prealloc_size += (((2 * 16 * i_mb_count * sizeof (int16_t)) + ((32) - 1)) & ~((32) - 1));
        }
        while (0);
        do {
            frame->mv16x16 = (void *) prealloc_size;
            preallocs[prealloc_idx++] = (uint8_t **) &frame->mv16x16;
            prealloc_size += (((2 * (i_mb_count + 1) * sizeof (int16_t)) + ((32) - 1)) & ~((32) - 1));
        }
        while (0);
        do {
            frame->ref[0] = (void *) prealloc_size;
            preallocs[prealloc_idx++] = (uint8_t **) &frame->ref[0];
            prealloc_size += (((4 * i_mb_count * sizeof (int8_t)) + ((32) - 1)) & ~((32) - 1));
        }
        while (0);
        if (h->param.i_bframe) {
            do {
                frame->mv[1] = (void *) prealloc_size;
                preallocs[prealloc_idx++] = (uint8_t **) &frame->mv[1];
                prealloc_size += (((2 * 16 * i_mb_count * sizeof (int16_t)) + ((32) - 1)) & ~((32) - 1));
            }
            while (0);
            do {
                frame->ref[1] = (void *) prealloc_size;
                preallocs[prealloc_idx++] = (uint8_t **) &frame->ref[1];
                prealloc_size += (((4 * i_mb_count * sizeof (int8_t)) + ((32) - 1)) & ~((32) - 1));
            }
            while (0);
        }
        else {
            frame->mv[1] = ((void *) 0);
            frame->ref[1] = ((void *) 0);
        }
        do {
            frame->i_row_bits = (void *) prealloc_size;
            preallocs[prealloc_idx++] = (uint8_t **) &frame->i_row_bits;
            prealloc_size += (((i_lines / 16 * sizeof (int)) + ((32) - 1)) & ~((32) - 1));
        }
        while (0);
        do {
            frame->f_row_qp = (void *) prealloc_size;
            preallocs[prealloc_idx++] = (uint8_t **) &frame->f_row_qp;
            prealloc_size += (((i_lines / 16 * sizeof (float)) + ((32) - 1)) & ~((32) - 1));
        }
        while (0);
        do {
            frame->f_row_qscale = (void *) prealloc_size;
            preallocs[prealloc_idx++] = (uint8_t **) &frame->f_row_qscale;
            prealloc_size += (((i_lines / 16 * sizeof (float)) + ((32) - 1)) & ~((32) - 1));
        }
        while (0);
        if (h->param.analyse.i_me_method >= 3)
            do {
                frame->buffer[3] = (void *) prealloc_size;
                preallocs[prealloc_idx++] = (uint8_t **) &frame->buffer[3];
                prealloc_size += (((frame->i_stride[0] * (frame->i_lines[0] + 2 * i_padv) * sizeof (uint16_t) << h->frames.b_have_sub8x8_esa) + ((32) - 1)) & ~((32) - 1));
            }
            while (0);
        if (0)
            do {
                frame->field = (void *) prealloc_size;
                preallocs[prealloc_idx++] = (uint8_t **) &frame->field;
                prealloc_size += (((i_mb_count * sizeof (uint8_t)) + ((32) - 1)) & ~((32) - 1));
            }
            while (0);
        if (h->param.analyse.b_mb_info)
            do {
                frame->effective_qp = (void *) prealloc_size;
                preallocs[prealloc_idx++] = (uint8_t **) &frame->effective_qp;
                prealloc_size += (((i_mb_count * sizeof (uint8_t)) + ((32) - 1)) & ~((32) - 1));
            }
            while (0);
    }
    else {
        if (h->frames.b_have_lowres) {
            int luma_plane_size = align_plane_size (frame->i_stride_lowres * (frame->i_lines[0] / 2 + 2 * 32), disalign);
            do {
                frame->buffer_lowres[0] = (void *) prealloc_size;
                preallocs[prealloc_idx++] = (uint8_t **) &frame->buffer_lowres[0];
                prealloc_size += (((4 * luma_plane_size * sizeof (pixel)) + ((32) - 1)) & ~((32) - 1));
            }
            while (0);
            for (int j = 0; j <= !!h->param.i_bframe; j++)
                for (int i = 0; i <= h->param.i_bframe; i++) {
                    do {
                        frame->lowres_mvs[j][i] = (void *) prealloc_size;
                        preallocs[prealloc_idx++] = (uint8_t **) &frame->lowres_mvs[j][i];
                        prealloc_size += (((2 * h->mb.i_mb_count * sizeof (int16_t)) + ((32) - 1)) & ~((32) - 1));
                    }
                    while (0);
                    do {
                        frame->lowres_mv_costs[j][i] = (void *) prealloc_size;
                        preallocs[prealloc_idx++] = (uint8_t **) &frame->lowres_mv_costs[j][i];
                        prealloc_size += (((h->mb.i_mb_count * sizeof (int)) + ((32) - 1)) & ~((32) - 1));
                    }
                    while (0);
                }
            do {
                frame->i_propagate_cost = (void *) prealloc_size;
                preallocs[prealloc_idx++] = (uint8_t **) &frame->i_propagate_cost;
                prealloc_size += ((((i_mb_count + 7) * sizeof (uint16_t)) + ((32) - 1)) & ~((32) - 1));
            }
            while (0);
            for (int j = 0; j <= h->param.i_bframe + 1; j++)
                for (int i = 0; i <= h->param.i_bframe + 1; i++)
                    do {
                        frame->lowres_costs[j][i] = (void *) prealloc_size;
                        preallocs[prealloc_idx++] = (uint8_t **) &frame->lowres_costs[j][i];
                        prealloc_size += ((((i_mb_count + 3) * sizeof (uint16_t)) + ((32) - 1)) & ~((32) - 1));
                    }
                    while (0);
        }
        if (h->param.rc.i_aq_mode) {
            do {
                frame->f_qp_offset = (void *) prealloc_size;
                preallocs[prealloc_idx++] = (uint8_t **) &frame->f_qp_offset;
                prealloc_size += (((h->mb.i_mb_count * sizeof (float)) + ((32) - 1)) & ~((32) - 1));
            }
            while (0);
            do {
                frame->f_qp_offset_aq = (void *) prealloc_size;
                preallocs[prealloc_idx++] = (uint8_t **) &frame->f_qp_offset_aq;
                prealloc_size += (((h->mb.i_mb_count * sizeof (float)) + ((32) - 1)) & ~((32) - 1));
            }
            while (0);
            if (h->frames.b_have_lowres)
                do {
                    frame->i_inv_qscale_factor = (void *) prealloc_size;
                    preallocs[prealloc_idx++] = (uint8_t **) &frame->i_inv_qscale_factor;
                    prealloc_size += ((((h->mb.i_mb_count + 3) * sizeof (uint16_t)) + ((32) - 1)) & ~((32) - 1));
                }
                while (0);
        }
    }
    do {
        do {
            frame->base = x264_malloc (prealloc_size);
            if (!frame->base)
                goto fail;
        }
        while (0);
        while (prealloc_idx--)
            *preallocs[prealloc_idx] += (intptr_t) frame->base;
    }
    while (0);
    if (i_csp == 0x0003 || i_csp == 0x0006) {
        int chroma_padv = i_padv >> (i_csp == 0x0003);
        frame->plane[1] = frame->buffer[1] + frame->i_stride[1] * chroma_padv + 32;
        if (0)
            frame->plane_fld[1] = frame->buffer_fld[1] + frame->i_stride[1] * chroma_padv + 32;
    }
    for (int p = 0; p < luma_plane_count; p++) {
        int luma_plane_size = align_plane_size (frame->i_stride[p] * (frame->i_lines[p] + 2 * i_padv), disalign);
        if (h->param.analyse.i_subpel_refine && b_fdec) {
            for (int i = 0; i < 4; i++) {
                frame->filtered[p][i] = frame->buffer[p] + i * luma_plane_size + frame->i_stride[p] * i_padv + 32;
                frame->filtered_fld[p][i] = frame->buffer_fld[p] + i * luma_plane_size + frame->i_stride[p] * i_padv + 32;
            }
            frame->plane[p] = frame->filtered[p][0];
            frame->plane_fld[p] = frame->filtered_fld[p][0];
        }
        else {
            frame->filtered[p][0] = frame->plane[p] = frame->buffer[p] + frame->i_stride[p] * i_padv + 32;
            frame->filtered_fld[p][0] = frame->plane_fld[p] = frame->buffer_fld[p] + frame->i_stride[p] * i_padv + 32;
        }
    }
    if (b_fdec) {
        (((x264_union32_t *) (frame->mv16x16[0]))->i) = 0;
        frame->mv16x16++;
        if (h->param.analyse.i_me_method >= 3)
            frame->integral = (uint16_t *) frame->buffer[3] + frame->i_stride[0] * i_padv + 32;
    }
    else {
        if (h->frames.b_have_lowres) {
            int luma_plane_size = align_plane_size (frame->i_stride_lowres * (frame->i_lines[0] / 2 + 2 * 32), disalign);
            for (int i = 0; i < 4; i++)
                frame->lowres[i] = frame->buffer_lowres[0] + (frame->i_stride_lowres * 32 + 32) + i * luma_plane_size;
            for (int j = 0; j <= !!h->param.i_bframe; j++)
                for (int i = 0; i <= h->param.i_bframe; i++)
                    memset (frame->lowres_mvs[j][i], 0, 2 * h->mb.i_mb_count * sizeof (int16_t));
            frame->i_intra_cost = frame->lowres_costs[0][0];
            memset (frame -> i_intra_cost, - 1, (i_mb_count + 3) * sizeof (uint16_t));
            if (h->param.rc.i_aq_mode)
                memset (frame->i_inv_qscale_factor, 0, (h->mb.i_mb_count + 3) * sizeof (uint16_t));
        }
    }
    if (0)
        goto fail;
    if (0)
        goto fail;
    frame->opencl.ocl = h->opencl.ocl;
    return frame;
fail :
    x264_free (frame);
    return ((void *) 0);
}

void x264_frame_delete (x264_frame_t *frame) {
    if (!frame->b_duplicate) {
        x264_free (frame -> base);
        if (frame->param && frame->param->param_free)
            frame->param->param_free (frame->param);
        if (frame->mb_info_free)
            frame->mb_info_free (frame->mb_info);
        if (frame->extra_sei.sei_free) {
            for (int i = 0; i < frame->extra_sei.num_payloads; i++)
                frame->extra_sei.sei_free (frame->extra_sei.payloads[i].payload);
            frame->extra_sei.sei_free (frame->extra_sei.payloads);
        };
        ;
        x264_opencl_frame_delete (frame);
    }
    x264_free (frame);
}

int get_plane_ptr (x264_t *h, x264_picture_t *src, uint8_t **pix, int *stride, int plane, int xshift, int yshift) {
    int width = h->param.i_width >> xshift;
    int height = h->param.i_height >> yshift;
    *pix = src->img.plane[plane];
    *stride = src->img.i_stride[plane];
    if (src->img.i_csp & 0x1000) {
        *pix += (height - 1) * *stride;
        *stride = -*stride;
    }
    if (width > abs (*stride)) {
        x264_log (h, 0, "Input picture width (%d) is greater than stride (%d)\n", width, * stride);
        return -1;
    }
    return 0;
}

int x264_frame_copy_picture (x264_t *h, x264_frame_t *dst, x264_picture_t *src) {
    int i_csp = src->img.i_csp & 0x00ff;
    if (dst->i_csp != x264_frame_internal_csp (i_csp)) {
        x264_log (h, 0, "Invalid input colorspace\n");
        return -1;
    }
    if (src->img.i_csp & 0x2000) {
        x264_log (h, 0, "This build of x264 requires 8-bit input. Rebuild to support high depth input.\n");
        return -1;
    }
    if (8 != 10 && i_csp == 0x0007) {
        x264_log (h, 0, "v210 input is only compatible with bit-depth of 10 bits\n");
        return -1;
    }
    dst->i_type = src->i_type;
    dst->i_qpplus1 = src->i_qpplus1;
    dst->i_pts = dst->i_reordered_pts = src->i_pts;
    dst->param = src->param;
    dst->i_pic_struct = src->i_pic_struct;
    dst->extra_sei = src->extra_sei;
    dst->opaque = src->opaque;
    dst->mb_info = h->param.analyse.b_mb_info ? src->prop.mb_info : ((void *) 0);
    dst->mb_info_free = h->param.analyse.b_mb_info ? src->prop.mb_info_free : ((void *) 0);
    uint8_t *pix [3];
    int stride [3];
    if (i_csp == 0x0007) {
        stride[0] = src->img.i_stride[0];
        pix[0] = src->img.plane[0];
        h->mc.plane_copy_deinterleave_v210 (dst->plane[0], dst->i_stride[0], dst->plane[1], dst->i_stride[1], (uint32_t *) pix[0], stride[0] / sizeof (uint32_t), h->param.i_width, h->param.i_height);
    }
    else if (i_csp >= 0x000a) {
        stride[0] = src->img.i_stride[0];
        pix[0] = src->img.plane[0];
        if (src->img.i_csp & 0x1000) {
            pix[0] += (h->param.i_height - 1) * stride[0];
            stride[0] = -stride[0];
        }
        int b = i_csp == 0x000c;
        h->mc.plane_copy_deinterleave_rgb (dst->plane[1 + b], dst->i_stride[1 + b], dst->plane[0], dst->i_stride[0], dst->plane[2 - b], dst->i_stride[2 - b], (pixel *) pix[0], stride[0] / sizeof (pixel), i_csp == 0x000b ? 4 : 3, h->param.i_width, h->param.i_height);
    }
    else {
        int v_shift = h->mb.chroma_v_shift;
        do {
            if (get_plane_ptr (h, src, &pix[0], &stride[0], 0, 0, 0) < 0)
                return -1;
        }
        while (0);
        h->mc.plane_copy (dst->plane[0], dst->i_stride[0], (pixel *) pix[0], stride[0] / sizeof (pixel), h->param.i_width, h->param.i_height);
        if (i_csp == 0x0003 || i_csp == 0x0006) {
            do {
                if (get_plane_ptr (h, src, &pix[1], &stride[1], 1, 0, v_shift) < 0)
                    return -1;
            }
            while (0);
            h->mc.plane_copy (dst->plane[1], dst->i_stride[1], (pixel *) pix[1], stride[1] / sizeof (pixel), h->param.i_width, h->param.i_height >> v_shift);
        }
        else if (i_csp == 0x0001 || i_csp == 0x0004 || i_csp == 0x0002 || i_csp == 0x0005) {
            int uv_swap = i_csp == 0x0002 || i_csp == 0x0005;
            do {
                if (get_plane_ptr (h, src, &pix[1], &stride[1], uv_swap ? 2 : 1, 1, v_shift) < 0)
                    return -1;
            }
            while (0);
            do {
                if (get_plane_ptr (h, src, &pix[2], &stride[2], uv_swap ? 1 : 2, 1, v_shift) < 0)
                    return -1;
            }
            while (0);
            h->mc.plane_copy_interleave (dst->plane[1], dst->i_stride[1], (pixel *) pix[1], stride[1] / sizeof (pixel), (pixel *) pix[2], stride[2] / sizeof (pixel), h->param.i_width >> 1, h->param.i_height >> v_shift);
        }
        else {
            do {
                if (get_plane_ptr (h, src, &pix[1], &stride[1], i_csp == 0x0008 ? 1 : 2, 0, 0) < 0)
                    return -1;
            }
            while (0);
            do {
                if (get_plane_ptr (h, src, &pix[2], &stride[2], i_csp == 0x0008 ? 2 : 1, 0, 0) < 0)
                    return -1;
            }
            while (0);
            h->mc.plane_copy (dst->plane[1], dst->i_stride[1], (pixel *) pix[1], stride[1] / sizeof (pixel), h->param.i_width, h->param.i_height);
            h->mc.plane_copy (dst->plane[2], dst->i_stride[2], (pixel *) pix[2], stride[2] / sizeof (pixel), h->param.i_width, h->param.i_height);
        }
    }
    return 0;
}

void pixel_memset (pixel *dst, pixel *src, int len, int size) {
    uint8_t *dstp = (uint8_t *) dst;
    uint32_t v1 = *src;
    uint32_t v2 = size == 1 ? v1 + (v1 << 8) : (((x264_union16_t *) (src))->i);
    uint32_t v4 = size <= 2 ? v2 + (v2 << 16) : (((x264_union32_t *) (src))->i);
    int i = 0;
    len *= size;
    if ((intptr_t) dstp & (sizeof (void *) - 1)) {
        if (size <= 2 && ((intptr_t) dstp & 3)) {
            if (size == 1 && ((intptr_t) dstp & 1))
                dstp[i++] = v1;
            if ((intptr_t) dstp & 2) {
                (((x264_union16_t *) (dstp + i))->i) = v2;
                i += 2;
            }
        }
        if (sizeof (void *) == 8 && (intptr_t) dstp & 4) {
            (((x264_union32_t *) (dstp + i))->i) = v4;
            i += 4;
        }
    }
    if (sizeof (void *) == 8) {
        uint64_t v8 = v4 + ((uint64_t) v4 << 32);
        for (; i < len - 7; i += 8)
            (((x264_union64_t *) (dstp + i))->i) = v8;
    }
    for (; i < len - 3; i += 4)
        (((x264_union32_t *) (dstp + i))->i) = v4;
    if (size <= 2) {
        if (i < len - 1) {
            (((x264_union16_t *) (dstp + i))->i) = v2;
            i += 2;
        }
        if (size == 1 && i != len)
            dstp[i] = v1;
    }
}

void plane_expand_border (pixel *pix, int i_stride, int i_width, int i_height, int i_padh, int i_padv, int b_pad_top, int b_pad_bottom, int b_chroma) {
    for (int y = 0; y < i_height; y++) {
        pixel_memset ((pix + (- i_padh) + (y) * i_stride), (pix + (0) + (y) * i_stride), i_padh >> b_chroma, sizeof (pixel) << b_chroma);
        pixel_memset ((pix + (i_width) + (y) * i_stride), (pix + (i_width - 1 - b_chroma) + (y) * i_stride), i_padh >> b_chroma, sizeof (pixel) << b_chroma);
    }
    if (b_pad_top)
        for (int y = 0; y < i_padv; y++)
            memcpy ((pix + (-i_padh) + (-y - 1) * i_stride), (pix + (-i_padh) + (0) * i_stride), (i_width + 2 * i_padh) * sizeof (pixel));
    if (b_pad_bottom)
        for (int y = 0; y < i_padv; y++)
            memcpy ((pix + (-i_padh) + (i_height + y) * i_stride), (pix + (-i_padh) + (i_height - 1) * i_stride), (i_width + 2 * i_padh) * sizeof (pixel));
}

void x264_frame_expand_border (x264_t *h, x264_frame_t *frame, int mb_y) {
    int pad_top = mb_y == 0;
    int pad_bot = mb_y == h->mb.i_mb_height - (1 << 0);
    int b_start = mb_y == h->i_threadslice_start;
    int b_end = mb_y == h->i_threadslice_end - (1 << 0);
    if (mb_y & 0)
        return;
    for (int i = 0; i < frame->i_plane; i++) {
        int h_shift = i && h->mb.chroma_h_shift;
        int v_shift = i && h->mb.chroma_v_shift;
        int stride = frame->i_stride[i];
        int width = 16 * h->mb.i_mb_width;
        int height = (pad_bot ? 16 * (h->mb.i_mb_height - mb_y) >> 0 : 16) >> v_shift;
        int padh = 32;
        int padv = 32 >> v_shift;
        if (b_end && !b_start)
            height += 4 >> (v_shift + 0);
        pixel *pix;
        int starty = 16 * mb_y - 4 * !b_start;
        if (0) {
            pix = frame->plane_fld[i] + (starty * stride >> v_shift);
            plane_expand_border (pix, stride * 2, width, height, padh, padv, pad_top, pad_bot, h_shift);
            plane_expand_border (pix + stride, stride * 2, width, height, padh, padv, pad_top, pad_bot, h_shift);
            height = (pad_bot ? 16 * (h->mb.i_mb_height - mb_y) : 32) >> v_shift;
            if (b_end && !b_start)
                height += 4 >> v_shift;
            pix = frame->plane[i] + (starty * stride >> v_shift);
            plane_expand_border (pix, stride, width, height, padh, padv, pad_top, pad_bot, h_shift);
        }
        else {
            pix = frame->plane[i] + (starty * stride >> v_shift);
            plane_expand_border (pix, stride, width, height, padh, padv, pad_top, pad_bot, h_shift);
        }
    }
}

void x264_frame_expand_border_filtered (x264_t *h, x264_frame_t *frame, int mb_y, int b_end) {
    int b_start = !mb_y;
    int width = 16 * h->mb.i_mb_width + 8;
    int height = b_end ? (16 * (h->mb.i_mb_height - mb_y) >> 0) + 16 : 16;
    int padh = 32 - 4;
    int padv = 32 - 8;
    for (int p = 0; p < ((h->sps->i_chroma_format_idc == CHROMA_444) ? 3 : 1); p++)
        for (int i = 1; i < 4; i++) {
            int stride = frame->i_stride[p];
            pixel *pix;
            if (0) {
                pix = frame->filtered_fld[p][i] + (16 * mb_y - 16) * stride - 4;
                plane_expand_border (pix, stride * 2, width, height, padh, padv, b_start, b_end, 0);
                plane_expand_border (pix + stride, stride * 2, width, height, padh, padv, b_start, b_end, 0);
            }
            pix = frame->filtered[p][i] + (16 * mb_y - 8) * stride - 4;
            plane_expand_border (pix, stride, width, height << 0, padh, padv, b_start, b_end, 0);
        }
}

void x264_frame_expand_border_lowres (x264_frame_t *frame) {
    for (int i = 0; i < 4; i++)
        plane_expand_border (frame->lowres[i], frame->i_stride_lowres, frame->i_width_lowres, frame->i_lines_lowres, 32, 32, 1, 1, 0);
}

void x264_frame_expand_border_chroma (x264_t *h, x264_frame_t *frame, int plane) {
    int v_shift = h->mb.chroma_v_shift;
    plane_expand_border (frame -> plane [plane], frame -> i_stride [plane], 16 * h -> mb.i_mb_width, 16 * h -> mb.i_mb_height >> v_shift, 32, 32 >> v_shift, 1, 1, h -> mb.chroma_h_shift);
}

void x264_frame_expand_border_mod16 (x264_t *h, x264_frame_t *frame) {
    for (int i = 0; i < frame->i_plane; i++) {
        int i_width = h->param.i_width;
        int h_shift = i && h->mb.chroma_h_shift;
        int v_shift = i && h->mb.chroma_v_shift;
        int i_height = h->param.i_height >> v_shift;
        int i_padx = (h->mb.i_mb_width * 16 - h->param.i_width);
        int i_pady = (h->mb.i_mb_height * 16 - h->param.i_height) >> v_shift;
        if (i_padx) {
            for (int y = 0; y < i_height; y++)
                pixel_memset (&frame->plane[i][y * frame->i_stride[i] + i_width], &frame->plane[i][y * frame->i_stride[i] + i_width - 1 - h_shift], i_padx >> h_shift, sizeof (pixel) << h_shift);
        }
        if (i_pady) {
            for (int y = i_height; y < i_height + i_pady; y++)
                memcpy (&frame->plane[i][y * frame->i_stride[i]], &frame->plane[i][(i_height - (~y & 0) - 1) * frame->i_stride[i]], (i_width + i_padx) * sizeof (pixel));
        }
    }
}

void x264_expand_border_mbpair (x264_t *h, int mb_x, int mb_y) {
    for (int i = 0; i < h->fenc->i_plane; i++) {
        int v_shift = i && h->mb.chroma_v_shift;
        int stride = h->fenc->i_stride[i];
        int height = h->param.i_height >> v_shift;
        int pady = (h->mb.i_mb_height * 16 - h->param.i_height) >> v_shift;
        pixel *fenc = h->fenc->plane[i] + 16 * mb_x;
        for (int y = height; y < height + pady; y++)
            memcpy (fenc +y * stride, fenc +(height - 1) * stride, 16 * sizeof (pixel));
    }
}

void x264_frame_cond_broadcast (x264_frame_t *frame, int i_lines_completed) {
    frame->i_lines_completed = i_lines_completed;
}

void x264_frame_cond_wait (x264_frame_t *frame, int i_lines_completed) {
    while (frame->i_lines_completed < i_lines_completed)
        ;
}

void x264_threadslice_cond_broadcast (x264_t *h, int pass) {
    h->i_threadslice_pass = pass;
    if (pass > 0)
        ;
}

void x264_threadslice_cond_wait (x264_t *h, int pass) {
    while (h->i_threadslice_pass < pass)
        ;
}

int x264_frame_new_slice (x264_t *h, x264_frame_t *frame) {
    if (h->param.i_slice_count_max) {
        int slice_count;
        if (h->param.b_sliced_threads)
            slice_count = x264_pthread_fetch_and_add (&frame->i_slice_count, 1, &frame->mutex);
        else
            slice_count = frame->i_slice_count++;
        if (slice_count >= h->param.i_slice_count_max)
            return -1;
    }
    return 0;
}

void x264_frame_push (x264_frame_t **list, x264_frame_t *frame) {
    int i = 0;
    while (list[i])
        i++;
    list[i] = frame;
}

x264_frame_t *x264_frame_pop (x264_frame_t **list) {
    x264_frame_t *frame;
    int i = 0;
    ((list[0]) ? (void) (0) : __assert_fail ("list[0]", "DonorSourceCode/frame.c", 713, __PRETTY_FUNCTION__));
    while (list[i + 1])
        i++;
    frame = list[i];
    list[i] = ((void *) 0);
    return frame;
}

void x264_frame_unshift (x264_frame_t **list, x264_frame_t *frame) {
    int i = 0;
    while (list[i])
        i++;
    while (i--)
        list[i + 1] = list[i];
    list[0] = frame;
}

x264_frame_t *x264_frame_shift (x264_frame_t **list) {
    x264_frame_t *frame = list[0];
    int i;
    for (i = 0; list[i]; i++)
        list[i] = list[i + 1];
    ((frame) ? (void) (0) : __assert_fail ("frame", "DonorSourceCode/frame.c", 735, __PRETTY_FUNCTION__));
    return frame;
}

void x264_frame_push_unused (x264_t *h, x264_frame_t *frame) {
    ((frame->i_reference_count > 0) ? (void) (0) : __assert_fail ("frame->i_reference_count > 0", "DonorSourceCode/frame.c", 741, __PRETTY_FUNCTION__));
    frame->i_reference_count--;
    if (frame->i_reference_count == 0)
        x264_frame_push (h->frames.unused[frame->b_fdec], frame);
}

x264_frame_t *x264_frame_pop_unused (x264_t *h, int b_fdec) {
    x264_frame_t *frame;
    if (h->frames.unused[b_fdec][0])
        frame = x264_frame_pop (h->frames.unused[b_fdec]);
    else
        frame = x264_frame_new (h, b_fdec);
    if (!frame)
        return ((void *) 0);
    frame->b_last_minigop_bframe = 0;
    frame->i_reference_count = 1;
    frame->b_intra_calculated = 0;
    frame->b_scenecut = 1;
    frame->b_keyframe = 0;
    frame->b_corrupt = 0;
    frame->i_slice_count = h->param.b_sliced_threads ? h->param.i_threads : 1;
    memset (frame -> weight, 0, sizeof (frame -> weight));
    memset (frame -> f_weighted_cost_delta, 0, sizeof (frame -> f_weighted_cost_delta));
    return frame;
}

void x264_frame_push_blank_unused (x264_t *h, x264_frame_t *frame) {
    ((frame->i_reference_count > 0) ? (void) (0) : __assert_fail ("frame->i_reference_count > 0", "DonorSourceCode/frame.c", 772, __PRETTY_FUNCTION__));
    frame->i_reference_count--;
    if (frame->i_reference_count == 0)
        x264_frame_push (h->frames.blank_unused, frame);
}

x264_frame_t *x264_frame_pop_blank_unused (x264_t *h) {
    x264_frame_t *frame;
    if (h->frames.blank_unused[0])
        frame = x264_frame_pop (h->frames.blank_unused);
    else
        frame = x264_malloc (sizeof (x264_frame_t));
    if (!frame)
        return ((void *) 0);
    frame->b_duplicate = 1;
    frame->i_reference_count = 1;
    return frame;
}

void x264_weight_scale_plane (x264_t *h, pixel *dst, intptr_t i_dst_stride, pixel *src, intptr_t i_src_stride, int i_width, int i_height, x264_weight_t *w) {
    while (i_height > 0) {
        int x;
        for (x = 0; x < i_width - 8; x += 16)
            w->weightfn[16 >> 2] (dst +x, i_dst_stride, src +x, i_src_stride, w, ((i_height) < (16) ? (i_height) : (16)));
        if (x < i_width)
            w->weightfn[8 >> 2] (dst +x, i_dst_stride, src +x, i_src_stride, w, ((i_height) < (16) ? (i_height) : (16)));
        i_height -= 16;
        dst += 16 * i_dst_stride;
        src += 16 * i_src_stride;
    }
}

void x264_frame_delete_list (x264_frame_t **list) {
    int i = 0;
    if (!list)
        return;
    while (list[i])
        x264_frame_delete (list[i++]);
    x264_free (list);
}

int x264_sync_frame_list_init (x264_sync_frame_list_t *slist, int max_size) {
    if (max_size < 0)
        return -1;
    slist->i_max_size = max_size;
    slist->i_size = 0;
    do {
        do {
            slist->list = x264_malloc ((max_size + 1) * sizeof (x264_frame_t *));
            if (!slist->list)
                goto fail;
        }
        while (0);
        memset (slist -> list, 0, (max_size + 1) * sizeof (x264_frame_t *));
    }
    while (0);
    if (0 || 0 || 0)
        return -1;
    return 0;
fail :
    return -1;
}

void x264_sync_frame_list_delete (x264_sync_frame_list_t *slist) {
    x264_frame_delete_list (slist -> list);
}

void x264_sync_frame_list_push (x264_sync_frame_list_t *slist, x264_frame_t *frame) {
    while (slist->i_size == slist->i_max_size)
        ;
    slist->list[slist->i_size++] = frame;
}

x264_frame_t *x264_sync_frame_list_pop (x264_sync_frame_list_t *slist) {
    x264_frame_t *frame;
    while (!slist->i_size)
        ;
    frame = slist->list[--slist->i_size];
    slist->list[slist->i_size] = ((void *) 0);
    return frame;
}

