
#include "common.h"

#include "analyse.h"

void x264_lookahead_shift (x264_sync_frame_list_t *dst, x264_sync_frame_list_t *src, int count) {
    int i = count;
    while (i--) {
        ((dst->i_size < dst->i_max_size) ? (void) (0) : __assert_fail ("dst->i_size < dst->i_max_size", "DonorSourceCode/lookahead.c", 47, __PRETTY_FUNCTION__));
        ((src->i_size) ? (void) (0) : __assert_fail ("src->i_size", "DonorSourceCode/lookahead.c", 48, __PRETTY_FUNCTION__));
        dst->list[dst->i_size++] = x264_frame_shift (src->list);
        src->i_size--;
    }
    if (count) {
        ;
        ;
    }
}

void x264_lookahead_update_last_nonb (x264_t *h, x264_frame_t *new_nonb) {
    if (h->lookahead->last_nonb)
        x264_frame_push_unused (h, h->lookahead->last_nonb);
    h->lookahead->last_nonb = new_nonb;
    new_nonb->i_reference_count++;
}

int x264_lookahead_init (x264_t *h, int i_slicetype_length) {
    x264_lookahead_t *look;
    do {
        do {
            look = x264_malloc (sizeof (x264_lookahead_t));
            if (!look)
                goto fail;
        }
        while (0);
        memset (look, 0, sizeof (x264_lookahead_t));
    }
    while (0);
    for (int i = 0; i < h->param.i_threads; i++)
        h->thread[i]->lookahead = look;
    look->i_last_keyframe = -h->param.i_keyint_max;
    look->b_analyse_keyframe = (h->param.rc.b_mb_tree || (h->param.rc.i_vbv_buffer_size && h->param.rc.i_lookahead)) && !h->param.rc.b_stat_read;
    look->i_slicetype_length = i_slicetype_length;
    if (x264_sync_frame_list_init (&look->ifbuf, h->param.i_sync_lookahead + 3) || x264_sync_frame_list_init (&look->next, h->frames.i_delay + 3) || x264_sync_frame_list_init (&look->ofbuf, h->frames.i_delay + 3))
        goto fail;
    if (!h->param.i_sync_lookahead)
        return 0;
    x264_t *look_h = h->thread[h->param.i_threads];
    *look_h = *h;
    if (x264_macroblock_cache_allocate (look_h))
        goto fail;
    if (x264_macroblock_thread_allocate (look_h, 1) < 0)
        goto fail;
    if (0)
        goto fail;
    look->b_thread_active = 1;
    return 0;
fail :
    x264_free (look);
    return -1;
}

void x264_lookahead_delete (x264_t *h) {
    if (h->param.i_sync_lookahead) {
        ;
        h->lookahead->b_exit_thread = 1;
        ;
        ;
        ;
        x264_macroblock_cache_free (h -> thread [h -> param.i_threads]);
        x264_macroblock_thread_free (h -> thread [h -> param.i_threads], 1);
        x264_free (h -> thread [h -> param.i_threads]);
    }
    x264_sync_frame_list_delete (& h -> lookahead -> ifbuf);
    x264_sync_frame_list_delete (& h -> lookahead -> next);
    if (h->lookahead->last_nonb)
        x264_frame_push_unused (h, h->lookahead->last_nonb);
    x264_sync_frame_list_delete (& h -> lookahead -> ofbuf);
    x264_free (h -> lookahead);
}

void x264_lookahead_put_frame (x264_t *h, x264_frame_t *frame) {
    if (h->param.i_sync_lookahead)
        x264_sync_frame_list_push (&h->lookahead->ifbuf, frame);
    else
        x264_sync_frame_list_push (&h->lookahead->next, frame);
}

int x264_lookahead_is_empty (x264_t *h) {
    ;
    ;
    int b_empty = !h->lookahead->next.i_size && !h->lookahead->ofbuf.i_size;
    ;
    ;
    return b_empty;
}

void x264_lookahead_encoder_shift (x264_t *h) {
    if (!h->lookahead->ofbuf.i_size)
        return;
    int i_frames = h->lookahead->ofbuf.list[0]->i_bframes + 1;
    while (i_frames--) {
        x264_frame_push (h -> frames.current, x264_frame_shift (h -> lookahead -> ofbuf.list));
        h->lookahead->ofbuf.i_size--;
    };
}

void x264_lookahead_get_frames (x264_t *h) {
    if (h->param.i_sync_lookahead) {
        ;
        while (!h->lookahead->ofbuf.i_size && h->lookahead->b_thread_active)
            ;
        x264_lookahead_encoder_shift (h);
        ;
    }
    else {
        if (h->frames.current[0] || !h->lookahead->next.i_size)
            return;
        x264_slicetype_decide (h);
        x264_lookahead_update_last_nonb (h, h -> lookahead -> next.list [0]);
        int shift_frames = h->lookahead->next.list[0]->i_bframes + 1;
        x264_lookahead_shift (& h -> lookahead -> ofbuf, & h -> lookahead -> next, shift_frames);
        if (h->lookahead->b_analyse_keyframe && ((h->lookahead->last_nonb->i_type) == 0x0002 || (h->lookahead->last_nonb->i_type) == 0x0001))
            x264_slicetype_analyse (h, shift_frames);
        x264_lookahead_encoder_shift (h);
    }
}

