/*****************************************************************************
 * macroblock.h: macroblock common functions
 *****************************************************************************
 * Copyright (C) 2005-2014 x264 project
 *
 * Authors: Loren Merritt <lorenm@u.washington.edu>
 *          Laurent Aimar <fenrir@via.ecp.fr>
 *          Fiona Glaser <fiona@x264.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 * This program is also available under a commercial proprietary license.
 * For more information, contact us at licensing@x264.com.
 *****************************************************************************/

#ifndef X264_MACROBLOCK_H
#define X264_MACROBLOCK_H

#include "common.h"

/* XXX mb_type isn't the one written in the bitstream -> only internal usage */
#define IS_INTRA(type) ( (type) == I_4x4 || (type) == I_8x8 || (type) == I_16x16 || (type) == I_PCM )
#define IS_SKIP(type)  ( (type) == P_SKIP || (type) == B_SKIP )
#define IS_DIRECT(type)  ( (type) == B_DIRECT )

#define IS_SUB4x4(type) ( (type ==D_L0_4x4)||(type ==D_L1_4x4)||(type ==D_BI_4x4))
#define IS_SUB4x8(type) ( (type ==D_L0_4x8)||(type ==D_L1_4x8)||(type ==D_BI_4x8))
#define IS_SUB8x4(type) ( (type ==D_L0_8x4)||(type ==D_L1_8x4)||(type ==D_BI_8x4))
#define IS_SUB8x8(type) ( (type ==D_L0_8x8)||(type ==D_L1_8x8)||(type ==D_BI_8x8)||(type ==D_DIRECT_8x8))

#define pack_pixel_1to2 pack8to16
#define pack_pixel_2to4 pack16to32


extern const uint8_t x264_mb_type_fix[19];
extern const uint8_t x264_mb_type_list_table[19][2][2];
extern const uint8_t x264_pred_i4x4_neighbors[12];
extern const uint8_t x264_mb_partition_listX_table[2][17];
extern const uint8_t x264_mb_partition_count_table[17];
extern const uint8_t x264_mb_partition_pixel_table[17];
extern const uint8_t x264_zigzag_scan4[2][16];
extern const uint8_t x264_zigzag_scan8[2][64];
extern const uint8_t block_idx_x[16];
extern const uint8_t block_idx_y[16];
extern const uint8_t block_idx_xy[4][4];
extern const uint8_t block_idx_xy_1d[16];
extern const uint8_t block_idx_yx_1d[16];
extern const uint8_t block_idx_xy_fenc[16];
extern const uint16_t block_idx_xy_fdec[16];
extern const uint8_t i_chroma_qp_table[69+1+12*2];
extern const uint8_t ctx_cat_plane[6][3];
extern const uint8_t x264_transform_allowed[19];





int x264_macroblock_cache_allocate( x264_t *h );
void x264_macroblock_cache_free( x264_t *h );
int x264_macroblock_thread_allocate( x264_t *h, int b_lookahead );
void x264_macroblock_thread_free( x264_t *h, int b_lookahead );
void x264_macroblock_slice_init( x264_t *h );
void x264_macroblock_thread_init( x264_t *h );
void x264_macroblock_cache_load_progressive( x264_t *h, int mb_x, int mb_y );
void x264_macroblock_cache_load_interlaced( x264_t *h, int mb_x, int mb_y );
void x264_macroblock_deblock_strength( x264_t *h );
void x264_macroblock_cache_save( x264_t *h );
void x264_macroblock_bipred_init( x264_t *h );
void x264_prefetch_fenc( x264_t *h, x264_frame_t *fenc, int i_mb_x, int i_mb_y );
void x264_copy_column8( pixel *dst, pixel *src );
void x264_mb_predict_mv_16x16( x264_t *h, int i_list, int i_ref, int16_t mvp[2] );
void x264_mb_predict_mv_pskip( x264_t *h, int16_t mv[2] );
void x264_mb_predict_mv( x264_t *h, int i_list, int idx, int i_width, int16_t mvp[2] );
int x264_mb_predict_mv_direct16x16( x264_t *h, int *b_changed );
void x264_mb_predict_mv_ref16x16( x264_t *h, int i_list, int i_ref, int16_t mvc[8][2], int *i_mvc );
void x264_mb_mc( x264_t *h );
void x264_mb_mc_8x8( x264_t *h, int i8 );
uint32_t pack16to32( uint32_t a, uint32_t b );
uint32_t pack8to16( uint32_t a, uint32_t b );
uint32_t pack8to32( uint32_t a, uint32_t b, uint32_t c, uint32_t d );
uint32_t pack16to32_mask( int a, int b );
uint64_t pack32to64( uint32_t a, uint32_t b );
int x264_mb_predict_intra4x4_mode( x264_t *h, int idx );
int x264_mb_predict_non_zero_code( x264_t *h, int idx );
int x264_mb_transform_8x8_allowed( x264_t *h );














#endif

