
#include "common.h"

#include "macroblockENCODER.h"

void zigzag_scan_2x2_dc (dctcoef level [4], dctcoef dct [4]) {
    level[0] = dct[0 * 2 + 0];
    level[1] = dct[1 * 2 + 0];
    level[2] = dct[0 * 2 + 1];
    level[3] = dct[1 * 2 + 1];
}

void zigzag_scan_2x4_dc (dctcoef level [8], dctcoef dct [8]) {
    level[0] = dct[0];
    level[1] = dct[2];
    level[2] = dct[1];
    level[3] = dct[4];
    level[4] = dct[6];
    level[5] = dct[3];
    level[6] = dct[5];
    level[7] = dct[7];
}

void idct_dequant_2x2_dc (dctcoef dct [4], dctcoef dct4x4 [4] [16], int dequant_mf [6] [16], int i_qp) {
    int d0 = dct[0] + dct[1];
    int d1 = dct[2] + dct[3];
    int d2 = dct[0] - dct[1];
    int d3 = dct[2] - dct[3];
    int dmf = dequant_mf[i_qp % 6][0] << i_qp / 6;
    dct4x4[0][0] = (d0 + d1) * dmf >> 5;
    dct4x4[1][0] = (d0 - d1) * dmf >> 5;
    dct4x4[2][0] = (d2 + d3) * dmf >> 5;
    dct4x4[3][0] = (d2 - d3) * dmf >> 5;
}

void idct_dequant_2x2_dconly (dctcoef dct [4], int dequant_mf [6] [16], int i_qp) {
    int d0 = dct[0] + dct[1];
    int d1 = dct[2] + dct[3];
    int d2 = dct[0] - dct[1];
    int d3 = dct[2] - dct[3];
    int dmf = dequant_mf[i_qp % 6][0] << i_qp / 6;
    dct[0] = (d0 + d1) * dmf >> 5;
    dct[1] = (d0 - d1) * dmf >> 5;
    dct[2] = (d2 + d3) * dmf >> 5;
    dct[3] = (d2 - d3) * dmf >> 5;
}

void dct2x2dc (dctcoef d [4], dctcoef dct4x4 [4] [16]) {
    int d0 = dct4x4[0][0] + dct4x4[1][0];
    int d1 = dct4x4[2][0] + dct4x4[3][0];
    int d2 = dct4x4[0][0] - dct4x4[1][0];
    int d3 = dct4x4[2][0] - dct4x4[3][0];
    d[0] = d0 + d1;
    d[2] = d2 + d3;
    d[1] = d0 - d1;
    d[3] = d2 - d3;
    dct4x4[0][0] = 0;
    dct4x4[1][0] = 0;
    dct4x4[2][0] = 0;
    dct4x4[3][0] = 0;
}

int array_non_zero (dctcoef *v, int i_count) {
    if (sizeof (void *) == 8) {
        for (int i = 0; i < i_count; i += 8 / sizeof (dctcoef))
            if ((((x264_union64_t *) (&v[i]))->i))
                return 1;
    }
    else {
        for (int i = 0; i < i_count; i += 4 / sizeof (dctcoef))
            if ((((x264_union32_t *) (&v[i]))->i))
                return 1;
    }
    return 0;
}

void x264_mb_encode_i16x16 (x264_t *h, int p, int i_qp) {
    pixel *p_src = h->mb.pic.p_fenc[p];
    pixel *p_dst = h->mb.pic.p_fdec[p];
    uint8_t dct4x4_u [sizeof (dctcoef[16][16]) + 31];
    dctcoef (*dct4x4) [16] = (void *) ((intptr_t) (dct4x4_u + 31) & ~31);
    uint8_t dct_dc4x4_u [sizeof (dctcoef[16]) + 31];
    dctcoef (*dct_dc4x4) = (void *) ((intptr_t) (dct_dc4x4_u + 31) & ~31);
    int nz, block_cbp = 0;
    int decimate_score = h->mb.b_dct_decimate ? 0 : 9;
    int i_quant_cat = p ? CQM_4IC : CQM_4IY;
    int i_mode = h->mb.i_intra16x16_pred_mode;
    if (h->mb.b_lossless)
        x264_predict_lossless_16x16 (h, p, i_mode);
    else
        h->predict_16x16[i_mode] (h->mb.pic.p_fdec[p]);
    if (h->mb.b_lossless) {
        for (int i = 0; i < 16; i++) {
            int oe = block_idx_xy_fenc[i];
            int od = block_idx_xy_fdec[i];
            nz = h->zigzagf.sub_4x4ac (h->dct.luma4x4[16 * p + i], p_src +oe, p_dst +od, &dct_dc4x4[block_idx_yx_1d[i]]);
            h->mb.cache.non_zero_count[x264_scan8[16 * p + i]] = nz;
            block_cbp |= nz;
        }
        h->mb.i_cbp_luma |= block_cbp * 0xf;
        h->mb.cache.non_zero_count[x264_scan8[48 + p]] = array_non_zero (dct_dc4x4, 16);
        h->zigzagf.scan_4x4 (h->dct.luma16x16_dc[p], dct_dc4x4);
        return;
    }
    return;
}

int x264_mb_optimize_chroma_dc (x264_t *h, dctcoef *dct_dc, int dequant_mf [6] [16], int i_qp, int chroma422) {
    int dmf = dequant_mf[i_qp % 6][0] << i_qp / 6;
    if (dmf > 32 * 64)
        return 1;
    if (chroma422)
        return h->quantf.optimize_chroma_2x4_dc (dct_dc, dmf);
    else
        return h->quantf.optimize_chroma_2x2_dc (dct_dc, dmf);
}

void x264_mb_encode_chroma_internal (x264_t *h, int b_inter, int i_qp, int chroma422) {
    int nz, nz_dc;
    int b_decimate = b_inter && h->mb.b_dct_decimate;
    int (*dequant_mf) [16] = h->dequant4_mf[CQM_4IC + b_inter];
    dctcoef dct_dc [8];
    h->mb.i_cbp_chroma = 0;
    h->nr_count[2] += h->mb.b_noise_reduction * 4;
    (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[16]]))->i) = 0;
    (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[18]]))->i) = 0;
    (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[32]]))->i) = 0;
    (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[34]]))->i) = 0;
    if (chroma422) {
        (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[24]]))->i) = 0;
        (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[26]]))->i) = 0;
        (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[40]]))->i) = 0;
        (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[42]]))->i) = 0;
    }
    for (int ch = 0; ch < 2; ch++) {
        pixel *p_src = h->mb.pic.p_fenc[1 + ch];
        pixel *p_dst = h->mb.pic.p_fdec[1 + ch];
        int i_decimate_score = b_decimate ? 0 : 7;
        int nz_ac = 0;
        uint8_t dct4x4_u [sizeof (dctcoef[8][16]) + 31];
        dctcoef (*dct4x4) [16] = (void *) ((intptr_t) (dct4x4_u + 31) & ~31);
        if (h->mb.b_lossless) {
            static const uint8_t chroma422_scan [8] = {0, 2, 1, 5, 3, 6, 4, 7};
            for (int i = 0; i < (chroma422 ? 8 : 4); i++) {
                int oe = 4 * (i & 1) + 4 * (i >> 1) * 16;
                int od = 4 * (i & 1) + 4 * (i >> 1) * 32;
                nz = h->zigzagf.sub_4x4ac (h->dct.luma4x4[16 + i + (chroma422 ? i & 4 : 0) + ch * 16], p_src +oe, p_dst +od, &h->dct.chroma_dc[ch][chroma422 ? chroma422_scan[i] : i]);
                h->mb.cache.non_zero_count[x264_scan8[16 + i + (chroma422 ? i & 4 : 0) + ch * 16]] = nz;
                h->mb.i_cbp_chroma |= nz;
            }
            h->mb.cache.non_zero_count[x264_scan8[49 + ch]] = array_non_zero (h->dct.chroma_dc[ch], chroma422 ? 8 : 4);
            continue;
        }
        for (int i = 0; i <= chroma422; i++)
            h->dctf.sub8x8_dct (&dct4x4[4 * i], p_src +8 * i * 16, p_dst +8 * i * 32);
        if (h->mb.b_noise_reduction)
            for (int i = 0; i < (chroma422 ? 8 : 4); i++)
                h->quantf.denoise_dct (dct4x4[i], h->nr_residual_sum[2], h->nr_offset[2], 16);
        if (chroma422)
            h->dctf.dct2x4dc (dct_dc, dct4x4);
        else
            dct2x2dc (dct_dc, dct4x4);
        for (int i8x8 = 0; i8x8 < (chroma422 ? 2 : 1); i8x8++) {
            if (h->mb.b_trellis) {
                for (int i4x4 = 0; i4x4 < 4; i4x4++) {
                }
            }
            else {
                nz = h->quantf.quant_4x4x4 (&dct4x4[i8x8 * 4], h->quant4_mf[CQM_4IC + b_inter][i_qp], h->quant4_bias[CQM_4IC + b_inter][i_qp]);
                nz_ac |= nz;
                for (int i4x4 = 0, msk = nz, skip; msk && (skip = x264_ctz_4bit (msk), i4x4 += skip, msk >>= skip + 1, 1); i4x4++) {
                    int idx = 16 + ch * 16 + i8x8 * 8 + i4x4;
                    h->zigzagf.scan_4x4 (h->dct.luma4x4[idx], dct4x4[i8x8 * 4 + i4x4]);
                    h->quantf.dequant_4x4 (dct4x4[i8x8 * 4 + i4x4], dequant_mf, i_qp);
                    if (i_decimate_score < 7)
                        i_decimate_score += h->quantf.decimate_score15 (h->dct.luma4x4[idx]);
                    h->mb.cache.non_zero_count[x264_scan8[idx]] = 1;
                }
            }
        }
        if (h->mb.b_trellis)
            printf ("ERROR");
        else {
            nz_dc = 0;
            for (int i = 0; i <= chroma422; i++)
                nz_dc |= h->quantf.quant_2x2_dc (&dct_dc[4 * i], h->quant4_mf[CQM_4IC + b_inter][i_qp + 3 * chroma422][0] >> 1, h->quant4_bias[CQM_4IC + b_inter][i_qp + 3 * chroma422][0] << 1);
        }
        h->mb.cache.non_zero_count[x264_scan8[49 + ch]] = nz_dc;
        if (i_decimate_score < 7 || !nz_ac) {
            (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 + 16 * ch]]))->i) = 0;
            (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[18 + 16 * ch]]))->i) = 0;
            if (chroma422) {
                (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[24 + 16 * ch]]))->i) = 0;
                (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[26 + 16 * ch]]))->i) = 0;
            }
            if (!nz_dc)
                continue;
            if (!x264_mb_optimize_chroma_dc (h, dct_dc, dequant_mf, i_qp +3 * chroma422, chroma422)) {
                h->mb.cache.non_zero_count[x264_scan8[49 + ch]] = 0;
                continue;
            }
            if (chroma422) {
                zigzag_scan_2x4_dc (h -> dct.chroma_dc [ch], dct_dc);
                h->quantf.idct_dequant_2x4_dconly (dct_dc, dequant_mf, i_qp +3);
            }
            else {
                zigzag_scan_2x2_dc (h -> dct.chroma_dc [ch], dct_dc);
                idct_dequant_2x2_dconly (dct_dc, dequant_mf, i_qp);
            }
            for (int i = 0; i <= chroma422; i++)
                h->dctf.add8x8_idct_dc (p_dst +8 * i * 32, &dct_dc[4 * i]);
        }
        else {
            h->mb.i_cbp_chroma = 1;
            if (nz_dc) {
                if (chroma422) {
                    zigzag_scan_2x4_dc (h -> dct.chroma_dc [ch], dct_dc);
                    h->quantf.idct_dequant_2x4_dc (dct_dc, dct4x4, dequant_mf, i_qp +3);
                }
                else {
                    zigzag_scan_2x2_dc (h -> dct.chroma_dc [ch], dct_dc);
                    idct_dequant_2x2_dc (dct_dc, dct4x4, dequant_mf, i_qp);
                }
            }
            for (int i = 0; i <= chroma422; i++)
                h->dctf.add8x8_idct (p_dst +8 * i * 32, &dct4x4[4 * i]);
        }
    }
    h->mb.i_cbp_chroma += (h->mb.cache.non_zero_count[x264_scan8[49 + 0]] | h->mb.cache.non_zero_count[x264_scan8[49 + 1]] | h->mb.i_cbp_chroma);
}

void x264_mb_encode_chroma (x264_t *h, int b_inter, int i_qp) {
    if (h->sps->i_chroma_format_idc == CHROMA_420)
        x264_mb_encode_chroma_internal (h, b_inter, i_qp, 0);
    else
        x264_mb_encode_chroma_internal (h, b_inter, i_qp, 1);
}

void x264_macroblock_encode_skip (x264_t *h) {
    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[0]]))->i) = 0;
    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[2]]))->i) = 0;
    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[8]]))->i) = 0;
    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[10]]))->i) = 0;
    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 + 0]]))->i) = 0;
    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 + 2]]))->i) = 0;
    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[32 + 0]]))->i) = 0;
    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[32 + 2]]))->i) = 0;
    if (h->sps->i_chroma_format_idc >= CHROMA_422) {
        (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 + 8]]))->i) = 0;
        (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 + 10]]))->i) = 0;
        (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[32 + 8]]))->i) = 0;
        (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[32 + 10]]))->i) = 0;
    }
    h->mb.i_cbp_luma = 0;
    h->mb.i_cbp_chroma = 0;
    h->mb.cbp[h->mb.i_mb_xy] = 0;
}

void x264_predict_lossless_chroma (x264_t *h, int i_mode) {
    int height = 16 >> h->mb.chroma_v_shift;
    if (i_mode == I_PRED_CHROMA_V) {
        h->mc.copy[PIXEL_8x8] (h->mb.pic.p_fdec[1], 32, h->mb.pic.p_fenc[1] - 16, 16, height);
        h->mc.copy[PIXEL_8x8] (h->mb.pic.p_fdec[2], 32, h->mb.pic.p_fenc[2] - 16, 16, height);
        memcpy (h -> mb.pic.p_fdec [1], h -> mb.pic.p_fdec [1] - 32, 8 * sizeof (pixel));
        memcpy (h -> mb.pic.p_fdec [2], h -> mb.pic.p_fdec [2] - 32, 8 * sizeof (pixel));
    }
    else if (i_mode == I_PRED_CHROMA_H) {
        h->mc.copy[PIXEL_8x8] (h->mb.pic.p_fdec[1], 32, h->mb.pic.p_fenc[1] - 1, 16, height);
        h->mc.copy[PIXEL_8x8] (h->mb.pic.p_fdec[2], 32, h->mb.pic.p_fenc[2] - 1, 16, height);
        x264_copy_column8 (h -> mb.pic.p_fdec [1] + 4 * 32, h -> mb.pic.p_fdec [1] + 4 * 32 - 1);
        x264_copy_column8 (h -> mb.pic.p_fdec [2] + 4 * 32, h -> mb.pic.p_fdec [2] + 4 * 32 - 1);
        if (h->sps->i_chroma_format_idc == CHROMA_422) {
            x264_copy_column8 (h -> mb.pic.p_fdec [1] + 12 * 32, h -> mb.pic.p_fdec [1] + 12 * 32 - 1);
            x264_copy_column8 (h -> mb.pic.p_fdec [2] + 12 * 32, h -> mb.pic.p_fdec [2] + 12 * 32 - 1);
        }
    }
    else {
        h->predict_chroma[i_mode] (h->mb.pic.p_fdec[1]);
        h->predict_chroma[i_mode] (h->mb.pic.p_fdec[2]);
    }
}

void x264_predict_lossless_4x4 (x264_t *h, pixel *p_dst, int p, int idx, int i_mode) {
    int stride = h->fenc->i_stride[p] << 0;
    pixel *p_src = h->mb.pic.p_fenc_plane[p] + block_idx_x[idx] * 4 + block_idx_y[idx] * 4 * stride;
    if (i_mode == I_PRED_4x4_V)
        h->mc.copy[PIXEL_4x4] (p_dst, 32, p_src -stride, stride, 4);
    else if (i_mode == I_PRED_4x4_H)
        h->mc.copy[PIXEL_4x4] (p_dst, 32, p_src -1, stride, 4);
    else
        h->predict_4x4[i_mode] (p_dst);
}

void x264_predict_lossless_8x8 (x264_t *h, pixel *p_dst, int p, int idx, int i_mode, pixel edge [36]) {
    int stride = h->fenc->i_stride[p] << 0;
    pixel *p_src = h->mb.pic.p_fenc_plane[p] + (idx & 1) * 8 + (idx >> 1) * 8 * stride;
    if (i_mode == I_PRED_8x8_V)
        h->mc.copy[PIXEL_8x8] (p_dst, 32, p_src -stride, stride, 8);
    else if (i_mode == I_PRED_8x8_H)
        h->mc.copy[PIXEL_8x8] (p_dst, 32, p_src -1, stride, 8);
    else
        h->predict_8x8[i_mode] (p_dst, edge);
}

void x264_predict_lossless_16x16 (x264_t *h, int p, int i_mode) {
    int stride = h->fenc->i_stride[p] << 0;
    if (i_mode == I_PRED_16x16_V)
        h->mc.copy[PIXEL_16x16] (h->mb.pic.p_fdec[p], 32, h->mb.pic.p_fenc_plane[p] - stride, stride, 16);
    else if (i_mode == I_PRED_16x16_H)
        h->mc.copy_16x16_unaligned (h->mb.pic.p_fdec[p], 32, h->mb.pic.p_fenc_plane[p] - 1, stride, 16);
    else
        h->predict_16x16[i_mode] (h->mb.pic.p_fdec[p]);
}

void x264_macroblock_encode_internal (x264_t *h, int plane_count, int chroma) {
    int i_qp = h->mb.i_qp;
    int b_decimate = h->mb.b_dct_decimate;
    int b_force_no_skip = 0;
    int nz;
    h->mb.i_cbp_luma = 0;
    for (int p = 0; p < plane_count; p++)
        h->mb.cache.non_zero_count[x264_scan8[48 + p]] = 0;
    if (h->mb.i_type == I_PCM) {
        for (int p = 0; p < plane_count; p++)
            h->mc.copy[PIXEL_16x16] (h->mb.pic.p_fdec[p], 32, h->mb.pic.p_fenc[p], 16, 16);
        if (chroma) {
            int height = 16 >> h->mb.chroma_v_shift;
            h->mc.copy[PIXEL_8x8] (h->mb.pic.p_fdec[1], 32, h->mb.pic.p_fenc[1], 16, height);
            h->mc.copy[PIXEL_8x8] (h->mb.pic.p_fdec[2], 32, h->mb.pic.p_fenc[2], 16, height);
        }
        return;
    }
    if (!h->mb.b_allow_skip) {
        b_force_no_skip = 1;
        if (((h->mb.i_type) == P_SKIP || (h->mb.i_type) == B_SKIP)) {
            if (h->mb.i_type == P_SKIP)
                h->mb.i_type = P_L0;
            else if (h->mb.i_type == B_SKIP)
                h->mb.i_type = B_DIRECT;
        }
    }
    if (h->mb.i_type == P_SKIP) {
        if (!h->mb.b_skip_mc) {
            int mvx = x264_clip3 (h->mb.cache.mv[0][x264_scan8[0]][0], h->mb.mv_min[0], h->mb.mv_max[0]);
            int mvy = x264_clip3 (h->mb.cache.mv[0][x264_scan8[0]][1], h->mb.mv_min[1], h->mb.mv_max[1]);
            for (int p = 0; p < plane_count; p++)
                h->mc.mc_luma (h->mb.pic.p_fdec[p], 32, &h->mb.pic.p_fref[0][0][p * 4], h->mb.pic.i_stride[p], mvx, mvy, 16, 16, &h->sh.weight[0][p]);
            if (chroma) {
                int v_shift = h->mb.chroma_v_shift;
                int height = 16 >> v_shift;
                if (mvx | mvy)
                    h->mc.mc_chroma (h->mb.pic.p_fdec[1], h->mb.pic.p_fdec[2], 32, h->mb.pic.p_fref[0][0][4], h->mb.pic.i_stride[1], mvx, 2 * mvy >> v_shift, 8, height);
                else
                    h->mc.load_deinterleave_chroma_fdec (h->mb.pic.p_fdec[1], h->mb.pic.p_fref[0][0][4], h->mb.pic.i_stride[1], height);
                if (h->sh.weight[0][1].weightfn)
                    h->sh.weight[0][1].weightfn[8 >> 2] (h->mb.pic.p_fdec[1], 32, h->mb.pic.p_fdec[1], 32, &h->sh.weight[0][1], height);
                if (h->sh.weight[0][2].weightfn)
                    h->sh.weight[0][2].weightfn[8 >> 2] (h->mb.pic.p_fdec[2], 32, h->mb.pic.p_fdec[2], 32, &h->sh.weight[0][2], height);
            }
        }
        x264_macroblock_encode_skip (h);
        return;
    }
    if (h->mb.i_type == B_SKIP) {
        if (!h->mb.b_skip_mc)
            x264_mb_mc (h);
        x264_macroblock_encode_skip (h);
        return;
    }
    if (h->mb.i_type == I_16x16) {
        h->mb.b_transform_8x8 = 0;
        for (int p = 0; p < plane_count; p++, i_qp = h->mb.i_chroma_qp)
            x264_mb_encode_i16x16 (h, p, i_qp);
    }
    else if (h->mb.i_type == I_8x8) {
        h->mb.b_transform_8x8 = 1;
        if (h->mb.i_skip_intra) {
            h->mc.copy[PIXEL_16x16] (h->mb.pic.p_fdec[0], 32, h->mb.pic.i8x8_fdec_buf, 16, 16);
            (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[0]]))->i) = h->mb.pic.i8x8_nnz_buf[0];
            (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[2]]))->i) = h->mb.pic.i8x8_nnz_buf[1];
            (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[8]]))->i) = h->mb.pic.i8x8_nnz_buf[2];
            (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[10]]))->i) = h->mb.pic.i8x8_nnz_buf[3];
            h->mb.i_cbp_luma = h->mb.pic.i8x8_cbp;
            if (h->mb.i_skip_intra == 2)
                h->mc.memcpy_aligned (h->dct.luma8x8, h->mb.pic.i8x8_dct_buf, sizeof (h->mb.pic.i8x8_dct_buf));
        }
        for (int p = 0; p < plane_count; p++, i_qp = h->mb.i_chroma_qp) {
            for (int i = (p == 0 && h->mb.i_skip_intra) ? 3 : 0; i < 4; i++) {
                int i_mode = h->mb.cache.intra4x4_pred_mode[x264_scan8[4 * i]];
                x264_mb_encode_i8x8 (h, p, i, i_qp, i_mode, ((void *) 0), 1);
            }
        }
    }
    else if (h->mb.i_type == I_4x4) {
        h->mb.b_transform_8x8 = 0;
        if (h->mb.i_skip_intra) {
            h->mc.copy[PIXEL_16x16] (h->mb.pic.p_fdec[0], 32, h->mb.pic.i4x4_fdec_buf, 16, 16);
            (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[0]]))->i) = h->mb.pic.i4x4_nnz_buf[0];
            (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[2]]))->i) = h->mb.pic.i4x4_nnz_buf[1];
            (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[8]]))->i) = h->mb.pic.i4x4_nnz_buf[2];
            (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[10]]))->i) = h->mb.pic.i4x4_nnz_buf[3];
            h->mb.i_cbp_luma = h->mb.pic.i4x4_cbp;
            if (h->mb.i_skip_intra == 2)
                h->mc.memcpy_aligned (h->dct.luma4x4, h->mb.pic.i4x4_dct_buf, sizeof (h->mb.pic.i4x4_dct_buf));
        }
        for (int p = 0; p < plane_count; p++, i_qp = h->mb.i_chroma_qp) {
            for (int i = (p == 0 && h->mb.i_skip_intra) ? 15 : 0; i < 16; i++) {
                pixel *p_dst = &h->mb.pic.p_fdec[p][block_idx_xy_fdec[i]];
                int i_mode = h->mb.cache.intra4x4_pred_mode[x264_scan8[i]];
                if ((h->mb.i_neighbour4[i] & (MB_TOPRIGHT | MB_TOP)) == MB_TOP)
                    (((x264_union32_t *) (&p_dst[4 - 32]))->i) = ((p_dst[3 - 32]) * 0x01010101U);
                x264_mb_encode_i4x4 (h, p, i, i_qp, i_mode, 1);
            }
        }
    }
    else {
        int i_decimate_mb = 0;
        if (!h->mb.b_skip_mc)
            x264_mb_mc (h);
        if (h->mb.b_lossless) {
            if (h->mb.b_transform_8x8)
                for (int p = 0; p < plane_count; p++)
                    for (int i8x8 = 0; i8x8 < 4; i8x8++) {
                        int x = i8x8 & 1;
                        int y = i8x8 >> 1;
                        nz = h->zigzagf.sub_8x8 (h->dct.luma8x8[p * 4 + i8x8], h->mb.pic.p_fenc[p] + 8 * x + 8 * y * 16, h->mb.pic.p_fdec[p] + 8 * x + 8 * y * 32);
                        do {
                            (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8x8 * 4] + 0]))->i) = (nz) *0x0101;
                            (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8x8 * 4] + 8]))->i) = (nz) *0x0101;
                        }
                        while (0);
                        h->mb.i_cbp_luma |= nz << i8x8;
                    }
            else
                for (int p = 0; p < plane_count; p++)
                    for (int i4x4 = 0; i4x4 < 16; i4x4++) {
                        nz = h->zigzagf.sub_4x4 (h->dct.luma4x4[p * 16 + i4x4], h->mb.pic.p_fenc[p] + block_idx_xy_fenc[i4x4], h->mb.pic.p_fdec[p] + block_idx_xy_fdec[i4x4]);
                        h->mb.cache.non_zero_count[x264_scan8[p * 16 + i4x4]] = nz;
                        h->mb.i_cbp_luma |= nz << (i4x4 >> 2);
                    }
        }
        else if (h->mb.b_transform_8x8) {
            uint8_t dct8x8_u [sizeof (dctcoef[4][64]) + 31];
            dctcoef (*dct8x8) [64] = (void *) ((intptr_t) (dct8x8_u + 31) & ~31);
            b_decimate &= !h->mb.b_trellis || !h->param.b_cabac;
            for (int p = 0; p < plane_count; p++, i_qp = h->mb.i_chroma_qp) {
                do {
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * p] + 0 * 8]))->i) = 0;
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * p] + 1 * 8]))->i) = 0;
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * p] + 2 * 8]))->i) = 0;
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * p] + 3 * 8]))->i) = 0;
                }
                while (0);
                h->dctf.sub16x16_dct8 (dct8x8, h->mb.pic.p_fenc[p], h->mb.pic.p_fdec[p]);
                h->nr_count[1 + !!p * 2] += h->mb.b_noise_reduction * 4;
                int plane_cbp = 0;
                for (int idx = 0; idx < 4; idx++) {
                    nz = x264_quant_8x8 (h, dct8x8[idx], i_qp, ctx_cat_plane[DCT_LUMA_8x8][p], 0, p, idx);
                    if (nz) {
                        h->zigzagf.scan_8x8 (h->dct.luma8x8[p * 4 + idx], dct8x8[idx]);
                        if (b_decimate) {
                            int i_decimate_8x8 = h->quantf.decimate_score64 (h->dct.luma8x8[p * 4 + idx]);
                            i_decimate_mb += i_decimate_8x8;
                            if (i_decimate_8x8 >= 4)
                                plane_cbp |= 1 << idx;
                        }
                        else
                            plane_cbp |= 1 << idx;
                    }
                }
                if (i_decimate_mb >= 6 || !b_decimate) {
                    h->mb.i_cbp_luma |= plane_cbp;
                    for (int idx = 0, msk = plane_cbp, skip; msk && (skip = x264_ctz_4bit (msk), idx += skip, msk >>= skip + 1, 1); idx++) {
                        h->quantf.dequant_8x8 (dct8x8[idx], h->dequant8_mf[p ? CQM_8PC : CQM_8PY], i_qp);
                        h->dctf.add8x8_idct8 (&h->mb.pic.p_fdec[p][8 * (idx & 1) + 8 * (idx >> 1) * 32], dct8x8[idx]);
                        do {
                            (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + idx * 4] + 0]))->i) = (1) * 0x0101;
                            (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + idx * 4] + 8]))->i) = (1) * 0x0101;
                        }
                        while (0);
                    }
                }
            }
        }
        else {
            uint8_t dct4x4_u [sizeof (dctcoef[16][16]) + 31];
            dctcoef (*dct4x4) [16] = (void *) ((intptr_t) (dct4x4_u + 31) & ~31);
            for (int p = 0; p < plane_count; p++, i_qp = h->mb.i_chroma_qp) {
                do {
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * p] + 0 * 8]))->i) = 0;
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * p] + 1 * 8]))->i) = 0;
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * p] + 2 * 8]))->i) = 0;
                    (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * p] + 3 * 8]))->i) = 0;
                }
                while (0);
                h->dctf.sub16x16_dct (dct4x4, h->mb.pic.p_fenc[p], h->mb.pic.p_fdec[p]);
                if (h->mb.b_noise_reduction) {
                    h->nr_count[0 + !!p * 2] += 16;
                    for (int idx = 0; idx < 16; idx++)
                        h->quantf.denoise_dct (dct4x4[idx], h->nr_residual_sum[0 + !!p * 2], h->nr_offset[0 + !!p * 2], 16);
                }
                int plane_cbp = 0;
                for (int i8x8 = 0; i8x8 < 4; i8x8++) {
                    int i_decimate_8x8 = b_decimate ? 0 : 6;
                    int nnz8x8 = 0;
                    if (h->mb.b_trellis) {
                        for (int i4x4 = 0; i4x4 < 4; i4x4++) {
                            int idx = i8x8 * 4 + i4x4;
                        }
                    }
                    else {
                        nnz8x8 = nz = h->quantf.quant_4x4x4 (&dct4x4[i8x8 * 4], h->quant4_mf[CQM_4PY][i_qp], h->quant4_bias[CQM_4PY][i_qp]);
                        if (nz) {
                            for (int idx = i8x8 * 4, msk = nz, skip; msk && (skip = x264_ctz_4bit (msk), idx += skip, msk >>= skip + 1, 1); idx++) {
                                h->zigzagf.scan_4x4 (h->dct.luma4x4[p * 16 + idx], dct4x4[idx]);
                                h->quantf.dequant_4x4 (dct4x4[idx], h->dequant4_mf[p ? CQM_4PC : CQM_4PY], i_qp);
                                if (i_decimate_8x8 < 6)
                                    i_decimate_8x8 += h->quantf.decimate_score16 (h->dct.luma4x4[p * 16 + idx]);
                                h->mb.cache.non_zero_count[x264_scan8[p * 16 + idx]] = 1;
                            }
                        }
                    }
                    if (nnz8x8) {
                        i_decimate_mb += i_decimate_8x8;
                        if (i_decimate_8x8 < 4)
                            do {
                                (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8x8 * 4] + 0]))->i) = (0) * 0x0101;
                                (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8x8 * 4] + 8]))->i) = (0) * 0x0101;
                            }
                            while (0);
                        else
                            plane_cbp |= 1 << i8x8;
                    }
                }
                if (i_decimate_mb < 6) {
                    plane_cbp = 0;
                    do {
                        (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * p] + 0 * 8]))->i) = 0;
                        (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * p] + 1 * 8]))->i) = 0;
                        (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * p] + 2 * 8]))->i) = 0;
                        (((x264_union32_t *) (&h->mb.cache.non_zero_count[x264_scan8[16 * p] + 3 * 8]))->i) = 0;
                    }
                    while (0);
                }
                else {
                    h->mb.i_cbp_luma |= plane_cbp;
                    for (int i8x8 = 0, msk = plane_cbp, skip; msk && (skip = x264_ctz_4bit (msk), i8x8 += skip, msk >>= skip + 1, 1); i8x8++) {
                        h->dctf.add8x8_idct (&h->mb.pic.p_fdec[p][(i8x8 & 1) * 8 + (i8x8 >> 1) * 8 * 32], &dct4x4[i8x8 * 4]);
                    }
                }
            }
        }
    }
    if (chroma) {
        if (((h->mb.i_type) == I_4x4 || (h->mb.i_type) == I_8x8 || (h->mb.i_type) == I_16x16 || (h->mb.i_type) == I_PCM)) {
            int i_mode = h->mb.i_chroma_pred_mode;
            if (h->mb.b_lossless)
                x264_predict_lossless_chroma (h, i_mode);
            else {
                h->predict_chroma[i_mode] (h->mb.pic.p_fdec[1]);
                h->predict_chroma[i_mode] (h->mb.pic.p_fdec[2]);
            }
        }
        x264_mb_encode_chroma (h, ! ((h -> mb.i_type) == I_4x4 || (h -> mb.i_type) == I_8x8 || (h -> mb.i_type) == I_16x16 || (h -> mb.i_type) == I_PCM), h -> mb.i_chroma_qp);
    }
    else
        h->mb.i_cbp_chroma = 0;
    int cbp = h->mb.i_cbp_chroma << 4 | h->mb.i_cbp_luma;
    if (h->param.b_cabac)
        cbp |= h->mb.cache.non_zero_count[x264_scan8[48]] << 8 | h->mb.cache.non_zero_count[x264_scan8[49 + 0]] << 9 | h->mb.cache.non_zero_count[x264_scan8[49 + 1]] << 10;
    h->mb.cbp[h->mb.i_mb_xy] = cbp;
    if (!b_force_no_skip) {
        if (h->mb.i_type == P_L0 && h->mb.i_partition == D_16x16 && !(h->mb.i_cbp_luma | h->mb.i_cbp_chroma) && (((x264_union32_t *) (h->mb.cache.mv[0][x264_scan8[0]]))->i) == (((x264_union32_t *) (h->mb.cache.pskip_mv))->i) && h->mb.cache.ref[0][x264_scan8[0]] == 0) {
            h->mb.i_type = P_SKIP;
        }
        if (h->mb.i_type == B_DIRECT && !(h->mb.i_cbp_luma | h->mb.i_cbp_chroma)) {
            h->mb.i_type = B_SKIP;
        }
    }
}

void x264_macroblock_encode (x264_t *h) {
    if ((h->sps->i_chroma_format_idc == CHROMA_444))
        x264_macroblock_encode_internal (h, 3, 0);
    else
        x264_macroblock_encode_internal (h, 1, 1);
}

int x264_macroblock_probe_skip_internal (x264_t *h, int b_bidir, int plane_count, int chroma) {
    uint8_t dct4x4_u [sizeof (dctcoef[8][16]) + 31];
    dctcoef (*dct4x4) [16] = (void *) ((intptr_t) (dct4x4_u + 31) & ~31);
    dctcoef dctscan [16];
    int16_t mvp [2];
    int i_qp = h->mb.i_qp;
    for (int p = 0; p < plane_count; p++, i_qp = h->mb.i_chroma_qp) {
        int quant_cat = p ? CQM_4PC : CQM_4PY;
        if (!b_bidir) {
            mvp[0] = x264_clip3 (h->mb.cache.pskip_mv[0], h->mb.mv_min[0], h->mb.mv_max[0]);
            mvp[1] = x264_clip3 (h->mb.cache.pskip_mv[1], h->mb.mv_min[1], h->mb.mv_max[1]);
            h->mc.mc_luma (h->mb.pic.p_fdec[p], 32, &h->mb.pic.p_fref[0][0][p * 4], h->mb.pic.i_stride[p], mvp[0], mvp[1], 16, 16, &h->sh.weight[0][p]);
        }
        for (int i8x8 = 0, i_decimate_mb = 0; i8x8 < 4; i8x8++) {
            int fenc_offset = (i8x8 & 1) * 8 + (i8x8 >> 1) * 16 * 8;
            int fdec_offset = (i8x8 & 1) * 8 + (i8x8 >> 1) * 32 * 8;
            h->dctf.sub8x8_dct (dct4x4, h->mb.pic.p_fenc[p] + fenc_offset, h->mb.pic.p_fdec[p] + fdec_offset);
            if (h->mb.b_noise_reduction)
                for (int i4x4 = 0; i4x4 < 4; i4x4++)
                    h->quantf.denoise_dct (dct4x4[i4x4], h->nr_residual_sum[0 + !!p * 2], h->nr_offset[0 + !!p * 2], 16);
            int nz = h->quantf.quant_4x4x4 (dct4x4, h->quant4_mf[quant_cat][i_qp], h->quant4_bias[quant_cat][i_qp]);
            for (int idx = 0, msk = nz, skip; msk && (skip = x264_ctz_4bit (msk), idx += skip, msk >>= skip + 1, 1); idx++) {
                h->zigzagf.scan_4x4 (dctscan, dct4x4[idx]);
                i_decimate_mb += h->quantf.decimate_score16 (dctscan);
                if (i_decimate_mb >= 6)
                    return 0;
            }
        }
    }
    if (chroma == CHROMA_420 || chroma == CHROMA_422) {
        i_qp = h->mb.i_chroma_qp;
        int chroma422 = chroma == CHROMA_422;
        int thresh = chroma422 ? (x264_lambda2_tab[i_qp] + 16) >> 5 : (x264_lambda2_tab[i_qp] + 32) >> 6;
        int ssd;
        dctcoef dct_dc [8];
        if (!b_bidir) {
            if ((((x264_union32_t *) (mvp))->i))
                h->mc.mc_chroma (h->mb.pic.p_fdec[1], h->mb.pic.p_fdec[2], 32, h->mb.pic.p_fref[0][0][4], h->mb.pic.i_stride[1], mvp[0], mvp[1] << chroma422, 8, chroma422 ? 16 : 8);
            else
                h->mc.load_deinterleave_chroma_fdec (h->mb.pic.p_fdec[1], h->mb.pic.p_fref[0][0][4], h->mb.pic.i_stride[1], chroma422 ? 16 : 8);
        }
        for (int ch = 0; ch < 2; ch++) {
            pixel *p_src = h->mb.pic.p_fenc[1 + ch];
            pixel *p_dst = h->mb.pic.p_fdec[1 + ch];
            if (!b_bidir && h->sh.weight[0][1 + ch].weightfn)
                h->sh.weight[0][1 + ch].weightfn[8 >> 2] (h->mb.pic.p_fdec[1 + ch], 32, h->mb.pic.p_fdec[1 + ch], 32, &h->sh.weight[0][1 + ch], chroma422 ? 16 : 8);
            ssd = h->pixf.ssd[chroma422 ? PIXEL_8x16 : PIXEL_8x8] (p_dst, 32, p_src, 16);
            if (ssd < thresh)
                continue;
            if (h->mb.b_noise_reduction) {
                for (int i = 0; i <= chroma422; i++)
                    h->dctf.sub8x8_dct (&dct4x4[4 * i], p_src +8 * i * 16, p_dst +8 * i * 32);
                for (int i4x4 = 0; i4x4 < (chroma422 ? 8 : 4); i4x4++) {
                    h->quantf.denoise_dct (dct4x4[i4x4], h->nr_residual_sum[2], h->nr_offset[2], 16);
                    dct_dc[i4x4] = dct4x4[i4x4][0];
                    dct4x4[i4x4][0] = 0;
                }
            }
            else {
                if (chroma422)
                    h->dctf.sub8x16_dct_dc (dct_dc, p_src, p_dst);
                else
                    h->dctf.sub8x8_dct_dc (dct_dc, p_src, p_dst);
            }
            for (int i = 0; i <= chroma422; i++)
                if (h->quantf.quant_2x2_dc (&dct_dc[4 * i], h->quant4_mf[CQM_4PC][i_qp + 3 * chroma422][0] >> 1, h->quant4_bias[CQM_4PC][i_qp + 3 * chroma422][0] << 1))
                    return 0;
            if (ssd < thresh * 4)
                continue;
            if (!h->mb.b_noise_reduction)
                for (int i = 0; i <= chroma422; i++) {
                    h->dctf.sub8x8_dct (&dct4x4[4 * i], p_src +8 * i * 16, p_dst +8 * i * 32);
                    dct4x4[i * 4 + 0][0] = 0;
                    dct4x4[i * 4 + 1][0] = 0;
                    dct4x4[i * 4 + 2][0] = 0;
                    dct4x4[i * 4 + 3][0] = 0;
                }
            for (int i8x8 = 0, i_decimate_mb = 0; i8x8 < (chroma422 ? 2 : 1); i8x8++) {
                int nz = h->quantf.quant_4x4x4 (&dct4x4[i8x8 * 4], h->quant4_mf[CQM_4PC][i_qp], h->quant4_bias[CQM_4PC][i_qp]);
                for (int idx = i8x8 * 4, msk = nz, skip; msk && (skip = x264_ctz_4bit (msk), idx += skip, msk >>= skip + 1, 1); idx++) {
                    h->zigzagf.scan_4x4 (dctscan, dct4x4[idx]);
                    i_decimate_mb += h->quantf.decimate_score15 (dctscan);
                    if (i_decimate_mb >= 7)
                        return 0;
                }
            }
        }
    }
    h->mb.b_skip_mc = 1;
    return 1;
}

int x264_macroblock_probe_skip (x264_t *h, int b_bidir) {
    if (h->sps->i_chroma_format_idc == CHROMA_444)
        return x264_macroblock_probe_skip_internal (h, b_bidir, 3, CHROMA_444);
    else if (h->sps->i_chroma_format_idc == CHROMA_422)
        return x264_macroblock_probe_skip_internal (h, b_bidir, 1, CHROMA_422);
    else
        return x264_macroblock_probe_skip_internal (h, b_bidir, 1, CHROMA_420);
}

void x264_noise_reduction_update (x264_t *h) {
    h->nr_offset = h->nr_offset_denoise;
    h->nr_residual_sum = h->nr_residual_sum_buf[0];
    h->nr_count = h->nr_count_buf[0];
    for (int cat = 0; cat < 3 + (h->sps->i_chroma_format_idc == CHROMA_444); cat++) {
        int dct8x8 = cat & 1;
        int size = dct8x8 ? 64 : 16;
        const uint32_t *weight = dct8x8 ? x264_dct8_weight2_tab : x264_dct4_weight2_tab;
        if (h->nr_count[cat] > (dct8x8 ? (1 << 16) : (1 << 18))) {
            for (int i = 0; i < size; i++)
                h->nr_residual_sum[cat][i] >>= 1;
            h->nr_count[cat] >>= 1;
        }
        for (int i = 0; i < size; i++)
            h->nr_offset[cat][i] = ((uint64_t) h->param.analyse.i_noise_reduction * h->nr_count[cat] + h->nr_residual_sum[cat][i] / 2) / ((uint64_t) h->nr_residual_sum[cat][i] * weight[i] / 256 + 1);
        h->nr_offset[cat][0] = 0;
    }
}

void x264_macroblock_encode_p8x8_internal (x264_t *h, int i8, int plane_count, int chroma) {
    int b_decimate = h->mb.b_dct_decimate;
    int i_qp = h->mb.i_qp;
    int x = i8 & 1;
    int y = i8 >> 1;
    int nz;
    int chroma422 = chroma == CHROMA_422;
    h->mb.i_cbp_chroma = 0;
    h->mb.i_cbp_luma &= ~(1 << i8);
    if (!h->mb.b_skip_mc)
        x264_mb_mc_8x8 (h, i8);
    if (h->mb.b_lossless) {
        for (int p = 0; p < plane_count; p++) {
            pixel *p_fenc = h->mb.pic.p_fenc[p] + 8 * x + 8 * y * 16;
            pixel *p_fdec = h->mb.pic.p_fdec[p] + 8 * x + 8 * y * 32;
            int nnz8x8 = 0;
            if (h->mb.b_transform_8x8) {
                nnz8x8 = h->zigzagf.sub_8x8 (h->dct.luma8x8[4 * p + i8], p_fenc, p_fdec);
                do {
                    (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8 * 4] + 0]))->i) = (nnz8x8) *0x0101;
                    (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8 * 4] + 8]))->i) = (nnz8x8) *0x0101;
                }
                while (0);
            }
            else {
                for (int i4 = i8 * 4; i4 < i8 * 4 + 4; i4++) {
                    nz = h->zigzagf.sub_4x4 (h->dct.luma4x4[16 * p + i4], h->mb.pic.p_fenc[p] + block_idx_xy_fenc[i4], h->mb.pic.p_fdec[p] + block_idx_xy_fdec[i4]);
                    h->mb.cache.non_zero_count[x264_scan8[16 * p + i4]] = nz;
                    nnz8x8 |= nz;
                }
            }
            h->mb.i_cbp_luma |= nnz8x8 << i8;
        }
        if (chroma == CHROMA_420 || chroma == CHROMA_422) {
            for (int ch = 0; ch < 2; ch++) {
                dctcoef dc;
                pixel *p_fenc = h->mb.pic.p_fenc[1 + ch] + 4 * x + (chroma422 ? 8 : 4) * y * 16;
                pixel *p_fdec = h->mb.pic.p_fdec[1 + ch] + 4 * x + (chroma422 ? 8 : 4) * y * 32;
                for (int i4x4 = 0; i4x4 <= chroma422; i4x4++) {
                    int offset = chroma422 ? 8 * y + 2 * i4x4 + x : i8;
                    nz = h->zigzagf.sub_4x4ac (h->dct.luma4x4[16 + offset + ch * 16], p_fenc +4 * i4x4 * 16, p_fdec +4 * i4x4 * 32, &dc);
                    h->mb.cache.non_zero_count[x264_scan8[16 + offset + ch * 16]] = nz;
                }
            }
            h->mb.i_cbp_chroma = 0x02;
        }
    }
    else {
        if (h->mb.b_transform_8x8) {
            for (int p = 0; p < plane_count; p++, i_qp = h->mb.i_chroma_qp) {
                int quant_cat = p ? CQM_8PC : CQM_8PY;
                pixel *p_fenc = h->mb.pic.p_fenc[p] + 8 * x + 8 * y * 16;
                pixel *p_fdec = h->mb.pic.p_fdec[p] + 8 * x + 8 * y * 32;
                uint8_t dct8x8_u [sizeof (dctcoef[64]) + 31];
                dctcoef (*dct8x8) = (void *) ((intptr_t) (dct8x8_u + 31) & ~31);
                h->dctf.sub8x8_dct8 (dct8x8, p_fenc, p_fdec);
                int nnz8x8 = x264_quant_8x8 (h, dct8x8, i_qp, ctx_cat_plane[DCT_LUMA_8x8][p], 0, p, i8);
                if (nnz8x8) {
                    h->zigzagf.scan_8x8 (h->dct.luma8x8[4 * p + i8], dct8x8);
                    if (b_decimate && !h->mb.b_trellis)
                        nnz8x8 = 4 <= h->quantf.decimate_score64 (h->dct.luma8x8[4 * p + i8]);
                    if (nnz8x8) {
                        h->quantf.dequant_8x8 (dct8x8, h->dequant8_mf[quant_cat], i_qp);
                        h->dctf.add8x8_idct8 (p_fdec, dct8x8);
                        do {
                            (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8 * 4] + 0]))->i) = (1) * 0x0101;
                            (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8 * 4] + 8]))->i) = (1) * 0x0101;
                        }
                        while (0);
                        h->mb.i_cbp_luma |= 1 << i8;
                    }
                    else
                        do {
                            (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8 * 4] + 0]))->i) = (0) * 0x0101;
                            (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8 * 4] + 8]))->i) = (0) * 0x0101;
                        }
                        while (0);
                }
                else
                    do {
                        (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8 * 4] + 0]))->i) = (0) * 0x0101;
                        (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8 * 4] + 8]))->i) = (0) * 0x0101;
                    }
                    while (0);
            }
        }
        else {
            for (int p = 0; p < plane_count; p++, i_qp = h->mb.i_chroma_qp) {
                int quant_cat = p ? CQM_4PC : CQM_4PY;
                pixel *p_fenc = h->mb.pic.p_fenc[p] + 8 * x + 8 * y * 16;
                pixel *p_fdec = h->mb.pic.p_fdec[p] + 8 * x + 8 * y * 32;
                int i_decimate_8x8 = b_decimate ? 0 : 4;
                uint8_t dct4x4_u [sizeof (dctcoef[4][16]) + 31];
                dctcoef (*dct4x4) [16] = (void *) ((intptr_t) (dct4x4_u + 31) & ~31);
                int nnz8x8 = 0;
                h->dctf.sub8x8_dct (dct4x4, p_fenc, p_fdec);
                do {
                    (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8 * 4] + 0]))->i) = (0) * 0x0101;
                    (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8 * 4] + 8]))->i) = (0) * 0x0101;
                }
                while (0);
                if (h->mb.b_noise_reduction)
                    for (int idx = 0; idx < 4; idx++)
                        h->quantf.denoise_dct (dct4x4[idx], h->nr_residual_sum[0 + !!p * 2], h->nr_offset[0 + !!p * 2], 16);
                if (h->mb.b_trellis) {
                    for (int i4x4 = 0; i4x4 < 4; i4x4++) {
                    }
                }
                else {
                    nnz8x8 = nz = h->quantf.quant_4x4x4 (dct4x4, h->quant4_mf[quant_cat][i_qp], h->quant4_bias[quant_cat][i_qp]);
                    if (nz) {
                        for (int i4x4 = 0, msk = nz, skip; msk && (skip = x264_ctz_4bit (msk), i4x4 += skip, msk >>= skip + 1, 1); i4x4++) {
                            h->zigzagf.scan_4x4 (h->dct.luma4x4[p * 16 + i8 * 4 + i4x4], dct4x4[i4x4]);
                            h->quantf.dequant_4x4 (dct4x4[i4x4], h->dequant4_mf[quant_cat], i_qp);
                            if (i_decimate_8x8 < 4)
                                i_decimate_8x8 += h->quantf.decimate_score16 (h->dct.luma4x4[p * 16 + i8 * 4 + i4x4]);
                            h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8 * 4 + i4x4]] = 1;
                        }
                    }
                }
                if (nnz8x8) {
                    if (i_decimate_8x8 < 4)
                        do {
                            (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8 * 4] + 0]))->i) = (0) * 0x0101;
                            (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + i8 * 4] + 8]))->i) = (0) * 0x0101;
                        }
                        while (0);
                    else {
                        h->dctf.add8x8_idct (p_fdec, dct4x4);
                        h->mb.i_cbp_luma |= 1 << i8;
                    }
                }
            }
        }
        if (chroma == CHROMA_420 || chroma == CHROMA_422) {
            i_qp = h->mb.i_chroma_qp;
            for (int ch = 0; ch < 2; ch++) {
                uint8_t dct4x4_u [sizeof (dctcoef[2][16]) + 31];
                dctcoef (*dct4x4) [16] = (void *) ((intptr_t) (dct4x4_u + 31) & ~31);
                pixel *p_fenc = h->mb.pic.p_fenc[1 + ch] + 4 * x + (chroma422 ? 8 : 4) * y * 16;
                pixel *p_fdec = h->mb.pic.p_fdec[1 + ch] + 4 * x + (chroma422 ? 8 : 4) * y * 32;
                for (int i4x4 = 0; i4x4 <= chroma422; i4x4++) {
                    h->dctf.sub4x4_dct (dct4x4[i4x4], p_fenc +4 * i4x4 * 16, p_fdec +4 * i4x4 * 32);
                    if (h->mb.b_noise_reduction)
                        h->quantf.denoise_dct (dct4x4[i4x4], h->nr_residual_sum[2], h->nr_offset[2], 16);
                    dct4x4[i4x4][0] = 0;
                    nz = h->quantf.quant_4x4 (dct4x4[i4x4], h->quant4_mf[CQM_4PC][i_qp], h->quant4_bias[CQM_4PC][i_qp]);
                    int offset = chroma422 ? ((5 * i8) & 0x09) + 2 * i4x4 : i8;
                    h->mb.cache.non_zero_count[x264_scan8[16 + offset + ch * 16]] = nz;
                    if (nz) {
                        h->zigzagf.scan_4x4 (h->dct.luma4x4[16 + offset + ch * 16], dct4x4[i4x4]);
                        h->quantf.dequant_4x4 (dct4x4[i4x4], h->dequant4_mf[CQM_4PC], i_qp);
                        h->dctf.add4x4_idct (p_fdec +4 * i4x4 * 32, dct4x4[i4x4]);
                    }
                }
            }
            h->mb.i_cbp_chroma = 0x02;
        }
    }
}

void x264_macroblock_encode_p8x8 (x264_t *h, int i8) {
    if ((h->sps->i_chroma_format_idc == CHROMA_444))
        x264_macroblock_encode_p8x8_internal (h, i8, 3, CHROMA_444);
    else if (h->sps->i_chroma_format_idc == CHROMA_422)
        x264_macroblock_encode_p8x8_internal (h, i8, 1, CHROMA_422);
    else
        x264_macroblock_encode_p8x8_internal (h, i8, 1, CHROMA_420);
}

void x264_macroblock_encode_p4x4_internal (x264_t *h, int i4, int plane_count) {
    int i_qp = h->mb.i_qp;
    for (int p = 0; p < plane_count; p++, i_qp = h->mb.i_chroma_qp) {
        int quant_cat = p ? CQM_4PC : CQM_4PY;
        pixel *p_fenc = &h->mb.pic.p_fenc[p][block_idx_xy_fenc[i4]];
        pixel *p_fdec = &h->mb.pic.p_fdec[p][block_idx_xy_fdec[i4]];
        int nz;
        if (h->mb.b_lossless) {
            nz = h->zigzagf.sub_4x4 (h->dct.luma4x4[p * 16 + i4], p_fenc, p_fdec);
            h->mb.cache.non_zero_count[x264_scan8[p * 16 + i4]] = nz;
        }
        else {
            uint8_t dct4x4_u [sizeof (dctcoef[16]) + 31];
            dctcoef (*dct4x4) = (void *) ((intptr_t) (dct4x4_u + 31) & ~31);
            h->dctf.sub4x4_dct (dct4x4, p_fenc, p_fdec);
            nz = x264_quant_4x4 (h, dct4x4, i_qp, ctx_cat_plane[DCT_LUMA_4x4][p], 0, p, i4);
            h->mb.cache.non_zero_count[x264_scan8[p * 16 + i4]] = nz;
            if (nz) {
                h->zigzagf.scan_4x4 (h->dct.luma4x4[p * 16 + i4], dct4x4);
                h->quantf.dequant_4x4 (dct4x4, h->dequant4_mf[quant_cat], i_qp);
                h->dctf.add4x4_idct (p_fdec, dct4x4);
            }
        }
    }
}

void x264_macroblock_encode_p4x4 (x264_t *h, int i8) {
    if ((h->sps->i_chroma_format_idc == CHROMA_444))
        x264_macroblock_encode_p4x4_internal (h, i8, 3);
    else
        x264_macroblock_encode_p4x4_internal (h, i8, 1);
}

int x264_quant_4x4 (x264_t *h, dctcoef dct [16], int i_qp, int ctx_block_cat, int b_intra, int p, int idx) {
    int i_quant_cat = b_intra ? (p ? CQM_4IC : CQM_4IY) : (p ? CQM_4PC : CQM_4PY);
    if (h->mb.b_noise_reduction)
        h->quantf.denoise_dct (dct, h->nr_residual_sum[0 + !!p * 2], h->nr_offset[0 + !!p * 2], 16);
    return h->quantf.quant_4x4 (dct, h->quant4_mf[i_quant_cat][i_qp], h->quant4_bias[i_quant_cat][i_qp]);
}

int x264_quant_8x8 (x264_t *h, dctcoef dct [64], int i_qp, int ctx_block_cat, int b_intra, int p, int idx) {
    return 0;
}

void x264_mb_encode_i4x4 (x264_t *h, int p, int idx, int i_qp, int i_mode, int b_predict) {
    int nz;
    pixel *p_src = &h->mb.pic.p_fenc[p][block_idx_xy_fenc[idx]];
    pixel *p_dst = &h->mb.pic.p_fdec[p][block_idx_xy_fdec[idx]];
    ALIGNED_ARRAY_N (dctcoef, dct4x4, [16]);
    if (b_predict) {
        if (h->mb.b_lossless)
            x264_predict_lossless_4x4 (h, p_dst, p, idx, i_mode);
        else
            h->predict_4x4[i_mode] (p_dst);
    }
    if (h->mb.b_lossless) {
        nz = h->zigzagf.sub_4x4 (h->dct.luma4x4[p * 16 + idx], p_src, p_dst);
        h->mb.cache.non_zero_count[x264_scan8[p * 16 + idx]] = nz;
        h->mb.i_cbp_luma |= nz << (idx >> 2);
        return;
    }
    h->dctf.sub4x4_dct (dct4x4, p_src, p_dst);
    nz = x264_quant_4x4 (h, dct4x4, i_qp, ctx_cat_plane[DCT_LUMA_4x4][p], 1, p, idx);
    h->mb.cache.non_zero_count[x264_scan8[p * 16 + idx]] = nz;
    if (nz) {
        h->mb.i_cbp_luma |= 1 << (idx >> 2);
        h->zigzagf.scan_4x4 (h->dct.luma4x4[p * 16 + idx], dct4x4);
        h->quantf.dequant_4x4 (dct4x4, h->dequant4_mf[p ? CQM_4IC : CQM_4IY], i_qp);
        h->dctf.add4x4_idct (p_dst, dct4x4);
    }
}

void x264_mb_encode_i8x8 (x264_t *h, int p, int idx, int i_qp, int i_mode, pixel *edge, int b_predict) {
    int x = idx & 1;
    int y = idx >> 1;
    int nz;
    pixel *p_src = &h->mb.pic.p_fenc[p][8 * x + 8 * y * FENC_STRIDE];
    pixel *p_dst = &h->mb.pic.p_fdec[p][8 * x + 8 * y * FDEC_STRIDE];
    ALIGNED_ARRAY_N (dctcoef, dct8x8, [64]);
    ALIGNED_ARRAY_32 (pixel, edge_buf, [36]);
    if (b_predict) {
        if (!edge) {
            h->predict_8x8_filter (p_dst, edge_buf, h->mb.i_neighbour8[idx], x264_pred_i4x4_neighbors[i_mode]);
            edge = edge_buf;
        }
        if (h->mb.b_lossless)
            x264_predict_lossless_8x8 (h, p_dst, p, idx, i_mode, edge);
        else
            h->predict_8x8[i_mode] (p_dst, edge);
    }
    if (h->mb.b_lossless) {
        nz = h->zigzagf.sub_8x8 (h->dct.luma8x8[p * 4 + idx], p_src, p_dst);
        do {
            M16 (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + idx * 4] + 0]) = (nz) *0x0101;
            M16 (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + idx * 4] + 8]) = (nz) *0x0101;
        }
        while (0);
        h->mb.i_cbp_luma |= nz << idx;
        return;
    }
    h->dctf.sub8x8_dct8 (dct8x8, p_src, p_dst);
    nz = x264_quant_8x8 (h, dct8x8, i_qp, ctx_cat_plane[DCT_LUMA_8x8][p], 1, p, idx);
    if (nz) {
        h->mb.i_cbp_luma |= 1 << idx;
        h->zigzagf.scan_8x8 (h->dct.luma8x8[p * 4 + idx], dct8x8);
        h->quantf.dequant_8x8 (dct8x8, h->dequant8_mf[p ? CQM_8IC : CQM_8IY], i_qp);
        h->dctf.add8x8_idct8 (p_dst, dct8x8);
        do {
            M16 (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + idx * 4] + 0]) = (1) * 0x0101;
            M16 (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + idx * 4] + 8]) = (1) * 0x0101;
        }
        while (0);
    }
    else
        do {
            M16 (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + idx * 4] + 0]) = (0) * 0x0101;
            M16 (&h->mb.cache.non_zero_count[x264_scan8[p * 16 + idx * 4] + 8]) = (0) * 0x0101;
        }
        while (0);
}

