
#include "common.h"

void pixel_avg (pixel *dst, intptr_t i_dst_stride, pixel *src1, intptr_t i_src1_stride, pixel *src2, intptr_t i_src2_stride, int i_width, int i_height) {
    for (int y = 0; y < i_height; y++) {
        for (int x = 0; x < i_width; x++)
            dst[x] = (src1[x] + src2[x] + 1) >> 1;
        dst += i_dst_stride;
        src1 += i_src1_stride;
        src2 += i_src2_stride;
    }
}

void pixel_avg_wxh (pixel *dst, intptr_t i_dst, pixel *src1, intptr_t i_src1, pixel *src2, intptr_t i_src2, int width, int height) {
    for (int y = 0; y < height; y++) {
        for (int x = 0; x < width; x++)
            dst[x] = (src1[x] + src2[x] + 1) >> 1;
        src1 += i_src1;
        src2 += i_src2;
        dst += i_dst;
    }
}

void pixel_avg_weight_wxh (pixel *dst, intptr_t i_dst, pixel *src1, intptr_t i_src1, pixel *src2, intptr_t i_src2, int width, int height, int i_weight1) {
    int i_weight2 = 64 - i_weight1;
    for (int y = 0; y < height; y++, dst += i_dst, src1 += i_src1, src2 += i_src2)
        for (int x = 0; x < width; x++)
            dst[x] = x264_clip_pixel ((src1[x] * i_weight1 + src2[x] * i_weight2 + (1 << 5)) >> 6);
}

void pixel_avg_16x16 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2, pixel *pix3, intptr_t i_stride_pix3, int weight) {
    if (weight == 32)
        pixel_avg_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 16, 16);
    else
        pixel_avg_weight_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 16, 16, weight);
}

void pixel_avg_16x8 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2, pixel *pix3, intptr_t i_stride_pix3, int weight) {
    if (weight == 32)
        pixel_avg_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 16, 8);
    else
        pixel_avg_weight_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 16, 8, weight);
}

void pixel_avg_8x16 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2, pixel *pix3, intptr_t i_stride_pix3, int weight) {
    if (weight == 32)
        pixel_avg_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 8, 16);
    else
        pixel_avg_weight_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 8, 16, weight);
}

void pixel_avg_8x8 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2, pixel *pix3, intptr_t i_stride_pix3, int weight) {
    if (weight == 32)
        pixel_avg_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 8, 8);
    else
        pixel_avg_weight_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 8, 8, weight);
}

void pixel_avg_8x4 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2, pixel *pix3, intptr_t i_stride_pix3, int weight) {
    if (weight == 32)
        pixel_avg_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 8, 4);
    else
        pixel_avg_weight_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 8, 4, weight);
}

void pixel_avg_4x16 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2, pixel *pix3, intptr_t i_stride_pix3, int weight) {
    if (weight == 32)
        pixel_avg_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 4, 16);
    else
        pixel_avg_weight_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 4, 16, weight);
}

void pixel_avg_4x8 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2, pixel *pix3, intptr_t i_stride_pix3, int weight) {
    if (weight == 32)
        pixel_avg_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 4, 8);
    else
        pixel_avg_weight_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 4, 8, weight);
}

void pixel_avg_4x4 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2, pixel *pix3, intptr_t i_stride_pix3, int weight) {
    if (weight == 32)
        pixel_avg_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 4, 4);
    else
        pixel_avg_weight_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 4, 4, weight);
}

void pixel_avg_4x2 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2, pixel *pix3, intptr_t i_stride_pix3, int weight) {
    if (weight == 32)
        pixel_avg_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 4, 2);
    else
        pixel_avg_weight_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 4, 2, weight);
}

void pixel_avg_2x8 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2, pixel *pix3, intptr_t i_stride_pix3, int weight) {
    if (weight == 32)
        pixel_avg_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 2, 8);
    else
        pixel_avg_weight_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 2, 8, weight);
}

void pixel_avg_2x4 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2, pixel *pix3, intptr_t i_stride_pix3, int weight) {
    if (weight == 32)
        pixel_avg_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 2, 4);
    else
        pixel_avg_weight_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 2, 4, weight);
}

void pixel_avg_2x2 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2, pixel *pix3, intptr_t i_stride_pix3, int weight) {
    if (weight == 32)
        pixel_avg_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 2, 2);
    else
        pixel_avg_weight_wxh (pix1, i_stride_pix1, pix2, i_stride_pix2, pix3, i_stride_pix3, 2, 2, weight);
}

void x264_weight_cache (x264_t *h, x264_weight_t *w) {
    w->weightfn = h->mc.weight;
}

void mc_weight (pixel *dst, intptr_t i_dst_stride, pixel *src, intptr_t i_src_stride, const x264_weight_t *weight, int i_width, int i_height) {
    int offset = weight->i_offset << (8 - 8);
    int scale = weight->i_scale;
    int denom = weight->i_denom;
    if (denom >= 1) {
        for (int y = 0; y < i_height; y++, dst += i_dst_stride, src += i_src_stride)
            for (int x = 0; x < i_width; x++)
                dst[x] = x264_clip_pixel (((src[x] * scale + (1 << (denom - 1))) >> denom) + offset);
    }
    else {
        for (int y = 0; y < i_height; y++, dst += i_dst_stride, src += i_src_stride)
            for (int x = 0; x < i_width; x++)
                dst[x] = x264_clip_pixel (src[x] * scale + offset);
    }
}

void mc_weight_w20 (pixel *dst, intptr_t i_dst_stride, pixel *src, intptr_t i_src_stride, const x264_weight_t *weight, int height) {
    mc_weight (dst, i_dst_stride, src, i_src_stride, weight, 20, height);
}

void mc_weight_w16 (pixel *dst, intptr_t i_dst_stride, pixel *src, intptr_t i_src_stride, const x264_weight_t *weight, int height) {
    mc_weight (dst, i_dst_stride, src, i_src_stride, weight, 16, height);
}

void mc_weight_w12 (pixel *dst, intptr_t i_dst_stride, pixel *src, intptr_t i_src_stride, const x264_weight_t *weight, int height) {
    mc_weight (dst, i_dst_stride, src, i_src_stride, weight, 12, height);
}

void mc_weight_w8 (pixel *dst, intptr_t i_dst_stride, pixel *src, intptr_t i_src_stride, const x264_weight_t *weight, int height) {
    mc_weight (dst, i_dst_stride, src, i_src_stride, weight, 8, height);
}

void mc_weight_w4 (pixel *dst, intptr_t i_dst_stride, pixel *src, intptr_t i_src_stride, const x264_weight_t *weight, int height) {
    mc_weight (dst, i_dst_stride, src, i_src_stride, weight, 4, height);
}

void mc_weight_w2 (pixel *dst, intptr_t i_dst_stride, pixel *src, intptr_t i_src_stride, const x264_weight_t *weight, int height) {
    mc_weight (dst, i_dst_stride, src, i_src_stride, weight, 2, height);
}

weight_fn_t x264_mc_weight_wtab [6] = {mc_weight_w2, mc_weight_w4, mc_weight_w8, mc_weight_w12, mc_weight_w16, mc_weight_w20,};
const x264_weight_t x264_weight_none [3] = {{{0}}};

void mc_copy (pixel *src, intptr_t i_src_stride, pixel *dst, intptr_t i_dst_stride, int i_width, int i_height) {
    for (int y = 0; y < i_height; y++) {
        memcpy (dst, src, i_width * sizeof (pixel));
        src += i_src_stride;
        dst += i_dst_stride;
    }
}

void hpel_filter (pixel *dsth, pixel *dstv, pixel *dstc, pixel *src, intptr_t stride, int width, int height, int16_t *buf) {
    const int pad = (8 > 9) ? (-10 * ((1 << 8) - 1)) : 0;
    for (int y = 0; y < height; y++) {
        for (int x = -2; x < width + 3; x++) {
            int v = ((src)[x - 2 * stride] + (src)[x + 3 * stride] - 5 * ((src)[x - stride] + (src)[x + 2 * stride]) + 20 * ((src)[x] + (src)[x + stride]));
            dstv[x] = x264_clip_pixel ((v + 16) >> 5);
            buf[x + 2] = v + pad;
        }
        for (int x = 0; x < width; x++)
            dstc[x] = x264_clip_pixel ((((buf + 2)[x - 2 * 1] + (buf + 2)[x + 3 * 1] - 5 * ((buf + 2)[x - 1] + (buf + 2)[x + 2 * 1]) + 20 * ((buf + 2)[x] + (buf + 2)[x + 1])) - 32 * pad + 512) >> 10);
        for (int x = 0; x < width; x++)
            dsth[x] = x264_clip_pixel ((((src)[x - 2 * 1] + (src)[x + 3 * 1] - 5 * ((src)[x - 1] + (src)[x + 2 * 1]) + 20 * ((src)[x] + (src)[x + 1])) + 16) >> 5);
        dsth += stride;
        dstv += stride;
        dstc += stride;
        src += stride;
    }
}

const uint8_t hpel_ref0 [16] = {0, 1, 1, 1, 0, 1, 1, 1, 2, 3, 3, 3, 0, 1, 1, 1};
const uint8_t hpel_ref1 [16] = {0, 0, 0, 0, 2, 2, 3, 2, 2, 2, 3, 2, 2, 2, 3, 2};

void mc_luma (pixel *dst, intptr_t i_dst_stride, pixel *src [4], intptr_t i_src_stride, int mvx, int mvy, int i_width, int i_height, const x264_weight_t *weight) {
    int qpel_idx = ((mvy & 3) << 2) + (mvx & 3);
    int offset = (mvy >> 2) * i_src_stride + (mvx >> 2);
    pixel *src1 = src[hpel_ref0[qpel_idx]] + offset + ((mvy & 3) == 3) * i_src_stride;
    if (qpel_idx & 5) {
        pixel *src2 = src[hpel_ref1[qpel_idx]] + offset + ((mvx & 3) == 3);
        pixel_avg (dst, i_dst_stride, src1, i_src_stride, src2, i_src_stride, i_width, i_height);
        if (weight->weightfn)
            mc_weight (dst, i_dst_stride, dst, i_dst_stride, weight, i_width, i_height);
    }
    else if (weight->weightfn)
        mc_weight (dst, i_dst_stride, src1, i_src_stride, weight, i_width, i_height);
    else
        mc_copy (src1, i_src_stride, dst, i_dst_stride, i_width, i_height);
}

pixel *get_ref (pixel *dst, intptr_t *i_dst_stride, pixel *src [4], intptr_t i_src_stride, int mvx, int mvy, int i_width, int i_height, const x264_weight_t *weight) {
    int qpel_idx = ((mvy & 3) << 2) + (mvx & 3);
    int offset = (mvy >> 2) * i_src_stride + (mvx >> 2);
    pixel *src1 = src[hpel_ref0[qpel_idx]] + offset + ((mvy & 3) == 3) * i_src_stride;
    if (qpel_idx & 5) {
        pixel *src2 = src[hpel_ref1[qpel_idx]] + offset + ((mvx & 3) == 3);
        pixel_avg (dst, * i_dst_stride, src1, i_src_stride, src2, i_src_stride, i_width, i_height);
        if (weight->weightfn)
            mc_weight (dst, *i_dst_stride, dst, *i_dst_stride, weight, i_width, i_height);
        return dst;
    }
    else if (weight->weightfn) {
        mc_weight (dst, * i_dst_stride, src1, i_src_stride, weight, i_width, i_height);
        return dst;
    }
    else {
        *i_dst_stride = i_src_stride;
        return src1;
    }
}

void mc_chroma (pixel *dstu, pixel *dstv, intptr_t i_dst_stride, pixel *src, intptr_t i_src_stride, int mvx, int mvy, int i_width, int i_height) {
    pixel *srcp;
    int d8x = mvx & 0x07;
    int d8y = mvy & 0x07;
    int cA = (8 - d8x) * (8 - d8y);
    int cB = d8x * (8 - d8y);
    int cC = (8 - d8x) * d8y;
    int cD = d8x * d8y;
    src += (mvy >> 3) * i_src_stride + (mvx >> 3) * 2;
    srcp = &src[i_src_stride];
    for (int y = 0; y < i_height; y++) {
        for (int x = 0; x < i_width; x++) {
            dstu[x] = (cA * src[2 * x] + cB * src[2 * x + 2] + cC * srcp[2 * x] + cD * srcp[2 * x + 2] + 32) >> 6;
            dstv[x] = (cA * src[2 * x + 1] + cB * src[2 * x + 3] + cC * srcp[2 * x + 1] + cD * srcp[2 * x + 3] + 32) >> 6;
        }
        dstu += i_dst_stride;
        dstv += i_dst_stride;
        src = srcp;
        srcp += i_src_stride;
    }
}

void mc_copy_w16 (pixel *dst, intptr_t i_dst, pixel *src, intptr_t i_src, int i_height) {
    mc_copy (src, i_src, dst, i_dst, 16, i_height);
}

void mc_copy_w8 (pixel *dst, intptr_t i_dst, pixel *src, intptr_t i_src, int i_height) {
    mc_copy (src, i_src, dst, i_dst, 8, i_height);
}

void mc_copy_w4 (pixel *dst, intptr_t i_dst, pixel *src, intptr_t i_src, int i_height) {
    mc_copy (src, i_src, dst, i_dst, 4, i_height);
}

void x264_plane_copy_c (pixel *dst, intptr_t i_dst, pixel *src, intptr_t i_src, int w, int h) {
    while (h--) {
        memcpy (dst, src, w * sizeof (pixel));
        dst += i_dst;
        src += i_src;
    }
}

void x264_plane_copy_interleave_c (pixel *dst, intptr_t i_dst, pixel *srcu, intptr_t i_srcu, pixel *srcv, intptr_t i_srcv, int w, int h) {
    for (int y = 0; y < h; y++, dst += i_dst, srcu += i_srcu, srcv += i_srcv)
        for (int x = 0; x < w; x++) {
            dst[2 * x] = srcu[x];
            dst[2 * x + 1] = srcv[x];
        }
}

void x264_plane_copy_deinterleave_c (pixel *dstu, intptr_t i_dstu, pixel *dstv, intptr_t i_dstv, pixel *src, intptr_t i_src, int w, int h) {
    for (int y = 0; y < h; y++, dstu += i_dstu, dstv += i_dstv, src += i_src)
        for (int x = 0; x < w; x++) {
            dstu[x] = src[2 * x];
            dstv[x] = src[2 * x + 1];
        }
}

void x264_plane_copy_deinterleave_rgb_c (pixel *dsta, intptr_t i_dsta, pixel *dstb, intptr_t i_dstb, pixel *dstc, intptr_t i_dstc, pixel *src, intptr_t i_src, int pw, int w, int h) {
    for (int y = 0; y < h; y++, dsta += i_dsta, dstb += i_dstb, dstc += i_dstc, src += i_src) {
        for (int x = 0; x < w; x++) {
            dsta[x] = src[x * pw];
            dstb[x] = src[x * pw + 1];
            dstc[x] = src[x * pw + 2];
        }
    }
}

void x264_plane_copy_deinterleave_v210_c (pixel *dsty, intptr_t i_dsty, pixel *dstc, intptr_t i_dstc, uint32_t *src, intptr_t i_src, int w, int h) {
    for (int l = 0; l < h; l++) {
        pixel *dsty0 = dsty;
        pixel *dstc0 = dstc;
        uint32_t *src0 = src;
        for (int n = 0; n < w; n += 3) {
            *(dstc0++) = *src0 & 0x03FF;
            *(dsty0++) = (*src0 >> 10) & 0x03FF;
            *(dstc0++) = (*src0 >> 20) & 0x03FF;
            src0++;
            *(dsty0++) = *src0 & 0x03FF;
            *(dstc0++) = (*src0 >> 10) & 0x03FF;
            *(dsty0++) = (*src0 >> 20) & 0x03FF;
            src0++;
        }
        dsty += i_dsty;
        dstc += i_dstc;
        src += i_src;
    }
}

void store_interleave_chroma (pixel *dst, intptr_t i_dst, pixel *srcu, pixel *srcv, int height) {
    for (int y = 0; y < height; y++, dst += i_dst, srcu += 32, srcv += 32)
        for (int x = 0; x < 8; x++) {
            dst[2 * x] = srcu[x];
            dst[2 * x + 1] = srcv[x];
        }
}

void load_deinterleave_chroma_fenc (pixel *dst, pixel *src, intptr_t i_src, int height) {
    x264_plane_copy_deinterleave_c (dst, 16, dst + 16 / 2, 16, src, i_src, 8, height);
}

void load_deinterleave_chroma_fdec (pixel *dst, pixel *src, intptr_t i_src, int height) {
    x264_plane_copy_deinterleave_c (dst, 32, dst + 32 / 2, 32, src, i_src, 8, height);
}

void prefetch_fenc_null (pixel *pix_y, intptr_t stride_y, pixel *pix_uv, intptr_t stride_uv, int mb_x) {
}

void prefetch_ref_null (pixel *pix, intptr_t stride, int parity) {
}

void memzero_aligned (void *dst, size_t n) {
    memset (dst, 0, n);
}

void integral_init4h (uint16_t *sum, pixel *pix, intptr_t stride) {
    int v = pix[0] + pix[1] + pix[2] + pix[3];
    for (int x = 0; x < stride - 4; x++) {
        sum[x] = v + sum[x - stride];
        v += pix[x + 4] - pix[x];
    }
}

void integral_init8h (uint16_t *sum, pixel *pix, intptr_t stride) {
    int v = pix[0] + pix[1] + pix[2] + pix[3] + pix[4] + pix[5] + pix[6] + pix[7];
    for (int x = 0; x < stride - 8; x++) {
        sum[x] = v + sum[x - stride];
        v += pix[x + 8] - pix[x];
    }
}

void integral_init4v (uint16_t *sum8, uint16_t *sum4, intptr_t stride) {
    for (int x = 0; x < stride - 8; x++)
        sum4[x] = sum8[x + 4 * stride] - sum8[x];
    for (int x = 0; x < stride - 8; x++)
        sum8[x] = sum8[x + 8 * stride] + sum8[x + 8 * stride + 4] - sum8[x] - sum8[x + 4];
}

void integral_init8v (uint16_t *sum8, intptr_t stride) {
    for (int x = 0; x < stride - 8; x++)
        sum8[x] = sum8[x + 8 * stride] - sum8[x];
}

void x264_frame_init_lowres (x264_t *h, x264_frame_t *frame) {
    pixel *src = frame->plane[0];
    int i_stride = frame->i_stride[0];
    int i_height = frame->i_lines[0];
    int i_width = frame->i_width[0];
    for (int y = 0; y < i_height; y++)
        src[i_width + y * i_stride] = src[i_width - 1 + y * i_stride];
    memcpy (src + i_stride * i_height, src + i_stride * (i_height - 1), (i_width + 1) * sizeof (pixel));
    h->mc.frame_init_lowres_core (src, frame->lowres[0], frame->lowres[1], frame->lowres[2], frame->lowres[3], i_stride, frame->i_stride_lowres, frame->i_width_lowres, frame->i_lines_lowres);
    x264_frame_expand_border_lowres (frame);
    memset (frame -> i_cost_est, - 1, sizeof (frame -> i_cost_est));
    for (int y = 0; y < h->param.i_bframe + 2; y++)
        for (int x = 0; x < h->param.i_bframe + 2; x++)
            frame->i_row_satds[y][x][0] = -1;
    for (int y = 0; y <= !!h->param.i_bframe; y++)
        for (int x = 0; x <= h->param.i_bframe; x++)
            frame->lowres_mvs[y][x][0][0] = 0x7FFF;
}

void frame_init_lowres_core (pixel *src0, pixel *dst0, pixel *dsth, pixel *dstv, pixel *dstc, intptr_t src_stride, intptr_t dst_stride, int width, int height) {
    for (int y = 0; y < height; y++) {
        pixel *src1 = src0 + src_stride;
        pixel *src2 = src1 + src_stride;
        for (int x = 0; x < width; x++) {
            dst0[x] = ((((src0[2 * x] + src1[2 * x] + 1) >> 1) + ((src0[2 * x + 1] + src1[2 * x + 1] + 1) >> 1) + 1) >> 1);
            dsth[x] = ((((src0[2 * x + 1] + src1[2 * x + 1] + 1) >> 1) + ((src0[2 * x + 2] + src1[2 * x + 2] + 1) >> 1) + 1) >> 1);
            dstv[x] = ((((src1[2 * x] + src2[2 * x] + 1) >> 1) + ((src1[2 * x + 1] + src2[2 * x + 1] + 1) >> 1) + 1) >> 1);
            dstc[x] = ((((src1[2 * x + 1] + src2[2 * x + 1] + 1) >> 1) + ((src1[2 * x + 2] + src2[2 * x + 2] + 1) >> 1) + 1) >> 1);
        }
        src0 += src_stride * 2;
        dst0 += dst_stride;
        dsth += dst_stride;
        dstv += dst_stride;
        dstc += dst_stride;
    }
}

void mbtree_propagate_cost (int16_t *dst, uint16_t *propagate_in, uint16_t *intra_costs, uint16_t *inter_costs, uint16_t *inv_qscales, float *fps_factor, int len) {
    float fps = *fps_factor;
    for (int i = 0; i < len; i++) {
        int intra_cost = intra_costs[i];
        int inter_cost = ((intra_costs[i]) < (inter_costs[i] & ((1 << 14) - 1)) ? (intra_costs[i]) : (inter_costs[i] & ((1 << 14) - 1)));
        float propagate_intra = intra_cost * inv_qscales[i];
        float propagate_amount = propagate_in[i] + propagate_intra * fps;
        float propagate_num = intra_cost - inter_cost;
        float propagate_denom = intra_cost;
        dst[i] = (((int) (propagate_amount * propagate_num / propagate_denom + 0.5f)) < (32767) ? ((int) (propagate_amount * propagate_num / propagate_denom + 0.5f)) : (32767));
    }
}

void mbtree_propagate_list (x264_t *h, uint16_t *ref_costs, int16_t (*mvs) [2], int16_t *propagate_amount, uint16_t *lowres_costs, int bipred_weight, int mb_y, int len, int list) {
    unsigned  stride = h->mb.i_mb_stride;
    unsigned  width = h->mb.i_mb_width;
    unsigned  height = h->mb.i_mb_height;
    for (unsigned  i = 0; i < len; i++) {
        int lists_used = lowres_costs[i] >> 14;
        if (!(lists_used & (1 << list)))
            continue;
        int listamount = propagate_amount[i];
        if (lists_used == 3)
            listamount = (listamount * bipred_weight + 32) >> 6;
        if (!(((x264_union32_t *) (mvs[i]))->i)) {
            (ref_costs[mb_y * stride + i]) = (((ref_costs[mb_y * stride + i]) + (listamount)) < ((1 << 15) - 1) ? ((ref_costs[mb_y * stride + i]) + (listamount)) : ((1 << 15) - 1));
            continue;
        }
        int x = mvs[i][0];
        int y = mvs[i][1];
        unsigned  mbx = (x >> 5) + i;
        unsigned  mby = (y >> 5) + mb_y;
        unsigned  idx0 = mbx + mby * stride;
        unsigned  idx2 = idx0 + stride;
        x &= 31;
        y &= 31;
        int idx0weight = (32 - y) * (32 - x);
        int idx1weight = (32 - y) * x;
        int idx2weight = y * (32 - x);
        int idx3weight = y * x;
        idx0weight = (idx0weight * listamount + 512) >> 10;
        idx1weight = (idx1weight * listamount + 512) >> 10;
        idx2weight = (idx2weight * listamount + 512) >> 10;
        idx3weight = (idx3weight * listamount + 512) >> 10;
        if (mbx < width - 1 && mby < height - 1) {
            (ref_costs[idx0 + 0]) = (((ref_costs[idx0 + 0]) + (idx0weight)) < ((1 << 15) - 1) ? ((ref_costs[idx0 + 0]) + (idx0weight)) : ((1 << 15) - 1));
            (ref_costs[idx0 + 1]) = (((ref_costs[idx0 + 1]) + (idx1weight)) < ((1 << 15) - 1) ? ((ref_costs[idx0 + 1]) + (idx1weight)) : ((1 << 15) - 1));
            (ref_costs[idx2 + 0]) = (((ref_costs[idx2 + 0]) + (idx2weight)) < ((1 << 15) - 1) ? ((ref_costs[idx2 + 0]) + (idx2weight)) : ((1 << 15) - 1));
            (ref_costs[idx2 + 1]) = (((ref_costs[idx2 + 1]) + (idx3weight)) < ((1 << 15) - 1) ? ((ref_costs[idx2 + 1]) + (idx3weight)) : ((1 << 15) - 1));
        }
        else {
            if (mby < height) {
                if (mbx < width)
                    (ref_costs[idx0 + 0]) = (((ref_costs[idx0 + 0]) + (idx0weight)) < ((1 << 15) - 1) ? ((ref_costs[idx0 + 0]) + (idx0weight)) : ((1 << 15) - 1));
                if (mbx + 1 < width)
                    (ref_costs[idx0 + 1]) = (((ref_costs[idx0 + 1]) + (idx1weight)) < ((1 << 15) - 1) ? ((ref_costs[idx0 + 1]) + (idx1weight)) : ((1 << 15) - 1));
            }
            if (mby + 1 < height) {
                if (mbx < width)
                    (ref_costs[idx2 + 0]) = (((ref_costs[idx2 + 0]) + (idx2weight)) < ((1 << 15) - 1) ? ((ref_costs[idx2 + 0]) + (idx2weight)) : ((1 << 15) - 1));
                if (mbx + 1 < width)
                    (ref_costs[idx2 + 1]) = (((ref_costs[idx2 + 1]) + (idx3weight)) < ((1 << 15) - 1) ? ((ref_costs[idx2 + 1]) + (idx3weight)) : ((1 << 15) - 1));
            }
        }
    }
}

void x264_mc_init (int cpu, x264_mc_functions_t *pf, int cpu_independent) {
    pf->mc_luma = mc_luma;
    pf->get_ref = get_ref;
    pf->mc_chroma = mc_chroma;
    pf->avg[PIXEL_16x16] = pixel_avg_16x16;
    pf->avg[PIXEL_16x8] = pixel_avg_16x8;
    pf->avg[PIXEL_8x16] = pixel_avg_8x16;
    pf->avg[PIXEL_8x8] = pixel_avg_8x8;
    pf->avg[PIXEL_8x4] = pixel_avg_8x4;
    pf->avg[PIXEL_4x16] = pixel_avg_4x16;
    pf->avg[PIXEL_4x8] = pixel_avg_4x8;
    pf->avg[PIXEL_4x4] = pixel_avg_4x4;
    pf->avg[PIXEL_4x2] = pixel_avg_4x2;
    pf->avg[PIXEL_2x8] = pixel_avg_2x8;
    pf->avg[PIXEL_2x4] = pixel_avg_2x4;
    pf->avg[PIXEL_2x2] = pixel_avg_2x2;
    pf->weight = x264_mc_weight_wtab;
    pf->offsetadd = x264_mc_weight_wtab;
    pf->offsetsub = x264_mc_weight_wtab;
    pf->weight_cache = x264_weight_cache;
    pf->copy_16x16_unaligned = mc_copy_w16;
    pf->copy[PIXEL_16x16] = mc_copy_w16;
    pf->copy[PIXEL_8x8] = mc_copy_w8;
    pf->copy[PIXEL_4x4] = mc_copy_w4;
    pf->store_interleave_chroma = store_interleave_chroma;
    pf->load_deinterleave_chroma_fenc = load_deinterleave_chroma_fenc;
    pf->load_deinterleave_chroma_fdec = load_deinterleave_chroma_fdec;
    pf->plane_copy = x264_plane_copy_c;
    pf->plane_copy_interleave = x264_plane_copy_interleave_c;
    pf->plane_copy_deinterleave = x264_plane_copy_deinterleave_c;
    pf->plane_copy_deinterleave_rgb = x264_plane_copy_deinterleave_rgb_c;
    pf->plane_copy_deinterleave_v210 = x264_plane_copy_deinterleave_v210_c;
    pf->hpel_filter = hpel_filter;
    pf->prefetch_fenc_420 = prefetch_fenc_null;
    pf->prefetch_fenc_422 = prefetch_fenc_null;
    pf->prefetch_ref = prefetch_ref_null;
    pf->memcpy_aligned = memcpy;
    pf->memzero_aligned = memzero_aligned;
    pf->frame_init_lowres_core = frame_init_lowres_core;
    pf->integral_init4h = integral_init4h;
    pf->integral_init8h = integral_init8h;
    pf->integral_init4v = integral_init4v;
    pf->integral_init8v = integral_init8v;
    pf->mbtree_propagate_cost = mbtree_propagate_cost;
    pf->mbtree_propagate_list = mbtree_propagate_list;
    if (cpu_independent) {
        pf->mbtree_propagate_cost = mbtree_propagate_cost;
        pf->mbtree_propagate_list = mbtree_propagate_list;
    }
}

void x264_frame_filter (x264_t *h, x264_frame_t *frame, int mb_y, int b_end) {
    const int b_interlaced = 0;
    int start = mb_y * 16 - 8;
    int height = (b_end ? frame->i_lines[0] + 16 * 0 : (mb_y + b_interlaced) * 16) + 8;
    if (mb_y & b_interlaced)
        return;
    for (int p = 0; p < ((h->sps->i_chroma_format_idc == CHROMA_444) ? 3 : 1); p++) {
        int stride = frame->i_stride[p];
        const int width = frame->i_width[p];
        int offs = start * stride - 8;
        if (!b_interlaced || h->mb.b_adaptive_mbaff)
            h->mc.hpel_filter (frame->filtered[p][1] + offs, frame->filtered[p][2] + offs, frame->filtered[p][3] + offs, frame->plane[p] + offs, stride, width +16, height -start, h->scratch_buffer);
        if (b_interlaced) {
            stride = frame->i_stride[p] << 1;
            start = (mb_y * 16 >> 1) - 8;
            int height_fld = ((b_end ? frame->i_lines[p] : mb_y * 16) >> 1) + 8;
            offs = start * stride - 8;
            for (int i = 0; i < 2; i++, offs += frame->i_stride[p]) {
                h->mc.hpel_filter (frame->filtered_fld[p][1] + offs, frame->filtered_fld[p][2] + offs, frame->filtered_fld[p][3] + offs, frame->plane_fld[p] + offs, stride, width +16, height_fld -start, h->scratch_buffer);
            }
        }
    }
    if (frame->integral) {
        int stride = frame->i_stride[0];
        if (start < 0) {
            memset (frame -> integral - 32 * stride - 32, 0, stride * sizeof (uint16_t));
            start = -32;
        }
        if (b_end)
            height += 32 - 9;
        for (int y = start; y < height; y++) {
            pixel *pix = frame->plane[0] + y * stride - 32;
            uint16_t *sum8 = frame->integral + (y + 1) * stride - 32;
            uint16_t *sum4;
            if (h->frames.b_have_sub8x8_esa) {
                h->mc.integral_init4h (sum8, pix, stride);
                sum8 -= 8 * stride;
                sum4 = sum8 + stride * (frame->i_lines[0] + 32 * 2);
                if (y >= 8 - 32)
                    h->mc.integral_init4v (sum8, sum4, stride);
            }
            else {
                h->mc.integral_init8h (sum8, pix, stride);
                if (y >= 8 - 32)
                    h->mc.integral_init8v (sum8 -8 * stride, stride);
            }
        }
    }
}

