/*****************************************************************************
 * opencl.h: OpenCL structures and defines
 *****************************************************************************
 * Copyright (C) 2012-2014 x264 project
 *
 * Authors: Steve Borho <sborho@multicorewareinc.com>
 *          Anton Mitrofanov <BugMaster@narod.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 * This program is also available under a commercial proprietary license.
 * For more information, contact us at licensing@x264.com.
 *****************************************************************************/

#ifndef X264_OPENCL_H
#define X264_OPENCL_H

#include "cl.h"



/* Number of downscale resolutions to use for motion search */
#define NUM_IMAGE_SCALES 4

/* Number of PCIe copies that can be queued before requiring a flush */
#define MAX_FINISH_COPIES 1024

/* Size (in bytes) of the page-locked buffer used for PCIe xfers */
#define PAGE_LOCKED_BUF_SIZE 32 * 1024 * 1024

/* Platform API */
typedef cl_int ( *clGetPlatformIDs_func)
( cl_uint ,
    cl_platform_id * ,
    cl_uint * );

typedef cl_int ( *clGetPlatformInfo_func)
( cl_platform_id ,
    cl_platform_info ,
    size_t ,
    void * ,
    size_t * );


typedef cl_int ( *clGetDeviceIDs_func)
( cl_platform_id ,
    cl_device_type ,
    cl_uint ,
    cl_device_id * ,
    cl_uint * );

typedef cl_int ( *clGetDeviceInfo_func)
( cl_device_id ,
    cl_device_info ,
    size_t ,
    void * ,
    size_t * );

typedef cl_int ( *clCreateSubDevices_func)
( cl_device_id ,
    const cl_device_partition_property * ,
    cl_uint ,
    cl_device_id * ,
    cl_uint * );

typedef cl_int ( *clRetainDevice_func)
( cl_device_id );

typedef cl_int ( *clReleaseDevice_func)
( cl_device_id );


typedef cl_context ( *clCreateContext_func)
( const cl_context_properties * ,
    cl_uint ,
    const cl_device_id * ,
    void ( * )(const char *, const void *, size_t, void *),
    void * ,
    cl_int * );

typedef cl_context ( *clCreateContextFromType_func)
( const cl_context_properties * ,
    cl_device_type ,
    void ( * )(const char *, const void *, size_t, void *),
    void * ,
    cl_int * );

typedef cl_int ( *clRetainContext_func)
( cl_context );

typedef cl_int ( *clReleaseContext_func)
( cl_context );

typedef cl_int ( *clGetContextInfo_func)
( cl_context ,
    cl_context_info ,
    size_t ,
    void * ,
    size_t * );


typedef cl_command_queue ( *clCreateCommandQueue_func)
( cl_context ,
    cl_device_id ,
    cl_command_queue_properties ,
    cl_int * );

typedef cl_int ( *clRetainCommandQueue_func)
( cl_command_queue );

typedef cl_int ( *clReleaseCommandQueue_func)
( cl_command_queue );

typedef cl_int ( *clGetCommandQueueInfo_func)
( cl_command_queue ,
    cl_command_queue_info ,
    size_t ,
    void * ,
    size_t * );


typedef cl_mem ( *clCreateBuffer_func)
( cl_context ,
    cl_mem_flags ,
    size_t ,
    void * ,
    cl_int * );

typedef cl_mem ( *clCreateSubBuffer_func)
( cl_mem ,
    cl_mem_flags ,
    cl_buffer_create_type ,
    const void * ,
    cl_int * );

typedef cl_mem ( *clCreateImage_func)
( cl_context ,
    cl_mem_flags ,
    const cl_image_format * ,
    const cl_image_desc * ,
    void * ,
    cl_int * );

typedef cl_int ( *clRetainMemObject_func)
( cl_mem );

typedef cl_int ( *clReleaseMemObject_func)
( cl_mem );

typedef cl_int ( *clGetSupportedImageFormats_func)
( cl_context ,
    cl_mem_flags ,
    cl_mem_object_type ,
    cl_uint ,
    cl_image_format * ,
    cl_uint * );

typedef cl_int ( *clGetMemObjectInfo_func)
( cl_mem ,
    cl_mem_info ,
    size_t ,
    void * ,
    size_t * );

typedef cl_int ( *clGetImageInfo_func)
( cl_mem ,
    cl_image_info ,
    size_t ,
    void * ,
    size_t * );

typedef cl_int ( *clSetMemObjectDestructorCallback_func)
( cl_mem ,
    void ( * )( cl_mem , void* ),
    void * );


typedef cl_sampler ( *clCreateSampler_func)
( cl_context ,
    cl_bool ,
    cl_addressing_mode ,
    cl_filter_mode ,
    cl_int * );

typedef cl_int ( *clRetainSampler_func)
( cl_sampler );

typedef cl_int ( *clReleaseSampler_func)
( cl_sampler );

typedef cl_int ( *clGetSamplerInfo_func)
( cl_sampler ,
    cl_sampler_info ,
    size_t ,
    void * ,
    size_t * );


typedef cl_program ( *clCreateProgramWithSource_func)
( cl_context ,
    cl_uint ,
    const char ** ,
    const size_t * ,
    cl_int * );

typedef cl_program ( *clCreateProgramWithBinary_func)
( cl_context ,
    cl_uint ,
    const cl_device_id * ,
    const size_t * ,
    const unsigned char ** ,
    cl_int * ,
    cl_int * );

typedef cl_program ( *clCreateProgramWithBuiltInKernels_func)
( cl_context ,
    cl_uint ,
    const cl_device_id * ,
    const char * ,
    cl_int * );

typedef cl_int ( *clRetainProgram_func)
( cl_program );

typedef cl_int ( *clReleaseProgram_func)
( cl_program );

typedef cl_int ( *clBuildProgram_func)
( cl_program ,
    cl_uint ,
    const cl_device_id * ,
    const char * ,
    void ( * )(cl_program , void * ),
    void * );

typedef cl_int ( *clCompileProgram_func)
( cl_program ,
    cl_uint ,
    const cl_device_id * ,
    const char * ,
    cl_uint ,
    const cl_program * ,
    const char ** ,
    void ( * )(cl_program , void * ),
    void * );

typedef cl_program ( *clLinkProgram_func)
( cl_context ,
    cl_uint ,
    const cl_device_id * ,
    const char * ,
    cl_uint ,
    const cl_program * ,
    void ( * )(cl_program , void * ),
    void * ,
    cl_int * );


typedef cl_int ( *clUnloadPlatformCompiler_func)
( cl_platform_id );

typedef cl_int ( *clGetProgramInfo_func)
( cl_program ,
    cl_program_info ,
    size_t ,
    void * ,
    size_t * );

typedef cl_int ( *clGetProgramBuildInfo_func)
( cl_program ,
    cl_device_id ,
    cl_program_build_info ,
    size_t ,
    void * ,
    size_t * );


typedef cl_kernel ( *clCreateKernel_func)
( cl_program ,
    const char * ,
    cl_int * );

typedef cl_int ( *clCreateKernelsInProgram_func)
( cl_program ,
    cl_uint ,
    cl_kernel * ,
    cl_uint * );

typedef cl_int ( *clRetainKernel_func)
( cl_kernel );

typedef cl_int ( *clReleaseKernel_func)
( cl_kernel );

typedef cl_int ( *clSetKernelArg_func)
( cl_kernel ,
    cl_uint ,
    size_t ,
    const void * );

typedef cl_int ( *clGetKernelInfo_func)
( cl_kernel ,
    cl_kernel_info ,
    size_t ,
    void * ,
    size_t * );

typedef cl_int ( *clGetKernelArgInfo_func)
( cl_kernel ,
    cl_uint ,
    cl_kernel_arg_info ,
    size_t ,
    void * ,
    size_t * );

typedef cl_int ( *clGetKernelWorkGroupInfo_func)
( cl_kernel ,
    cl_device_id ,
    cl_kernel_work_group_info ,
    size_t ,
    void * ,
    size_t * );


typedef cl_int ( *clWaitForEvents_func)
( cl_uint ,
    const cl_event * );

typedef cl_int ( *clGetEventInfo_func)
( cl_event ,
    cl_event_info ,
    size_t ,
    void * ,
    size_t * );

typedef cl_event ( *clCreateUserEvent_func)
( cl_context ,
    cl_int * );

typedef cl_int ( *clRetainEvent_func)
( cl_event );

typedef cl_int ( *clReleaseEvent_func)
( cl_event );

typedef cl_int ( *clSetUserEventStatus_func)
( cl_event ,
    cl_int );

typedef cl_int ( *clSetEventCallback_func)
( cl_event ,
    cl_int ,
    void ( * )(cl_event, cl_int, void *),
    void * );


typedef cl_int ( *clGetEventProfilingInfo_func)
( cl_event ,
    cl_profiling_info ,
    size_t ,
    void * ,
    size_t * );


typedef cl_int ( *clFlush_func)
( cl_command_queue );

typedef cl_int ( *clFinish_func)
( cl_command_queue );


typedef cl_int ( *clEnqueueReadBuffer_func)
( cl_command_queue ,
    cl_mem ,
    cl_bool ,
    size_t ,
    size_t ,
    void * ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueReadBufferRect_func)
( cl_command_queue ,
    cl_mem ,
    cl_bool ,
    const size_t * ,
    const size_t * ,
    const size_t * ,
    size_t ,
    size_t ,
    size_t ,
    size_t ,
    void * ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueWriteBuffer_func)
( cl_command_queue ,
    cl_mem ,
    cl_bool ,
    size_t ,
    size_t ,
    const void * ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueWriteBufferRect_func)
( cl_command_queue ,
    cl_mem ,
    cl_bool ,
    const size_t * ,
    const size_t * ,
    const size_t * ,
    size_t ,
    size_t ,
    size_t ,
    size_t ,
    const void * ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueFillBuffer_func)
( cl_command_queue ,
    cl_mem ,
    const void * ,
    size_t ,
    size_t ,
    size_t ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueCopyBuffer_func)
( cl_command_queue ,
    cl_mem ,
    cl_mem ,
    size_t ,
    size_t ,
    size_t ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueCopyBufferRect_func)
( cl_command_queue ,
    cl_mem ,
    cl_mem ,
    const size_t * ,
    const size_t * ,
    const size_t * ,
    size_t ,
    size_t ,
    size_t ,
    size_t ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueReadImage_func)
( cl_command_queue ,
    cl_mem ,
    cl_bool ,
    const size_t * ,
    const size_t * ,
    size_t ,
    size_t ,
    void * ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueWriteImage_func)
( cl_command_queue ,
    cl_mem ,
    cl_bool ,
    const size_t * ,
    const size_t * ,
    size_t ,
    size_t ,
    const void * ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueFillImage_func)
( cl_command_queue ,
    cl_mem ,
    const void * ,
    const size_t * ,
    const size_t * ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueCopyImage_func)
( cl_command_queue ,
    cl_mem ,
    cl_mem ,
    const size_t * ,
    const size_t * ,
    const size_t * ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueCopyImageToBuffer_func)
( cl_command_queue ,
    cl_mem ,
    cl_mem ,
    const size_t * ,
    const size_t * ,
    size_t ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueCopyBufferToImage_func)
( cl_command_queue ,
    cl_mem ,
    cl_mem ,
    size_t ,
    const size_t * ,
    const size_t * ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef void * ( *clEnqueueMapBuffer_func)
( cl_command_queue ,
    cl_mem ,
    cl_bool ,
    cl_map_flags ,
    size_t ,
    size_t ,
    cl_uint ,
    const cl_event * ,
    cl_event * ,
    cl_int * );

typedef void * ( *clEnqueueMapImage_func)
( cl_command_queue ,
    cl_mem ,
    cl_bool ,
    cl_map_flags ,
    const size_t * ,
    const size_t * ,
    size_t * ,
    size_t * ,
    cl_uint ,
    const cl_event * ,
    cl_event * ,
    cl_int * );

typedef cl_int ( *clEnqueueUnmapMemObject_func)
( cl_command_queue ,
    cl_mem ,
    void * ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueMigrateMemObjects_func)
( cl_command_queue ,
    cl_uint ,
    const cl_mem * ,
    cl_mem_migration_flags ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueNDRangeKernel_func)
( cl_command_queue ,
    cl_kernel ,
    cl_uint ,
    const size_t * ,
    const size_t * ,
    const size_t * ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueTask_func)
( cl_command_queue ,
    cl_kernel ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueNativeKernel_func)
( cl_command_queue ,
    void ( * )(void *),
    void * ,
    size_t ,
    cl_uint ,
    const cl_mem * ,
    const void ** ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueMarkerWithWaitList_func)
( cl_command_queue ,
    cl_uint ,
    const cl_event * ,
    cl_event * );

typedef cl_int ( *clEnqueueBarrierWithWaitList_func)
( cl_command_queue ,
    cl_uint ,
    const cl_event * ,
    cl_event * );
typedef void * ( *clGetExtensionFunctionAddressForPlatform_func)
( cl_platform_id ,
    const char * );



typedef cl_mem ( *clCreateImage2D_func)
( cl_context ,
    cl_mem_flags ,
    const cl_image_format * ,
    size_t ,
    size_t ,
    size_t ,
    void * ,
    cl_int * );

typedef cl_mem ( *clCreateImage3D_func)
( cl_context ,
    cl_mem_flags ,
    const cl_image_format * ,
    size_t ,
    size_t ,
    size_t ,
    size_t ,
    size_t ,
    void * ,
    cl_int * );

typedef cl_int ( *clEnqueueMarker_func)
( cl_command_queue ,
    cl_event * );

typedef cl_int ( *clEnqueueWaitForEvents_func)
( cl_command_queue ,
    cl_uint ,
    const cl_event * );

typedef cl_int ( *clEnqueueBarrier_func)
( cl_command_queue );

typedef cl_int ( *clUnloadCompiler_func)
( void);

typedef void * ( *clGetExtensionFunctionAddress_func)
( const char * );

struct _x264_opencl_function_t
{
        void *library;

    clBuildProgram_func clBuildProgram;
    clCreateBuffer_func clCreateBuffer;
    clCreateCommandQueue_func clCreateCommandQueue;
    clCreateContext_func clCreateContext;
    clCreateImage2D_func clCreateImage2D;
    clCreateKernel_func clCreateKernel;
    clCreateProgramWithBinary_func clCreateProgramWithBinary;
    clCreateProgramWithSource_func clCreateProgramWithSource;
    clEnqueueCopyBuffer_func clEnqueueCopyBuffer;
    clEnqueueMapBuffer_func clEnqueueMapBuffer;
    clEnqueueNDRangeKernel_func clEnqueueNDRangeKernel;
    clEnqueueReadBuffer_func clEnqueueReadBuffer;
    clEnqueueWriteBuffer_func clEnqueueWriteBuffer;
    clFinish_func clFinish;
    clGetCommandQueueInfo_func clGetCommandQueueInfo;
    clGetDeviceIDs_func clGetDeviceIDs;
    clGetDeviceInfo_func clGetDeviceInfo;
    clGetKernelWorkGroupInfo_func clGetKernelWorkGroupInfo;
    clGetPlatformIDs_func clGetPlatformIDs;
    clGetProgramBuildInfo_func clGetProgramBuildInfo;
    clGetProgramInfo_func clGetProgramInfo;
    clGetSupportedImageFormats_func clGetSupportedImageFormats;
    clReleaseCommandQueue_func clReleaseCommandQueue;
    clReleaseContext_func clReleaseContext;
    clReleaseKernel_func clReleaseKernel;
    clReleaseMemObject_func clReleaseMemObject;
    clReleaseProgram_func clReleaseProgram;
    clSetKernelArg_func clSetKernelArg;
};

typedef struct _x264_opencl_function_t x264_opencl_function_t;

struct _x264_opencl_t
{
    x264_opencl_function_t *ocl;

    cl_context       context;
    cl_device_id     device;
    cl_command_queue queue;

    cl_program  lookahead_program;
    cl_int      last_buf;

    cl_mem      page_locked_buffer;
    char       *page_locked_ptr;
    int         pl_occupancy;

    struct
    {
        void *src;
        void *dest;
        int   bytes;
    } copies[MAX_FINISH_COPIES];
    int         num_copies;

    int         b_device_AMD_SI;
    int         b_fatal_error;
    int         lookahead_thread_pri;
    int         opencl_thread_pri;

    /* downscale lowres luma */
    cl_kernel   downscale_hpel_kernel;
    cl_kernel   downscale_kernel1;
    cl_kernel   downscale_kernel2;
    cl_mem      luma_16x16_image[2];

    /* weightp filtering */
    cl_kernel   weightp_hpel_kernel;
    cl_kernel   weightp_scaled_images_kernel;
    cl_mem      weighted_scaled_images[NUM_IMAGE_SCALES];
    cl_mem      weighted_luma_hpel;

    /* intra */
    cl_kernel   memset_kernel;
    cl_kernel   intra_kernel;
    cl_kernel   rowsum_intra_kernel;
    cl_mem      row_satds[2];

    /* hierarchical motion estimation */
    cl_kernel   hme_kernel;
    cl_kernel   subpel_refine_kernel;
    cl_mem      mv_buffers[2];
    cl_mem      lowres_mv_costs;
    cl_mem      mvp_buffer;

    /* bidir */
    cl_kernel   mode_select_kernel;
    cl_kernel   rowsum_inter_kernel;
    cl_mem      lowres_costs[2];
    cl_mem      frame_stats[2]; /* cost_est, cost_est_aq, intra_mbs */
};

typedef struct _x264_opencl_t x264_opencl_t;


struct _x264_frame_opencl_t
{
    x264_opencl_function_t *ocl;

    cl_mem scaled_image2Ds[NUM_IMAGE_SCALES];
    cl_mem luma_hpel;
    cl_mem inv_qscale_factor;
    cl_mem intra_cost;
    cl_mem lowres_mvs0;
    cl_mem lowres_mvs1;
    cl_mem lowres_mv_costs0;
    cl_mem lowres_mv_costs1;
};

typedef struct _x264_frame_opencl_t x264_frame_opencl_t;


typedef struct x264_frame x264_frame;

x264_opencl_function_t *x264_opencl_load_library( void );
void x264_opencl_close_library( x264_opencl_function_t *ocl );

int x264_opencl_lookahead_init( x264_t *h );
void x264_opencl_lookahead_delete( x264_t *h );

void x264_opencl_frame_delete( x264_frame *frame );

#endif
