
#include "common.h"
const uint8_t x264_size2pixel [5] [5] = {{0,}, {0, PIXEL_4x4, PIXEL_8x4, 0, 0}, {0, PIXEL_4x8, PIXEL_8x8, 0, PIXEL_16x8}, {0,}, {0, 0, PIXEL_8x16, 0, PIXEL_16x16}};
const uint8_t x264_luma2chroma_pixel [4] [7] = {{0}, {PIXEL_8x8, PIXEL_8x4, PIXEL_4x8, PIXEL_4x4, PIXEL_4x2, PIXEL_2x4, PIXEL_2x2}, {PIXEL_8x16, PIXEL_8x8, PIXEL_4x16, PIXEL_4x8, PIXEL_4x4, PIXEL_2x8, PIXEL_2x4}, {PIXEL_16x16, PIXEL_16x8, PIXEL_8x16, PIXEL_8x8, PIXEL_8x4, PIXEL_4x8, PIXEL_4x4},};

int x264_pixel_sad_16x16 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 16; y++) {
        for (int x = 0; x < 16; x++) {
            i_sum += abs (pix1[x] - pix2[x]);
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_sad_16x8 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 8; y++) {
        for (int x = 0; x < 16; x++) {
            i_sum += abs (pix1[x] - pix2[x]);
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_sad_8x16 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 16; y++) {
        for (int x = 0; x < 8; x++) {
            i_sum += abs (pix1[x] - pix2[x]);
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_sad_8x8 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 8; y++) {
        for (int x = 0; x < 8; x++) {
            i_sum += abs (pix1[x] - pix2[x]);
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_sad_8x4 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 4; y++) {
        for (int x = 0; x < 8; x++) {
            i_sum += abs (pix1[x] - pix2[x]);
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_sad_4x16 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 16; y++) {
        for (int x = 0; x < 4; x++) {
            i_sum += abs (pix1[x] - pix2[x]);
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_sad_4x8 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 8; y++) {
        for (int x = 0; x < 4; x++) {
            i_sum += abs (pix1[x] - pix2[x]);
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_sad_4x4 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 4; y++) {
        for (int x = 0; x < 4; x++) {
            i_sum += abs (pix1[x] - pix2[x]);
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_ssd_16x16 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 16; y++) {
        for (int x = 0; x < 16; x++) {
            int d = pix1[x] - pix2[x];
            i_sum += d * d;
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_ssd_16x8 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 8; y++) {
        for (int x = 0; x < 16; x++) {
            int d = pix1[x] - pix2[x];
            i_sum += d * d;
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_ssd_8x16 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 16; y++) {
        for (int x = 0; x < 8; x++) {
            int d = pix1[x] - pix2[x];
            i_sum += d * d;
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_ssd_8x8 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 8; y++) {
        for (int x = 0; x < 8; x++) {
            int d = pix1[x] - pix2[x];
            i_sum += d * d;
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_ssd_8x4 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 4; y++) {
        for (int x = 0; x < 8; x++) {
            int d = pix1[x] - pix2[x];
            i_sum += d * d;
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_ssd_4x16 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 16; y++) {
        for (int x = 0; x < 4; x++) {
            int d = pix1[x] - pix2[x];
            i_sum += d * d;
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_ssd_4x8 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 8; y++) {
        for (int x = 0; x < 4; x++) {
            int d = pix1[x] - pix2[x];
            i_sum += d * d;
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

int x264_pixel_ssd_4x4 (pixel *pix1, intptr_t i_stride_pix1, pixel *pix2, intptr_t i_stride_pix2) {
    int i_sum = 0;
    for (int y = 0; y < 4; y++) {
        for (int x = 0; x < 4; x++) {
            int d = pix1[x] - pix2[x];
            i_sum += d * d;
        }
        pix1 += i_stride_pix1;
        pix2 += i_stride_pix2;
    }
    return i_sum;
}

uint64_t x264_pixel_ssd_wxh (x264_pixel_function_t *pf, pixel *pix1, intptr_t i_pix1, pixel *pix2, intptr_t i_pix2, int i_width, int i_height) {
    uint64_t i_ssd = 0;
    int y;
    int align = !(((intptr_t) pix1 | (intptr_t) pix2 | i_pix1 | i_pix2) & 15);
    for (y = 0; y < i_height - 15; y += 16) {
        int x = 0;
        if (align)
            for (; x < i_width - 15; x += 16)
                i_ssd += pf->ssd[PIXEL_16x16] (pix1 +y * i_pix1 + x, i_pix1, pix2 +y * i_pix2 + x, i_pix2);
        ;
        for (; x < i_width - 7; x += 8)
            i_ssd += pf->ssd[PIXEL_8x16] (pix1 +y * i_pix1 + x, i_pix1, pix2 +y * i_pix2 + x, i_pix2);
        ;
    }
    if (y < i_height - 7)
        for (int x = 0; x < i_width - 7; x += 8)
            i_ssd += pf->ssd[PIXEL_8x8] (pix1 +y * i_pix1 + x, i_pix1, pix2 +y * i_pix2 + x, i_pix2);
    ;
    if (i_width & 7) {
        for (y = 0; y < (i_height & ~7); y++)
            for (int x = i_width & ~7; x < i_width; x++) {
                int d = pix1[y * i_pix1 + x] - pix2[y * i_pix2 + x];
                i_ssd += d * d;
            };
    }
    if (i_height & 7) {
        for (y = i_height & ~7; y < i_height; y++)
            for (int x = 0; x < i_width; x++) {
                int d = pix1[y * i_pix1 + x] - pix2[y * i_pix2 + x];
                i_ssd += d * d;
            };
    }
    return i_ssd;
}

void pixel_ssd_nv12_core (pixel *pixuv1, intptr_t stride1, pixel *pixuv2, intptr_t stride2, int width, int height, uint64_t *ssd_u, uint64_t *ssd_v) {
    *ssd_u = 0, *ssd_v = 0;
    for (int y = 0; y < height; y++, pixuv1 += stride1, pixuv2 += stride2)
        for (int x = 0; x < width; x++) {
            int du = pixuv1[2 * x] - pixuv2[2 * x];
            int dv = pixuv1[2 * x + 1] - pixuv2[2 * x + 1];
            *ssd_u += du * du;
            *ssd_v += dv * dv;
        }
}

void x264_pixel_ssd_nv12 (x264_pixel_function_t *pf, pixel *pix1, intptr_t i_pix1, pixel *pix2, intptr_t i_pix2, int i_width, int i_height, uint64_t *ssd_u, uint64_t *ssd_v) {
    pf->ssd_nv12_core (pix1, i_pix1, pix2, i_pix2, i_width &~7, i_height, ssd_u, ssd_v);
    if (i_width & 7) {
        uint64_t tmp [2];
        pixel_ssd_nv12_core (pix1 + (i_width & ~ 7), i_pix1, pix2 + (i_width & ~ 7), i_pix2, i_width & 7, i_height, & tmp [0], & tmp [1]);
        *ssd_u += tmp[0];
        *ssd_v += tmp[1];
    }
}

uint64_t x264_pixel_var_16x16 (pixel *pix, intptr_t i_stride) {
    uint32_t sum = 0, sqr = 0;
    for (int y = 0; y < 16; y++) {
        for (int x = 0; x < 16; x++) {
            sum += pix[x];
            sqr += pix[x] * pix[x];
        }
        pix += i_stride;
    }
    return sum + ((uint64_t) sqr << 32);
}

uint64_t x264_pixel_var_8x16 (pixel *pix, intptr_t i_stride) {
    uint32_t sum = 0, sqr = 0;
    for (int y = 0; y < 16; y++) {
        for (int x = 0; x < 8; x++) {
            sum += pix[x];
            sqr += pix[x] * pix[x];
        }
        pix += i_stride;
    }
    return sum + ((uint64_t) sqr << 32);
}

uint64_t x264_pixel_var_8x8 (pixel *pix, intptr_t i_stride) {
    uint32_t sum = 0, sqr = 0;
    for (int y = 0; y < 8; y++) {
        for (int x = 0; x < 8; x++) {
            sum += pix[x];
            sqr += pix[x] * pix[x];
        }
        pix += i_stride;
    }
    return sum + ((uint64_t) sqr << 32);
}

int x264_pixel_var2_8x16 (pixel *pix1, intptr_t i_stride1, pixel *pix2, intptr_t i_stride2, int *ssd) {
    int var = 0, sum = 0, sqr = 0;
    for (int y = 0; y < 16; y++) {
        for (int x = 0; x < 8; x++) {
            int diff = pix1[x] - pix2[x];
            sum += diff;
            sqr += diff * diff;
        }
        pix1 += i_stride1;
        pix2 += i_stride2;
    }
    var = sqr - ((int64_t) sum * sum >> 7);
    *ssd = sqr;
    return var;
}

int x264_pixel_var2_8x8 (pixel *pix1, intptr_t i_stride1, pixel *pix2, intptr_t i_stride2, int *ssd) {
    int var = 0, sum = 0, sqr = 0;
    for (int y = 0; y < 8; y++) {
        for (int x = 0; x < 8; x++) {
            int diff = pix1[x] - pix2[x];
            sum += diff;
            sqr += diff * diff;
        }
        pix1 += i_stride1;
        pix2 += i_stride2;
    }
    var = sqr - ((int64_t) sum * sum >> 6);
    *ssd = sqr;
    return var;
}

typedef uint16_t sum_t;
typedef uint32_t sum2_t;

sum2_t abs2 (sum2_t a) {
    sum2_t s = ((a >> ((8 * sizeof (sum_t)) - 1)) & (((sum2_t) 1 << (8 * sizeof (sum_t))) + 1)) * ((sum_t) -1);
    return (a + s) ^ s;
}

int x264_pixel_satd_4x4 (pixel *pix1, intptr_t i_pix1, pixel *pix2, intptr_t i_pix2) {
    sum2_t tmp [4] [2];
    sum2_t a0, a1, a2, a3, b0, b1;
    sum2_t sum = 0;
    for (int i = 0; i < 4; i++, pix1 += i_pix1, pix2 += i_pix2) {
        a0 = pix1[0] - pix2[0];
        a1 = pix1[1] - pix2[1];
        b0 = (a0 + a1) + ((a0 - a1) << (8 * sizeof (sum_t)));
        a2 = pix1[2] - pix2[2];
        a3 = pix1[3] - pix2[3];
        b1 = (a2 + a3) + ((a2 - a3) << (8 * sizeof (sum_t)));
        tmp[i][0] = b0 + b1;
        tmp[i][1] = b0 - b1;
    }
    for (int i = 0; i < 2; i++) {
        {
            sum2_t t0 = tmp[0][i] + tmp[1][i];
            sum2_t t1 = tmp[0][i] - tmp[1][i];
            sum2_t t2 = tmp[2][i] + tmp[3][i];
            sum2_t t3 = tmp[2][i] - tmp[3][i];
            a0 = t0 + t2;
            a2 = t0 - t2;
            a1 = t1 + t3;
            a3 = t1 - t3;
        };
        a0 = abs2 (a0) + abs2 (a1) + abs2 (a2) + abs2 (a3);
        sum += ((sum_t) a0) + (a0 >> (8 * sizeof (sum_t)));
    }
    return sum >> 1;
}

int x264_pixel_satd_8x4 (pixel *pix1, intptr_t i_pix1, pixel *pix2, intptr_t i_pix2) {
    sum2_t tmp [4] [4];
    sum2_t a0, a1, a2, a3;
    sum2_t sum = 0;
    for (int i = 0; i < 4; i++, pix1 += i_pix1, pix2 += i_pix2) {
        a0 = (pix1[0] - pix2[0]) + ((sum2_t) (pix1[4] - pix2[4]) << (8 * sizeof (sum_t)));
        a1 = (pix1[1] - pix2[1]) + ((sum2_t) (pix1[5] - pix2[5]) << (8 * sizeof (sum_t)));
        a2 = (pix1[2] - pix2[2]) + ((sum2_t) (pix1[6] - pix2[6]) << (8 * sizeof (sum_t)));
        a3 = (pix1[3] - pix2[3]) + ((sum2_t) (pix1[7] - pix2[7]) << (8 * sizeof (sum_t)));
        {
            sum2_t t0 = a0 + a1;
            sum2_t t1 = a0 - a1;
            sum2_t t2 = a2 + a3;
            sum2_t t3 = a2 - a3;
            tmp[i][0] = t0 + t2;
            tmp[i][2] = t0 - t2;
            tmp[i][1] = t1 + t3;
            tmp[i][3] = t1 - t3;
        };
    }
    for (int i = 0; i < 4; i++) {
        {
            sum2_t t0 = tmp[0][i] + tmp[1][i];
            sum2_t t1 = tmp[0][i] - tmp[1][i];
            sum2_t t2 = tmp[2][i] + tmp[3][i];
            sum2_t t3 = tmp[2][i] - tmp[3][i];
            a0 = t0 + t2;
            a2 = t0 - t2;
            a1 = t1 + t3;
            a3 = t1 - t3;
        };
        sum += abs2 (a0) + abs2 (a1) + abs2 (a2) + abs2 (a3);
    }
    return (((sum_t) sum) + (sum >> (8 * sizeof (sum_t)))) >> 1;
}

int x264_pixel_satd_16x16 (pixel *pix1, intptr_t i_pix1, pixel *pix2, intptr_t i_pix2) {
    int sum = x264_pixel_satd_8x4 (pix1, i_pix1, pix2, i_pix2) + x264_pixel_satd_8x4 (pix1 +4 * i_pix1, i_pix1, pix2 +4 * i_pix2, i_pix2);
    if (16 == 16)
        sum += x264_pixel_satd_8x4 (pix1 +8, i_pix1, pix2 +8, i_pix2) + x264_pixel_satd_8x4 (pix1 +8 + 4 * i_pix1, i_pix1, pix2 +8 + 4 * i_pix2, i_pix2);
    if (16 == 16)
        sum += x264_pixel_satd_8x4 (pix1 +8 * i_pix1, i_pix1, pix2 +8 * i_pix2, i_pix2) + x264_pixel_satd_8x4 (pix1 +12 * i_pix1, i_pix1, pix2 +12 * i_pix2, i_pix2);
    if (16 == 16 && 16 == 16)
        sum += x264_pixel_satd_8x4 (pix1 +8 + 8 * i_pix1, i_pix1, pix2 +8 + 8 * i_pix2, i_pix2) + x264_pixel_satd_8x4 (pix1 +8 + 12 * i_pix1, i_pix1, pix2 +8 + 12 * i_pix2, i_pix2);
    return sum;
}

int x264_pixel_satd_16x8 (pixel *pix1, intptr_t i_pix1, pixel *pix2, intptr_t i_pix2) {
    int sum = x264_pixel_satd_8x4 (pix1, i_pix1, pix2, i_pix2) + x264_pixel_satd_8x4 (pix1 +4 * i_pix1, i_pix1, pix2 +4 * i_pix2, i_pix2);
    if (16 == 16)
        sum += x264_pixel_satd_8x4 (pix1 +8, i_pix1, pix2 +8, i_pix2) + x264_pixel_satd_8x4 (pix1 +8 + 4 * i_pix1, i_pix1, pix2 +8 + 4 * i_pix2, i_pix2);
    if (8 == 16)
        sum += x264_pixel_satd_8x4 (pix1 +8 * i_pix1, i_pix1, pix2 +8 * i_pix2, i_pix2) + x264_pixel_satd_8x4 (pix1 +12 * i_pix1, i_pix1, pix2 +12 * i_pix2, i_pix2);
    if (16 == 16 && 8 == 16)
        sum += x264_pixel_satd_8x4 (pix1 +8 + 8 * i_pix1, i_pix1, pix2 +8 + 8 * i_pix2, i_pix2) + x264_pixel_satd_8x4 (pix1 +8 + 12 * i_pix1, i_pix1, pix2 +8 + 12 * i_pix2, i_pix2);
    return sum;
}

int x264_pixel_satd_8x16 (pixel *pix1, intptr_t i_pix1, pixel *pix2, intptr_t i_pix2) {
    int sum = x264_pixel_satd_8x4 (pix1, i_pix1, pix2, i_pix2) + x264_pixel_satd_8x4 (pix1 +4 * i_pix1, i_pix1, pix2 +4 * i_pix2, i_pix2);
    if (8 == 16)
        sum += x264_pixel_satd_8x4 (pix1 +8, i_pix1, pix2 +8, i_pix2) + x264_pixel_satd_8x4 (pix1 +8 + 4 * i_pix1, i_pix1, pix2 +8 + 4 * i_pix2, i_pix2);
    if (16 == 16)
        sum += x264_pixel_satd_8x4 (pix1 +8 * i_pix1, i_pix1, pix2 +8 * i_pix2, i_pix2) + x264_pixel_satd_8x4 (pix1 +12 * i_pix1, i_pix1, pix2 +12 * i_pix2, i_pix2);
    if (8 == 16 && 16 == 16)
        sum += x264_pixel_satd_8x4 (pix1 +8 + 8 * i_pix1, i_pix1, pix2 +8 + 8 * i_pix2, i_pix2) + x264_pixel_satd_8x4 (pix1 +8 + 12 * i_pix1, i_pix1, pix2 +8 + 12 * i_pix2, i_pix2);
    return sum;
}

int x264_pixel_satd_8x8 (pixel *pix1, intptr_t i_pix1, pixel *pix2, intptr_t i_pix2) {
    int sum = x264_pixel_satd_8x4 (pix1, i_pix1, pix2, i_pix2) + x264_pixel_satd_8x4 (pix1 +4 * i_pix1, i_pix1, pix2 +4 * i_pix2, i_pix2);
    if (8 == 16)
        sum += x264_pixel_satd_8x4 (pix1 +8, i_pix1, pix2 +8, i_pix2) + x264_pixel_satd_8x4 (pix1 +8 + 4 * i_pix1, i_pix1, pix2 +8 + 4 * i_pix2, i_pix2);
    if (8 == 16)
        sum += x264_pixel_satd_8x4 (pix1 +8 * i_pix1, i_pix1, pix2 +8 * i_pix2, i_pix2) + x264_pixel_satd_8x4 (pix1 +12 * i_pix1, i_pix1, pix2 +12 * i_pix2, i_pix2);
    if (8 == 16 && 8 == 16)
        sum += x264_pixel_satd_8x4 (pix1 +8 + 8 * i_pix1, i_pix1, pix2 +8 + 8 * i_pix2, i_pix2) + x264_pixel_satd_8x4 (pix1 +8 + 12 * i_pix1, i_pix1, pix2 +8 + 12 * i_pix2, i_pix2);
    return sum;
}

int x264_pixel_satd_4x16 (pixel *pix1, intptr_t i_pix1, pixel *pix2, intptr_t i_pix2) {
    int sum = x264_pixel_satd_4x4 (pix1, i_pix1, pix2, i_pix2) + x264_pixel_satd_4x4 (pix1 +4 * i_pix1, i_pix1, pix2 +4 * i_pix2, i_pix2);
    if (4 == 16)
        sum += x264_pixel_satd_4x4 (pix1 +8, i_pix1, pix2 +8, i_pix2) + x264_pixel_satd_4x4 (pix1 +8 + 4 * i_pix1, i_pix1, pix2 +8 + 4 * i_pix2, i_pix2);
    if (16 == 16)
        sum += x264_pixel_satd_4x4 (pix1 +8 * i_pix1, i_pix1, pix2 +8 * i_pix2, i_pix2) + x264_pixel_satd_4x4 (pix1 +12 * i_pix1, i_pix1, pix2 +12 * i_pix2, i_pix2);
    if (4 == 16 && 16 == 16)
        sum += x264_pixel_satd_4x4 (pix1 +8 + 8 * i_pix1, i_pix1, pix2 +8 + 8 * i_pix2, i_pix2) + x264_pixel_satd_4x4 (pix1 +8 + 12 * i_pix1, i_pix1, pix2 +8 + 12 * i_pix2, i_pix2);
    return sum;
}

int x264_pixel_satd_4x8 (pixel *pix1, intptr_t i_pix1, pixel *pix2, intptr_t i_pix2) {
    int sum = x264_pixel_satd_4x4 (pix1, i_pix1, pix2, i_pix2) + x264_pixel_satd_4x4 (pix1 +4 * i_pix1, i_pix1, pix2 +4 * i_pix2, i_pix2);
    if (4 == 16)
        sum += x264_pixel_satd_4x4 (pix1 +8, i_pix1, pix2 +8, i_pix2) + x264_pixel_satd_4x4 (pix1 +8 + 4 * i_pix1, i_pix1, pix2 +8 + 4 * i_pix2, i_pix2);
    if (8 == 16)
        sum += x264_pixel_satd_4x4 (pix1 +8 * i_pix1, i_pix1, pix2 +8 * i_pix2, i_pix2) + x264_pixel_satd_4x4 (pix1 +12 * i_pix1, i_pix1, pix2 +12 * i_pix2, i_pix2);
    if (4 == 16 && 8 == 16)
        sum += x264_pixel_satd_4x4 (pix1 +8 + 8 * i_pix1, i_pix1, pix2 +8 + 8 * i_pix2, i_pix2) + x264_pixel_satd_4x4 (pix1 +8 + 12 * i_pix1, i_pix1, pix2 +8 + 12 * i_pix2, i_pix2);
    return sum;
}

int sa8d_8x8 (pixel *pix1, intptr_t i_pix1, pixel *pix2, intptr_t i_pix2) {
    sum2_t tmp [8] [4];
    sum2_t a0, a1, a2, a3, a4, a5, a6, a7, b0, b1, b2, b3;
    sum2_t sum = 0;
    for (int i = 0; i < 8; i++, pix1 += i_pix1, pix2 += i_pix2) {
        a0 = pix1[0] - pix2[0];
        a1 = pix1[1] - pix2[1];
        b0 = (a0 + a1) + ((a0 - a1) << (8 * sizeof (sum_t)));
        a2 = pix1[2] - pix2[2];
        a3 = pix1[3] - pix2[3];
        b1 = (a2 + a3) + ((a2 - a3) << (8 * sizeof (sum_t)));
        a4 = pix1[4] - pix2[4];
        a5 = pix1[5] - pix2[5];
        b2 = (a4 + a5) + ((a4 - a5) << (8 * sizeof (sum_t)));
        a6 = pix1[6] - pix2[6];
        a7 = pix1[7] - pix2[7];
        b3 = (a6 + a7) + ((a6 - a7) << (8 * sizeof (sum_t)));
        {
            sum2_t t0 = b0 + b1;
            sum2_t t1 = b0 - b1;
            sum2_t t2 = b2 + b3;
            sum2_t t3 = b2 - b3;
            tmp[i][0] = t0 + t2;
            tmp[i][2] = t0 - t2;
            tmp[i][1] = t1 + t3;
            tmp[i][3] = t1 - t3;
        };
    }
    for (int i = 0; i < 4; i++) {
        {
            sum2_t t0 = tmp[0][i] + tmp[1][i];
            sum2_t t1 = tmp[0][i] - tmp[1][i];
            sum2_t t2 = tmp[2][i] + tmp[3][i];
            sum2_t t3 = tmp[2][i] - tmp[3][i];
            a0 = t0 + t2;
            a2 = t0 - t2;
            a1 = t1 + t3;
            a3 = t1 - t3;
        };
        {
            sum2_t t0 = tmp[4][i] + tmp[5][i];
            sum2_t t1 = tmp[4][i] - tmp[5][i];
            sum2_t t2 = tmp[6][i] + tmp[7][i];
            sum2_t t3 = tmp[6][i] - tmp[7][i];
            a4 = t0 + t2;
            a6 = t0 - t2;
            a5 = t1 + t3;
            a7 = t1 - t3;
        };
        b0 = abs2 (a0 +a4) + abs2 (a0 -a4);
        b0 += abs2 (a1 +a5) + abs2 (a1 -a5);
        b0 += abs2 (a2 +a6) + abs2 (a2 -a6);
        b0 += abs2 (a3 +a7) + abs2 (a3 -a7);
        sum += (sum_t) b0 + (b0 >> (8 * sizeof (sum_t)));
    }
    return sum;
}

int x264_pixel_sa8d_8x8 (pixel *pix1, intptr_t i_pix1, pixel *pix2, intptr_t i_pix2) {
    int sum = sa8d_8x8 (pix1, i_pix1, pix2, i_pix2);
    return (sum + 2) >> 2;
}

int x264_pixel_sa8d_16x16 (pixel *pix1, intptr_t i_pix1, pixel *pix2, intptr_t i_pix2) {
    int sum = sa8d_8x8 (pix1, i_pix1, pix2, i_pix2) + sa8d_8x8 (pix1 +8, i_pix1, pix2 +8, i_pix2) + sa8d_8x8 (pix1 +8 * i_pix1, i_pix1, pix2 +8 * i_pix2, i_pix2) + sa8d_8x8 (pix1 +8 + 8 * i_pix1, i_pix1, pix2 +8 + 8 * i_pix2, i_pix2);
    return (sum + 2) >> 2;
}

uint64_t pixel_hadamard_ac (pixel *pix, intptr_t stride) {
    sum2_t tmp [32];
    sum2_t a0, a1, a2, a3, dc;
    sum2_t sum4 = 0, sum8 = 0;
    for (int i = 0; i < 8; i++, pix += stride) {
        sum2_t *t = tmp + (i & 3) + (i & 4) * 4;
        a0 = (pix[0] + pix[1]) + ((sum2_t) (pix[0] - pix[1]) << (8 * sizeof (sum_t)));
        a1 = (pix[2] + pix[3]) + ((sum2_t) (pix[2] - pix[3]) << (8 * sizeof (sum_t)));
        t[0] = a0 + a1;
        t[4] = a0 - a1;
        a2 = (pix[4] + pix[5]) + ((sum2_t) (pix[4] - pix[5]) << (8 * sizeof (sum_t)));
        a3 = (pix[6] + pix[7]) + ((sum2_t) (pix[6] - pix[7]) << (8 * sizeof (sum_t)));
        t[8] = a2 + a3;
        t[12] = a2 - a3;
    }
    for (int i = 0; i < 8; i++) {
        {
            sum2_t t0 = tmp[i * 4 + 0] + tmp[i * 4 + 1];
            sum2_t t1 = tmp[i * 4 + 0] - tmp[i * 4 + 1];
            sum2_t t2 = tmp[i * 4 + 2] + tmp[i * 4 + 3];
            sum2_t t3 = tmp[i * 4 + 2] - tmp[i * 4 + 3];
            a0 = t0 + t2;
            a2 = t0 - t2;
            a1 = t1 + t3;
            a3 = t1 - t3;
        };
        tmp[i * 4 + 0] = a0;
        tmp[i * 4 + 1] = a1;
        tmp[i * 4 + 2] = a2;
        tmp[i * 4 + 3] = a3;
        sum4 += abs2 (a0) + abs2 (a1) + abs2 (a2) + abs2 (a3);
    }
    for (int i = 0; i < 8; i++) {
        {
            sum2_t t0 = tmp[i] + tmp[8 + i];
            sum2_t t1 = tmp[i] - tmp[8 + i];
            sum2_t t2 = tmp[16 + i] + tmp[24 + i];
            sum2_t t3 = tmp[16 + i] - tmp[24 + i];
            a0 = t0 + t2;
            a2 = t0 - t2;
            a1 = t1 + t3;
            a3 = t1 - t3;
        };
        sum8 += abs2 (a0) + abs2 (a1) + abs2 (a2) + abs2 (a3);
    }
    dc = (sum_t) (tmp[0] + tmp[8] + tmp[16] + tmp[24]);
    sum4 = (sum_t) sum4 + (sum4 >> (8 * sizeof (sum_t))) - dc;
    sum8 = (sum_t) sum8 + (sum8 >> (8 * sizeof (sum_t))) - dc;
    return ((uint64_t) sum8 << 32) + sum4;
}

uint64_t x264_pixel_hadamard_ac_16x16 (pixel *pix, intptr_t stride) {
    uint64_t sum = pixel_hadamard_ac (pix, stride);
    if (16 == 16)
        sum += pixel_hadamard_ac (pix +8, stride);
    if (16 == 16)
        sum += pixel_hadamard_ac (pix +8 * stride, stride);
    if (16 == 16 && 16 == 16)
        sum += pixel_hadamard_ac (pix +8 * stride + 8, stride);
    return ((sum >> 34) << 32) + ((uint32_t) sum >> 1);
}

uint64_t x264_pixel_hadamard_ac_16x8 (pixel *pix, intptr_t stride) {
    uint64_t sum = pixel_hadamard_ac (pix, stride);
    if (16 == 16)
        sum += pixel_hadamard_ac (pix +8, stride);
    if (8 == 16)
        sum += pixel_hadamard_ac (pix +8 * stride, stride);
    if (16 == 16 && 8 == 16)
        sum += pixel_hadamard_ac (pix +8 * stride + 8, stride);
    return ((sum >> 34) << 32) + ((uint32_t) sum >> 1);
}

uint64_t x264_pixel_hadamard_ac_8x16 (pixel *pix, intptr_t stride) {
    uint64_t sum = pixel_hadamard_ac (pix, stride);
    if (8 == 16)
        sum += pixel_hadamard_ac (pix +8, stride);
    if (16 == 16)
        sum += pixel_hadamard_ac (pix +8 * stride, stride);
    if (8 == 16 && 16 == 16)
        sum += pixel_hadamard_ac (pix +8 * stride + 8, stride);
    return ((sum >> 34) << 32) + ((uint32_t) sum >> 1);
}

uint64_t x264_pixel_hadamard_ac_8x8 (pixel *pix, intptr_t stride) {
    uint64_t sum = pixel_hadamard_ac (pix, stride);
    if (8 == 16)
        sum += pixel_hadamard_ac (pix +8, stride);
    if (8 == 16)
        sum += pixel_hadamard_ac (pix +8 * stride, stride);
    if (8 == 16 && 8 == 16)
        sum += pixel_hadamard_ac (pix +8 * stride + 8, stride);
    return ((sum >> 34) << 32) + ((uint32_t) sum >> 1);
}

void x264_pixel_sad_x3_16x16 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_sad_16x16 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_16x16 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_16x16 (fenc, 16, pix2, i_stride);
}

void x264_pixel_sad_x4_16x16 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_sad_16x16 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_16x16 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_16x16 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_sad_16x16 (fenc, 16, pix3, i_stride);
}

void x264_pixel_sad_x3_16x8 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_sad_16x8 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_16x8 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_16x8 (fenc, 16, pix2, i_stride);
}

void x264_pixel_sad_x4_16x8 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_sad_16x8 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_16x8 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_16x8 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_sad_16x8 (fenc, 16, pix3, i_stride);
}

void x264_pixel_sad_x3_8x16 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_sad_8x16 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_8x16 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_8x16 (fenc, 16, pix2, i_stride);
}

void x264_pixel_sad_x4_8x16 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_sad_8x16 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_8x16 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_8x16 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_sad_8x16 (fenc, 16, pix3, i_stride);
}

void x264_pixel_sad_x3_8x8 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_sad_8x8 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_8x8 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_8x8 (fenc, 16, pix2, i_stride);
}

void x264_pixel_sad_x4_8x8 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_sad_8x8 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_8x8 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_8x8 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_sad_8x8 (fenc, 16, pix3, i_stride);
}

void x264_pixel_sad_x3_8x4 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_sad_8x4 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_8x4 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_8x4 (fenc, 16, pix2, i_stride);
}

void x264_pixel_sad_x4_8x4 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_sad_8x4 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_8x4 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_8x4 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_sad_8x4 (fenc, 16, pix3, i_stride);
}

void x264_pixel_sad_x3_4x8 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_sad_4x8 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_4x8 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_4x8 (fenc, 16, pix2, i_stride);
}

void x264_pixel_sad_x4_4x8 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_sad_4x8 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_4x8 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_4x8 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_sad_4x8 (fenc, 16, pix3, i_stride);
}

void x264_pixel_sad_x3_4x4 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_sad_4x4 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_4x4 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_4x4 (fenc, 16, pix2, i_stride);
}

void x264_pixel_sad_x4_4x4 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_sad_4x4 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_sad_4x4 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_sad_4x4 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_sad_4x4 (fenc, 16, pix3, i_stride);
}

void x264_pixel_satd_x3_16x16 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_satd_16x16 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_16x16 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_16x16 (fenc, 16, pix2, i_stride);
}

void x264_pixel_satd_x4_16x16 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_satd_16x16 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_16x16 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_16x16 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_satd_16x16 (fenc, 16, pix3, i_stride);
}

void x264_pixel_satd_x3_16x8 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_satd_16x8 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_16x8 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_16x8 (fenc, 16, pix2, i_stride);
}

void x264_pixel_satd_x4_16x8 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_satd_16x8 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_16x8 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_16x8 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_satd_16x8 (fenc, 16, pix3, i_stride);
}

void x264_pixel_satd_x3_8x16 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_satd_8x16 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_8x16 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_8x16 (fenc, 16, pix2, i_stride);
}

void x264_pixel_satd_x4_8x16 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_satd_8x16 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_8x16 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_8x16 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_satd_8x16 (fenc, 16, pix3, i_stride);
}

void x264_pixel_satd_x3_8x8 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_satd_8x8 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_8x8 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_8x8 (fenc, 16, pix2, i_stride);
}

void x264_pixel_satd_x4_8x8 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_satd_8x8 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_8x8 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_8x8 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_satd_8x8 (fenc, 16, pix3, i_stride);
}

void x264_pixel_satd_x3_8x4 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_satd_8x4 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_8x4 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_8x4 (fenc, 16, pix2, i_stride);
}

void x264_pixel_satd_x4_8x4 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_satd_8x4 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_8x4 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_8x4 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_satd_8x4 (fenc, 16, pix3, i_stride);
}

void x264_pixel_satd_x3_4x8 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_satd_4x8 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_4x8 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_4x8 (fenc, 16, pix2, i_stride);
}

void x264_pixel_satd_x4_4x8 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_satd_4x8 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_4x8 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_4x8 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_satd_4x8 (fenc, 16, pix3, i_stride);
}

void x264_pixel_satd_x3_4x4 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, intptr_t i_stride, int scores [3]) {
    scores[0] = x264_pixel_satd_4x4 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_4x4 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_4x4 (fenc, 16, pix2, i_stride);
}

void x264_pixel_satd_x4_4x4 (pixel *fenc, pixel *pix0, pixel *pix1, pixel *pix2, pixel *pix3, intptr_t i_stride, int scores [4]) {
    scores[0] = x264_pixel_satd_4x4 (fenc, 16, pix0, i_stride);
    scores[1] = x264_pixel_satd_4x4 (fenc, 16, pix1, i_stride);
    scores[2] = x264_pixel_satd_4x4 (fenc, 16, pix2, i_stride);
    scores[3] = x264_pixel_satd_4x4 (fenc, 16, pix3, i_stride);
}

void x264_intra_sad_x3_8x8 (pixel *fenc, pixel edge [36], int res [3]) {
    pixel pix [8 * 32];
    x264_predict_8x8_v_c (pix, edge);
    res[0] = x264_pixel_sad_8x8 (pix, 32, fenc, 16);
    x264_predict_8x8_h_c (pix, edge);
    res[1] = x264_pixel_sad_8x8 (pix, 32, fenc, 16);
    x264_predict_8x8_dc_c (pix, edge);
    res[2] = x264_pixel_sad_8x8 (pix, 32, fenc, 16);
}

void x264_intra_sa8d_x3_8x8 (pixel *fenc, pixel edge [36], int res [3]) {
    pixel pix [8 * 32];
    x264_predict_8x8_v_c (pix, edge);
    res[0] = x264_pixel_sa8d_8x8 (pix, 32, fenc, 16);
    x264_predict_8x8_h_c (pix, edge);
    res[1] = x264_pixel_sa8d_8x8 (pix, 32, fenc, 16);
    x264_predict_8x8_dc_c (pix, edge);
    res[2] = x264_pixel_sa8d_8x8 (pix, 32, fenc, 16);
}

void x264_intra_sad_x3_4x4 (pixel *fenc, pixel *fdec, int res [3]) {
    x264_predict_4x4_v_c (fdec);
    res[0] = x264_pixel_sad_4x4 (fdec, 32, fenc, 16);
    x264_predict_4x4_h_c (fdec);
    res[1] = x264_pixel_sad_4x4 (fdec, 32, fenc, 16);
    x264_predict_4x4_dc_c (fdec);
    res[2] = x264_pixel_sad_4x4 (fdec, 32, fenc, 16);
}

void x264_intra_satd_x3_4x4 (pixel *fenc, pixel *fdec, int res [3]) {
    x264_predict_4x4_v_c (fdec);
    res[0] = x264_pixel_satd_4x4 (fdec, 32, fenc, 16);
    x264_predict_4x4_h_c (fdec);
    res[1] = x264_pixel_satd_4x4 (fdec, 32, fenc, 16);
    x264_predict_4x4_dc_c (fdec);
    res[2] = x264_pixel_satd_4x4 (fdec, 32, fenc, 16);
}

void x264_intra_sad_x3_8x8c (pixel *fenc, pixel *fdec, int res [3]) {
    x264_predict_8x8c_dc_c (fdec);
    res[0] = x264_pixel_sad_8x8 (fdec, 32, fenc, 16);
    x264_predict_8x8c_h_c (fdec);
    res[1] = x264_pixel_sad_8x8 (fdec, 32, fenc, 16);
    x264_predict_8x8c_v_c (fdec);
    res[2] = x264_pixel_sad_8x8 (fdec, 32, fenc, 16);
}

void x264_intra_satd_x3_8x8c (pixel *fenc, pixel *fdec, int res [3]) {
    x264_predict_8x8c_dc_c (fdec);
    res[0] = x264_pixel_satd_8x8 (fdec, 32, fenc, 16);
    x264_predict_8x8c_h_c (fdec);
    res[1] = x264_pixel_satd_8x8 (fdec, 32, fenc, 16);
    x264_predict_8x8c_v_c (fdec);
    res[2] = x264_pixel_satd_8x8 (fdec, 32, fenc, 16);
}

void x264_intra_sad_x3_8x16c (pixel *fenc, pixel *fdec, int res [3]) {
    x264_predict_8x16c_dc_c (fdec);
    res[0] = x264_pixel_sad_8x16 (fdec, 32, fenc, 16);
    x264_predict_8x16c_h_c (fdec);
    res[1] = x264_pixel_sad_8x16 (fdec, 32, fenc, 16);
    x264_predict_8x16c_v_c (fdec);
    res[2] = x264_pixel_sad_8x16 (fdec, 32, fenc, 16);
}

void x264_intra_satd_x3_8x16c (pixel *fenc, pixel *fdec, int res [3]) {
    x264_predict_8x16c_dc_c (fdec);
    res[0] = x264_pixel_satd_8x16 (fdec, 32, fenc, 16);
    x264_predict_8x16c_h_c (fdec);
    res[1] = x264_pixel_satd_8x16 (fdec, 32, fenc, 16);
    x264_predict_8x16c_v_c (fdec);
    res[2] = x264_pixel_satd_8x16 (fdec, 32, fenc, 16);
}

void x264_intra_sad_x3_16x16 (pixel *fenc, pixel *fdec, int res [3]) {
    x264_predict_16x16_v_c (fdec);
    res[0] = x264_pixel_sad_16x16 (fdec, 32, fenc, 16);
    x264_predict_16x16_h_c (fdec);
    res[1] = x264_pixel_sad_16x16 (fdec, 32, fenc, 16);
    x264_predict_16x16_dc_c (fdec);
    res[2] = x264_pixel_sad_16x16 (fdec, 32, fenc, 16);
}

void x264_intra_satd_x3_16x16 (pixel *fenc, pixel *fdec, int res [3]) {
    x264_predict_16x16_v_c (fdec);
    res[0] = x264_pixel_satd_16x16 (fdec, 32, fenc, 16);
    x264_predict_16x16_h_c (fdec);
    res[1] = x264_pixel_satd_16x16 (fdec, 32, fenc, 16);
    x264_predict_16x16_dc_c (fdec);
    res[2] = x264_pixel_satd_16x16 (fdec, 32, fenc, 16);
}

void ssim_4x4x2_core (const pixel *pix1, intptr_t stride1, const pixel *pix2, intptr_t stride2, int sums [2] [4]) {
    for (int z = 0; z < 2; z++) {
        uint32_t s1 = 0, s2 = 0, ss = 0, s12 = 0;
        for (int y = 0; y < 4; y++)
            for (int x = 0; x < 4; x++) {
                int a = pix1[x + y * stride1];
                int b = pix2[x + y * stride2];
                s1 += a;
                s2 += b;
                ss += a * a;
                ss += b * b;
                s12 += a * b;
            }
        sums[z][0] = s1;
        sums[z][1] = s2;
        sums[z][2] = ss;
        sums[z][3] = s12;
        pix1 += 4;
        pix2 += 4;
    }
}

float ssim_end1 (int s1, int s2, int ss, int s12) {
    const int ssim_c1 = (int) ( .01 * .01 * ((1 << 8) - 1) * ((1 << 8) - 1) * 64 + .5);
    const int ssim_c2 = (int) ( .03 * .03 * ((1 << 8) - 1) * ((1 << 8) - 1) * 64 * 63 + .5);
    int fs1 = s1;
    int fs2 = s2;
    int fss = ss;
    int fs12 = s12;
    int vars = fss * 64 - fs1 * fs1 - fs2 * fs2;
    int covar = fs12 * 64 - fs1 * fs2;
    return (float) (2 * fs1 * fs2 + ssim_c1) * (float) (2 * covar + ssim_c2) / ((float) (fs1 * fs1 + fs2 * fs2 + ssim_c1) * (float) (vars + ssim_c2));
}

float ssim_end4 (int sum0 [5] [4], int sum1 [5] [4], int width) {
    float ssim = 0.0;
    for (int i = 0; i < width; i++)
        ssim += ssim_end1 (sum0[i][0] + sum0[i + 1][0] + sum1[i][0] + sum1[i + 1][0], sum0[i][1] + sum0[i + 1][1] + sum1[i][1] + sum1[i + 1][1], sum0[i][2] + sum0[i + 1][2] + sum1[i][2] + sum1[i + 1][2], sum0[i][3] + sum0[i + 1][3] + sum1[i][3] + sum1[i + 1][3]);
    return ssim;
}

float x264_pixel_ssim_wxh (x264_pixel_function_t *pf, pixel *pix1, intptr_t stride1, pixel *pix2, intptr_t stride2, int width, int height, void *buf, int *cnt) {
    int z = 0;
    float ssim = 0.0;
    int (*sum0) [4] = buf;
    int (*sum1) [4] = sum0 + (width >> 2) + 3;
    width >>= 2;
    height >>= 2;
    for (int y = 1; y < height; y++) {
        for (; z <= y; z++) {
            do {
                void *t = sum0;
                sum0 = sum1;
                sum1 = t;
            }
            while (0);
            for (int x = 0; x < width; x += 2)
                pf->ssim_4x4x2_core (&pix1[4 * (x + z * stride1)], stride1, &pix2[4 * (x + z * stride2)], stride2, &sum0[x]);
        }
        for (int x = 0; x < width - 1; x += 4)
            ssim += pf->ssim_end4 (sum0 +x, sum1 +x, ((4) < (width - x - 1) ? (4) : (width - x - 1)));
    }
    *cnt = (height - 1) * (width - 1);
    return ssim;
}

int pixel_vsad (pixel *src, intptr_t stride, int height) {
    int score = 0;
    for (int i = 1; i < height; i++, src += stride)
        for (int j = 0; j < 16; j++)
            score += abs (src[j] - src[j + stride]);
    return score;
}

int x264_field_vsad (x264_t *h, int mb_x, int mb_y) {
    int score_field, score_frame;
    int stride = h->fenc->i_stride[0];
    int mb_stride = h->mb.i_mb_stride;
    pixel *fenc = h->fenc->plane[0] + 16 * (mb_x + mb_y * stride);
    int mb_xy = mb_x + mb_y * mb_stride;
    int mbpair_height = ((h->param.i_height - mb_y * 16) < (32) ? (h->param.i_height - mb_y * 16) : (32));
    score_frame = h->pixf.vsad (fenc, stride, mbpair_height);
    score_field = h->pixf.vsad (fenc, stride *2, mbpair_height >> 1);
    score_field += h->pixf.vsad (fenc +stride, stride *2, mbpair_height >> 1);
    if (mb_x > 0)
        score_field += 512 - h->mb.field[mb_xy - 1] * 1024;
    if (mb_y > 0)
        score_field += 512 - h->mb.field[mb_xy - mb_stride] * 1024;
    return (score_field < score_frame);
}

int pixel_asd8 (pixel *pix1, intptr_t stride1, pixel *pix2, intptr_t stride2, int height) {
    int sum = 0;
    for (int y = 0; y < height; y++, pix1 += stride1, pix2 += stride2)
        for (int x = 0; x < 8; x++)
            sum += pix1[x] - pix2[x];
    return abs (sum);
}

int x264_pixel_ads4 (int enc_dc [4], uint16_t *sums, int delta, uint16_t *cost_mvx, int16_t *mvs, int width, int thresh) {
    int nmv = 0;
    for (int i = 0; i < width; i++, sums++) {
        int ads = abs (enc_dc[0] - sums[0]) + abs (enc_dc[1] - sums[8]) + abs (enc_dc[2] - sums[delta]) + abs (enc_dc[3] - sums[delta + 8]) + cost_mvx[i];
        if (ads < thresh)
            mvs[nmv++] = i;
    }
    return nmv;
}

int x264_pixel_ads2 (int enc_dc [2], uint16_t *sums, int delta, uint16_t *cost_mvx, int16_t *mvs, int width, int thresh) {
    int nmv = 0;
    for (int i = 0; i < width; i++, sums++) {
        int ads = abs (enc_dc[0] - sums[0]) + abs (enc_dc[1] - sums[delta]) + cost_mvx[i];
        if (ads < thresh)
            mvs[nmv++] = i;
    }
    return nmv;
}

int x264_pixel_ads1 (int enc_dc [1], uint16_t *sums, int delta, uint16_t *cost_mvx, int16_t *mvs, int width, int thresh) {
    int nmv = 0;
    for (int i = 0; i < width; i++, sums++) {
        int ads = abs (enc_dc[0] - sums[0]) + cost_mvx[i];
        if (ads < thresh)
            mvs[nmv++] = i;
    }
    return nmv;
}

void x264_pixel_init (int cpu, x264_pixel_function_t *pixf) {
    memset (pixf, 0, sizeof (* pixf));
    pixf->sad[PIXEL_16x16] = x264_pixel_sad_16x16;
    pixf->sad[PIXEL_16x8] = x264_pixel_sad_16x8;
    pixf->sad[PIXEL_8x16] = x264_pixel_sad_8x16;
    pixf->sad[PIXEL_8x8] = x264_pixel_sad_8x8;
    pixf->sad[PIXEL_8x4] = x264_pixel_sad_8x4;
    pixf->sad[PIXEL_4x8] = x264_pixel_sad_4x8;
    pixf->sad[PIXEL_4x4] = x264_pixel_sad_4x4;
    pixf->sad[PIXEL_4x16] = x264_pixel_sad_4x16;
    ;
    pixf->sad_aligned[PIXEL_16x16] = x264_pixel_sad_16x16;
    pixf->sad_aligned[PIXEL_16x8] = x264_pixel_sad_16x8;
    pixf->sad_aligned[PIXEL_8x16] = x264_pixel_sad_8x16;
    pixf->sad_aligned[PIXEL_8x8] = x264_pixel_sad_8x8;
    pixf->sad_aligned[PIXEL_8x4] = x264_pixel_sad_8x4;
    pixf->sad_aligned[PIXEL_4x8] = x264_pixel_sad_4x8;
    pixf->sad_aligned[PIXEL_4x4] = x264_pixel_sad_4x4;
    pixf->sad_aligned[PIXEL_4x16] = x264_pixel_sad_4x16;
    ;
    pixf->sad_x3[PIXEL_16x16] = x264_pixel_sad_x3_16x16;
    pixf->sad_x3[PIXEL_16x8] = x264_pixel_sad_x3_16x8;
    pixf->sad_x3[PIXEL_8x16] = x264_pixel_sad_x3_8x16;
    pixf->sad_x3[PIXEL_8x8] = x264_pixel_sad_x3_8x8;
    pixf->sad_x3[PIXEL_8x4] = x264_pixel_sad_x3_8x4;
    pixf->sad_x3[PIXEL_4x8] = x264_pixel_sad_x3_4x8;
    pixf->sad_x3[PIXEL_4x4] = x264_pixel_sad_x3_4x4;
    ;
    pixf->sad_x4[PIXEL_16x16] = x264_pixel_sad_x4_16x16;
    pixf->sad_x4[PIXEL_16x8] = x264_pixel_sad_x4_16x8;
    pixf->sad_x4[PIXEL_8x16] = x264_pixel_sad_x4_8x16;
    pixf->sad_x4[PIXEL_8x8] = x264_pixel_sad_x4_8x8;
    pixf->sad_x4[PIXEL_8x4] = x264_pixel_sad_x4_8x4;
    pixf->sad_x4[PIXEL_4x8] = x264_pixel_sad_x4_4x8;
    pixf->sad_x4[PIXEL_4x4] = x264_pixel_sad_x4_4x4;
    ;
    pixf->ssd[PIXEL_16x16] = x264_pixel_ssd_16x16;
    pixf->ssd[PIXEL_16x8] = x264_pixel_ssd_16x8;
    pixf->ssd[PIXEL_8x16] = x264_pixel_ssd_8x16;
    pixf->ssd[PIXEL_8x8] = x264_pixel_ssd_8x8;
    pixf->ssd[PIXEL_8x4] = x264_pixel_ssd_8x4;
    pixf->ssd[PIXEL_4x8] = x264_pixel_ssd_4x8;
    pixf->ssd[PIXEL_4x4] = x264_pixel_ssd_4x4;
    pixf->ssd[PIXEL_4x16] = x264_pixel_ssd_4x16;
    ;
    pixf->satd[PIXEL_16x16] = x264_pixel_satd_16x16;
    pixf->satd[PIXEL_16x8] = x264_pixel_satd_16x8;
    pixf->satd[PIXEL_8x16] = x264_pixel_satd_8x16;
    pixf->satd[PIXEL_8x8] = x264_pixel_satd_8x8;
    pixf->satd[PIXEL_8x4] = x264_pixel_satd_8x4;
    pixf->satd[PIXEL_4x8] = x264_pixel_satd_4x8;
    pixf->satd[PIXEL_4x4] = x264_pixel_satd_4x4;
    pixf->satd[PIXEL_4x16] = x264_pixel_satd_4x16;
    ;
    pixf->satd_x3[PIXEL_16x16] = x264_pixel_satd_x3_16x16;
    pixf->satd_x3[PIXEL_16x8] = x264_pixel_satd_x3_16x8;
    pixf->satd_x3[PIXEL_8x16] = x264_pixel_satd_x3_8x16;
    pixf->satd_x3[PIXEL_8x8] = x264_pixel_satd_x3_8x8;
    pixf->satd_x3[PIXEL_8x4] = x264_pixel_satd_x3_8x4;
    pixf->satd_x3[PIXEL_4x8] = x264_pixel_satd_x3_4x8;
    pixf->satd_x3[PIXEL_4x4] = x264_pixel_satd_x3_4x4;
    ;
    pixf->satd_x4[PIXEL_16x16] = x264_pixel_satd_x4_16x16;
    pixf->satd_x4[PIXEL_16x8] = x264_pixel_satd_x4_16x8;
    pixf->satd_x4[PIXEL_8x16] = x264_pixel_satd_x4_8x16;
    pixf->satd_x4[PIXEL_8x8] = x264_pixel_satd_x4_8x8;
    pixf->satd_x4[PIXEL_8x4] = x264_pixel_satd_x4_8x4;
    pixf->satd_x4[PIXEL_4x8] = x264_pixel_satd_x4_4x8;
    pixf->satd_x4[PIXEL_4x4] = x264_pixel_satd_x4_4x4;
    ;
    pixf->hadamard_ac[PIXEL_16x16] = x264_pixel_hadamard_ac_16x16;
    pixf->hadamard_ac[PIXEL_16x8] = x264_pixel_hadamard_ac_16x8;
    pixf->hadamard_ac[PIXEL_8x16] = x264_pixel_hadamard_ac_8x16;
    pixf->hadamard_ac[PIXEL_8x8] = x264_pixel_hadamard_ac_8x8;
    ;
    pixf->ads[PIXEL_16x16] = x264_pixel_ads4;
    pixf->ads[PIXEL_16x8] = x264_pixel_ads2;
    pixf->ads[PIXEL_8x8] = x264_pixel_ads1;
    ;
    pixf->sa8d[PIXEL_16x16] = x264_pixel_sa8d_16x16;
    pixf->sa8d[PIXEL_8x8] = x264_pixel_sa8d_8x8;
    pixf->var[PIXEL_16x16] = x264_pixel_var_16x16;
    pixf->var[PIXEL_8x16] = x264_pixel_var_8x16;
    pixf->var[PIXEL_8x8] = x264_pixel_var_8x8;
    pixf->var2[PIXEL_8x16] = x264_pixel_var2_8x16;
    pixf->var2[PIXEL_8x8] = x264_pixel_var2_8x8;
    pixf->ssd_nv12_core = pixel_ssd_nv12_core;
    pixf->ssim_4x4x2_core = ssim_4x4x2_core;
    pixf->ssim_end4 = ssim_end4;
    pixf->vsad = pixel_vsad;
    pixf->asd8 = pixel_asd8;
    pixf->intra_sad_x3_4x4 = x264_intra_sad_x3_4x4;
    pixf->intra_satd_x3_4x4 = x264_intra_satd_x3_4x4;
    pixf->intra_sad_x3_8x8 = x264_intra_sad_x3_8x8;
    pixf->intra_sa8d_x3_8x8 = x264_intra_sa8d_x3_8x8;
    pixf->intra_sad_x3_8x8c = x264_intra_sad_x3_8x8c;
    pixf->intra_satd_x3_8x8c = x264_intra_satd_x3_8x8c;
    pixf->intra_sad_x3_8x16c = x264_intra_sad_x3_8x16c;
    pixf->intra_satd_x3_8x16c = x264_intra_satd_x3_8x16c;
    pixf->intra_sad_x3_16x16 = x264_intra_sad_x3_16x16;
    pixf->intra_satd_x3_16x16 = x264_intra_satd_x3_16x16;
    pixf->ads[PIXEL_8x16] = pixf->ads[PIXEL_8x4] = pixf->ads[PIXEL_4x8] = pixf->ads[PIXEL_16x8];
    pixf->ads[PIXEL_4x4] = pixf->ads[PIXEL_8x8];
}

