
#include "common.h"
const uint8_t x264_mb_chroma_pred_mode_fix [7] = {I_PRED_CHROMA_DC, I_PRED_CHROMA_H, I_PRED_CHROMA_V, I_PRED_CHROMA_P, I_PRED_CHROMA_DC, I_PRED_CHROMA_DC, I_PRED_CHROMA_DC};
const uint8_t x264_mb_pred_mode16x16_fix [7] = {I_PRED_16x16_V, I_PRED_16x16_H, I_PRED_16x16_DC, I_PRED_16x16_P, I_PRED_16x16_DC, I_PRED_16x16_DC, I_PRED_16x16_DC};
const int8_t x264_mb_pred_mode4x4_fix [13] = {-1, I_PRED_4x4_V, I_PRED_4x4_H, I_PRED_4x4_DC, I_PRED_4x4_DDL, I_PRED_4x4_DDR, I_PRED_4x4_VR, I_PRED_4x4_HD, I_PRED_4x4_VL, I_PRED_4x4_HU, I_PRED_4x4_DC, I_PRED_4x4_DC, I_PRED_4x4_DC};

void x264_predict_16x16_dc_c (pixel *src) {
    int dc = 0;
    for (int i = 0; i < 16; i++) {
        dc += src[-1 + i * 32];
        dc += src[i - 32];
    }
    pixel4 dcsplat = (((dc + 16) >> 5) * 0x01010101U);
    for (int i = 0; i < 16; i++) {
        (((x264_union32_t *) (src + 0))->i) = dcsplat;
        (((x264_union32_t *) (src + 4))->i) = dcsplat;
        (((x264_union32_t *) (src + 8))->i) = dcsplat;
        (((x264_union32_t *) (src + 12))->i) = dcsplat;
        src += 32;
    };
}

void x264_predict_16x16_dc_left_c (pixel *src) {
    int dc = 0;
    for (int i = 0; i < 16; i++)
        dc += src[-1 + i * 32];
    pixel4 dcsplat = (((dc + 8) >> 4) * 0x01010101U);
    for (int i = 0; i < 16; i++) {
        (((x264_union32_t *) (src + 0))->i) = dcsplat;
        (((x264_union32_t *) (src + 4))->i) = dcsplat;
        (((x264_union32_t *) (src + 8))->i) = dcsplat;
        (((x264_union32_t *) (src + 12))->i) = dcsplat;
        src += 32;
    };
}

void x264_predict_16x16_dc_top_c (pixel *src) {
    int dc = 0;
    for (int i = 0; i < 16; i++)
        dc += src[i - 32];
    pixel4 dcsplat = (((dc + 8) >> 4) * 0x01010101U);
    for (int i = 0; i < 16; i++) {
        (((x264_union32_t *) (src + 0))->i) = dcsplat;
        (((x264_union32_t *) (src + 4))->i) = dcsplat;
        (((x264_union32_t *) (src + 8))->i) = dcsplat;
        (((x264_union32_t *) (src + 12))->i) = dcsplat;
        src += 32;
    };
}

void x264_predict_16x16_dc_128_c (pixel *src) {
    for (int i = 0; i < 16; i++) {
        (((x264_union32_t *) (src + 0))->i) = ((1 << (8 - 1)) * 0x01010101U);
        (((x264_union32_t *) (src + 4))->i) = ((1 << (8 - 1)) * 0x01010101U);
        (((x264_union32_t *) (src + 8))->i) = ((1 << (8 - 1)) * 0x01010101U);
        (((x264_union32_t *) (src + 12))->i) = ((1 << (8 - 1)) * 0x01010101U);
        src += 32;
    };
}

void x264_predict_16x16_h_c (pixel *src) {
    for (int i = 0; i < 16; i++) {
        const pixel4 v = ((src[-1]) * 0x01010101U);
        (((x264_union32_t *) (src + 0))->i) = v;
        (((x264_union32_t *) (src + 4))->i) = v;
        (((x264_union32_t *) (src + 8))->i) = v;
        (((x264_union32_t *) (src + 12))->i) = v;
        src += 32;
    }
}

void x264_predict_16x16_v_c (pixel *src) {
    pixel4 v0 = (((x264_union32_t *) (&src[0 - 32]))->i);
    pixel4 v1 = (((x264_union32_t *) (&src[4 - 32]))->i);
    pixel4 v2 = (((x264_union32_t *) (&src[8 - 32]))->i);
    pixel4 v3 = (((x264_union32_t *) (&src[12 - 32]))->i);
    for (int i = 0; i < 16; i++) {
        (((x264_union32_t *) (src + 0))->i) = v0;
        (((x264_union32_t *) (src + 4))->i) = v1;
        (((x264_union32_t *) (src + 8))->i) = v2;
        (((x264_union32_t *) (src + 12))->i) = v3;
        src += 32;
    }
}

void x264_predict_16x16_p_c (pixel *src) {
    int H = 0, V = 0;
    for (int i = 0; i <= 7; i++) {
        H += (i + 1) * (src[8 + i - 32] - src[6 - i - 32]);
        V += (i + 1) * (src[-1 + (8 + i) * 32] - src[-1 + (6 - i) * 32]);
    }
    int a = 16 * (src[-1 + 15 * 32] + src[15 - 32]);
    int b = (5 * H + 32) >> 6;
    int c = (5 * V + 32) >> 6;
    int i00 = a - b * 7 - c * 7 + 16;
    for (int y = 0; y < 16; y++) {
        int pix = i00;
        for (int x = 0; x < 16; x++) {
            src[x] = x264_clip_pixel (pix >> 5);
            pix += b;
        }
        src += 32;
        i00 += c;
    }
}

void x264_predict_8x8c_dc_128_c (pixel *src) {
    for (int y = 0; y < 8; y++) {
        (((x264_union32_t *) (src + 0))->i) = ((1 << (8 - 1)) * 0x01010101U);
        (((x264_union32_t *) (src + 4))->i) = ((1 << (8 - 1)) * 0x01010101U);
        src += 32;
    }
}

void x264_predict_8x8c_dc_left_c (pixel *src) {
    int dc0 = 0, dc1 = 0;
    for (int y = 0; y < 4; y++) {
        dc0 += src[y * 32 - 1];
        dc1 += src[(y + 4) * 32 - 1];
    }
    pixel4 dc0splat = (((dc0 + 2) >> 2) * 0x01010101U);
    pixel4 dc1splat = (((dc1 + 2) >> 2) * 0x01010101U);
    for (int y = 0; y < 4; y++) {
        (((x264_union32_t *) (src + 0))->i) = dc0splat;
        (((x264_union32_t *) (src + 4))->i) = dc0splat;
        src += 32;
    }
    for (int y = 0; y < 4; y++) {
        (((x264_union32_t *) (src + 0))->i) = dc1splat;
        (((x264_union32_t *) (src + 4))->i) = dc1splat;
        src += 32;
    }
}

void x264_predict_8x8c_dc_top_c (pixel *src) {
    int dc0 = 0, dc1 = 0;
    for (int x = 0; x < 4; x++) {
        dc0 += src[x - 32];
        dc1 += src[x + 4 - 32];
    }
    pixel4 dc0splat = (((dc0 + 2) >> 2) * 0x01010101U);
    pixel4 dc1splat = (((dc1 + 2) >> 2) * 0x01010101U);
    for (int y = 0; y < 8; y++) {
        (((x264_union32_t *) (src + 0))->i) = dc0splat;
        (((x264_union32_t *) (src + 4))->i) = dc1splat;
        src += 32;
    }
}

void x264_predict_8x8c_dc_c (pixel *src) {
    int s0 = 0, s1 = 0, s2 = 0, s3 = 0;
    for (int i = 0; i < 4; i++) {
        s0 += src[i - 32];
        s1 += src[i + 4 - 32];
        s2 += src[-1 + i * 32];
        s3 += src[-1 + (i + 4) * 32];
    }
    pixel4 dc0 = (((s0 + s2 + 4) >> 3) * 0x01010101U);
    pixel4 dc1 = (((s1 + 2) >> 2) * 0x01010101U);
    pixel4 dc2 = (((s3 + 2) >> 2) * 0x01010101U);
    pixel4 dc3 = (((s1 + s3 + 4) >> 3) * 0x01010101U);
    for (int y = 0; y < 4; y++) {
        (((x264_union32_t *) (src + 0))->i) = dc0;
        (((x264_union32_t *) (src + 4))->i) = dc1;
        src += 32;
    }
    for (int y = 0; y < 4; y++) {
        (((x264_union32_t *) (src + 0))->i) = dc2;
        (((x264_union32_t *) (src + 4))->i) = dc3;
        src += 32;
    }
}

void x264_predict_8x8c_h_c (pixel *src) {
    for (int i = 0; i < 8; i++) {
        pixel4 v = ((src[-1]) * 0x01010101U);
        (((x264_union32_t *) (src + 0))->i) = v;
        (((x264_union32_t *) (src + 4))->i) = v;
        src += 32;
    }
}

void x264_predict_8x8c_v_c (pixel *src) {
    pixel4 v0 = (((x264_union32_t *) (src + 0 - 32))->i);
    pixel4 v1 = (((x264_union32_t *) (src + 4 - 32))->i);
    for (int i = 0; i < 8; i++) {
        (((x264_union32_t *) (src + 0))->i) = v0;
        (((x264_union32_t *) (src + 4))->i) = v1;
        src += 32;
    }
}

void x264_predict_8x8c_p_c (pixel *src) {
    int H = 0, V = 0;
    for (int i = 0; i < 4; i++) {
        H += (i + 1) * (src[4 + i - 32] - src[2 - i - 32]);
        V += (i + 1) * (src[-1 + (i + 4) * 32] - src[-1 + (2 - i) * 32]);
    }
    int a = 16 * (src[-1 + 7 * 32] + src[7 - 32]);
    int b = (17 * H + 16) >> 5;
    int c = (17 * V + 16) >> 5;
    int i00 = a - 3 * b - 3 * c + 16;
    for (int y = 0; y < 8; y++) {
        int pix = i00;
        for (int x = 0; x < 8; x++) {
            src[x] = x264_clip_pixel (pix >> 5);
            pix += b;
        }
        src += 32;
        i00 += c;
    }
}

void x264_predict_8x16c_dc_128_c (pixel *src) {
    for (int y = 0; y < 16; y++) {
        (((x264_union32_t *) (src + 0))->i) = ((1 << (8 - 1)) * 0x01010101U);
        (((x264_union32_t *) (src + 4))->i) = ((1 << (8 - 1)) * 0x01010101U);
        src += 32;
    }
}

void x264_predict_8x16c_dc_left_c (pixel *src) {
    for (int i = 0; i < 4; i++) {
        int dc = 0;
        for (int y = 0; y < 4; y++)
            dc += src[y * 32 - 1];
        pixel4 dcsplat = (((dc + 2) >> 2) * 0x01010101U);
        for (int y = 0; y < 4; y++) {
            (((x264_union32_t *) (src + 0))->i) = dcsplat;
            (((x264_union32_t *) (src + 4))->i) = dcsplat;
            src += 32;
        }
    }
}

void x264_predict_8x16c_dc_top_c (pixel *src) {
    int dc0 = 0, dc1 = 0;
    for (int x = 0; x < 4; x++) {
        dc0 += src[x - 32];
        dc1 += src[x + 4 - 32];
    }
    pixel4 dc0splat = (((dc0 + 2) >> 2) * 0x01010101U);
    pixel4 dc1splat = (((dc1 + 2) >> 2) * 0x01010101U);
    for (int y = 0; y < 16; y++) {
        (((x264_union32_t *) (src + 0))->i) = dc0splat;
        (((x264_union32_t *) (src + 4))->i) = dc1splat;
        src += 32;
    }
}

void x264_predict_8x16c_dc_c (pixel *src) {
    int s0 = 0, s1 = 0, s2 = 0, s3 = 0, s4 = 0, s5 = 0;
    for (int i = 0; i < 4; i++) {
        s0 += src[i + 0 - 32];
        s1 += src[i + 4 - 32];
        s2 += src[-1 + (i + 0) * 32];
        s3 += src[-1 + (i + 4) * 32];
        s4 += src[-1 + (i + 8) * 32];
        s5 += src[-1 + (i + 12) * 32];
    }
    pixel4 dc0 = (((s0 + s2 + 4) >> 3) * 0x01010101U);
    pixel4 dc1 = (((s1 + 2) >> 2) * 0x01010101U);
    pixel4 dc2 = (((s3 + 2) >> 2) * 0x01010101U);
    pixel4 dc3 = (((s1 + s3 + 4) >> 3) * 0x01010101U);
    pixel4 dc4 = (((s4 + 2) >> 2) * 0x01010101U);
    pixel4 dc5 = (((s1 + s4 + 4) >> 3) * 0x01010101U);
    pixel4 dc6 = (((s5 + 2) >> 2) * 0x01010101U);
    pixel4 dc7 = (((s1 + s5 + 4) >> 3) * 0x01010101U);
    for (int y = 0; y < 4; y++) {
        (((x264_union32_t *) (src + 0))->i) = dc0;
        (((x264_union32_t *) (src + 4))->i) = dc1;
        src += 32;
    }
    for (int y = 0; y < 4; y++) {
        (((x264_union32_t *) (src + 0))->i) = dc2;
        (((x264_union32_t *) (src + 4))->i) = dc3;
        src += 32;
    }
    for (int y = 0; y < 4; y++) {
        (((x264_union32_t *) (src + 0))->i) = dc4;
        (((x264_union32_t *) (src + 4))->i) = dc5;
        src += 32;
    }
    for (int y = 0; y < 4; y++) {
        (((x264_union32_t *) (src + 0))->i) = dc6;
        (((x264_union32_t *) (src + 4))->i) = dc7;
        src += 32;
    }
}

void x264_predict_8x16c_h_c (pixel *src) {
    for (int i = 0; i < 16; i++) {
        pixel4 v = ((src[-1]) * 0x01010101U);
        (((x264_union32_t *) (src + 0))->i) = v;
        (((x264_union32_t *) (src + 4))->i) = v;
        src += 32;
    }
}

void x264_predict_8x16c_v_c (pixel *src) {
    pixel4 v0 = (((x264_union32_t *) (src + 0 - 32))->i);
    pixel4 v1 = (((x264_union32_t *) (src + 4 - 32))->i);
    for (int i = 0; i < 16; i++) {
        (((x264_union32_t *) (src + 0))->i) = v0;
        (((x264_union32_t *) (src + 4))->i) = v1;
        src += 32;
    }
}

void x264_predict_8x16c_p_c (pixel *src) {
    int H = 0;
    int V = 0;
    for (int i = 0; i < 4; i++)
        H += (i + 1) * (src[4 + i - 32] - src[2 - i - 32]);
    for (int i = 0; i < 8; i++)
        V += (i + 1) * (src[-1 + (i + 8) * 32] - src[-1 + (6 - i) * 32]);
    int a = 16 * (src[-1 + 15 * 32] + src[7 - 32]);
    int b = (17 * H + 16) >> 5;
    int c = (5 * V + 32) >> 6;
    int i00 = a - 3 * b - 7 * c + 16;
    for (int y = 0; y < 16; y++) {
        int pix = i00;
        for (int x = 0; x < 8; x++) {
            src[x] = x264_clip_pixel (pix >> 5);
            pix += b;
        }
        src += 32;
        i00 += c;
    }
}

void x264_predict_4x4_dc_128_c (pixel *src) {
    (((x264_union32_t *) (&src[(0) + (0) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (1) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (2) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (3) * 32]))->i) = ((1 << (8 - 1)) * 0x01010101U);
    ;
}

void x264_predict_4x4_dc_left_c (pixel *src) {
    pixel4 dc = (((src[(-1) + (0) * 32] + src[(-1) + (1) * 32] + src[(-1) + (2) * 32] + src[(-1) + (3) * 32] + 2) >> 2) * 0x01010101U);
    (((x264_union32_t *) (&src[(0) + (0) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (1) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (2) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (3) * 32]))->i) = dc;
    ;
}

void x264_predict_4x4_dc_top_c (pixel *src) {
    pixel4 dc = (((src[(0) + (-1) * 32] + src[(1) + (-1) * 32] + src[(2) + (-1) * 32] + src[(3) + (-1) * 32] + 2) >> 2) * 0x01010101U);
    (((x264_union32_t *) (&src[(0) + (0) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (1) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (2) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (3) * 32]))->i) = dc;
    ;
}

void x264_predict_4x4_dc_c (pixel *src) {
    pixel4 dc = (((src[(-1) + (0) * 32] + src[(-1) + (1) * 32] + src[(-1) + (2) * 32] + src[(-1) + (3) * 32] + src[(0) + (-1) * 32] + src[(1) + (-1) * 32] + src[(2) + (-1) * 32] + src[(3) + (-1) * 32] + 4) >> 3) * 0x01010101U);
    (((x264_union32_t *) (&src[(0) + (0) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (1) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (2) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (3) * 32]))->i) = dc;
    ;
}

void x264_predict_4x4_h_c (pixel *src) {
    (((x264_union32_t *) (&src[(0) + (0) * 32]))->i) = ((src[(-1) + (0) * 32]) * 0x01010101U);
    (((x264_union32_t *) (&src[(0) + (1) * 32]))->i) = ((src[(-1) + (1) * 32]) * 0x01010101U);
    (((x264_union32_t *) (&src[(0) + (2) * 32]))->i) = ((src[(-1) + (2) * 32]) * 0x01010101U);
    (((x264_union32_t *) (&src[(0) + (3) * 32]))->i) = ((src[(-1) + (3) * 32]) * 0x01010101U);
}

void x264_predict_4x4_v_c (pixel *src) {
    (((x264_union32_t *) (&src[(0) + (0) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (1) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (2) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (3) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (-1) * 32]))->i);
    ;
}

void x264_predict_4x4_ddl_c (pixel *src) {
    int t0 = src[(0) + (-1) * 32];
    int t1 = src[(1) + (-1) * 32];
    int t2 = src[(2) + (-1) * 32];
    int t3 = src[(3) + (-1) * 32];
    int t4 = src[(4) + (-1) * 32];
    int t5 = src[(5) + (-1) * 32];
    int t6 = src[(6) + (-1) * 32];
    int t7 = src[(7) + (-1) * 32];
    src[(0) + (0) * 32] = (((t0) +2 * (t1) +(t2)+2) >> 2);
    src[(1) + (0) * 32] = src[(0) + (1) * 32] = (((t1) +2 * (t2) +(t3)+2) >> 2);
    src[(2) + (0) * 32] = src[(1) + (1) * 32] = src[(0) + (2) * 32] = (((t2) +2 * (t3) +(t4)+2) >> 2);
    src[(3) + (0) * 32] = src[(2) + (1) * 32] = src[(1) + (2) * 32] = src[(0) + (3) * 32] = (((t3) +2 * (t4) +(t5)+2) >> 2);
    src[(3) + (1) * 32] = src[(2) + (2) * 32] = src[(1) + (3) * 32] = (((t4) +2 * (t5) +(t6)+2) >> 2);
    src[(3) + (2) * 32] = src[(2) + (3) * 32] = (((t5) +2 * (t6) +(t7)+2) >> 2);
    src[(3) + (3) * 32] = (((t6) +2 * (t7) +(t7)+2) >> 2);
}

void x264_predict_4x4_ddr_c (pixel *src) {
    int lt = src[(-1) + (-1) * 32];
    int l0 = src[(-1) + (0) * 32];
    int l1 = src[(-1) + (1) * 32];
    int l2 = src[(-1) + (2) * 32];
    int l3 = src[(-1) + (3) * 32];
    int t0 = src[(0) + (-1) * 32];
    int t1 = src[(1) + (-1) * 32];
    int t2 = src[(2) + (-1) * 32];
    int t3 = src[(3) + (-1) * 32];
    src[(3) + (0) * 32] = (((t3) +2 * (t2) +(t1)+2) >> 2);
    src[(2) + (0) * 32] = src[(3) + (1) * 32] = (((t2) +2 * (t1) +(t0)+2) >> 2);
    src[(1) + (0) * 32] = src[(2) + (1) * 32] = src[(3) + (2) * 32] = (((t1) +2 * (t0) +(lt)+2) >> 2);
    src[(0) + (0) * 32] = src[(1) + (1) * 32] = src[(2) + (2) * 32] = src[(3) + (3) * 32] = (((t0) +2 * (lt) +(l0)+2) >> 2);
    src[(0) + (1) * 32] = src[(1) + (2) * 32] = src[(2) + (3) * 32] = (((lt) +2 * (l0) +(l1)+2) >> 2);
    src[(0) + (2) * 32] = src[(1) + (3) * 32] = (((l0) +2 * (l1) +(l2)+2) >> 2);
    src[(0) + (3) * 32] = (((l1) +2 * (l2) +(l3)+2) >> 2);
}

void x264_predict_4x4_vr_c (pixel *src) {
    int lt = src[(-1) + (-1) * 32];
    int l0 = src[(-1) + (0) * 32];
    int l1 = src[(-1) + (1) * 32];
    int l2 = src[(-1) + (2) * 32];
    int l3 = src[(-1) + (3) * 32];
    int t0 = src[(0) + (-1) * 32];
    int t1 = src[(1) + (-1) * 32];
    int t2 = src[(2) + (-1) * 32];
    int t3 = src[(3) + (-1) * 32];
    src[(0) + (3) * 32] = (((l2) +2 * (l1) +(l0)+2) >> 2);
    src[(0) + (2) * 32] = (((l1) +2 * (l0) +(lt)+2) >> 2);
    src[(0) + (1) * 32] = src[(1) + (3) * 32] = (((l0) +2 * (lt) +(t0)+2) >> 2);
    src[(0) + (0) * 32] = src[(1) + (2) * 32] = (((lt) +(t0)+1) >> 1);
    src[(1) + (1) * 32] = src[(2) + (3) * 32] = (((lt) +2 * (t0) +(t1)+2) >> 2);
    src[(1) + (0) * 32] = src[(2) + (2) * 32] = (((t0) +(t1)+1) >> 1);
    src[(2) + (1) * 32] = src[(3) + (3) * 32] = (((t0) +2 * (t1) +(t2)+2) >> 2);
    src[(2) + (0) * 32] = src[(3) + (2) * 32] = (((t1) +(t2)+1) >> 1);
    src[(3) + (1) * 32] = (((t1) +2 * (t2) +(t3)+2) >> 2);
    src[(3) + (0) * 32] = (((t2) +(t3)+1) >> 1);
}

void x264_predict_4x4_hd_c (pixel *src) {
    int lt = src[(-1) + (-1) * 32];
    int l0 = src[(-1) + (0) * 32];
    int l1 = src[(-1) + (1) * 32];
    int l2 = src[(-1) + (2) * 32];
    int l3 = src[(-1) + (3) * 32];
    int t0 = src[(0) + (-1) * 32];
    int t1 = src[(1) + (-1) * 32];
    int t2 = src[(2) + (-1) * 32];
    int t3 = src[(3) + (-1) * 32];
    src[(0) + (3) * 32] = (((l2) +(l3)+1) >> 1);
    src[(1) + (3) * 32] = (((l1) +2 * (l2) +(l3)+2) >> 2);
    src[(0) + (2) * 32] = src[(2) + (3) * 32] = (((l1) +(l2)+1) >> 1);
    src[(1) + (2) * 32] = src[(3) + (3) * 32] = (((l0) +2 * (l1) +(l2)+2) >> 2);
    src[(0) + (1) * 32] = src[(2) + (2) * 32] = (((l0) +(l1)+1) >> 1);
    src[(1) + (1) * 32] = src[(3) + (2) * 32] = (((lt) +2 * (l0) +(l1)+2) >> 2);
    src[(0) + (0) * 32] = src[(2) + (1) * 32] = (((lt) +(l0)+1) >> 1);
    src[(1) + (0) * 32] = src[(3) + (1) * 32] = (((t0) +2 * (lt) +(l0)+2) >> 2);
    src[(2) + (0) * 32] = (((t1) +2 * (t0) +(lt)+2) >> 2);
    src[(3) + (0) * 32] = (((t2) +2 * (t1) +(t0)+2) >> 2);
}

void x264_predict_4x4_vl_c (pixel *src) {
    int t0 = src[(0) + (-1) * 32];
    int t1 = src[(1) + (-1) * 32];
    int t2 = src[(2) + (-1) * 32];
    int t3 = src[(3) + (-1) * 32];
    int t4 = src[(4) + (-1) * 32];
    int t5 = src[(5) + (-1) * 32];
    int t6 = src[(6) + (-1) * 32];
    int t7 = src[(7) + (-1) * 32];
    src[(0) + (0) * 32] = (((t0) +(t1)+1) >> 1);
    src[(0) + (1) * 32] = (((t0) +2 * (t1) +(t2)+2) >> 2);
    src[(1) + (0) * 32] = src[(0) + (2) * 32] = (((t1) +(t2)+1) >> 1);
    src[(1) + (1) * 32] = src[(0) + (3) * 32] = (((t1) +2 * (t2) +(t3)+2) >> 2);
    src[(2) + (0) * 32] = src[(1) + (2) * 32] = (((t2) +(t3)+1) >> 1);
    src[(2) + (1) * 32] = src[(1) + (3) * 32] = (((t2) +2 * (t3) +(t4)+2) >> 2);
    src[(3) + (0) * 32] = src[(2) + (2) * 32] = (((t3) +(t4)+1) >> 1);
    src[(3) + (1) * 32] = src[(2) + (3) * 32] = (((t3) +2 * (t4) +(t5)+2) >> 2);
    src[(3) + (2) * 32] = (((t4) +(t5)+1) >> 1);
    src[(3) + (3) * 32] = (((t4) +2 * (t5) +(t6)+2) >> 2);
}

void x264_predict_4x4_hu_c (pixel *src) {
    int l0 = src[(-1) + (0) * 32];
    int l1 = src[(-1) + (1) * 32];
    int l2 = src[(-1) + (2) * 32];
    int l3 = src[(-1) + (3) * 32];
    src[(0) + (0) * 32] = (((l0) +(l1)+1) >> 1);
    src[(1) + (0) * 32] = (((l0) +2 * (l1) +(l2)+2) >> 2);
    src[(2) + (0) * 32] = src[(0) + (1) * 32] = (((l1) +(l2)+1) >> 1);
    src[(3) + (0) * 32] = src[(1) + (1) * 32] = (((l1) +2 * (l2) +(l3)+2) >> 2);
    src[(2) + (1) * 32] = src[(0) + (2) * 32] = (((l2) +(l3)+1) >> 1);
    src[(3) + (1) * 32] = src[(1) + (2) * 32] = (((l2) +2 * (l3) +(l3)+2) >> 2);
    src[(3) + (2) * 32] = src[(1) + (3) * 32] = src[(0) + (3) * 32] = src[(2) + (2) * 32] = src[(2) + (3) * 32] = src[(3) + (3) * 32] = l3;
}

void x264_predict_8x8_filter_c (pixel *src, pixel edge [36], int i_neighbor, int i_filters) {
    int have_lt = i_neighbor & MB_TOPLEFT;
    if (i_filters & MB_LEFT) {
        edge[15] = (src[(0) + (-1) * 32] + 2 * src[(-1) + (-1) * 32] + src[(-1) + (0) * 32] + 2) >> 2;
        edge[14] = ((have_lt ? src[(-1) + (-1) * 32] : src[(-1) + (0) * 32]) + 2 * src[(-1) + (0) * 32] + src[(-1) + (1) * 32] + 2) >> 2;
        edge[14 - 1] = (((src[(-1) + (1 - 1) * 32]) + 2 * (src[(-1) + (1) * 32]) + (src[(-1) + (1 + 1) * 32]) + 2) >> 2);
        edge[14 - 2] = (((src[(-1) + (2 - 1) * 32]) + 2 * (src[(-1) + (2) * 32]) + (src[(-1) + (2 + 1) * 32]) + 2) >> 2);
        edge[14 - 3] = (((src[(-1) + (3 - 1) * 32]) + 2 * (src[(-1) + (3) * 32]) + (src[(-1) + (3 + 1) * 32]) + 2) >> 2);
        edge[14 - 4] = (((src[(-1) + (4 - 1) * 32]) + 2 * (src[(-1) + (4) * 32]) + (src[(-1) + (4 + 1) * 32]) + 2) >> 2);
        edge[14 - 5] = (((src[(-1) + (5 - 1) * 32]) + 2 * (src[(-1) + (5) * 32]) + (src[(-1) + (5 + 1) * 32]) + 2) >> 2);
        edge[14 - 6] = (((src[(-1) + (6 - 1) * 32]) + 2 * (src[(-1) + (6) * 32]) + (src[(-1) + (6 + 1) * 32]) + 2) >> 2);
        edge[6] = edge[7] = (src[(-1) + (6) * 32] + 3 * src[(-1) + (7) * 32] + 2) >> 2;
    }
    if (i_filters & MB_TOP) {
        int have_tr = i_neighbor & MB_TOPRIGHT;
        edge[16] = ((have_lt ? src[(-1) + (-1) * 32] : src[(0) + (-1) * 32]) + 2 * src[(0) + (-1) * 32] + src[(1) + (-1) * 32] + 2) >> 2;
        edge[16 + 1] = (((src[(1 - 1) + (-1) * 32]) + 2 * (src[(1) + (-1) * 32]) + (src[(1 + 1) + (-1) * 32]) + 2) >> 2);
        edge[16 + 2] = (((src[(2 - 1) + (-1) * 32]) + 2 * (src[(2) + (-1) * 32]) + (src[(2 + 1) + (-1) * 32]) + 2) >> 2);
        edge[16 + 3] = (((src[(3 - 1) + (-1) * 32]) + 2 * (src[(3) + (-1) * 32]) + (src[(3 + 1) + (-1) * 32]) + 2) >> 2);
        edge[16 + 4] = (((src[(4 - 1) + (-1) * 32]) + 2 * (src[(4) + (-1) * 32]) + (src[(4 + 1) + (-1) * 32]) + 2) >> 2);
        edge[16 + 5] = (((src[(5 - 1) + (-1) * 32]) + 2 * (src[(5) + (-1) * 32]) + (src[(5 + 1) + (-1) * 32]) + 2) >> 2);
        edge[16 + 6] = (((src[(6 - 1) + (-1) * 32]) + 2 * (src[(6) + (-1) * 32]) + (src[(6 + 1) + (-1) * 32]) + 2) >> 2);
        edge[23] = (src[(6) + (-1) * 32] + 2 * src[(7) + (-1) * 32] + (have_tr ? src[(8) + (-1) * 32] : src[(7) + (-1) * 32]) + 2) >> 2;
        if (i_filters & MB_TOPRIGHT) {
            if (have_tr) {
                edge[16 + 8] = (((src[(8 - 1) + (-1) * 32]) + 2 * (src[(8) + (-1) * 32]) + (src[(8 + 1) + (-1) * 32]) + 2) >> 2);
                edge[16 + 9] = (((src[(9 - 1) + (-1) * 32]) + 2 * (src[(9) + (-1) * 32]) + (src[(9 + 1) + (-1) * 32]) + 2) >> 2);
                edge[16 + 10] = (((src[(10 - 1) + (-1) * 32]) + 2 * (src[(10) + (-1) * 32]) + (src[(10 + 1) + (-1) * 32]) + 2) >> 2);
                edge[16 + 11] = (((src[(11 - 1) + (-1) * 32]) + 2 * (src[(11) + (-1) * 32]) + (src[(11 + 1) + (-1) * 32]) + 2) >> 2);
                edge[16 + 12] = (((src[(12 - 1) + (-1) * 32]) + 2 * (src[(12) + (-1) * 32]) + (src[(12 + 1) + (-1) * 32]) + 2) >> 2);
                edge[16 + 13] = (((src[(13 - 1) + (-1) * 32]) + 2 * (src[(13) + (-1) * 32]) + (src[(13 + 1) + (-1) * 32]) + 2) >> 2);
                edge[16 + 14] = (((src[(14 - 1) + (-1) * 32]) + 2 * (src[(14) + (-1) * 32]) + (src[(14 + 1) + (-1) * 32]) + 2) >> 2);
                edge[31] = edge[32] = (src[(14) + (-1) * 32] + 3 * src[(15) + (-1) * 32] + 2) >> 2;
            }
            else {
                (((x264_union32_t *) (edge + 24))->i) = ((src[(7) + (-1) * 32]) * 0x01010101U);
                (((x264_union32_t *) (edge + 28))->i) = ((src[(7) + (-1) * 32]) * 0x01010101U);
                edge[32] = src[(7) + (-1) * 32];
            }
        }
    }
}

void x264_predict_8x8_dc_128_c (pixel *src, pixel edge [36]) {
    for (int y = 0; y < 8; y++) {
        (((x264_union32_t *) (src + 0))->i) = ((1 << (8 - 1)) * 0x01010101U);
        (((x264_union32_t *) (src + 4))->i) = ((1 << (8 - 1)) * 0x01010101U);
        src += 32;
    };
}

void x264_predict_8x8_dc_left_c (pixel *src, pixel edge [36]) {
    int l0 = edge[14 - 0];
    int l1 = edge[14 - 1];
    int l2 = edge[14 - 2];
    int l3 = edge[14 - 3];
    int l4 = edge[14 - 4];
    int l5 = edge[14 - 5];
    int l6 = edge[14 - 6];
    int l7 = edge[14 - 7];
    pixel4 dc = (((l0 + l1 + l2 + l3 + l4 + l5 + l6 + l7 + 4) >> 3) * 0x01010101U);
    for (int y = 0; y < 8; y++) {
        (((x264_union32_t *) (src + 0))->i) = dc;
        (((x264_union32_t *) (src + 4))->i) = dc;
        src += 32;
    };
}

void x264_predict_8x8_dc_top_c (pixel *src, pixel edge [36]) {
    int t0 = edge[16 + 0];
    int t1 = edge[16 + 1];
    int t2 = edge[16 + 2];
    int t3 = edge[16 + 3];
    int t4 = edge[16 + 4];
    int t5 = edge[16 + 5];
    int t6 = edge[16 + 6];
    int t7 = edge[16 + 7];
    pixel4 dc = (((t0 + t1 + t2 + t3 + t4 + t5 + t6 + t7 + 4) >> 3) * 0x01010101U);
    for (int y = 0; y < 8; y++) {
        (((x264_union32_t *) (src + 0))->i) = dc;
        (((x264_union32_t *) (src + 4))->i) = dc;
        src += 32;
    };
}

void x264_predict_8x8_dc_c (pixel *src, pixel edge [36]) {
    int l0 = edge[14 - 0];
    int l1 = edge[14 - 1];
    int l2 = edge[14 - 2];
    int l3 = edge[14 - 3];
    int l4 = edge[14 - 4];
    int l5 = edge[14 - 5];
    int l6 = edge[14 - 6];
    int l7 = edge[14 - 7];
    int t0 = edge[16 + 0];
    int t1 = edge[16 + 1];
    int t2 = edge[16 + 2];
    int t3 = edge[16 + 3];
    int t4 = edge[16 + 4];
    int t5 = edge[16 + 5];
    int t6 = edge[16 + 6];
    int t7 = edge[16 + 7];
    pixel4 dc = (((l0 + l1 + l2 + l3 + l4 + l5 + l6 + l7 + t0 + t1 + t2 + t3 + t4 + t5 + t6 + t7 + 8) >> 4) * 0x01010101U);
    for (int y = 0; y < 8; y++) {
        (((x264_union32_t *) (src + 0))->i) = dc;
        (((x264_union32_t *) (src + 4))->i) = dc;
        src += 32;
    };
}

void x264_predict_8x8_h_c (pixel *src, pixel edge [36]) {
    int l0 = edge[14 - 0];
    int l1 = edge[14 - 1];
    int l2 = edge[14 - 2];
    int l3 = edge[14 - 3];
    int l4 = edge[14 - 4];
    int l5 = edge[14 - 5];
    int l6 = edge[14 - 6];
    int l7 = edge[14 - 7];
    (((x264_union32_t *) (src + 0 * 32 + 0))->i) = (((x264_union32_t *) (src + 0 * 32 + 4))->i) = ((l0) *0x01010101U);
    ;
    (((x264_union32_t *) (src + 1 * 32 + 0))->i) = (((x264_union32_t *) (src + 1 * 32 + 4))->i) = ((l1) *0x01010101U);
    ;
    (((x264_union32_t *) (src + 2 * 32 + 0))->i) = (((x264_union32_t *) (src + 2 * 32 + 4))->i) = ((l2) *0x01010101U);
    ;
    (((x264_union32_t *) (src + 3 * 32 + 0))->i) = (((x264_union32_t *) (src + 3 * 32 + 4))->i) = ((l3) *0x01010101U);
    ;
    (((x264_union32_t *) (src + 4 * 32 + 0))->i) = (((x264_union32_t *) (src + 4 * 32 + 4))->i) = ((l4) *0x01010101U);
    ;
    (((x264_union32_t *) (src + 5 * 32 + 0))->i) = (((x264_union32_t *) (src + 5 * 32 + 4))->i) = ((l5) *0x01010101U);
    ;
    (((x264_union32_t *) (src + 6 * 32 + 0))->i) = (((x264_union32_t *) (src + 6 * 32 + 4))->i) = ((l6) *0x01010101U);
    ;
    (((x264_union32_t *) (src + 7 * 32 + 0))->i) = (((x264_union32_t *) (src + 7 * 32 + 4))->i) = ((l7) *0x01010101U);
    ;
}

void x264_predict_8x8_v_c (pixel *src, pixel edge [36]) {
    pixel4 top [2] = {(((x264_union32_t *) (edge + 16))->i), (((x264_union32_t *) (edge + 20))->i)};
    for (int y = 0; y < 8; y++) {
        (((x264_union32_t *) (src + y * 32 + 0))->i) = top[0];
        (((x264_union32_t *) (src + y * 32 + 4))->i) = top[1];
    }
}

void x264_predict_8x8_ddl_c (pixel *src, pixel edge [36]) {
    int t0 = edge[16 + 0];
    int t1 = edge[16 + 1];
    int t2 = edge[16 + 2];
    int t3 = edge[16 + 3];
    int t4 = edge[16 + 4];
    int t5 = edge[16 + 5];
    int t6 = edge[16 + 6];
    int t7 = edge[16 + 7];
    int t8 = edge[16 + 8];
    int t9 = edge[16 + 9];
    int t10 = edge[16 + 10];
    int t11 = edge[16 + 11];
    int t12 = edge[16 + 12];
    int t13 = edge[16 + 13];
    int t14 = edge[16 + 14];
    int t15 = edge[16 + 15];
    src[(0) + (0) * 32] = (((t0) +2 * (t1) +(t2)+2) >> 2);
    src[(0) + (1) * 32] = src[(1) + (0) * 32] = (((t1) +2 * (t2) +(t3)+2) >> 2);
    src[(0) + (2) * 32] = src[(1) + (1) * 32] = src[(2) + (0) * 32] = (((t2) +2 * (t3) +(t4)+2) >> 2);
    src[(0) + (3) * 32] = src[(1) + (2) * 32] = src[(2) + (1) * 32] = src[(3) + (0) * 32] = (((t3) +2 * (t4) +(t5)+2) >> 2);
    src[(0) + (4) * 32] = src[(1) + (3) * 32] = src[(2) + (2) * 32] = src[(3) + (1) * 32] = src[(4) + (0) * 32] = (((t4) +2 * (t5) +(t6)+2) >> 2);
    src[(0) + (5) * 32] = src[(1) + (4) * 32] = src[(2) + (3) * 32] = src[(3) + (2) * 32] = src[(4) + (1) * 32] = src[(5) + (0) * 32] = (((t5) +2 * (t6) +(t7)+2) >> 2);
    src[(0) + (6) * 32] = src[(1) + (5) * 32] = src[(2) + (4) * 32] = src[(3) + (3) * 32] = src[(4) + (2) * 32] = src[(5) + (1) * 32] = src[(6) + (0) * 32] = (((t6) +2 * (t7) +(t8)+2) >> 2);
    src[(0) + (7) * 32] = src[(1) + (6) * 32] = src[(2) + (5) * 32] = src[(3) + (4) * 32] = src[(4) + (3) * 32] = src[(5) + (2) * 32] = src[(6) + (1) * 32] = src[(7) + (0) * 32] = (((t7) +2 * (t8) +(t9)+2) >> 2);
    src[(1) + (7) * 32] = src[(2) + (6) * 32] = src[(3) + (5) * 32] = src[(4) + (4) * 32] = src[(5) + (3) * 32] = src[(6) + (2) * 32] = src[(7) + (1) * 32] = (((t8) +2 * (t9) +(t10)+2) >> 2);
    src[(2) + (7) * 32] = src[(3) + (6) * 32] = src[(4) + (5) * 32] = src[(5) + (4) * 32] = src[(6) + (3) * 32] = src[(7) + (2) * 32] = (((t9) +2 * (t10) +(t11)+2) >> 2);
    src[(3) + (7) * 32] = src[(4) + (6) * 32] = src[(5) + (5) * 32] = src[(6) + (4) * 32] = src[(7) + (3) * 32] = (((t10) +2 * (t11) +(t12)+2) >> 2);
    src[(4) + (7) * 32] = src[(5) + (6) * 32] = src[(6) + (5) * 32] = src[(7) + (4) * 32] = (((t11) +2 * (t12) +(t13)+2) >> 2);
    src[(5) + (7) * 32] = src[(6) + (6) * 32] = src[(7) + (5) * 32] = (((t12) +2 * (t13) +(t14)+2) >> 2);
    src[(6) + (7) * 32] = src[(7) + (6) * 32] = (((t13) +2 * (t14) +(t15)+2) >> 2);
    src[(7) + (7) * 32] = (((t14) +2 * (t15) +(t15)+2) >> 2);
}

void x264_predict_8x8_ddr_c (pixel *src, pixel edge [36]) {
    int t0 = edge[16 + 0];
    int t1 = edge[16 + 1];
    int t2 = edge[16 + 2];
    int t3 = edge[16 + 3];
    int t4 = edge[16 + 4];
    int t5 = edge[16 + 5];
    int t6 = edge[16 + 6];
    int t7 = edge[16 + 7];
    int l0 = edge[14 - 0];
    int l1 = edge[14 - 1];
    int l2 = edge[14 - 2];
    int l3 = edge[14 - 3];
    int l4 = edge[14 - 4];
    int l5 = edge[14 - 5];
    int l6 = edge[14 - 6];
    int l7 = edge[14 - 7];
    int lt = edge[15];
    src[(0) + (7) * 32] = (((l7) +2 * (l6) +(l5)+2) >> 2);
    src[(0) + (6) * 32] = src[(1) + (7) * 32] = (((l6) +2 * (l5) +(l4)+2) >> 2);
    src[(0) + (5) * 32] = src[(1) + (6) * 32] = src[(2) + (7) * 32] = (((l5) +2 * (l4) +(l3)+2) >> 2);
    src[(0) + (4) * 32] = src[(1) + (5) * 32] = src[(2) + (6) * 32] = src[(3) + (7) * 32] = (((l4) +2 * (l3) +(l2)+2) >> 2);
    src[(0) + (3) * 32] = src[(1) + (4) * 32] = src[(2) + (5) * 32] = src[(3) + (6) * 32] = src[(4) + (7) * 32] = (((l3) +2 * (l2) +(l1)+2) >> 2);
    src[(0) + (2) * 32] = src[(1) + (3) * 32] = src[(2) + (4) * 32] = src[(3) + (5) * 32] = src[(4) + (6) * 32] = src[(5) + (7) * 32] = (((l2) +2 * (l1) +(l0)+2) >> 2);
    src[(0) + (1) * 32] = src[(1) + (2) * 32] = src[(2) + (3) * 32] = src[(3) + (4) * 32] = src[(4) + (5) * 32] = src[(5) + (6) * 32] = src[(6) + (7) * 32] = (((l1) +2 * (l0) +(lt)+2) >> 2);
    src[(0) + (0) * 32] = src[(1) + (1) * 32] = src[(2) + (2) * 32] = src[(3) + (3) * 32] = src[(4) + (4) * 32] = src[(5) + (5) * 32] = src[(6) + (6) * 32] = src[(7) + (7) * 32] = (((l0) +2 * (lt) +(t0)+2) >> 2);
    src[(1) + (0) * 32] = src[(2) + (1) * 32] = src[(3) + (2) * 32] = src[(4) + (3) * 32] = src[(5) + (4) * 32] = src[(6) + (5) * 32] = src[(7) + (6) * 32] = (((lt) +2 * (t0) +(t1)+2) >> 2);
    src[(2) + (0) * 32] = src[(3) + (1) * 32] = src[(4) + (2) * 32] = src[(5) + (3) * 32] = src[(6) + (4) * 32] = src[(7) + (5) * 32] = (((t0) +2 * (t1) +(t2)+2) >> 2);
    src[(3) + (0) * 32] = src[(4) + (1) * 32] = src[(5) + (2) * 32] = src[(6) + (3) * 32] = src[(7) + (4) * 32] = (((t1) +2 * (t2) +(t3)+2) >> 2);
    src[(4) + (0) * 32] = src[(5) + (1) * 32] = src[(6) + (2) * 32] = src[(7) + (3) * 32] = (((t2) +2 * (t3) +(t4)+2) >> 2);
    src[(5) + (0) * 32] = src[(6) + (1) * 32] = src[(7) + (2) * 32] = (((t3) +2 * (t4) +(t5)+2) >> 2);
    src[(6) + (0) * 32] = src[(7) + (1) * 32] = (((t4) +2 * (t5) +(t6)+2) >> 2);
    src[(7) + (0) * 32] = (((t5) +2 * (t6) +(t7)+2) >> 2);
}

void x264_predict_8x8_vr_c (pixel *src, pixel edge [36]) {
    int t0 = edge[16 + 0];
    int t1 = edge[16 + 1];
    int t2 = edge[16 + 2];
    int t3 = edge[16 + 3];
    int t4 = edge[16 + 4];
    int t5 = edge[16 + 5];
    int t6 = edge[16 + 6];
    int t7 = edge[16 + 7];
    int l0 = edge[14 - 0];
    int l1 = edge[14 - 1];
    int l2 = edge[14 - 2];
    int l3 = edge[14 - 3];
    int l4 = edge[14 - 4];
    int l5 = edge[14 - 5];
    int l6 = edge[14 - 6];
    int l7 = edge[14 - 7];
    int lt = edge[15];
    src[(0) + (6) * 32] = (((l5) +2 * (l4) +(l3)+2) >> 2);
    src[(0) + (7) * 32] = (((l6) +2 * (l5) +(l4)+2) >> 2);
    src[(0) + (4) * 32] = src[(1) + (6) * 32] = (((l3) +2 * (l2) +(l1)+2) >> 2);
    src[(0) + (5) * 32] = src[(1) + (7) * 32] = (((l4) +2 * (l3) +(l2)+2) >> 2);
    src[(0) + (2) * 32] = src[(1) + (4) * 32] = src[(2) + (6) * 32] = (((l1) +2 * (l0) +(lt)+2) >> 2);
    src[(0) + (3) * 32] = src[(1) + (5) * 32] = src[(2) + (7) * 32] = (((l2) +2 * (l1) +(l0)+2) >> 2);
    src[(0) + (1) * 32] = src[(1) + (3) * 32] = src[(2) + (5) * 32] = src[(3) + (7) * 32] = (((l0) +2 * (lt) +(t0)+2) >> 2);
    src[(0) + (0) * 32] = src[(1) + (2) * 32] = src[(2) + (4) * 32] = src[(3) + (6) * 32] = (((lt) +(t0)+1) >> 1);
    src[(1) + (1) * 32] = src[(2) + (3) * 32] = src[(3) + (5) * 32] = src[(4) + (7) * 32] = (((lt) +2 * (t0) +(t1)+2) >> 2);
    src[(1) + (0) * 32] = src[(2) + (2) * 32] = src[(3) + (4) * 32] = src[(4) + (6) * 32] = (((t0) +(t1)+1) >> 1);
    src[(2) + (1) * 32] = src[(3) + (3) * 32] = src[(4) + (5) * 32] = src[(5) + (7) * 32] = (((t0) +2 * (t1) +(t2)+2) >> 2);
    src[(2) + (0) * 32] = src[(3) + (2) * 32] = src[(4) + (4) * 32] = src[(5) + (6) * 32] = (((t1) +(t2)+1) >> 1);
    src[(3) + (1) * 32] = src[(4) + (3) * 32] = src[(5) + (5) * 32] = src[(6) + (7) * 32] = (((t1) +2 * (t2) +(t3)+2) >> 2);
    src[(3) + (0) * 32] = src[(4) + (2) * 32] = src[(5) + (4) * 32] = src[(6) + (6) * 32] = (((t2) +(t3)+1) >> 1);
    src[(4) + (1) * 32] = src[(5) + (3) * 32] = src[(6) + (5) * 32] = src[(7) + (7) * 32] = (((t2) +2 * (t3) +(t4)+2) >> 2);
    src[(4) + (0) * 32] = src[(5) + (2) * 32] = src[(6) + (4) * 32] = src[(7) + (6) * 32] = (((t3) +(t4)+1) >> 1);
    src[(5) + (1) * 32] = src[(6) + (3) * 32] = src[(7) + (5) * 32] = (((t3) +2 * (t4) +(t5)+2) >> 2);
    src[(5) + (0) * 32] = src[(6) + (2) * 32] = src[(7) + (4) * 32] = (((t4) +(t5)+1) >> 1);
    src[(6) + (1) * 32] = src[(7) + (3) * 32] = (((t4) +2 * (t5) +(t6)+2) >> 2);
    src[(6) + (0) * 32] = src[(7) + (2) * 32] = (((t5) +(t6)+1) >> 1);
    src[(7) + (1) * 32] = (((t5) +2 * (t6) +(t7)+2) >> 2);
    src[(7) + (0) * 32] = (((t6) +(t7)+1) >> 1);
}

void x264_predict_8x8_hd_c (pixel *src, pixel edge [36]) {
    int t0 = edge[16 + 0];
    int t1 = edge[16 + 1];
    int t2 = edge[16 + 2];
    int t3 = edge[16 + 3];
    int t4 = edge[16 + 4];
    int t5 = edge[16 + 5];
    int t6 = edge[16 + 6];
    int t7 = edge[16 + 7];
    int l0 = edge[14 - 0];
    int l1 = edge[14 - 1];
    int l2 = edge[14 - 2];
    int l3 = edge[14 - 3];
    int l4 = edge[14 - 4];
    int l5 = edge[14 - 5];
    int l6 = edge[14 - 6];
    int l7 = edge[14 - 7];
    int lt = edge[15];
    int p1 = pack8to16 ((((l6) +(l7)+1) >> 1), (((l5) +2 * (l6) +(l7)+2) >> 2));
    int p2 = pack8to16 ((((l5) +(l6)+1) >> 1), (((l4) +2 * (l5) +(l6)+2) >> 2));
    int p3 = pack8to16 ((((l4) +(l5)+1) >> 1), (((l3) +2 * (l4) +(l5)+2) >> 2));
    int p4 = pack8to16 ((((l3) +(l4)+1) >> 1), (((l2) +2 * (l3) +(l4)+2) >> 2));
    int p5 = pack8to16 ((((l2) +(l3)+1) >> 1), (((l1) +2 * (l2) +(l3)+2) >> 2));
    int p6 = pack8to16 ((((l1) +(l2)+1) >> 1), (((l0) +2 * (l1) +(l2)+2) >> 2));
    int p7 = pack8to16 ((((l0) +(l1)+1) >> 1), (((lt) +2 * (l0) +(l1)+2) >> 2));
    int p8 = pack8to16 ((((lt) +(l0)+1) >> 1), (((l0) +2 * (lt) +(t0)+2) >> 2));
    int p9 = pack8to16 ((((t1) +2 * (t0) +(lt)+2) >> 2), (((t2) +2 * (t1) +(t0)+2) >> 2));
    int p10 = pack8to16 ((((t3) +2 * (t2) +(t1)+2) >> 2), (((t4) +2 * (t3) +(t2)+2) >> 2));
    int p11 = pack8to16 ((((t5) +2 * (t4) +(t3)+2) >> 2), (((t6) +2 * (t5) +(t4)+2) >> 2));
    (((x264_union32_t *) (&src[(0) + (7) * 32]))->i) = pack16to32 (p1, p2);
    (((x264_union32_t *) (&src[(0) + (6) * 32]))->i) = pack16to32 (p2, p3);
    (((x264_union32_t *) (&src[(4) + (7) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (5) * 32]))->i) = pack16to32 (p3, p4);
    (((x264_union32_t *) (&src[(4) + (6) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (4) * 32]))->i) = pack16to32 (p4, p5);
    (((x264_union32_t *) (&src[(4) + (5) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (3) * 32]))->i) = pack16to32 (p5, p6);
    (((x264_union32_t *) (&src[(4) + (4) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (2) * 32]))->i) = pack16to32 (p6, p7);
    (((x264_union32_t *) (&src[(4) + (3) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (1) * 32]))->i) = pack16to32 (p7, p8);
    (((x264_union32_t *) (&src[(4) + (2) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (0) * 32]))->i) = pack16to32 (p8, p9);
    (((x264_union32_t *) (&src[(4) + (1) * 32]))->i) = pack16to32 (p9, p10);
    (((x264_union32_t *) (&src[(4) + (0) * 32]))->i) = pack16to32 (p10, p11);
}

void x264_predict_8x8_vl_c (pixel *src, pixel edge [36]) {
    int t0 = edge[16 + 0];
    int t1 = edge[16 + 1];
    int t2 = edge[16 + 2];
    int t3 = edge[16 + 3];
    int t4 = edge[16 + 4];
    int t5 = edge[16 + 5];
    int t6 = edge[16 + 6];
    int t7 = edge[16 + 7];
    int t8 = edge[16 + 8];
    int t9 = edge[16 + 9];
    int t10 = edge[16 + 10];
    int t11 = edge[16 + 11];
    int t12 = edge[16 + 12];
    int t13 = edge[16 + 13];
    int t14 = edge[16 + 14];
    int t15 = edge[16 + 15];
    src[(0) + (0) * 32] = (((t0) +(t1)+1) >> 1);
    src[(0) + (1) * 32] = (((t0) +2 * (t1) +(t2)+2) >> 2);
    src[(0) + (2) * 32] = src[(1) + (0) * 32] = (((t1) +(t2)+1) >> 1);
    src[(0) + (3) * 32] = src[(1) + (1) * 32] = (((t1) +2 * (t2) +(t3)+2) >> 2);
    src[(0) + (4) * 32] = src[(1) + (2) * 32] = src[(2) + (0) * 32] = (((t2) +(t3)+1) >> 1);
    src[(0) + (5) * 32] = src[(1) + (3) * 32] = src[(2) + (1) * 32] = (((t2) +2 * (t3) +(t4)+2) >> 2);
    src[(0) + (6) * 32] = src[(1) + (4) * 32] = src[(2) + (2) * 32] = src[(3) + (0) * 32] = (((t3) +(t4)+1) >> 1);
    src[(0) + (7) * 32] = src[(1) + (5) * 32] = src[(2) + (3) * 32] = src[(3) + (1) * 32] = (((t3) +2 * (t4) +(t5)+2) >> 2);
    src[(1) + (6) * 32] = src[(2) + (4) * 32] = src[(3) + (2) * 32] = src[(4) + (0) * 32] = (((t4) +(t5)+1) >> 1);
    src[(1) + (7) * 32] = src[(2) + (5) * 32] = src[(3) + (3) * 32] = src[(4) + (1) * 32] = (((t4) +2 * (t5) +(t6)+2) >> 2);
    src[(2) + (6) * 32] = src[(3) + (4) * 32] = src[(4) + (2) * 32] = src[(5) + (0) * 32] = (((t5) +(t6)+1) >> 1);
    src[(2) + (7) * 32] = src[(3) + (5) * 32] = src[(4) + (3) * 32] = src[(5) + (1) * 32] = (((t5) +2 * (t6) +(t7)+2) >> 2);
    src[(3) + (6) * 32] = src[(4) + (4) * 32] = src[(5) + (2) * 32] = src[(6) + (0) * 32] = (((t6) +(t7)+1) >> 1);
    src[(3) + (7) * 32] = src[(4) + (5) * 32] = src[(5) + (3) * 32] = src[(6) + (1) * 32] = (((t6) +2 * (t7) +(t8)+2) >> 2);
    src[(4) + (6) * 32] = src[(5) + (4) * 32] = src[(6) + (2) * 32] = src[(7) + (0) * 32] = (((t7) +(t8)+1) >> 1);
    src[(4) + (7) * 32] = src[(5) + (5) * 32] = src[(6) + (3) * 32] = src[(7) + (1) * 32] = (((t7) +2 * (t8) +(t9)+2) >> 2);
    src[(5) + (6) * 32] = src[(6) + (4) * 32] = src[(7) + (2) * 32] = (((t8) +(t9)+1) >> 1);
    src[(5) + (7) * 32] = src[(6) + (5) * 32] = src[(7) + (3) * 32] = (((t8) +2 * (t9) +(t10)+2) >> 2);
    src[(6) + (6) * 32] = src[(7) + (4) * 32] = (((t9) +(t10)+1) >> 1);
    src[(6) + (7) * 32] = src[(7) + (5) * 32] = (((t9) +2 * (t10) +(t11)+2) >> 2);
    src[(7) + (6) * 32] = (((t10) +(t11)+1) >> 1);
    src[(7) + (7) * 32] = (((t10) +2 * (t11) +(t12)+2) >> 2);
}

void x264_predict_8x8_hu_c (pixel *src, pixel edge [36]) {
    int l0 = edge[14 - 0];
    int l1 = edge[14 - 1];
    int l2 = edge[14 - 2];
    int l3 = edge[14 - 3];
    int l4 = edge[14 - 4];
    int l5 = edge[14 - 5];
    int l6 = edge[14 - 6];
    int l7 = edge[14 - 7];
    int p1 = pack8to16 ((((l0) +(l1)+1) >> 1), (((l0) +2 * (l1) +(l2)+2) >> 2));
    int p2 = pack8to16 ((((l1) +(l2)+1) >> 1), (((l1) +2 * (l2) +(l3)+2) >> 2));
    int p3 = pack8to16 ((((l2) +(l3)+1) >> 1), (((l2) +2 * (l3) +(l4)+2) >> 2));
    int p4 = pack8to16 ((((l3) +(l4)+1) >> 1), (((l3) +2 * (l4) +(l5)+2) >> 2));
    int p5 = pack8to16 ((((l4) +(l5)+1) >> 1), (((l4) +2 * (l5) +(l6)+2) >> 2));
    int p6 = pack8to16 ((((l5) +(l6)+1) >> 1), (((l5) +2 * (l6) +(l7)+2) >> 2));
    int p7 = pack8to16 ((((l6) +(l7)+1) >> 1), (((l6) +2 * (l7) +(l7)+2) >> 2));
    int p8 = pack8to16 (l7, l7);
    (((x264_union32_t *) (&src[(0) + (0) * 32]))->i) = pack16to32 (p1, p2);
    (((x264_union32_t *) (&src[(0) + (1) * 32]))->i) = pack16to32 (p2, p3);
    (((x264_union32_t *) (&src[(4) + (0) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (2) * 32]))->i) = pack16to32 (p3, p4);
    (((x264_union32_t *) (&src[(4) + (1) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (3) * 32]))->i) = pack16to32 (p4, p5);
    (((x264_union32_t *) (&src[(4) + (2) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (4) * 32]))->i) = pack16to32 (p5, p6);
    (((x264_union32_t *) (&src[(4) + (3) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (5) * 32]))->i) = pack16to32 (p6, p7);
    (((x264_union32_t *) (&src[(4) + (4) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (6) * 32]))->i) = pack16to32 (p7, p8);
    (((x264_union32_t *) (&src[(4) + (5) * 32]))->i) = (((x264_union32_t *) (&src[(4) + (6) * 32]))->i) = (((x264_union32_t *) (&src[(0) + (7) * 32]))->i) = (((x264_union32_t *) (&src[(4) + (7) * 32]))->i) = pack16to32 (p8, p8);
}

void x264_predict_16x16_init (int cpu, x264_predict_t pf [7]) {
    pf[I_PRED_16x16_V] = x264_predict_16x16_v_c;
    pf[I_PRED_16x16_H] = x264_predict_16x16_h_c;
    pf[I_PRED_16x16_DC] = x264_predict_16x16_dc_c;
    pf[I_PRED_16x16_P] = x264_predict_16x16_p_c;
    pf[I_PRED_16x16_DC_LEFT] = x264_predict_16x16_dc_left_c;
    pf[I_PRED_16x16_DC_TOP] = x264_predict_16x16_dc_top_c;
    pf[I_PRED_16x16_DC_128] = x264_predict_16x16_dc_128_c;
}

void x264_predict_8x8c_init (int cpu, x264_predict_t pf [7]) {
    pf[I_PRED_CHROMA_V] = x264_predict_8x8c_v_c;
    pf[I_PRED_CHROMA_H] = x264_predict_8x8c_h_c;
    pf[I_PRED_CHROMA_DC] = x264_predict_8x8c_dc_c;
    pf[I_PRED_CHROMA_P] = x264_predict_8x8c_p_c;
    pf[I_PRED_CHROMA_DC_LEFT] = x264_predict_8x8c_dc_left_c;
    pf[I_PRED_CHROMA_DC_TOP] = x264_predict_8x8c_dc_top_c;
    pf[I_PRED_CHROMA_DC_128] = x264_predict_8x8c_dc_128_c;
}

void x264_predict_8x16c_init (int cpu, x264_predict_t pf [7]) {
    pf[I_PRED_CHROMA_V] = x264_predict_8x16c_v_c;
    pf[I_PRED_CHROMA_H] = x264_predict_8x16c_h_c;
    pf[I_PRED_CHROMA_DC] = x264_predict_8x16c_dc_c;
    pf[I_PRED_CHROMA_P] = x264_predict_8x16c_p_c;
    pf[I_PRED_CHROMA_DC_LEFT] = x264_predict_8x16c_dc_left_c;
    pf[I_PRED_CHROMA_DC_TOP] = x264_predict_8x16c_dc_top_c;
    pf[I_PRED_CHROMA_DC_128] = x264_predict_8x16c_dc_128_c;
}

void x264_predict_8x8_init (int cpu, x264_predict8x8_t pf [12], x264_predict_8x8_filter_t *predict_filter) {
    pf[I_PRED_8x8_V] = x264_predict_8x8_v_c;
    pf[I_PRED_8x8_H] = x264_predict_8x8_h_c;
    pf[I_PRED_8x8_DC] = x264_predict_8x8_dc_c;
    pf[I_PRED_8x8_DDL] = x264_predict_8x8_ddl_c;
    pf[I_PRED_8x8_DDR] = x264_predict_8x8_ddr_c;
    pf[I_PRED_8x8_VR] = x264_predict_8x8_vr_c;
    pf[I_PRED_8x8_HD] = x264_predict_8x8_hd_c;
    pf[I_PRED_8x8_VL] = x264_predict_8x8_vl_c;
    pf[I_PRED_8x8_HU] = x264_predict_8x8_hu_c;
    pf[I_PRED_8x8_DC_LEFT] = x264_predict_8x8_dc_left_c;
    pf[I_PRED_8x8_DC_TOP] = x264_predict_8x8_dc_top_c;
    pf[I_PRED_8x8_DC_128] = x264_predict_8x8_dc_128_c;
    *predict_filter = x264_predict_8x8_filter_c;
}

void x264_predict_4x4_init (int cpu, x264_predict_t pf [12]) {
    pf[I_PRED_4x4_V] = x264_predict_4x4_v_c;
    pf[I_PRED_4x4_H] = x264_predict_4x4_h_c;
    pf[I_PRED_4x4_DC] = x264_predict_4x4_dc_c;
    pf[I_PRED_4x4_DDL] = x264_predict_4x4_ddl_c;
    pf[I_PRED_4x4_DDR] = x264_predict_4x4_ddr_c;
    pf[I_PRED_4x4_VR] = x264_predict_4x4_vr_c;
    pf[I_PRED_4x4_HD] = x264_predict_4x4_hd_c;
    pf[I_PRED_4x4_VL] = x264_predict_4x4_vl_c;
    pf[I_PRED_4x4_HU] = x264_predict_4x4_hu_c;
    pf[I_PRED_4x4_DC_LEFT] = x264_predict_4x4_dc_left_c;
    pf[I_PRED_4x4_DC_TOP] = x264_predict_4x4_dc_top_c;
    pf[I_PRED_4x4_DC_128] = x264_predict_4x4_dc_128_c;
}

