
#include "common.h"

int quant_8x8 (dctcoef dct [64], udctcoef mf [64], udctcoef bias [64]) {
    int nz = 0;
    for (int i = 0; i < 64; i++) {
        if ((dct[i]) > 0)
            (dct[i]) = (bias[i] + (dct[i])) * (mf[i]) >> 16;
        else
            (dct[i]) = -((bias[i] - (dct[i])) * (mf[i]) >> 16);
        nz |= (dct[i]);
    };
    return !!nz;
}

int quant_4x4 (dctcoef dct [16], udctcoef mf [16], udctcoef bias [16]) {
    int nz = 0;
    for (int i = 0; i < 16; i++) {
        if ((dct[i]) > 0)
            (dct[i]) = (bias[i] + (dct[i])) * (mf[i]) >> 16;
        else
            (dct[i]) = -((bias[i] - (dct[i])) * (mf[i]) >> 16);
        nz |= (dct[i]);
    };
    return !!nz;
}

int quant_4x4x4 (dctcoef dct [4] [16], udctcoef mf [16], udctcoef bias [16]) {
    int nza = 0;
    for (int j = 0; j < 4; j++) {
        int nz = 0;
        for (int i = 0; i < 16; i++) {
            if ((dct[j][i]) > 0)
                (dct[j][i]) = (bias[i] + (dct[j][i])) * (mf[i]) >> 16;
            else
                (dct[j][i]) = -((bias[i] - (dct[j][i])) * (mf[i]) >> 16);
            nz |= (dct[j][i]);
        };
        nza |= (!!nz) << j;
    }
    return nza;
}

int quant_4x4_dc (dctcoef dct [16], int mf, int bias) {
    int nz = 0;
    for (int i = 0; i < 16; i++) {
        if ((dct[i]) > 0)
            (dct[i]) = (bias + (dct[i])) * (mf) >> 16;
        else
            (dct[i]) = -((bias - (dct[i])) * (mf) >> 16);
        nz |= (dct[i]);
    };
    return !!nz;
}

int quant_2x2_dc (dctcoef dct [4], int mf, int bias) {
    int nz = 0;
    {
        if ((dct[0]) > 0)
            (dct[0]) = (bias + (dct[0])) * (mf) >> 16;
        else
            (dct[0]) = -((bias - (dct[0])) * (mf) >> 16);
        nz |= (dct[0]);
    };
    {
        if ((dct[1]) > 0)
            (dct[1]) = (bias + (dct[1])) * (mf) >> 16;
        else
            (dct[1]) = -((bias - (dct[1])) * (mf) >> 16);
        nz |= (dct[1]);
    };
    {
        if ((dct[2]) > 0)
            (dct[2]) = (bias + (dct[2])) * (mf) >> 16;
        else
            (dct[2]) = -((bias - (dct[2])) * (mf) >> 16);
        nz |= (dct[2]);
    };
    {
        if ((dct[3]) > 0)
            (dct[3]) = (bias + (dct[3])) * (mf) >> 16;
        else
            (dct[3]) = -((bias - (dct[3])) * (mf) >> 16);
        nz |= (dct[3]);
    };
    return !!nz;
}

void dequant_4x4 (dctcoef dct [16], int dequant_mf [6] [16], int i_qp) {
    const int i_mf = i_qp % 6;
    const int i_qbits = i_qp / 6 - 4;
    if (i_qbits >= 0) {
        for (int i = 0; i < 16; i++)
            dct[i] = (dct[i] * dequant_mf[i_mf][i]) << i_qbits;
    }
    else {
        const int f = 1 << (-i_qbits - 1);
        for (int i = 0; i < 16; i++)
            dct[i] = (dct[i] * dequant_mf[i_mf][i] + f) >> (-i_qbits);
    }
}

void dequant_8x8 (dctcoef dct [64], int dequant_mf [6] [64], int i_qp) {
    const int i_mf = i_qp % 6;
    const int i_qbits = i_qp / 6 - 6;
    if (i_qbits >= 0) {
        for (int i = 0; i < 64; i++)
            dct[i] = (dct[i] * dequant_mf[i_mf][i]) << i_qbits;
    }
    else {
        const int f = 1 << (-i_qbits - 1);
        for (int i = 0; i < 64; i++)
            dct[i] = (dct[i] * dequant_mf[i_mf][i] + f) >> (-i_qbits);
    }
}

void dequant_4x4_dc (dctcoef dct [16], int dequant_mf [6] [16], int i_qp) {
    const int i_qbits = i_qp / 6 - 6;
    if (i_qbits >= 0) {
        const int i_dmf = dequant_mf[i_qp % 6][0] << i_qbits;
        for (int i = 0; i < 16; i++)
            dct[i] *= i_dmf;
    }
    else {
        const int i_dmf = dequant_mf[i_qp % 6][0];
        const int f = 1 << (-i_qbits - 1);
        for (int i = 0; i < 16; i++)
            dct[i] = (dct[i] * i_dmf + f) >> (-i_qbits);
    }
}

void idct_dequant_2x4_dc (dctcoef dct [8], dctcoef dct4x4 [8] [16], int dequant_mf [6] [16], int i_qp) {
    int a0 = dct[0] + dct[1];
    int a1 = dct[2] + dct[3];
    int a2 = dct[4] + dct[5];
    int a3 = dct[6] + dct[7];
    int a4 = dct[0] - dct[1];
    int a5 = dct[2] - dct[3];
    int a6 = dct[4] - dct[5];
    int a7 = dct[6] - dct[7];
    int b0 = a0 + a1;
    int b1 = a2 + a3;
    int b2 = a4 + a5;
    int b3 = a6 + a7;
    int b4 = a0 - a1;
    int b5 = a2 - a3;
    int b6 = a4 - a5;
    int b7 = a6 - a7;
    int dmf = dequant_mf[i_qp % 6][0] << i_qp / 6;
    dct4x4[0][0] = ((b0 + b1) * dmf + 32) >> 6;
    dct4x4[1][0] = ((b2 + b3) * dmf + 32) >> 6;
    dct4x4[2][0] = ((b0 - b1) * dmf + 32) >> 6;
    dct4x4[3][0] = ((b2 - b3) * dmf + 32) >> 6;
    dct4x4[4][0] = ((b4 - b5) * dmf + 32) >> 6;
    dct4x4[5][0] = ((b6 - b7) * dmf + 32) >> 6;
    dct4x4[6][0] = ((b4 + b5) * dmf + 32) >> 6;
    dct4x4[7][0] = ((b6 + b7) * dmf + 32) >> 6;
}

void idct_dequant_2x4_dconly (dctcoef dct [8], int dequant_mf [6] [16], int i_qp) {
    int a0 = dct[0] + dct[1];
    int a1 = dct[2] + dct[3];
    int a2 = dct[4] + dct[5];
    int a3 = dct[6] + dct[7];
    int a4 = dct[0] - dct[1];
    int a5 = dct[2] - dct[3];
    int a6 = dct[4] - dct[5];
    int a7 = dct[6] - dct[7];
    int b0 = a0 + a1;
    int b1 = a2 + a3;
    int b2 = a4 + a5;
    int b3 = a6 + a7;
    int b4 = a0 - a1;
    int b5 = a2 - a3;
    int b6 = a4 - a5;
    int b7 = a6 - a7;
    int dmf = dequant_mf[i_qp % 6][0] << i_qp / 6;
    dct[0] = ((b0 + b1) * dmf + 32) >> 6;
    dct[1] = ((b2 + b3) * dmf + 32) >> 6;
    dct[2] = ((b0 - b1) * dmf + 32) >> 6;
    dct[3] = ((b2 - b3) * dmf + 32) >> 6;
    dct[4] = ((b4 - b5) * dmf + 32) >> 6;
    dct[5] = ((b6 - b7) * dmf + 32) >> 6;
    dct[6] = ((b4 + b5) * dmf + 32) >> 6;
    dct[7] = ((b6 + b7) * dmf + 32) >> 6;
}

void optimize_chroma_idct_dequant_2x4 (dctcoef out [8], dctcoef dct [8], int dmf) {
    int a0 = dct[0] + dct[1];
    int a1 = dct[2] + dct[3];
    int a2 = dct[4] + dct[5];
    int a3 = dct[6] + dct[7];
    int a4 = dct[0] - dct[1];
    int a5 = dct[2] - dct[3];
    int a6 = dct[4] - dct[5];
    int a7 = dct[6] - dct[7];
    int b0 = a0 + a1;
    int b1 = a2 + a3;
    int b2 = a4 + a5;
    int b3 = a6 + a7;
    int b4 = a0 - a1;
    int b5 = a2 - a3;
    int b6 = a4 - a5;
    int b7 = a6 - a7;
    out[0] = ((b0 + b1) * dmf + 2080) >> 6;
    out[1] = ((b2 + b3) * dmf + 2080) >> 6;
    out[2] = ((b0 - b1) * dmf + 2080) >> 6;
    out[3] = ((b2 - b3) * dmf + 2080) >> 6;
    out[4] = ((b4 - b5) * dmf + 2080) >> 6;
    out[5] = ((b6 - b7) * dmf + 2080) >> 6;
    out[6] = ((b4 + b5) * dmf + 2080) >> 6;
    out[7] = ((b6 + b7) * dmf + 2080) >> 6;
}

void optimize_chroma_idct_dequant_2x2 (dctcoef out [4], dctcoef dct [4], int dmf) {
    int d0 = dct[0] + dct[1];
    int d1 = dct[2] + dct[3];
    int d2 = dct[0] - dct[1];
    int d3 = dct[2] - dct[3];
    out[0] = ((d0 + d1) * dmf >> 5) + 32;
    out[1] = ((d0 - d1) * dmf >> 5) + 32;
    out[2] = ((d2 + d3) * dmf >> 5) + 32;
    out[3] = ((d2 - d3) * dmf >> 5) + 32;
}

int optimize_chroma_round (dctcoef *ref, dctcoef *dct, int dequant_mf, int chroma422) {
    dctcoef out [8];
    if (chroma422)
        optimize_chroma_idct_dequant_2x4 (out, dct, dequant_mf);
    else
        optimize_chroma_idct_dequant_2x2 (out, dct, dequant_mf);
    int sum = 0;
    for (int i = 0; i < (chroma422 ? 8 : 4); i++)
        sum |= ref[i] ^ out[i];
    return sum >> 6;
}

int optimize_chroma_dc_internal (dctcoef *dct, int dequant_mf, int chroma422) {
    dctcoef dct_orig [8];
    int coeff, nz;
    if (chroma422)
        optimize_chroma_idct_dequant_2x4 (dct_orig, dct, dequant_mf);
    else
        optimize_chroma_idct_dequant_2x2 (dct_orig, dct, dequant_mf);
    int sum = 0;
    for (int i = 0; i < (chroma422 ? 8 : 4); i++)
        sum |= dct_orig[i];
    if (!(sum >> 6))
        return 0;
    for (nz = 0, coeff = (chroma422 ? 7 : 3); coeff >= 0; coeff--) {
        int level = dct[coeff];
        int sign = level >> 31 | 1;
        while (level) {
            dct[coeff] = level - sign;
            if (optimize_chroma_round (dct_orig, dct, dequant_mf, chroma422)) {
                nz = 1;
                dct[coeff] = level;
                break;
            }
            level -= sign;
        }
    }
    return nz;
}

int optimize_chroma_2x2_dc (dctcoef dct [4], int dequant_mf) {
    return optimize_chroma_dc_internal (dct, dequant_mf, 0);
}

int optimize_chroma_2x4_dc (dctcoef dct [8], int dequant_mf) {
    return optimize_chroma_dc_internal (dct, dequant_mf, 1);
}

void x264_denoise_dct (dctcoef *dct, uint32_t *sum, udctcoef *offset, int size) {
    for (int i = 0; i < size; i++) {
        int level = dct[i];
        int sign = level >> 31;
        level = (level + sign) ^ sign;
        sum[i] += level;
        level -= offset[i];
        dct[i] = level < 0 ? 0 : (level ^ sign) - sign;
    }
}

const uint8_t x264_decimate_table4 [16] = {3, 2, 2, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
const uint8_t x264_decimate_table8 [64] = {3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

int x264_decimate_score_internal (dctcoef *dct, int i_max) {
    const uint8_t *ds_table = (i_max == 64) ? x264_decimate_table8 : x264_decimate_table4;
    int i_score = 0;
    int idx = i_max - 1;
    while (idx >= 0 && dct[idx] == 0)
        idx--;
    while (idx >= 0) {
        int i_run;
        if ((unsigned ) (dct[idx--] + 1) > 2)
            return 9;
        i_run = 0;
        while (idx >= 0 && dct[idx] == 0) {
            idx--;
            i_run++;
        }
        i_score += ds_table[i_run];
    }
    return i_score;
}

int x264_decimate_score15 (dctcoef *dct) {
    return x264_decimate_score_internal (dct +1, 15);
}

int x264_decimate_score16 (dctcoef *dct) {
    return x264_decimate_score_internal (dct, 16);
}

int x264_decimate_score64 (dctcoef *dct) {
    return x264_decimate_score_internal (dct, 64);
}

int x264_coeff_last4 (dctcoef *l) {
    int i_last = 4 - 1;
    while (i_last >= 0 && l[i_last] == 0)
        i_last--;
    return i_last;
}

int x264_coeff_last8 (dctcoef *l) {
    int i_last = 8 - 1;
    while (i_last >= 0 && l[i_last] == 0)
        i_last--;
    return i_last;
}

int x264_coeff_last15 (dctcoef *l) {
    int i_last = 15 - 1;
    while (i_last >= 0 && l[i_last] == 0)
        i_last--;
    return i_last;
}

int x264_coeff_last16 (dctcoef *l) {
    int i_last = 16 - 1;
    while (i_last >= 0 && l[i_last] == 0)
        i_last--;
    return i_last;
}

int x264_coeff_last64 (dctcoef *l) {
    int i_last = 64 - 1;
    while (i_last >= 0 && l[i_last] == 0)
        i_last--;
    return i_last;
}

int x264_coeff_level_run4 (dctcoef *dct, x264_run_level_t *runlevel) {
    int i_last = runlevel->last = x264_coeff_last4 (dct);
    int i_total = 0;
    int mask = 0;
    do {
        runlevel->level[i_total++] = dct[i_last];
        mask |= 1 << (i_last);
        while (--i_last >= 0 && dct[i_last] == 0)
            ;
    }
    while (i_last >= 0);
    runlevel->mask = mask;
    return i_total;
}

int x264_coeff_level_run8 (dctcoef *dct, x264_run_level_t *runlevel) {
    int i_last = runlevel->last = x264_coeff_last8 (dct);
    int i_total = 0;
    int mask = 0;
    do {
        runlevel->level[i_total++] = dct[i_last];
        mask |= 1 << (i_last);
        while (--i_last >= 0 && dct[i_last] == 0)
            ;
    }
    while (i_last >= 0);
    runlevel->mask = mask;
    return i_total;
}

int x264_coeff_level_run15 (dctcoef *dct, x264_run_level_t *runlevel) {
    int i_last = runlevel->last = x264_coeff_last15 (dct);
    int i_total = 0;
    int mask = 0;
    do {
        runlevel->level[i_total++] = dct[i_last];
        mask |= 1 << (i_last);
        while (--i_last >= 0 && dct[i_last] == 0)
            ;
    }
    while (i_last >= 0);
    runlevel->mask = mask;
    return i_total;
}

int x264_coeff_level_run16 (dctcoef *dct, x264_run_level_t *runlevel) {
    int i_last = runlevel->last = x264_coeff_last16 (dct);
    int i_total = 0;
    int mask = 0;
    do {
        runlevel->level[i_total++] = dct[i_last];
        mask |= 1 << (i_last);
        while (--i_last >= 0 && dct[i_last] == 0)
            ;
    }
    while (i_last >= 0);
    runlevel->mask = mask;
    return i_total;
}

void x264_quant_init (x264_t *h, int cpu, x264_quant_function_t *pf) {
    pf->quant_8x8 = quant_8x8;
    pf->quant_4x4 = quant_4x4;
    pf->quant_4x4x4 = quant_4x4x4;
    pf->quant_4x4_dc = quant_4x4_dc;
    pf->quant_2x2_dc = quant_2x2_dc;
    pf->dequant_4x4 = dequant_4x4;
    pf->dequant_4x4_dc = dequant_4x4_dc;
    pf->dequant_8x8 = dequant_8x8;
    pf->idct_dequant_2x4_dc = idct_dequant_2x4_dc;
    pf->idct_dequant_2x4_dconly = idct_dequant_2x4_dconly;
    pf->optimize_chroma_2x2_dc = optimize_chroma_2x2_dc;
    pf->optimize_chroma_2x4_dc = optimize_chroma_2x4_dc;
    pf->denoise_dct = x264_denoise_dct;
    pf->decimate_score15 = x264_decimate_score15;
    pf->decimate_score16 = x264_decimate_score16;
    pf->decimate_score64 = x264_decimate_score64;
    pf->coeff_last4 = x264_coeff_last4;
    pf->coeff_last8 = x264_coeff_last8;
    pf->coeff_last[DCT_LUMA_AC] = x264_coeff_last15;
    pf->coeff_last[DCT_LUMA_4x4] = x264_coeff_last16;
    pf->coeff_last[DCT_LUMA_8x8] = x264_coeff_last64;
    pf->coeff_level_run4 = x264_coeff_level_run4;
    pf->coeff_level_run8 = x264_coeff_level_run8;
    pf->coeff_level_run[DCT_LUMA_AC] = x264_coeff_level_run15;
    pf->coeff_level_run[DCT_LUMA_4x4] = x264_coeff_level_run16;
    pf->coeff_last[DCT_LUMA_DC] = pf->coeff_last[DCT_CHROMAU_DC] = pf->coeff_last[DCT_CHROMAV_DC] = pf->coeff_last[DCT_CHROMAU_4x4] = pf->coeff_last[DCT_CHROMAV_4x4] = pf->coeff_last[DCT_LUMA_4x4];
    pf->coeff_last[DCT_CHROMA_AC] = pf->coeff_last[DCT_CHROMAU_AC] = pf->coeff_last[DCT_CHROMAV_AC] = pf->coeff_last[DCT_LUMA_AC];
    pf->coeff_last[DCT_CHROMAU_8x8] = pf->coeff_last[DCT_CHROMAV_8x8] = pf->coeff_last[DCT_LUMA_8x8];
    pf->coeff_level_run[DCT_LUMA_DC] = pf->coeff_level_run[DCT_CHROMAU_DC] = pf->coeff_level_run[DCT_CHROMAV_DC] = pf->coeff_level_run[DCT_CHROMAU_4x4] = pf->coeff_level_run[DCT_CHROMAV_4x4] = pf->coeff_level_run[DCT_LUMA_4x4];
    pf->coeff_level_run[DCT_CHROMA_AC] = pf->coeff_level_run[DCT_CHROMAU_AC] = pf->coeff_level_run[DCT_CHROMAV_AC] = pf->coeff_level_run[DCT_LUMA_AC];
}

