
#include "common.h"

#include "ratecontrol.h"

#include "me.h"
struct _ratecontrol_entry_t {
    int pict_type;
    int frame_type;
    int kept_as_ref;
    double qscale;
    int mv_bits;
    int tex_bits;
    int misc_bits;
    uint64_t expected_bits;
    double expected_vbv;
    double new_qscale;
    int new_qp;
    int i_count;
    int p_count;
    int s_count;
    float blurred_complexity;
    char direct_mode;
    int16_t weight [3] [2];
    int16_t i_weight_denom [2];
    int refcount [16];
    int refs;
    int64_t i_duration;
    int64_t i_cpb_duration;
};
typedef struct _ratecontrol_entry_t ratecontrol_entry_t;
struct _predictor_t {
    float coeff_min;
    float coeff;
    float count;
    float decay;
    float offset;
};
typedef struct _predictor_t predictor_t;
struct x264_ratecontrol_t {
    int b_abr;
    int b_2pass;
    int b_vbv;
    int b_vbv_min_rate;
    double fps;
    double bitrate;
    double rate_tolerance;
    double qcompress;
    int nmb;
    int qp_constant [3];
    ratecontrol_entry_t *rce;
    int qp;
    float qpm;
    float qpa_rc;
    float qpa_rc_prev;
    int qpa_aq;
    int qpa_aq_prev;
    float qp_novbv;
    double buffer_size;
    int64_t buffer_fill_final;
    double buffer_fill;
    double buffer_rate;
    double vbv_max_rate;
    predictor_t *pred;
    int single_frame_vbv;
    float rate_factor_max_increment;
    int last_satd;
    double last_rceq;
    double cplxr_sum;
    double expected_bits_sum;
    int64_t filler_bits_sum;
    double wanted_bits_window;
    double cbr_decay;
    double short_term_cplxsum;
    double short_term_cplxcount;
    double rate_factor_constant;
    double ip_offset;
    double pb_offset;
    FILE *p_stat_file_out;
    char *psz_stat_file_tmpname;
    FILE *p_mbtree_stat_file_out;
    char *psz_mbtree_stat_file_tmpname;
    char *psz_mbtree_stat_file_name;
    FILE *p_mbtree_stat_file_in;
    int num_entries;
    ratecontrol_entry_t *entry;
    double last_qscale;
    double last_qscale_for [3];
    int last_non_b_pict_type;
    double accum_p_qp;
    double accum_p_norm;
    double last_accum_p_norm;
    double lmin [3];
    double lmax [3];
    double lstep;
    struct {
        uint16_t *qp_buffer [2];
        int qpbuf_pos;
        int src_mb_count;
        int rescale_enabled;
        float *scale_buffer [2];
        int filtersize [2];
        float *coeffs [2];
        int *pos [2];
        int srcdim [2];
    } mbtree;
    float frame_size_estimated;
    double frame_size_maximum;
    double frame_size_planned;
    double slice_size_planned;
    predictor_t *row_pred;
    predictor_t row_preds [3] [2];
    predictor_t *pred_b_from_p;
    int bframes;
    int bframe_bits;
    int i_zones;
    x264_zone_t *zones;
    x264_zone_t *prev_zone;
    int initial_cpb_removal_delay;
    int initial_cpb_removal_delay_offset;
    double nrt_first_access_unit;
    double previous_cpb_final_arrival_time;
    uint64_t hrd_multiply_denom;
};
int parse_zones (x264_t * h);
int init_pass2 (x264_t *);
float rate_estimate_qscale (x264_t * h);
int update_vbv (x264_t * h, int bits);
void update_vbv_plan (x264_t * h, int overhead);
float predict_size (predictor_t * p, float q, float var);
void update_predictor (predictor_t * p, float q, float var, float bits);

float qp2qscale (float qp) {
    return 0.85f * powf (2.0f, (qp - 12.0f) / 6.0f);
}

float qscale2qp (float qscale) {
    return 12.0f + 6.0f * log2f (qscale / 0.85f);
}

double qscale2bits (ratecontrol_entry_t *rce, double qscale) {
    if (qscale < 0.1)
        qscale = 0.1;
    return (rce->tex_bits + .1) * pow (rce->qscale / qscale, 1.1) + rce->mv_bits * pow (((rce->qscale) > (1) ? (rce->qscale) : (1)) / ((qscale) > (1) ? (qscale) : (1)), 0.5) + rce->misc_bits;
}

uint32_t ac_energy_var (uint64_t sum_ssd, int shift, x264_frame_t *frame, int i, int b_store) {
    uint32_t sum = sum_ssd;
    uint32_t ssd = sum_ssd >> 32;
    if (b_store) {
        frame->i_pixel_sum[i] += sum;
        frame->i_pixel_ssd[i] += ssd;
    }
    return ssd - ((uint64_t) sum * sum >> shift);
}

uint32_t ac_energy_plane (x264_t *h, int mb_x, int mb_y, x264_frame_t *frame, int i, int b_chroma, int b_field, int b_store) {
    int height = b_chroma ? 16 >> h->mb.chroma_v_shift : 16;
    int stride = frame->i_stride[i];
    int offset = b_field ? 16 * mb_x + height * (mb_y & ~1) * stride + (mb_y & 1) * stride : 16 * mb_x + height * mb_y * stride;
    stride <<= b_field;
    if (b_chroma) {
        pixel pix [16 * 16];
        int chromapix = h->luma2chroma_pixel[PIXEL_16x16];
        int shift = 7 - h->mb.chroma_v_shift;
        h->mc.load_deinterleave_chroma_fenc (pix, frame->plane[1] + offset, stride, height);
        return ac_energy_var (h->pixf.var[chromapix] (pix, 16), shift, frame, 1, b_store) + ac_energy_var (h->pixf.var[chromapix] (pix +16 / 2, 16), shift, frame, 2, b_store);
    }
    else
        return ac_energy_var (h->pixf.var[PIXEL_16x16] (frame->plane[i] + offset, stride), 8, frame, i, b_store);
}

uint32_t x264_ac_energy_mb (x264_t *h, int mb_x, int mb_y, x264_frame_t *frame) {
    uint32_t var;
    x264_prefetch_fenc (h, frame, mb_x, mb_y);
    if (h->mb.b_adaptive_mbaff) {
        uint32_t var_interlaced, var_progressive;
        var_interlaced = ac_energy_plane (h, mb_x, mb_y, frame, 0, 0, 1, 1);
        var_progressive = ac_energy_plane (h, mb_x, mb_y, frame, 0, 0, 0, 0);
        if ((h->sps->i_chroma_format_idc == CHROMA_444)) {
            var_interlaced += ac_energy_plane (h, mb_x, mb_y, frame, 1, 0, 1, 1);
            var_progressive += ac_energy_plane (h, mb_x, mb_y, frame, 1, 0, 0, 0);
            var_interlaced += ac_energy_plane (h, mb_x, mb_y, frame, 2, 0, 1, 1);
            var_progressive += ac_energy_plane (h, mb_x, mb_y, frame, 2, 0, 0, 0);
        }
        else {
            var_interlaced += ac_energy_plane (h, mb_x, mb_y, frame, 1, 1, 1, 1);
            var_progressive += ac_energy_plane (h, mb_x, mb_y, frame, 1, 1, 0, 0);
        }
        var = ((var_interlaced) < (var_progressive) ? (var_interlaced) : (var_progressive));
    }
    else {
        var = ac_energy_plane (h, mb_x, mb_y, frame, 0, 0, 0, 1);
        if ((h->sps->i_chroma_format_idc == CHROMA_444)) {
            var += ac_energy_plane (h, mb_x, mb_y, frame, 1, 0, 0, 1);
            var += ac_energy_plane (h, mb_x, mb_y, frame, 2, 0, 0, 1);
        }
        else
            var += ac_energy_plane (h, mb_x, mb_y, frame, 1, 1, 0, 1);
    };
    return var;
}

void x264_adaptive_quant_frame (x264_t *h, x264_frame_t *frame, float *quant_offsets) {
    float strength;
    float avg_adj = 0.f;
    for (int i = 0; i < 3; i++) {
        frame->i_pixel_sum[i] = 0;
        frame->i_pixel_ssd[i] = 0;
    }
    if (h->param.rc.i_aq_mode == 0 || h->param.rc.f_aq_strength == 0) {
        if (h->param.rc.i_aq_mode && h->param.rc.f_aq_strength == 0) {
            if (quant_offsets) {
                for (int mb_xy = 0; mb_xy < h->mb.i_mb_count; mb_xy++)
                    frame->f_qp_offset[mb_xy] = frame->f_qp_offset_aq[mb_xy] = quant_offsets[mb_xy];
                if (h->frames.b_have_lowres)
                    for (int mb_xy = 0; mb_xy < h->mb.i_mb_count; mb_xy++)
                        frame->i_inv_qscale_factor[mb_xy] = x264_exp2fix8 (frame->f_qp_offset[mb_xy]);
            }
            else {
                memset (frame -> f_qp_offset, 0, h -> mb.i_mb_count * sizeof (float));
                memset (frame -> f_qp_offset_aq, 0, h -> mb.i_mb_count * sizeof (float));
                if (h->frames.b_have_lowres)
                    for (int mb_xy = 0; mb_xy < h->mb.i_mb_count; mb_xy++)
                        frame->i_inv_qscale_factor[mb_xy] = 256;
            }
        }
        if (h->param.analyse.i_weighted_pred) {
            for (int mb_y = 0; mb_y < h->mb.i_mb_height; mb_y++)
                for (int mb_x = 0; mb_x < h->mb.i_mb_width; mb_x++)
                    x264_ac_energy_mb (h, mb_x, mb_y, frame);
        }
        else
            return;
    }
    else {
        if (h->param.rc.i_aq_mode == 2) {
            float bit_depth_correction = powf (1 << (8 - 8), 0.5f);
            float avg_adj_pow2 = 0.f;
            for (int mb_y = 0; mb_y < h->mb.i_mb_height; mb_y++)
                for (int mb_x = 0; mb_x < h->mb.i_mb_width; mb_x++) {
                    uint32_t energy = x264_ac_energy_mb (h, mb_x, mb_y, frame);
                    float qp_adj = powf (energy +1, 0.125f);
                    frame->f_qp_offset[mb_x + mb_y * h->mb.i_mb_stride] = qp_adj;
                    avg_adj += qp_adj;
                    avg_adj_pow2 += qp_adj * qp_adj;
                }
            avg_adj /= h->mb.i_mb_count;
            avg_adj_pow2 /= h->mb.i_mb_count;
            strength = h->param.rc.f_aq_strength * avg_adj / bit_depth_correction;
            avg_adj = avg_adj - 0.5f * (avg_adj_pow2 - (14.f * bit_depth_correction)) / avg_adj;
        }
        else
            strength = h->param.rc.f_aq_strength * 1.0397f;
        for (int mb_y = 0; mb_y < h->mb.i_mb_height; mb_y++)
            for (int mb_x = 0; mb_x < h->mb.i_mb_width; mb_x++) {
                float qp_adj;
                int mb_xy = mb_x + mb_y * h->mb.i_mb_stride;
                if (h->param.rc.i_aq_mode == 2) {
                    qp_adj = frame->f_qp_offset[mb_xy];
                    qp_adj = strength * (qp_adj - avg_adj);
                }
                else {
                    uint32_t energy = x264_ac_energy_mb (h, mb_x, mb_y, frame);
                    qp_adj = strength * (x264_log2 (((energy) > (1) ? (energy) : (1))) - (14.427f + 2 * (8 - 8)));
                }
                if (quant_offsets)
                    qp_adj += quant_offsets[mb_xy];
                frame->f_qp_offset[mb_xy] = frame->f_qp_offset_aq[mb_xy] = qp_adj;
                if (h->frames.b_have_lowres)
                    frame->i_inv_qscale_factor[mb_xy] = x264_exp2fix8 (qp_adj);
            }
    }
    for (int i = 0; i < 3; i++) {
        uint64_t ssd = frame->i_pixel_ssd[i];
        uint64_t sum = frame->i_pixel_sum[i];
        int width = 16 * h->mb.i_mb_width >> (i && h->mb.chroma_h_shift);
        int height = 16 * h->mb.i_mb_height >> (i && h->mb.chroma_v_shift);
        frame->i_pixel_ssd[i] = ssd - (sum * sum + width * height / 2) / (width * height);
    }
}

int x264_macroblock_tree_rescale_init (x264_t *h, x264_ratecontrol_t *rc) {
    float srcdim [2] = {rc->mbtree.srcdim[0] / 16.f, rc->mbtree.srcdim[1] / 16.f};
    float dstdim [2] = {h->param.i_width / 16.f, h->param.i_height / 16.f};
    int srcdimi [2] = {ceil (srcdim[0]), ceil (srcdim[1])};
    int dstdimi [2] = {ceil (dstdim[0]), ceil (dstdim[1])};
    if (0) {
        srcdimi[1] = (srcdimi[1] + 1) & ~1;
        dstdimi[1] = (dstdimi[1] + 1) & ~1;
    }
    rc->mbtree.src_mb_count = srcdimi[0] * srcdimi[1];
    do {
        rc->mbtree.qp_buffer[0] = x264_malloc (rc->mbtree.src_mb_count * sizeof (uint16_t));
        if (!rc->mbtree.qp_buffer[0])
            goto fail;
    }
    while (0);
    if (h->param.i_bframe_pyramid && h->param.rc.b_stat_read)
        do {
            rc->mbtree.qp_buffer[1] = x264_malloc (rc->mbtree.src_mb_count * sizeof (uint16_t));
            if (!rc->mbtree.qp_buffer[1])
                goto fail;
        }
        while (0);
    rc->mbtree.qpbuf_pos = -1;
    if (srcdimi[0] == dstdimi[0] && srcdimi[1] == dstdimi[1])
        return 0;
    rc->mbtree.rescale_enabled = 1;
    do {
        rc->mbtree.scale_buffer[0] = x264_malloc (srcdimi[0] * srcdimi[1] * sizeof (float));
        if (!rc->mbtree.scale_buffer[0])
            goto fail;
    }
    while (0);
    do {
        rc->mbtree.scale_buffer[1] = x264_malloc (dstdimi[0] * srcdimi[1] * sizeof (float));
        if (!rc->mbtree.scale_buffer[1])
            goto fail;
    }
    while (0);
    for (int i = 0; i < 2; i++) {
        if (srcdim[i] > dstdim[i])
            rc->mbtree.filtersize[i] = 1 + (2 * srcdimi[i] + dstdimi[i] - 1) / dstdimi[i];
        else
            rc->mbtree.filtersize[i] = 3;
        do {
            rc->mbtree.coeffs[i] = x264_malloc (rc->mbtree.filtersize[i] * dstdimi[i] * sizeof (float));
            if (!rc->mbtree.coeffs[i])
                goto fail;
        }
        while (0);
        do {
            rc->mbtree.pos[i] = x264_malloc (dstdimi[i] * sizeof (int));
            if (!rc->mbtree.pos[i])
                goto fail;
        }
        while (0);
        float inc = srcdim[i] / dstdim[i];
        float dmul = inc > 1.f ? dstdim[i] / srcdim[i] : 1.f;
        float dstinsrc = 0.5f * inc - 0.5f;
        int filtersize = rc->mbtree.filtersize[i];
        for (int j = 0; j < dstdimi[i]; j++) {
            int pos = dstinsrc - (filtersize - 2.f) * 0.5f;
            float sum = 0.0;
            rc->mbtree.pos[i][j] = pos;
            for (int k = 0; k < filtersize; k++) {
                float d = fabs (pos +k - dstinsrc) * dmul;
                float coeff = ((1.f - d) > (0) ? (1.f - d) : (0));
                rc->mbtree.coeffs[i][j * filtersize + k] = coeff;
                sum += coeff;
            }
            sum = 1.0f / sum;
            for (int k = 0; k < filtersize; k++)
                rc->mbtree.coeffs[i][j * filtersize + k] *= sum;
            dstinsrc += inc;
        }
    }
    rc->mbtree.srcdim[0] = srcdimi[0];
    rc->mbtree.srcdim[1] = srcdimi[1];
    return 0;
fail :
    return -1;
}

void x264_macroblock_tree_rescale_destroy (x264_ratecontrol_t *rc) {
    for (int i = 0; i < 2; i++) {
        x264_free (rc -> mbtree.qp_buffer [i]);
        x264_free (rc -> mbtree.scale_buffer [i]);
        x264_free (rc -> mbtree.coeffs [i]);
        x264_free (rc -> mbtree.pos [i]);
    }
}

float tapfilter (float *src, int pos, int max, int stride, float *coeff, int filtersize) {
    float sum = 0.f;
    for (int i = 0; i < filtersize; i++, pos++)
        sum += src[x264_clip3 (pos, 0, max -1) * stride] * coeff[i];
    return sum;
}

void x264_macroblock_tree_rescale (x264_t *h, x264_ratecontrol_t *rc, float *dst) {
    float *input, *output;
    int filtersize, stride, height;
    input = rc->mbtree.scale_buffer[0];
    output = rc->mbtree.scale_buffer[1];
    filtersize = rc->mbtree.filtersize[0];
    stride = rc->mbtree.srcdim[0];
    height = rc->mbtree.srcdim[1];
    for (int y = 0; y < height; y++, input += stride, output += h->mb.i_mb_width) {
        float *coeff = rc->mbtree.coeffs[0];
        for (int x = 0; x < h->mb.i_mb_width; x++, coeff += filtersize)
            output[x] = tapfilter (input, rc->mbtree.pos[0][x], stride, 1, coeff, filtersize);
    }
    input = rc->mbtree.scale_buffer[1];
    output = dst;
    filtersize = rc->mbtree.filtersize[1];
    stride = h->mb.i_mb_width;
    height = rc->mbtree.srcdim[1];
    for (int x = 0; x < h->mb.i_mb_width; x++, input++, output++) {
        float *coeff = rc->mbtree.coeffs[1];
        for (int y = 0; y < h->mb.i_mb_height; y++, coeff += filtersize)
            output[y * stride] = tapfilter (input, rc->mbtree.pos[1][y], height, stride, coeff, filtersize);
    }
}

int x264_macroblock_tree_read (x264_t *h, x264_frame_t *frame, float *quant_offsets) {
    x264_ratecontrol_t *rc = h->rc;
    uint8_t i_type_actual = rc->entry[frame->i_frame].pict_type;
    if (rc->entry[frame->i_frame].kept_as_ref) {
        uint8_t i_type;
        if (rc->mbtree.qpbuf_pos < 0) {
            do {
                rc->mbtree.qpbuf_pos++;
                if (!fread (&i_type, 1, 1, rc->p_mbtree_stat_file_in))
                    goto fail;
                if (fread (rc->mbtree.qp_buffer[rc->mbtree.qpbuf_pos], sizeof (uint16_t), rc->mbtree.src_mb_count, rc->p_mbtree_stat_file_in) != rc->mbtree.src_mb_count)
                    goto fail;
                if (i_type != i_type_actual && rc->mbtree.qpbuf_pos == 1) {
                    x264_log (h, 0, "MB-tree frametype %d doesn't match actual frametype %d.\n", i_type, i_type_actual);
                    return -1;
                }
            }
            while (i_type != i_type_actual);
        }
        float *dst = rc->mbtree.rescale_enabled ? rc->mbtree.scale_buffer[0] : frame->f_qp_offset;
        for (int i = 0; i < rc->mbtree.src_mb_count; i++) {
            int16_t qp_fix8 = endian_fix16 (rc->mbtree.qp_buffer[rc->mbtree.qpbuf_pos][i]);
            dst[i] = qp_fix8 * (1.f / 256.f);
        }
        if (rc->mbtree.rescale_enabled)
            x264_macroblock_tree_rescale (h, rc, frame->f_qp_offset);
        if (h->frames.b_have_lowres)
            for (int i = 0; i < h->mb.i_mb_count; i++)
                frame->i_inv_qscale_factor[i] = x264_exp2fix8 (frame->f_qp_offset[i]);
        rc->mbtree.qpbuf_pos--;
    }
    else
        x264_adaptive_quant_frame (h, frame, quant_offsets);
    return 0;
fail :
    x264_log (h, 0, "Incomplete MB-tree stats file.\n");
    return -1;
}

int x264_reference_build_list_optimal (x264_t *h) {
    ratecontrol_entry_t *rce = h->rc->rce;
    x264_frame_t *frames [16];
    x264_weight_t weights [16] [3];
    int refcount [16];
    if (rce->refs != h->i_ref[0])
        return -1;
    memcpy (frames, h -> fref [0], sizeof (frames));
    memcpy (refcount, rce -> refcount, sizeof (refcount));
    memcpy (weights, h -> fenc -> weight, sizeof (weights));
    memset (& h -> fenc -> weight [1] [0], 0, sizeof (x264_weight_t [15] [3]));
    for (int ref = 1; ref < h->i_ref[0]; ref++) {
        int max = -1;
        int bestref = 1;
        for (int i = 1; i < h->i_ref[0]; i++)
            if ((refcount[i]) > (max)) {
                (max) = (refcount[i]);
                (bestref) = (i);
            };
        refcount[bestref] = -1;
        h->fref[0][ref] = frames[bestref];
        memcpy (h -> fenc -> weight [ref], weights [bestref], sizeof (weights [bestref]));
    }
    return 0;
}

char *x264_strcat_filename (char *input, char *suffix) {
    char *output = x264_malloc (strlen (input) +strlen (suffix) + 1);
    if (!output)
        return ((void *) 0);
    strcpy (output, input);
    strcat (output, suffix);
    return output;
}

void x264_ratecontrol_init_reconfigurable (x264_t *h, int b_init) {
    x264_ratecontrol_t *rc = h->rc;
    if (!b_init && rc->b_2pass)
        return;
    if (h->param.rc.i_rc_method == 1) {
        double base_cplx = h->mb.i_mb_count * (h->param.i_bframe ? 120 : 80);
        double mbtree_offset = h->param.rc.b_mb_tree ? (1.0 - h->param.rc.f_qcompress) * 13.5 : 0;
        rc->rate_factor_constant = pow (base_cplx, 1 - rc->qcompress) / qp2qscale (h->param.rc.f_rf_constant + mbtree_offset + (6 * (8 - 8)));
    }
    if (h->param.rc.i_vbv_max_bitrate > 0 && h->param.rc.i_vbv_buffer_size > 0) {
        if (rc->b_vbv_min_rate)
            h->param.rc.i_vbv_max_bitrate = h->param.rc.i_bitrate;
        if (h->param.rc.i_vbv_buffer_size < (int) (h->param.rc.i_vbv_max_bitrate / rc->fps)) {
            h->param.rc.i_vbv_buffer_size = h->param.rc.i_vbv_max_bitrate / rc->fps;
            x264_log (h, 1, "VBV buffer size cannot be smaller than one frame, using %d kbit\n", h -> param.rc.i_vbv_buffer_size);
        }
        int kilobit_size = h->param.i_avcintra_class ? 1024 : 1000;
        int vbv_buffer_size = h->param.rc.i_vbv_buffer_size * kilobit_size;
        int vbv_max_bitrate = h->param.rc.i_vbv_max_bitrate * kilobit_size;
        if (h->param.i_nal_hrd && b_init) {
            h->sps->vui.hrd.i_cpb_cnt = 1;
            h->sps->vui.hrd.b_cbr_hrd = h->param.i_nal_hrd == 2;
            h->sps->vui.hrd.i_time_offset_length = 0;
            h->sps->vui.hrd.i_bit_rate_scale = x264_clip3 (__builtin_ctz (vbv_max_bitrate) -6, 0, 15);
            h->sps->vui.hrd.i_bit_rate_value = vbv_max_bitrate >> (h->sps->vui.hrd.i_bit_rate_scale + 6);
            h->sps->vui.hrd.i_bit_rate_unscaled = h->sps->vui.hrd.i_bit_rate_value << (h->sps->vui.hrd.i_bit_rate_scale + 6);
            h->sps->vui.hrd.i_cpb_size_scale = x264_clip3 (__builtin_ctz (vbv_buffer_size) -4, 0, 15);
            h->sps->vui.hrd.i_cpb_size_value = vbv_buffer_size >> (h->sps->vui.hrd.i_cpb_size_scale + 4);
            h->sps->vui.hrd.i_cpb_size_unscaled = h->sps->vui.hrd.i_cpb_size_value << (h->sps->vui.hrd.i_cpb_size_scale + 4);
            int max_cpb_output_delay = ((h->param.i_keyint_max * 0.5 * h->sps->vui.i_time_scale / h->sps->vui.i_num_units_in_tick) < (2147483647) ? (h->param.i_keyint_max * 0.5 * h->sps->vui.i_time_scale / h->sps->vui.i_num_units_in_tick) : (2147483647));
            int max_dpb_output_delay = h->sps->vui.i_max_dec_frame_buffering * 0.5 * h->sps->vui.i_time_scale / h->sps->vui.i_num_units_in_tick;
            int max_delay = (int) (90000.0 * (double) h->sps->vui.hrd.i_cpb_size_unscaled / h->sps->vui.hrd.i_bit_rate_unscaled + 0.5);
            h->sps->vui.hrd.i_initial_cpb_removal_delay_length = 2 + x264_clip3 (32 - __builtin_clz (max_delay), 4, 22);
            h->sps->vui.hrd.i_cpb_removal_delay_length = x264_clip3 (32 - __builtin_clz (max_cpb_output_delay), 4, 31);
            h->sps->vui.hrd.i_dpb_output_delay_length = x264_clip3 (32 - __builtin_clz (max_dpb_output_delay), 4, 31);
            vbv_buffer_size = h->sps->vui.hrd.i_cpb_size_unscaled;
            vbv_max_bitrate = h->sps->vui.hrd.i_bit_rate_unscaled;
        }
        else if (h->param.i_nal_hrd && !b_init) {
            x264_log (h, 1, "VBV parameters cannot be changed when NAL HRD is in use\n");
            return;
        }
        h->sps->vui.hrd.i_bit_rate_unscaled = vbv_max_bitrate;
        h->sps->vui.hrd.i_cpb_size_unscaled = vbv_buffer_size;
        if (rc->b_vbv_min_rate)
            rc->bitrate = (double) h->param.rc.i_bitrate * kilobit_size;
        rc->buffer_rate = vbv_max_bitrate / rc->fps;
        rc->vbv_max_rate = vbv_max_bitrate;
        rc->buffer_size = vbv_buffer_size;
        rc->single_frame_vbv = rc->buffer_rate * 1.1 > rc->buffer_size;
        rc->cbr_decay = 1.0 - rc->buffer_rate / rc->buffer_size * 0.5 * ((0) > (1.5 - rc->buffer_rate * rc->fps / rc->bitrate) ? (0) : (1.5 - rc->buffer_rate * rc->fps / rc->bitrate));
        if (h->param.rc.i_rc_method == 1 && h->param.rc.f_rf_constant_max) {
            rc->rate_factor_max_increment = h->param.rc.f_rf_constant_max - h->param.rc.f_rf_constant;
            if (rc->rate_factor_max_increment <= 0) {
                x264_log (h, 1, "CRF max must be greater than CRF\n");
                rc->rate_factor_max_increment = 0;
            }
        }
        if (b_init) {
            if (h->param.rc.f_vbv_buffer_init > 1.)
                h->param.rc.f_vbv_buffer_init = x264_clip3f (h->param.rc.f_vbv_buffer_init / h->param.rc.i_vbv_buffer_size, 0, 1);
            h->param.rc.f_vbv_buffer_init = x264_clip3f (((h->param.rc.f_vbv_buffer_init) > (rc->buffer_rate / rc->buffer_size) ? (h->param.rc.f_vbv_buffer_init) : (rc->buffer_rate / rc->buffer_size)), 0, 1);
            rc->buffer_fill_final = rc->buffer_size * h->param.rc.f_vbv_buffer_init * h->sps->vui.i_time_scale;
            rc->b_vbv = 1;
            rc->b_vbv_min_rate = !rc->b_2pass && h->param.rc.i_rc_method == 2 && h->param.rc.i_vbv_max_bitrate <= h->param.rc.i_bitrate;
        }
    }
}

int x264_ratecontrol_new (x264_t *h) {
    x264_ratecontrol_t *rc;
    ;
    do {
        do {
            h->rc = x264_malloc (h->param.i_threads * sizeof (x264_ratecontrol_t));
            if (!h->rc)
                goto fail;
        }
        while (0);
        memset (h -> rc, 0, h -> param.i_threads * sizeof (x264_ratecontrol_t));
    }
    while (0);
    rc = h->rc;
    rc->b_abr = h->param.rc.i_rc_method != 0 && !h->param.rc.b_stat_read;
    rc->b_2pass = h->param.rc.i_rc_method == 2 && h->param.rc.b_stat_read;
    if (h->param.i_fps_num > 0 && h->param.i_fps_den > 0)
        rc->fps = (float) h->param.i_fps_num / h->param.i_fps_den;
    else
        rc->fps = 25.0;
    if (h->param.rc.b_mb_tree) {
        h->param.rc.f_pb_factor = 1;
        rc->qcompress = 1;
    }
    else
        rc->qcompress = h->param.rc.f_qcompress;
    rc->bitrate = h->param.rc.i_bitrate * (h->param.i_avcintra_class ? 1024.: 1000.);
    rc->rate_tolerance = h->param.rc.f_rate_tolerance;
    rc->nmb = h->mb.i_mb_count;
    rc->last_non_b_pict_type = -1;
    rc->cbr_decay = 1.0;
    if (h->param.rc.i_rc_method == 1 && h->param.rc.b_stat_read) {
        x264_log (h, 0, "constant rate-factor is incompatible with 2pass.\n");
        return -1;
    }
    x264_ratecontrol_init_reconfigurable (h, 1);
    if (h->param.i_nal_hrd) {
        uint64_t denom = (uint64_t) h->sps->vui.hrd.i_bit_rate_unscaled * h->sps->vui.i_time_scale;
        uint64_t num = 180000;
        x264_reduce_fraction64 (& num, & denom);
        rc->hrd_multiply_denom = 180000 / num;
        double bits_required = log2 (180000 / rc->hrd_multiply_denom) + log2 (h->sps->vui.i_time_scale) + log2 (h->sps->vui.hrd.i_cpb_size_unscaled);
        if (bits_required >= 63) {
            x264_log (h, 0, "HRD with very large timescale and bufsize not supported\n");
            return -1;
        }
    }
    if (rc->rate_tolerance < 0.01) {
        x264_log (h, 1, "bitrate tolerance too small, using .01\n");
        rc->rate_tolerance = 0.01;
    }
    h->mb.b_variable_qp = rc->b_vbv || h->param.rc.i_aq_mode;
    if (rc->b_abr) {
        rc->accum_p_norm = .01;
        rc->accum_p_qp = ((h->param.rc.i_rc_method == 1 ? h->param.rc.f_rf_constant : 24) + (6 * (8 - 8))) * rc->accum_p_norm;
        rc->cplxr_sum = .01 * pow (7.0e5, rc->qcompress) * pow (h->mb.i_mb_count, 0.5);
        rc->wanted_bits_window = 1.0 * rc->bitrate / rc->fps;
        rc->last_non_b_pict_type = SLICE_TYPE_I;
    }
    rc->ip_offset = 6.0 * log2f (h->param.rc.f_ip_factor);
    rc->pb_offset = 6.0 * log2f (h->param.rc.f_pb_factor);
    rc->qp_constant[SLICE_TYPE_P] = h->param.rc.i_qp_constant;
    rc->qp_constant[SLICE_TYPE_I] = x264_clip3 (h->param.rc.i_qp_constant - rc->ip_offset + 0.5, 0, ((51 + (6 * (8 - 8))) + 18));
    rc->qp_constant[SLICE_TYPE_B] = x264_clip3 (h->param.rc.i_qp_constant + rc->pb_offset + 0.5, 0, ((51 + (6 * (8 - 8))) + 18));
    h->mb.ip_offset = rc->ip_offset + 0.5;
    rc->lstep = pow (2, h->param.rc.i_qp_step / 6.0);
    rc->last_qscale = qp2qscale (26);
    int num_preds = h->param.b_sliced_threads * h->param.i_threads + 1;
    do {
        rc->pred = x264_malloc (5 * sizeof (predictor_t) * num_preds);
        if (!rc->pred)
            goto fail;
    }
    while (0);
    do {
        rc->pred_b_from_p = x264_malloc (sizeof (predictor_t));
        if (!rc->pred_b_from_p)
            goto fail;
    }
    while (0);
    for (int i = 0; i < 3; i++) {
        rc->last_qscale_for[i] = qp2qscale (((h->param.rc.i_rc_method == 1 ? h->param.rc.f_rf_constant : 24) + (6 * (8 - 8))));
        rc->lmin[i] = qp2qscale (h->param.rc.i_qp_min);
        rc->lmax[i] = qp2qscale (h->param.rc.i_qp_max);
        for (int j = 0; j < num_preds; j++) {
            rc->pred[i + j * 5].coeff_min = 2.0 / 4;
            rc->pred[i + j * 5].coeff = 2.0;
            rc->pred[i + j * 5].count = 1.0;
            rc->pred[i + j * 5].decay = 0.5;
            rc->pred[i + j * 5].offset = 0.0;
        }
        for (int j = 0; j < 2; j++) {
            rc->row_preds[i][j].coeff_min = .25 / 4;
            rc->row_preds[i][j].coeff = .25;
            rc->row_preds[i][j].count = 1.0;
            rc->row_preds[i][j].decay = 0.5;
            rc->row_preds[i][j].offset = 0.0;
        }
    }
    *rc->pred_b_from_p = rc->pred[0];
    if (parse_zones (h) < 0) {
        x264_log (h, 0, "failed to parse zones\n");
        return -1;
    }
    if (h->param.rc.b_stat_read) {
        char *p, *stats_in, *stats_buf;
        ((h->param.rc.psz_stat_in) ? (void) (0) : __assert_fail ("h->param.rc.psz_stat_in", "DonorSourceCode/ratecontrol.c", 838, __PRETTY_FUNCTION__));
        stats_buf = stats_in = x264_slurp_file (h->param.rc.psz_stat_in);
        if (!stats_buf) {
            x264_log (h, 0, "ratecontrol_init: can't open stats file\n");
            return -1;
        }
        if (h->param.rc.b_mb_tree) {
            char *mbtree_stats_in = x264_strcat_filename (h->param.rc.psz_stat_in, ".mbtree");
            if (!mbtree_stats_in)
                return -1;
            rc->p_mbtree_stat_file_in = fopen (mbtree_stats_in, "rb");
            x264_free (mbtree_stats_in);
            if (!rc->p_mbtree_stat_file_in) {
                x264_log (h, 0, "ratecontrol_init: can't open mbtree stats file\n");
                return -1;
            }
        }
        if (strncmp (stats_buf, "#options:", 9)) {
            x264_log (h, 0, "options list in stats file not valid\n");
            return -1;
        }
        float res_factor, res_factor_bits;
        {
            int i, j;
            uint32_t k, l;
            char *opts = stats_buf;
            stats_in = strchr (stats_buf, '\n');
            if (!stats_in)
                return -1;
            *stats_in = '\0';
            stats_in++;
            if (sscanf (opts, "#options: %dx%d", &i, &j) != 2) {
                x264_log (h, 0, "resolution specified in stats file not valid\n");
                return -1;
            }
            else if (h->param.rc.b_mb_tree) {
                rc->mbtree.srcdim[0] = i;
                rc->mbtree.srcdim[1] = j;
            }
            res_factor = (float) h->param.i_width * h->param.i_height / (i * j);
            res_factor_bits = powf (res_factor, 0.7);
            if (!(p = strstr (opts, "timebase=")) || sscanf (p, "timebase=%u/%u", &k, &l) != 2) {
                x264_log (h, 0, "timebase specified in stats file not valid\n");
                return -1;
            }
            if (k != h->param.i_timebase_num || l != h->param.i_timebase_den) {
                x264_log (h, 0, "timebase mismatch with 1st pass (%u/%u vs %u/%u)\n", h -> param.i_timebase_num, h -> param.i_timebase_den, k, l);
                return -1;
            }
            {
                if ((p = strstr (opts, "bitdepth" "=")) && sscanf (p, "bitdepth" "=%d", &i) && 8 != i) {
                    x264_log (h, 0, "different " "bitdepth" " setting than first pass (%d vs %d)\n", 8, i);
                    return -1;
                }
            };
            {
                if ((p = strstr (opts, "weightp" "=")) && sscanf (p, "weightp" "=%d", &i) && ((0) > (h->param.analyse.i_weighted_pred) ? (0) : (h->param.analyse.i_weighted_pred)) != i) {
                    x264_log (h, 0, "different " "weightp" " setting than first pass (%d vs %d)\n", ((0) > (h -> param.analyse.i_weighted_pred) ? (0) : (h -> param.analyse.i_weighted_pred)), i);
                    return -1;
                }
            };
            {
                if ((p = strstr (opts, "bframes" "=")) && sscanf (p, "bframes" "=%d", &i) && h->param.i_bframe != i) {
                    x264_log (h, 0, "different " "bframes" " setting than first pass (%d vs %d)\n", h -> param.i_bframe, i);
                    return -1;
                }
            };
            {
                if ((p = strstr (opts, "b_pyramid" "=")) && sscanf (p, "b_pyramid" "=%d", &i) && h->param.i_bframe_pyramid != i) {
                    x264_log (h, 0, "different " "b_pyramid" " setting than first pass (%d vs %d)\n", h -> param.i_bframe_pyramid, i);
                    return -1;
                }
            };
            {
                if ((p = strstr (opts, "intra_refresh" "=")) && sscanf (p, "intra_refresh" "=%d", &i) && h->param.b_intra_refresh != i) {
                    x264_log (h, 0, "different " "intra_refresh" " setting than first pass (%d vs %d)\n", h -> param.b_intra_refresh, i);
                    return -1;
                }
            };
            {
                if ((p = strstr (opts, "open_gop" "=")) && sscanf (p, "open_gop" "=%d", &i) && h->param.b_open_gop != i) {
                    x264_log (h, 0, "different " "open_gop" " setting than first pass (%d vs %d)\n", h -> param.b_open_gop, i);
                    return -1;
                }
            };
            {
                if ((p = strstr (opts, "bluray_compat" "=")) && sscanf (p, "bluray_compat" "=%d", &i) && h->param.b_bluray_compat != i) {
                    x264_log (h, 0, "different " "bluray_compat" " setting than first pass (%d vs %d)\n", h -> param.b_bluray_compat, i);
                    return -1;
                }
            };
            if ((p = strstr (opts, "interlaced="))) {
                char *current = h->param.b_interlaced ? h->param.b_tff ? "tff" : "bff" : h->param.b_fake_interlaced ? "fake" : "0";
                char buf [5];
                sscanf (p, "interlaced=%4s", buf);
                if (strcmp (current, buf)) {
                    x264_log (h, 0, "different interlaced setting than first pass (%s vs %s)\n", current, buf);
                    return -1;
                }
            }
            if ((p = strstr (opts, "keyint="))) {
                p += 7;
                char buf [13] = "infinite ";
                if (h->param.i_keyint_max != (1 << 30))
                    sprintf (buf, "%d ", h->param.i_keyint_max);
                if (strncmp (p, buf, strlen (buf))) {
                    x264_log (h, 0, "different keyint setting than first pass (%.*s vs %.*s)\n", strlen (buf) - 1, buf, strcspn (p, " "), p);
                    return -1;
                }
            }
            if (strstr (opts, "qp=0") && h->param.rc.i_rc_method == 2)
                x264_log (h, 1, "1st pass was lossless, bitrate prediction will be inaccurate\n");
            if (!strstr (opts, "direct=3") && h->param.analyse.i_direct_mv_pred == 3) {
                x264_log (h, 1, "direct=auto not used on the first pass\n");
                h->mb.b_direct_auto_write = 1;
            }
            if ((p = strstr (opts, "b_adapt=")) && sscanf (p, "b_adapt=%d", &i) && i >= 0 && i <= 2)
                h->param.i_bframe_adaptive = i;
            else if (h->param.i_bframe) {
                x264_log (h, 0, "b_adapt method specified in stats file not valid\n");
                return -1;
            }
            if ((h->param.rc.b_mb_tree || h->param.rc.i_vbv_buffer_size) && (p = strstr (opts, "rc_lookahead=")) && sscanf (p, "rc_lookahead=%d", &i))
                h->param.rc.i_lookahead = i;
        }
        p = stats_in;
        int num_entries;
        for (num_entries = -1; p; num_entries++)
            p = strchr (p +1, ';');
        if (!num_entries) {
            x264_log (h, 0, "empty stats file\n");
            return -1;
        }
        rc->num_entries = num_entries;
        if (h->param.i_frame_total < rc->num_entries && h->param.i_frame_total > 0) {
            x264_log (h, 1, "2nd pass has fewer frames than 1st pass (%d vs %d)\n", h -> param.i_frame_total, rc -> num_entries);
        }
        if (h->param.i_frame_total > rc->num_entries) {
            x264_log (h, 0, "2nd pass has more frames than 1st pass (%d vs %d)\n", h -> param.i_frame_total, rc -> num_entries);
            return -1;
        }
        do {
            do {
                rc->entry = x264_malloc (rc->num_entries * sizeof (ratecontrol_entry_t));
                if (!rc->entry)
                    goto fail;
            }
            while (0);
            memset (rc -> entry, 0, rc -> num_entries * sizeof (ratecontrol_entry_t));
        }
        while (0);
        for (int i = 0; i < rc->num_entries; i++) {
            ratecontrol_entry_t *rce = &rc->entry[i];
            rce->pict_type = SLICE_TYPE_P;
            rce->qscale = rce->new_qscale = qp2qscale (20);
            rce->misc_bits = rc->nmb + 10;
            rce->new_qp = 0;
        }
        p = stats_in;
        double total_qp_aq = 0;
        for (int i = 0; i < rc->num_entries; i++) {
            ratecontrol_entry_t *rce;
            int frame_number;
            char pict_type;
            int e;
            char *next;
            float qp_rc, qp_aq;
            int ref;
            next = strchr (p, ';');
            if (next)
                *next++ = 0;
            e = sscanf (p, " in:%d ", &frame_number);
            if (frame_number < 0 || frame_number >= rc->num_entries) {
                x264_log (h, 0, "bad frame number (%d) at stats line %d\n", frame_number, i);
                return -1;
            }
            rce = &rc->entry[frame_number];
            rce->direct_mode = 0;
            e += sscanf (p, " in:%*d out:%*d type:%c dur:%" "l" "d" " cpbdur:%" "l" "d" " q:%f aq:%f tex:%d mv:%d misc:%d imb:%d pmb:%d smb:%d d:%c", &pict_type, &rce->i_duration, &rce->i_cpb_duration, &qp_rc, &qp_aq, &rce->tex_bits, &rce->mv_bits, &rce->misc_bits, &rce->i_count, &rce->p_count, &rce->s_count, &rce->direct_mode);
            rce->tex_bits *= res_factor_bits;
            rce->mv_bits *= res_factor_bits;
            rce->misc_bits *= res_factor_bits;
            rce->i_count *= res_factor;
            rce->p_count *= res_factor;
            rce->s_count *= res_factor;
            p = strstr (p, "ref:");
            if (!p)
                goto parse_error;
            p += 4;
            for (ref = 0; ref < 16; ref++) {
                if (sscanf (p, " %d", &rce->refcount[ref]) != 1)
                    break;
                p = strchr (p +1, ' ');
                if (!p)
                    goto parse_error;
            }
            rce->refs = ref;
            rce->i_weight_denom[0] = rce->i_weight_denom[1] = -1;
            char *w = strchr (p, 'w');
            if (w) {
                int count = sscanf (w, "w:%hd,%hd,%hd,%hd,%hd,%hd,%hd,%hd", &rce->i_weight_denom[0], &rce->weight[0][0], &rce->weight[0][1], &rce->i_weight_denom[1], &rce->weight[1][0], &rce->weight[1][1], &rce->weight[2][0], &rce->weight[2][1]);
                if (count == 3)
                    rce->i_weight_denom[1] = -1;
                else if (count != 8)
                    rce->i_weight_denom[0] = rce->i_weight_denom[1] = -1;
            }
            if (pict_type != 'b')
                rce->kept_as_ref = 1;
            switch (pict_type) {
            case 'I' :
                rce->frame_type = 0x0001;
                rce->pict_type = SLICE_TYPE_I;
                break;
            case 'i' :
                rce->frame_type = 0x0002;
                rce->pict_type = SLICE_TYPE_I;
                break;
            case 'P' :
                rce->frame_type = 0x0003;
                rce->pict_type = SLICE_TYPE_P;
                break;
            case 'B' :
                rce->frame_type = 0x0004;
                rce->pict_type = SLICE_TYPE_B;
                break;
            case 'b' :
                rce->frame_type = 0x0005;
                rce->pict_type = SLICE_TYPE_B;
                break;
            default :
                e = -1;
                break;
            }
            if (e < 13) {
            parse_error :
                x264_log (h, 0, "statistics are damaged at line %d, parser out=%d\n", i, e);
                return -1;
            }
            rce->qscale = qp2qscale (qp_rc);
            total_qp_aq += qp_aq;
            p = next;
        }
        if (!h->param.b_stitchable)
            h->pps->i_pic_init_qp = ((((int) (total_qp_aq / rc->num_entries + 0.5))) < ((51 + (6 * (8 - 8)))) ? (((int) (total_qp_aq / rc->num_entries + 0.5))) : ((51 + (6 * (8 - 8)))));
        x264_free (stats_buf);
        if (h->param.rc.i_rc_method == 2) {
            if (init_pass2 (h) < 0)
                return -1;
        }
    }
    if (h->param.rc.b_stat_write) {
        char *p;
        rc->psz_stat_file_tmpname = x264_strcat_filename (h->param.rc.psz_stat_out, ".temp");
        if (!rc->psz_stat_file_tmpname)
            return -1;
        rc->p_stat_file_out = fopen (rc->psz_stat_file_tmpname, "wb");
        if (rc->p_stat_file_out == ((void *) 0)) {
            x264_log (h, 0, "ratecontrol_init: can't open stats file\n");
            return -1;
        }
        p = x264_param2string (&h->param, 1);
        if (p)
            fprintf (rc->p_stat_file_out, "#options: %s\n", p);
        x264_free (p);
        if (h->param.rc.b_mb_tree && !h->param.rc.b_stat_read) {
            rc->psz_mbtree_stat_file_tmpname = x264_strcat_filename (h->param.rc.psz_stat_out, ".mbtree.temp");
            rc->psz_mbtree_stat_file_name = x264_strcat_filename (h->param.rc.psz_stat_out, ".mbtree");
            if (!rc->psz_mbtree_stat_file_tmpname || !rc->psz_mbtree_stat_file_name)
                return -1;
            rc->p_mbtree_stat_file_out = fopen (rc->psz_mbtree_stat_file_tmpname, "wb");
            if (rc->p_mbtree_stat_file_out == ((void *) 0)) {
                x264_log (h, 0, "ratecontrol_init: can't open mbtree stats file\n");
                return -1;
            }
        }
    }
    if (h->param.rc.b_mb_tree && (h->param.rc.b_stat_read || h->param.rc.b_stat_write)) {
        if (!h->param.rc.b_stat_read) {
            rc->mbtree.srcdim[0] = h->param.i_width;
            rc->mbtree.srcdim[1] = h->param.i_height;
        }
        if (x264_macroblock_tree_rescale_init (h, rc) < 0)
            return -1;
    }
    for (int i = 0; i < h->param.i_threads; i++) {
        h->thread[i]->rc = rc + i;
        if (i) {
            rc[i] = rc[0];
            h->thread[i]->param = h->param;
            h->thread[i]->mb.b_variable_qp = h->mb.b_variable_qp;
            h->thread[i]->mb.ip_offset = h->mb.ip_offset;
        }
    }
    return 0;
fail :
    return -1;
}

int parse_zone (x264_t *h, x264_zone_t *z, char *p) {
    int len = 0;
    char *tok, *saveptr = ((void *) 0);
    z->param = ((void *) 0);
    z->f_bitrate_factor = 1;
    if (3 <= sscanf (p, "%d,%d,q=%d%n", &z->i_start, &z->i_end, &z->i_qp, &len))
        z->b_force_qp = 1;
    else if (3 <= sscanf (p, "%d,%d,b=%f%n", &z->i_start, &z->i_end, &z->f_bitrate_factor, &len))
        z->b_force_qp = 0;
    else if (2 <= sscanf (p, "%d,%d%n", &z->i_start, &z->i_end, &len))
        z->b_force_qp = 0;
    else {
        x264_log (h, 0, "invalid zone: \"%s\"\n", p);
        return -1;
    }
    p += len;
    if (!*p)
        return 0;
    do {
        z->param = x264_malloc (sizeof (x264_param_t));
        if (!z->param)
            goto fail;
    }
    while (0);
    memcpy (z -> param, & h -> param, sizeof (x264_param_t));
    z->param->param_free = x264_free;
    while ((tok = strtok_r (p, ",", &saveptr))) {
        char *val = strchr (tok, '=');
        if (val) {
            *val = '\0';
            val++;
        }
        if (x264_param_parse (z->param, tok, val)) {
            x264_log (h, 0, "invalid zone param: %s = %s\n", tok, val);
            return -1;
        }
        p = ((void *) 0);
    }
    return 0;
fail :
    return -1;
}

int parse_zones (x264_t *h) {
    x264_ratecontrol_t *rc = h->rc;
    if (h->param.rc.psz_zones && !h->param.rc.i_zones) {
        char *psz_zones, *p;
        do {
            psz_zones = x264_malloc (strlen (h->param.rc.psz_zones) + 1);
            if (!psz_zones)
                goto fail;
        }
        while (0);
        strcpy (psz_zones, h -> param.rc.psz_zones);
        h->param.rc.i_zones = 1;
        for (p = psz_zones; *p; p++)
            h->param.rc.i_zones += (*p == '/');
        do {
            h->param.rc.zones = x264_malloc (h->param.rc.i_zones * sizeof (x264_zone_t));
            if (!h->param.rc.zones)
                goto fail;
        }
        while (0);
        p = psz_zones;
        for (int i = 0; i < h->param.rc.i_zones; i++) {
            int i_tok = strcspn (p, "/");
            p[i_tok] = 0;
            if (parse_zone (h, &h->param.rc.zones[i], p))
                return -1;
            p += i_tok + 1;
        }
        x264_free (psz_zones);
    }
    if (h->param.rc.i_zones > 0) {
        for (int i = 0; i < h->param.rc.i_zones; i++) {
            x264_zone_t z = h->param.rc.zones[i];
            if (z.i_start < 0 || z.i_start > z.i_end) {
                x264_log (h, 0, "invalid zone: start=%d end=%d\n", z.i_start, z.i_end);
                return -1;
            }
            else if (!z.b_force_qp && z.f_bitrate_factor <= 0) {
                x264_log (h, 0, "invalid zone: bitrate_factor=%f\n", z.f_bitrate_factor);
                return -1;
            }
        }
        rc->i_zones = h->param.rc.i_zones + 1;
        do {
            rc->zones = x264_malloc (rc->i_zones * sizeof (x264_zone_t));
            if (!rc->zones)
                goto fail;
        }
        while (0);
        memcpy (rc -> zones + 1, h -> param.rc.zones, (rc -> i_zones - 1) * sizeof (x264_zone_t));
        rc->zones[0].i_start = 0;
        rc->zones[0].i_end = 2147483647;
        rc->zones[0].b_force_qp = 0;
        rc->zones[0].f_bitrate_factor = 1;
        do {
            rc->zones[0].param = x264_malloc (sizeof (x264_param_t));
            if (!rc->zones[0].param)
                goto fail;
        }
        while (0);
        memcpy (rc -> zones [0].param, & h -> param, sizeof (x264_param_t));
        for (int i = 1; i < rc->i_zones; i++) {
            if (!rc->zones[i].param)
                rc->zones[i].param = rc->zones[0].param;
        }
    }
    return 0;
fail :
    return -1;
}

x264_zone_t *get_zone (x264_t *h, int frame_num) {
    for (int i = h->rc->i_zones - 1; i >= 0; i--) {
        x264_zone_t *z = &h->rc->zones[i];
        if (frame_num >= z->i_start && frame_num <= z->i_end)
            return z;
    }
    return ((void *) 0);
}

void x264_ratecontrol_summary (x264_t *h) {
    x264_ratecontrol_t *rc = h->rc;
    if (rc->b_abr && h->param.rc.i_rc_method == 2 && rc->cbr_decay > .9999) {
        double base_cplx = h->mb.i_mb_count * (h->param.i_bframe ? 120 : 80);
        double mbtree_offset = h->param.rc.b_mb_tree ? (1.0 - h->param.rc.f_qcompress) * 13.5 : 0;
        x264_log (h, 2, "final ratefactor: %.2f\n", qscale2qp (pow (base_cplx, 1 - rc -> qcompress) * rc -> cplxr_sum / rc -> wanted_bits_window) - mbtree_offset - (6 * (8 - 8)));
    }
}

void x264_ratecontrol_delete (x264_t *h) {
    x264_ratecontrol_t *rc = h->rc;
    int b_regular_file;
    if (rc->p_stat_file_out) {
        b_regular_file = x264_is_regular_file (rc->p_stat_file_out);
        fclose (rc -> p_stat_file_out);
        if (h->i_frame >= rc->num_entries && b_regular_file)
            if (rename (rc->psz_stat_file_tmpname, h->param.rc.psz_stat_out) != 0) {
                x264_log (h, 0, "failed to rename \"%s\" to \"%s\"\n", rc -> psz_stat_file_tmpname, h -> param.rc.psz_stat_out);
            }
        x264_free (rc -> psz_stat_file_tmpname);
    }
    if (rc->p_mbtree_stat_file_out) {
        b_regular_file = x264_is_regular_file (rc->p_mbtree_stat_file_out);
        fclose (rc -> p_mbtree_stat_file_out);
        if (h->i_frame >= rc->num_entries && b_regular_file)
            if (rename (rc->psz_mbtree_stat_file_tmpname, rc->psz_mbtree_stat_file_name) != 0) {
                x264_log (h, 0, "failed to rename \"%s\" to \"%s\"\n", rc -> psz_mbtree_stat_file_tmpname, rc -> psz_mbtree_stat_file_name);
            }
        x264_free (rc -> psz_mbtree_stat_file_tmpname);
        x264_free (rc -> psz_mbtree_stat_file_name);
    }
    if (rc->p_mbtree_stat_file_in)
        fclose (rc->p_mbtree_stat_file_in);
    x264_free (rc -> pred);
    x264_free (rc -> pred_b_from_p);
    x264_free (rc -> entry);
    x264_macroblock_tree_rescale_destroy (rc);
    if (rc->zones) {
        x264_free (rc -> zones [0].param);
        for (int i = 1; i < rc->i_zones; i++)
            if (rc->zones[i].param != rc->zones[0].param && rc->zones[i].param->param_free)
                rc->zones[i].param->param_free (rc->zones[i].param);
        x264_free (rc -> zones);
    }
    x264_free (rc);
}

void accum_p_qp_update (x264_t *h, float qp) {
    x264_ratecontrol_t *rc = h->rc;
    rc->accum_p_qp *= .95;
    rc->accum_p_norm *= .95;
    rc->accum_p_norm += 1;
    if (h->sh.i_type == SLICE_TYPE_I)
        rc->accum_p_qp += qp + rc->ip_offset;
    else
        rc->accum_p_qp += qp;
}

void x264_ratecontrol_start (x264_t *h, int i_force_qp, int overhead) {
    x264_ratecontrol_t *rc = h->rc;
    ratecontrol_entry_t *rce = ((void *) 0);
    x264_zone_t *zone = get_zone (h, h->fenc->i_frame);
    float q;
    ;
    if (zone && (!rc->prev_zone || zone->param != rc->prev_zone->param))
        x264_encoder_reconfig_apply (h, zone->param);
    rc->prev_zone = zone;
    if (h->param.rc.b_stat_read) {
        int frame = h->fenc->i_frame;
        ((frame >= 0 && frame < rc->num_entries) ? (void) (0) : __assert_fail ("frame >= 0 && frame < rc->num_entries", "DonorSourceCode/ratecontrol.c", 1382, __PRETTY_FUNCTION__));
        rce = h->rc->rce = &h->rc->entry[frame];
        if (h->sh.i_type == SLICE_TYPE_B && h->param.analyse.i_direct_mv_pred == 3) {
            h->sh.b_direct_spatial_mv_pred = (rce->direct_mode == 's');
            h->mb.b_direct_auto_read = (rce->direct_mode == 's' || rce->direct_mode == 't');
        }
    }
    if (rc->b_vbv) {
        memset (h -> fdec -> i_row_bits, 0, h -> mb.i_mb_height * sizeof (int));
        memset (h -> fdec -> f_row_qp, 0, h -> mb.i_mb_height * sizeof (float));
        memset (h -> fdec -> f_row_qscale, 0, h -> mb.i_mb_height * sizeof (float));
        rc->row_pred = rc->row_preds[h->sh.i_type];
        rc->buffer_rate = h->fenc->i_cpb_duration * rc->vbv_max_rate * h->sps->vui.i_num_units_in_tick / h->sps->vui.i_time_scale;
        update_vbv_plan (h, overhead);
        const x264_level_t *l = x264_levels;
        while (l->level_idc != 0 && l->level_idc != h->param.i_level_idc)
            l++;
        int mincr = l->mincr;
        if (h->param.b_bluray_compat)
            mincr = 4;
        if (h->sps->i_profile_idc > PROFILE_HIGH)
            rc->frame_size_maximum = 1e9;
        else {
            if (h->i_frame == 0) {
                double fr = 1./ 172;
                int pic_size_in_mbs = h->mb.i_mb_width * h->mb.i_mb_height;
                rc->frame_size_maximum = 384 * 8 * ((pic_size_in_mbs) > (fr * l->mbps) ? (pic_size_in_mbs) : (fr * l->mbps)) / mincr;
            }
            else {
                rc->frame_size_maximum = 384 * 8 * ((double) h->fenc->i_cpb_duration * h->sps->vui.i_num_units_in_tick / h->sps->vui.i_time_scale) * l->mbps / mincr;
            }
        }
    }
    if (h->sh.i_type != SLICE_TYPE_B)
        rc->bframes = h->fenc->i_bframes;
    if (rc->b_abr) {
        q = qscale2qp (rate_estimate_qscale (h));
    }
    else if (rc->b_2pass) {
        rce->new_qscale = rate_estimate_qscale (h);
        q = qscale2qp (rce->new_qscale);
    }
    else {
        if (h->sh.i_type == SLICE_TYPE_B && h->fdec->b_kept_as_ref)
            q = (rc->qp_constant[SLICE_TYPE_B] + rc->qp_constant[SLICE_TYPE_P]) / 2;
        else
            q = rc->qp_constant[h->sh.i_type];
        if (zone) {
            if (zone->b_force_qp)
                q += zone->i_qp - rc->qp_constant[SLICE_TYPE_P];
            else
                q -= 6 * log2f (zone->f_bitrate_factor);
        }
    }
    if (i_force_qp != 0)
        q = i_force_qp - 1;
    q = x264_clip3f (q, h->param.rc.i_qp_min, h->param.rc.i_qp_max);
    rc->qpa_rc = rc->qpa_rc_prev = rc->qpa_aq = rc->qpa_aq_prev = 0;
    rc->qp = x264_clip3 (q +0.5f, 0, ((51 + (6 * (8 - 8))) + 18));
    h->fdec->f_qp_avg_rc = h->fdec->f_qp_avg_aq = rc->qpm = q;
    if (rce)
        rce->new_qp = rc->qp;
    accum_p_qp_update (h, rc -> qpm);
    if (h->sh.i_type != SLICE_TYPE_B)
        rc->last_non_b_pict_type = h->sh.i_type;
}

float predict_row_size (x264_t *h, int y, float qscale) {
    x264_ratecontrol_t *rc = h->rc;
    float pred_s = predict_size (&rc->row_pred[0], qscale, h->fdec->i_row_satd[y]);
    if (h->sh.i_type == SLICE_TYPE_I || qscale >= h->fref[0][0]->f_row_qscale[y]) {
        if (h->sh.i_type == SLICE_TYPE_P && h->fref[0][0]->i_type == h->fdec->i_type && h->fref[0][0]->f_row_qscale[y] > 0 && h->fref[0][0]->i_row_satd[y] > 0 && (abs (h->fref[0][0]->i_row_satd[y] - h->fdec->i_row_satd[y]) < h->fdec->i_row_satd[y] / 2)) {
            float pred_t = h->fref[0][0]->i_row_bits[y] * h->fdec->i_row_satd[y] / h->fref[0][0]->i_row_satd[y] * h->fref[0][0]->f_row_qscale[y] / qscale;
            return (pred_s + pred_t) * 0.5f;
        }
        return pred_s;
    }
    else {
        float pred_intra = predict_size (&rc->row_pred[1], qscale, h->fdec->i_row_satds[0][0][y]);
        return pred_intra + pred_s;
    }
}

int row_bits_so_far (x264_t *h, int y) {
    int bits = 0;
    for (int i = h->i_threadslice_start; i <= y; i++)
        bits += h->fdec->i_row_bits[i];
    return bits;
}

float predict_row_size_sum (x264_t *h, int y, float qp) {
    float qscale = qp2qscale (qp);
    float bits = row_bits_so_far (h, y);
    for (int i = y + 1; i < h->i_threadslice_end; i++)
        bits += predict_row_size (h, i, qscale);
    return bits;
}

int x264_ratecontrol_mb (x264_t *h, int bits) {
    x264_ratecontrol_t *rc = h->rc;
    const int y = h->mb.i_mb_y;
    h->fdec->i_row_bits[y] += bits;
    rc->qpa_aq += h->mb.i_qp;
    if (h->mb.i_mb_x != h->mb.i_mb_width - 1)
        return 0;
    ;
    rc->qpa_rc += rc->qpm * h->mb.i_mb_width;
    if (!rc->b_vbv)
        return 0;
    float qscale = qp2qscale (rc->qpm);
    h->fdec->f_row_qp[y] = rc->qpm;
    h->fdec->f_row_qscale[y] = qscale;
    update_predictor (& rc -> row_pred [0], qscale, h -> fdec -> i_row_satd [y], h -> fdec -> i_row_bits [y]);
    if (h->sh.i_type == SLICE_TYPE_P && rc->qpm < h->fref[0][0]->f_row_qp[y])
        update_predictor (&rc->row_pred[1], qscale, h->fdec->i_row_satds[0][0][y], h->fdec->i_row_bits[y]);
    if (0 && !(y & 1))
        return 0;
    int can_reencode_row = h->sh.i_first_mb <= ((h->mb.i_mb_y - 0) * h->mb.i_mb_stride);
    float prev_row_qp = h->fdec->f_row_qp[y];
    float qp_absolute_max = h->param.rc.i_qp_max;
    if (rc->rate_factor_max_increment)
        qp_absolute_max = ((qp_absolute_max) < (rc->qp_novbv + rc->rate_factor_max_increment) ? (qp_absolute_max) : (rc->qp_novbv + rc->rate_factor_max_increment));
    float qp_max = ((prev_row_qp + h->param.rc.i_qp_step) < (qp_absolute_max) ? (prev_row_qp + h->param.rc.i_qp_step) : (qp_absolute_max));
    float qp_min = ((prev_row_qp - h->param.rc.i_qp_step) > (h->param.rc.i_qp_min) ? (prev_row_qp - h->param.rc.i_qp_step) : (h->param.rc.i_qp_min));
    float step_size = 0.5f;
    float buffer_left_planned = rc->buffer_fill - rc->frame_size_planned;
    float slice_size_planned = h->param.b_sliced_threads ? rc->slice_size_planned : rc->frame_size_planned;
    float max_frame_error = ((0.05f) > (1.0f / h->mb.i_mb_height) ? (0.05f) : (1.0f / h->mb.i_mb_height));
    float size_of_other_slices = 0;
    if (h->param.b_sliced_threads) {
        float size_of_other_slices_planned = 0;
        for (int i = 0; i < h->param.i_threads; i++)
            if (h != h->thread[i]) {
                size_of_other_slices += h->thread[i]->rc->frame_size_estimated;
                size_of_other_slices_planned += h->thread[i]->rc->slice_size_planned;
            }
        float weight = rc->slice_size_planned / rc->frame_size_planned;
        size_of_other_slices = (size_of_other_slices - size_of_other_slices_planned) * weight + size_of_other_slices_planned;
    }
    if (y < h->i_threadslice_end - 1) {
        if (h->sh.i_type == SLICE_TYPE_B) {
            qp_min = ((qp_min) > (((h->fref[0][0]->f_row_qp[y + 1]) > (h->fref[1][0]->f_row_qp[y + 1]) ? (h->fref[0][0]->f_row_qp[y + 1]) : (h->fref[1][0]->f_row_qp[y + 1]))) ? (qp_min) : (((h->fref[0][0]->f_row_qp[y + 1]) > (h->fref[1][0]->f_row_qp[y + 1]) ? (h->fref[0][0]->f_row_qp[y + 1]) : (h->fref[1][0]->f_row_qp[y + 1]))));
            rc->qpm = ((rc->qpm) > (qp_min) ? (rc->qpm) : (qp_min));
        }
        float rc_tol = buffer_left_planned / h->param.i_threads * rc->rate_tolerance;
        float b1 = predict_row_size_sum (h, y, rc->qpm) + size_of_other_slices;
        if (row_bits_so_far (h, y) < 0.05f * slice_size_planned)
            qp_max = qp_absolute_max = prev_row_qp;
        if (h->sh.i_type != SLICE_TYPE_I)
            rc_tol *= 0.5f;
        if (!rc->b_vbv_min_rate)
            qp_min = ((qp_min) > (rc->qp_novbv) ? (qp_min) : (rc->qp_novbv));
        while (rc->qpm < qp_max && ((b1 > rc->frame_size_planned + rc_tol) || (rc->buffer_fill - b1 < buffer_left_planned * 0.5f) || (b1 > rc->frame_size_planned && rc->qpm < rc->qp_novbv))) {
            rc->qpm += step_size;
            b1 = predict_row_size_sum (h, y, rc->qpm) + size_of_other_slices;
        }
        while (rc->qpm > qp_min && (rc->qpm > h->fdec->f_row_qp[0] || rc->single_frame_vbv) && ((b1 < rc->frame_size_planned * 0.8f && rc->qpm <= prev_row_qp) || b1 < (rc->buffer_fill - rc->buffer_size + rc->buffer_rate) * 1.1f)) {
            rc->qpm -= step_size;
            b1 = predict_row_size_sum (h, y, rc->qpm) + size_of_other_slices;
        }
        while ((rc->qpm < qp_absolute_max) && ((rc->buffer_fill - b1 < rc->buffer_rate * max_frame_error) || (rc->frame_size_maximum - b1 < rc->frame_size_maximum * max_frame_error))) {
            rc->qpm += step_size;
            b1 = predict_row_size_sum (h, y, rc->qpm) + size_of_other_slices;
        }
        h->rc->frame_size_estimated = b1 - size_of_other_slices;
        if (rc->qpm > qp_max && prev_row_qp < qp_max && can_reencode_row) {
            rc->qpm = x264_clip3f ((prev_row_qp + rc->qpm) * 0.5f, prev_row_qp +1.0f, qp_max);
            rc->qpa_rc = rc->qpa_rc_prev;
            rc->qpa_aq = rc->qpa_aq_prev;
            h->fdec->i_row_bits[y] = 0;
            h->fdec->i_row_bits[y - 0] = 0;
            return -1;
        }
    }
    else {
        h->rc->frame_size_estimated = predict_row_size_sum (h, y, rc->qpm);
        if ((h->rc->frame_size_estimated + size_of_other_slices) > (rc->buffer_fill - rc->buffer_rate * max_frame_error) && rc->qpm < qp_max && can_reencode_row) {
            rc->qpm = qp_max;
            rc->qpa_rc = rc->qpa_rc_prev;
            rc->qpa_aq = rc->qpa_aq_prev;
            h->fdec->i_row_bits[y] = 0;
            h->fdec->i_row_bits[y - 0] = 0;
            return -1;
        }
    }
    rc->qpa_rc_prev = rc->qpa_rc;
    rc->qpa_aq_prev = rc->qpa_aq;
    return 0;
}

int x264_ratecontrol_qp (x264_t *h) {
    ;
    return x264_clip3 (h->rc->qpm + 0.5f, h->param.rc.i_qp_min, h->param.rc.i_qp_max);
}

int x264_ratecontrol_mb_qp (x264_t *h) {
    ;
    float qp = h->rc->qpm;
    if (h->param.rc.i_aq_mode) {
        float qp_offset = h->fdec->b_kept_as_ref ? h->fenc->f_qp_offset[h->mb.i_mb_xy] : h->fenc->f_qp_offset_aq[h->mb.i_mb_xy];
        if (qp > (51 + (6 * (8 - 8))))
            qp_offset *= (((51 + (6 * (8 - 8))) + 18) - qp) / (((51 + (6 * (8 - 8))) + 18) - (51 + (6 * (8 - 8))));
        qp += qp_offset;
    }
    return x264_clip3 (qp +0.5f, h->param.rc.i_qp_min, h->param.rc.i_qp_max);
}

int x264_ratecontrol_slice_type (x264_t *h, int frame_num) {
    x264_ratecontrol_t *rc = h->rc;
    if (h->param.rc.b_stat_read) {
        if (frame_num >= rc->num_entries) {
            h->param.rc.i_qp_constant = (h->stat.i_frame_count[SLICE_TYPE_P] == 0) ? 24 + (6 * (8 - 8)) : 1 + h->stat.f_frame_qp[SLICE_TYPE_P] / h->stat.i_frame_count[SLICE_TYPE_P];
            rc->qp_constant[SLICE_TYPE_P] = x264_clip3 (h->param.rc.i_qp_constant, 0, ((51 + (6 * (8 - 8))) + 18));
            rc->qp_constant[SLICE_TYPE_I] = x264_clip3 ((int) (qscale2qp (qp2qscale (h->param.rc.i_qp_constant) / fabs (h->param.rc.f_ip_factor)) + 0.5), 0, ((51 + (6 * (8 - 8))) + 18));
            rc->qp_constant[SLICE_TYPE_B] = x264_clip3 ((int) (qscale2qp (qp2qscale (h->param.rc.i_qp_constant) * fabs (h->param.rc.f_pb_factor)) + 0.5), 0, ((51 + (6 * (8 - 8))) + 18));
            x264_log (h, 0, "2nd pass has more frames than 1st pass (%d)\n", rc -> num_entries);
            x264_log (h, 0, "continuing anyway, at constant QP=%d\n", h -> param.rc.i_qp_constant);
            if (h->param.i_bframe_adaptive)
                x264_log (h, 0, "disabling adaptive B-frames\n");
            for (int i = 0; i < h->param.i_threads; i++) {
                h->thread[i]->rc->b_abr = 0;
                h->thread[i]->rc->b_2pass = 0;
                h->thread[i]->param.rc.i_rc_method = 0;
                h->thread[i]->param.rc.b_stat_read = 0;
                h->thread[i]->param.i_bframe_adaptive = 0;
                h->thread[i]->param.i_scenecut_threshold = 0;
                h->thread[i]->param.rc.b_mb_tree = 0;
                if (h->thread[i]->param.i_bframe > 1)
                    h->thread[i]->param.i_bframe = 1;
            }
            return 0x0000;
        }
        return rc->entry[frame_num].frame_type;
    }
    else
        return 0x0000;
}

void x264_ratecontrol_set_weights (x264_t *h, x264_frame_t *frm) {
    ratecontrol_entry_t *rce = &h->rc->entry[frm->i_frame];
    if (h->param.analyse.i_weighted_pred <= 0)
        return;
    if (rce->i_weight_denom[0] >= 0) {
        (frm->weight[0][0]).i_scale = (rce->weight[0][0]);
        (frm->weight[0][0]).i_denom = (rce->i_weight_denom[0]);
        (frm->weight[0][0]).i_offset = (rce->weight[0][1]);
        if (1)
            h->mc.weight_cache (h, &frm->weight[0][0]);
        else
            frm->weight[0][0].weightfn = ((void *) 0);
    };
    if (rce->i_weight_denom[1] >= 0) {
        {
            (frm->weight[0][1]).i_scale = (rce->weight[1][0]);
            (frm->weight[0][1]).i_denom = (rce->i_weight_denom[1]);
            (frm->weight[0][1]).i_offset = (rce->weight[1][1]);
            if (1)
                h->mc.weight_cache (h, &frm->weight[0][1]);
            else
                frm->weight[0][1].weightfn = ((void *) 0);
        };
        {
            (frm->weight[0][2]).i_scale = (rce->weight[2][0]);
            (frm->weight[0][2]).i_denom = (rce->i_weight_denom[1]);
            (frm->weight[0][2]).i_offset = (rce->weight[2][1]);
            if (1)
                h->mc.weight_cache (h, &frm->weight[0][2]);
            else
                frm->weight[0][2].weightfn = ((void *) 0);
        };
    }
}

int x264_ratecontrol_end (x264_t *h, int bits, int *filler) {
    x264_ratecontrol_t *rc = h->rc;
    const int *mbs = h->stat.frame.i_mb_count;
    ;
    h->stat.frame.i_mb_count_skip = mbs[P_SKIP] + mbs[B_SKIP];
    h->stat.frame.i_mb_count_i = mbs[I_16x16] + mbs[I_8x8] + mbs[I_4x4];
    h->stat.frame.i_mb_count_p = mbs[P_L0] + mbs[P_8x8];
    for (int i = B_DIRECT; i < B_8x8; i++)
        h->stat.frame.i_mb_count_p += mbs[i];
    h->fdec->f_qp_avg_rc = rc->qpa_rc /= h->mb.i_mb_count;
    h->fdec->f_qp_avg_aq = (float) rc->qpa_aq / h->mb.i_mb_count;
    h->fdec->f_crf_avg = h->param.rc.f_rf_constant + h->fdec->f_qp_avg_rc - rc->qp_novbv;
    if (h->param.rc.b_stat_write) {
        char c_type = h->sh.i_type == SLICE_TYPE_I ? (h->fenc->i_poc == 0 ? 'I' : 'i') : h->sh.i_type == SLICE_TYPE_P ? 'P' : h->fenc->b_kept_as_ref ? 'B' : 'b';
        int dir_frame = h->stat.frame.i_direct_score[1] - h->stat.frame.i_direct_score[0];
        int dir_avg = h->stat.i_direct_score[1] - h->stat.i_direct_score[0];
        char c_direct = h->mb.b_direct_auto_write ? (dir_frame > 0 ? 's' : dir_frame < 0 ? 't' : dir_avg > 0 ? 's' : dir_avg < 0 ? 't' : '-') : '-';
        if (fprintf (rc->p_stat_file_out, "in:%d out:%d type:%c dur:%" "l" "d" " cpbdur:%" "l" "d" " q:%.2f aq:%.2f tex:%d mv:%d misc:%d imb:%d pmb:%d smb:%d d:%c ref:", h->fenc->i_frame, h->i_frame, c_type, h->fenc->i_duration, h->fenc->i_cpb_duration, rc->qpa_rc, h->fdec->f_qp_avg_aq, h->stat.frame.i_tex_bits, h->stat.frame.i_mv_bits, h->stat.frame.i_misc_bits, h->stat.frame.i_mb_count_i, h->stat.frame.i_mb_count_p, h->stat.frame.i_mb_count_skip, c_direct) < 0)
            goto fail;
        int use_old_stats = h->param.rc.b_stat_read && rc->rce->refs > 1;
        for (int i = 0; i < (use_old_stats ? rc->rce->refs : h->i_ref[0]); i++) {
            int refcount = use_old_stats ? rc->rce->refcount[i] : 0 ? h->stat.frame.i_mb_count_ref[0][i * 2] + h->stat.frame.i_mb_count_ref[0][i * 2 + 1] : h->stat.frame.i_mb_count_ref[0][i];
            if (fprintf (rc->p_stat_file_out, "%d ", refcount) < 0)
                goto fail;
        }
        if (h->param.analyse.i_weighted_pred >= 1 && h->sh.weight[0][0].weightfn) {
            if (fprintf (rc->p_stat_file_out, "w:%d,%d,%d", h->sh.weight[0][0].i_denom, h->sh.weight[0][0].i_scale, h->sh.weight[0][0].i_offset) < 0)
                goto fail;
            if (h->sh.weight[0][1].weightfn || h->sh.weight[0][2].weightfn) {
                if (fprintf (rc->p_stat_file_out, ",%d,%d,%d,%d,%d ", h->sh.weight[0][1].i_denom, h->sh.weight[0][1].i_scale, h->sh.weight[0][1].i_offset, h->sh.weight[0][2].i_scale, h->sh.weight[0][2].i_offset) < 0)
                    goto fail;
            }
            else if (fprintf (rc->p_stat_file_out, " ") < 0)
                goto fail;
        }
        if (fprintf (rc->p_stat_file_out, ";\n") < 0)
            goto fail;
        if (h->param.rc.b_mb_tree && h->fenc->b_kept_as_ref && !h->param.rc.b_stat_read) {
            uint8_t i_type = h->sh.i_type;
            for (int i = 0; i < h->mb.i_mb_count; i++)
                rc->mbtree.qp_buffer[0][i] = endian_fix16 (h->fenc->f_qp_offset[i] * 256.0);
            if (fwrite (&i_type, 1, 1, rc->p_mbtree_stat_file_out) < 1)
                goto fail;
            if (fwrite (rc->mbtree.qp_buffer[0], sizeof (uint16_t), h->mb.i_mb_count, rc->p_mbtree_stat_file_out) < h->mb.i_mb_count)
                goto fail;
        }
    }
    if (rc->b_abr) {
        if (h->sh.i_type != SLICE_TYPE_B)
            rc->cplxr_sum += bits * qp2qscale (rc->qpa_rc) / rc->last_rceq;
        else {
            rc->cplxr_sum += bits * qp2qscale (rc->qpa_rc) / (rc->last_rceq * fabs (h->param.rc.f_pb_factor));
        }
        rc->cplxr_sum *= rc->cbr_decay;
        rc->wanted_bits_window += h->fenc->f_duration * rc->bitrate;
        rc->wanted_bits_window *= rc->cbr_decay;
    }
    if (rc->b_2pass)
        rc->expected_bits_sum += qscale2bits (rc->rce, qp2qscale (rc->rce->new_qp));
    if (h->mb.b_variable_qp) {
        if (h->sh.i_type == SLICE_TYPE_B) {
            rc->bframe_bits += bits;
            if (h->fenc->b_last_minigop_bframe) {
                update_predictor (rc -> pred_b_from_p, qp2qscale (rc -> qpa_rc), h -> fref [1] [h -> i_ref [1] - 1] -> i_satd, rc -> bframe_bits / rc -> bframes);
                rc->bframe_bits = 0;
            }
        }
    }
    *filler = update_vbv (h, bits);
    rc->filler_bits_sum += *filler * 8;
    if (h->sps->vui.b_nal_hrd_parameters_present) {
        if (h->fenc->i_frame == 0) {
            h->fenc->hrd_timing.cpb_initial_arrival_time = 0;
            rc->initial_cpb_removal_delay = h->initial_cpb_removal_delay;
            rc->initial_cpb_removal_delay_offset = h->initial_cpb_removal_delay_offset;
            h->fenc->hrd_timing.cpb_removal_time = rc->nrt_first_access_unit = (double) rc->initial_cpb_removal_delay / 90000;
        }
        else {
            h->fenc->hrd_timing.cpb_removal_time = rc->nrt_first_access_unit + (double) (h->fenc->i_cpb_delay - h->i_cpb_delay_pir_offset) * h->sps->vui.i_num_units_in_tick / h->sps->vui.i_time_scale;
            double cpb_earliest_arrival_time = h->fenc->hrd_timing.cpb_removal_time - (double) rc->initial_cpb_removal_delay / 90000;
            if (h->fenc->b_keyframe) {
                rc->nrt_first_access_unit = h->fenc->hrd_timing.cpb_removal_time;
                rc->initial_cpb_removal_delay = h->initial_cpb_removal_delay;
                rc->initial_cpb_removal_delay_offset = h->initial_cpb_removal_delay_offset;
            }
            else
                cpb_earliest_arrival_time -= (double) rc->initial_cpb_removal_delay_offset / 90000;
            if (h->sps->vui.hrd.b_cbr_hrd)
                h->fenc->hrd_timing.cpb_initial_arrival_time = rc->previous_cpb_final_arrival_time;
            else
                h->fenc->hrd_timing.cpb_initial_arrival_time = ((rc->previous_cpb_final_arrival_time) > (cpb_earliest_arrival_time) ? (rc->previous_cpb_final_arrival_time) : (cpb_earliest_arrival_time));
        }
        int filler_bits = *filler ? ((((5 + 1) - h->param.b_annexb)) > (*filler) ? (((5 + 1) - h->param.b_annexb)) : (*filler)) * 8 : 0;
        h->fenc->hrd_timing.cpb_final_arrival_time = rc->previous_cpb_final_arrival_time = h->fenc->hrd_timing.cpb_initial_arrival_time + (double) (bits + filler_bits) / h->sps->vui.hrd.i_bit_rate_unscaled;
        h->fenc->hrd_timing.dpb_output_time = (double) h->fenc->i_dpb_output_delay * h->sps->vui.i_num_units_in_tick / h->sps->vui.i_time_scale + h->fenc->hrd_timing.cpb_removal_time;
    }
    return 0;
fail :
    x264_log (h, 0, "ratecontrol_end: stats file could not be written to\n");
    return -1;
}

double get_qscale (x264_t *h, ratecontrol_entry_t *rce, double rate_factor, int frame_num) {
    x264_ratecontrol_t *rcc = h->rc;
    x264_zone_t *zone = get_zone (h, frame_num);
    double q;
    if (h->param.rc.b_mb_tree) {
        double timescale = (double) h->sps->vui.i_num_units_in_tick / h->sps->vui.i_time_scale;
        q = pow ((0.04f / ((h->param.i_frame_packing == 5) + 1)) / x264_clip3f (rce->i_duration * timescale, (0.01f / ((h->param.i_frame_packing == 5) + 1)), (1.00f / ((h->param.i_frame_packing == 5) + 1))), 1 - h->param.rc.f_qcompress);
    }
    else
        q = pow (rce->blurred_complexity, 1 - rcc->qcompress);
    if (!(sizeof (q) == sizeof (float) ? __finitef (q) : sizeof (q) == sizeof (double) ? __finite (q) : __finitel (q)) || rce->tex_bits + rce->mv_bits == 0)
        q = rcc->last_qscale_for[rce->pict_type];
    else {
        rcc->last_rceq = q;
        q /= rate_factor;
        rcc->last_qscale = q;
    }
    if (zone) {
        if (zone->b_force_qp)
            q = qp2qscale (zone->i_qp);
        else
            q /= zone->f_bitrate_factor;
    }
    return q;
}

double get_diff_limited_q (x264_t *h, ratecontrol_entry_t *rce, double q, int frame_num) {
    x264_ratecontrol_t *rcc = h->rc;
    const int pict_type = rce->pict_type;
    x264_zone_t *zone = get_zone (h, frame_num);
    const double last_p_q = rcc->last_qscale_for[SLICE_TYPE_P];
    const double last_non_b_q = rcc->last_qscale_for[rcc->last_non_b_pict_type];
    if (pict_type == SLICE_TYPE_I) {
        double iq = q;
        double pq = qp2qscale (rcc->accum_p_qp / rcc->accum_p_norm);
        double ip_factor = fabs (h->param.rc.f_ip_factor);
        if (rcc->accum_p_norm <= 0)
            q = iq;
        else if (h->param.rc.f_ip_factor < 0)
            q = iq / ip_factor;
        else if (rcc->accum_p_norm >= 1)
            q = pq / ip_factor;
        else
            q = rcc->accum_p_norm * pq / ip_factor + (1 - rcc->accum_p_norm) * iq;
    }
    else if (pict_type == SLICE_TYPE_B) {
        if (h->param.rc.f_pb_factor > 0)
            q = last_non_b_q;
        if (!rce->kept_as_ref)
            q *= fabs (h->param.rc.f_pb_factor);
    }
    else if (pict_type == SLICE_TYPE_P && rcc->last_non_b_pict_type == SLICE_TYPE_P && rce->tex_bits == 0) {
        q = last_p_q;
    }
    if (rcc->last_non_b_pict_type == pict_type && (pict_type != SLICE_TYPE_I || rcc->last_accum_p_norm < 1)) {
        double last_q = rcc->last_qscale_for[pict_type];
        double max_qscale = last_q * rcc->lstep;
        double min_qscale = last_q / rcc->lstep;
        if (q > max_qscale)
            q = max_qscale;
        else if (q < min_qscale)
            q = min_qscale;
    }
    rcc->last_qscale_for[pict_type] = q;
    if (pict_type != SLICE_TYPE_B)
        rcc->last_non_b_pict_type = pict_type;
    if (pict_type == SLICE_TYPE_I) {
        rcc->last_accum_p_norm = rcc->accum_p_norm;
        rcc->accum_p_norm = 0;
        rcc->accum_p_qp = 0;
    }
    if (pict_type == SLICE_TYPE_P) {
        float mask = 1 - pow ((float) rce->i_count / rcc->nmb, 2);
        rcc->accum_p_qp = mask * (qscale2qp (q) + rcc->accum_p_qp);
        rcc->accum_p_norm = mask * (1 + rcc->accum_p_norm);
    }
    if (zone) {
        if (zone->b_force_qp)
            q = qp2qscale (zone->i_qp);
        else
            q /= zone->f_bitrate_factor;
    }
    return q;
}

float predict_size (predictor_t *p, float q, float var) {
    return (p->coeff * var + p->offset) / (q * p->count);
}

void update_predictor (predictor_t *p, float q, float var, float bits) {
    float range = 1.5;
    if (var < 10)
        return;
    float old_coeff = p->coeff / p->count;
    float new_coeff = ((bits * q / var) > (p->coeff_min) ? (bits * q / var) : (p->coeff_min));
    float new_coeff_clipped = x264_clip3f (new_coeff, old_coeff / range, old_coeff *range);
    float new_offset = bits * q - new_coeff_clipped * var;
    if (new_offset >= 0)
        new_coeff = new_coeff_clipped;
    else
        new_offset = 0;
    p->count *= p->decay;
    p->coeff *= p->decay;
    p->offset *= p->decay;
    p->count++;
    p->coeff += new_coeff;
    p->offset += new_offset;
}

int update_vbv (x264_t *h, int bits) {
    int filler = 0;
    int bitrate = h->sps->vui.hrd.i_bit_rate_unscaled;
    x264_ratecontrol_t *rcc = h->rc;
    x264_ratecontrol_t *rct = h->thread[0]->rc;
    uint64_t buffer_size = (uint64_t) h->sps->vui.hrd.i_cpb_size_unscaled * h->sps->vui.i_time_scale;
    if (rcc->last_satd >= h->mb.i_mb_count)
        update_predictor (&rct->pred[h->sh.i_type], qp2qscale (rcc->qpa_rc), rcc->last_satd, bits);
    if (!rcc->b_vbv)
        return filler;
    rct->buffer_fill_final -= (uint64_t) bits * h->sps->vui.i_time_scale;
    if (rct->buffer_fill_final < 0) {
        double underflow = (double) rct->buffer_fill_final / h->sps->vui.i_time_scale;
        if (rcc->rate_factor_max_increment && rcc->qpm >= rcc->qp_novbv + rcc->rate_factor_max_increment)
            x264_log (h, 3, "VBV underflow due to CRF-max (frame %d, %.0f bits)\n", h->i_frame, underflow);
        else
            x264_log (h, 1, "VBV underflow (frame %d, %.0f bits)\n", h->i_frame, underflow);
    }
    rct->buffer_fill_final = ((rct->buffer_fill_final) > (0) ? (rct->buffer_fill_final) : (0));
    if (h->param.i_avcintra_class)
        rct->buffer_fill_final += buffer_size;
    else
        rct->buffer_fill_final += (uint64_t) bitrate * h->sps->vui.i_num_units_in_tick * h->fenc->i_cpb_duration;
    if (h->param.rc.b_filler && rct->buffer_fill_final > buffer_size) {
        int64_t scale = (int64_t) h->sps->vui.i_time_scale * 8;
        filler = (rct->buffer_fill_final - buffer_size + scale - 1) / scale;
        bits = h->param.i_avcintra_class ? filler * 8 : ((((5 + 1) - h->param.b_annexb)) > (filler) ? (((5 + 1) - h->param.b_annexb)) : (filler)) * 8;
        rct->buffer_fill_final -= (uint64_t) bits * h->sps->vui.i_time_scale;
    }
    else
        rct->buffer_fill_final = ((rct->buffer_fill_final) < (buffer_size) ? (rct->buffer_fill_final) : (buffer_size));
    return filler;
}

void x264_hrd_fullness (x264_t *h) {
    x264_ratecontrol_t *rct = h->thread[0]->rc;
    uint64_t denom = (uint64_t) h->sps->vui.hrd.i_bit_rate_unscaled * h->sps->vui.i_time_scale / rct->hrd_multiply_denom;
    uint64_t cpb_state = rct->buffer_fill_final;
    uint64_t cpb_size = (uint64_t) h->sps->vui.hrd.i_cpb_size_unscaled * h->sps->vui.i_time_scale;
    uint64_t multiply_factor = 180000 / rct->hrd_multiply_denom;
    if (rct->buffer_fill_final < 0 || rct->buffer_fill_final > cpb_size) {
        x264_log (h, 1, "CPB %s: %.0lf bits in a %.0lf-bit buffer\n", rct -> buffer_fill_final < 0 ? "underflow" : "overflow", (float) rct -> buffer_fill_final / denom, (float) cpb_size / denom);
    }
    h->initial_cpb_removal_delay = (multiply_factor * cpb_state + denom) / (2 * denom);
    h->initial_cpb_removal_delay_offset = (multiply_factor * cpb_size + denom) / (2 * denom) - h->initial_cpb_removal_delay;
}

void update_vbv_plan (x264_t *h, int overhead) {
    x264_ratecontrol_t *rcc = h->rc;
    rcc->buffer_fill = h->thread[0]->rc->buffer_fill_final / h->sps->vui.i_time_scale;
    if (h->i_thread_frames > 1) {
        int j = h->rc - h->thread[0]->rc;
        for (int i = 1; i < h->i_thread_frames; i++) {
            x264_t *t = h->thread[(j + i) % h->i_thread_frames];
            double bits = t->rc->frame_size_planned;
            if (!t->b_thread_active)
                continue;
            bits = ((bits) > (t->rc->frame_size_estimated) ? (bits) : (t->rc->frame_size_estimated));
            rcc->buffer_fill -= bits;
            rcc->buffer_fill = ((rcc->buffer_fill) > (0) ? (rcc->buffer_fill) : (0));
            rcc->buffer_fill += t->rc->buffer_rate;
            rcc->buffer_fill = ((rcc->buffer_fill) < (rcc->buffer_size) ? (rcc->buffer_fill) : (rcc->buffer_size));
        }
    }
    rcc->buffer_fill = ((rcc->buffer_fill) < (rcc->buffer_size) ? (rcc->buffer_fill) : (rcc->buffer_size));
    rcc->buffer_fill -= overhead;
}

double clip_qscale (x264_t *h, int pict_type, double q) {
    x264_ratecontrol_t *rcc = h->rc;
    double lmin = rcc->lmin[pict_type];
    double lmax = rcc->lmax[pict_type];
    if (rcc->rate_factor_max_increment)
        lmax = ((lmax) < (qp2qscale (rcc->qp_novbv + rcc->rate_factor_max_increment)) ? (lmax) : (qp2qscale (rcc->qp_novbv + rcc->rate_factor_max_increment)));
    double q0 = q;
    if (rcc->b_vbv && rcc->last_satd > 0) {
        if (h->param.rc.i_lookahead) {
            int terminate = 0;
            for (int iterations = 0; iterations < 1000 && terminate != 3; iterations++) {
                double frame_q [3];
                double cur_bits = predict_size (&rcc->pred[h->sh.i_type], q, rcc->last_satd);
                double buffer_fill_cur = rcc->buffer_fill - cur_bits;
                double target_fill;
                double total_duration = 0;
                frame_q[0] = h->sh.i_type == SLICE_TYPE_I ? q * h->param.rc.f_ip_factor : q;
                frame_q[1] = frame_q[0] * h->param.rc.f_pb_factor;
                frame_q[2] = frame_q[0] / h->param.rc.f_ip_factor;
                for (int j = 0; buffer_fill_cur >= 0 && buffer_fill_cur <= rcc->buffer_size; j++) {
                    total_duration += h->fenc->f_planned_cpb_duration[j];
                    buffer_fill_cur += rcc->vbv_max_rate * h->fenc->f_planned_cpb_duration[j];
                    int i_type = h->fenc->i_planned_type[j];
                    int i_satd = h->fenc->i_planned_satd[j];
                    if (i_type == 0x0000)
                        break;
                    i_type = ((i_type) == 0x0002 || (i_type) == 0x0001) ? SLICE_TYPE_I : ((i_type) == 0x0005 || (i_type) == 0x0004) ? SLICE_TYPE_B : SLICE_TYPE_P;
                    cur_bits = predict_size (&rcc->pred[i_type], frame_q[i_type], i_satd);
                    buffer_fill_cur -= cur_bits;
                }
                target_fill = ((rcc->buffer_fill + total_duration * rcc->vbv_max_rate * 0.5) < (rcc->buffer_size * 0.5) ? (rcc->buffer_fill + total_duration * rcc->vbv_max_rate * 0.5) : (rcc->buffer_size * 0.5));
                if (buffer_fill_cur < target_fill) {
                    q *= 1.01;
                    terminate |= 1;
                    continue;
                }
                target_fill = x264_clip3f (rcc->buffer_fill - total_duration * rcc->vbv_max_rate * 0.5, rcc->buffer_size * 0.8, rcc->buffer_size);
                if (rcc->b_vbv_min_rate && buffer_fill_cur > target_fill) {
                    q /= 1.01;
                    terminate |= 2;
                    continue;
                }
                break;
            }
        }
        else {
            if ((pict_type == SLICE_TYPE_P || (pict_type == SLICE_TYPE_I && rcc->last_non_b_pict_type == SLICE_TYPE_I)) && rcc->buffer_fill / rcc->buffer_size < 0.5) {
                q /= x264_clip3f (2.0 * rcc->buffer_fill / rcc->buffer_size, 0.5, 1.0);
            }
            double bits = predict_size (&rcc->pred[h->sh.i_type], q, rcc->last_satd);
            double qf = 1.0;
            double max_fill_factor = h->param.rc.i_vbv_buffer_size >= 5 * h->param.rc.i_vbv_max_bitrate / rcc->fps ? 2 : 1;
            double min_fill_factor = rcc->single_frame_vbv ? 1 : 2;
            if (bits > rcc->buffer_fill / max_fill_factor)
                qf = x264_clip3f (rcc->buffer_fill / (max_fill_factor * bits), 0.2, 1.0);
            q /= qf;
            bits *= qf;
            if (bits < rcc->buffer_rate / min_fill_factor)
                q *= bits * min_fill_factor / rcc->buffer_rate;
            q = ((q0) > (q) ? (q0) : (q));
        }
        double bits = predict_size (&rcc->pred[h->sh.i_type], q, rcc->last_satd);
        if (bits > rcc->frame_size_maximum)
            q *= bits / rcc->frame_size_maximum;
        bits = predict_size (&rcc->pred[h->sh.i_type], q, rcc->last_satd);
        if (h->sh.i_type == SLICE_TYPE_P && !rcc->single_frame_vbv) {
            int nb = rcc->bframes;
            double pbbits = bits;
            double bbits = predict_size (rcc->pred_b_from_p, q *h->param.rc.f_pb_factor, rcc->last_satd);
            double space;
            double bframe_cpb_duration = 0;
            double minigop_cpb_duration;
            for (int i = 0; i < nb; i++)
                bframe_cpb_duration += h->fenc->f_planned_cpb_duration[1 + i];
            if (bbits * nb > bframe_cpb_duration * rcc->vbv_max_rate)
                nb = 0;
            pbbits += nb * bbits;
            minigop_cpb_duration = bframe_cpb_duration + h->fenc->f_planned_cpb_duration[0];
            space = rcc->buffer_fill + minigop_cpb_duration * rcc->vbv_max_rate - rcc->buffer_size;
            if (pbbits < space) {
                q *= ((pbbits / space) > (bits / (0.5 * rcc->buffer_size)) ? (pbbits / space) : (bits / (0.5 * rcc->buffer_size)));
            }
            q = ((q0 / 2) > (q) ? (q0 / 2) : (q));
        }
        if (!rcc->b_vbv_min_rate)
            q = ((q0) > (q) ? (q0) : (q));
    }
    if (lmin == lmax)
        return lmin;
    else if (rcc->b_2pass) {
        double min2 = log (lmin);
        double max2 = log (lmax);
        q = (log (q) - min2) / (max2 - min2) - 0.5;
        q = 1.0 / (1.0 + exp (-4 * q));
        q = q * (max2 - min2) + min2;
        return exp (q);
    }
    else
        return x264_clip3f (q, lmin, lmax);
}

float rate_estimate_qscale (x264_t *h) {
    float q;
    x264_ratecontrol_t *rcc = h->rc;
    ratecontrol_entry_t rce = rce;
    int pict_type = h->sh.i_type;
    int64_t total_bits = 8 * (h->stat.i_frame_size[SLICE_TYPE_I] + h->stat.i_frame_size[SLICE_TYPE_P] + h->stat.i_frame_size[SLICE_TYPE_B]) - rcc->filler_bits_sum;
    if (rcc->b_2pass) {
        rce = *rcc->rce;
        if (pict_type != rce.pict_type) {
            x264_log (h, 0, "slice=%c but 2pass stats say %c\n", slice_type_to_char [pict_type], slice_type_to_char [rce.pict_type]);
        }
    }
    if (pict_type == SLICE_TYPE_B) {
        int i0 = ((h->fref_nearest[0]->i_type) == 0x0002 || (h->fref_nearest[0]->i_type) == 0x0001);
        int i1 = ((h->fref_nearest[1]->i_type) == 0x0002 || (h->fref_nearest[1]->i_type) == 0x0001);
        int dt0 = abs (h->fenc->i_poc - h->fref_nearest[0]->i_poc);
        int dt1 = abs (h->fenc->i_poc - h->fref_nearest[1]->i_poc);
        float q0 = h->fref_nearest[0]->f_qp_avg_rc;
        float q1 = h->fref_nearest[1]->f_qp_avg_rc;
        if (h->fref_nearest[0]->i_type == 0x0004)
            q0 -= rcc->pb_offset / 2;
        if (h->fref_nearest[1]->i_type == 0x0004)
            q1 -= rcc->pb_offset / 2;
        if (i0 && i1)
            q = (q0 + q1) / 2 + rcc->ip_offset;
        else if (i0)
            q = q1;
        else if (i1)
            q = q0;
        else
            q = (q0 * dt1 + q1 * dt0) / (dt0 + dt1);
        if (h->fenc->b_kept_as_ref)
            q += rcc->pb_offset / 2;
        else
            q += rcc->pb_offset;
        if (rcc->b_2pass && rcc->b_vbv)
            rcc->frame_size_planned = qscale2bits (&rce, qp2qscale (q));
        else
            rcc->frame_size_planned = predict_size (rcc->pred_b_from_p, qp2qscale (q), h->fref[1][h->i_ref[1] - 1]->i_satd);
        if (rcc->b_vbv)
            rcc->frame_size_planned = ((rcc->frame_size_planned) < (rcc->frame_size_maximum) ? (rcc->frame_size_planned) : (rcc->frame_size_maximum));
        h->rc->frame_size_estimated = rcc->frame_size_planned;
        if (rcc->b_vbv)
            rcc->last_satd = x264_rc_analyse_slice (h);
        rcc->qp_novbv = q;
        return qp2qscale (q);
    }
    else {
        double abr_buffer = 2 * rcc->rate_tolerance * rcc->bitrate;
        if (rcc->b_2pass) {
            double lmin = rcc->lmin[pict_type];
            double lmax = rcc->lmax[pict_type];
            int64_t diff;
            int64_t predicted_bits = total_bits;
            if (rcc->b_vbv) {
                if (h->i_thread_frames > 1) {
                    int j = h->rc - h->thread[0]->rc;
                    for (int i = 1; i < h->i_thread_frames; i++) {
                        x264_t *t = h->thread[(j + i) % h->i_thread_frames];
                        double bits = t->rc->frame_size_planned;
                        if (!t->b_thread_active)
                            continue;
                        bits = ((bits) > (t->rc->frame_size_estimated) ? (bits) : (t->rc->frame_size_estimated));
                        predicted_bits += (int64_t) bits;
                    }
                }
            }
            else {
                if (h->i_frame < h->i_thread_frames)
                    predicted_bits += (int64_t) h->i_frame * rcc->bitrate / rcc->fps;
                else
                    predicted_bits += (int64_t) (h->i_thread_frames - 1) * rcc->bitrate / rcc->fps;
            }
            if (rcc->num_entries > h->i_frame) {
                double final_bits = rcc->entry[rcc->num_entries - 1].expected_bits;
                double video_pos = rce.expected_bits / final_bits;
                double scale_factor = sqrt ((1 - video_pos) * rcc->num_entries);
                abr_buffer *= 0.5 * ((scale_factor) > (0.5) ? (scale_factor) : (0.5));
            }
            diff = predicted_bits - (int64_t) rce.expected_bits;
            q = rce.new_qscale;
            q /= x264_clip3f ((double) (abr_buffer - diff) / abr_buffer, .5, 2);
            if (((h->i_frame + 1 - h->i_thread_frames) >= rcc->fps) && (rcc->expected_bits_sum > 0)) {
                double cur_time = (double) h->i_frame / rcc->num_entries;
                double w = x264_clip3f (cur_time *100, 0.0, 1.0);
                q *= pow ((double) total_bits / rcc->expected_bits_sum, w);
            }
            rcc->qp_novbv = qscale2qp (q);
            if (rcc->b_vbv) {
                double expected_size = qscale2bits (&rce, q);
                double expected_vbv = rcc->buffer_fill + rcc->buffer_rate - expected_size;
                double expected_fullness = rce.expected_vbv / rcc->buffer_size;
                double qmax = q * (2 - expected_fullness);
                double size_constraint = 1 + expected_fullness;
                qmax = ((qmax) > (rce.new_qscale) ? (qmax) : (rce.new_qscale));
                if (expected_fullness < .05)
                    qmax = lmax;
                qmax = ((qmax) < (lmax) ? (qmax) : (lmax));
                while (((expected_vbv < rce.expected_vbv / size_constraint) && (q < qmax)) || ((expected_vbv < 0) && (q < lmax))) {
                    q *= 1.05;
                    expected_size = qscale2bits (&rce, q);
                    expected_vbv = rcc->buffer_fill + rcc->buffer_rate - expected_size;
                }
                rcc->last_satd = x264_rc_analyse_slice (h);
            }
            q = x264_clip3f (q, lmin, lmax);
        }
        else {
            double wanted_bits, overflow = 1;
            rcc->last_satd = x264_rc_analyse_slice (h);
            rcc->short_term_cplxsum *= 0.5;
            rcc->short_term_cplxcount *= 0.5;
            rcc->short_term_cplxsum += rcc->last_satd / (x264_clip3f (h->fenc->f_duration, (0.01f / ((h->param.i_frame_packing == 5) + 1)), (1.00f / ((h->param.i_frame_packing == 5) + 1))) / (0.04f / ((h->param.i_frame_packing == 5) + 1)));
            rcc->short_term_cplxcount++;
            rce.tex_bits = rcc->last_satd;
            rce.blurred_complexity = rcc->short_term_cplxsum / rcc->short_term_cplxcount;
            rce.mv_bits = 0;
            rce.p_count = rcc->nmb;
            rce.i_count = 0;
            rce.s_count = 0;
            rce.qscale = 1;
            rce.pict_type = pict_type;
            rce.i_duration = h->fenc->i_duration;
            if (h->param.rc.i_rc_method == 1) {
                q = get_qscale (h, &rce, rcc->rate_factor_constant, h->fenc->i_frame);
            }
            else {
                q = get_qscale (h, &rce, rcc->wanted_bits_window / rcc->cplxr_sum, h->fenc->i_frame);
                if (!rcc->b_vbv_min_rate && rcc->last_satd) {
                    int i_frame_done = h->i_frame + 1 - h->i_thread_frames;
                    double time_done = i_frame_done / rcc->fps;
                    if (h->param.b_vfr_input && i_frame_done > 0)
                        time_done = ((double) (h->fenc->i_reordered_pts - h->i_reordered_pts_delay)) * h->param.i_timebase_num / h->param.i_timebase_den;
                    wanted_bits = time_done * rcc->bitrate;
                    if (wanted_bits > 0) {
                        abr_buffer *= ((1) > (sqrt (time_done)) ? (1) : (sqrt (time_done)));
                        overflow = x264_clip3f (1.0 + (total_bits - wanted_bits) / abr_buffer, .5, 2);
                        q *= overflow;
                    }
                }
            }
            if (pict_type == SLICE_TYPE_I && h->param.i_keyint_max > 1 && rcc->last_non_b_pict_type != SLICE_TYPE_I) {
                q = qp2qscale (rcc->accum_p_qp / rcc->accum_p_norm);
                q /= fabs (h->param.rc.f_ip_factor);
            }
            else if (h->i_frame > 0) {
                if (h->param.rc.i_rc_method != 1) {
                    double lmin = rcc->last_qscale_for[pict_type] / rcc->lstep;
                    double lmax = rcc->last_qscale_for[pict_type] * rcc->lstep;
                    if (overflow > 1.1 && h->i_frame > 3)
                        lmax *= rcc->lstep;
                    else if (overflow < 0.9)
                        lmin /= rcc->lstep;
                    q = x264_clip3f (q, lmin, lmax);
                }
            }
            else if (h->param.rc.i_rc_method == 1 && rcc->qcompress != 1) {
                q = qp2qscale (((h->param.rc.i_rc_method == 1 ? h->param.rc.f_rf_constant : 24) + (6 * (8 - 8)))) / fabs (h->param.rc.f_ip_factor);
            }
            rcc->qp_novbv = qscale2qp (q);
            q = clip_qscale (h, pict_type, q);
        }
        rcc->last_qscale_for[pict_type] = rcc->last_qscale = q;
        if (!(rcc->b_2pass && !rcc->b_vbv) && h->fenc->i_frame == 0)
            rcc->last_qscale_for[SLICE_TYPE_P] = q * fabs (h->param.rc.f_ip_factor);
        if (rcc->b_2pass && rcc->b_vbv)
            rcc->frame_size_planned = qscale2bits (&rce, q);
        else
            rcc->frame_size_planned = predict_size (&rcc->pred[h->sh.i_type], q, rcc->last_satd);
        if (rcc->single_frame_vbv)
            rcc->frame_size_planned = rcc->buffer_rate;
        if (rcc->b_vbv)
            rcc->frame_size_planned = ((rcc->frame_size_planned) < (rcc->frame_size_maximum) ? (rcc->frame_size_planned) : (rcc->frame_size_maximum));
        h->rc->frame_size_estimated = rcc->frame_size_planned;
        return q;
    }
}

void x264_threads_normalize_predictors (x264_t *h) {
    double totalsize = 0;
    for (int i = 0; i < h->param.i_threads; i++)
        totalsize += h->thread[i]->rc->slice_size_planned;
    double factor = h->rc->frame_size_planned / totalsize;
    for (int i = 0; i < h->param.i_threads; i++)
        h->thread[i]->rc->slice_size_planned *= factor;
}

void x264_threads_distribute_ratecontrol (x264_t *h) {
    int row;
    x264_ratecontrol_t *rc = h->rc;
    ;
    float qscale = qp2qscale (rc->qpm);
    if (h->i_frame == 0)
        for (int i = 0; i < h->param.i_threads; i++) {
            x264_t *t = h->thread[i];
            if (t != h)
                memcpy (t->rc->row_preds, rc->row_preds, sizeof (rc->row_preds));
        }
    for (int i = 0; i < h->param.i_threads; i++) {
        x264_t *t = h->thread[i];
        if (t != h)
            memcpy (t->rc, rc, __builtin_offsetof (x264_ratecontrol_t, row_pred));
        t->rc->row_pred = t->rc->row_preds[h->sh.i_type];
        if (rc->b_vbv && rc->frame_size_planned) {
            int size = 0;
            for (row = t->i_threadslice_start; row < t->i_threadslice_end; row++)
                size += h->fdec->i_row_satd[row];
            t->rc->slice_size_planned = predict_size (&rc->pred[h->sh.i_type + (i + 1) * 5], qscale, size);
        }
        else
            t->rc->slice_size_planned = 0;
    }
    if (rc->b_vbv && rc->frame_size_planned) {
        x264_threads_normalize_predictors (h);
        if (rc->single_frame_vbv) {
            for (int i = 0; i < h->param.i_threads; i++) {
                x264_t *t = h->thread[i];
                float max_frame_error = ((0.05) > (1.0 / (t->i_threadslice_end - t->i_threadslice_start)) ? (0.05) : (1.0 / (t->i_threadslice_end - t->i_threadslice_start)));
                t->rc->slice_size_planned += 2 * max_frame_error * rc->frame_size_planned;
            }
            x264_threads_normalize_predictors (h);
        }
        for (int i = 0; i < h->param.i_threads; i++)
            h->thread[i]->rc->frame_size_estimated = h->thread[i]->rc->slice_size_planned;
    }
}

void x264_threads_merge_ratecontrol (x264_t *h) {
    x264_ratecontrol_t *rc = h->rc;
    ;
    for (int i = 0; i < h->param.i_threads; i++) {
        x264_t *t = h->thread[i];
        x264_ratecontrol_t *rct = h->thread[i]->rc;
        if (h->param.rc.i_vbv_buffer_size) {
            int size = 0;
            for (int row = t->i_threadslice_start; row < t->i_threadslice_end; row++)
                size += h->fdec->i_row_satd[row];
            int bits = t->stat.frame.i_mv_bits + t->stat.frame.i_tex_bits + t->stat.frame.i_misc_bits;
            int mb_count = (t->i_threadslice_end - t->i_threadslice_start) * h->mb.i_mb_width;
            update_predictor (& rc -> pred [h -> sh.i_type + (i + 1) * 5], qp2qscale (rct -> qpa_rc / mb_count), size, bits);
        }
        if (!i)
            continue;
        rc->qpa_rc += rct->qpa_rc;
        rc->qpa_aq += rct->qpa_aq;
    }
}

void x264_thread_sync_ratecontrol (x264_t *cur, x264_t *prev, x264_t *next) {
    if (cur != prev) {
        memcpy (& cur -> rc -> accum_p_qp, & prev -> rc -> accum_p_qp, sizeof (cur -> rc -> accum_p_qp));
        memcpy (& cur -> rc -> accum_p_norm, & prev -> rc -> accum_p_norm, sizeof (cur -> rc -> accum_p_norm));
        memcpy (& cur -> rc -> last_satd, & prev -> rc -> last_satd, sizeof (cur -> rc -> last_satd));
        memcpy (& cur -> rc -> last_rceq, & prev -> rc -> last_rceq, sizeof (cur -> rc -> last_rceq));
        memcpy (& cur -> rc -> last_qscale_for, & prev -> rc -> last_qscale_for, sizeof (cur -> rc -> last_qscale_for));
        memcpy (& cur -> rc -> last_non_b_pict_type, & prev -> rc -> last_non_b_pict_type, sizeof (cur -> rc -> last_non_b_pict_type));
        memcpy (& cur -> rc -> short_term_cplxsum, & prev -> rc -> short_term_cplxsum, sizeof (cur -> rc -> short_term_cplxsum));
        memcpy (& cur -> rc -> short_term_cplxcount, & prev -> rc -> short_term_cplxcount, sizeof (cur -> rc -> short_term_cplxcount));
        memcpy (& cur -> rc -> bframes, & prev -> rc -> bframes, sizeof (cur -> rc -> bframes));
        memcpy (& cur -> rc -> prev_zone, & prev -> rc -> prev_zone, sizeof (cur -> rc -> prev_zone));
        memcpy (& cur -> rc -> mbtree.qpbuf_pos, & prev -> rc -> mbtree.qpbuf_pos, sizeof (cur -> rc -> mbtree.qpbuf_pos));
        memcpy (& cur -> rc -> bitrate, & prev -> rc -> bitrate, sizeof (cur -> rc -> bitrate));
        memcpy (& cur -> rc -> buffer_size, & prev -> rc -> buffer_size, sizeof (cur -> rc -> buffer_size));
        memcpy (& cur -> rc -> buffer_rate, & prev -> rc -> buffer_rate, sizeof (cur -> rc -> buffer_rate));
        memcpy (& cur -> rc -> vbv_max_rate, & prev -> rc -> vbv_max_rate, sizeof (cur -> rc -> vbv_max_rate));
        memcpy (& cur -> rc -> single_frame_vbv, & prev -> rc -> single_frame_vbv, sizeof (cur -> rc -> single_frame_vbv));
        memcpy (& cur -> rc -> cbr_decay, & prev -> rc -> cbr_decay, sizeof (cur -> rc -> cbr_decay));
        memcpy (& cur -> rc -> rate_factor_constant, & prev -> rc -> rate_factor_constant, sizeof (cur -> rc -> rate_factor_constant));
        memcpy (& cur -> rc -> rate_factor_max_increment, & prev -> rc -> rate_factor_max_increment, sizeof (cur -> rc -> rate_factor_max_increment));
    }
    if (cur != next) {
        next->rc->cplxr_sum = cur->rc->cplxr_sum;
        next->rc->expected_bits_sum = cur->rc->expected_bits_sum;
        next->rc->filler_bits_sum = cur->rc->filler_bits_sum;
        next->rc->wanted_bits_window = cur->rc->wanted_bits_window;
        next->rc->bframe_bits = cur->rc->bframe_bits;
        next->rc->initial_cpb_removal_delay = cur->rc->initial_cpb_removal_delay;
        next->rc->initial_cpb_removal_delay_offset = cur->rc->initial_cpb_removal_delay_offset;
        next->rc->nrt_first_access_unit = cur->rc->nrt_first_access_unit;
        next->rc->previous_cpb_final_arrival_time = cur->rc->previous_cpb_final_arrival_time;
    }
}

int find_underflow (x264_t *h, double *fills, int *t0, int *t1, int over) {
    x264_ratecontrol_t *rcc = h->rc;
    const double buffer_min = .1 * rcc->buffer_size;
    const double buffer_max = .9 * rcc->buffer_size;
    double fill = fills[*t0 - 1];
    double parity = over ? 1.: -1.;
    int start = -1, end = -1;
    for (int i = *t0; i < rcc->num_entries; i++) {
        fill += (rcc->entry[i].i_cpb_duration * rcc->vbv_max_rate * h->sps->vui.i_num_units_in_tick / h->sps->vui.i_time_scale - qscale2bits (&rcc->entry[i], rcc->entry[i].new_qscale)) * parity;
        fill = x264_clip3f (fill, 0, rcc->buffer_size);
        fills[i] = fill;
        if (fill <= buffer_min || i == 0) {
            if (end >= 0)
                break;
            start = i;
        }
        else if (fill >= buffer_max && start >= 0)
            end = i;
    }
    *t0 = start;
    *t1 = end;
    return start >= 0 && end >= 0;
}

int fix_underflow (x264_t *h, int t0, int t1, double adjustment, double qscale_min, double qscale_max) {
    x264_ratecontrol_t *rcc = h->rc;
    double qscale_orig, qscale_new;
    int adjusted = 0;
    if (t0 > 0)
        t0++;
    for (int i = t0; i <= t1; i++) {
        qscale_orig = rcc->entry[i].new_qscale;
        qscale_orig = x264_clip3f (qscale_orig, qscale_min, qscale_max);
        qscale_new = qscale_orig * adjustment;
        qscale_new = x264_clip3f (qscale_new, qscale_min, qscale_max);
        rcc->entry[i].new_qscale = qscale_new;
        adjusted = adjusted || (qscale_new != qscale_orig);
    }
    return adjusted;
}

double count_expected_bits (x264_t *h) {
    x264_ratecontrol_t *rcc = h->rc;
    double expected_bits = 0;
    for (int i = 0; i < rcc->num_entries; i++) {
        ratecontrol_entry_t *rce = &rcc->entry[i];
        rce->expected_bits = expected_bits;
        expected_bits += qscale2bits (rce, rce->new_qscale);
    }
    return expected_bits;
}

int vbv_pass2 (x264_t *h, double all_available_bits) {
    x264_ratecontrol_t *rcc = h->rc;
    double *fills;
    double expected_bits = 0;
    double adjustment;
    double prev_bits = 0;
    int t0, t1;
    double qscale_min = qp2qscale (h->param.rc.i_qp_min);
    double qscale_max = qp2qscale (h->param.rc.i_qp_max);
    int iterations = 0;
    int adj_min, adj_max;
    do {
        fills = x264_malloc ((rcc->num_entries + 1) * sizeof (double));
        if (!fills)
            goto fail;
    }
    while (0);
    fills++;
    do {
        iterations++;
        prev_bits = expected_bits;
        if (expected_bits) {
            adjustment = ((((expected_bits / all_available_bits) < (0.999) ? (expected_bits / all_available_bits) : (0.999))) > (0.9) ? (((expected_bits / all_available_bits) < (0.999) ? (expected_bits / all_available_bits) : (0.999))) : (0.9));
            fills[-1] = rcc->buffer_size * h->param.rc.f_vbv_buffer_init;
            t0 = 0;
            adj_min = 1;
            while (adj_min && find_underflow (h, fills, &t0, &t1, 1)) {
                adj_min = fix_underflow (h, t0, t1, adjustment, qscale_min, qscale_max);
                t0 = t1;
            }
        }
        fills[-1] = rcc->buffer_size * (1.- h->param.rc.f_vbv_buffer_init);
        t0 = 0;
        adj_max = 1;
        while (adj_max && find_underflow (h, fills, &t0, &t1, 0))
            adj_max = fix_underflow (h, t0, t1, 1.001, qscale_min, qscale_max);
        expected_bits = count_expected_bits (h);
    }
    while ((expected_bits < .995 * all_available_bits) && ((int64_t) (expected_bits + .5) > (int64_t) (prev_bits + .5)));
    if (!adj_max)
        x264_log (h, 1, "vbv-maxrate issue, qpmax or vbv-maxrate too low\n");
    for (int i = 0; i < rcc->num_entries; i++)
        rcc->entry[i].expected_vbv = rcc->buffer_size - fills[i];
    x264_free (fills - 1);
    return 0;
fail :
    return -1;
}

int init_pass2 (x264_t *h) {
    x264_ratecontrol_t *rcc = h->rc;
    uint64_t all_const_bits = 0;
    double timescale = (double) h->sps->vui.i_num_units_in_tick / h->sps->vui.i_time_scale;
    double duration = 0;
    for (int i = 0; i < rcc->num_entries; i++)
        duration += rcc->entry[i].i_duration;
    duration *= timescale;
    uint64_t all_available_bits = h->param.rc.i_bitrate * 1000.* duration;
    double rate_factor, step_mult;
    double qblur = h->param.rc.f_qblur;
    double cplxblur = h->param.rc.f_complexity_blur;
    const int filter_size = (int) (qblur * 4) | 1;
    double expected_bits;
    double *qscale, *blurred_qscale;
    double base_cplx = h->mb.i_mb_count * (h->param.i_bframe ? 120 : 80);
    for (int i = 0; i < rcc->num_entries; i++) {
        ratecontrol_entry_t *rce = &rcc->entry[i];
        all_const_bits += rce->misc_bits;
    }
    if (all_available_bits < all_const_bits) {
        x264_log (h, 0, "requested bitrate is too low. estimated minimum is %d kbps\n", (int) (all_const_bits * rcc -> fps / (rcc -> num_entries * 1000.)));
        return -1;
    }
    for (int i = 0; i < rcc->num_entries; i++) {
        ratecontrol_entry_t *rce = &rcc->entry[i];
        double weight_sum = 0;
        double cplx_sum = 0;
        double weight = 1.0;
        double gaussian_weight;
        for (int j = 1; j < cplxblur * 2 && j < rcc->num_entries - i; j++) {
            ratecontrol_entry_t *rcj = &rcc->entry[i + j];
            double frame_duration = x264_clip3f (rcj->i_duration * timescale, (0.01f / ((h->param.i_frame_packing == 5) + 1)), (1.00f / ((h->param.i_frame_packing == 5) + 1))) / (0.04f / ((h->param.i_frame_packing == 5) + 1));
            weight *= 1 - pow ((float) rcj->i_count / rcc->nmb, 2);
            if (weight < .0001)
                break;
            gaussian_weight = weight * exp (-j * j / 200.0);
            weight_sum += gaussian_weight;
            cplx_sum += gaussian_weight * (qscale2bits (rcj, 1) - rcj->misc_bits) / frame_duration;
        }
        weight = 1.0;
        for (int j = 0; j <= cplxblur * 2 && j <= i; j++) {
            ratecontrol_entry_t *rcj = &rcc->entry[i - j];
            double frame_duration = x264_clip3f (rcj->i_duration * timescale, (0.01f / ((h->param.i_frame_packing == 5) + 1)), (1.00f / ((h->param.i_frame_packing == 5) + 1))) / (0.04f / ((h->param.i_frame_packing == 5) + 1));
            gaussian_weight = weight * exp (-j * j / 200.0);
            weight_sum += gaussian_weight;
            cplx_sum += gaussian_weight * (qscale2bits (rcj, 1) - rcj->misc_bits) / frame_duration;
            weight *= 1 - pow ((float) rcj->i_count / rcc->nmb, 2);
            if (weight < .0001)
                break;
        }
        rce->blurred_complexity = cplx_sum / weight_sum;
    }
    do {
        qscale = x264_malloc (sizeof (double) * rcc->num_entries);
        if (!qscale)
            goto fail;
    }
    while (0);
    if (filter_size > 1)
        do {
            blurred_qscale = x264_malloc (sizeof (double) * rcc->num_entries);
            if (!blurred_qscale)
                goto fail;
        }
        while (0);
    else
        blurred_qscale = qscale;
    expected_bits = 1;
    for (int i = 0; i < rcc->num_entries; i++) {
        double q = get_qscale (h, &rcc->entry[i], 1.0, i);
        expected_bits += qscale2bits (&rcc->entry[i], q);
        rcc->last_qscale_for[rcc->entry[i].pict_type] = q;
    }
    step_mult = all_available_bits / expected_bits;
    rate_factor = 0;
    for (double step = 1E4 * step_mult; step > 1E-7 * step_mult; step *= 0.5) {
        expected_bits = 0;
        rate_factor += step;
        rcc->last_non_b_pict_type = -1;
        rcc->last_accum_p_norm = 1;
        rcc->accum_p_norm = 0;
        rcc->last_qscale_for[0] = rcc->last_qscale_for[1] = rcc->last_qscale_for[2] = pow (base_cplx, 1 - rcc->qcompress) / rate_factor;
        for (int i = 0; i < rcc->num_entries; i++) {
            qscale[i] = get_qscale (h, &rcc->entry[i], rate_factor, -1);
            rcc->last_qscale_for[rcc->entry[i].pict_type] = qscale[i];
        }
        for (int i = rcc->num_entries - 1; i >= 0; i--) {
            qscale[i] = get_diff_limited_q (h, &rcc->entry[i], qscale[i], i);
            ((qscale[i] >= 0) ? (void) (0) : __assert_fail ("qscale[i] >= 0", "DonorSourceCode/ratecontrol.c", 2946, __PRETTY_FUNCTION__));
        }
        if (filter_size > 1) {
            ((filter_size % 2 == 1) ? (void) (0) : __assert_fail ("filter_size%2 == 1", "DonorSourceCode/ratecontrol.c", 2952, __PRETTY_FUNCTION__));
            for (int i = 0; i < rcc->num_entries; i++) {
                ratecontrol_entry_t *rce = &rcc->entry[i];
                double q = 0.0, sum = 0.0;
                for (int j = 0; j < filter_size; j++) {
                    int idx = i + j - filter_size / 2;
                    double d = idx - i;
                    double coeff = qblur == 0 ? 1.0 : exp (-d * d / (qblur * qblur));
                    if (idx < 0 || idx >= rcc->num_entries)
                        continue;
                    if (rce->pict_type != rcc->entry[idx].pict_type)
                        continue;
                    q += qscale[idx] * coeff;
                    sum += coeff;
                }
                blurred_qscale[i] = q / sum;
            }
        }
        for (int i = 0; i < rcc->num_entries; i++) {
            ratecontrol_entry_t *rce = &rcc->entry[i];
            rce->new_qscale = clip_qscale (h, rce->pict_type, blurred_qscale[i]);
            ((rce->new_qscale >= 0) ? (void) (0) : __assert_fail ("rce->new_qscale >= 0", "DonorSourceCode/ratecontrol.c", 2979, __PRETTY_FUNCTION__));
            expected_bits += qscale2bits (rce, rce->new_qscale);
        }
        if (expected_bits > all_available_bits)
            rate_factor -= step;
    }
    x264_free (qscale);
    if (filter_size > 1)
        x264_free (blurred_qscale);
    if (rcc->b_vbv)
        if (vbv_pass2 (h, all_available_bits))
            return -1;
    expected_bits = count_expected_bits (h);
    if (fabs (expected_bits / all_available_bits - 1.0) > 0.01) {
        double avgq = 0;
        for (int i = 0; i < rcc->num_entries; i++)
            avgq += rcc->entry[i].new_qscale;
        avgq = qscale2qp (avgq / rcc->num_entries);
        if (expected_bits > all_available_bits || !rcc->b_vbv)
            x264_log (h, 1, "Error: 2pass curve failed to converge\n");
        x264_log (h, 1, "target: %.2f kbit/s, expected: %.2f kbit/s, avg QP: %.4f\n", (float) h -> param.rc.i_bitrate, expected_bits * rcc -> fps / (rcc -> num_entries * 1000.), avgq);
        if (expected_bits < all_available_bits && avgq < h->param.rc.i_qp_min + 2) {
            if (h->param.rc.i_qp_min > 0)
                x264_log (h, 1, "try reducing target bitrate or reducing qp_min (currently %d)\n", h->param.rc.i_qp_min);
            else
                x264_log (h, 1, "try reducing target bitrate\n");
        }
        else if (expected_bits > all_available_bits && avgq > h->param.rc.i_qp_max - 2) {
            if (h->param.rc.i_qp_max < ((51 + (6 * (8 - 8))) + 18))
                x264_log (h, 1, "try increasing target bitrate or increasing qp_max (currently %d)\n", h->param.rc.i_qp_max);
            else
                x264_log (h, 1, "try increasing target bitrate\n");
        }
        else if (!(rcc->b_2pass && rcc->b_vbv))
            x264_log (h, 1, "internal error\n");
    }
    return 0;
fail :
    return -1;
}

