/*****************************************************************************
 * rdo.c: rate-distortion optimization
 *****************************************************************************
 * Copyright (C) 2005-2014 x264 project
 *
 * Authors: Loren Merritt <lorenm@u.washington.edu>
 *          Fiona Glaser <fiona@x264.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 * This program is also available under a commercial proprietary license.
 * For more information, contact us at licensing@x264.com.
 *****************************************************************************/

/* duplicate all the writer functions, just calculating bit cost
 * instead of writing the bitstream.
 * TODO: use these for fast 1st pass too. */


#ifndef X264_ENCODER_RDO_H
#define X264_ENCODER_RDO_H


/* Transition and size tables for abs<9 MVD and residual coding */
/* Consist of i_prefix-2 1s, one zero, and a bypass sign bit */
uint8_t x264_cabac_transition_unary[15][128];
uint16_t x264_cabac_size_unary[15][128];
/* Transition and size tables for abs>9 MVD */
/* Consist of 5 1s and a bypass sign bit */
 uint8_t cabac_transition_5ones[128];
 uint16_t cabac_size_5ones[128];










 const uint8_t cbp_to_golomb[2][2][48] =
{
    {{ 0, 1, 2, 5, 3, 6, 14, 10, 4, 15, 7, 11, 8, 12, 13, 9 },
     { 1, 10, 11, 6, 12, 7, 14, 2, 13, 15, 8, 3, 9, 4, 5, 0 }},
    {{ 0, 2, 3, 7, 4, 8, 17, 13, 5, 18, 9, 14, 10, 15, 16, 11,
       1, 32, 33, 36, 34, 37, 44, 40, 35, 45, 38, 41, 39, 42, 43, 19,
       6, 24, 25, 20, 26, 21, 46, 28, 27, 47, 22, 29, 23, 30, 31, 12 },
     { 3, 29, 30, 17, 31, 18, 37, 8, 32, 38, 19, 9, 20, 10, 11, 2,
      16, 33, 34, 21, 35, 22, 39, 4, 36, 40, 23, 5, 24, 6, 7, 1,
      41, 42, 43, 25, 44, 26, 46, 12, 45, 47, 27, 13, 28, 14, 15, 0 }}
};

 const uint8_t mb_type_b_to_golomb[3][9]=
{
    { 4, 8, 12, 10, 6, 14, 16, 18, 20 },
    { 5, 9, 13, 11, 7, 15, 17, 19, 21 },
    { 1, -1, -1, -1, 2, -1, -1, -1, 3 }
};

 const uint8_t subpartition_p_to_golomb[4]=
{
    3, 1, 2, 0
};

 const uint8_t subpartition_b_to_golomb[13]=
{
    10, 4, 5, 1, 11, 6, 7, 2, 12, 8, 9, 3, 0
};








 const uint8_t ct_index[17] = {0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,3};











  void x264_cabac_mb_type_intra_RDO( x264_t *h, x264_cabac_t *cb, int i_mb_type,
                    int ctx0, int ctx1, int ctx2, int ctx3, int ctx4, int ctx5 )
{
    if( i_mb_type == I_4x4 || i_mb_type == I_8x8 )
    {
        x264_cabac_size_decision_noup(cb,ctx0,0);
    }







    else
    {
        int i_pred = x264_mb_pred_mode16x16_fix[h->mb.i_intra16x16_pred_mode];

        x264_cabac_size_decision_noup(cb,ctx0,1);
        ((cb)->f8_bits_encoded += 7);

        x264_cabac_size_decision_noup(cb,ctx1,!!h->mb.i_cbp_luma);
        if( h->mb.i_cbp_chroma == 0 )
            x264_cabac_size_decision_noup(cb,ctx2,0);
        else
        {
            x264_cabac_size_decision(cb,ctx2,1);
            x264_cabac_size_decision_noup(cb,ctx3,h->mb.i_cbp_chroma>>1);
        }
        x264_cabac_size_decision(cb,ctx4,i_pred>>1);
        x264_cabac_size_decision_noup(cb,ctx5,i_pred&1);
    }
}
 void x264_cabac_intra4x4_pred_mode_RDO( x264_cabac_t *cb, int i_pred, int i_mode )
{
    if( i_pred == i_mode )
        x264_cabac_size_decision(cb,68,1);
    else
    {
        x264_cabac_size_decision(cb,68,0);
        if( i_mode > i_pred )
            i_mode--;
        x264_cabac_size_decision(cb,69,(i_mode )&0x01);
        x264_cabac_size_decision(cb,69,(i_mode >> 1)&0x01);
        x264_cabac_size_decision(cb,69,(i_mode >> 2));
    }
}

 void x264_cabac_intra_chroma_pred_mode_RDO( x264_t *h, x264_cabac_t *cb )
{
    int i_mode = x264_mb_chroma_pred_mode_fix[h->mb.i_chroma_pred_mode];
    int ctx = 0;


    if( (h->mb.i_neighbour & MB_LEFT) && h->mb.chroma_pred_mode[h->mb.i_mb_left_xy[0]] != 0 )
        ctx++;
    if( (h->mb.i_neighbour & MB_TOP) && h->mb.chroma_pred_mode[h->mb.i_mb_top_xy] != 0 )
        ctx++;

    x264_cabac_size_decision_noup(cb,64 + ctx,i_mode > 0);
    if( i_mode > 0 )
    {
        x264_cabac_size_decision(cb,64 + 3,i_mode > 1);
        if( i_mode > 1 )
            x264_cabac_size_decision_noup(cb,64 + 3,i_mode > 2);
    }
}

 void x264_cabac_cbp_luma_RDO( x264_t *h, x264_cabac_t *cb )
{
    int cbp = h->mb.i_cbp_luma;
    int cbp_l = h->mb.cache.i_cbp_left;
    int cbp_t = h->mb.cache.i_cbp_top;
    x264_cabac_size_decision(cb,76 - ((cbp_l >> 1) & 1) - ((cbp_t >> 1) & 2),(cbp >> 0) & 1);
    x264_cabac_size_decision(cb,76 - ((cbp >> 0) & 1) - ((cbp_t >> 2) & 2),(cbp >> 1) & 1);
    x264_cabac_size_decision(cb,76 - ((cbp_l >> 3) & 1) - ((cbp << 1) & 2),(cbp >> 2) & 1);
    x264_cabac_size_decision_noup(cb,76 - ((cbp >> 2) & 1) - ((cbp >> 0) & 2),(cbp >> 3) & 1);
}

 void x264_cabac_cbp_chroma_RDO( x264_t *h, x264_cabac_t *cb )
{
    int cbp_a = h->mb.cache.i_cbp_left & 0x30;
    int cbp_b = h->mb.cache.i_cbp_top & 0x30;
    int ctx = 0;

    if( cbp_a && h->mb.cache.i_cbp_left != -1 ) ctx++;
    if( cbp_b && h->mb.cache.i_cbp_top != -1 ) ctx+=2;
    if( h->mb.i_cbp_chroma == 0 )
        x264_cabac_size_decision_noup(cb,77 + ctx,0);
    else
    {
        x264_cabac_size_decision_noup(cb,77 + ctx,1);

        ctx = 4;
        if( cbp_a == 0x20 ) ctx++;
        if( cbp_b == 0x20 ) ctx += 2;
        x264_cabac_size_decision_noup(cb,77 + ctx,h->mb.i_cbp_chroma >> 1);
    }
}

 void x264_cabac_qp_delta_RDO( x264_t *h, x264_cabac_t *cb )
{
    int i_dqp = h->mb.i_qp - h->mb.i_last_qp;
    int ctx;




    if( h->mb.i_type == I_16x16 && !h->mb.cbp[h->mb.i_mb_xy] && h->mb.i_qp > h->mb.i_last_qp )
    {



        i_dqp = 0;
    }

    ctx = h->mb.i_last_dqp && (h->mb.type[h->mb.i_mb_prev_xy] == I_16x16 || (h->mb.cbp[h->mb.i_mb_prev_xy]&0x3f));

    if( i_dqp != 0 )
    {


        i_dqp *= 2;
        int val = 1 - i_dqp;
        if( val < 0 ) val = i_dqp;
        val--;

        if( val >= (51+(6*(8 -8))) && val != (51+(6*(8 -8)))+1 )
            val = 2*(51+(6*(8 -8)))+1 - val;
        do
        {
            x264_cabac_size_decision(cb,60 + ctx,1);
            ctx = 2+(ctx>>1);
        } while( --val );
    }
    x264_cabac_size_decision_noup(cb,60 + ctx,0);
}
  void x264_cabac_subpartition_p_RDO( x264_cabac_t *cb, int i_sub )
{
    if( i_sub == D_L0_8x8 )
    {
        x264_cabac_size_decision(cb,21,1);
        return;
    }
    x264_cabac_size_decision(cb,21,0);
    if( i_sub == D_L0_8x4 )
        x264_cabac_size_decision(cb,22,0);
    else
    {
        x264_cabac_size_decision(cb,22,1);
        x264_cabac_size_decision(cb,23,i_sub == D_L0_4x8);
    }
}

   void x264_cabac_subpartition_b_RDO( x264_cabac_t *cb, int i_sub )
{
    if( i_sub == D_DIRECT_8x8 )
    {
        x264_cabac_size_decision(cb,36,0);
        return;
    }
    x264_cabac_size_decision(cb,36,1);
    if( i_sub == D_BI_8x8 )
    {
        x264_cabac_size_decision(cb,37,1);
        x264_cabac_size_decision(cb,38,0);
        x264_cabac_size_decision(cb,39,0);
        x264_cabac_size_decision(cb,39,0);
        return;
    }
    x264_cabac_size_decision(cb,37,0);
    x264_cabac_size_decision(cb,39,i_sub == D_L1_8x8);
}

   void x264_cabac_transform_size_RDO( x264_t *h, x264_cabac_t *cb )
{
    int ctx = 399 + h->mb.cache.i_neighbour_transform_size;
    x264_cabac_size_decision_noup(cb,ctx,h->mb.b_transform_8x8);
}

   void x264_cabac_ref_internal_RDO( x264_t *h, x264_cabac_t *cb, int i_list, int idx, int bframe )
{
    const int i8 = x264_scan8[idx];
    const int i_refa = h->mb.cache.ref[i_list][i8 - 1];
    const int i_refb = h->mb.cache.ref[i_list][i8 - 8];
    int ctx = 0;

    if( i_refa > 0 && (!bframe || !h->mb.cache.skip[i8 - 1]) )
        ctx++;
    if( i_refb > 0 && (!bframe || !h->mb.cache.skip[i8 - 8]) )
        ctx += 2;

    for( int i_ref = h->mb.cache.ref[i_list][i8]; i_ref > 0; i_ref-- )
    {
        x264_cabac_size_decision(cb,54 + ctx,1);
        ctx = (ctx>>2)+4;
    }
    x264_cabac_size_decision(cb,54 + ctx,0);
}

  void x264_cabac_ref_p_RDO( x264_t *h, x264_cabac_t *cb, int idx )
{
    x264_cabac_ref_internal_RDO( h, cb, 0, idx, 0 );
}
  void x264_cabac_ref_b_RDO( x264_t *h, x264_cabac_t *cb, int i_list, int idx )
{
    x264_cabac_ref_internal_RDO( h, cb, i_list, idx, 1 );
}

   int x264_cabac_mvd_cpn_RDO( x264_t *h, x264_cabac_t *cb, int i_list, int idx, int l, int mvd, int ctx )
{
    int ctxbase = l ? 47 : 40;

    if( mvd == 0 )
    {
        x264_cabac_size_decision(cb,ctxbase + ctx,0);
        return 0;
    }

    int i_abs = abs( mvd );
    x264_cabac_size_decision(cb,ctxbase + ctx,1);

    if( i_abs <= 3 )
    {
        for( int i = 1; i < i_abs; i++ )
            x264_cabac_size_decision(cb,ctxbase + i + 2,1);
        x264_cabac_size_decision(cb,ctxbase + i_abs + 2,0);
        ((cb)->f8_bits_encoded += 256);
    }
    else
    {
        x264_cabac_size_decision(cb,ctxbase + 3,1);
        x264_cabac_size_decision(cb,ctxbase + 4,1);
        x264_cabac_size_decision(cb,ctxbase + 5,1);
        if( i_abs < 9 )
        {
            cb->f8_bits_encoded += x264_cabac_size_unary[i_abs - 3][cb->state[ctxbase+6]];
            cb->state[ctxbase+6] = x264_cabac_transition_unary[i_abs - 3][cb->state[ctxbase+6]];
        }
        else
        {
            cb->f8_bits_encoded += cabac_size_5ones[cb->state[ctxbase+6]];
            cb->state[ctxbase+6] = cabac_transition_5ones[cb->state[ctxbase+6]];
            ((cb)->f8_bits_encoded += (bs_size_ue_big(i_abs - 9 +(1<<3)-1)-3)<<8);
        }
    }
    return ( (i_abs)<(66) ? (i_abs) : (66) );
}

  uint16_t x264_cabac_mvd_RDO( x264_t *h, x264_cabac_t *cb, int i_list, int idx, int width )
{
    int16_t mvp[2] ;
    int mdx, mdy;


    x264_mb_predict_mv( h, i_list, idx, width, mvp );
    mdx = h->mb.cache.mv[i_list][x264_scan8[idx]][0] - mvp[0];
    mdy = h->mb.cache.mv[i_list][x264_scan8[idx]][1] - mvp[1];
    uint16_t amvd = x264_cabac_mvd_sum(h->mb.cache.mvd[i_list][x264_scan8[idx] - 1],
                                       h->mb.cache.mvd[i_list][x264_scan8[idx] - 8]);


    mdx = x264_cabac_mvd_cpn_RDO( h, cb, i_list, idx, 0, mdx, amvd&0xFF );
    mdy = x264_cabac_mvd_cpn_RDO( h, cb, i_list, idx, 1, mdy, amvd>>8 );

    return pack8to16(mdx,mdy);
}
  void x264_cabac_8x8_mvd_RDO( x264_t *h, x264_cabac_t *cb, int i )
{
    switch( h->mb.i_sub_partition[i] )
    {
        case D_L0_8x8:
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4*i,2); x264_macroblock_cache_mvd( h, block_idx_x[4*i], block_idx_y[4*i], 2, 2, 0, mvd );} while(0);
            break;
        case D_L0_8x4:
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4*i+0,2); x264_macroblock_cache_mvd( h, block_idx_x[4*i+0], block_idx_y[4*i+0], 2, 1, 0, mvd );} while(0);
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4*i+2,2); x264_macroblock_cache_mvd( h, block_idx_x[4*i+2], block_idx_y[4*i+2], 2, 1, 0, mvd );} while(0);
            break;
        case D_L0_4x8:
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4*i+0,1); x264_macroblock_cache_mvd( h, block_idx_x[4*i+0], block_idx_y[4*i+0], 1, 2, 0, mvd );} while(0);
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4*i+1,1); x264_macroblock_cache_mvd( h, block_idx_x[4*i+1], block_idx_y[4*i+1], 1, 2, 0, mvd );} while(0);
            break;
        case D_L0_4x4:
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4*i+0,1); x264_macroblock_cache_mvd( h, block_idx_x[4*i+0], block_idx_y[4*i+0], 1, 1, 0, mvd );} while(0);
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4*i+1,1); x264_macroblock_cache_mvd( h, block_idx_x[4*i+1], block_idx_y[4*i+1], 1, 1, 0, mvd );} while(0);
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4*i+2,1); x264_macroblock_cache_mvd( h, block_idx_x[4*i+2], block_idx_y[4*i+2], 1, 1, 0, mvd );} while(0);
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4*i+3,1); x264_macroblock_cache_mvd( h, block_idx_x[4*i+3], block_idx_y[4*i+3], 1, 1, 0, mvd );} while(0);
            break;
        default:
            ((0) ? (void) (0) : __assert_fail ("0", "DonorSourceCode/cabacENCODER.c", 377, __PRETTY_FUNCTION__));
    }
}

   void x264_cabac_mb_header_i_RDO( x264_t *h, x264_cabac_t *cb, int i_mb_type, int slice_type, int chroma )
{
    if( slice_type == SLICE_TYPE_I )
    {
        int ctx = 0;
        if( (h->mb.i_neighbour & MB_LEFT) && h->mb.i_mb_type_left[0] != I_4x4 )
            ctx++;
        if( (h->mb.i_neighbour & MB_TOP) && h->mb.i_mb_type_top != I_4x4 )
            ctx++;

        x264_cabac_mb_type_intra_RDO( h, cb, i_mb_type, 3+ctx, 3+3, 3+4, 3+5, 3+6, 3+7 );
    }
    else if( slice_type == SLICE_TYPE_P )
    {

        x264_cabac_size_decision_noup(cb,14,1);


        x264_cabac_mb_type_intra_RDO( h, cb, i_mb_type, 17+0, 17+1, 17+2, 17+2, 17+3, 17+3 );
    }
    else if( slice_type == SLICE_TYPE_B )
    {

        x264_cabac_size_decision_noup(cb,27+3,1);
        x264_cabac_size_decision_noup(cb,27+4,1);
        x264_cabac_size_decision(cb,27+5,1);
        x264_cabac_size_decision(cb,27+5,0);
        x264_cabac_size_decision(cb,27+5,1);


        x264_cabac_mb_type_intra_RDO( h, cb, i_mb_type, 32+0, 32+1, 32+2, 32+2, 32+3, 32+3 );
    }

    if( i_mb_type == I_PCM )
        return;

    if( i_mb_type != I_16x16 )
    {
        if( h->pps->b_transform_8x8_mode )
            x264_cabac_transform_size_RDO( h, cb );

        int di = h->mb.b_transform_8x8 ? 4 : 1;
        for( int i = 0; i < 16; i += di )
        {
            const int i_pred = x264_mb_predict_intra4x4_mode( h, i );
            const int i_mode = x264_mb_pred_mode4x4_fix[(h->mb.cache.intra4x4_pred_mode[x264_scan8[i]])+1];
            x264_cabac_intra4x4_pred_mode_RDO( cb, i_pred, i_mode );
        }
    }

    if( chroma )
        x264_cabac_intra_chroma_pred_mode_RDO( h, cb );
}

   void x264_cabac_mb_header_p_RDO( x264_t *h, x264_cabac_t *cb, int i_mb_type, int chroma )
{
    if( i_mb_type == P_L0 )
    {
        x264_cabac_size_decision_noup(cb,14,0);
        if( h->mb.i_partition == D_16x16 )
        {
            x264_cabac_size_decision_noup(cb,15,0);
            x264_cabac_size_decision_noup(cb,16,0);
            if( h->mb.pic.i_fref[0] > 1 )
                x264_cabac_ref_p_RDO( h, cb, 0 );
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,0,4); x264_macroblock_cache_mvd( h, block_idx_x[0], block_idx_y[0], 4, 4, 0, mvd );} while(0);
        }
        else if( h->mb.i_partition == D_16x8 )
        {
            x264_cabac_size_decision_noup(cb,15,1);
            x264_cabac_size_decision_noup(cb,17,1);
            if( h->mb.pic.i_fref[0] > 1 )
            {
                x264_cabac_ref_p_RDO( h, cb, 0 );
                x264_cabac_ref_p_RDO( h, cb, 8 );
            }
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,0,4); x264_macroblock_cache_mvd( h, block_idx_x[0], block_idx_y[0], 4, 2, 0, mvd );} while(0);
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,8,4); x264_macroblock_cache_mvd( h, block_idx_x[8], block_idx_y[8], 4, 2, 0, mvd );} while(0);
        }
        else
        {
            x264_cabac_size_decision_noup(cb,15,1);
            x264_cabac_size_decision_noup(cb,17,0);
            if( h->mb.pic.i_fref[0] > 1 )
            {
                x264_cabac_ref_p_RDO( h, cb, 0 );
                x264_cabac_ref_p_RDO( h, cb, 4 );
            }
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,0,2); x264_macroblock_cache_mvd( h, block_idx_x[0], block_idx_y[0], 2, 4, 0, mvd );} while(0);
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4,2); x264_macroblock_cache_mvd( h, block_idx_x[4], block_idx_y[4], 2, 4, 0, mvd );} while(0);
        }
    }
    else if( i_mb_type == P_8x8 )
    {
        x264_cabac_size_decision_noup(cb,14,0);
        x264_cabac_size_decision_noup(cb,15,0);
        x264_cabac_size_decision_noup(cb,16,1);


        for( int i = 0; i < 4; i++ )
            x264_cabac_subpartition_p_RDO( cb, h->mb.i_sub_partition[i] );


        if( h->mb.pic.i_fref[0] > 1 )
        {
            x264_cabac_ref_p_RDO( h, cb, 0 );
            x264_cabac_ref_p_RDO( h, cb, 4 );
            x264_cabac_ref_p_RDO( h, cb, 8 );
            x264_cabac_ref_p_RDO( h, cb, 12 );
        }

        for( int i = 0; i < 4; i++ )
            x264_cabac_8x8_mvd_RDO( h, cb, i );
    }
    else
        x264_cabac_mb_header_i_RDO( h, cb, i_mb_type, SLICE_TYPE_P, chroma );
}

   void x264_cabac_mb_header_b_RDO( x264_t *h, x264_cabac_t *cb, int i_mb_type, int chroma )
{
    int ctx = 0;
    if( (h->mb.i_neighbour & MB_LEFT) && h->mb.i_mb_type_left[0] != B_SKIP && h->mb.i_mb_type_left[0] != B_DIRECT )
        ctx++;
    if( (h->mb.i_neighbour & MB_TOP) && h->mb.i_mb_type_top != B_SKIP && h->mb.i_mb_type_top != B_DIRECT )
        ctx++;

    if( i_mb_type == B_DIRECT )
    {
        x264_cabac_size_decision_noup(cb,27+ctx,0);
        return;
    }
    x264_cabac_size_decision_noup(cb,27+ctx,1);

    if( i_mb_type == B_8x8 )
    {
        x264_cabac_size_decision_noup(cb,27+3,1);
        x264_cabac_size_decision_noup(cb,27+4,1);
        x264_cabac_size_decision(cb,27+5,1);
        x264_cabac_size_decision(cb,27+5,1);
        x264_cabac_size_decision_noup(cb,27+5,1);


        for( int i = 0; i < 4; i++ )
            x264_cabac_subpartition_b_RDO( cb, h->mb.i_sub_partition[i] );


        if( h->mb.pic.i_fref[0] > 1 )
            for( int i = 0; i < 4; i++ )
                if( x264_mb_partition_listX_table[0][ h->mb.i_sub_partition[i] ] )
                    x264_cabac_ref_b_RDO( h, cb, 0, 4*i );

        if( h->mb.pic.i_fref[1] > 1 )
            for( int i = 0; i < 4; i++ )
                if( x264_mb_partition_listX_table[1][ h->mb.i_sub_partition[i] ] )
                    x264_cabac_ref_b_RDO( h, cb, 1, 4*i );

        for( int i = 0; i < 4; i++ )
            if( x264_mb_partition_listX_table[0][ h->mb.i_sub_partition[i] ] )
                do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4*i,2); x264_macroblock_cache_mvd( h, block_idx_x[4*i], block_idx_y[4*i], 2, 2, 0, mvd );} while(0);

        for( int i = 0; i < 4; i++ )
            if( x264_mb_partition_listX_table[1][ h->mb.i_sub_partition[i] ] )
                do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,1,4*i,2); x264_macroblock_cache_mvd( h, block_idx_x[4*i], block_idx_y[4*i], 2, 2, 1, mvd );} while(0);
    }
    else if( i_mb_type >= B_L0_L0 && i_mb_type <= B_BI_BI )
    {

         const uint8_t i_mb_bits[9*3] =
        {
            0x31, 0x29, 0x4,
            0x35, 0x2d, 0,
            0x43, 0x63, 0,
            0x3d, 0x2f, 0,
            0x39, 0x25, 0x6,
            0x53, 0x73, 0,
            0x4b, 0x6b, 0,
            0x5b, 0x7b, 0,
            0x47, 0x67, 0x21
        };

        const int idx = (i_mb_type - B_L0_L0) * 3 + (h->mb.i_partition - D_16x8);
        int bits = i_mb_bits[idx];

        x264_cabac_size_decision_noup(cb,27+3,bits&1);
        x264_cabac_size_decision(cb,27+5-(bits&1),(bits>>1)&1); bits >>= 2;
        if( bits != 1 )
        {
            x264_cabac_size_decision(cb,27+5,bits&1); bits >>= 1;
            x264_cabac_size_decision(cb,27+5,bits&1); bits >>= 1;
            x264_cabac_size_decision(cb,27+5,bits&1); bits >>= 1;
            if( bits != 1 )
                x264_cabac_size_decision_noup(cb,27+5,bits&1);
        }

        const uint8_t (*b_list)[2] = x264_mb_type_list_table[i_mb_type];
        if( h->mb.pic.i_fref[0] > 1 )
        {
            if( b_list[0][0] )
                x264_cabac_ref_b_RDO( h, cb, 0, 0 );
            if( b_list[0][1] && h->mb.i_partition != D_16x16 )
                x264_cabac_ref_b_RDO( h, cb, 0, 8 >> (h->mb.i_partition == D_8x16) );
        }
        if( h->mb.pic.i_fref[1] > 1 )
        {
            if( b_list[1][0] )
                x264_cabac_ref_b_RDO( h, cb, 1, 0 );
            if( b_list[1][1] && h->mb.i_partition != D_16x16 )
                x264_cabac_ref_b_RDO( h, cb, 1, 8 >> (h->mb.i_partition == D_8x16) );
        }
        for( int i_list = 0; i_list < 2; i_list++ )
        {
            if( h->mb.i_partition == D_16x16 )
            {
                if( b_list[i_list][0] ) do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,i_list,0,4); x264_macroblock_cache_mvd( h, block_idx_x[0], block_idx_y[0], 4, 4, i_list, mvd );} while(0);
            }
            else if( h->mb.i_partition == D_16x8 )
            {
                if( b_list[i_list][0] ) do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,i_list,0,4); x264_macroblock_cache_mvd( h, block_idx_x[0], block_idx_y[0], 4, 2, i_list, mvd );} while(0);
                if( b_list[i_list][1] ) do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,i_list,8,4); x264_macroblock_cache_mvd( h, block_idx_x[8], block_idx_y[8], 4, 2, i_list, mvd );} while(0);
            }
            else
            {
                if( b_list[i_list][0] ) do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,i_list,0,2); x264_macroblock_cache_mvd( h, block_idx_x[0], block_idx_y[0], 2, 4, i_list, mvd );} while(0);
                if( b_list[i_list][1] ) do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,i_list,4,2); x264_macroblock_cache_mvd( h, block_idx_x[4], block_idx_y[4], 2, 4, i_list, mvd );} while(0);
            }
        }
    }
    else
        x264_cabac_mb_header_i_RDO( h, cb, i_mb_type, SLICE_TYPE_B, chroma );
}

 int   x264_cabac_cbf_ctxidxinc_RDO( x264_t *h, int i_cat, int i_idx, int b_intra, int b_dc )
{
     const uint16_t base_ctx[14] = {85,89,93,97,101,1012,460,464,468,1016,472,476,480,1020};

    if( b_dc )
    {
        i_idx -= 48;
        if( i_cat == DCT_CHROMA_DC )
        {
            int i_nza = h->mb.cache.i_cbp_left != -1 ? (h->mb.cache.i_cbp_left >> (8 + i_idx)) & 1 : b_intra;
            int i_nzb = h->mb.cache.i_cbp_top != -1 ? (h->mb.cache.i_cbp_top >> (8 + i_idx)) & 1 : b_intra;
            return base_ctx[i_cat] + 2*i_nzb + i_nza;
        }
        else
        {
            int i_nza = (h->mb.cache.i_cbp_left >> (8 + i_idx)) & 1;
            int i_nzb = (h->mb.cache.i_cbp_top >> (8 + i_idx)) & 1;
            return base_ctx[i_cat] + 2*i_nzb + i_nza;
        }
    }
    else
    {
        int i_nza = h->mb.cache.non_zero_count[x264_scan8[i_idx] - 1];
        int i_nzb = h->mb.cache.non_zero_count[x264_scan8[i_idx] - 8];
        if( __builtin_constant_p(b_intra) && !b_intra )
            return base_ctx[i_cat] + ((2*i_nzb + i_nza)&0x7f);
        else
        {
            i_nza &= 0x7f + (b_intra << 7);
            i_nzb &= 0x7f + (b_intra << 7);
            return base_ctx[i_cat] + 2*!!i_nzb + !!i_nza;
        }
    }
}
const uint8_t x264_significant_coeff_flag_offset_8x8[2][64] =
{{
    0, 1, 2, 3, 4, 5, 5, 4, 4, 3, 3, 4, 4, 4, 5, 5,
    4, 4, 4, 4, 3, 3, 6, 7, 7, 7, 8, 9,10, 9, 8, 7,
    7, 6,11,12,13,11, 6, 7, 8, 9,14,10, 9, 8, 6,11,
   12,13,11, 6, 9,14,10, 9,11,12,13,11,14,10,12
},{
    0, 1, 1, 2, 2, 3, 3, 4, 5, 6, 7, 7, 7, 8, 4, 5,
    6, 9,10,10, 8,11,12,11, 9, 9,10,10, 8,11,12,11,
    9, 9,10,10, 8,11,12,11, 9, 9,10,10, 8,13,13, 9,
    9,10,10, 8,13,13, 9, 9,10,10,14,14,14,14,14
}};
const uint8_t x264_last_coeff_flag_offset_8x8[63] =
{
    0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4,
    5, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8
};
const uint8_t x264_coeff_flag_offset_chroma_422_dc[7] = { 0, 0, 1, 1, 2, 2, 2 };
const uint16_t x264_significant_coeff_flag_offset[2][16] =
{
    { 105+0, 105+15, 105+29, 105+44, 105+47, 402, 484+0, 484+15, 484+29, 660, 528+0, 528+15, 528+29, 718, 0, 0 },
    { 277+0, 277+15, 277+29, 277+44, 277+47, 436, 776+0, 776+15, 776+29, 675, 820+0, 820+15, 820+29, 733, 0, 0 }
};
const uint16_t x264_last_coeff_flag_offset[2][16] =
{
    { 166+0, 166+15, 166+29, 166+44, 166+47, 417, 572+0, 572+15, 572+29, 690, 616+0, 616+15, 616+29, 748, 0, 0 },
    { 338+0, 338+15, 338+29, 338+44, 338+47, 451, 864+0, 864+15, 864+29, 699, 908+0, 908+15, 908+29, 757, 0, 0 }
};
const uint16_t x264_coeff_abs_level_m1_offset[16] =
{
    227+0, 227+10, 227+20, 227+30, 227+39, 426, 952+0, 952+10, 952+20, 708, 982+0, 982+10, 982+20, 766
};
const uint8_t x264_count_cat_m1[14] = {15, 14, 15, 3, 14, 63, 15, 14, 15, 63, 15, 14, 15, 63};





 const uint8_t coeff_abs_level1_ctx_RDO[8] = { 1, 2, 3, 4, 0, 0, 0, 0 };

 const uint8_t coeff_abs_levelgt1_ctx_RDO[8] = { 5, 5, 5, 5, 6, 7, 8, 9 };


 const uint8_t coeff_abs_levelgt1_ctx_chroma_dc_RDO[8] = { 5, 5, 5, 5, 6, 7, 8, 8 };

 const uint8_t coeff_abs_level_transition_RDO[2][8] = {

    { 1, 2, 3, 3, 4, 5, 6, 7 },

    { 4, 4, 4, 4, 5, 6, 7, 7 }
};
 void   x264_cabac_block_residual_internal_RDO( x264_t *h, x264_cabac_t *cb, int ctx_block_cat, dctcoef *l, int b_8x8, int chroma422dc )
{
    const uint8_t *sig_offset = x264_significant_coeff_flag_offset_8x8[0];
    int ctx_sig = x264_significant_coeff_flag_offset[0][ctx_block_cat];
    int ctx_last = x264_last_coeff_flag_offset[0][ctx_block_cat];
    int ctx_level = x264_coeff_abs_level_m1_offset[ctx_block_cat];
    int last = h->quantf.coeff_last[ctx_block_cat]( l );
    int coeff_abs = abs(l[last]);
    int ctx = coeff_abs_level1_ctx_RDO[0] + ctx_level;
    int node_ctx;
    const uint8_t *levelgt1_ctx = chroma422dc ? coeff_abs_levelgt1_ctx_chroma_dc_RDO : coeff_abs_levelgt1_ctx_RDO;

    if( last != (b_8x8 ? 63 : chroma422dc ? 7 : x264_count_cat_m1[ctx_block_cat]) )
    {
        x264_cabac_size_decision(cb,ctx_sig + (b_8x8 ? sig_offset[last] : chroma422dc ? x264_coeff_flag_offset_chroma_422_dc[last] : last),1)
                                                                                                         ;
        x264_cabac_size_decision(cb,ctx_last + (b_8x8 ? x264_last_coeff_flag_offset_8x8[last] : chroma422dc ? x264_coeff_flag_offset_chroma_422_dc[last] : last),1)
                                                                                                         ;
    }

    if( coeff_abs > 1 )
    {
        x264_cabac_size_decision(cb,ctx,1);
        ctx = levelgt1_ctx[0] + ctx_level;
        if( coeff_abs < 15 )
        {
            cb->f8_bits_encoded += x264_cabac_size_unary[coeff_abs-1][cb->state[ctx]];
            cb->state[ctx] = x264_cabac_transition_unary[coeff_abs-1][cb->state[ctx]];
        }
        else
        {
            cb->f8_bits_encoded += x264_cabac_size_unary[14][cb->state[ctx]];
            cb->state[ctx] = x264_cabac_transition_unary[14][cb->state[ctx]];
            ((cb)->f8_bits_encoded += (bs_size_ue_big(coeff_abs - 15 +(1<<0)-1)-0)<<8);
        }
        node_ctx = coeff_abs_level_transition_RDO[1][0];
    }
    else
    {
        x264_cabac_size_decision(cb,ctx,0);
        node_ctx = coeff_abs_level_transition_RDO[0][0];
        ((cb)->f8_bits_encoded += 256);
    }

    for( int i = last-1 ; i >= 0; i-- )
    {
        if( l[i] )
        {
            coeff_abs = abs(l[i]);
            x264_cabac_size_decision(cb,ctx_sig + (b_8x8 ? sig_offset[i] : chroma422dc ? x264_coeff_flag_offset_chroma_422_dc[i] : i),1)
                                                                                                       ;
            x264_cabac_size_decision(cb,ctx_last + (b_8x8 ? x264_last_coeff_flag_offset_8x8[i] : chroma422dc ? x264_coeff_flag_offset_chroma_422_dc[i] : i),0)
                                                                                                       ;
            ctx = coeff_abs_level1_ctx_RDO[node_ctx] + ctx_level;

            if( coeff_abs > 1 )
            {
                x264_cabac_size_decision(cb,ctx,1);
                ctx = levelgt1_ctx[node_ctx] + ctx_level;
                if( coeff_abs < 15 )
                {
                    cb->f8_bits_encoded += x264_cabac_size_unary[coeff_abs-1][cb->state[ctx]];
                    cb->state[ctx] = x264_cabac_transition_unary[coeff_abs-1][cb->state[ctx]];
                }
                else
                {
                    cb->f8_bits_encoded += x264_cabac_size_unary[14][cb->state[ctx]];
                    cb->state[ctx] = x264_cabac_transition_unary[14][cb->state[ctx]];
                    ((cb)->f8_bits_encoded += (bs_size_ue_big(coeff_abs - 15 +(1<<0)-1)-0)<<8);
                }
                node_ctx = coeff_abs_level_transition_RDO[1][node_ctx];
            }
            else
            {
                x264_cabac_size_decision(cb,ctx,0);
                node_ctx = coeff_abs_level_transition_RDO[0][node_ctx];
                ((cb)->f8_bits_encoded += 256);
            }
        }
        else
            x264_cabac_size_decision(cb,ctx_sig + (b_8x8 ? sig_offset[i] : chroma422dc ? x264_coeff_flag_offset_chroma_422_dc[i] : i),0)
                                                                                                       ;
    }
}

void x264_cabac_block_residual_8x8_rd_c( x264_t *h, x264_cabac_t *cb, int ctx_block_cat, dctcoef *l )
{
    x264_cabac_block_residual_internal_RDO( h, cb, ctx_block_cat, l, 1, 0 );
}
void x264_cabac_block_residual_rd_c( x264_t *h, x264_cabac_t *cb, int ctx_block_cat, dctcoef *l )
{
    x264_cabac_block_residual_internal_RDO( h, cb, ctx_block_cat, l, 0, 0 );
}

   void x264_cabac_block_residual_8x8( x264_t *h, x264_cabac_t *cb, int ctx_block_cat, dctcoef *l )
{



    x264_cabac_block_residual_8x8_rd_c( h, cb, ctx_block_cat, l );

}
   void x264_cabac_block_residual_RDO( x264_t *h, x264_cabac_t *cb, int ctx_block_cat, dctcoef *l )
{



    x264_cabac_block_residual_rd_c( h, cb, ctx_block_cat, l );

}

 void x264_cabac_block_residual_422_dc_RDO( x264_t *h, x264_cabac_t *cb, int ctx_block_cat, dctcoef *l )
{
    x264_cabac_block_residual_internal_RDO( h, cb, DCT_CHROMA_DC, l, 0, 1 );
}
   void x264_macroblock_write_cabac_internal_RDO( x264_t *h, x264_cabac_t *cb, int plane_count, int chroma )
{
    const int i_mb_type = h->mb.i_type;
    if( h->sh.i_type == SLICE_TYPE_P )
        x264_cabac_mb_header_p_RDO( h, cb, i_mb_type, chroma );
    else if( h->sh.i_type == SLICE_TYPE_B )
        x264_cabac_mb_header_b_RDO( h, cb, i_mb_type, chroma );
    else
        x264_cabac_mb_header_i_RDO( h, cb, i_mb_type, SLICE_TYPE_I, chroma );
    if( i_mb_type != I_16x16 )
    {
        x264_cabac_cbp_luma_RDO( h, cb );
        if( chroma )
            x264_cabac_cbp_chroma_RDO( h, cb );
    }

    if( x264_mb_transform_8x8_allowed( h ) && h->mb.i_cbp_luma )
    {
        x264_cabac_transform_size_RDO( h, cb );
    }

    if( h->mb.i_cbp_luma || (chroma && h->mb.i_cbp_chroma) || i_mb_type == I_16x16 )
    {
        const int b_intra = ( (i_mb_type) == I_4x4 || (i_mb_type) == I_8x8 || (i_mb_type) == I_16x16 || (i_mb_type) == I_PCM );
        x264_cabac_qp_delta_RDO( h, cb );


        if( i_mb_type == I_16x16 )
        {

            for( int p = 0; p < plane_count; p++ )
            {
                do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, ctx_cat_plane[DCT_LUMA_DC][p], 48 +p, 1, 1 ); if( h->mb.cache.non_zero_count[x264_scan8[48 +p]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, ctx_cat_plane[DCT_LUMA_DC][p], h->dct.luma16x16_dc[p] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);


                if( h->mb.i_cbp_luma )
                    for( int i = p*16; i < p*16+16; i++ )
                        do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, ctx_cat_plane[DCT_LUMA_AC][p], i, 1, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[i]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, ctx_cat_plane[DCT_LUMA_AC][p], h->dct.luma4x4[i]+1 ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
            }
        }
        else if( h->mb.b_transform_8x8 )
        {
            if( plane_count == 3 )
            {
                uint8_t nnzbak[3][8] ;
                if( (h->mb.i_neighbour & MB_LEFT) && !h->mb.mb_transform_size[h->mb.i_mb_left_xy[0]] ){ nnzbak[0][0] = h->mb.cache.non_zero_count[x264_scan8[16*0+ 0] - 1]; h->mb.cache.non_zero_count[x264_scan8[16*0+ 0] - 1] = 0x80; nnzbak[0][1] = h->mb.cache.non_zero_count[x264_scan8[16*0+ 2] - 1]; h->mb.cache.non_zero_count[x264_scan8[16*0+ 2] - 1] = 0x80; nnzbak[1][0] = h->mb.cache.non_zero_count[x264_scan8[16*1+ 0] - 1]; h->mb.cache.non_zero_count[x264_scan8[16*1+ 0] - 1] = 0x80; nnzbak[1][1] = h->mb.cache.non_zero_count[x264_scan8[16*1+ 2] - 1]; h->mb.cache.non_zero_count[x264_scan8[16*1+ 2] - 1] = 0x80; nnzbak[2][0] = h->mb.cache.non_zero_count[x264_scan8[16*2+ 0] - 1]; h->mb.cache.non_zero_count[x264_scan8[16*2+ 0] - 1] = 0x80; nnzbak[2][1] = h->mb.cache.non_zero_count[x264_scan8[16*2+ 2] - 1]; h->mb.cache.non_zero_count[x264_scan8[16*2+ 2] - 1] = 0x80;}if( (h->mb.i_neighbour & MB_LEFT) && !h->mb.mb_transform_size[h->mb.i_mb_left_xy[1]] ){ nnzbak[0][2] = h->mb.cache.non_zero_count[x264_scan8[16*0+ 8] - 1]; h->mb.cache.non_zero_count[x264_scan8[16*0+ 8] - 1] = 0x80; nnzbak[0][3] = h->mb.cache.non_zero_count[x264_scan8[16*0+10] - 1]; h->mb.cache.non_zero_count[x264_scan8[16*0+10] - 1] = 0x80; nnzbak[1][2] = h->mb.cache.non_zero_count[x264_scan8[16*1+ 8] - 1]; h->mb.cache.non_zero_count[x264_scan8[16*1+ 8] - 1] = 0x80; nnzbak[1][3] = h->mb.cache.non_zero_count[x264_scan8[16*1+10] - 1]; h->mb.cache.non_zero_count[x264_scan8[16*1+10] - 1] = 0x80; nnzbak[2][2] = h->mb.cache.non_zero_count[x264_scan8[16*2+ 8] - 1]; h->mb.cache.non_zero_count[x264_scan8[16*2+ 8] - 1] = 0x80; nnzbak[2][3] = h->mb.cache.non_zero_count[x264_scan8[16*2+10] - 1]; h->mb.cache.non_zero_count[x264_scan8[16*2+10] - 1] = 0x80;}if( (h->mb.i_neighbour & MB_TOP) && !h->mb.mb_transform_size[h->mb.i_mb_top_xy] ){ (((x264_union32_t*)(&nnzbak[0][4]))->i) = (((x264_union32_t*)(&h->mb.cache.non_zero_count[x264_scan8[16*0] - 8]))->i); (((x264_union32_t*)(&h->mb.cache.non_zero_count[x264_scan8[16*0] - 8]))->i) = 0x80808080U; (((x264_union32_t*)(&nnzbak[1][4]))->i) = (((x264_union32_t*)(&h->mb.cache.non_zero_count[x264_scan8[16*1] - 8]))->i); (((x264_union32_t*)(&h->mb.cache.non_zero_count[x264_scan8[16*1] - 8]))->i) = 0x80808080U; (((x264_union32_t*)(&nnzbak[2][4]))->i) = (((x264_union32_t*)(&h->mb.cache.non_zero_count[x264_scan8[16*2] - 8]))->i); (((x264_union32_t*)(&h->mb.cache.non_zero_count[x264_scan8[16*2] - 8]))->i) = 0x80808080U;}

                for( int p = 0; p < 3; p++ )
                    for( int i = 0, msk = h->mb.i_cbp_luma, skip; msk && (skip = x264_ctz_4bit(msk), i += skip, msk >>= skip+1, 1); i++ )
                        do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, ctx_cat_plane[DCT_LUMA_8x8][p], i*4+p*16, b_intra, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[i*4+p*16]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_8x8( h, cb, ctx_cat_plane[DCT_LUMA_8x8][p], h->dct.luma8x8[i+p*4] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);

                if( (h->mb.i_neighbour & MB_LEFT) && !h->mb.mb_transform_size[h->mb.i_mb_left_xy[0]] ){ h->mb.cache.non_zero_count[x264_scan8[16*0+ 0] - 1] = nnzbak[0][0]; h->mb.cache.non_zero_count[x264_scan8[16*0+ 2] - 1] = nnzbak[0][1]; h->mb.cache.non_zero_count[x264_scan8[16*1+ 0] - 1] = nnzbak[1][0]; h->mb.cache.non_zero_count[x264_scan8[16*1+ 2] - 1] = nnzbak[1][1]; h->mb.cache.non_zero_count[x264_scan8[16*2+ 0] - 1] = nnzbak[2][0]; h->mb.cache.non_zero_count[x264_scan8[16*2+ 2] - 1] = nnzbak[2][1];}if( (h->mb.i_neighbour & MB_LEFT) && !h->mb.mb_transform_size[h->mb.i_mb_left_xy[1]] ){ h->mb.cache.non_zero_count[x264_scan8[16*0+ 8] - 1] = nnzbak[0][2]; h->mb.cache.non_zero_count[x264_scan8[16*0+10] - 1] = nnzbak[0][3]; h->mb.cache.non_zero_count[x264_scan8[16*1+ 8] - 1] = nnzbak[1][2]; h->mb.cache.non_zero_count[x264_scan8[16*1+10] - 1] = nnzbak[1][3]; h->mb.cache.non_zero_count[x264_scan8[16*2+ 8] - 1] = nnzbak[2][2]; h->mb.cache.non_zero_count[x264_scan8[16*2+10] - 1] = nnzbak[2][3];}if( (h->mb.i_neighbour & MB_TOP) && !h->mb.mb_transform_size[h->mb.i_mb_top_xy] ){ (((x264_union32_t*)(&h->mb.cache.non_zero_count[x264_scan8[16*0] - 8]))->i) = (((x264_union32_t*)(&nnzbak[0][4]))->i); (((x264_union32_t*)(&h->mb.cache.non_zero_count[x264_scan8[16*1] - 8]))->i) = (((x264_union32_t*)(&nnzbak[1][4]))->i); (((x264_union32_t*)(&h->mb.cache.non_zero_count[x264_scan8[16*2] - 8]))->i) = (((x264_union32_t*)(&nnzbak[2][4]))->i);}
            }
            else
            {
                for( int i = 0, msk = h->mb.i_cbp_luma, skip; msk && (skip = x264_ctz_4bit(msk), i += skip, msk >>= skip+1, 1); i++ )
                    x264_cabac_block_residual_8x8( h, cb, DCT_LUMA_8x8, h->dct.luma8x8[i] );
            }
        }
        else
        {
            for( int p = 0; p < plane_count; p++ )
                for( int i8x8 = 0, msk = h->mb.i_cbp_luma, skip; msk && (skip = x264_ctz_4bit(msk), i8x8 += skip, msk >>= skip+1, 1); i8x8++ )
                    for( int i = 0; i < 4; i++ )
                        do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, ctx_cat_plane[DCT_LUMA_4x4][p], i+i8x8*4+p*16, b_intra, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[i+i8x8*4+p*16]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, ctx_cat_plane[DCT_LUMA_4x4][p], h->dct.luma4x4[i+i8x8*4+p*16] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
        }

        if( chroma && h->mb.i_cbp_chroma )
        {
            if( h->sps->i_chroma_format_idc == CHROMA_422 )
            {
                do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_DC, 49 +(0), b_intra, 1 ); if( h->mb.cache.non_zero_count[x264_scan8[49 +(0)]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_422_dc_RDO( h, cb, DCT_CHROMA_DC, h->dct.chroma_dc[0] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
                do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_DC, 49 +(1), b_intra, 1 ); if( h->mb.cache.non_zero_count[x264_scan8[49 +(1)]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_422_dc_RDO( h, cb, DCT_CHROMA_DC, h->dct.chroma_dc[1] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
            }
            else
            {
                do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_DC, 49 +0, b_intra, 1 ); if( h->mb.cache.non_zero_count[x264_scan8[49 +0]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, DCT_CHROMA_DC, h->dct.chroma_dc[0] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
                do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_DC, 49 +1, b_intra, 1 ); if( h->mb.cache.non_zero_count[x264_scan8[49 +1]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, DCT_CHROMA_DC, h->dct.chroma_dc[1] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
            }

            if( h->mb.i_cbp_chroma == 2 )
            {
                int step = 8 << h->mb.chroma_v_shift;
                for( int i = 16; i < 3*16; i += step )
                    for( int j = i; j < i+4; j++ )
                        do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_AC, j, b_intra, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[j]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, DCT_CHROMA_AC, h->dct.luma4x4[j]+1 ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
            }
        }
    }




}

void  x264_macroblock_size_cabac( x264_t *h, x264_cabac_t *cb )
{
    if( (h->sps->i_chroma_format_idc == CHROMA_444) )
        x264_macroblock_write_cabac_internal_RDO( h, cb, 3, 0 );
    else
        x264_macroblock_write_cabac_internal_RDO( h, cb, 1, 1 );
}
 void x264_partition_size_cabac( x264_t *h, x264_cabac_t *cb, int i8, int i_pixel )
{
    const int i_mb_type = h->mb.i_type;
    int b_8x16 = h->mb.i_partition == D_8x16;
    int plane_count = (h->sps->i_chroma_format_idc == CHROMA_444) ? 3 : 1;

    if( i_mb_type == P_8x8 )
    {
        x264_cabac_8x8_mvd_RDO( h, cb, i8 );
        x264_cabac_subpartition_p_RDO( cb, h->mb.i_sub_partition[i8] );
    }
    else if( i_mb_type == P_L0 )
        do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4*i8,4>>b_8x16); x264_macroblock_cache_mvd( h, block_idx_x[4*i8], block_idx_y[4*i8], 4>>b_8x16, 2<<b_8x16, 0, mvd );} while(0);
    else if( i_mb_type > B_DIRECT && i_mb_type < B_8x8 )
    {
        if( x264_mb_type_list_table[ i_mb_type ][0][!!i8] ) do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4*i8,4>>b_8x16); x264_macroblock_cache_mvd( h, block_idx_x[4*i8], block_idx_y[4*i8], 4>>b_8x16, 2<<b_8x16, 0, mvd );} while(0);
        if( x264_mb_type_list_table[ i_mb_type ][1][!!i8] ) do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,1,4*i8,4>>b_8x16); x264_macroblock_cache_mvd( h, block_idx_x[4*i8], block_idx_y[4*i8], 4>>b_8x16, 2<<b_8x16, 1, mvd );} while(0);
    }
    else
    {
        if( x264_mb_partition_listX_table[0][ h->mb.i_sub_partition[i8] ] )
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,4*i8,2); x264_macroblock_cache_mvd( h, block_idx_x[4*i8], block_idx_y[4*i8], 2, 2, 0, mvd );} while(0);
        if( x264_mb_partition_listX_table[1][ h->mb.i_sub_partition[i8] ] )
            do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,1,4*i8,2); x264_macroblock_cache_mvd( h, block_idx_x[4*i8], block_idx_y[4*i8], 2, 2, 1, mvd );} while(0);
    }

    for( int j = (i_pixel < PIXEL_8x8); j >= 0; j-- )
    {
        if( h->mb.i_cbp_luma & (1 << i8) )
        {
            if( h->mb.b_transform_8x8 )
            {
                if( (h->sps->i_chroma_format_idc == CHROMA_444) )
                    for( int p = 0; p < 3; p++ )
                        do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, ctx_cat_plane[DCT_LUMA_8x8][p], i8*4+p*16, 0, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[i8*4+p*16]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_8x8( h, cb, ctx_cat_plane[DCT_LUMA_8x8][p], h->dct.luma8x8[i8+p*4] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
                else
                    x264_cabac_block_residual_8x8( h, cb, DCT_LUMA_8x8, h->dct.luma8x8[i8] );
            }
            else
                for( int p = 0; p < plane_count; p++ )
                    for( int i4 = 0; i4 < 4; i4++ )
                        do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, ctx_cat_plane[DCT_LUMA_4x4][p], i4+i8*4+p*16, 0, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[i4+i8*4+p*16]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, ctx_cat_plane[DCT_LUMA_4x4][p], h->dct.luma4x4[i4+i8*4+p*16] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
        }

        if( h->mb.i_cbp_chroma )
        {
            if( h->sps->i_chroma_format_idc == CHROMA_422 )
            {
                int offset = (5*i8) & 0x09;
                do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_AC, 16+offset, 0, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[16+offset]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, DCT_CHROMA_AC, h->dct.luma4x4[16+offset]+1 ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
                do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_AC, 18+offset, 0, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[18+offset]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, DCT_CHROMA_AC, h->dct.luma4x4[18+offset]+1 ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
                do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_AC, 32+offset, 0, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[32+offset]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, DCT_CHROMA_AC, h->dct.luma4x4[32+offset]+1 ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
                do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_AC, 34+offset, 0, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[34+offset]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, DCT_CHROMA_AC, h->dct.luma4x4[34+offset]+1 ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
            }
            else
            {
                do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_AC, 16+i8, 0, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[16+i8]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, DCT_CHROMA_AC, h->dct.luma4x4[16+i8]+1 ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
                do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_AC, 32+i8, 0, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[32+i8]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, DCT_CHROMA_AC, h->dct.luma4x4[32+i8]+1 ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
            }
        }

        i8 += x264_pixel_size[i_pixel].h >> 3;
    }
}

 void x264_subpartition_size_cabac( x264_t *h, x264_cabac_t *cb, int i4, int i_pixel )
{
    int b_8x4 = i_pixel == PIXEL_8x4;
    int plane_count = (h->sps->i_chroma_format_idc == CHROMA_444) ? 3 : 1;
    if( i_pixel == PIXEL_4x4 )
        do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,i4,1); x264_macroblock_cache_mvd( h, block_idx_x[i4], block_idx_y[i4], 1, 1, 0, mvd );} while(0);
    else
        do{ uint16_t mvd = x264_cabac_mvd_RDO(h,cb,0,i4,1+b_8x4); x264_macroblock_cache_mvd( h, block_idx_x[i4], block_idx_y[i4], 1+b_8x4, 2-b_8x4, 0, mvd );} while(0);
    for( int p = 0; p < plane_count; p++ )
    {
        do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, ctx_cat_plane[DCT_LUMA_4x4][p], p*16+i4, 0, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[p*16+i4]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, ctx_cat_plane[DCT_LUMA_4x4][p], h->dct.luma4x4[p*16+i4] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
        if( i_pixel != PIXEL_4x4 )
            do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, ctx_cat_plane[DCT_LUMA_4x4][p], p*16+i4+2-b_8x4, 0, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[p*16+i4+2-b_8x4]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, ctx_cat_plane[DCT_LUMA_4x4][p], h->dct.luma4x4[p*16+i4+2-b_8x4] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
    }
}

 void x264_partition_i8x8_size_cabac( x264_t *h, x264_cabac_t *cb, int i8, int i_mode )
{
    const int i_pred = x264_mb_predict_intra4x4_mode( h, 4*i8 );
    i_mode = x264_mb_pred_mode4x4_fix[(i_mode)+1];
    x264_cabac_intra4x4_pred_mode_RDO( cb, i_pred, i_mode );
    x264_cabac_cbp_luma_RDO( h, cb );
    if( h->mb.i_cbp_luma & (1 << i8) )
    {
        if( (h->sps->i_chroma_format_idc == CHROMA_444) )
            for( int p = 0; p < 3; p++ )
                do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, ctx_cat_plane[DCT_LUMA_8x8][p], i8*4+p*16, 1, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[i8*4+p*16]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_8x8( h, cb, ctx_cat_plane[DCT_LUMA_8x8][p], h->dct.luma8x8[i8+p*4] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
        else
            x264_cabac_block_residual_8x8( h, cb, DCT_LUMA_8x8, h->dct.luma8x8[i8] );
    }
}

 void x264_partition_i4x4_size_cabac( x264_t *h, x264_cabac_t *cb, int i4, int i_mode )
{
    const int i_pred = x264_mb_predict_intra4x4_mode( h, i4 );
    int plane_count = (h->sps->i_chroma_format_idc == CHROMA_444) ? 3 : 1;
    i_mode = x264_mb_pred_mode4x4_fix[(i_mode)+1];
    x264_cabac_intra4x4_pred_mode_RDO( cb, i_pred, i_mode );
    for( int p = 0; p < plane_count; p++ )
        do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, ctx_cat_plane[DCT_LUMA_4x4][p], i4+p*16, 1, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[i4+p*16]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, ctx_cat_plane[DCT_LUMA_4x4][p], h->dct.luma4x4[i4+p*16] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
}

 void x264_chroma_size_cabac( x264_t *h, x264_cabac_t *cb )
{
    x264_cabac_intra_chroma_pred_mode_RDO( h, cb );
    x264_cabac_cbp_chroma_RDO( h, cb );
    if( h->mb.i_cbp_chroma )
    {
        if( h->sps->i_chroma_format_idc == CHROMA_422 )
        {
            do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_DC, 49 +(0), 1, 1 ); if( h->mb.cache.non_zero_count[x264_scan8[49 +(0)]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_422_dc_RDO( h, cb, DCT_CHROMA_DC, h->dct.chroma_dc[0] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
            do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_DC, 49 +(1), 1, 1 ); if( h->mb.cache.non_zero_count[x264_scan8[49 +(1)]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_422_dc_RDO( h, cb, DCT_CHROMA_DC, h->dct.chroma_dc[1] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
        }
        else
        {
            do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_DC, 49 +0, 1, 1 ); if( h->mb.cache.non_zero_count[x264_scan8[49 +0]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, DCT_CHROMA_DC, h->dct.chroma_dc[0] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
            do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_DC, 49 +1, 1, 1 ); if( h->mb.cache.non_zero_count[x264_scan8[49 +1]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, DCT_CHROMA_DC, h->dct.chroma_dc[1] ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
        }

        if( h->mb.i_cbp_chroma == 2 )
        {
            int step = 8 << h->mb.chroma_v_shift;
            for( int i = 16; i < 3*16; i += step )
                for( int j = i; j < i+4; j++ )
                    do{ int ctxidxinc = x264_cabac_cbf_ctxidxinc_RDO( h, DCT_CHROMA_AC, j, 1, 0 ); if( h->mb.cache.non_zero_count[x264_scan8[j]] ) { x264_cabac_size_decision(cb,ctxidxinc,1); x264_cabac_block_residual_RDO( h, cb, DCT_CHROMA_AC, h->dct.luma4x4[j]+1 ); } else x264_cabac_size_decision(cb,ctxidxinc,0);} while(0);
        }
    }
}












#define COPY_CABAC h->mc.memcpy_aligned( &cabac_tmp.f8_bits_encoded, &h->cabac.f8_bits_encoded, \
        sizeof(x264_cabac_t) - offsetof(x264_cabac_t,f8_bits_encoded) - (CHROMA444 ? 0 : (1024+12)-460) )
#define COPY_CABAC_PART( pos, size )\
        memcpy( &cb->state[pos], &h->cabac.state[pos], size )

   uint64_t cached_hadamard( x264_t *h, int size, int x, int y )
{
     const uint8_t hadamard_shift_x[4] = {4, 4, 3, 3};
     const uint8_t hadamard_shift_y[4] = {4-0, 3-0, 4-1, 3-1};
     const uint8_t hadamard_offset[4] = {0, 1, 3, 5};
    int cache_index = (x >> hadamard_shift_x[size]) + (y >> hadamard_shift_y[size])
                    + hadamard_offset[size];
    uint64_t res = h->mb.pic.fenc_hadamard_cache[cache_index];
    if( res )
        return res - 1;
    else
    {
        pixel *fenc = h->mb.pic.p_fenc[0] + x + y*16;
        res = h->pixf.hadamard_ac[size]( fenc, 16 );
        h->mb.pic.fenc_hadamard_cache[cache_index] = res + 1;
        return res;
    }
}

   int cached_satd( x264_t *h, int size, int x, int y )
{
     const uint8_t satd_shift_x[3] = {3, 2, 2};
     const uint8_t satd_shift_y[3] = {2-1, 3-2, 2-2};
     const uint8_t satd_offset[3] = {0, 8, 16};
     pixel zero[16]  = {0};
    int cache_index = (x >> satd_shift_x[size - PIXEL_8x4]) + (y >> satd_shift_y[size - PIXEL_8x4])
                    + satd_offset[size - PIXEL_8x4];
    int res = h->mb.pic.fenc_satd_cache[cache_index];
    if( res )
        return res - 1;
    else
    {
        pixel *fenc = h->mb.pic.p_fenc[0] + x + y*16;
        int dc = h->pixf.sad[size]( fenc, 16, zero, 0 ) >> 1;
        res = h->pixf.satd[size]( fenc, 16, zero, 0 ) - dc;
        h->mb.pic.fenc_satd_cache[cache_index] = res + 1;
        return res;
    }
}
  int ssd_plane( x264_t *h, int size, int p, int x, int y )
{
     pixel zero[16]  = {0};
    int satd = 0;
    pixel *fdec = h->mb.pic.p_fdec[p] + x + y*32;
    pixel *fenc = h->mb.pic.p_fenc[p] + x + y*16;
    if( p == 0 && h->mb.i_psy_rd )
    {

        if( size <= PIXEL_8x8 )
        {
            uint64_t fdec_acs = h->pixf.hadamard_ac[size]( fdec, 32 );
            uint64_t fenc_acs = cached_hadamard( h, size, x, y );
            satd = abs((int32_t)fdec_acs - (int32_t)fenc_acs)
                 + abs((int32_t)(fdec_acs>>32) - (int32_t)(fenc_acs>>32));
            satd >>= 1;
        }
        else
        {
            int dc = h->pixf.sad[size]( fdec, 32, zero, 0 ) >> 1;
            satd = abs(h->pixf.satd[size]( fdec, 32, zero, 0 ) - dc - cached_satd( h, size, x, y ));
        }
        satd = (satd * h->mb.i_psy_rd * h->mb.i_psy_rd_lambda + 128) >> 8;
    }
    return h->pixf.ssd[size](fenc, 16, fdec, 32) + satd;
}

  int ssd_mb( x264_t *h )
{
    int chroma_size = h->luma2chroma_pixel[PIXEL_16x16];
    int chroma_ssd = ssd_plane(h, chroma_size, 1, 0, 0) + ssd_plane(h, chroma_size, 2, 0, 0);
    chroma_ssd = ((uint64_t)chroma_ssd * h->mb.i_chroma_lambda2_offset + 128) >> 8;
    return ssd_plane(h, PIXEL_16x16, 0, 0, 0) + chroma_ssd;
}

 int x264_rd_cost_mb( x264_t *h, int i_lambda2 )
{
    int b_transform_bak = h->mb.b_transform_8x8;
    int i_ssd;
    int i_bits;
    int type_bak = h->mb.i_type;

    x264_macroblock_encode( h );


    i_ssd = ssd_mb( h );

    if( ( (h->mb.i_type) == P_SKIP || (h->mb.i_type) == B_SKIP ) )
    {
        i_bits = (1 * i_lambda2 + 128) >> 8;
    }
    else if( h->param.b_cabac )
    {
        x264_cabac_t cabac_tmp;
        h->mc.memcpy_aligned( &cabac_tmp.f8_bits_encoded, &h->cabac.f8_bits_encoded, sizeof(x264_cabac_t) - __builtin_offsetof (x264_cabac_t, f8_bits_encoded) - ((h->sps->i_chroma_format_idc == CHROMA_444) ? 0 : (1024+12)-460) );
        x264_macroblock_size_cabac( h, &cabac_tmp );
        i_bits = ( (uint64_t)cabac_tmp.f8_bits_encoded * i_lambda2 + 32768 ) >> 16;
    }
    else
    {
    }

    h->mb.b_transform_8x8 = b_transform_bak;
    h->mb.i_type = type_bak;

    return i_ssd + i_bits;
}



 uint64_t x264_rd_cost_subpart( x264_t *h, int i_lambda2, int i4, int i_pixel )
{
    uint64_t i_ssd, i_bits;

    x264_macroblock_encode_p4x4( h, i4 );
    if( i_pixel == PIXEL_8x4 )
        x264_macroblock_encode_p4x4( h, i4+1 );
    if( i_pixel == PIXEL_4x8 )
        x264_macroblock_encode_p4x4( h, i4+2 );

    i_ssd = ssd_plane( h, i_pixel, 0, block_idx_x[i4]*4, block_idx_y[i4]*4 );
    if( (h->sps->i_chroma_format_idc == CHROMA_444) )
    {
        int chromassd = ssd_plane( h, i_pixel, 1, block_idx_x[i4]*4, block_idx_y[i4]*4 )
                      + ssd_plane( h, i_pixel, 2, block_idx_x[i4]*4, block_idx_y[i4]*4 );
        chromassd = ((uint64_t)chromassd * h->mb.i_chroma_lambda2_offset + 128) >> 8;
        i_ssd += chromassd;
    }

    if( h->param.b_cabac )
    {
        x264_cabac_t cabac_tmp;
        h->mc.memcpy_aligned( &cabac_tmp.f8_bits_encoded, &h->cabac.f8_bits_encoded, sizeof(x264_cabac_t) - __builtin_offsetof (x264_cabac_t, f8_bits_encoded) - ((h->sps->i_chroma_format_idc == CHROMA_444) ? 0 : (1024+12)-460) );
        x264_subpartition_size_cabac( h, &cabac_tmp, i4, i_pixel );
        i_bits = ( (uint64_t)cabac_tmp.f8_bits_encoded * i_lambda2 + 128 ) >> 8;
    }
    //else
        //i_bits = x264_subpartition_size_cavlc( h, i4, i_pixel );

    return (i_ssd<<8) + i_bits;
}

uint64_t x264_rd_cost_part( x264_t *h, int i_lambda2, int i4, int i_pixel )
{
    uint64_t i_ssd, i_bits;
    int i8 = i4 >> 2;

    if( i_pixel == PIXEL_16x16 )
    {
        int i_cost = x264_rd_cost_mb( h, i_lambda2 );
        return i_cost;
    }

    if( i_pixel > PIXEL_8x8 )
        return x264_rd_cost_subpart( h, i_lambda2, i4, i_pixel );

    h->mb.i_cbp_luma = 0;

    x264_macroblock_encode_p8x8( h, i8 );
    if( i_pixel == PIXEL_16x8 )
        x264_macroblock_encode_p8x8( h, i8+1 );
    if( i_pixel == PIXEL_8x16 )
        x264_macroblock_encode_p8x8( h, i8+2 );

    int ssd_x = 8*(i8&1);
    int ssd_y = 8*(i8>>1);
    i_ssd = ssd_plane( h, i_pixel, 0, ssd_x, ssd_y );
    int chromapix = h->luma2chroma_pixel[i_pixel];
    int chromassd = ssd_plane( h, chromapix, 1, ssd_x>>h->mb.chroma_h_shift, ssd_y>>h->mb.chroma_v_shift )
                  + ssd_plane( h, chromapix, 2, ssd_x>>h->mb.chroma_h_shift, ssd_y>>h->mb.chroma_v_shift );
    i_ssd += ((uint64_t)chromassd * h->mb.i_chroma_lambda2_offset + 128) >> 8;

    if( h->param.b_cabac )
    {
        x264_cabac_t cabac_tmp;
        h->mc.memcpy_aligned( &cabac_tmp.f8_bits_encoded, &h->cabac.f8_bits_encoded, sizeof(x264_cabac_t) - __builtin_offsetof (x264_cabac_t, f8_bits_encoded) - ((h->sps->i_chroma_format_idc == CHROMA_444) ? 0 : (1024+12)-460) );
        x264_partition_size_cabac( h, &cabac_tmp, i8, i_pixel );
        i_bits = ( (uint64_t)cabac_tmp.f8_bits_encoded * i_lambda2 + 128 ) >> 8;
    }
    //else
        //i_bits = x264_partition_size_cavlc( h, i8, i_pixel ) * i_lambda2;

    return (i_ssd<<8) + i_bits;
}

 uint64_t x264_rd_cost_i8x8( x264_t *h, int i_lambda2, int i8, int i_mode, pixel edge[4][32] )
{
    uint64_t i_ssd, i_bits;
    int plane_count = (h->sps->i_chroma_format_idc == CHROMA_444) ? 3 : 1;
    int i_qp = h->mb.i_qp;
    h->mb.i_cbp_luma &= ~(1<<i8);
    h->mb.b_transform_8x8 = 1;

    for( int p = 0; p < plane_count; p++ )
    {
        x264_mb_encode_i8x8( h, p, i8, i_qp, i_mode, edge[p], 1 );
        i_qp = h->mb.i_chroma_qp;
    }

    i_ssd = ssd_plane( h, PIXEL_8x8, 0, (i8&1)*8, (i8>>1)*8 );
    if( (h->sps->i_chroma_format_idc == CHROMA_444) )
    {
        int chromassd = ssd_plane( h, PIXEL_8x8, 1, (i8&1)*8, (i8>>1)*8 )
                      + ssd_plane( h, PIXEL_8x8, 2, (i8&1)*8, (i8>>1)*8 );
        chromassd = ((uint64_t)chromassd * h->mb.i_chroma_lambda2_offset + 128) >> 8;
        i_ssd += chromassd;
    }

    if( h->param.b_cabac )
    {
        x264_cabac_t cabac_tmp;
        h->mc.memcpy_aligned( &cabac_tmp.f8_bits_encoded, &h->cabac.f8_bits_encoded, sizeof(x264_cabac_t) - __builtin_offsetof (x264_cabac_t, f8_bits_encoded) - ((h->sps->i_chroma_format_idc == CHROMA_444) ? 0 : (1024+12)-460) );
        x264_partition_i8x8_size_cabac( h, &cabac_tmp, i8, i_mode );
        i_bits = ( (uint64_t)cabac_tmp.f8_bits_encoded * i_lambda2 + 128 ) >> 8;
    }
    //else
        //i_bits = x264_partition_i8x8_size_cavlc( h, i8, i_mode ) * i_lambda2;

    return (i_ssd<<8) + i_bits;
}

 uint64_t x264_rd_cost_i4x4( x264_t *h, int i_lambda2, int i4, int i_mode )
{
    uint64_t i_ssd, i_bits;
    int plane_count = (h->sps->i_chroma_format_idc == CHROMA_444) ? 3 : 1;
    int i_qp = h->mb.i_qp;

    for( int p = 0; p < plane_count; p++ )
    {
        x264_mb_encode_i4x4( h, p, i4, i_qp, i_mode, 1 );
        i_qp = h->mb.i_chroma_qp;
    }

    i_ssd = ssd_plane( h, PIXEL_4x4, 0, block_idx_x[i4]*4, block_idx_y[i4]*4 );
    if( (h->sps->i_chroma_format_idc == CHROMA_444) )
    {
        int chromassd = ssd_plane( h, PIXEL_4x4, 1, block_idx_x[i4]*4, block_idx_y[i4]*4 )
                      + ssd_plane( h, PIXEL_4x4, 2, block_idx_x[i4]*4, block_idx_y[i4]*4 );
        chromassd = ((uint64_t)chromassd * h->mb.i_chroma_lambda2_offset + 128) >> 8;
        i_ssd += chromassd;
    }

    if( h->param.b_cabac )
    {
        x264_cabac_t cabac_tmp;
        h->mc.memcpy_aligned( &cabac_tmp.f8_bits_encoded, &h->cabac.f8_bits_encoded, sizeof(x264_cabac_t) - __builtin_offsetof (x264_cabac_t, f8_bits_encoded) - ((h->sps->i_chroma_format_idc == CHROMA_444) ? 0 : (1024+12)-460) );
        x264_partition_i4x4_size_cabac( h, &cabac_tmp, i4, i_mode );
        i_bits = ( (uint64_t)cabac_tmp.f8_bits_encoded * i_lambda2 + 128 ) >> 8;
    }
    //else
        //i_bits = x264_partition_i4x4_size_cavlc( h, i4, i_mode ) * i_lambda2;

    return (i_ssd<<8) + i_bits;
}

 uint64_t x264_rd_cost_chroma( x264_t *h, int i_lambda2, int i_mode, int b_dct )
{
    uint64_t i_ssd, i_bits;

    if( b_dct )
        x264_mb_encode_chroma( h, 0, h->mb.i_chroma_qp );

    int chromapix = h->luma2chroma_pixel[PIXEL_16x16];
    i_ssd = ssd_plane( h, chromapix, 1, 0, 0 )
          + ssd_plane( h, chromapix, 2, 0, 0 );

    h->mb.i_chroma_pred_mode = i_mode;

    if( h->param.b_cabac )
    {
        x264_cabac_t cabac_tmp;
        h->mc.memcpy_aligned( &cabac_tmp.f8_bits_encoded, &h->cabac.f8_bits_encoded, sizeof(x264_cabac_t) - __builtin_offsetof (x264_cabac_t, f8_bits_encoded) - ((h->sps->i_chroma_format_idc == CHROMA_444) ? 0 : (1024+12)-460) );
        x264_chroma_size_cabac( h, &cabac_tmp );
        i_bits = ( (uint64_t)cabac_tmp.f8_bits_encoded * i_lambda2 + 128 ) >> 8;
    }
    //else
        //i_bits = x264_chroma_size_cavlc( h ) * i_lambda2;

    return (i_ssd<<8) + i_bits;
}
void x264_rdo_init( void )
{
    for( int i_prefix = 0; i_prefix < 15; i_prefix++ )
    {
        for( int i_ctx = 0; i_ctx < 128; i_ctx++ )
        {
            int f8_bits = 0;
            uint8_t ctx = i_ctx;

            for( int i = 1; i < i_prefix; i++ )
                f8_bits += x264_cabac_size_decision2( &ctx, 1 );
            if( i_prefix > 0 && i_prefix < 14 )
                f8_bits += x264_cabac_size_decision2( &ctx, 0 );
            f8_bits += 1 << 8;

            x264_cabac_size_unary[i_prefix][i_ctx] = f8_bits;
            x264_cabac_transition_unary[i_prefix][i_ctx] = ctx;
        }
    }
    for( int i_ctx = 0; i_ctx < 128; i_ctx++ )
    {
        int f8_bits = 0;
        uint8_t ctx = i_ctx;

        for( int i = 0; i < 5; i++ )
            f8_bits += x264_cabac_size_decision2( &ctx, 1 );
        f8_bits += 1 << 8;

        cabac_size_5ones[i_ctx] = f8_bits;
        cabac_transition_5ones[i_ctx] = ctx;
    }
}

 struct _trellis_node_t
{
    uint64_t score;
    int level_idx;
    uint8_t cabac_state[4];
};

typedef struct _trellis_node_t trellis_node_t;


struct _trellis_level_t
{
    uint16_t next;
    uint16_t abs_level;
} ;

typedef struct _trellis_level_t trellis_level_t;

 const uint8_t x264_zigzag_scan2x2[4] = { 0, 1, 2, 3 };
 const uint8_t x264_zigzag_scan2x4[8] = { 0, 2, 1, 4, 6, 3, 5, 7 };

#endif