/*****************************************************************************
 * slicetype.c: lookahead analysis
 *****************************************************************************
 * Copyright (C) 2005-2014 x264 project
 *
 * Authors: Fiona Glaser <fiona@x264.com>
 *          Loren Merritt <lorenm@u.washington.edu>
 *          Dylan Yudaken <dyudaken@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 * This program is also available under a commercial proprietary license.
 * For more information, contact us at licensing@x264.com.
 *****************************************************************************/


#ifndef X264_ENCODER_SLICETYPE_H
#define X264_ENCODER_SLICETYPE_H

#include "common.h"
#include "macroblockENCODER.h"
#include "me.h"

// Indexed by pic_struct values
 const uint8_t delta_tfi_divisor[10] = { 0, 2, 1, 1, 2, 2, 3, 3, 4, 6 };

 int x264_slicetype_frame_cost( x264_t *h, x264_mb_analysis_t *a,
                                      x264_frame_t **frames, int p0, int p1, int b,
                                      int b_intra_penalty );

void x264_weights_analyse( x264_t *h, x264_frame_t *fenc, x264_frame_t *ref, int b_lookahead );

int x264_opencl_lowres_init( x264_t *h, x264_frame_t *fenc, int lambda );
int x264_opencl_motionsearch( x264_t *h, x264_frame_t **frames, int b, int ref, int b_islist1, int lambda, const x264_weight_t *w );
int x264_opencl_finalize_cost( x264_t *h, int lambda, x264_frame_t **frames, int p0, int p1, int b, int dist_scale_factor );
int x264_opencl_precalculate_frame_cost( x264_t *h, x264_frame_t **frames, int lambda, int p0, int p1, int b );
void x264_opencl_flush( x264_t *h );
void x264_opencl_slicetype_prep( x264_t *h, x264_frame_t **frames, int num_frames, int lambda );
void x264_opencl_slicetype_end( x264_t *h );

 void x264_lowres_context_init( x264_t *h, x264_mb_analysis_t *a )
{
    a->i_qp = (12+(6*(8 -8)));
    a->i_lambda = x264_lambda_tab[ a->i_qp ];
    x264_mb_analyse_load_costs( h, a );
    if( h->param.analyse.i_subpel_refine > 1 )
    {
        h->mb.i_me_method = ( (1)<(h->param.analyse.i_me_method) ? (1) : (h->param.analyse.i_me_method) );
        h->mb.i_subpel_refine = 4;
    }
    else
    {
        h->mb.i_me_method = 0;
        h->mb.i_subpel_refine = 2;
    }
    h->mb.b_chroma_me = 0;
}


 void x264_weight_get_h264( int weight_nonh264, int offset, x264_weight_t *w )
{
    w->i_offset = offset;
    w->i_denom = 7;
    w->i_scale = weight_nonh264;
    while( w->i_denom > 0 && (w->i_scale > 127) )
    {
        w->i_denom--;
        w->i_scale >>= 1;
    }
    w->i_scale = ( (w->i_scale)<(127) ? (w->i_scale) : (127) );
}

  pixel *x264_weight_cost_init_luma( x264_t *h, x264_frame_t *fenc, x264_frame_t *ref, pixel *dest )
{
    int ref0_distance = fenc->i_frame - ref->i_frame - 1;


    if( fenc->lowres_mvs[0][ref0_distance][0][0] != 0x7FFF )
    {
        int i_stride = fenc->i_stride_lowres;
        int i_lines = fenc->i_lines_lowres;
        int i_width = fenc->i_width_lowres;
        int i_mb_xy = 0;
        pixel *p = dest;

        for( int y = 0; y < i_lines; y += 8, p += i_stride*8 )
            for( int x = 0; x < i_width; x += 8, i_mb_xy++ )
            {
                int mvx = fenc->lowres_mvs[0][ref0_distance][i_mb_xy][0];
                int mvy = fenc->lowres_mvs[0][ref0_distance][i_mb_xy][1];
                h->mc.mc_luma( p+x, i_stride, ref->lowres, i_stride,
                               mvx+(x<<2), mvy+(y<<2), 8, 8, x264_weight_none );
            }
        ;
        return dest;
    }
    ;
    return ref->lowres[0];
}







  void x264_weight_cost_init_chroma( x264_t *h, x264_frame_t *fenc, x264_frame_t *ref, pixel *dstu, pixel *dstv )
{
    int ref0_distance = fenc->i_frame - ref->i_frame - 1;
    int i_stride = fenc->i_stride[1];
    int i_offset = i_stride / 2;
    int i_lines = fenc->i_lines[1];
    int i_width = fenc->i_width[1];
    int v_shift = h->mb.chroma_v_shift;
    int cw = 8*h->mb.i_mb_width;
    int ch = 16*h->mb.i_mb_height >> v_shift;
    int height = 16 >> v_shift;

    if( fenc->lowres_mvs[0][ref0_distance][0][0] != 0x7FFF )
    {
        x264_frame_expand_border_chroma( h, ref, 1 );
        for( int y = 0, mb_xy = 0, pel_offset_y = 0; y < i_lines; y += height, pel_offset_y = y*i_stride )
            for( int x = 0, pel_offset_x = 0; x < i_width; x += 8, mb_xy++, pel_offset_x += 8 )
            {
                pixel *pixu = dstu + pel_offset_y + pel_offset_x;
                pixel *pixv = dstv + pel_offset_y + pel_offset_x;
                pixel *src1 = ref->plane[1] + pel_offset_y + pel_offset_x*2;
                int mvx = fenc->lowres_mvs[0][ref0_distance][mb_xy][0];
                int mvy = fenc->lowres_mvs[0][ref0_distance][mb_xy][1];
                h->mc.mc_chroma( pixu, pixv, i_stride, src1, i_stride, mvx, 2*mvy>>v_shift, 8, height );
            }
    }
    else
        h->mc.plane_copy_deinterleave( dstu, i_stride, dstv, i_stride, ref->plane[1], i_stride, cw, ch );
    h->mc.plane_copy_deinterleave( dstu+i_offset, i_stride, dstv+i_offset, i_stride, fenc->plane[1], i_stride, cw, ch );
    ;
}

  pixel *x264_weight_cost_init_chroma444( x264_t *h, x264_frame_t *fenc, x264_frame_t *ref, pixel *dst, int p )
{
    int ref0_distance = fenc->i_frame - ref->i_frame - 1;
    int i_stride = fenc->i_stride[p];
    int i_lines = fenc->i_lines[p];
    int i_width = fenc->i_width[p];

    if( fenc->lowres_mvs[0][ref0_distance][0][0] != 0x7FFF )
    {
        x264_frame_expand_border_chroma( h, ref, p );
        for( int y = 0, mb_xy = 0, pel_offset_y = 0; y < i_lines; y += 16, pel_offset_y = y*i_stride )
            for( int x = 0, pel_offset_x = 0; x < i_width; x += 16, mb_xy++, pel_offset_x += 16 )
            {
                pixel *pix = dst + pel_offset_y + pel_offset_x;
                pixel *src = ref->plane[p] + pel_offset_y + pel_offset_x;
                int mvx = fenc->lowres_mvs[0][ref0_distance][mb_xy][0] / 2;
                int mvy = fenc->lowres_mvs[0][ref0_distance][mb_xy][1] / 2;


                h->mc.copy_16x16_unaligned( pix, i_stride, src+mvx+mvy*i_stride, i_stride, 16 );
            }
        ;
        return dst;
    }
    ;
    return ref->plane[p];
}

 int x264_weight_slice_header_cost( x264_t *h, x264_weight_t *w, int b_chroma )
{

    int lambda = x264_lambda_tab[(12+(6*(8 -8)))];

    if( b_chroma )
        lambda *= 4;
    int numslices;
    if( h->param.i_slice_count )
        numslices = h->param.i_slice_count;
    else if( h->param.i_slice_max_mbs )
        numslices = (h->mb.i_mb_width * h->mb.i_mb_height + h->param.i_slice_max_mbs-1) / h->param.i_slice_max_mbs;
    else
        numslices = 1;



    int denom_cost = bs_size_ue( w[0].i_denom ) * (2 - b_chroma);
    return lambda * numslices * ( 10 + denom_cost + 2 * (bs_size_se( w[0].i_scale ) + bs_size_se( w[0].i_offset )) );
}

  unsigned int x264_weight_cost_luma( x264_t *h, x264_frame_t *fenc, pixel *src, x264_weight_t *w )
{
    unsigned int cost = 0;
    int i_stride = fenc->i_stride_lowres;
    int i_lines = fenc->i_lines_lowres;
    int i_width = fenc->i_width_lowres;
    pixel *fenc_plane = fenc->lowres[0];
    pixel buf [8*8] ;
    int pixoff = 0;
    int i_mb = 0;

    if( w )
    {
        for( int y = 0; y < i_lines; y += 8, pixoff = y*i_stride )
            for( int x = 0; x < i_width; x += 8, i_mb++, pixoff += 8)
            {
                w->weightfn[8>>2]( buf, 8, &src[pixoff], i_stride, w, 8 );
                int cmp = h->pixf.mbcmp[PIXEL_8x8]( buf, 8, &fenc_plane[pixoff], i_stride );
                cost += ( (cmp)<(fenc->i_intra_cost[i_mb]) ? (cmp) : (fenc->i_intra_cost[i_mb]) );
            }
        cost += x264_weight_slice_header_cost( h, w, 0 );
    }
    else
        for( int y = 0; y < i_lines; y += 8, pixoff = y*i_stride )
            for( int x = 0; x < i_width; x += 8, i_mb++, pixoff += 8 )
            {
                int cmp = h->pixf.mbcmp[PIXEL_8x8]( &src[pixoff], i_stride, &fenc_plane[pixoff], i_stride );
                cost += ( (cmp)<(fenc->i_intra_cost[i_mb]) ? (cmp) : (fenc->i_intra_cost[i_mb]) );
            }
    ;
    return cost;
}

  unsigned int x264_weight_cost_chroma( x264_t *h, x264_frame_t *fenc, pixel *ref, x264_weight_t *w )
{
    unsigned int cost = 0;
    int i_stride = fenc->i_stride[1];
    int i_lines = fenc->i_lines[1];
    int i_width = fenc->i_width[1];
    pixel *src = ref + (i_stride >> 1);
    pixel buf [8*16] ;
    int pixoff = 0;
    int height = 16 >> h->mb.chroma_v_shift;
    if( w )
    {
        for( int y = 0; y < i_lines; y += height, pixoff = y*i_stride )
            for( int x = 0; x < i_width; x += 8, pixoff += 8 )
            {
                w->weightfn[8>>2]( buf, 8, &ref[pixoff], i_stride, w, height );





                cost += h->pixf.asd8( buf, 8, &src[pixoff], i_stride, height );
            }
        cost += x264_weight_slice_header_cost( h, w, 1 );
    }
    else
        for( int y = 0; y < i_lines; y += height, pixoff = y*i_stride )
            for( int x = 0; x < i_width; x += 8, pixoff += 8 )
                cost += h->pixf.asd8( &ref[pixoff], i_stride, &src[pixoff], i_stride, height );
    ;
    return cost;
}

  unsigned int x264_weight_cost_chroma444( x264_t *h, x264_frame_t *fenc, pixel *ref, x264_weight_t *w, int p )
{
    unsigned int cost = 0;
    int i_stride = fenc->i_stride[p];
    int i_lines = fenc->i_lines[p];
    int i_width = fenc->i_width[p];
    pixel *src = fenc->plane[p];
    pixel buf [16*16] ;
    int pixoff = 0;
    if( w )
    {
        for( int y = 0; y < i_lines; y += 16, pixoff = y*i_stride )
            for( int x = 0; x < i_width; x += 16, pixoff += 16 )
            {
                w->weightfn[16>>2]( buf, 16, &ref[pixoff], i_stride, w, 16 );
                cost += h->pixf.mbcmp[PIXEL_16x16]( buf, 16, &src[pixoff], i_stride );
            }
        cost += x264_weight_slice_header_cost( h, w, 1 );
    }
    else
        for( int y = 0; y < i_lines; y += 16, pixoff = y*i_stride )
            for( int x = 0; x < i_width; x += 16, pixoff += 16 )
                cost += h->pixf.mbcmp[PIXEL_16x16]( &ref[pixoff], i_stride, &src[pixoff], i_stride );
    ;
    return cost;
}

void x264_weights_analyse( x264_t *h, x264_frame_t *fenc, x264_frame_t *ref, int b_lookahead )
{
    int i_delta_index = fenc->i_frame - ref->i_frame - 1;

    const float epsilon = 1.f/128.f;
    x264_weight_t *weights = fenc->weight[0];
    { (weights[0]).i_scale = (1); (weights[0]).i_denom = (0); (weights[0]).i_offset = (0); if( 0 ) h->mc.weight_cache( h, &weights[0] ); else weights[0].weightfn = ((void *)0);};
    { (weights[1]).i_scale = (1); (weights[1]).i_denom = (0); (weights[1]).i_offset = (0); if( 0 ) h->mc.weight_cache( h, &weights[1] ); else weights[1].weightfn = ((void *)0);};
    { (weights[2]).i_scale = (1); (weights[2]).i_denom = (0); (weights[2]).i_offset = (0); if( 0 ) h->mc.weight_cache( h, &weights[2] ); else weights[2].weightfn = ((void *)0);};
    int chroma_initted = 0;
    float guess_scale[3];
    float fenc_mean[3];
    float ref_mean[3];
    for( int plane = 0; plane <= 2*!b_lookahead; plane++ )
    {
        float fenc_var = fenc->i_pixel_ssd[plane] + !ref->i_pixel_ssd[plane];
        float ref_var = ref->i_pixel_ssd[plane] + !ref->i_pixel_ssd[plane];
        guess_scale[plane] = sqrtf( fenc_var / ref_var );
        fenc_mean[plane] = (float)fenc->i_pixel_sum[plane] / (fenc->i_lines[!!plane] * fenc->i_width[!!plane]) / (1 << (8 - 8));
        ref_mean[plane] = (float) ref->i_pixel_sum[plane] / (fenc->i_lines[!!plane] * fenc->i_width[!!plane]) / (1 << (8 - 8));
    }

    int chroma_denom = 7;
    if( !b_lookahead )
    {

        while( chroma_denom > 0 )
        {
            float thresh = 127.f / (1<<chroma_denom);
            if( guess_scale[1] < thresh && guess_scale[2] < thresh )
                break;
            chroma_denom--;
        }
    }


    for( int plane = 0; plane <= 2 && !( plane && ( !weights[0].weightfn || b_lookahead ) ); plane++ )
    {
        int minoff, minscale, mindenom;
        unsigned int minscore, origscore;
        int found;


        if( fabsf( ref_mean[plane] - fenc_mean[plane] ) < 0.5f && fabsf( 1.f - guess_scale[plane] ) < epsilon )
        {
            { (weights[plane]).i_scale = (1); (weights[plane]).i_denom = (0); (weights[plane]).i_offset = (0); if( 0 ) h->mc.weight_cache( h, &weights[plane] ); else weights[plane].weightfn = ((void *)0);};
            continue;
        }

        if( plane )
        {
            weights[plane].i_denom = chroma_denom;
            weights[plane].i_scale = x264_clip3( round( guess_scale[plane] * (1<<chroma_denom) ), 0, 255 );
            if( weights[plane].i_scale > 127 )
            {
                weights[1].weightfn = weights[2].weightfn = ((void *)0);
                break;
            }
        }
        else
            x264_weight_get_h264( round( guess_scale[plane] * 128 ), 0, &weights[plane] );

        found = 0;
        mindenom = weights[plane].i_denom;
        minscale = weights[plane].i_scale;
        minoff = 0;

        pixel *mcbuf;
        if( !plane )
        {
            if( !fenc->b_intra_calculated )
            {
                x264_mb_analysis_t a;
                x264_lowres_context_init( h, &a );
                x264_slicetype_frame_cost( h, &a, &fenc, 0, 0, 0, 0 );
            }
            mcbuf = x264_weight_cost_init_luma( h, fenc, ref, h->mb.p_weight_buf[0] );
            origscore = minscore = x264_weight_cost_luma( h, fenc, mcbuf, ((void *)0) );
        }
        else
        {
            if( (h->sps->i_chroma_format_idc == CHROMA_444) )
            {
                mcbuf = x264_weight_cost_init_chroma444( h, fenc, ref, h->mb.p_weight_buf[0], plane );
                origscore = minscore = x264_weight_cost_chroma444( h, fenc, mcbuf, ((void *)0), plane );
            }
            else
            {
                pixel *dstu = h->mb.p_weight_buf[0];
                pixel *dstv = h->mb.p_weight_buf[0]+fenc->i_stride[1]*fenc->i_lines[1];
                if( !chroma_initted++ )
                    x264_weight_cost_init_chroma( h, fenc, ref, dstu, dstv );
                mcbuf = plane == 1 ? dstu : dstv;
                origscore = minscore = x264_weight_cost_chroma( h, fenc, mcbuf, ((void *)0) );
            }
        }

        if( !minscore )
            continue;


         const uint8_t weight_check_distance[][2] =
        {
            {0,0},{0,0},{0,1},{0,1},
            {0,1},{0,1},{0,1},{1,1},
            {1,1},{2,1},{2,1},{4,2}
        };
        int scale_dist = b_lookahead ? 0 : weight_check_distance[h->param.analyse.i_subpel_refine][0];
        int offset_dist = b_lookahead ? 0 : weight_check_distance[h->param.analyse.i_subpel_refine][1];

        int start_scale = x264_clip3( minscale - scale_dist, 0, 127 );
        int end_scale = x264_clip3( minscale + scale_dist, 0, 127 );
        for( int i_scale = start_scale; i_scale <= end_scale; i_scale++ )
        {
            int cur_scale = i_scale;
            int cur_offset = fenc_mean[plane] - ref_mean[plane] * cur_scale / (1 << mindenom) + 0.5f * b_lookahead;
            if( cur_offset < - 128 || cur_offset > 127 )
            {



                cur_offset = x264_clip3( cur_offset, -128, 127 );
                cur_scale = (1 << mindenom) * (fenc_mean[plane] - cur_offset) / ref_mean[plane] + 0.5f;
                cur_scale = x264_clip3( cur_scale, 0, 127 );
            }
            int start_offset = x264_clip3( cur_offset - offset_dist, -128, 127 );
            int end_offset = x264_clip3( cur_offset + offset_dist, -128, 127 );
            for( int i_off = start_offset; i_off <= end_offset; i_off++ )
            {
                { (weights[plane]).i_scale = (cur_scale); (weights[plane]).i_denom = (mindenom); (weights[plane]).i_offset = (i_off); if( 1 ) h->mc.weight_cache( h, &weights[plane] ); else weights[plane].weightfn = ((void *)0);};
                unsigned int s;
                if( plane )
                {
                    if( (h->sps->i_chroma_format_idc == CHROMA_444) )
                        s = x264_weight_cost_chroma444( h, fenc, mcbuf, &weights[plane], plane );
                    else
                        s = x264_weight_cost_chroma( h, fenc, mcbuf, &weights[plane] );
                }
                else
                    s = x264_weight_cost_luma( h, fenc, mcbuf, &weights[plane] );
                if((s)<(minscore)){ (minscore)=(s); (minscale)=(cur_scale); (minoff)=(i_off); (found)=(1);};


                if( minoff == start_offset && i_off != start_offset )
                    break;
            }
        }
        ;


        if( !plane )
        {
            while( mindenom > 0 && !(minscale&1) )
            {
                mindenom--;
                minscale >>= 1;
            }
        }



        if( !found || (minscale == 1 << mindenom && minoff == 0) || (float)minscore / origscore > 0.998f )
        {
            { (weights[plane]).i_scale = (1); (weights[plane]).i_denom = (0); (weights[plane]).i_offset = (0); if( 0 ) h->mc.weight_cache( h, &weights[plane] ); else weights[plane].weightfn = ((void *)0);};
            continue;
        }
        else
            { (weights[plane]).i_scale = (minscale); (weights[plane]).i_denom = (mindenom); (weights[plane]).i_offset = (minoff); if( 1 ) h->mc.weight_cache( h, &weights[plane] ); else weights[plane].weightfn = ((void *)0);};

        if( h->param.analyse.i_weighted_pred == (-1) && weights[0].weightfn && !plane )
            fenc->f_weighted_cost_delta[i_delta_index] = (float)minscore / origscore;
    }


    if( weights[1].weightfn || weights[2].weightfn )
    {
        int denom = weights[1].weightfn ? weights[1].i_denom : weights[2].i_denom;
        int both_weighted = weights[1].weightfn && weights[2].weightfn;


        while( (!both_weighted && denom==7) ||
               (denom > 0 && !(weights[1].weightfn && (weights[1].i_scale&1))
                         && !(weights[2].weightfn && (weights[2].i_scale&1))) )
        {
            denom--;
            for( int i = 1; i <= 2; i++ )
                if( weights[i].weightfn )
                {
                    weights[i].i_scale >>= 1;
                    weights[i].i_denom = denom;
                }
        }
    }
    for( int i = 1; i <= 2; i++ )
        if( weights[i].weightfn )
            h->mc.weight_cache( h, &weights[i] );

    if( weights[0].weightfn && b_lookahead )
    {

        pixel *src = ref->buffer_lowres[0];
        pixel *dst = h->mb.p_weight_buf[0];
        int width = ref->i_width_lowres + 32*2;
        int height = ref->i_lines_lowres + 32*2;
        x264_weight_scale_plane( h, dst, ref->i_stride_lowres, src, ref->i_stride_lowres,
                                 width, height, &weights[0] );
        fenc->weighted[0] = h->mb.p_weight_buf[0] + 32 + ref->i_stride_lowres * 32;
    }
}
 void x264_slicetype_mb_cost( x264_t *h, x264_mb_analysis_t *a,
                                    x264_frame_t **frames, int p0, int p1, int b,
                                    int dist_scale_factor, int do_search[2], const x264_weight_t *w,
                                    int *output_inter, int *output_intra )
{
    x264_frame_t *fref0 = frames[p0];
    x264_frame_t *fref1 = frames[p1];
    x264_frame_t *fenc = frames[b];
    const int b_bidir = (b < p1);
    const int i_mb_x = h->mb.i_mb_x;
    const int i_mb_y = h->mb.i_mb_y;
    const int i_mb_stride = h->mb.i_mb_width;
    const int i_mb_xy = i_mb_x + i_mb_y * i_mb_stride;
    const int i_stride = fenc->i_stride_lowres;
    const int i_pel_offset = 8 * (i_mb_x + i_mb_y * i_stride);
    const int i_bipred_weight = h->param.analyse.b_weighted_bipred ? 64 - (dist_scale_factor>>2) : 32;
    int16_t (*fenc_mvs[2])[2] = { &fenc->lowres_mvs[0][b-p0-1][i_mb_xy], &fenc->lowres_mvs[1][p1-b-1][i_mb_xy] };
    int (*fenc_costs[2]) = { &fenc->lowres_mv_costs[0][b-p0-1][i_mb_xy], &fenc->lowres_mv_costs[1][p1-b-1][i_mb_xy] };
    int b_frame_score_mb = (i_mb_x > 0 && i_mb_x < h->mb.i_mb_width - 1 &&
                            i_mb_y > 0 && i_mb_y < h->mb.i_mb_height - 1) ||
                            h->mb.i_mb_width <= 2 || h->mb.i_mb_height <= 2;

    pixel pix1 [9*32] ;
    pixel *pix2 = pix1+8;
    x264_me_t m[2];
    int i_bcost = (1<<28);
    int list_used = 0;

    int lowres_penalty = 4;

    h->mb.pic.p_fenc[0] = h->mb.pic.fenc_buf;
    h->mc.copy[PIXEL_8x8]( h->mb.pic.p_fenc[0], 16, &fenc->lowres[0][i_pel_offset], i_stride, 8 );

    if( p0 == p1 )
        goto lowres_intra_mb;


    h->mb.mv_limit_fpel[0][0] = -8*h->mb.i_mb_x - 4;
    h->mb.mv_limit_fpel[1][0] = 8*( h->mb.i_mb_width - h->mb.i_mb_x - 1 ) + 4;
    h->mb.mv_min_spel[0] = 4*( h->mb.mv_limit_fpel[0][0] - 8 );
    h->mb.mv_max_spel[0] = 4*( h->mb.mv_limit_fpel[1][0] + 8 );
    if( h->mb.i_mb_x >= h->mb.i_mb_width - 2 )
    {
        h->mb.mv_limit_fpel[0][1] = -8*h->mb.i_mb_y - 4;
        h->mb.mv_limit_fpel[1][1] = 8*( h->mb.i_mb_height - h->mb.i_mb_y - 1 ) + 4;
        h->mb.mv_min_spel[1] = 4*( h->mb.mv_limit_fpel[0][1] - 8 );
        h->mb.mv_max_spel[1] = 4*( h->mb.mv_limit_fpel[1][1] + 8 );
    }
    m[0].i_pixel = PIXEL_8x8;
    m[0].p_cost_mv = a->p_cost_mv;
    m[0].i_stride[0] = i_stride;
    m[0].p_fenc[0] = h->mb.pic.p_fenc[0];
    m[0].weight = w;
    m[0].i_ref = 0;
    { (m[0].p_fref)[0] = &(fref0->lowres)[0][i_pel_offset]; (m[0].p_fref)[1] = &(fref0->lowres)[1][i_pel_offset]; (m[0].p_fref)[2] = &(fref0->lowres)[2][i_pel_offset]; (m[0].p_fref)[3] = &(fref0->lowres)[3][i_pel_offset]; };
    m[0].p_fref_w = m[0].p_fref[0];
    if( w[0].weightfn )
        (m[0].p_fref_w) = &(fenc->weighted[0])[i_pel_offset];;

    if( b_bidir )
    {
        int16_t *mvr = fref1->lowres_mvs[0][p1-p0-1][i_mb_xy];
        int16_t dmv [2] [2] ;

        m[1].i_pixel = PIXEL_8x8;
        m[1].p_cost_mv = a->p_cost_mv;
        m[1].i_stride[0] = i_stride;
        m[1].p_fenc[0] = h->mb.pic.p_fenc[0];
        m[1].i_ref = 0;
        m[1].weight = x264_weight_none;
        { (m[1].p_fref)[0] = &(fref1->lowres)[0][i_pel_offset]; (m[1].p_fref)[1] = &(fref1->lowres)[1][i_pel_offset]; (m[1].p_fref)[2] = &(fref1->lowres)[2][i_pel_offset]; (m[1].p_fref)[3] = &(fref1->lowres)[3][i_pel_offset]; };
        m[1].p_fref_w = m[1].p_fref[0];

        dmv[0][0] = ( mvr[0] * dist_scale_factor + 128 ) >> 8;
        dmv[0][1] = ( mvr[1] * dist_scale_factor + 128 ) >> 8;
        dmv[1][0] = dmv[0][0] - mvr[0];
        dmv[1][1] = dmv[0][1] - mvr[1];
        { dmv[0][0] = x264_clip3( dmv[0][0], h->mb.mv_min_spel[0], h->mb.mv_max_spel[0] ); dmv[0][1] = x264_clip3( dmv[0][1], h->mb.mv_min_spel[1], h->mb.mv_max_spel[1] ); };
        { dmv[1][0] = x264_clip3( dmv[1][0], h->mb.mv_min_spel[0], h->mb.mv_max_spel[0] ); dmv[1][1] = x264_clip3( dmv[1][1], h->mb.mv_min_spel[1], h->mb.mv_max_spel[1] ); };
        if( h->param.analyse.i_subpel_refine <= 1 )
            (((x264_union64_t*)(dmv))->i) &= ~0x0001000100010001ULL;

        { int i_cost; if( h->param.analyse.i_subpel_refine <= 1 ) { int hpel_idx1 = (((dmv[0])[0]&2)>>1) + ((dmv[0])[1]&2); int hpel_idx2 = (((dmv[1])[0]&2)>>1) + ((dmv[1])[1]&2); pixel *src1 = m[0].p_fref[hpel_idx1] + ((dmv[0])[0]>>2) + ((dmv[0])[1]>>2) * m[0].i_stride[0]; pixel *src2 = m[1].p_fref[hpel_idx2] + ((dmv[1])[0]>>2) + ((dmv[1])[1]>>2) * m[1].i_stride[0]; h->mc.avg[PIXEL_8x8]( pix1, 16, src1, m[0].i_stride[0], src2, m[1].i_stride[0], i_bipred_weight ); } else { intptr_t stride1 = 16, stride2 = 16; pixel *src1, *src2; src1 = h->mc.get_ref( pix1, &stride1, m[0].p_fref, m[0].i_stride[0], (dmv[0])[0], (dmv[0])[1], 8, 8, w ); src2 = h->mc.get_ref( pix2, &stride2, m[1].p_fref, m[1].i_stride[0], (dmv[1])[0], (dmv[1])[1], 8, 8, w ); h->mc.avg[PIXEL_8x8]( pix1, 16, src1, stride1, src2, stride2, i_bipred_weight ); } i_cost = 0 * a->i_lambda + h->pixf.mbcmp[PIXEL_8x8]( m[0].p_fenc[0], 16, pix1, 16 ); if((i_cost)<(i_bcost)){ (i_bcost)=(i_cost); (list_used)=(3);}; };
        if( (((x264_union64_t*)(dmv))->i) )
        {
            int i_cost;
            h->mc.avg[PIXEL_8x8]( pix1, 16, m[0].p_fref[0], m[0].i_stride[0], m[1].p_fref[0], m[1].i_stride[0], i_bipred_weight );
            i_cost = h->pixf.mbcmp[PIXEL_8x8]( m[0].p_fenc[0], 16, pix1, 16 );
            if((i_cost)<(i_bcost)){ (i_bcost)=(i_cost); (list_used)=(3);};
        }
    }

    for( int l = 0; l < 1 + b_bidir; l++ )
    {
        if( do_search[l] )
        {
            int i_mvc = 0;
            int16_t (*fenc_mv)[2] = fenc_mvs[l];
            int16_t mvc[4][2] ;


            (((x264_union32_t*)(mvc[0]))->i) = 0;
            (((x264_union32_t*)(mvc[2]))->i) = 0;

            if( i_mb_x < h->mb.i_mb_width - 1 )
                { (((x264_union32_t*)(mvc[i_mvc]))->i) = (((x264_union32_t*)(fenc_mv[1]))->i); i_mvc++; };
            if( i_mb_y < h->i_threadslice_end - 1 )
            {
                { (((x264_union32_t*)(mvc[i_mvc]))->i) = (((x264_union32_t*)(fenc_mv[i_mb_stride]))->i); i_mvc++; };
                if( i_mb_x > 0 )
                    { (((x264_union32_t*)(mvc[i_mvc]))->i) = (((x264_union32_t*)(fenc_mv[i_mb_stride-1]))->i); i_mvc++; };
                if( i_mb_x < h->mb.i_mb_width - 1 )
                    { (((x264_union32_t*)(mvc[i_mvc]))->i) = (((x264_union32_t*)(fenc_mv[i_mb_stride+1]))->i); i_mvc++; };
            }

            if( i_mvc <= 1 )
                (((x264_union32_t*)(m[l].mvp))->i) = (((x264_union32_t*)(mvc[0]))->i);
            else
                x264_median_mv( m[l].mvp, mvc[0], mvc[1], mvc[2] );



            if( !(((x264_union32_t*)(m[l].mvp))->i) )
            {
                m[l].cost = h->pixf.mbcmp[PIXEL_8x8]( m[l].p_fenc[0], 16, m[l].p_fref[0], m[l].i_stride[0] );
                if( m[l].cost < 64 )
                {
                    (((x264_union32_t*)(m[l].mv))->i) = 0;
                    goto skip_motionest;
                }
            }

            x264_me_search_ref( h, &m[l], mvc, i_mvc, ((void *)0) );
            m[l].cost -= a->p_cost_mv[0];
            if( (((x264_union32_t*)(m[l].mv))->i) )
                m[l].cost += 5 * a->i_lambda;

skip_motionest:
            (((x264_union32_t*)(fenc_mvs[l]))->i) = (((x264_union32_t*)(m[l].mv))->i);
            *fenc_costs[l] = m[l].cost;
        }
        else
        {
            (((x264_union32_t*)(m[l].mv))->i) = (((x264_union32_t*)(fenc_mvs[l]))->i);
            m[l].cost = *fenc_costs[l];
        }
        if((m[l].cost)<(i_bcost)){ (i_bcost)=(m[l].cost); (list_used)=(l+1);};
    }

    if( b_bidir && ( (((x264_union32_t*)(m[0].mv))->i) || (((x264_union32_t*)(m[1].mv))->i) ) )
        { int i_cost; if( h->param.analyse.i_subpel_refine <= 1 ) { int hpel_idx1 = (((m[0].mv)[0]&2)>>1) + ((m[0].mv)[1]&2); int hpel_idx2 = (((m[1].mv)[0]&2)>>1) + ((m[1].mv)[1]&2); pixel *src1 = m[0].p_fref[hpel_idx1] + ((m[0].mv)[0]>>2) + ((m[0].mv)[1]>>2) * m[0].i_stride[0]; pixel *src2 = m[1].p_fref[hpel_idx2] + ((m[1].mv)[0]>>2) + ((m[1].mv)[1]>>2) * m[1].i_stride[0]; h->mc.avg[PIXEL_8x8]( pix1, 16, src1, m[0].i_stride[0], src2, m[1].i_stride[0], i_bipred_weight ); } else { intptr_t stride1 = 16, stride2 = 16; pixel *src1, *src2; src1 = h->mc.get_ref( pix1, &stride1, m[0].p_fref, m[0].i_stride[0], (m[0].mv)[0], (m[0].mv)[1], 8, 8, w ); src2 = h->mc.get_ref( pix2, &stride2, m[1].p_fref, m[1].i_stride[0], (m[1].mv)[0], (m[1].mv)[1], 8, 8, w ); h->mc.avg[PIXEL_8x8]( pix1, 16, src1, stride1, src2, stride2, i_bipred_weight ); } i_cost = 5 * a->i_lambda + h->pixf.mbcmp[PIXEL_8x8]( m[0].p_fenc[0], 16, pix1, 16 ); if((i_cost)<(i_bcost)){ (i_bcost)=(i_cost); (list_used)=(3);}; };

lowres_intra_mb:
    if( !fenc->b_intra_calculated )
    {
        pixel edge [36] ;
        pixel *pix = &pix1[8+32];
        pixel *src = &fenc->lowres[0][i_pel_offset];
        const int intra_penalty = 5 * a->i_lambda;
        int satds[3];
        int pixoff = 4 / sizeof(pixel);


        memcpy( pix-32, src-i_stride, 16 * sizeof(pixel) );
        for( int i = -1; i < 8; i++ )
            (((x264_union32_t*)(&pix[i*32 -pixoff]))->i) = (((x264_union32_t*)(&src[i*i_stride-pixoff]))->i);

        h->pixf.intra_mbcmp_x3_8x8c( h->mb.pic.p_fenc[0], pix, satds );
        int i_icost = ( ((satds[0]))<(( ((satds[1]))<((satds[2])) ? ((satds[1])) : ((satds[2])) )) ? ((satds[0])) : (( ((satds[1]))<((satds[2])) ? ((satds[1])) : ((satds[2])) )) );

        if( h->param.analyse.i_subpel_refine > 1 )
        {
            h->predict_8x8c[I_PRED_CHROMA_P]( pix );
            int satd = h->pixf.mbcmp[PIXEL_8x8]( pix, 32, h->mb.pic.p_fenc[0], 16 );
            i_icost = ( (i_icost)<(satd) ? (i_icost) : (satd) );
            h->predict_8x8_filter( pix, edge, ALL_NEIGHBORS, ALL_NEIGHBORS );
            for( int i = 3; i < 9; i++ )
            {
                h->predict_8x8[i]( pix, edge );
                satd = h->pixf.mbcmp[PIXEL_8x8]( pix, 32, h->mb.pic.p_fenc[0], 16 );
                i_icost = ( (i_icost)<(satd) ? (i_icost) : (satd) );
            }
        }

        i_icost += intra_penalty + lowres_penalty;
        fenc->i_intra_cost[i_mb_xy] = i_icost;
        int i_icost_aq = i_icost;
        if( h->param.rc.i_aq_mode )
            i_icost_aq = (i_icost_aq * fenc->i_inv_qscale_factor[i_mb_xy] + 128) >> 8;
        output_intra[(4 + (h->mb.i_mb_y - h->i_threadslice_start))] += i_icost_aq;
        if( b_frame_score_mb )
        {
            output_intra[0] += i_icost;
            output_intra[1] += i_icost_aq;
        }
    }
    i_bcost += lowres_penalty;



    if( !b_bidir )
    {
        int i_icost = fenc->i_intra_cost[i_mb_xy];
        int b_intra = i_icost < i_bcost;
        if( b_intra )
        {
            i_bcost = i_icost;
            list_used = 0;
        }
        if( b_frame_score_mb )
            output_inter[2] += b_intra;
    }


    if( p0 != p1 )
    {
        int i_bcost_aq = i_bcost;
        if( h->param.rc.i_aq_mode )
            i_bcost_aq = (i_bcost_aq * fenc->i_inv_qscale_factor[i_mb_xy] + 128) >> 8;
        output_inter[(4 + (h->mb.i_mb_y - h->i_threadslice_start))] += i_bcost_aq;
        if( b_frame_score_mb )
        {

            output_inter[0] += i_bcost;
            output_inter[1] += i_bcost_aq;
        }
    }

    fenc->lowres_costs[b-p0][p1-b][i_mb_xy] = ( (i_bcost)<(((1<<14)-1)) ? (i_bcost) : (((1<<14)-1)) ) + (list_used << 14);
}







struct _x264_slicetype_slice_t
{
    x264_t *h;
    x264_mb_analysis_t *a;
    x264_frame_t **frames;
    int p0;
    int p1;
    int b;
    int dist_scale_factor;
    int *do_search;
    const x264_weight_t *w;
    int *output_inter;
    int *output_intra;
} ;

typedef struct _x264_slicetype_slice_t x264_slicetype_slice_t;

 void x264_slicetype_slice_cost( x264_slicetype_slice_t *s )
{
    x264_t *h = s->h;






    int do_edges = h->param.rc.b_mb_tree || h->param.rc.i_vbv_buffer_size || h->mb.i_mb_width <= 2 || h->mb.i_mb_height <= 2;

    int start_y = ( (h->i_threadslice_end - 1)<(h->mb.i_mb_height - 2 + do_edges) ? (h->i_threadslice_end - 1) : (h->mb.i_mb_height - 2 + do_edges) );
    int end_y = ( (h->i_threadslice_start)>(1 - do_edges) ? (h->i_threadslice_start) : (1 - do_edges) );
    int start_x = h->mb.i_mb_width - 2 + do_edges;
    int end_x = 1 - do_edges;

    for( h->mb.i_mb_y = start_y; h->mb.i_mb_y >= end_y; h->mb.i_mb_y-- )
        for( h->mb.i_mb_x = start_x; h->mb.i_mb_x >= end_x; h->mb.i_mb_x-- )
            x264_slicetype_mb_cost( h, s->a, s->frames, s->p0, s->p1, s->b, s->dist_scale_factor,
                                    s->do_search, s->w, s->output_inter, s->output_intra );
}

 int x264_slicetype_frame_cost( x264_t *h, x264_mb_analysis_t *a,
                                      x264_frame_t **frames, int p0, int p1, int b,
                                      int b_intra_penalty )
{
    int i_score = 0;
    int do_search[2];
    const x264_weight_t *w = x264_weight_none;
    x264_frame_t *fenc = frames[b];





    if( fenc->i_cost_est[b-p0][p1-b] >= 0 && (!h->param.rc.i_vbv_buffer_size || fenc->i_row_satds[b-p0][p1-b][0] != -1) )
        i_score = fenc->i_cost_est[b-p0][p1-b];
    else
    {
        int dist_scale_factor = 128;


        do_search[0] = b != p0 && fenc->lowres_mvs[0][b-p0-1][0][0] == 0x7FFF;
        do_search[1] = b != p1 && fenc->lowres_mvs[1][p1-b-1][0][0] == 0x7FFF;
        if( do_search[0] )
        {
            if( h->param.analyse.i_weighted_pred && b == p1 )
            {
                ;
                x264_weights_analyse( h, fenc, frames[p0], 1 );
                w = fenc->weight[0];
            }
            fenc->lowres_mvs[0][b-p0-1][0][0] = 0;
        }
        if( do_search[1] ) fenc->lowres_mvs[1][p1-b-1][0][0] = 0;

        if( p1 != p0 )
            dist_scale_factor = ( ((b-p0) << 8) + ((p1-p0) >> 1) ) / (p1-p0);

        int output_buf_size = h->mb.i_mb_height + (4 + 32) * h->param.i_lookahead_threads;
        int *output_inter[16 +1];
        int *output_intra[16 +1];
        output_inter[0] = h->scratch_buffer2;
        output_intra[0] = output_inter[0] + output_buf_size;

        if( h->param.b_opencl )
        {
        }
        else
        {
            if( h->param.i_lookahead_threads > 1 )
            {
                x264_slicetype_slice_t s[16];
            }
            else
            {
                h->i_threadslice_start = 0;
                h->i_threadslice_end = h->mb.i_mb_height;
                memset( output_inter[0], 0, (output_buf_size - 32) * sizeof(int) );
                memset( output_intra[0], 0, (output_buf_size - 32) * sizeof(int) );
                output_inter[0][3] = output_intra[0][3] = h->mb.i_mb_height;
                x264_slicetype_slice_t s = (x264_slicetype_slice_t){ h, a, frames, p0, p1, b, dist_scale_factor, do_search, w,
                    output_inter[0], output_intra[0] };
                x264_slicetype_slice_cost( &s );
            }


            if( b == p1 )
                fenc->i_intra_mbs[b-p0] = 0;
            if( !fenc->b_intra_calculated )
            {
                fenc->i_cost_est[0][0] = 0;
                fenc->i_cost_est_aq[0][0] = 0;
            }
            fenc->i_cost_est[b-p0][p1-b] = 0;
            fenc->i_cost_est_aq[b-p0][p1-b] = 0;

            int *row_satd_inter = fenc->i_row_satds[b-p0][p1-b];
            int *row_satd_intra = fenc->i_row_satds[0][0];
            for( int i = 0; i < h->param.i_lookahead_threads; i++ )
            {
                if( b == p1 )
                    fenc->i_intra_mbs[b-p0] += output_inter[i][2];
                if( !fenc->b_intra_calculated )
                {
                    fenc->i_cost_est[0][0] += output_intra[i][0];
                    fenc->i_cost_est_aq[0][0] += output_intra[i][1];
                }

                fenc->i_cost_est[b-p0][p1-b] += output_inter[i][0];
                fenc->i_cost_est_aq[b-p0][p1-b] += output_inter[i][1];

                if( h->param.rc.i_vbv_buffer_size )
                {
                    int row_count = output_inter[i][3];
                    memcpy( row_satd_inter, output_inter[i] + 4, row_count * sizeof(int) );
                    if( !fenc->b_intra_calculated )
                        memcpy( row_satd_intra, output_intra[i] + 4, row_count * sizeof(int) );
                    row_satd_inter += row_count;
                    row_satd_intra += row_count;
                }
            }

            i_score = fenc->i_cost_est[b-p0][p1-b];
            if( b != p1 )
                i_score = (uint64_t)i_score * 100 / (120 + h->param.i_bframe_bias);
            else
                fenc->b_intra_calculated = 1;

            fenc->i_cost_est[b-p0][p1-b] = i_score;
            ;
        }
    }

    if( b_intra_penalty )
    {

        int nmb = (h->mb.i_mb_width > 2 && h->mb.i_mb_height > 2 ? (h->mb.i_mb_width - 2) * (h->mb.i_mb_height - 2) : h->mb.i_mb_width * h->mb.i_mb_height);
        i_score += (uint64_t)i_score * fenc->i_intra_mbs[b-p0] / (nmb * 8);
    }
    return i_score;
}



 int x264_slicetype_frame_cost_recalculate( x264_t *h, x264_frame_t **frames, int p0, int p1, int b )
{
    int i_score = 0;
    int *row_satd = frames[b]->i_row_satds[b-p0][p1-b];
    float *qp_offset = ((frames[b]->i_type)==0x0005 || (frames[b]->i_type)==0x0004) ? frames[b]->f_qp_offset_aq : frames[b]->f_qp_offset;
    ;
    for( h->mb.i_mb_y = h->mb.i_mb_height - 1; h->mb.i_mb_y >= 0; h->mb.i_mb_y-- )
    {
        row_satd[ h->mb.i_mb_y ] = 0;
        for( h->mb.i_mb_x = h->mb.i_mb_width - 1; h->mb.i_mb_x >= 0; h->mb.i_mb_x-- )
        {
            int i_mb_xy = h->mb.i_mb_x + h->mb.i_mb_y*h->mb.i_mb_stride;
            int i_mb_cost = frames[b]->lowres_costs[b-p0][p1-b][i_mb_xy] & ((1<<14)-1);
            float qp_adj = qp_offset[i_mb_xy];
            i_mb_cost = (i_mb_cost * x264_exp2fix8(qp_adj) + 128) >> 8;
            row_satd[ h->mb.i_mb_y ] += i_mb_cost;
            if( (h->mb.i_mb_y > 0 && h->mb.i_mb_y < h->mb.i_mb_height - 1 &&
                 h->mb.i_mb_x > 0 && h->mb.i_mb_x < h->mb.i_mb_width - 1) ||
                 h->mb.i_mb_width <= 2 || h->mb.i_mb_height <= 2 )
            {
                i_score += i_mb_cost;
            }
        }
    }
    return i_score;
}




 void x264_macroblock_tree_finish( x264_t *h, x264_frame_t *frame, float average_duration, int ref0_distance )
{
    int fps_factor = round( CLIP_DURATION(average_duration) / CLIP_DURATION(frame->f_duration) * 256 / 0.5f );
    float weightdelta = 0.0;
    if( ref0_distance && frame->f_weighted_cost_delta[ref0_distance-1] > 0 )
        weightdelta = (1.0 - frame->f_weighted_cost_delta[ref0_distance-1]);



    float strength = 5.0f * (1.0f - h->param.rc.f_qcompress);
    for( int mb_index = 0; mb_index < h->mb.i_mb_count; mb_index++ )
    {
        int intra_cost = (frame->i_intra_cost[mb_index] * frame->i_inv_qscale_factor[mb_index] + 128) >> 8;
        if( intra_cost )
        {
            int propagate_cost = (frame->i_propagate_cost[mb_index] * fps_factor + 128) >> 8;
            float log2_ratio = x264_log2(intra_cost + propagate_cost) - x264_log2(intra_cost) + weightdelta;
            frame->f_qp_offset[mb_index] = frame->f_qp_offset_aq[mb_index] - strength * log2_ratio;
        }
    }
}

 void x264_macroblock_tree_propagate( x264_t *h, x264_frame_t **frames, float average_duration, int p0, int p1, int b, int referenced )
{
    uint16_t *ref_costs[2] = {frames[p0]->i_propagate_cost,frames[p1]->i_propagate_cost};
    int dist_scale_factor = ( ((b-p0) << 8) + ((p1-p0) >> 1) ) / (p1-p0);
    int i_bipred_weight = h->param.analyse.b_weighted_bipred ? 64 - (dist_scale_factor>>2) : 32;
    int16_t (*mvs[2])[2] = { frames[b]->lowres_mvs[0][b-p0-1], frames[b]->lowres_mvs[1][p1-b-1] };
    int bipred_weights[2] = {i_bipred_weight, 64 - i_bipred_weight};
    int16_t *buf = h->scratch_buffer;
    uint16_t *propagate_cost = frames[b]->i_propagate_cost;
    uint16_t *lowres_costs = frames[b]->lowres_costs[b-p0][p1-b];

    ;
    float fps_factor = CLIP_DURATION(frames[b]->f_duration) / (CLIP_DURATION(average_duration) * 256.0f) * 0.5f;


    if( !referenced )
        memset( frames[b]->i_propagate_cost, 0, h->mb.i_mb_width * sizeof(uint16_t) );

    for( h->mb.i_mb_y = 0; h->mb.i_mb_y < h->mb.i_mb_height; h->mb.i_mb_y++ )
    {
        int mb_index = h->mb.i_mb_y*h->mb.i_mb_stride;
        h->mc.mbtree_propagate_cost( buf, propagate_cost,
            frames[b]->i_intra_cost+mb_index, lowres_costs+mb_index,
            frames[b]->i_inv_qscale_factor+mb_index, &fps_factor, h->mb.i_mb_width );
        if( referenced )
            propagate_cost += h->mb.i_mb_width;

        h->mc.mbtree_propagate_list( h, ref_costs[0], &mvs[0][mb_index], buf, &lowres_costs[mb_index],
                                     bipred_weights[0], h->mb.i_mb_y, h->mb.i_mb_width, 0 );
        if( b != p1 )
        {
            h->mc.mbtree_propagate_list( h, ref_costs[1], &mvs[1][mb_index], buf, &lowres_costs[mb_index],
                                         bipred_weights[1], h->mb.i_mb_y, h->mb.i_mb_width, 1 );
        }
    }

    if( h->param.rc.i_vbv_buffer_size && h->param.rc.i_lookahead && referenced )
        x264_macroblock_tree_finish( h, frames[b], average_duration, b == p1 ? b - p0 : 0 );
}

 void x264_macroblock_tree( x264_t *h, x264_mb_analysis_t *a, x264_frame_t **frames, int num_frames, int b_intra )
{
    int idx = !b_intra;
    int last_nonb, cur_nonb = 1;
    int bframes = 0;

    ;
    float total_duration = 0.0;
    for( int j = 0; j <= num_frames; j++ )
        total_duration += frames[j]->f_duration;
    float average_duration = total_duration / (num_frames + 1);

    int i = num_frames;

    if( b_intra )
        x264_slicetype_frame_cost( h, a, frames, 0, 0, 0, 0 );

    while( i > 0 && frames[i]->i_type == 0x0005 )
        i--;
    last_nonb = i;




    if( !h->param.rc.i_lookahead )
    {
        if( b_intra )
        {
            memset( frames[0]->i_propagate_cost, 0, h->mb.i_mb_count * sizeof(uint16_t) );
            memcpy( frames[0]->f_qp_offset, frames[0]->f_qp_offset_aq, h->mb.i_mb_count * sizeof(float) );
            return;
        }
        do{ uint16_t* t = frames[last_nonb]->i_propagate_cost; frames[last_nonb]->i_propagate_cost = frames[0]->i_propagate_cost; frames[0]->i_propagate_cost = t; } while(0);
        memset( frames[0]->i_propagate_cost, 0, h->mb.i_mb_count * sizeof(uint16_t) );
    }
    else
    {
        if( last_nonb < idx )
            return;
        memset( frames[last_nonb]->i_propagate_cost, 0, h->mb.i_mb_count * sizeof(uint16_t) );
    }

    while( i-- > idx )
    {
        cur_nonb = i;
        while( frames[cur_nonb]->i_type == 0x0005 && cur_nonb > 0 )
            cur_nonb--;
        if( cur_nonb < idx )
            break;
        x264_slicetype_frame_cost( h, a, frames, cur_nonb, last_nonb, last_nonb, 0 );
        memset( frames[cur_nonb]->i_propagate_cost, 0, h->mb.i_mb_count * sizeof(uint16_t) );
        bframes = last_nonb - cur_nonb - 1;
        if( h->param.i_bframe_pyramid && bframes > 1 )
        {
            int middle = (bframes + 1)/2 + cur_nonb;
            x264_slicetype_frame_cost( h, a, frames, cur_nonb, last_nonb, middle, 0 );
            memset( frames[middle]->i_propagate_cost, 0, h->mb.i_mb_count * sizeof(uint16_t) );
            while( i > cur_nonb )
            {
                int p0 = i > middle ? middle : cur_nonb;
                int p1 = i < middle ? middle : last_nonb;
                if( i != middle )
                {
                    x264_slicetype_frame_cost( h, a, frames, p0, p1, i, 0 );
                    x264_macroblock_tree_propagate( h, frames, average_duration, p0, p1, i, 0 );
                }
                i--;
            }
            x264_macroblock_tree_propagate( h, frames, average_duration, cur_nonb, last_nonb, middle, 1 );
        }
        else
        {
            while( i > cur_nonb )
            {
                x264_slicetype_frame_cost( h, a, frames, cur_nonb, last_nonb, i, 0 );
                x264_macroblock_tree_propagate( h, frames, average_duration, cur_nonb, last_nonb, i, 0 );
                i--;
            }
        }
        x264_macroblock_tree_propagate( h, frames, average_duration, cur_nonb, last_nonb, last_nonb, 1 );
        last_nonb = cur_nonb;
    }

    if( !h->param.rc.i_lookahead )
    {
        x264_slicetype_frame_cost( h, a, frames, 0, last_nonb, last_nonb, 0 );
        x264_macroblock_tree_propagate( h, frames, average_duration, 0, last_nonb, last_nonb, 1 );
        do{ uint16_t* t = frames[last_nonb]->i_propagate_cost; frames[last_nonb]->i_propagate_cost = frames[0]->i_propagate_cost; frames[0]->i_propagate_cost = t; } while(0);
    }

    x264_macroblock_tree_finish( h, frames[last_nonb], average_duration, last_nonb );
    if( h->param.i_bframe_pyramid && bframes > 1 && !h->param.rc.i_vbv_buffer_size )
        x264_macroblock_tree_finish( h, frames[last_nonb+(bframes+1)/2], average_duration, 0 );
}

 int x264_vbv_frame_cost( x264_t *h, x264_mb_analysis_t *a, x264_frame_t **frames, int p0, int p1, int b )
{
    int cost = x264_slicetype_frame_cost( h, a, frames, p0, p1, b, 0 );
    if( h->param.rc.i_aq_mode )
    {
        if( h->param.rc.b_mb_tree )
            return x264_slicetype_frame_cost_recalculate( h, frames, p0, p1, b );
        else
            return frames[b]->i_cost_est_aq[b-p0][p1-b];
    }
    return cost;
}

 void x264_calculate_durations( x264_t *h, x264_frame_t *cur_frame, x264_frame_t *prev_frame, int64_t *i_cpb_delay, int64_t *i_coded_fields )
{
    cur_frame->i_cpb_delay = *i_cpb_delay;
    cur_frame->i_dpb_output_delay = cur_frame->i_field_cnt - *i_coded_fields;


    cur_frame->i_dpb_output_delay += h->sps->vui.i_num_reorder_frames*2;


    if( cur_frame->i_dpb_output_delay < 0 )
    {
        cur_frame->i_cpb_delay += cur_frame->i_dpb_output_delay;
        cur_frame->i_dpb_output_delay = 0;
        if( prev_frame )
            prev_frame->i_cpb_duration += cur_frame->i_dpb_output_delay;
    }


    if( cur_frame->b_keyframe && !h->param.b_intra_refresh )
        *i_cpb_delay = 0;

    *i_cpb_delay += cur_frame->i_duration;
    *i_coded_fields += cur_frame->i_duration;
    cur_frame->i_cpb_duration = cur_frame->i_duration;
}

 void x264_vbv_lookahead( x264_t *h, x264_mb_analysis_t *a, x264_frame_t **frames, int num_frames, int keyframe )
{
    int last_nonb = 0, cur_nonb = 1, idx = 0;
    x264_frame_t *prev_frame = ((void *)0);
    int prev_frame_idx = 0;
    while( cur_nonb < num_frames && frames[cur_nonb]->i_type == 0x0005 )
        cur_nonb++;
    int next_nonb = keyframe ? last_nonb : cur_nonb;

    if( frames[cur_nonb]->i_coded_fields_lookahead >= 0 )
    {
        h->i_coded_fields_lookahead = frames[cur_nonb]->i_coded_fields_lookahead;
        h->i_cpb_delay_lookahead = frames[cur_nonb]->i_cpb_delay_lookahead;
    }

    while( cur_nonb < num_frames )
    {

        if( next_nonb != cur_nonb )
        {
            int p0 = ((frames[cur_nonb]->i_type)==0x0002 || (frames[cur_nonb]->i_type)==0x0001) ? cur_nonb : last_nonb;
            frames[next_nonb]->i_planned_satd[idx] = x264_vbv_frame_cost( h, a, frames, p0, cur_nonb, cur_nonb );
            frames[next_nonb]->i_planned_type[idx] = frames[cur_nonb]->i_type;
            frames[cur_nonb]->i_coded_fields_lookahead = h->i_coded_fields_lookahead;
            frames[cur_nonb]->i_cpb_delay_lookahead = h->i_cpb_delay_lookahead;
            x264_calculate_durations( h, frames[cur_nonb], prev_frame, &h->i_cpb_delay_lookahead, &h->i_coded_fields_lookahead );
            if( prev_frame )
            {
                frames[next_nonb]->f_planned_cpb_duration[prev_frame_idx] = (double)prev_frame->i_cpb_duration *
                                                                            h->sps->vui.i_num_units_in_tick / h->sps->vui.i_time_scale;
            }
            frames[next_nonb]->f_planned_cpb_duration[idx] = (double)frames[cur_nonb]->i_cpb_duration *
                                                             h->sps->vui.i_num_units_in_tick / h->sps->vui.i_time_scale;
            prev_frame = frames[cur_nonb];
            prev_frame_idx = idx;
            idx++;
        }

        for( int i = last_nonb+1; i < cur_nonb; i++, idx++ )
        {
            frames[next_nonb]->i_planned_satd[idx] = x264_vbv_frame_cost( h, a, frames, last_nonb, cur_nonb, i );
            frames[next_nonb]->i_planned_type[idx] = 0x0005;
            frames[i]->i_coded_fields_lookahead = h->i_coded_fields_lookahead;
            frames[i]->i_cpb_delay_lookahead = h->i_cpb_delay_lookahead;
            x264_calculate_durations( h, frames[i], prev_frame, &h->i_cpb_delay_lookahead, &h->i_coded_fields_lookahead );
            if( prev_frame )
            {
                frames[next_nonb]->f_planned_cpb_duration[prev_frame_idx] = (double)prev_frame->i_cpb_duration *
                                                                            h->sps->vui.i_num_units_in_tick / h->sps->vui.i_time_scale;
            }
            frames[next_nonb]->f_planned_cpb_duration[idx] = (double)frames[i]->i_cpb_duration *
                                                             h->sps->vui.i_num_units_in_tick / h->sps->vui.i_time_scale;
            prev_frame = frames[i];
            prev_frame_idx = idx;
        }
        last_nonb = cur_nonb;
        cur_nonb++;
        while( cur_nonb <= num_frames && frames[cur_nonb]->i_type == 0x0005 )
            cur_nonb++;
    }
    frames[next_nonb]->i_planned_type[idx] = 0x0000;
}

 int x264_slicetype_path_cost( x264_t *h, x264_mb_analysis_t *a, x264_frame_t **frames, char *path, int threshold )
{
    int loc = 1;
    int cost = 0;
    int cur_p = 0;
    path--;
    while( path[loc] )
    {
        int next_p = loc;

        while( path[next_p] != 'P' )
            next_p++;


        cost += x264_slicetype_frame_cost( h, a, frames, cur_p, next_p, next_p, 0 );

        if( cost > threshold )
            break;

        if( h->param.i_bframe_pyramid && next_p - cur_p > 2 )
        {
            int middle = cur_p + (next_p - cur_p)/2;
            cost += x264_slicetype_frame_cost( h, a, frames, cur_p, next_p, middle, 0 );
            for( int next_b = loc; next_b < middle && cost < threshold; next_b++ )
                cost += x264_slicetype_frame_cost( h, a, frames, cur_p, middle, next_b, 0 );
            for( int next_b = middle+1; next_b < next_p && cost < threshold; next_b++ )
                cost += x264_slicetype_frame_cost( h, a, frames, middle, next_p, next_b, 0 );
        }
        else
            for( int next_b = loc; next_b < next_p && cost < threshold; next_b++ )
                cost += x264_slicetype_frame_cost( h, a, frames, cur_p, next_p, next_b, 0 );

        loc = next_p + 1;
        cur_p = next_p;
    }
    return cost;
}





 void x264_slicetype_path( x264_t *h, x264_mb_analysis_t *a, x264_frame_t **frames, int length, char (*best_paths)[250 +1] )
{
    char paths[2][250 +1];
    int num_paths = ( (h->param.i_bframe+1)<(length) ? (h->param.i_bframe+1) : (length) );
    int best_cost = (1<<28);
    int idx = 0;


    for( int path = 0; path < num_paths; path++ )
    {

        int len = length - (path + 1);
        memcpy( paths[idx], best_paths[len % (16 +1)], len );
        memset( paths[idx]+len, 'B', path );
        strcpy( paths[idx]+len+path, "P" );


        int cost = x264_slicetype_path_cost( h, a, frames, paths[idx], best_cost );
        if( cost < best_cost )
        {
            best_cost = cost;
            idx ^= 1;
        }
    }


    memcpy( best_paths[length % (16 +1)], paths[idx^1], length );
}

 int scenecut_internal( x264_t *h, x264_mb_analysis_t *a, x264_frame_t **frames, int p0, int p1, int real_scenecut )
{
    x264_frame_t *frame = frames[p1];


    if( real_scenecut && h->param.i_frame_packing == 5 && (frame->i_frame&1) )
        return 0;

    x264_slicetype_frame_cost( h, a, frames, p0, p1, p1, 0 );

    int icost = frame->i_cost_est[0][0];
    int pcost = frame->i_cost_est[p1-p0][0];
    float f_bias;
    int i_gop_size = frame->i_frame - h->lookahead->i_last_keyframe;
    float f_thresh_max = h->param.i_scenecut_threshold / 100.0;

    float f_thresh_min = f_thresh_max * 0.25;
    int res;

    if( h->param.i_keyint_min == h->param.i_keyint_max )
        f_thresh_min = f_thresh_max;
    if( i_gop_size <= h->param.i_keyint_min / 4 || h->param.b_intra_refresh )
        f_bias = f_thresh_min / 4;
    else if( i_gop_size <= h->param.i_keyint_min )
        f_bias = f_thresh_min * i_gop_size / h->param.i_keyint_min;
    else
    {
        f_bias = f_thresh_min
                 + ( f_thresh_max - f_thresh_min )
                 * ( i_gop_size - h->param.i_keyint_min )
                 / ( h->param.i_keyint_max - h->param.i_keyint_min );
    }

    res = pcost >= (1.0 - f_bias) * icost;
    if( res && real_scenecut )
    {
        int imb = frame->i_intra_mbs[p1-p0];
        int pmb = (h->mb.i_mb_width > 2 && h->mb.i_mb_height > 2 ? (h->mb.i_mb_width - 2) * (h->mb.i_mb_height - 2) : h->mb.i_mb_width * h->mb.i_mb_height) - imb;
        x264_log( h, 3, "scene cut at %d Icost:%d Pcost:%d ratio:%.4f bias:%.4f gop:%d (imb:%d pmb:%d)\n",
                  frame->i_frame,
                  icost, pcost, 1. - (double)pcost / icost,
                  f_bias, i_gop_size, imb, pmb );
    }
    return res;
}

 int scenecut( x264_t *h, x264_mb_analysis_t *a, x264_frame_t **frames, int p0, int p1, int real_scenecut, int num_frames, int i_max_search )
{

    if( real_scenecut && h->param.i_bframe )
    {
        int origmaxp1 = p0 + 1;

        if( h->param.i_bframe_adaptive == 2 )

            origmaxp1 += h->param.i_bframe;
        else
            origmaxp1++;
        int maxp1 = ( (origmaxp1)<(num_frames) ? (origmaxp1) : (num_frames) );




        for( int curp1 = p1; curp1 <= maxp1; curp1++ )
            if( !scenecut_internal( h, a, frames, p0, curp1, 0 ) )

                for( int i = curp1; i > p0; i-- )
                    frames[i]->b_scenecut = 0;






        for( int curp0 = p0; curp0 <= maxp1; curp0++ )
            if( origmaxp1 > i_max_search || (curp0 < maxp1 && scenecut_internal( h, a, frames, curp0, maxp1, 0 )) )

                    frames[curp0]->b_scenecut = 0;
    }


    if( !frames[p1]->b_scenecut )
        return 0;
    return scenecut_internal( h, a, frames, p0, p1, real_scenecut );
}

void x264_slicetype_analyse( x264_t *h, int intra_minigop )
{
    x264_mb_analysis_t a;
    x264_frame_t *frames[250 +3] = { ((void *)0), };
    int num_frames, orig_num_frames, keyint_limit, framecnt;
    int i_mb_count = (h->mb.i_mb_width > 2 && h->mb.i_mb_height > 2 ? (h->mb.i_mb_width - 2) * (h->mb.i_mb_height - 2) : h->mb.i_mb_width * h->mb.i_mb_height);
    int cost1p0, cost2p0, cost1b1, cost2p1;
    int i_max_search = ( (h->lookahead->next.i_size)<(250) ? (h->lookahead->next.i_size) : (250) );
    int vbv_lookahead = h->param.rc.i_vbv_buffer_size && h->param.rc.i_lookahead;




    if( h->param.b_deterministic )
        i_max_search = ( (i_max_search)<(h->lookahead->i_slicetype_length + 1 - intra_minigop) ? (i_max_search) : (h->lookahead->i_slicetype_length + 1 - intra_minigop) );
    int keyframe = !!intra_minigop;

    ((h->frames.b_have_lowres) ? (void) (0) : __assert_fail ("h->frames.b_have_lowres", "DonorSourceCode/slicetype.h", 1416, __PRETTY_FUNCTION__));

    if( !h->lookahead->last_nonb )
        return;
    frames[0] = h->lookahead->last_nonb;
    for( framecnt = 0; framecnt < i_max_search && h->lookahead->next.list[framecnt]->i_type == 0x0000; framecnt++ )
        frames[framecnt+1] = h->lookahead->next.list[framecnt];

    x264_lowres_context_init( h, &a );

    if( !framecnt )
    {
        if( h->param.rc.b_mb_tree )
            x264_macroblock_tree( h, &a, frames, 0, keyframe );
        return;
    }

    keyint_limit = h->param.i_keyint_max - frames[0]->i_frame + h->lookahead->i_last_keyframe - 1;
    orig_num_frames = num_frames = h->param.b_intra_refresh ? framecnt : ( (framecnt)<(keyint_limit) ? (framecnt) : (keyint_limit) );





    if( (h->param.analyse.b_psy && h->param.rc.b_mb_tree) || vbv_lookahead )
        num_frames = framecnt;
    else if( h->param.b_open_gop && num_frames < framecnt )
        num_frames++;
    else if( num_frames == 0 )
    {
        frames[1]->i_type = 0x0002;
        return;
    }

    int num_bframes = 0;
    int num_analysed_frames = num_frames;
    int reset_start;
    if( h->param.i_scenecut_threshold && scenecut( h, &a, frames, 0, 1, 1, orig_num_frames, i_max_search ) )
    {
        frames[1]->i_type = 0x0002;
        return;
    }

    x264_opencl_slicetype_prep( h, frames, num_frames, a.i_lambda );

    if( h->param.i_bframe )
    {

    }
    else
    {
        for( int j = 1; j <= num_frames; j++ )
            frames[j]->i_type = 0x0003;
        reset_start = !keyframe + 1;
        num_bframes = 0;
    }



    if( h->param.rc.b_mb_tree )
        x264_macroblock_tree( h, &a, frames, ( (num_frames)<(h->param.i_keyint_max) ? (num_frames) : (h->param.i_keyint_max) ), keyframe );


    if( !h->param.b_intra_refresh )
        for( int i = keyint_limit+1; i <= num_frames; i += h->param.i_keyint_max )
        {
            frames[i]->i_type = 0x0002;
            reset_start = ( (reset_start)<(i+1) ? (reset_start) : (i+1) );
            if( h->param.b_open_gop && h->param.b_bluray_compat )
                while( ((frames[i-1]->i_type)==0x0005 || (frames[i-1]->i_type)==0x0004) )
                    i--;
        }

    if( vbv_lookahead )
        x264_vbv_lookahead( h, &a, frames, num_frames, keyframe );


    for( int j = reset_start; j <= num_frames; j++ )
        frames[j]->i_type = 0x0000;
    x264_opencl_slicetype_end( h );
}

void x264_slicetype_decide( x264_t *h )
{
    x264_frame_t *frames[16 +2];
    x264_frame_t *frm;
    int bframes;
    int brefs;

    if( !h->lookahead->next.i_size )
        return;

    int lookahead_size = h->lookahead->next.i_size;

    for( int i = 0; i < h->lookahead->next.i_size; i++ )
    {
        if( h->param.b_vfr_input )
        {
            if( lookahead_size-- > 1 )
                h->lookahead->next.list[i]->i_duration = 2 * (h->lookahead->next.list[i+1]->i_pts - h->lookahead->next.list[i]->i_pts);
            else
                h->lookahead->next.list[i]->i_duration = h->i_prev_duration;
        }
        else
            h->lookahead->next.list[i]->i_duration = delta_tfi_divisor[h->lookahead->next.list[i]->i_pic_struct];
        h->i_prev_duration = h->lookahead->next.list[i]->i_duration;
        h->lookahead->next.list[i]->f_duration = (double)h->lookahead->next.list[i]->i_duration
                                               * h->sps->vui.i_num_units_in_tick
                                               / h->sps->vui.i_time_scale;

        if( h->lookahead->next.list[i]->i_frame > h->i_disp_fields_last_frame && lookahead_size > 0 )
        {
            h->lookahead->next.list[i]->i_field_cnt = h->i_disp_fields;
            h->i_disp_fields += h->lookahead->next.list[i]->i_duration;
            h->i_disp_fields_last_frame = h->lookahead->next.list[i]->i_frame;
        }
        else if( lookahead_size == 0 )
        {
            h->lookahead->next.list[i]->i_field_cnt = h->i_disp_fields;
            h->lookahead->next.list[i]->i_duration = h->i_prev_duration;
        }
    }

    if( h->param.rc.b_stat_read )
    {

        for( int i = 0; i < h->lookahead->next.i_size; i++ )
            h->lookahead->next.list[i]->i_type =
                x264_ratecontrol_slice_type( h, h->lookahead->next.list[i]->i_frame );
    }
    else if( (h->param.i_bframe && h->param.i_bframe_adaptive)
             || h->param.i_scenecut_threshold
             || h->param.rc.b_mb_tree
             || (h->param.rc.i_vbv_buffer_size && h->param.rc.i_lookahead) )
        x264_slicetype_analyse( h, 0 );

    for( bframes = 0, brefs = 0;; bframes++ )
    {
        frm = h->lookahead->next.list[bframes];
        if( frm->i_type == 0x0004 && h->param.i_bframe_pyramid < 2 &&
            brefs == h->param.i_bframe_pyramid )
        {
            frm->i_type = 0x0005;
            x264_log( h, 1, "B-ref at frame %d incompatible with B-pyramid %s \n",
                      frm->i_frame, x264_b_pyramid_names[h->param.i_bframe_pyramid] );
        }


        else if( frm->i_type == 0x0004 && h->param.i_bframe_pyramid == 2 &&
            brefs && h->param.i_frame_reference <= (brefs+3) )
        {
            frm->i_type = 0x0005;
            x264_log( h, 1, "B-ref at frame %d incompatible with B-pyramid %s and %d reference frames\n",
                      frm->i_frame, x264_b_pyramid_names[h->param.i_bframe_pyramid], h->param.i_frame_reference );
        }

        if( frm->i_type == 0x0006 )
            frm->i_type = h->param.b_open_gop ? 0x0002 : 0x0001;


        if( (!h->param.b_intra_refresh || frm->i_frame == 0) && frm->i_frame - h->lookahead->i_last_keyframe >= h->param.i_keyint_max )
        {
            if( frm->i_type == 0x0000 || frm->i_type == 0x0002 )
                frm->i_type = h->param.b_open_gop && h->lookahead->i_last_keyframe >= 0 ? 0x0002 : 0x0001;
            int warn = frm->i_type != 0x0001;
            if( warn && h->param.b_open_gop )
                warn &= frm->i_type != 0x0002;
            if( warn )
            {
                x264_log( h, 1, "specified frame type (%d) at %d is not compatible with keyframe interval\n", frm->i_type, frm->i_frame );
                frm->i_type = h->param.b_open_gop && h->lookahead->i_last_keyframe >= 0 ? 0x0002 : 0x0001;
            }
        }
        if( frm->i_type == 0x0002 && frm->i_frame - h->lookahead->i_last_keyframe >= h->param.i_keyint_min )
        {
            if( h->param.b_open_gop )
            {
                h->lookahead->i_last_keyframe = frm->i_frame;
                if( h->param.b_bluray_compat )
                    h->lookahead->i_last_keyframe -= bframes;
                frm->b_keyframe = 1;
            }
            else
                frm->i_type = 0x0001;
        }
        if( frm->i_type == 0x0001 )
        {

            h->lookahead->i_last_keyframe = frm->i_frame;
            frm->b_keyframe = 1;
            if( bframes > 0 )
            {
                bframes--;
                h->lookahead->next.list[bframes]->i_type = 0x0003;
            }
        }

        if( bframes == h->param.i_bframe ||
            !h->lookahead->next.list[bframes+1] )
        {
            if( ((frm->i_type)==0x0005 || (frm->i_type)==0x0004) )
                x264_log( h, 1, "specified frame type is not compatible with max B-frames\n" );
            if( frm->i_type == 0x0000
                || ((frm->i_type)==0x0005 || (frm->i_type)==0x0004) )
                frm->i_type = 0x0003;
        }

        if( frm->i_type == 0x0004 )
            brefs++;

        if( frm->i_type == 0x0000 )
            frm->i_type = 0x0005;

        else if( !((frm->i_type)==0x0005 || (frm->i_type)==0x0004) ) break;
    }

    if( bframes )
        h->lookahead->next.list[bframes-1]->b_last_minigop_bframe = 1;
    h->lookahead->next.list[bframes]->i_bframes = bframes;


    if( h->param.i_bframe_pyramid && bframes > 1 && !brefs )
    {
        h->lookahead->next.list[bframes/2]->i_type = 0x0004;
        brefs++;
    }


    if( h->param.rc.i_rc_method != 0 )
    {
        x264_mb_analysis_t a;
        int p0, p1, b;
        p1 = b = bframes + 1;

        x264_lowres_context_init( h, &a );

        frames[0] = h->lookahead->last_nonb;
        memcpy( &frames[1], h->lookahead->next.list, (bframes+1) * sizeof(x264_frame_t*) );
        if( ((h->lookahead->next.list[bframes]->i_type)==0x0002 || (h->lookahead->next.list[bframes]->i_type)==0x0001) )
            p0 = bframes + 1;
        else
            p0 = 0;

        x264_slicetype_frame_cost( h, &a, frames, p0, p1, b, 0 );

        if( (p0 != p1 || bframes) && h->param.rc.i_vbv_buffer_size )
        {

            x264_slicetype_frame_cost( h, &a, frames, b, b, b, 0 );


            p0 = 0;
            for( b = 1; b <= bframes; b++ )
            {
                if( frames[b]->i_type == 0x0005 )
                    for( p1 = b; frames[p1]->i_type == 0x0005; )
                        p1++;
                else
                    p1 = bframes + 1;
                x264_slicetype_frame_cost( h, &a, frames, p0, p1, b, 0 );
                if( frames[b]->i_type == 0x0004 )
                    p0 = b;
            }
        }
    }


    if( !h->param.rc.b_stat_read && h->lookahead->next.list[bframes]->i_type == 0x0003
        && h->param.analyse.i_weighted_pred >= 1 )
    {
        ;
        x264_weights_analyse( h, h->lookahead->next.list[bframes], h->lookahead->last_nonb, 0 );
    }



    int i_coded = h->lookahead->next.list[0]->i_frame;
    if( bframes )
    {
        int idx_list[] = { brefs+1, 1 };
        for( int i = 0; i < bframes; i++ )
        {
            int idx = idx_list[h->lookahead->next.list[i]->i_type == 0x0004]++;
            frames[idx] = h->lookahead->next.list[i];
            frames[idx]->i_reordered_pts = h->lookahead->next.list[idx]->i_pts;
        }
        frames[0] = h->lookahead->next.list[bframes];
        frames[0]->i_reordered_pts = h->lookahead->next.list[0]->i_pts;
        memcpy( h->lookahead->next.list, frames, (bframes+1) * sizeof(x264_frame_t*) );
    }

    for( int i = 0; i <= bframes; i++ )
    {
        h->lookahead->next.list[i]->i_coded = i_coded++;
        if( i )
        {
            x264_calculate_durations( h, h->lookahead->next.list[i], h->lookahead->next.list[i-1], &h->i_cpb_delay, &h->i_coded_fields );
            h->lookahead->next.list[0]->f_planned_cpb_duration[i-1] = (double)h->lookahead->next.list[i-1]->i_cpb_duration *
                                                                      h->sps->vui.i_num_units_in_tick / h->sps->vui.i_time_scale;
        }
        else
            x264_calculate_durations( h, h->lookahead->next.list[i], ((void *)0), &h->i_cpb_delay, &h->i_coded_fields );

        h->lookahead->next.list[0]->f_planned_cpb_duration[i] = (double)h->lookahead->next.list[i]->i_cpb_duration *
                                                                h->sps->vui.i_num_units_in_tick / h->sps->vui.i_time_scale;
    }
}

int x264_rc_analyse_slice( x264_t *h )
{
    int p0 = 0, p1, b;
    int cost;
    ;

    if( ((h->fenc->i_type)==0x0002 || (h->fenc->i_type)==0x0001) )
        p1 = b = 0;
    else if( h->fenc->i_type == 0x0003 )
        p1 = b = h->fenc->i_bframes + 1;
    else
    {
        p1 = (h->fref_nearest[1]->i_poc - h->fref_nearest[0]->i_poc)/2;
        b = (h->fenc->i_poc - h->fref_nearest[0]->i_poc)/2;
    }

    x264_frame_t **frames = &h->fenc - b;


    cost = frames[b]->i_cost_est[b-p0][p1-b];
    ((cost >= 0) ? (void) (0) : __assert_fail ("cost >= 0", "DonorSourceCode/slicetype.h", 1744, __PRETTY_FUNCTION__));

    if( h->param.rc.b_mb_tree && !h->param.rc.b_stat_read )
    {
        cost = x264_slicetype_frame_cost_recalculate( h, frames, p0, p1, b );
        if( b && h->param.rc.i_vbv_buffer_size )
            x264_slicetype_frame_cost_recalculate( h, frames, b, b, b );
    }

    else if( h->param.rc.i_aq_mode )
        cost = frames[b]->i_cost_est_aq[b-p0][p1-b];

    h->fenc->i_row_satd = h->fenc->i_row_satds[b-p0][p1-b];
    h->fdec->i_row_satd = h->fdec->i_row_satds[b-p0][p1-b];
    h->fdec->i_satd = cost;
    memcpy( h->fdec->i_row_satd, h->fenc->i_row_satd, h->mb.i_mb_height * sizeof(int) );
    if( !((h->fenc->i_type)==0x0002 || (h->fenc->i_type)==0x0001) )
        memcpy( h->fdec->i_row_satds[0][0], h->fenc->i_row_satds[0][0], h->mb.i_mb_height * sizeof(int) );

    if( h->param.b_intra_refresh && h->param.rc.i_vbv_buffer_size && h->fenc->i_type == 0x0003 )
    {
        int ip_factor = 256 * h->param.rc.f_ip_factor;
        for( int y = 0; y < h->mb.i_mb_height; y++ )
        {
            int mb_xy = y * h->mb.i_mb_stride + h->fdec->i_pir_start_col;
            for( int x = h->fdec->i_pir_start_col; x <= h->fdec->i_pir_end_col; x++, mb_xy++ )
            {
                int intra_cost = (h->fenc->i_intra_cost[mb_xy] * ip_factor + 128) >> 8;
                int inter_cost = h->fenc->lowres_costs[b-p0][p1-b][mb_xy] & ((1<<14)-1);
                int diff = intra_cost - inter_cost;
                if( h->param.rc.i_aq_mode )
                    h->fdec->i_row_satd[y] += (diff * frames[b]->i_inv_qscale_factor[mb_xy] + 128) >> 8;
                else
                    h->fdec->i_row_satd[y] += diff;
                cost += diff;
            }
        }
    }

    if( 8 > 8 )
        for( int y = 0; y < h->mb.i_mb_height; y++ )
            h->fdec->i_row_satd[y] >>= (8 - 8);

    return cost >> (8 - 8);
}

#endif