
#include <signal.h>

#define _GNU_SOURCE

#include <getopt.h>

#include "common.h"

#include "x264cli.h"

#include "input.h"

#include "output.h"

#include "filters.h"

#include <libavformat/avformat.h>

#include <libavutil/pixfmt.h>

#include <libavutil/pixdesc.h>

#include <libswscale/swscale.h>
volatile int b_ctrl_c = 0;

void sigint_handler (int a) {
    b_ctrl_c = 1;
}

struct _cli_opt_t {
    int b_progress;
    int i_seek;
    hnd_t hin;
    hnd_t hout;
    FILE *qpfile;
    FILE *tcfile_out;
    double timebase_convert_multiplier;
    int i_pulldown;
};
typedef struct _cli_opt_t cli_opt_t;
cli_input_t cli_input;
cli_output_t cli_output;
cli_vid_filter_t filter;
const char * const demuxer_names [] = {"auto", "raw", "y4m", "lavf", 0};
const char * const muxer_names [] = {"auto", "raw", "mkv", "flv", 0};
const char * const pulldown_names [] = {"none", "22", "32", "64", "double", "triple", "euro", 0};
const char * const log_level_names [] = {"none", "error", "warning", "info", "debug", 0};
const char * const output_csp_names [] = {"i420", "i422", "i444", "rgb", 0};
const char * const range_names [] = {"auto", "tv", "pc", 0};
struct _cli_pulldown_t {
    int mod;
    uint8_t pattern [24];
    float fps_factor;
};
typedef struct _cli_pulldown_t cli_pulldown_t;
enum pulldown_type_e {X264_PULLDOWN_22 = 1, X264_PULLDOWN_32, X264_PULLDOWN_64, X264_PULLDOWN_DOUBLE, X264_PULLDOWN_TRIPLE, X264_PULLDOWN_EURO};
const cli_pulldown_t pulldown_values [] = {
    [X264_PULLDOWN_22] = {1, {PIC_STRUCT_TOP_BOTTOM}, 1.0},
    [X264_PULLDOWN_32] = {4, {PIC_STRUCT_TOP_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP_BOTTOM, PIC_STRUCT_TOP_BOTTOM}, 1.25},
    [X264_PULLDOWN_64] = {2, {PIC_STRUCT_DOUBLE, PIC_STRUCT_TRIPLE}, 1.0},
    [X264_PULLDOWN_DOUBLE] = {1, {PIC_STRUCT_DOUBLE}, 2.0},
    [X264_PULLDOWN_TRIPLE] = {1, {PIC_STRUCT_TRIPLE}, 3.0},
    [X264_PULLDOWN_EURO] = {24, {PIC_STRUCT_TOP_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP, PIC_STRUCT_BOTTOM_TOP_BOTTOM, PIC_STRUCT_TOP_BOTTOM, PIC_STRUCT_TOP_BOTTOM, PIC_STRUCT_TOP_BOTTOM, PIC_STRUCT_TOP_BOTTOM, PIC_STRUCT_TOP_BOTTOM, PIC_STRUCT_TOP_BOTTOM, PIC_STRUCT_TOP_BOTTOM, PIC_STRUCT_TOP_BOTTOM, PIC_STRUCT_TOP_BOTTOM, PIC_STRUCT_TOP_BOTTOM, PIC_STRUCT_TOP_BOTTOM}, 25.0 / 24.0}
};
const float pulldown_frame_duration [10] = {0.0, 1, 0.5, 0.5, 1, 1, 1.5, 1.5, 2, 3};
void help (x264_param_t * defaults, int longhelp);
int parse (int argc, char * * argv, x264_param_t * param, cli_opt_t * opt);
int encode (x264_param_t * param, cli_opt_t * opt);
int cli_log_level;

void x264_cli_log (const char *name, int i_level, const char *fmt, ... ) {
    if (i_level > cli_log_level)
        return;
    char *s_level;
    switch (i_level) {
    case 0 :
        s_level = "error";
        break;
    case 1 :
        s_level = "warning";
        break;
    case 2 :
        s_level = "info";
        break;
    case 3 :
        s_level = "debug";
        break;
    default :
        s_level = "unknown";
        break;
    }
    fprintf (stderr, "%s [%s]: ", name, s_level);
    va_list arg;
    __builtin_va_start (arg, fmt);
    vfprintf (stderr, fmt, arg);
    __builtin_va_end (arg);
}

typedef int temp_length;

int main (temp_length argc, char **argv) {
    x264_param_t param;
    cli_opt_t opt = {0};
    int ret = 0;
    if (0) {
        x264_cli_log ("x264", 0, "unable to initialize threading\n");
        return -1;
    }
    if (parse (argc, argv, &param, &opt) < 0)
        ret = -1;
    signal (2, sigint_handler);
    if (!ret)
        ret = encode (&param, &opt);
    if (filter.free)
        filter.free (opt.hin);
    else if (opt.hin)
        cli_input.close_file (opt.hin);
    if (opt.hout)
        cli_output.close_file (opt.hout, 0, 0);
    if (opt.tcfile_out)
        fclose (opt.tcfile_out);
    if (opt.qpfile)
        fclose (opt.qpfile);
    return ret;
}

typedef enum {OPT_FRAMES = 256, OPT_SEEK, OPT_QPFILE, OPT_THREAD_INPUT, OPT_QUIET, OPT_NOPROGRESS, OPT_LONGHELP, OPT_PROFILE, OPT_PRESET, OPT_TUNE, OPT_SLOWFIRSTPASS, OPT_FULLHELP, OPT_FPS, OPT_MUXER, OPT_DEMUXER, OPT_INDEX, OPT_INTERLACED, OPT_TCFILE_IN, OPT_TCFILE_OUT, OPT_TIMEBASE, OPT_PULLDOWN, OPT_LOG_LEVEL, OPT_VIDEO_FILTER, OPT_INPUT_FMT, OPT_INPUT_RES, OPT_INPUT_CSP, OPT_INPUT_DEPTH, OPT_DTS_COMPRESSION, OPT_OUTPUT_CSP, OPT_INPUT_RANGE, OPT_RANGE} OptionsOPT;
char short_options [] = "8A:B:b:f:hI:i:m:o:p:q:r:t:Vvw";
struct option long_options [] = {{"help", 0, ((void *) 0), 'h'}, {"longhelp", 0, ((void *) 0), OPT_LONGHELP}, {"fullhelp", 0, ((void *) 0), OPT_FULLHELP}, {"version", 0, ((void *) 0), 'V'}, {"profile", 1, ((void *) 0), OPT_PROFILE}, {"preset", 1, ((void *) 0), OPT_PRESET}, {"tune", 1, ((void *) 0), OPT_TUNE}, {"slow-firstpass", 0, ((void *) 0), OPT_SLOWFIRSTPASS}, {"bitrate", 1, ((void *) 0), 'B'}, {"bframes", 1, ((void *) 0), 'b'}, {"b-adapt", 1, ((void *) 0), 0}, {"no-b-adapt", 0, ((void *) 0), 0}, {"b-bias", 1, ((void *) 0), 0}, {"b-pyramid", 1, ((void *) 0), 0}, {"open-gop", 0, ((void *) 0), 0}, {"bluray-compat", 0, ((void *) 0), 0}, {"avcintra-class", 1, ((void *) 0), 0}, {"min-keyint", 1, ((void *) 0), 'i'}, {"keyint", 1, ((void *) 0), 'I'}, {"intra-refresh", 0, ((void *) 0), 0}, {"scenecut", 1, ((void *) 0), 0}, {"no-scenecut", 0, ((void *) 0), 0}, {"nf", 0, ((void *) 0), 0}, {"no-deblock", 0, ((void *) 0), 0}, {"filter", 1, ((void *) 0), 0}, {"deblock", 1, ((void *) 0), 'f'}, {"interlaced", 0, ((void *) 0), OPT_INTERLACED}, {"tff", 0, ((void *) 0), OPT_INTERLACED}, {"bff", 0, ((void *) 0), OPT_INTERLACED}, {"no-interlaced", 0, ((void *) 0), OPT_INTERLACED}, {"constrained-intra", 0, ((void *) 0), 0}, {"cabac", 0, ((void *) 0), 0}, {"no-cabac", 0, ((void *) 0), 0}, {"qp", 1, ((void *) 0), 'q'}, {"qpmin", 1, ((void *) 0), 0}, {"qpmax", 1, ((void *) 0), 0}, {"qpstep", 1, ((void *) 0), 0}, {"crf", 1, ((void *) 0), 0}, {"rc-lookahead", 1, ((void *) 0), 0}, {"ref", 1, ((void *) 0), 'r'}, {"asm", 1, ((void *) 0), 0}, {"no-asm", 0, ((void *) 0), 0}, {"opencl", 0, ((void *) 0), 1}, {"opencl-clbin", 1, ((void *) 0), 0}, {"opencl-device", 1, ((void *) 0), 0}, {"sar", 1, ((void *) 0), 0}, {"fps", 1, ((void *) 0), OPT_FPS}, {"frames", 1, ((void *) 0), OPT_FRAMES}, {"seek", 1, ((void *) 0), OPT_SEEK}, {"output", 1, ((void *) 0), 'o'}, {"muxer", 1, ((void *) 0), OPT_MUXER}, {"demuxer", 1, ((void *) 0), OPT_DEMUXER}, {"stdout", 1, ((void *) 0), OPT_MUXER}, {"stdin", 1, ((void *) 0), OPT_DEMUXER}, {"index", 1, ((void *) 0), OPT_INDEX}, {"analyse", 1, ((void *) 0), 0}, {"partitions", 1, ((void *) 0), 'A'}, {"direct", 1, ((void *) 0), 0}, {"weightb", 0, ((void *) 0), 'w'}, {"no-weightb", 0, ((void *) 0), 0}, {"weightp", 1, ((void *) 0), 0}, {"me", 1, ((void *) 0), 0}, {"merange", 1, ((void *) 0), 0}, {"mvrange", 1, ((void *) 0), 0}, {"mvrange-thread", 1, ((void *) 0), 0}, {"subme", 1, ((void *) 0), 'm'}, {"psy-rd", 1, ((void *) 0), 0}, {"no-psy", 0, ((void *) 0), 0}, {"psy", 0, ((void *) 0), 0}, {"mixed-refs", 0, ((void *) 0), 0}, {"no-mixed-refs", 0, ((void *) 0), 0}, {"no-chroma-me", 0, ((void *) 0), 0}, {"8x8dct", 0, ((void *) 0), '8'}, {"no-8x8dct", 0, ((void *) 0), 0}, {"trellis", 1, ((void *) 0), 't'}, {"fast-pskip", 0, ((void *) 0), 0}, {"no-fast-pskip", 0, ((void *) 0), 0}, {"no-dct-decimate", 0, ((void *) 0), 0}, {"aq-strength", 1, ((void *) 0), 0}, {"aq-mode", 1, ((void *) 0), 0}, {"deadzone-inter", 1, ((void *) 0), 0}, {"deadzone-intra", 1, ((void *) 0), 0}, {"level", 1, ((void *) 0), 0}, {"ratetol", 1, ((void *) 0), 0}, {"vbv-maxrate", 1, ((void *) 0), 0}, {"vbv-bufsize", 1, ((void *) 0), 0}, {"vbv-init", 1, ((void *) 0), 0}, {"crf-max", 1, ((void *) 0), 0}, {"ipratio", 1, ((void *) 0), 0}, {"pbratio", 1, ((void *) 0), 0}, {"chroma-qp-offset", 1, ((void *) 0), 0}, {"pass", 1, ((void *) 0), 'p'}, {"stats", 1, ((void *) 0), 0}, {"qcomp", 1, ((void *) 0), 0}, {"mbtree", 0, ((void *) 0), 0}, {"no-mbtree", 0, ((void *) 0), 0}, {"qblur", 1, ((void *) 0), 0}, {"cplxblur", 1, ((void *) 0), 0}, {"zones", 1, ((void *) 0), 0}, {"qpfile", 1, ((void *) 0), OPT_QPFILE}, {"threads", 1, ((void *) 0), 0}, {"lookahead-threads", 1, ((void *) 0), 0}, {"sliced-threads", 0, ((void *) 0), 0}, {"no-sliced-threads", 0, ((void *) 0), 0}, {"slice-max-size", 1, ((void *) 0), 0}, {"slice-max-mbs", 1, ((void *) 0), 0}, {"slice-min-mbs", 1, ((void *) 0), 0}, {"slices", 1, ((void *) 0), 0}, {"slices-max", 1, ((void *) 0), 0}, {"thread-input", 0, ((void *) 0), OPT_THREAD_INPUT}, {"sync-lookahead", 1, ((void *) 0), 0}, {"non-deterministic", 0, ((void *)
  0), 0}, {"cpu-independent", 0, ((void *) 0), 0}, {"psnr", 0, ((void *) 0), 0}, {"ssim", 0, ((void *) 0), 0}, {"quiet", 0, ((void *) 0), OPT_QUIET}, {"verbose", 0, ((void *) 0), 'v'}, {"log-level", 1, ((void *) 0), OPT_LOG_LEVEL}, {"no-progress", 0, ((void *) 0), OPT_NOPROGRESS}, {"dump-yuv", 1, ((void *) 0), 0}, {"sps-id", 1, ((void *) 0), 0}, {"aud", 0, ((void *) 0), 0}, {"nr", 1, ((void *) 0), 0}, {"cqm", 1, ((void *) 0), 0}, {"cqmfile", 1, ((void *) 0), 0}, {"cqm4", 1, ((void *) 0), 0}, {"cqm4i", 1, ((void *) 0), 0}, {"cqm4iy", 1, ((void *) 0), 0}, {"cqm4ic", 1, ((void *) 0), 0}, {"cqm4p", 1, ((void *) 0), 0}, {"cqm4py", 1, ((void *) 0), 0}, {"cqm4pc", 1, ((void *) 0), 0}, {"cqm8", 1, ((void *) 0), 0}, {"cqm8i", 1, ((void *) 0), 0}, {"cqm8p", 1, ((void *) 0), 0}, {"overscan", 1, ((void *) 0), 0}, {"videoformat", 1, ((void *) 0), 0}, {"range", 1, ((void *) 0), OPT_RANGE}, {"colorprim", 1, ((void *) 0), 0}, {"transfer", 1, ((void *) 0), 0}, {"colormatrix", 1, ((void *) 0), 0}, {"chromaloc", 1, ((void *) 0), 0}, {"force-cfr", 0, ((void *) 0), 0}, {"tcfile-in", 1, ((void *) 0), OPT_TCFILE_IN}, {"tcfile-out", 1, ((void *) 0), OPT_TCFILE_OUT}, {"timebase", 1, ((void *) 0), OPT_TIMEBASE}, {"pic-struct", 0, ((void *) 0), 0}, {"crop-rect", 1, ((void *) 0), 0}, {"nal-hrd", 1, ((void *) 0), 0}, {"pulldown", 1, ((void *) 0), OPT_PULLDOWN}, {"fake-interlaced", 0, ((void *) 0), 0}, {"frame-packing", 1, ((void *) 0), 0}, {"vf", 1, ((void *) 0), OPT_VIDEO_FILTER}, {"video-filter", 1, ((void *) 0), OPT_VIDEO_FILTER}, {"input-fmt", 1, ((void *) 0), OPT_INPUT_FMT}, {"input-res", 1, ((void *) 0), OPT_INPUT_RES}, {"input-csp", 1, ((void *) 0), OPT_INPUT_CSP}, {"input-depth", 1, ((void *) 0), OPT_INPUT_DEPTH}, {"dts-compress", 0, ((void *) 0), OPT_DTS_COMPRESSION}, {"output-csp", 1, ((void *) 0), OPT_OUTPUT_CSP}, {"input-range", 1, ((void *) 0), OPT_INPUT_RANGE}, {"stitchable", 0, ((void *) 0), 0}, {"filler", 0, ((void *) 0), 0}, {0, 0, 0, 0}};

int select_output (const char *muxer, char *filename, x264_param_t *param) {
    const char *ext = get_filename_extension (filename);
    if (!strcmp (filename, "-") || strcasecmp (muxer, "auto"))
        ext = muxer;
    if (!strcasecmp (ext, "mp4")) {
        x264_cli_log ("x264", 0, "not compiled with MP4 output support\n");
        return -1;
    }
    else if (!strcasecmp (ext, "mkv")) {
        printf ("not supported");
        exit (0);
    }
    else if (!strcasecmp (ext, "flv")) {
        printf ("not supported");
        exit (0);
    }
    else
        cli_output = raw_output;
    return 0;
}

int select_input (const char *demuxer, char *used_demuxer, char *filename, hnd_t *p_handle, video_info_t *info, cli_input_opt_t *opt) {
    int b_auto = !strcasecmp (demuxer, "auto");
    const char *ext = b_auto ? get_filename_extension (filename) : "";
    int b_regular = strcmp (filename, "-");
    if (!b_regular && b_auto)
        ext = "raw";
    b_regular = b_regular && x264_is_regular_file_path (filename);
    if (b_regular) {
        FILE *f = fopen (filename, "r");
        if (f) {
            b_regular = x264_is_regular_file (f);
            fclose (f);
        }
    }
    const char *module = b_auto ? ext : demuxer;
    if (!strcasecmp (module, "avs") || !strcasecmp (ext, "d2v") || !strcasecmp (ext, "dga")) {
        x264_cli_log ("x264", 0, "not compiled with AVS input support\n");
        return -1;
    }
    else if (!strcasecmp (module, "y4m")) {
        printf ("not supported");
        exit (0);
    }
    else if (!strcasecmp (module, "raw") || !strcasecmp (ext, "yuv"))
        cli_input = raw_input;
    else {
        if ((b_auto || !strcasecmp (demuxer, "lavf")) && !lavf_input.open_file (filename, p_handle, info, opt)) {
            module = "lavf";
            b_auto = 0;
            cli_input = lavf_input;
        }
        if (b_auto && !raw_input.open_file (filename, p_handle, info, opt)) {
            module = "raw";
            b_auto = 0;
            cli_input = raw_input;
        }
        if (!(*p_handle)) {
            x264_cli_log ("x264", 0, "could not open input file `%s' via any method!\n", filename);
            return -1;
        }
    }
    strcpy (used_demuxer, module);
    return 0;
}

int init_vid_filters (char *sequence, hnd_t *handle, video_info_t *info, x264_param_t *param, int output_csp) {
    x264_register_vid_filters ();
    if (x264_init_vid_filter ("source", handle, &filter, info, param, ((void *) 0)))
        return -1;
    if (x264_init_vid_filter ("resize", handle, &filter, info, param, "normcsp"))
        return -1;
    if (x264_init_vid_filter ("fix_vfr_pts", handle, &filter, info, param, ((void *) 0)))
        return -1;
    for (char *p = sequence; p && *p;) {
        int tok_len = strcspn (p, "/");
        int p_len = strlen (p);
        p[tok_len] = 0;
        int name_len = strcspn (p, ":");
        p[name_len] = 0;
        name_len += name_len != tok_len;
        if (x264_init_vid_filter (p, handle, &filter, info, param, p +name_len))
            return -1;
        p += ((tok_len + 1) < (p_len) ? (tok_len + 1) : (p_len));
    }
    if (!param->i_width && !param->i_height) {
        param->i_height = info->height;
        param->i_width = info->width;
    }
    param->i_csp = info->csp;
    int csp = info->csp & 0x00ff;
    if (output_csp == 0x0001 && (csp < 0x0001 || csp > 0x0003))
        param->i_csp = 0x0001;
    else if (output_csp == 0x0004 && (csp < 0x0004 || csp > 0x0007))
        param->i_csp = 0x0004;
    else if (output_csp == 0x0008 && (csp < 0x0008 || csp > 0x0009))
        param->i_csp = 0x0008;
    else if (output_csp == 0x000c && (csp < 0x000a || csp > 0x000c))
        param->i_csp = 0x000c;
    param->i_csp |= info->csp & 0x2000;
    if (param->vui.b_fullrange == RANGE_AUTO)
        param->vui.b_fullrange = info->fullrange;
    if (x264_init_vid_filter ("resize", handle, &filter, info, param, ((void *) 0)))
        return -1;
    char args [20];
    sprintf (args, "bit_depth=%d", x264_bit_depth);
    if (x264_init_vid_filter ("depth", handle, &filter, info, param, args))
        return -1;
    return 0;
}

int parse_enum_name (const char *arg, const char * const *names, const char **dst) {
    for (int i = 0; names[i]; i++)
        if (!strcasecmp (arg, names[i])) {
            *dst = names[i];
            return 0;
        }
    return -1;
}

int parse_enum_value (const char *arg, const char * const *names, int *dst) {
    for (int i = 0; names[i]; i++)
        if (!strcasecmp (arg, names[i])) {
            *dst = i;
            return 0;
        }
    return -1;
}

int parse (int argc, char **argv, x264_param_t *param, cli_opt_t *opt) {
    char *input_filename = ((void *) 0);
    const char *demuxer = demuxer_names[0];
    char *output_filename = ((void *) 0);
    const char *muxer = muxer_names[0];
    char *tcfile_name = ((void *) 0);
    x264_param_t defaults;
    char *profile = ((void *) 0);
    char *vid_filters = ((void *) 0);
    int b_thread_input = 0;
    int b_turbo = 1;
    int b_user_ref = 0;
    int b_user_fps = 0;
    int b_user_interlaced = 0;
    cli_input_opt_t input_opt;
    cli_output_opt_t output_opt;
    char *preset = ((void *) 0);
    char *tune = ((void *) 0);
    x264_param_default (& defaults);
    cli_log_level = defaults.i_log_level;
    memset (& input_opt, 0, sizeof (cli_input_opt_t));
    memset (& output_opt, 0, sizeof (cli_output_opt_t));
    input_opt.bit_depth = 8;
    input_opt.input_range = input_opt.output_range = param->vui.b_fullrange = RANGE_AUTO;
    int output_csp = defaults.i_csp;
    opt->b_progress = 1;
    for (optind = 0;;) {
        int c = getopt_long (argc, argv, short_options, long_options, ((void *) 0));
        if (c == -1)
            break;
        if (c == OPT_PRESET)
            preset = optarg;
        if (c == OPT_TUNE)
            tune = optarg;
        else if (c == '?')
            return -1;
    }
    if (preset && !strcasecmp (preset, "placebo"))
        b_turbo = 0;
    if (x264_param_default_preset (param, preset, tune) < 0)
        return -1;
    for (optind = 0;;) {
        int b_error = 0;
        int long_options_index = -1;
        int c = getopt_long (argc, argv, short_options, long_options, &long_options_index);
        if (c == -1) {
            break;
        }
        switch (c) {
        case 'h' :
            exit (0);
        case OPT_LONGHELP :
            exit (0);
        case OPT_FULLHELP :
            exit (0);
        case 'V' :
            exit (0);
        case OPT_FRAMES :
            param->i_frame_total = ((atoi (optarg)) > (0) ? (atoi (optarg)) : (0));
            break;
        case OPT_SEEK :
            opt->i_seek = ((atoi (optarg)) > (0) ? (atoi (optarg)) : (0));
            break;
        case 'o' :
            output_filename = optarg;
            break;
        case OPT_MUXER :
            if (parse_enum_name (optarg, muxer_names, &muxer)) {
                x264_cli_log ("x264", 0, "Unknown muxer `%s'\n", optarg);
                return -1;
            }
            break;
        case OPT_DEMUXER :
            if (parse_enum_name (optarg, demuxer_names, &demuxer)) {
                x264_cli_log ("x264", 0, "Unknown demuxer `%s'\n", optarg);
                return -1;
            }
            break;
        case OPT_INDEX :
            input_opt.index_file = optarg;
            break;
        case OPT_QPFILE :
            opt->qpfile = fopen (optarg, "rb");
            if (!opt->qpfile) {
                x264_cli_log ("x264", 0, "can't open qpfile `%s'\n", optarg);
                return -1;
            }
            if (!x264_is_regular_file (opt->qpfile)) {
                x264_cli_log ("x264", 0, "qpfile incompatible with non-regular file `%s'\n", optarg);
                fclose (opt -> qpfile);
                return -1;
            }
            break;
        case OPT_THREAD_INPUT :
            b_thread_input = 1;
            break;
        case OPT_QUIET :
            cli_log_level = param->i_log_level = (-1);
            break;
        case 'v' :
            cli_log_level = param->i_log_level = 3;
            break;
        case OPT_LOG_LEVEL :
            if (!parse_enum_value (optarg, log_level_names, &cli_log_level))
                cli_log_level += (-1);
            else
                cli_log_level = atoi (optarg);
            param->i_log_level = cli_log_level;
            break;
        case OPT_NOPROGRESS :
            opt->b_progress = 0;
            break;
        case OPT_TUNE :
        case OPT_PRESET :
            break;
        case OPT_PROFILE :
            profile = optarg;
            break;
        case OPT_SLOWFIRSTPASS :
            b_turbo = 0;
            break;
        case 'r' :
            b_user_ref = 1;
            goto generic_option;
        case OPT_FPS :
            b_user_fps = 1;
            param->b_vfr_input = 0;
            goto generic_option;
        case OPT_INTERLACED :
            b_user_interlaced = 1;
            goto generic_option;
        case OPT_TCFILE_IN :
            tcfile_name = optarg;
            break;
        case OPT_TCFILE_OUT :
            opt->tcfile_out = fopen (optarg, "wb");
            if (!opt->tcfile_out) {
                x264_cli_log ("x264", 0, "can't open `%s'\n", optarg);
                return -1;
            }
            break;
        case OPT_TIMEBASE :
            input_opt.timebase = optarg;
            break;
        case OPT_PULLDOWN :
            if (parse_enum_value (optarg, pulldown_names, &opt->i_pulldown)) {
                x264_cli_log ("x264", 0, "Unknown pulldown `%s'\n", optarg);
                return -1;
            }
            break;
        case OPT_VIDEO_FILTER :
            vid_filters = optarg;
            break;
        case OPT_INPUT_FMT :
            input_opt.format = optarg;
            break;
        case OPT_INPUT_RES :
            input_opt.resolution = optarg;
            break;
        case OPT_INPUT_CSP :
            input_opt.colorspace = optarg;
            break;
        case OPT_INPUT_DEPTH :
            input_opt.bit_depth = atoi (optarg);
            break;
        case OPT_DTS_COMPRESSION :
            output_opt.use_dts_compress = 1;
            break;
        case OPT_OUTPUT_CSP :
            if (parse_enum_value (optarg, output_csp_names, &output_csp)) {
                x264_cli_log ("x264", 0, "Unknown output csp `%s'\n", optarg);
                return -1;
            }
            const uint8_t output_csp_fix [] = {0x0001, 0x0004, 0x0008, 0x000c};
            param->i_csp = output_csp = output_csp_fix[output_csp];
            break;
        case OPT_INPUT_RANGE :
            if (parse_enum_value (optarg, range_names, &input_opt.input_range)) {
                x264_cli_log ("x264", 0, "Unknown input range `%s'\n", optarg);
                return -1;
            }
            input_opt.input_range += RANGE_AUTO;
            break;
        case OPT_RANGE :
            if (parse_enum_value (optarg, range_names, &param->vui.b_fullrange)) {
                x264_cli_log ("x264", 0, "Unknown range `%s'\n", optarg);
                return -1;
            };
            input_opt.output_range = param->vui.b_fullrange += RANGE_AUTO;
            break;
        default :
        generic_option :
            {
                if (long_options_index < 0) {
                    for (int i = 0; long_options[i].name; i++)
                        if (long_options[i].val == c) {
                            long_options_index = i;
                            break;
                        }
                    if (long_options_index < 0) {
                        return -1;
                    }
                }
                b_error |= x264_param_parse (param, long_options[long_options_index].name, optarg);
            }
        }
        if (b_error) {
            const char *name = long_options_index > 0 ? long_options[long_options_index].name : argv[optind - 2];
            x264_cli_log ("x264", 0, "invalid argument: %s = %s\n", name, optarg);
            return -1;
        }
    }
    if (b_turbo)
        x264_param_apply_fastfirstpass (param);
    if (x264_param_apply_profile (param, profile) < 0)
        return -1;
    if (optind > argc - 1 || !output_filename) {
        x264_cli_log ("x264", 0, "No %s file. Run x264 --help for a list of options.\n", optind > argc - 1 ? "input" : "output");
        return -1;
    }
    if (select_output (muxer, output_filename, param))
        return -1;
    if (cli_output.open_file (output_filename, &opt->hout, &output_opt)) {
        x264_cli_log ("x264", 0, "could not open output file `%s'\n", output_filename);
        return -1;
    }
    input_filename = argv[optind++];
    video_info_t info = {0};
    char demuxername [5];
    info.csp = param->i_csp;
    info.fps_num = param->i_fps_num;
    info.fps_den = param->i_fps_den;
    info.fullrange = input_opt.input_range == RANGE_PC;
    info.interlaced = param->b_interlaced;
    if (param->vui.i_sar_width > 0 && param->vui.i_sar_height > 0) {
        info.sar_width = param->vui.i_sar_width;
        info.sar_height = param->vui.i_sar_height;
    }
    info.tff = param->b_tff;
    info.vfr = param->b_vfr_input;
    input_opt.seek = opt->i_seek;
    input_opt.progress = opt->b_progress;
    input_opt.output_csp = output_csp;
    if (select_input (demuxer, demuxername, input_filename, &opt->hin, &info, &input_opt))
        return -1;
    if (!opt->hin && cli_input.open_file (input_filename, &opt->hin, &info, &input_opt)) {
        x264_cli_log ("x264", 0, "could not open input file `%s'\n", input_filename);
        return -1;
    }
    x264_reduce_fraction (& info.sar_width, & info.sar_height);
    x264_reduce_fraction (& info.fps_num, & info.fps_den);
    x264_cli_log (demuxername, 2, "%dx%d%c %u:%u @ %u/%u fps (%cfr)\n", info.width, info.height, info.interlaced ? 'i' : 'p', info.sar_width, info.sar_height, info.fps_num, info.fps_den, info.vfr ? 'v' : 'c');
    if (tcfile_name) {
        printf ("not supported");
        exit (0);
    }
    else if (!info.vfr && input_opt.timebase) {
        x264_cli_log ("x264", 0, "--timebase is incompatible with cfr input\n");
        return -1;
    }
    if (param->vui.i_sar_width > 0 && param->vui.i_sar_height > 0) {
        info.sar_width = param->vui.i_sar_width;
        info.sar_height = param->vui.i_sar_height;
    }
    if (b_user_fps) {
        info.fps_num = param->i_fps_num;
        info.fps_den = param->i_fps_den;
    }
    if (!info.vfr) {
        info.timebase_num = info.fps_den;
        info.timebase_den = info.fps_num;
    }
    if (!tcfile_name && input_opt.timebase) {
        uint64_t i_user_timebase_num;
        uint64_t i_user_timebase_den;
        int ret = sscanf (input_opt.timebase, "%" "l" "u" "/%" "l" "u", &i_user_timebase_num, &i_user_timebase_den);
        if (!ret) {
            x264_cli_log ("x264", 0, "invalid argument: timebase = %s\n", input_opt.timebase);
            return -1;
        }
        else if (ret == 1) {
            i_user_timebase_num = info.timebase_num;
            i_user_timebase_den = strtoul (input_opt.timebase, ((void *) 0), 10);
        }
        if (i_user_timebase_num > (4294967295U) || i_user_timebase_den > (4294967295U)) {
            x264_cli_log ("x264", 0, "timebase you specified exceeds H.264 maximum\n");
            return -1;
        }
        opt->timebase_convert_multiplier = ((double) i_user_timebase_den / info.timebase_den) * ((double) info.timebase_num / i_user_timebase_num);
        info.timebase_num = i_user_timebase_num;
        info.timebase_den = i_user_timebase_den;
        info.vfr = 1;
    }
    if (b_user_interlaced) {
        info.interlaced = param->b_interlaced;
        info.tff = param->b_tff;
    }
    if (input_opt.input_range != RANGE_AUTO)
        info.fullrange = input_opt.input_range;
    if (init_vid_filters (vid_filters, &opt->hin, &info, param, output_csp))
        return -1;
    param->b_vfr_input = info.vfr;
    param->i_fps_num = info.fps_num;
    param->i_fps_den = info.fps_den;
    param->i_timebase_num = info.timebase_num;
    param->i_timebase_den = info.timebase_den;
    param->vui.i_sar_width = info.sar_width;
    param->vui.i_sar_height = info.sar_height;
    info.num_frames = ((info.num_frames - opt->i_seek) > (0) ? (info.num_frames - opt->i_seek) : (0));
    if ((!info.num_frames || param->i_frame_total < info.num_frames) && param->i_frame_total > 0)
        info.num_frames = param->i_frame_total;
    param->i_frame_total = info.num_frames;
    if (!b_user_interlaced && info.interlaced) {
        x264_cli_log ("x264", 1, "input appears to be interlaced, but not compiled with interlaced support\n");
    }
    int csp = param->i_csp & 0x00ff;
    if (csp >= 0x000a && csp <= 0x000c) {
        if (input_opt.output_range == RANGE_AUTO)
            param->vui.b_fullrange = RANGE_PC;
        if (!param->vui.b_fullrange) {
            x264_cli_log ("x264", 0, "RGB must be PC range");
            return -1;
        }
    }
    if (!b_user_ref) {
        int mbs = (((param->i_width) + 15) >> 4) * (((param->i_height) + 15) >> 4);
        for (int i = 0; x264_levels[i].level_idc != 0; i++)
            if (param->i_level_idc == x264_levels[i].level_idc) {
                while (mbs * param->i_frame_reference > x264_levels[i].dpb && param->i_frame_reference > 1)
                    param->i_frame_reference--;
                break;
            }
    }
    return 0;
}

int encode_frame (x264_t *h, hnd_t hout, x264_picture_t *pic, int64_t *last_dts) {
    x264_picture_t pic_out;
    x264_nal_t *nal;
    int i_nal;
    int i_frame_size = 0;
    i_frame_size = x264_encoder_encode (h, &nal, &i_nal, pic, &pic_out);
    if (i_frame_size < 0) {
        x264_cli_log ("x264", 0, "x264_encoder_encode failed\n");
        return -1;
    };
    if (i_frame_size) {
        i_frame_size = cli_output.write_frame (hout, nal[0].p_payload, i_frame_size, &pic_out);
        *last_dts = pic_out.i_dts;
    }
    return i_frame_size;
}

int64_t print_status (int64_t i_start, int64_t i_previous, int i_frame, int i_frame_total, int64_t i_file, x264_param_t *param, int64_t last_ts) {
    char buf [200];
    int64_t i_time = x264_mdate ();
    if (i_previous && i_time - i_previous < 250000)
        return i_previous;
    int64_t i_elapsed = i_time - i_start;
    double fps = i_elapsed > 0 ? i_frame * 1000000./ i_elapsed : 0;
    double bitrate;
    if (last_ts)
        bitrate = (double) i_file * 8 / ((double) last_ts * 1000 * param->i_timebase_num / param->i_timebase_den);
    else
        bitrate = (double) i_file * 8 / ((double) 1000 * param->i_fps_den / param->i_fps_num);
    if (i_frame_total) {
        int eta = i_elapsed * (i_frame_total - i_frame) / ((int64_t) i_frame * 1000000);
        sprintf (buf, "x264 [%.1f%%] %d/%d frames, %.2f fps, %.2f kb/s, eta %d:%02d:%02d", 100.* i_frame / i_frame_total, i_frame, i_frame_total, fps, bitrate, eta / 3600, (eta / 60) % 60, eta % 60);
    }
    else
        sprintf (buf, "x264 %d frames: %.2f fps, %.2f kb/s", i_frame, fps, bitrate);
    fprintf (stderr, "%s  \r", buf + 5);
    ;
    fflush (stderr);
    return i_time;
}

void convert_cli_to_lib_pic (x264_picture_t *lib, cli_pic_t *cli) {
    memcpy (lib -> img.i_stride, cli -> img.stride, sizeof (cli -> img.stride));
    memcpy (lib -> img.plane, cli -> img.plane, sizeof (cli -> img.plane));
    lib->img.i_plane = cli->img.planes;
    lib->img.i_csp = cli->img.csp;
    lib->i_pts = cli->pts;
}

int placeWrapper () {
    if (0 == 1) {
        x264_param_t *param;
        cli_opt_t *opt;
        x264_t *h = ((void *) 0);
        x264_picture_t pic;
        cli_pic_t cli_pic;
        const cli_pulldown_t *pulldown = ((void *) 0);
        int i_frame = 0;
        int i_frame_output = 0;
        int64_t i_end, i_previous = 0, i_start = 0;
        int64_t i_file = 0;
        int i_frame_size;
        int64_t last_dts = 0;
        int64_t prev_dts = 0;
        int64_t first_dts = 0;
        int pts_warning_cnt = 0;
        int64_t largest_pts = -1;
        int64_t second_largest_pts = -1;
        int64_t ticks_per_frame;
        double duration;
        double pulldown_pts = 0;
        int retval = 0;
        opt->b_progress &= param->i_log_level < 3;
        if (opt->i_pulldown && !param->b_vfr_input) {
            param->b_pulldown = 1;
            param->b_pic_struct = 1;
            pulldown = &pulldown_values[opt->i_pulldown];
            param->i_timebase_num = param->i_fps_den;
            if (fmod (param->i_fps_num * pulldown->fps_factor, 1)) {
                x264_cli_log ("x264", 0, "unsupported framerate for chosen pulldown\n");
                retval = -1;
                return -1;
            }
            param->i_timebase_den = param->i_fps_num * pulldown->fps_factor;
        }
        h = x264_encoder_open_142 (param);
        if (!h) {
            x264_cli_log ("x264", 0, "x264_encoder_open failed\n");
            retval = -1;
            return -1;
        };
        x264_encoder_parameters (h, param);
        if (cli_output.set_param (opt->hout, param)) {
            x264_cli_log ("x264", 0, "can't set outfile param\n");
            retval = -1;
            return -1;
        };
        i_start = x264_mdate ();
        ticks_per_frame = (int64_t) param->i_timebase_den * param->i_fps_den / param->i_timebase_num / param->i_fps_num;
        if (ticks_per_frame < 1 && !param->b_vfr_input) {
            x264_cli_log ("x264", 0, "ticks_per_frame invalid: %" "l" "d" "\n", ticks_per_frame);
            retval = -1;
            return -1;
        }
        ticks_per_frame = ((ticks_per_frame) > (1) ? (ticks_per_frame) : (1));
        if (!param->b_repeat_headers) {
            x264_nal_t *headers;
            int i_nal;
            if (x264_encoder_headers (h, &headers, &i_nal) < 0) {
                x264_cli_log ("x264", 0, "x264_encoder_headers failed\n");
                retval = -1;
                return -1;
            }
            if ((i_file = cli_output.write_headers (opt->hout, headers)) < 0) {
                x264_cli_log ("x264", 0, "error writing headers to output file\n");
                retval = -1;
                return -1;
            };
        }
        if (opt->tcfile_out)
            fprintf (opt->tcfile_out, "# timecode format v2\n");
        for (; !b_ctrl_c && (i_frame < param->i_frame_total || !param->i_frame_total); i_frame++) {
            if (filter.get_frame (opt->hin, &cli_pic, i_frame +opt->i_seek))
                break;
            x264_picture_init (& pic);
            convert_cli_to_lib_pic (& pic, & cli_pic);
            if (!param->b_vfr_input)
                pic.i_pts = i_frame;
            if (opt->i_pulldown && !param->b_vfr_input) {
                pic.i_pic_struct = pulldown->pattern[i_frame % pulldown->mod];
                pic.i_pts = (int64_t) (pulldown_pts + 0.5);
                pulldown_pts += pulldown_frame_duration[pic.i_pic_struct];
            }
            else if (opt->timebase_convert_multiplier)
                pic.i_pts = (int64_t) (pic.i_pts * opt->timebase_convert_multiplier + 0.5);
            if (pic.i_pts <= largest_pts) {
                if (cli_log_level >= 3 || pts_warning_cnt < 3)
                    x264_cli_log ("x264", 1, "non-strictly-monotonic pts at frame %d (%" "l" "d" " <= %" "l" "d" ")\n", i_frame, pic.i_pts, largest_pts);
                else if (pts_warning_cnt == 3)
                    x264_cli_log ("x264", 1, "too many nonmonotonic pts warnings, suppressing further ones\n");
                pts_warning_cnt++;
                pic.i_pts = largest_pts + ticks_per_frame;
            }
            second_largest_pts = largest_pts;
            largest_pts = pic.i_pts;
            if (opt->tcfile_out)
                fprintf (opt->tcfile_out, "%.6f\n", pic.i_pts * ((double) param->i_timebase_num / param->i_timebase_den) * 1e3);
            if (opt->qpfile)
                exit (0);
            prev_dts = last_dts;
            i_frame_size = encode_frame (h, opt->hout, &pic, &last_dts);
            if (i_frame_size < 0) {
                b_ctrl_c = 1;
                retval = -1;
            }
            else if (i_frame_size) {
                i_file += i_frame_size;
                i_frame_output++;
                if (i_frame_output == 1)
                    first_dts = prev_dts = last_dts;
            }
            if (filter.release_frame (opt->hin, &cli_pic, i_frame +opt->i_seek))
                break;
            if (opt->b_progress && i_frame_output)
                i_previous = print_status (i_start, i_previous, i_frame_output, param->i_frame_total, i_file, param, 2 * last_dts - prev_dts - first_dts);
        }
        while (!b_ctrl_c && x264_encoder_delayed_frames (h)) {
            prev_dts = last_dts;
            i_frame_size = encode_frame (h, opt->hout, ((void *) 0), &last_dts);
            if (i_frame_size < 0) {
                b_ctrl_c = 1;
                retval = -1;
            }
            else if (i_frame_size) {
                i_file += i_frame_size;
                i_frame_output++;
                if (i_frame_output == 1)
                    first_dts = prev_dts = last_dts;
            }
            if (opt->b_progress && i_frame_output)
                i_previous = print_status (i_start, i_previous, i_frame_output, param->i_frame_total, i_file, param, 2 * last_dts - prev_dts - first_dts);
        }
        if (pts_warning_cnt >= 3 && cli_log_level < 3)
            x264_cli_log ("x264", 1, "%d suppressed nonmonotonic pts warnings\n", pts_warning_cnt -3);
        if (i_frame_output == 1)
            duration = (double) param->i_fps_den / param->i_fps_num;
        else if (b_ctrl_c)
            duration = (double) (2 * last_dts - prev_dts - first_dts) * param->i_timebase_num / param->i_timebase_den;
        else
            duration = (double) (2 * largest_pts - second_largest_pts) * param->i_timebase_num / param->i_timebase_den;
        i_end = x264_mdate ();
        if (opt->b_progress)
            fprintf (stderr, "                                                                               \r");
        if (h)
            x264_encoder_close (h);
        fprintf (stderr, "\n");
        if (b_ctrl_c)
            fprintf (stderr, "aborted at input frame %d, output frame %d\n", opt->i_seek + i_frame, i_frame_output);
        cli_output.close_file (opt->hout, largest_pts, second_largest_pts);
        opt->hout = ((void *) 0);
        if (i_frame_output > 0) {
            double fps = (double) i_frame_output * (double) 1000000 / (double) (i_end - i_start);
            fprintf (stderr, "encoded %d frames, %.2f fps, %.2f kb/s\n", i_frame_output, fps, (double) i_file * 8 / (1000 * duration));
        }
    }
    return 0;
}

int encode (x264_param_t *param, cli_opt_t *opt) {
    x264_t *h = ((void *) 0);
    x264_picture_t pic;
    cli_pic_t cli_pic;
    const cli_pulldown_t *pulldown = ((void *) 0);
    int i_frame = 0;
    int i_frame_output = 0;
    int64_t i_end, i_previous = 0, i_start = 0;
    int64_t i_file = 0;
    int i_frame_size;
    int64_t last_dts = 0;
    int64_t prev_dts = 0;
    int64_t first_dts = 0;
    int pts_warning_cnt = 0;
    int64_t largest_pts = -1;
    int64_t second_largest_pts = -1;
    int64_t ticks_per_frame;
    double duration;
    double pulldown_pts = 0;
    int retval = 0;
    opt->b_progress &= param->i_log_level < 3;
    if (opt->i_pulldown && !param->b_vfr_input) {
        param->b_pulldown = 1;
        param->b_pic_struct = 1;
        pulldown = &pulldown_values[opt->i_pulldown];
        param->i_timebase_num = param->i_fps_den;
        if (fmod (param->i_fps_num * pulldown->fps_factor, 1)) {
            x264_cli_log ("x264", 0, "unsupported framerate for chosen pulldown\n");
            retval = -1;
            return -1;
        }
        param->i_timebase_den = param->i_fps_num * pulldown->fps_factor;
    }
    h = x264_encoder_open_142 (param);
    if (!h) {
        x264_cli_log ("x264", 0, "x264_encoder_open failed\n");
        retval = -1;
        return -1;
    };
    x264_encoder_parameters (h, param);
    if (cli_output.set_param (opt->hout, param)) {
        x264_cli_log ("x264", 0, "can't set outfile param\n");
        retval = -1;
        return -1;
    };
    i_start = x264_mdate ();
    ticks_per_frame = (int64_t) param->i_timebase_den * param->i_fps_den / param->i_timebase_num / param->i_fps_num;
    if (ticks_per_frame < 1 && !param->b_vfr_input) {
        x264_cli_log ("x264", 0, "ticks_per_frame invalid: %" "l" "d" "\n", ticks_per_frame);
        retval = -1;
        return -1;
    }
    ticks_per_frame = ((ticks_per_frame) > (1) ? (ticks_per_frame) : (1));
    if (!param->b_repeat_headers) {
        x264_nal_t *headers;
        int i_nal;
        if (x264_encoder_headers (h, &headers, &i_nal) < 0) {
            x264_cli_log ("x264", 0, "x264_encoder_headers failed\n");
            retval = -1;
            return -1;
        }
        if ((i_file = cli_output.write_headers (opt->hout, headers)) < 0) {
            x264_cli_log ("x264", 0, "error writing headers to output file\n");
            retval = -1;
            return -1;
        };
    }
    if (opt->tcfile_out)
        fprintf (opt->tcfile_out, "# timecode format v2\n");
    for (; !b_ctrl_c && (i_frame < param->i_frame_total || !param->i_frame_total); i_frame++) {
        if (filter.get_frame (opt->hin, &cli_pic, i_frame +opt->i_seek))
            break;
        x264_picture_init (& pic);
        convert_cli_to_lib_pic (& pic, & cli_pic);
        if (!param->b_vfr_input)
            pic.i_pts = i_frame;
        if (opt->i_pulldown && !param->b_vfr_input) {
            pic.i_pic_struct = pulldown->pattern[i_frame % pulldown->mod];
            pic.i_pts = (int64_t) (pulldown_pts + 0.5);
            pulldown_pts += pulldown_frame_duration[pic.i_pic_struct];
        }
        else if (opt->timebase_convert_multiplier)
            pic.i_pts = (int64_t) (pic.i_pts * opt->timebase_convert_multiplier + 0.5);
        if (pic.i_pts <= largest_pts) {
            if (cli_log_level >= 3 || pts_warning_cnt < 3)
                x264_cli_log ("x264", 1, "non-strictly-monotonic pts at frame %d (%" "l" "d" " <= %" "l" "d" ")\n", i_frame, pic.i_pts, largest_pts);
            else if (pts_warning_cnt == 3)
                x264_cli_log ("x264", 1, "too many nonmonotonic pts warnings, suppressing further ones\n");
            pts_warning_cnt++;
            pic.i_pts = largest_pts + ticks_per_frame;
        }
        second_largest_pts = largest_pts;
        largest_pts = pic.i_pts;
        if (opt->tcfile_out)
            fprintf (opt->tcfile_out, "%.6f\n", pic.i_pts * ((double) param->i_timebase_num / param->i_timebase_den) * 1e3);
        if (opt->qpfile)
            exit (0);
        prev_dts = last_dts;
        i_frame_size = encode_frame (h, opt->hout, &pic, &last_dts);
        if (i_frame_size < 0) {
            b_ctrl_c = 1;
            retval = -1;
        }
        else if (i_frame_size) {
            i_file += i_frame_size;
            i_frame_output++;
            if (i_frame_output == 1)
                first_dts = prev_dts = last_dts;
        }
        if (filter.release_frame (opt->hin, &cli_pic, i_frame +opt->i_seek))
            break;
        if (opt->b_progress && i_frame_output)
            i_previous = print_status (i_start, i_previous, i_frame_output, param->i_frame_total, i_file, param, 2 * last_dts - prev_dts - first_dts);
    }
    while (!b_ctrl_c && x264_encoder_delayed_frames (h)) {
        prev_dts = last_dts;
        i_frame_size = encode_frame (h, opt->hout, ((void *) 0), &last_dts);
        if (i_frame_size < 0) {
            b_ctrl_c = 1;
            retval = -1;
        }
        else if (i_frame_size) {
            i_file += i_frame_size;
            i_frame_output++;
            if (i_frame_output == 1)
                first_dts = prev_dts = last_dts;
        }
        if (opt->b_progress && i_frame_output)
            i_previous = print_status (i_start, i_previous, i_frame_output, param->i_frame_total, i_file, param, 2 * last_dts - prev_dts - first_dts);
    }
    if (pts_warning_cnt >= 3 && cli_log_level < 3)
        x264_cli_log ("x264", 1, "%d suppressed nonmonotonic pts warnings\n", pts_warning_cnt -3);
    if (i_frame_output == 1)
        duration = (double) param->i_fps_den / param->i_fps_num;
    else if (b_ctrl_c)
        duration = (double) (2 * last_dts - prev_dts - first_dts) * param->i_timebase_num / param->i_timebase_den;
    else
        duration = (double) (2 * largest_pts - second_largest_pts) * param->i_timebase_num / param->i_timebase_den;
    i_end = x264_mdate ();
    if (opt->b_progress)
        fprintf (stderr, "                                                                               \r");
    if (h)
        x264_encoder_close (h);
    fprintf (stderr, "\n");
    if (b_ctrl_c)
        fprintf (stderr, "aborted at input frame %d, output frame %d\n", opt->i_seek + i_frame, i_frame_output);
    cli_output.close_file (opt->hout, largest_pts, second_largest_pts);
    opt->hout = ((void *) 0);
    if (i_frame_output > 0) {
        double fps = (double) i_frame_output * (double) 1000000 / (double) (i_end - i_start);
        fprintf (stderr, "encoded %d frames, %.2f fps, %.2f kb/s\n", i_frame_output, fps, (double) i_file * 8 / (1000 * duration));
    }
    placeWrapper ();
    return retval;
}

