
#include "common.h"

#include "setENCODER.h"

#include "analyse.h"

#include "ratecontrol.h"

#include "macroblockENCODER.h"

#include "me.h"
int x264_encoder_frame_end (x264_t * h, x264_t * thread_current, x264_nal_t * * pp_nal, int * pi_nal, x264_picture_t * pic_out);

double x264_psnr (double sqe, double size) {
    double mse = sqe / (((1 << 8) - 1) * ((1 << 8) - 1) * size);
    if (mse <= 0.0000000001)
        return 100;
    return -10.0 * log10 (mse);
}

double x264_ssim (double ssim) {
    double inv_ssim = 1 - ssim;
    if (inv_ssim <= 0.0000000001)
        return 100;
    return -10.0 * log10 (inv_ssim);
}

int x264_threadpool_wait_all (x264_t *h) {
    for (int i = 0; i < h->param.i_threads; i++)
        if (h->thread[i]->b_thread_active) {
            h->thread[i]->b_thread_active = 0;
            if ((intptr_t) ((void *) 0) < 0)
                return -1;
        }
    return 0;
}

void x264_frame_dump (x264_t *h) {
    FILE *f = fopen (h->param.psz_dump_yuv, "r+b");
    if (!f)
        return;
    if (h->param.b_sliced_threads)
        x264_threadpool_wait_all (h);
    int frame_size = ((h->param.i_height * h->param.i_width * sizeof (pixel)) + 2 * ((h->param.i_height * h->param.i_width * sizeof (pixel)) >> (h->mb.chroma_h_shift + h->mb.chroma_v_shift)));
    fseeko (f, (uint64_t) h -> fdec -> i_frame * frame_size, 0);
    for (int p = 0; p < ((h->sps->i_chroma_format_idc == CHROMA_444) ? 3 : 1); p++)
        for (int y = 0; y < h->param.i_height; y++)
            fwrite (&h->fdec->plane[p][y * h->fdec->i_stride[p]], sizeof (pixel), h->param.i_width, f);
    if (!(h->sps->i_chroma_format_idc == CHROMA_444)) {
        int cw = h->param.i_width >> 1;
        int ch = h->param.i_height >> h->mb.chroma_v_shift;
        pixel *planeu = x264_malloc ((cw * ch * 2 + 32) * sizeof (pixel));
        if (planeu) {
            pixel *planev = planeu + cw * ch + 16;
            h->mc.plane_copy_deinterleave (planeu, cw, planev, cw, h->fdec->plane[1], h->fdec->i_stride[1], cw, ch);
            fwrite (planeu, 1, cw * ch * sizeof (pixel), f);
            fwrite (planev, 1, cw * ch * sizeof (pixel), f);
            x264_free (planeu);
        }
    }
    fclose (f);
}

void x264_slice_header_init (x264_t *h, x264_slice_header_t *sh, x264_sps_t *sps, x264_pps_t *pps, int i_idr_pic_id, int i_frame, int i_qp) {
    x264_param_t *param = &h->param;
    sh->sps = sps;
    sh->pps = pps;
    sh->i_first_mb = 0;
    sh->i_last_mb = h->mb.i_mb_count - 1;
    sh->i_pps_id = pps->i_id;
    sh->i_frame_num = i_frame;
    sh->b_mbaff = 0;
    sh->b_field_pic = 0;
    sh->b_bottom_field = 0;
    sh->i_idr_pic_id = i_idr_pic_id;
    sh->i_poc = 0;
    sh->i_delta_poc_bottom = 0;
    sh->i_delta_poc[0] = 0;
    sh->i_delta_poc[1] = 0;
    sh->i_redundant_pic_cnt = 0;
    h->mb.b_direct_auto_write = h->param.analyse.i_direct_mv_pred == 3 && h->param.i_bframe && (h->param.rc.b_stat_write || !h->param.rc.b_stat_read);
    if (!h->mb.b_direct_auto_read && sh->i_type == SLICE_TYPE_B) {
        if (h->fref[1][0]->i_poc_l0ref0 == h->fref[0][0]->i_poc) {
            if (h->mb.b_direct_auto_write)
                sh->b_direct_spatial_mv_pred = (h->stat.i_direct_score[1] > h->stat.i_direct_score[0]);
            else
                sh->b_direct_spatial_mv_pred = (param->analyse.i_direct_mv_pred == 1);
        }
        else {
            h->mb.b_direct_auto_write = 0;
            sh->b_direct_spatial_mv_pred = 1;
        }
    }
    sh->b_num_ref_idx_override = 0;
    sh->i_num_ref_idx_l0_active = 1;
    sh->i_num_ref_idx_l1_active = 1;
    sh->b_ref_pic_list_reordering[0] = h->b_ref_reorder[0];
    sh->b_ref_pic_list_reordering[1] = h->b_ref_reorder[1];
    for (int list = 0; list < 2; list++) {
        if (sh->b_ref_pic_list_reordering[list]) {
            int pred_frame_num = i_frame;
            for (int i = 0; i < h->i_ref[list]; i++) {
                int diff = h->fref[list][i]->i_frame_num - pred_frame_num;
                sh->ref_pic_list_order[list][i].idc = (diff > 0);
                sh->ref_pic_list_order[list][i].arg = (abs (diff) - 1) & ((1 << sps->i_log2_max_frame_num) - 1);
                pred_frame_num = h->fref[list][i]->i_frame_num;
            }
        }
    }
    sh->i_cabac_init_idc = param->i_cabac_init_idc;
    sh->i_qp = (((i_qp)) < ((51 + (6 * (8 - 8)))) ? ((i_qp)) : ((51 + (6 * (8 - 8)))));
    sh->i_qp_delta = sh->i_qp - pps->i_pic_init_qp;
    sh->b_sp_for_swidth = 0;
    sh->i_qs_delta = 0;
    int deblock_thresh = i_qp + 2 * ((param->i_deblocking_filter_alphac0) < (param->i_deblocking_filter_beta) ? (param->i_deblocking_filter_alphac0) : (param->i_deblocking_filter_beta));
    if (param->b_deblocking_filter && (h->mb.b_variable_qp || 15 < deblock_thresh))
        sh->i_disable_deblocking_filter_idc = param->b_sliced_threads ? 2 : 0;
    else
        sh->i_disable_deblocking_filter_idc = 1;
    sh->i_alpha_c0_offset = param->i_deblocking_filter_alphac0 << 1;
    sh->i_beta_offset = param->i_deblocking_filter_beta << 1;
}

void x264_slice_header_write (bs_t *s, x264_slice_header_t *sh, int i_nal_ref_idc) {
    if (sh->b_mbaff) {
        int first_x = sh->i_first_mb % sh->sps->i_mb_width;
        int first_y = sh->i_first_mb / sh->sps->i_mb_width;
        (((first_y & 1) == 0) ? (void) (0) : __assert_fail ("(first_y&1) == 0", "DonorSourceCode/encoder.c", 210, __PRETTY_FUNCTION__));
        bs_write_ue_big (s, (2 * first_x + sh -> sps -> i_mb_width * (first_y & ~ 1) + (first_y & 1)) >> 1);
    }
    else
        bs_write_ue_big (s, sh->i_first_mb);
    bs_write_ue_big (s, sh -> i_type + 5);
    bs_write_ue_big (s, sh -> i_pps_id);
    bs_write (s, sh -> sps -> i_log2_max_frame_num, sh -> i_frame_num & ((1 << sh -> sps -> i_log2_max_frame_num) - 1));
    if (!sh->sps->b_frame_mbs_only) {
        bs_write1 (s, sh -> b_field_pic);
        if (sh->b_field_pic)
            bs_write1 (s, sh->b_bottom_field);
    }
    if (sh->i_idr_pic_id >= 0)
        bs_write_ue_big (s, sh->i_idr_pic_id);
    if (sh->sps->i_poc_type == 0) {
        bs_write (s, sh -> sps -> i_log2_max_poc_lsb, sh -> i_poc & ((1 << sh -> sps -> i_log2_max_poc_lsb) - 1));
        if (sh->pps->b_pic_order && !sh->b_field_pic)
            bs_write_se (s, sh->i_delta_poc_bottom);
    }
    if (sh->pps->b_redundant_pic_cnt)
        bs_write_ue_big (s, sh->i_redundant_pic_cnt);
    if (sh->i_type == SLICE_TYPE_B)
        bs_write1 (s, sh->b_direct_spatial_mv_pred);
    if (sh->i_type == SLICE_TYPE_P || sh->i_type == SLICE_TYPE_B) {
        bs_write1 (s, sh -> b_num_ref_idx_override);
        if (sh->b_num_ref_idx_override) {
            bs_write_ue_big (s, sh -> i_num_ref_idx_l0_active - 1);
            if (sh->i_type == SLICE_TYPE_B)
                bs_write_ue_big (s, sh->i_num_ref_idx_l1_active - 1);
        }
    }
    if (sh->i_type != SLICE_TYPE_I) {
        bs_write1 (s, sh -> b_ref_pic_list_reordering [0]);
        if (sh->b_ref_pic_list_reordering[0]) {
            for (int i = 0; i < sh->i_num_ref_idx_l0_active; i++) {
                bs_write_ue_big (s, sh -> ref_pic_list_order [0] [i].idc);
                bs_write_ue_big (s, sh -> ref_pic_list_order [0] [i].arg);
            }
            bs_write_ue_big (s, 3);
        }
    }
    if (sh->i_type == SLICE_TYPE_B) {
        bs_write1 (s, sh -> b_ref_pic_list_reordering [1]);
        if (sh->b_ref_pic_list_reordering[1]) {
            for (int i = 0; i < sh->i_num_ref_idx_l1_active; i++) {
                bs_write_ue_big (s, sh -> ref_pic_list_order [1] [i].idc);
                bs_write_ue_big (s, sh -> ref_pic_list_order [1] [i].arg);
            }
            bs_write_ue_big (s, 3);
        }
    }
    sh->b_weighted_pred = 0;
    if (sh->pps->b_weighted_pred && sh->i_type == SLICE_TYPE_P) {
        sh->b_weighted_pred = sh->weight[0][0].weightfn || sh->weight[0][1].weightfn || sh->weight[0][2].weightfn;
        bs_write_ue_big (s, sh -> weight [0] [0].i_denom);
        bs_write_ue_big (s, sh -> weight [0] [1].i_denom);
        for (int i = 0; i < sh->i_num_ref_idx_l0_active; i++) {
            int luma_weight_l0_flag = !!sh->weight[i][0].weightfn;
            int chroma_weight_l0_flag = !!sh->weight[i][1].weightfn || !!sh->weight[i][2].weightfn;
            bs_write1 (s, luma_weight_l0_flag);
            if (luma_weight_l0_flag) {
                bs_write_se (s, sh -> weight [i] [0].i_scale);
                bs_write_se (s, sh -> weight [i] [0].i_offset);
            }
            bs_write1 (s, chroma_weight_l0_flag);
            if (chroma_weight_l0_flag) {
                for (int j = 1; j < 3; j++) {
                    bs_write_se (s, sh -> weight [i] [j].i_scale);
                    bs_write_se (s, sh -> weight [i] [j].i_offset);
                }
            }
        }
    }
    else if (sh->pps->b_weighted_bipred == 1 && sh->i_type == SLICE_TYPE_B) {
    }
    if (i_nal_ref_idc != 0) {
        if (sh->i_idr_pic_id >= 0) {
            bs_write1 (s, 0);
            bs_write1 (s, 0);
        }
        else {
            bs_write1 (s, sh -> i_mmco_command_count > 0);
            if (sh->i_mmco_command_count > 0) {
                for (int i = 0; i < sh->i_mmco_command_count; i++) {
                    bs_write_ue_big (s, 1);
                    bs_write_ue_big (s, sh -> mmco [i].i_difference_of_pic_nums - 1);
                }
                bs_write_ue_big (s, 0);
            }
        }
    }
    if (sh->pps->b_cabac && sh->i_type != SLICE_TYPE_I)
        bs_write_ue_big (s, sh->i_cabac_init_idc);
    bs_write_se (s, sh -> i_qp_delta);
    if (sh->pps->b_deblocking_filter_control) {
        bs_write_ue_big (s, sh -> i_disable_deblocking_filter_idc);
        if (sh->i_disable_deblocking_filter_idc != 1) {
            bs_write_se (s, sh -> i_alpha_c0_offset >> 1);
            bs_write_se (s, sh -> i_beta_offset >> 1);
        }
    }
}

int x264_bitstream_check_buffer_internal (x264_t *h, int size, int b_cabac, int i_nal) {
    if ((b_cabac && (h->cabac.p_end - h->cabac.p < size)) || (h->out.bs.p_end - h->out.bs.p < size)) {
        int buf_size = h->out.i_bitstream + size;
        uint8_t *buf = x264_malloc (buf_size);
        if (!buf)
            return -1;
        int aligned_size = h->out.i_bitstream & ~15;
        h->mc.memcpy_aligned (buf, h->out.p_bitstream, aligned_size);
        memcpy (buf + aligned_size, h -> out.p_bitstream + aligned_size, h -> out.i_bitstream - aligned_size);
        intptr_t delta = buf - h->out.p_bitstream;
        h->out.bs.p_start += delta;
        h->out.bs.p += delta;
        h->out.bs.p_end = buf + buf_size;
        h->cabac.p_start += delta;
        h->cabac.p += delta;
        h->cabac.p_end = buf + buf_size;
        for (int i = 0; i <= i_nal; i++)
            h->out.nal[i].p_payload += delta;
        x264_free (h -> out.p_bitstream);
        h->out.p_bitstream = buf;
        h->out.i_bitstream = buf_size;
    }
    return 0;
}

int x264_bitstream_check_buffer (x264_t *h) {
    int max_row_size = (2500 << 0) * h->mb.i_mb_width;
    return x264_bitstream_check_buffer_internal (h, max_row_size, h->param.b_cabac, h->out.i_nal);
}

int x264_bitstream_check_buffer_filler (x264_t *h, int filler) {
    filler += 32;
    return x264_bitstream_check_buffer_internal (h, filler, 0, -1);
}

int x264_validate_parameters (x264_t *h, int b_open) {
    if (!h->param.pf_log) {
        x264_log (((void *) 0), 0, "pf_log not set! did you forget to call x264_param_default?\n");
        return -1;
    }
    if (h->param.b_interlaced) {
        x264_log (h, 0, "not compiled with interlaced support\n");
        return -1;
    }
    if (h->param.i_width <= 0 || h->param.i_height <= 0) {
        x264_log (h, 0, "invalid width x height (%dx%d)\n", h -> param.i_width, h -> param.i_height);
        return -1;
    }
    int i_csp = h->param.i_csp & 0x00ff;
    if (i_csp <= 0x0000 || i_csp >= 0x000d) {
        x264_log (h, 0, "invalid CSP (only I420/YV12/NV12/I422/YV16/NV16/I444/YV24/BGR/BGRA/RGB supported)\n");
        return -1;
    }
    if (i_csp < 0x0008 && h->param.i_width % 2) {
        x264_log (h, 0, "width not divisible by 2 (%dx%d)\n", h -> param.i_width, h -> param.i_height);
        return -1;
    }
    if (i_csp < 0x0004 && 0 && h->param.i_height % 4) {
        x264_log (h, 0, "height not divisible by 4 (%dx%d)\n", h -> param.i_width, h -> param.i_height);
        return -1;
    }
    if ((i_csp < 0x0004 || 0) && h->param.i_height % 2) {
        x264_log (h, 0, "height not divisible by 2 (%dx%d)\n", h -> param.i_width, h -> param.i_height);
        return -1;
    }
    if ((h->param.crop_rect.i_left + h->param.crop_rect.i_right) >= h->param.i_width || (h->param.crop_rect.i_top + h->param.crop_rect.i_bottom) >= h->param.i_height) {
        x264_log (h, 0, "invalid crop-rect %u,%u,%u,%u\n", h -> param.crop_rect.i_left, h -> param.crop_rect.i_top, h -> param.crop_rect.i_right, h -> param.crop_rect.i_bottom);
        return -1;
    }
    if (h->param.vui.i_sar_width <= 0 || h->param.vui.i_sar_height <= 0) {
        h->param.vui.i_sar_width = 0;
        h->param.vui.i_sar_height = 0;
    }
    if (h->param.i_threads == 0)
        h->param.i_threads = x264_cpu_num_processors () * (h->param.b_sliced_threads ? 2 : 3) / 2;
    int max_sliced_threads = ((1) > ((h->param.i_height + 15) / 16 / 4) ? (1) : ((h->param.i_height + 15) / 16 / 4));
    if (h->param.i_threads > 1) {
        x264_log (h, 1, "not compiled with thread support!\n");
        h->param.i_threads = 1;
        if (h->param.b_sliced_threads)
            h->param.i_threads = ((h->param.i_threads) < (max_sliced_threads) ? (h->param.i_threads) : (max_sliced_threads));
    }
    h->param.i_threads = x264_clip3 (h->param.i_threads, 1, 128);
    if (h->param.i_threads == 1) {
        h->param.b_sliced_threads = 0;
        h->param.i_lookahead_threads = 1;
    }
    h->i_thread_frames = h->param.b_sliced_threads ? 1 : h->param.i_threads;
    if (h->i_thread_frames > 1)
        h->param.nalu_process = ((void *) 0);
    if (h->param.b_opencl) {
        if (h->param.i_width < 32 || h->param.i_height < 32) {
            x264_log (h, 1, "OpenCL: frame size is too small, disabling opencl\n");
            h->param.b_opencl = 0;
        }
        if (h->param.opencl_device_id && h->param.i_opencl_device) {
            x264_log (h, 1, "OpenCL: device id and device skip count configured; dropping skip\n");
            h->param.i_opencl_device = 0;
        }
    }
    h->param.i_keyint_max = x264_clip3 (h->param.i_keyint_max, 1, (1 << 30));
    if (h->param.i_keyint_max == 1) {
        h->param.b_intra_refresh = 0;
        h->param.analyse.i_weighted_pred = 0;
        h->param.i_frame_reference = 1;
        h->param.i_dpb_size = 1;
    }
    h->param.i_frame_packing = x264_clip3 (h->param.i_frame_packing, -1, 5);
    if (b_open) {
        int score = 0;
        score += h->param.analyse.i_me_range == 0;
        score += h->param.rc.i_qp_step == 3;
        score += h->param.i_keyint_max == 12;
        score += h->param.rc.i_qp_min == 2;
        score += h->param.rc.i_qp_max == 31;
        score += h->param.rc.f_qcompress == 0.5;
        score += fabs (h->param.rc.f_ip_factor - 1.25) < 0.01;
        score += fabs (h->param.rc.f_pb_factor - 1.25) < 0.01;
        score += h->param.analyse.inter == 0 && h->param.analyse.i_subpel_refine == 8;
        if (score >= 5) {
            x264_log (h, 0, "broken ffmpeg default settings detected\n");
            x264_log (h, 0, "use an encoding preset (e.g. -vpre medium)\n");
            x264_log (h, 0, "preset usage: -vpre <speed> -vpre <profile>\n");
            x264_log (h, 0, "speed presets are listed in x264 --help\n");
            x264_log (h, 0, "profile is optional; x264 defaults to high\n");
            return -1;
        }
    }
    if (h->param.rc.i_rc_method < 0 || h->param.rc.i_rc_method > 2) {
        x264_log (h, 0, "no ratecontrol method specified\n");
        return -1;
    }
    if (0)
        h->param.b_pic_struct = 1;
    if (h->param.i_avcintra_class) {
        if (8 != 10) {
            x264_log (h, 0, "%2d-bit AVC-Intra is not widely compatible\n", 8);
            x264_log (h, 0, "10-bit x264 is required to encode AVC-Intra\n");
            return -1;
        }
        int type = h->param.i_avcintra_class == 200 ? 2 : h->param.i_avcintra_class == 100 ? 1 : h->param.i_avcintra_class == 50 ? 0 : -1;
        if (type < 0) {
            x264_log (h, 0, "Invalid AVC-Intra class\n");
            return -1;
        }
        static const struct {
            uint16_t fps_num;
            uint16_t fps_den;
            uint8_t interlaced;
            uint16_t frame_size;
            const uint8_t *cqm_4ic;
            const uint8_t *cqm_8iy;
        } avcintra_lut [3] [2] [7] = {{{{60000, 1001, 0, 912, x264_cqm_avci50_4ic, x264_cqm_avci50_p_8iy}, {50, 1, 0, 1100, x264_cqm_avci50_4ic, x264_cqm_avci50_p_8iy}, {30000, 1001, 0, 912, x264_cqm_avci50_4ic, x264_cqm_avci50_p_8iy}, {25, 1, 0, 1100, x264_cqm_avci50_4ic, x264_cqm_avci50_p_8iy}, {24000, 1001, 0, 912, x264_cqm_avci50_4ic, x264_cqm_avci50_p_8iy}}, {{30000, 1001, 1, 1820, x264_cqm_avci50_4ic, x264_cqm_avci50_1080i_8iy}, {25, 1, 1, 2196, x264_cqm_avci50_4ic, x264_cqm_avci50_1080i_8iy}, {60000, 1001, 0, 1820, x264_cqm_avci50_4ic, x264_cqm_avci50_p_8iy}, {30000, 1001, 0, 1820, x264_cqm_avci50_4ic, x264_cqm_avci50_p_8iy}, {50, 1, 0, 2196, x264_cqm_avci50_4ic, x264_cqm_avci50_p_8iy}, {25, 1, 0, 2196, x264_cqm_avci50_4ic, x264_cqm_avci50_p_8iy}, {24000, 1001, 0, 1820, x264_cqm_avci50_4ic, x264_cqm_avci50_p_8iy}}}, {{{60000, 1001, 0, 1848, x264_cqm_avci100_720p_4ic, x264_cqm_avci100_720p_8iy}, {50, 1, 0, 2224, x264_cqm_avci100_720p_4ic, x264_cqm_avci100_720p_8iy}, {30000, 1001, 0, 1848, x264_cqm_avci100_720p_4ic, x264_cqm_avci100_720p_8iy}, {25, 1, 0, 2224, x264_cqm_avci100_720p_4ic, x264_cqm_avci100_720p_8iy}, {24000, 1001, 0, 1848, x264_cqm_avci100_720p_4ic, x264_cqm_avci100_720p_8iy}}, {{30000, 1001, 1, 3692, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080i_8iy}, {25, 1, 1, 4444, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080i_8iy}, {60000, 1001, 0, 3692, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080p_8iy}, {30000, 1001, 0, 3692, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080p_8iy}, {50, 1, 0, 4444, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080p_8iy}, {25, 1, 0, 4444, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080p_8iy}, {24000, 1001, 0, 3692, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080p_8iy}}}, {{{60000, 1001, 0, 3724, x264_cqm_avci100_720p_4ic, x264_cqm_avci100_720p_8iy}, {50, 1, 0, 4472, x264_cqm_avci100_720p_4ic, x264_cqm_avci100_720p_8iy}}, {{30000, 1001, 1, 7444, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080i_8iy}, {25, 1, 1, 8940, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080i_8iy}, {60000, 1001, 0, 7444, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080p_8iy}, {30000, 1001, 0, 7444, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080p_8iy}, {50, 1, 0, 8940, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080p_8iy}, {25, 1, 0, 8940, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080p_8iy}, {24000, 1001, 0, 7444, x264_cqm_avci100_1080_4ic, x264_cqm_avci100_1080p_8iy}}}};
        int res = -1;
        if (i_csp >= 0x0001 && i_csp < 0x0004 && !type) {
            if (h->param.i_width == 1440 && h->param.i_height == 1080)
                res = 1;
            else if (h->param.i_width == 960 && h->param.i_height == 720)
                res = 0;
        }
        else if (i_csp >= 0x0004 && i_csp < 0x0008 && type) {
            if (h->param.i_width == 1920 && h->param.i_height == 1080)
                res = 1;
            else if (h->param.i_width == 1280 && h->param.i_height == 720)
                res = 0;
        }
        else {
            x264_log (h, 0, "Invalid colorspace for AVC-Intra %d\n", h -> param.i_avcintra_class);
            return -1;
        }
        if (res < 0) {
            x264_log (h, 0, "Resolution %dx%d invalid for AVC-Intra %d\n", h -> param.i_width, h -> param.i_height, h -> param.i_avcintra_class);
            return -1;
        }
        if (h->param.nalu_process) {
            x264_log (h, 0, "nalu_process is not supported in AVC-Intra mode\n");
            return -1;
        }
        if (!h->param.b_repeat_headers) {
            x264_log (h, 0, "Separate headers not supported in AVC-Intra mode\n");
            return -1;
        }
        int i;
        uint32_t fps_num = h->param.i_fps_num, fps_den = h->param.i_fps_den;
        x264_reduce_fraction (& fps_num, & fps_den);
        for (i = 0; i < 7; i++) {
            if (avcintra_lut[type][res][i].fps_num == fps_num && avcintra_lut[type][res][i].fps_den == fps_den && avcintra_lut[type][res][i].interlaced == 0) {
                break;
            }
        }
        if (i == 7) {
            x264_log (h, 0, "FPS %d/%d%c not compatible with AVC-Intra\n", h -> param.i_fps_num, h -> param.i_fps_den, 0 ? 'i' : 'p');
            return -1;
        }
        h->param.i_keyint_max = 1;
        h->param.b_intra_refresh = 0;
        h->param.analyse.i_weighted_pred = 0;
        h->param.i_frame_reference = 1;
        h->param.i_dpb_size = 1;
        h->param.b_bluray_compat = 0;
        h->param.b_vfr_input = 0;
        h->param.b_aud = 1;
        h->param.vui.i_chroma_loc = 0;
        h->param.i_nal_hrd = 0;
        h->param.b_deblocking_filter = 0;
        h->param.b_stitchable = 1;
        h->param.b_pic_struct = 0;
        h->param.analyse.b_transform_8x8 = 1;
        h->param.analyse.intra = 0x0002;
        h->param.analyse.i_chroma_qp_offset = res && type ? 3 : 4;
        h->param.b_cabac = !type;
        h->param.rc.i_vbv_buffer_size = avcintra_lut[type][res][i].frame_size;
        h->param.rc.i_vbv_max_bitrate = h->param.rc.i_bitrate = h->param.rc.i_vbv_buffer_size * fps_num / fps_den;
        h->param.rc.i_rc_method = 2;
        h->param.rc.f_vbv_buffer_init = 1.0;
        h->param.rc.b_filler = 1;
        h->param.i_cqm_preset = 2;
        memcpy (h -> param.cqm_4iy, x264_cqm_jvt4i, sizeof (h -> param.cqm_4iy));
        memcpy (h -> param.cqm_4ic, avcintra_lut [type] [res] [i].cqm_4ic, sizeof (h -> param.cqm_4ic));
        memcpy (h -> param.cqm_8iy, avcintra_lut [type] [res] [i].cqm_8iy, sizeof (h -> param.cqm_8iy));
        h->param.i_slice_max_mbs = ((h->param.i_width + 15) / 16) * ((h->param.i_height + 15) / 16) / 10;
        h->param.i_slice_max_size = 0;
        if (h->param.b_sliced_threads) {
            if (res)
                h->param.i_threads = ((2) < (h->param.i_threads) ? (2) : (h->param.i_threads));
            else {
                h->param.i_threads = ((5) < (h->param.i_threads) ? (5) : (h->param.i_threads));
                if (h->param.i_threads < 5)
                    h->param.i_threads = 1;
            }
        }
        if (type)
            h->param.vui.i_sar_width = h->param.vui.i_sar_height = 1;
        else {
            h->param.vui.i_sar_width = 4;
            h->param.vui.i_sar_height = 3;
        }
        h->param.rc.i_qp_min = ((h->param.rc.i_qp_min) > ((6 * (8 - 8)) + 1) ? (h->param.rc.i_qp_min) : ((6 * (8 - 8)) + 1));
    }
    h->param.rc.f_rf_constant = x264_clip3f (h->param.rc.f_rf_constant, -(6 * (8 - 8)), 51);
    h->param.rc.f_rf_constant_max = x264_clip3f (h->param.rc.f_rf_constant_max, -(6 * (8 - 8)), 51);
    h->param.rc.i_qp_constant = x264_clip3 (h->param.rc.i_qp_constant, 0, ((51 + (6 * (8 - 8))) + 18));
    h->param.analyse.i_subpel_refine = x264_clip3 (h->param.analyse.i_subpel_refine, 0, 11);
    h->param.rc.f_ip_factor = ((h->param.rc.f_ip_factor) > (0.01f) ? (h->param.rc.f_ip_factor) : (0.01f));
    h->param.rc.f_pb_factor = ((h->param.rc.f_pb_factor) > (0.01f) ? (h->param.rc.f_pb_factor) : (0.01f));
    if (h->param.rc.i_rc_method == 1) {
        h->param.rc.i_qp_constant = h->param.rc.f_rf_constant + (6 * (8 - 8));
        h->param.rc.i_bitrate = 0;
    }
    if (b_open && (h->param.rc.i_rc_method == 0 || h->param.rc.i_rc_method == 1) && h->param.rc.i_qp_constant == 0) {
        h->mb.b_lossless = 1;
        h->param.i_cqm_preset = 0;
        h->param.psz_cqm_file = ((void *) 0);
        h->param.rc.i_rc_method = 0;
        h->param.rc.f_ip_factor = 1;
        h->param.rc.f_pb_factor = 1;
        h->param.analyse.b_psnr = 0;
        h->param.analyse.b_ssim = 0;
        h->param.analyse.i_chroma_qp_offset = 0;
        h->param.analyse.i_trellis = 0;
        h->param.analyse.b_fast_pskip = 0;
        h->param.analyse.i_noise_reduction = 0;
        h->param.analyse.b_psy = 0;
        h->param.i_bframe = 0;
        if (!h->param.b_cabac && h->param.analyse.i_subpel_refine < 6)
            h->param.analyse.b_transform_8x8 = 0;
        h->param.analyse.inter &= ~0x0002;
        h->param.analyse.intra &= ~0x0002;
    }
    if (h->param.rc.i_rc_method == 0) {
        float qp_p = h->param.rc.i_qp_constant;
        float qp_i = qp_p - 6 * log2f (h->param.rc.f_ip_factor);
        float qp_b = qp_p + 6 * log2f (h->param.rc.f_pb_factor);
        h->param.rc.i_qp_min = x264_clip3 ((int) ((((qp_p)) < ((((qp_i)) < ((qp_b)) ? ((qp_i)) : ((qp_b)))) ? ((qp_p)) : ((((qp_i)) < ((qp_b)) ? ((qp_i)) : ((qp_b)))))), 0, ((51 + (6 * (8 - 8))) + 18));
        h->param.rc.i_qp_max = x264_clip3 ((int) ((((qp_p)) > ((((qp_i)) > ((qp_b)) ? ((qp_i)) : ((qp_b)))) ? ((qp_p)) : ((((qp_i)) > ((qp_b)) ? ((qp_i)) : ((qp_b))))) + .999), 0, ((51 + (6 * (8 - 8))) + 18));
        h->param.rc.i_aq_mode = 0;
        h->param.rc.b_mb_tree = 0;
        h->param.rc.i_bitrate = 0;
    }
    h->param.rc.i_qp_max = x264_clip3 (h->param.rc.i_qp_max, 0, ((51 + (6 * (8 - 8))) + 18));
    h->param.rc.i_qp_min = x264_clip3 (h->param.rc.i_qp_min, 0, h->param.rc.i_qp_max);
    h->param.rc.i_qp_step = x264_clip3 (h->param.rc.i_qp_step, 2, ((51 + (6 * (8 - 8))) + 18));
    h->param.rc.i_bitrate = x264_clip3 (h->param.rc.i_bitrate, 0, 2000000);
    if (h->param.rc.i_rc_method == 2 && !h->param.rc.i_bitrate) {
        x264_log (h, 0, "bitrate not specified\n");
        return -1;
    }
    h->param.rc.i_vbv_buffer_size = x264_clip3 (h->param.rc.i_vbv_buffer_size, 0, 2000000);
    h->param.rc.i_vbv_max_bitrate = x264_clip3 (h->param.rc.i_vbv_max_bitrate, 0, 2000000);
    h->param.rc.f_vbv_buffer_init = x264_clip3f (h->param.rc.f_vbv_buffer_init, 0, 2000000);
    if (h->param.rc.i_vbv_buffer_size) {
        if (h->param.rc.i_rc_method == 0) {
            x264_log (h, 1, "VBV is incompatible with constant QP, ignored.\n");
            h->param.rc.i_vbv_max_bitrate = 0;
            h->param.rc.i_vbv_buffer_size = 0;
        }
        else if (h->param.rc.i_vbv_max_bitrate == 0) {
            if (h->param.rc.i_rc_method == 2) {
                x264_log (h, 1, "VBV maxrate unspecified, assuming CBR\n");
                h->param.rc.i_vbv_max_bitrate = h->param.rc.i_bitrate;
            }
            else {
                x264_log (h, 1, "VBV bufsize set but maxrate unspecified, ignored\n");
                h->param.rc.i_vbv_buffer_size = 0;
            }
        }
        else if (h->param.rc.i_vbv_max_bitrate < h->param.rc.i_bitrate && h->param.rc.i_rc_method == 2) {
            x264_log (h, 1, "max bitrate less than average bitrate, assuming CBR\n");
            h->param.rc.i_bitrate = h->param.rc.i_vbv_max_bitrate;
        }
    }
    else if (h->param.rc.i_vbv_max_bitrate) {
        x264_log (h, 1, "VBV maxrate specified, but no bufsize, ignored\n");
        h->param.rc.i_vbv_max_bitrate = 0;
    }
    h->param.i_slice_max_size = ((h->param.i_slice_max_size) > (0) ? (h->param.i_slice_max_size) : (0));
    h->param.i_slice_max_mbs = ((h->param.i_slice_max_mbs) > (0) ? (h->param.i_slice_max_mbs) : (0));
    h->param.i_slice_min_mbs = ((h->param.i_slice_min_mbs) > (0) ? (h->param.i_slice_min_mbs) : (0));
    if (h->param.i_slice_max_mbs)
        h->param.i_slice_min_mbs = ((h->param.i_slice_min_mbs) < (h->param.i_slice_max_mbs / 2) ? (h->param.i_slice_min_mbs) : (h->param.i_slice_max_mbs / 2));
    else if (!h->param.i_slice_max_size)
        h->param.i_slice_min_mbs = 0;
    if (0 && h->param.i_slice_min_mbs) {
        x264_log (h, 1, "interlace + slice-min-mbs is not implemented\n");
        h->param.i_slice_min_mbs = 0;
    }
    int mb_width = (h->param.i_width + 15) / 16;
    if (h->param.i_slice_min_mbs > mb_width) {
        x264_log (h, 1, "slice-min-mbs > row mb size (%d) not implemented\n", mb_width);
        h->param.i_slice_min_mbs = mb_width;
    }
    int max_slices = (h->param.i_height + ((16 << 0) - 1)) / (16 << 0);
    if (h->param.b_sliced_threads)
        h->param.i_slice_count = x264_clip3 (h->param.i_threads, 0, max_slices);
    else {
        h->param.i_slice_count = x264_clip3 (h->param.i_slice_count, 0, max_slices);
        if (h->param.i_slice_max_mbs || h->param.i_slice_max_size)
            h->param.i_slice_count = 0;
    }
    if (h->param.i_slice_count_max > 0)
        h->param.i_slice_count_max = ((h->param.i_slice_count) > (h->param.i_slice_count_max) ? (h->param.i_slice_count) : (h->param.i_slice_count_max));
    if (h->param.b_bluray_compat) {
        h->param.i_bframe_pyramid = ((1) < (h->param.i_bframe_pyramid) ? (1) : (h->param.i_bframe_pyramid));
        h->param.i_bframe = ((h->param.i_bframe) < (3) ? (h->param.i_bframe) : (3));
        h->param.b_aud = 1;
        h->param.i_nal_hrd = ((h->param.i_nal_hrd) > (1) ? (h->param.i_nal_hrd) : (1));
        h->param.i_slice_max_size = 0;
        h->param.i_slice_max_mbs = 0;
        h->param.b_intra_refresh = 0;
        h->param.i_frame_reference = ((h->param.i_frame_reference) < (6) ? (h->param.i_frame_reference) : (6));
        h->param.i_dpb_size = ((h->param.i_dpb_size) < (6) ? (h->param.i_dpb_size) : (6));
        h->param.i_keyint_min = 1;
        h->param.analyse.i_weighted_pred = ((h->param.analyse.i_weighted_pred) < (1) ? (h->param.analyse.i_weighted_pred) : (1));
        if (h->param.b_fake_interlaced)
            h->param.b_pic_struct = 1;
    }
    h->param.i_frame_reference = x264_clip3 (h->param.i_frame_reference, 1, 16);
    h->param.i_dpb_size = x264_clip3 (h->param.i_dpb_size, 1, 16);
    if (h->param.i_scenecut_threshold < 0)
        h->param.i_scenecut_threshold = 0;
    h->param.analyse.i_direct_mv_pred = x264_clip3 (h->param.analyse.i_direct_mv_pred, 0, 3);
    if (!h->param.analyse.i_subpel_refine && h->param.analyse.i_direct_mv_pred > 1) {
        x264_log (h, 1, "subme=0 + direct=temporal is not supported\n");
        h->param.analyse.i_direct_mv_pred = 1;
    }
    h->param.i_bframe = x264_clip3 (h->param.i_bframe, 0, ((16) < (h->param.i_keyint_max - 1) ? (16) : (h->param.i_keyint_max - 1)));
    h->param.i_bframe_bias = x264_clip3 (h->param.i_bframe_bias, -90, 100);
    if (h->param.i_bframe <= 1)
        h->param.i_bframe_pyramid = 0;
    h->param.i_bframe_pyramid = x264_clip3 (h->param.i_bframe_pyramid, 0, 2);
    h->param.i_bframe_adaptive = x264_clip3 (h->param.i_bframe_adaptive, 0, 2);
    if (!h->param.i_bframe) {
        h->param.i_bframe_adaptive = 0;
        h->param.analyse.i_direct_mv_pred = 0;
        h->param.analyse.b_weighted_bipred = 0;
        h->param.b_open_gop = 0;
    }
    if (h->param.b_intra_refresh && h->param.i_bframe_pyramid == 2) {
        x264_log (h, 1, "b-pyramid normal + intra-refresh is not supported\n");
        h->param.i_bframe_pyramid = 1;
    }
    if (h->param.b_intra_refresh && (h->param.i_frame_reference > 1 || h->param.i_dpb_size > 1)) {
        x264_log (h, 1, "ref > 1 + intra-refresh is not supported\n");
        h->param.i_frame_reference = 1;
        h->param.i_dpb_size = 1;
    }
    if (h->param.b_intra_refresh && h->param.b_open_gop) {
        x264_log (h, 1, "intra-refresh is not compatible with open-gop\n");
        h->param.b_open_gop = 0;
    }
    if (!h->param.i_fps_num || !h->param.i_fps_den) {
        h->param.i_fps_num = 25;
        h->param.i_fps_den = 1;
    }
    float fps = (float) h->param.i_fps_num / h->param.i_fps_den;
    if (h->param.i_keyint_min == 0)
        h->param.i_keyint_min = ((h->param.i_keyint_max / 10) < (fps) ? (h->param.i_keyint_max / 10) : (fps));
    h->param.i_keyint_min = x264_clip3 (h->param.i_keyint_min, 1, h->param.i_keyint_max / 2 + 1);
    h->param.rc.i_lookahead = x264_clip3 (h->param.rc.i_lookahead, 0, 250);
    {
        int maxrate = ((h->param.rc.i_vbv_max_bitrate) > (h->param.rc.i_bitrate) ? (h->param.rc.i_vbv_max_bitrate) : (h->param.rc.i_bitrate));
        float bufsize = maxrate ? (float) h->param.rc.i_vbv_buffer_size / maxrate : 0;
        h->param.rc.i_lookahead = ((h->param.rc.i_lookahead) < (((h->param.i_keyint_max) > (bufsize * fps) ? (h->param.i_keyint_max) : (bufsize * fps))) ? (h->param.rc.i_lookahead) : (((h->param.i_keyint_max) > (bufsize * fps) ? (h->param.i_keyint_max) : (bufsize * fps))));
    }
    if (!h->param.i_timebase_num || !h->param.i_timebase_den || !(h->param.b_vfr_input || h->param.b_pulldown)) {
        h->param.i_timebase_num = h->param.i_fps_den;
        h->param.i_timebase_den = h->param.i_fps_num;
    }
    h->param.rc.f_qcompress = x264_clip3f (h->param.rc.f_qcompress, 0.0, 1.0);
    if (h->param.i_keyint_max == 1 || h->param.rc.f_qcompress == 1)
        h->param.rc.b_mb_tree = 0;
    if ((!h->param.b_intra_refresh && h->param.i_keyint_max != (1 << 30)) && !h->param.rc.i_lookahead && h->param.rc.b_mb_tree) {
        x264_log (h, 1, "lookaheadless mb-tree requires intra refresh or infinite keyint\n");
        h->param.rc.b_mb_tree = 0;
    }
    if (b_open && h->param.rc.b_stat_read)
        h->param.rc.i_lookahead = 0;
    h->param.i_sync_lookahead = 0;
    h->param.i_deblocking_filter_alphac0 = x264_clip3 (h->param.i_deblocking_filter_alphac0, -6, 6);
    h->param.i_deblocking_filter_beta = x264_clip3 (h->param.i_deblocking_filter_beta, -6, 6);
    h->param.analyse.i_luma_deadzone[0] = x264_clip3 (h->param.analyse.i_luma_deadzone[0], 0, 32);
    h->param.analyse.i_luma_deadzone[1] = x264_clip3 (h->param.analyse.i_luma_deadzone[1], 0, 32);
    h->param.i_cabac_init_idc = x264_clip3 (h->param.i_cabac_init_idc, 0, 2);
    if (h->param.i_cqm_preset < 0 || h->param.i_cqm_preset > 2)
        h->param.i_cqm_preset = 0;
    if (h->param.analyse.i_me_method < 0 || h->param.analyse.i_me_method > 4)
        h->param.analyse.i_me_method = 1;
    h->param.analyse.i_me_range = x264_clip3 (h->param.analyse.i_me_range, 4, 1024);
    if (h->param.analyse.i_me_range > 16 && h->param.analyse.i_me_method <= 1)
        h->param.analyse.i_me_range = 16;
    if (h->param.analyse.i_me_method == 4 && (h->mb.b_lossless || h->param.analyse.i_subpel_refine <= 1))
        h->param.analyse.i_me_method = 3;
    h->param.analyse.b_mixed_references = h->param.analyse.b_mixed_references && h->param.i_frame_reference > 1;
    h->param.analyse.inter &= 0x0010 | 0x0020 | 0x0100 | 0x0001 | 0x0002;
    h->param.analyse.intra &= 0x0001 | 0x0002;
    if (!(h->param.analyse.inter & 0x0010))
        h->param.analyse.inter &= ~0x0020;
    if (!h->param.analyse.b_transform_8x8) {
        h->param.analyse.inter &= ~0x0002;
        h->param.analyse.intra &= ~0x0002;
    }
    h->param.analyse.i_trellis = x264_clip3 (h->param.analyse.i_trellis, 0, 2);
    h->param.rc.i_aq_mode = x264_clip3 (h->param.rc.i_aq_mode, 0, 2);
    h->param.rc.f_aq_strength = x264_clip3f (h->param.rc.f_aq_strength, 0, 3);
    if (h->param.rc.f_aq_strength == 0)
        h->param.rc.i_aq_mode = 0;
    if (h->param.i_log_level < 2) {
        h->param.analyse.b_psnr = 0;
        h->param.analyse.b_ssim = 0;
    }
    if (b_open && (h->param.analyse.b_psnr || h->param.analyse.b_ssim)) {
        char *s = ((void *) 0);
        if (h->param.analyse.b_psy) {
            s = h->param.analyse.b_psnr ? "psnr" : "ssim";
            x264_log (h, 1, "--%s used with psy on: results will be invalid!\n", s);
        }
        else if (!h->param.rc.i_aq_mode && h->param.analyse.b_ssim) {
            x264_log (h, 1, "--ssim used with AQ off: results will be invalid!\n");
            s = "ssim";
        }
        else if (h->param.rc.i_aq_mode && h->param.analyse.b_psnr) {
            x264_log (h, 1, "--psnr used with AQ on: results will be invalid!\n");
            s = "psnr";
        }
        if (s)
            x264_log (h, 1, "--tune %s should be used if attempting to benchmark %s!\n", s, s);
    }
    if (!h->param.analyse.b_psy) {
        h->param.analyse.f_psy_rd = 0;
        h->param.analyse.f_psy_trellis = 0;
    }
    h->param.analyse.f_psy_rd = x264_clip3f (h->param.analyse.f_psy_rd, 0, 10);
    h->param.analyse.f_psy_trellis = x264_clip3f (h->param.analyse.f_psy_trellis, 0, 10);
    h->mb.i_psy_rd = h->param.analyse.i_subpel_refine >= 6 ? ((int) (h->param.analyse.f_psy_rd * (1 << 8) + .5)) : 0;
    h->mb.i_psy_trellis = h->param.analyse.i_trellis ? ((int) (h->param.analyse.f_psy_trellis / 4 * (1 << 8) + .5)) : 0;
    h->param.analyse.i_chroma_qp_offset = x264_clip3 (h->param.analyse.i_chroma_qp_offset, -32, 32);
    if (b_open && i_csp >= 0x0008 && i_csp < 0x000a && h->param.analyse.b_psy)
        h->param.analyse.i_chroma_qp_offset += 6;
    if (b_open && h->mb.i_psy_rd && !h->param.i_avcintra_class)
        h->param.analyse.i_chroma_qp_offset -= h->param.analyse.f_psy_rd < 0.25 ? 1 : 2;
    if (b_open && h->mb.i_psy_trellis && !h->param.i_avcintra_class)
        h->param.analyse.i_chroma_qp_offset -= h->param.analyse.f_psy_trellis < 0.25 ? 1 : 2;
    h->param.analyse.i_chroma_qp_offset = x264_clip3 (h->param.analyse.i_chroma_qp_offset, -12, 12);
    if (!h->param.rc.i_aq_mode && h->param.rc.b_mb_tree) {
        h->param.rc.i_aq_mode = 1;
        h->param.rc.f_aq_strength = 0;
    }
    h->param.analyse.i_noise_reduction = x264_clip3 (h->param.analyse.i_noise_reduction, 0, 1 << 16);
    if (h->param.analyse.i_subpel_refine >= 10 && (h->param.analyse.i_trellis != 2 || !h->param.rc.i_aq_mode))
        h->param.analyse.i_subpel_refine = 9;
    {
        const x264_level_t *l = x264_levels;
        if (h->param.i_level_idc < 0) {
            int maxrate_bak = h->param.rc.i_vbv_max_bitrate;
            if (h->param.rc.i_rc_method == 2 && h->param.rc.i_vbv_buffer_size <= 0)
                h->param.rc.i_vbv_max_bitrate = h->param.rc.i_bitrate * 2;
            x264_sps_init (h -> sps, h -> param.i_sps_id, & h -> param);
            do
                h->param.i_level_idc = l->level_idc;
            while (l[1].level_idc && x264_validate_levels (h, 0) && l++);
            h->param.rc.i_vbv_max_bitrate = maxrate_bak;
        }
        else {
            while (l->level_idc && l->level_idc != h->param.i_level_idc)
                l++;
            if (l->level_idc == 0) {
                x264_log (h, 0, "invalid level_idc: %d\n", h -> param.i_level_idc);
                return -1;
            }
        }
        if (h->param.analyse.i_mv_range <= 0)
            h->param.analyse.i_mv_range = l->mv_range >> 0;
        else
            h->param.analyse.i_mv_range = x264_clip3 (h->param.analyse.i_mv_range, 32, 512 >> 0);
    }
    h->param.analyse.i_weighted_pred = x264_clip3 (h->param.analyse.i_weighted_pred, 0, 2);
    if (h->param.i_lookahead_threads == 0) {
        if (h->param.b_sliced_threads)
            h->param.i_lookahead_threads = h->param.i_threads;
        else {
            int badapt = h->param.i_bframe_adaptive == 2;
            int subme = ((h->param.analyse.i_subpel_refine / 3) < (3) ? (h->param.analyse.i_subpel_refine / 3) : (3)) + (h->param.analyse.i_subpel_refine > 1);
            int bframes = (((h->param.i_bframe - 1) / 3) < (3) ? ((h->param.i_bframe - 1) / 3) : (3));
            static const uint8_t lookahead_thread_div [2] [5] [4] = {{{6, 6, 6, 6}, {3, 3, 3, 3}, {4, 4, 4, 4}, {6, 6, 6, 6}, {12, 12, 12, 12}}, {{3, 2, 1, 1}, {2, 1, 1, 1}, {4, 3, 2, 1}, {6, 4, 3, 2}, {12, 9, 6, 4}}};
            h->param.i_lookahead_threads = h->param.i_threads / lookahead_thread_div[badapt][subme][bframes];
            h->param.i_lookahead_threads = ((h->param.i_lookahead_threads) < (h->param.i_height / 128) ? (h->param.i_lookahead_threads) : (h->param.i_height / 128));
        }
    }
    h->param.i_lookahead_threads = x264_clip3 (h->param.i_lookahead_threads, 1, ((max_sliced_threads) < (16) ? (max_sliced_threads) : (16)));
    if (0) {
        if (h->param.analyse.i_me_method >= 3) {
            x264_log (h, 1, "interlace + me=esa is not implemented\n");
            h->param.analyse.i_me_method = 2;
        }
        if (h->param.analyse.i_weighted_pred > 0) {
            x264_log (h, 1, "interlace + weightp is not implemented\n");
            h->param.analyse.i_weighted_pred = 0;
        }
    }
    if (!h->param.analyse.i_weighted_pred && h->param.rc.b_mb_tree && h->param.analyse.b_psy)
        h->param.analyse.i_weighted_pred = (-1);
    if (h->i_thread_frames > 1) {
        int r = h->param.analyse.i_mv_range_thread;
        int r2;
        if (r <= 0) {
            int max_range = (h->param.i_height + 24) / h->i_thread_frames - 24;
            r = max_range / 2;
        }
        r = ((r) > (h->param.analyse.i_me_range) ? (r) : (h->param.analyse.i_me_range));
        r = ((r) < (h->param.analyse.i_mv_range) ? (r) : (h->param.analyse.i_mv_range));
        r2 = (r & ~15) + ((-24) & 15);
        if (r2 < r)
            r2 += 16;
        x264_log (h, 3, "using mv_range_thread = %d\n", r2);
        h->param.analyse.i_mv_range_thread = r2;
    }
    if (h->param.rc.f_rate_tolerance < 0)
        h->param.rc.f_rate_tolerance = 0;
    if (h->param.rc.f_qblur < 0)
        h->param.rc.f_qblur = 0;
    if (h->param.rc.f_complexity_blur < 0)
        h->param.rc.f_complexity_blur = 0;
    h->param.i_sps_id &= 31;
    h->param.i_nal_hrd = x264_clip3 (h->param.i_nal_hrd, 0, 2);
    if (h->param.i_nal_hrd && !h->param.rc.i_vbv_buffer_size) {
        x264_log (h, 1, "NAL HRD parameters require VBV parameters\n");
        h->param.i_nal_hrd = 0;
    }
    if (h->param.i_nal_hrd == 2 && (h->param.rc.i_bitrate != h->param.rc.i_vbv_max_bitrate || !h->param.rc.i_vbv_max_bitrate)) {
        x264_log (h, 1, "CBR HRD requires constant bitrate\n");
        h->param.i_nal_hrd = 1;
    }
    if (h->param.i_nal_hrd == 2)
        h->param.rc.b_filler = 1;
    h->param.b_cabac = !!h->param.b_cabac;
    h->param.b_constrained_intra = !!h->param.b_constrained_intra;
    h->param.b_deblocking_filter = !!h->param.b_deblocking_filter;
    h->param.b_deterministic = !!h->param.b_deterministic;
    h->param.b_sliced_threads = !!h->param.b_sliced_threads;
    h->param.b_interlaced = !!h->param.b_interlaced;
    h->param.b_intra_refresh = !!h->param.b_intra_refresh;
    h->param.b_aud = !!h->param.b_aud;
    h->param.b_repeat_headers = !!h->param.b_repeat_headers;
    h->param.b_annexb = !!h->param.b_annexb;
    h->param.b_vfr_input = !!h->param.b_vfr_input;
    h->param.b_pulldown = !!h->param.b_pulldown;
    h->param.b_tff = !!h->param.b_tff;
    h->param.b_pic_struct = !!h->param.b_pic_struct;
    h->param.b_fake_interlaced = !!h->param.b_fake_interlaced;
    h->param.b_open_gop = !!h->param.b_open_gop;
    h->param.b_bluray_compat = !!h->param.b_bluray_compat;
    h->param.b_stitchable = !!h->param.b_stitchable;
    h->param.b_full_recon = !!h->param.b_full_recon;
    h->param.b_opencl = !!h->param.b_opencl;
    h->param.analyse.b_transform_8x8 = !!h->param.analyse.b_transform_8x8;
    h->param.analyse.b_weighted_bipred = !!h->param.analyse.b_weighted_bipred;
    h->param.analyse.b_chroma_me = !!h->param.analyse.b_chroma_me;
    h->param.analyse.b_mixed_references = !!h->param.analyse.b_mixed_references;
    h->param.analyse.b_fast_pskip = !!h->param.analyse.b_fast_pskip;
    h->param.analyse.b_dct_decimate = !!h->param.analyse.b_dct_decimate;
    h->param.analyse.b_psy = !!h->param.analyse.b_psy;
    h->param.analyse.b_psnr = !!h->param.analyse.b_psnr;
    h->param.analyse.b_ssim = !!h->param.analyse.b_ssim;
    h->param.rc.b_stat_write = !!h->param.rc.b_stat_write;
    h->param.rc.b_stat_read = !!h->param.rc.b_stat_read;
    h->param.rc.b_mb_tree = !!h->param.rc.b_mb_tree;
    h->param.rc.b_filler = !!h->param.rc.b_filler;
    return 0;
}

void mbcmp_init (x264_t *h) {
    int satd = !h->mb.b_lossless && h->param.analyse.i_subpel_refine > 1;
    memcpy (h -> pixf.mbcmp, satd ? h -> pixf.satd : h -> pixf.sad_aligned, sizeof (h -> pixf.mbcmp));
    memcpy (h -> pixf.mbcmp_unaligned, satd ? h -> pixf.satd : h -> pixf.sad, sizeof (h -> pixf.mbcmp_unaligned));
    h->pixf.intra_mbcmp_x3_16x16 = satd ? h->pixf.intra_satd_x3_16x16 : h->pixf.intra_sad_x3_16x16;
    h->pixf.intra_mbcmp_x3_8x16c = satd ? h->pixf.intra_satd_x3_8x16c : h->pixf.intra_sad_x3_8x16c;
    h->pixf.intra_mbcmp_x3_8x8c = satd ? h->pixf.intra_satd_x3_8x8c : h->pixf.intra_sad_x3_8x8c;
    h->pixf.intra_mbcmp_x3_8x8 = satd ? h->pixf.intra_sa8d_x3_8x8 : h->pixf.intra_sad_x3_8x8;
    h->pixf.intra_mbcmp_x3_4x4 = satd ? h->pixf.intra_satd_x3_4x4 : h->pixf.intra_sad_x3_4x4;
    h->pixf.intra_mbcmp_x9_4x4 = h->param.b_cpu_independent || h->mb.b_lossless ? ((void *) 0) : satd ? h->pixf.intra_satd_x9_4x4 : h->pixf.intra_sad_x9_4x4;
    h->pixf.intra_mbcmp_x9_8x8 = h->param.b_cpu_independent || h->mb.b_lossless ? ((void *) 0) : satd ? h->pixf.intra_sa8d_x9_8x8 : h->pixf.intra_sad_x9_8x8;
    satd &= h->param.analyse.i_me_method == 4;
    memcpy (h -> pixf.fpelcmp, satd ? h -> pixf.satd : h -> pixf.sad, sizeof (h -> pixf.fpelcmp));
    memcpy (h -> pixf.fpelcmp_x3, satd ? h -> pixf.satd_x3 : h -> pixf.sad_x3, sizeof (h -> pixf.fpelcmp_x3));
    memcpy (h -> pixf.fpelcmp_x4, satd ? h -> pixf.satd_x4 : h -> pixf.sad_x4, sizeof (h -> pixf.fpelcmp_x4));
}

void chroma_dsp_init (x264_t *h) {
    memcpy (h -> luma2chroma_pixel, x264_luma2chroma_pixel [h -> sps -> i_chroma_format_idc], sizeof (h -> luma2chroma_pixel));
    switch (h->sps->i_chroma_format_idc) {
    case CHROMA_420 :
        memcpy (h->predict_chroma, h->predict_8x8c, sizeof (h->predict_chroma));
        h->mc.prefetch_fenc = h->mc.prefetch_fenc_420;
        h->loopf.deblock_chroma[0] = h->loopf.deblock_h_chroma_420;
        h->loopf.deblock_chroma_intra[0] = h->loopf.deblock_h_chroma_420_intra;
        h->loopf.deblock_chroma_mbaff = h->loopf.deblock_chroma_420_mbaff;
        h->loopf.deblock_chroma_intra_mbaff = h->loopf.deblock_chroma_420_intra_mbaff;
        h->pixf.intra_mbcmp_x3_chroma = h->pixf.intra_mbcmp_x3_8x8c;
        h->quantf.coeff_last[DCT_CHROMA_DC] = h->quantf.coeff_last4;
        h->quantf.coeff_level_run[DCT_CHROMA_DC] = h->quantf.coeff_level_run4;
        break;
    case CHROMA_422 :
        memcpy (h->predict_chroma, h->predict_8x16c, sizeof (h->predict_chroma));
        h->mc.prefetch_fenc = h->mc.prefetch_fenc_422;
        h->loopf.deblock_chroma[0] = h->loopf.deblock_h_chroma_422;
        h->loopf.deblock_chroma_intra[0] = h->loopf.deblock_h_chroma_422_intra;
        h->loopf.deblock_chroma_mbaff = h->loopf.deblock_chroma_422_mbaff;
        h->loopf.deblock_chroma_intra_mbaff = h->loopf.deblock_chroma_422_intra_mbaff;
        h->pixf.intra_mbcmp_x3_chroma = h->pixf.intra_mbcmp_x3_8x16c;
        h->quantf.coeff_last[DCT_CHROMA_DC] = h->quantf.coeff_last8;
        h->quantf.coeff_level_run[DCT_CHROMA_DC] = h->quantf.coeff_level_run8;
        break;
    case CHROMA_444 :
        h->mc.prefetch_fenc = h->mc.prefetch_fenc_422;
        h->loopf.deblock_chroma_mbaff = h->loopf.deblock_luma_mbaff;
        h->loopf.deblock_chroma_intra_mbaff = h->loopf.deblock_luma_intra_mbaff;
        break;
    }
}

void x264_set_aspect_ratio (x264_t *h, x264_param_t *param, int initial) {
    if (param->vui.i_sar_width > 0 && param->vui.i_sar_height > 0) {
        uint32_t i_w = param->vui.i_sar_width;
        uint32_t i_h = param->vui.i_sar_height;
        uint32_t old_w = h->param.vui.i_sar_width;
        uint32_t old_h = h->param.vui.i_sar_height;
        x264_reduce_fraction (& i_w, & i_h);
        while (i_w > 65535 || i_h > 65535) {
            i_w /= 2;
            i_h /= 2;
        }
        x264_reduce_fraction (& i_w, & i_h);
        if (i_w != old_w || i_h != old_h || initial) {
            h->param.vui.i_sar_width = 0;
            h->param.vui.i_sar_height = 0;
            if (i_w == 0 || i_h == 0)
                x264_log (h, 1, "cannot create valid sample aspect ratio\n");
            else {
                x264_log (h, initial ? 2 : 3, "using SAR=%d/%d\n", i_w, i_h);
                h->param.vui.i_sar_width = i_w;
                h->param.vui.i_sar_height = i_h;
            }
        }
    }
}

x264_t *x264_encoder_open_142 (x264_param_t *param) {
    x264_t *h;
    char buf [1000], *p;
    int qp, i_slicetype_length;
    do {
        do {
            h = x264_malloc (sizeof (x264_t));
            if (!h)
                goto fail;
        }
        while (0);
        memset (h, 0, sizeof (x264_t));
    }
    while (0);
    memcpy (& h -> param, param, sizeof (x264_param_t));
    if (param->param_free)
        param->param_free (param);
    if (0) {
        x264_log (h, 0, "unable to initialize threading\n");
        goto fail;
    }
    if (x264_validate_parameters (h, 1) < 0)
        goto fail;
    if (h->param.psz_cqm_file)
        if (x264_cqm_parse_file (h, h->param.psz_cqm_file) < 0)
            goto fail;
    if (h->param.rc.psz_stat_out)
        h->param.rc.psz_stat_out = strdup (h->param.rc.psz_stat_out);
    if (h->param.rc.psz_stat_in)
        h->param.rc.psz_stat_in = strdup (h->param.rc.psz_stat_in);
    x264_reduce_fraction (& h -> param.i_fps_num, & h -> param.i_fps_den);
    x264_reduce_fraction (& h -> param.i_timebase_num, & h -> param.i_timebase_den);
    h->i_frame = -1;
    h->i_frame_num = 0;
    if (h->param.i_avcintra_class)
        h->i_idr_pic_id = 5;
    else
        h->i_idr_pic_id = 0;
    if ((uint64_t) h->param.i_timebase_den * 2 > (4294967295U)) {
        x264_log (h, 0, "Effective timebase denominator %u exceeds H.264 maximum\n", h -> param.i_timebase_den);
        goto fail;
    }
    x264_set_aspect_ratio (h, & h -> param, 1);
    x264_sps_init (h -> sps, h -> param.i_sps_id, & h -> param);
    x264_pps_init (h -> pps, h -> param.i_sps_id, & h -> param, h -> sps);
    x264_validate_levels (h, 1);
    h->chroma_qp_table = i_chroma_qp_table + 12 + h->pps->i_chroma_qp_index_offset;
    if (x264_cqm_init (h) < 0)
        goto fail;
    h->mb.i_mb_width = h->sps->i_mb_width;
    h->mb.i_mb_height = h->sps->i_mb_height;
    h->mb.i_mb_count = h->mb.i_mb_width * h->mb.i_mb_height;
    h->mb.chroma_h_shift = h->sps->i_chroma_format_idc == CHROMA_420 || h->sps->i_chroma_format_idc == CHROMA_422;
    h->mb.chroma_v_shift = h->sps->i_chroma_format_idc == CHROMA_420;
    h->mb.b_adaptive_mbaff = 0 && h->param.analyse.i_subpel_refine;
    if (h->param.i_bframe_adaptive == 2 && !h->param.rc.b_stat_read)
        h->frames.i_delay = ((h->param.i_bframe) > (3) ? (h->param.i_bframe) : (3)) * 4;
    else
        h->frames.i_delay = h->param.i_bframe;
    if (h->param.rc.b_mb_tree || h->param.rc.i_vbv_buffer_size)
        h->frames.i_delay = ((h->frames.i_delay) > (h->param.rc.i_lookahead) ? (h->frames.i_delay) : (h->param.rc.i_lookahead));
    i_slicetype_length = h->frames.i_delay;
    h->frames.i_delay += h->i_thread_frames - 1;
    h->frames.i_delay += h->param.i_sync_lookahead;
    h->frames.i_delay += h->param.b_vfr_input;
    h->frames.i_bframe_delay = h->param.i_bframe ? (h->param.i_bframe_pyramid ? 2 : 1) : 0;
    h->frames.i_max_ref0 = h->param.i_frame_reference;
    h->frames.i_max_ref1 = ((h->sps->vui.i_num_reorder_frames) < (h->param.i_frame_reference) ? (h->sps->vui.i_num_reorder_frames) : (h->param.i_frame_reference));
    h->frames.i_max_dpb = h->sps->vui.i_max_dec_frame_buffering;
    h->frames.b_have_lowres = !h->param.rc.b_stat_read && (h->param.rc.i_rc_method == 2 || h->param.rc.i_rc_method == 1 || h->param.i_bframe_adaptive || h->param.i_scenecut_threshold || h->param.rc.b_mb_tree || h->param.analyse.i_weighted_pred);
    h->frames.b_have_lowres |= h->param.rc.b_stat_read && h->param.rc.i_vbv_buffer_size > 0;
    h->frames.b_have_sub8x8_esa = !!(h->param.analyse.inter & 0x0020);
    h->frames.i_last_idr = h->frames.i_last_keyframe = -h->param.i_keyint_max;
    h->frames.i_input = 0;
    h->frames.i_largest_pts = h->frames.i_second_largest_pts = -1;
    h->frames.i_poc_last_open_gop = -1;
    do {
        do {
            h->frames.unused[0] = x264_malloc ((h->frames.i_delay + 3) * sizeof (x264_frame_t *));
            if (!h->frames.unused[0])
                goto fail;
        }
        while (0);
        memset (h -> frames.unused [0], 0, (h -> frames.i_delay + 3) * sizeof (x264_frame_t *));
    }
    while (0);
    do {
        do {
            h->frames.unused[1] = x264_malloc ((h->i_thread_frames + 16 + 4) * sizeof (x264_frame_t *));
            if (!h->frames.unused[1])
                goto fail;
        }
        while (0);
        memset (h -> frames.unused [1], 0, (h -> i_thread_frames + 16 + 4) * sizeof (x264_frame_t *));
    }
    while (0);
    do {
        do {
            h->frames.current = x264_malloc ((h->param.i_sync_lookahead + h->param.i_bframe + h->i_thread_frames + 3) * sizeof (x264_frame_t *));
            if (!h->frames.current)
                goto fail;
        }
        while (0);
        memset (h -> frames.current, 0, (h -> param.i_sync_lookahead + h -> param.i_bframe + h -> i_thread_frames + 3) * sizeof (x264_frame_t *));
    }
    while (0);
    if (h->param.analyse.i_weighted_pred > 0)
        do {
            do {
                h->frames.blank_unused = x264_malloc (h->i_thread_frames * 4 * sizeof (x264_frame_t *));
                if (!h->frames.blank_unused)
                    goto fail;
            }
            while (0);
            memset (h -> frames.blank_unused, 0, h -> i_thread_frames * 4 * sizeof (x264_frame_t *));
        }
        while (0);
    h->i_ref[0] = h->i_ref[1] = 0;
    h->i_cpb_delay = h->i_coded_fields = h->i_disp_fields = 0;
    h->i_prev_duration = ((uint64_t) h->param.i_fps_den * h->sps->vui.i_time_scale) / ((uint64_t) h->param.i_fps_num * h->sps->vui.i_num_units_in_tick);
    h->i_disp_fields_last_frame = -1;
    x264_rdo_init ();
    x264_predict_16x16_init (h -> param.cpu, h -> predict_16x16);
    x264_predict_8x8c_init (h -> param.cpu, h -> predict_8x8c);
    x264_predict_8x16c_init (h -> param.cpu, h -> predict_8x16c);
    x264_predict_8x8_init (h -> param.cpu, h -> predict_8x8, & h -> predict_8x8_filter);
    x264_predict_4x4_init (h -> param.cpu, h -> predict_4x4);
    x264_pixel_init (h -> param.cpu, & h -> pixf);
    x264_dct_init (h -> param.cpu, & h -> dctf);
    x264_zigzag_init (h -> param.cpu, & h -> zigzagf_progressive, & h -> zigzagf_interlaced);
    memcpy (& h -> zigzagf, 0 ? & h -> zigzagf_interlaced : & h -> zigzagf_progressive, sizeof (h -> zigzagf));
    x264_mc_init (h -> param.cpu, & h -> mc, h -> param.b_cpu_independent);
    x264_quant_init (h, h -> param.cpu, & h -> quantf);
    x264_deblock_init (h -> param.cpu, & h -> loopf, 0);
    x264_bitstream_init (h -> param.cpu, & h -> bsf);
    if (h->param.b_cabac)
        x264_cabac_init (h);
    mbcmp_init (h);
    chroma_dsp_init (h);
    p = buf + sprintf (buf, "using cpu capabilities:");
    for (int i = 0; x264_cpu_names[i].flags; i++) {
        if (!strcmp (x264_cpu_names[i].name, "SSE") && h->param.cpu & (0x0000010))
            continue;
        if (!strcmp (x264_cpu_names[i].name, "SSE2") && h->param.cpu & (0x0100000 | 0x0080000))
            continue;
        if (!strcmp (x264_cpu_names[i].name, "SSE3") && (h->param.cpu & 0x0000040 || !(h->param.cpu & 0x0040000)))
            continue;
        if (!strcmp (x264_cpu_names[i].name, "SSE4.1") && (h->param.cpu & 0x0000100))
            continue;
        if (!strcmp (x264_cpu_names[i].name, "BMI1") && (h->param.cpu & 0x0010000))
            continue;
        if ((h->param.cpu & x264_cpu_names[i].flags) == x264_cpu_names[i].flags && (!i || x264_cpu_names[i].flags != x264_cpu_names[i - 1].flags))
            p += sprintf (p, " %s", x264_cpu_names[i].name);
    }
    if (!h->param.cpu)
        p += sprintf (p, " none!");
    x264_log (h, 2, "%s\n", buf);
    float *logs = x264_analyse_prepare_costs (h);
    if (!logs)
        goto fail;
    for (qp = ((h->param.rc.i_qp_min) < ((51 + (6 * (8 - 8)))) ? (h->param.rc.i_qp_min) : ((51 + (6 * (8 - 8))))); qp <= h->param.rc.i_qp_max; qp++)
        if (x264_analyse_init_costs (h, logs, qp))
            goto fail;
    if (x264_analyse_init_costs (h, logs, (12 + (6 * (8 - 8)))))
        goto fail;
    x264_free (logs);
    static const uint16_t cost_mv_correct [7] = {24, 47, 95, 189, 379, 757, 1515};
    if (h->cost_mv[(12 + (6 * (8 - 8)))][2013] != cost_mv_correct[8 - 8]) {
        x264_log (h, 0, "MV cost test failed: x264 has been miscompiled!\n");
        goto fail;
    }
    volatile int temp = 392;
    if (__builtin_clz (temp) != 23) {
        x264_log (h, 0, "CLZ test failed: x264 has been miscompiled!\n");
        x264_log (h, 0, "Are you attempting to run an SSE4a/LZCNT-targeted build on a CPU that\n");
        x264_log (h, 0, "doesn't support it?\n");
        goto fail;
    }
    h->out.i_nal = 0;
    h->out.i_bitstream = ((1000000) > (h->param.i_width * h->param.i_height * 4 * (h->param.rc.i_rc_method == 2 ? pow (0.95, h->param.rc.i_qp_min) : pow (0.95, h->param.rc.i_qp_constant) * ((1) > (h->param.rc.f_ip_factor) ? (1) : (h->param.rc.f_ip_factor)))) ? (1000000) : (h->param.i_width * h->param.i_height * 4 * (h->param.rc.i_rc_method == 2 ? pow (0.95, h->param.rc.i_qp_min) : pow (0.95, h->param.rc.i_qp_constant) * ((1) > (h->param.rc.f_ip_factor) ? (1) : (h->param.rc.f_ip_factor)))));
    h->nal_buffer_size = h->out.i_bitstream * 3 / 2 + 4 + 64;
    do {
        h->nal_buffer = x264_malloc (h->nal_buffer_size);
        if (!h->nal_buffer)
            goto fail;
    }
    while (0);
    do {
        h->reconfig_h = x264_malloc (sizeof (x264_t));
        if (!h->reconfig_h)
            goto fail;
    }
    while (0);
    if (h->param.i_threads > 1 && -1)
        goto fail;
    if (h->param.i_lookahead_threads > 1 && -1)
        goto fail;
    if (h->param.b_opencl) {
        h->opencl.ocl = x264_opencl_load_library ();
        if (!h->opencl.ocl) {
            x264_log (h, 1, "failed to load OpenCL\n");
            h->param.b_opencl = 0;
        }
    }
    h->thread[0] = h;
    for (int i = 1; i < h->param.i_threads + !!h->param.i_sync_lookahead; i++)
        do {
            h->thread[i] = x264_malloc (sizeof (x264_t));
            if (!h->thread[i])
                goto fail;
        }
        while (0);
    if (h->param.i_lookahead_threads > 1)
        for (int i = 0; i < h->param.i_lookahead_threads; i++) {
            do {
                h->lookahead_thread[i] = x264_malloc (sizeof (x264_t));
                if (!h->lookahead_thread[i])
                    goto fail;
            }
            while (0);
            *h->lookahead_thread[i] = *h;
        }
    *h->reconfig_h = *h;
    for (int i = 0; i < h->param.i_threads; i++) {
        int init_nal_count = h->param.i_slice_count + 3;
        int allocate_threadlocal_data = !h->param.b_sliced_threads || !i;
        if (i > 0)
            *h->thread[i] = *h;
        if (0)
            goto fail;
        if (0)
            goto fail;
        if (allocate_threadlocal_data) {
            h->thread[i]->fdec = x264_frame_pop_unused (h, 1);
            if (!h->thread[i]->fdec)
                goto fail;
        }
        else
            h->thread[i]->fdec = h->thread[0]->fdec;
        do {
            h->thread[i]->out.p_bitstream = x264_malloc (h->out.i_bitstream);
            if (!h->thread[i]->out.p_bitstream)
                goto fail;
        }
        while (0);
        do {
            h->thread[i]->out.nal = x264_malloc (init_nal_count * sizeof (x264_nal_t));
            if (!h->thread[i]->out.nal)
                goto fail;
        }
        while (0);
        h->thread[i]->out.i_nals_allocated = init_nal_count;
        if (allocate_threadlocal_data && x264_macroblock_cache_allocate (h->thread[i]) < 0)
            goto fail;
    }
    if (h->param.b_opencl && x264_opencl_lookahead_init (h) < 0)
        h->param.b_opencl = 0;
    if (x264_lookahead_init (h, i_slicetype_length))
        goto fail;
    for (int i = 0; i < h->param.i_threads; i++)
        if (x264_macroblock_thread_allocate (h->thread[i], 0) < 0)
            goto fail;
    if (x264_ratecontrol_new (h) < 0)
        goto fail;
    if (h->param.i_nal_hrd) {
        x264_log (h, 3, "HRD bitrate: %i bits/sec\n", h -> sps -> vui.hrd.i_bit_rate_unscaled);
        x264_log (h, 3, "CPB size: %i bits\n", h -> sps -> vui.hrd.i_cpb_size_unscaled);
    }
    if (h->param.psz_dump_yuv) {
        FILE *f = fopen (h->param.psz_dump_yuv, "w");
        if (!f) {
            x264_log (h, 0, "dump_yuv: can't write to %s\n", h -> param.psz_dump_yuv);
            goto fail;
        }
        else if (!x264_is_regular_file (f)) {
            x264_log (h, 0, "dump_yuv: incompatible with non-regular file %s\n", h -> param.psz_dump_yuv);
            goto fail;
        }
        fclose (f);
    }
    const char *profile = h->sps->i_profile_idc == PROFILE_BASELINE ? "Constrained Baseline" : h->sps->i_profile_idc == PROFILE_MAIN ? "Main" : h->sps->i_profile_idc == PROFILE_HIGH ? "High" : h->sps->i_profile_idc == PROFILE_HIGH10 ? (h->sps->b_constraint_set3 == 1 ? "High 10 Intra" : "High 10") : h->sps->i_profile_idc == PROFILE_HIGH422 ? (h->sps->b_constraint_set3 == 1 ? "High 4:2:2 Intra" : "High 4:2:2") : h->sps->b_constraint_set3 == 1 ? "High 4:4:4 Intra" : "High 4:4:4 Predictive";
    char level [4];
    snprintf (level, sizeof (level), "%d.%d", h -> sps -> i_level_idc / 10, h -> sps -> i_level_idc % 10);
    if (h->sps->i_level_idc == 9 || (h->sps->i_level_idc == 11 && h->sps->b_constraint_set3 && (h->sps->i_profile_idc == PROFILE_BASELINE || h->sps->i_profile_idc == PROFILE_MAIN)))
        strcpy (level, "1b");
    if (h->sps->i_profile_idc < PROFILE_HIGH10) {
        x264_log (h, 2, "profile %s, level %s\n", profile, level);
    }
    else {
        static const char * const subsampling [4] = {"4:0:0", "4:2:0", "4:2:2", "4:4:4"};
        x264_log (h, 2, "profile %s, level %s, %s %d-bit\n", profile, level, subsampling [h -> sps -> i_chroma_format_idc], 8);
    }
    return h;
fail :
    x264_free (h);
    return ((void *) 0);
}

int x264_encoder_try_reconfig (x264_t *h, x264_param_t *param, int *rc_reconfig) {
    *rc_reconfig = 0;
    x264_set_aspect_ratio (h, param, 0);
    h->param.i_frame_reference = param->i_frame_reference;
    h->param.i_bframe_bias = param->i_bframe_bias;
    if (h->param.i_scenecut_threshold)
        h->param.i_scenecut_threshold = param->i_scenecut_threshold;
    h->param.b_deblocking_filter = param->b_deblocking_filter;
    h->param.i_deblocking_filter_alphac0 = param->i_deblocking_filter_alphac0;
    h->param.i_deblocking_filter_beta = param->i_deblocking_filter_beta;
    h->param.i_frame_packing = param->i_frame_packing;
    h->param.analyse.inter = param->analyse.inter;
    h->param.analyse.intra = param->analyse.intra;
    h->param.analyse.i_direct_mv_pred = param->analyse.i_direct_mv_pred;
    if (h->param.analyse.i_me_method < 3 || param->analyse.i_me_range < h->param.analyse.i_me_range)
        h->param.analyse.i_me_range = param->analyse.i_me_range;
    h->param.analyse.i_noise_reduction = param->analyse.i_noise_reduction;
    if (h->param.analyse.i_subpel_refine)
        h->param.analyse.i_subpel_refine = param->analyse.i_subpel_refine;
    h->param.analyse.i_trellis = param->analyse.i_trellis;
    h->param.analyse.b_chroma_me = param->analyse.b_chroma_me;
    h->param.analyse.b_dct_decimate = param->analyse.b_dct_decimate;
    h->param.analyse.b_fast_pskip = param->analyse.b_fast_pskip;
    h->param.analyse.b_mixed_references = param->analyse.b_mixed_references;
    h->param.analyse.f_psy_rd = param->analyse.f_psy_rd;
    h->param.analyse.f_psy_trellis = param->analyse.f_psy_trellis;
    h->param.crop_rect = param->crop_rect;
    if (h->param.analyse.i_me_method >= 3 || param->analyse.i_me_method < 3)
        h->param.analyse.i_me_method = param->analyse.i_me_method;
    if (h->param.analyse.i_me_method >= 3 && !h->frames.b_have_sub8x8_esa)
        h->param.analyse.inter &= ~0x0020;
    if (h->pps->b_transform_8x8_mode)
        h->param.analyse.b_transform_8x8 = param->analyse.b_transform_8x8;
    if (h->frames.i_max_ref1 > 1)
        h->param.i_bframe_pyramid = param->i_bframe_pyramid;
    h->param.i_slice_max_size = param->i_slice_max_size;
    h->param.i_slice_max_mbs = param->i_slice_max_mbs;
    h->param.i_slice_min_mbs = param->i_slice_min_mbs;
    h->param.i_slice_count = param->i_slice_count;
    h->param.i_slice_count_max = param->i_slice_count_max;
    h->param.b_tff = param->b_tff;
    if (h->param.rc.i_vbv_max_bitrate > 0 && h->param.rc.i_vbv_buffer_size > 0 && param->rc.i_vbv_max_bitrate > 0 && param->rc.i_vbv_buffer_size > 0) {
        *rc_reconfig |= h->param.rc.i_vbv_max_bitrate != param->rc.i_vbv_max_bitrate;
        *rc_reconfig |= h->param.rc.i_vbv_buffer_size != param->rc.i_vbv_buffer_size;
        *rc_reconfig |= h->param.rc.i_bitrate != param->rc.i_bitrate;
        h->param.rc.i_vbv_max_bitrate = param->rc.i_vbv_max_bitrate;
        h->param.rc.i_vbv_buffer_size = param->rc.i_vbv_buffer_size;
        h->param.rc.i_bitrate = param->rc.i_bitrate;
    }
    *rc_reconfig |= h->param.rc.f_rf_constant != param->rc.f_rf_constant;
    *rc_reconfig |= h->param.rc.f_rf_constant_max != param->rc.f_rf_constant_max;
    h->param.rc.f_rf_constant = param->rc.f_rf_constant;
    h->param.rc.f_rf_constant_max = param->rc.f_rf_constant_max;
    return x264_validate_parameters (h, 0);
}

int x264_encoder_reconfig_apply (x264_t *h, x264_param_t *param) {
    int rc_reconfig;
    int ret = x264_encoder_try_reconfig (h, param, &rc_reconfig);
    mbcmp_init (h);
    if (!ret)
        x264_sps_init (h->sps, h->param.i_sps_id, &h->param);
    if (!ret && rc_reconfig)
        x264_ratecontrol_init_reconfigurable (h, 0);
    return ret;
}

int x264_encoder_reconfig (x264_t *h, x264_param_t *param) {
    h = h->thread[h->thread[0]->i_thread_phase];
    x264_param_t param_save = h->reconfig_h->param;
    h->reconfig_h->param = h->param;
    int rc_reconfig;
    int ret = x264_encoder_try_reconfig (h->reconfig_h, param, &rc_reconfig);
    if (!ret)
        h->reconfig = 1;
    else
        h->reconfig_h->param = param_save;
    return ret;
}

void x264_encoder_parameters (x264_t *h, x264_param_t *param) {
    memcpy (param, & h -> thread [h -> i_thread_phase] -> param, sizeof (x264_param_t));
}

void x264_nal_start (x264_t *h, int i_type, int i_ref_idc) {
    x264_nal_t *nal = &h->out.nal[h->out.i_nal];
    nal->i_ref_idc = i_ref_idc;
    nal->i_type = i_type;
    nal->b_long_startcode = 1;
    nal->i_payload = 0;
    nal->p_payload = &h->out.p_bitstream[bs_pos (&h->out.bs) / 8];
    nal->i_padding = 0;
}

int x264_nal_check_buffer (x264_t *h) {
    if (h->out.i_nal >= h->out.i_nals_allocated) {
        x264_nal_t *new_out = x264_malloc (sizeof (x264_nal_t) * (h->out.i_nals_allocated * 2));
        if (!new_out)
            return -1;
        memcpy (new_out, h -> out.nal, sizeof (x264_nal_t) * (h -> out.i_nals_allocated));
        x264_free (h -> out.nal);
        h->out.nal = new_out;
        h->out.i_nals_allocated *= 2;
    }
    return 0;
}

int x264_nal_end (x264_t *h) {
    x264_nal_t *nal = &h->out.nal[h->out.i_nal];
    uint8_t *end = &h->out.p_bitstream[bs_pos (&h->out.bs) / 8];
    nal->i_payload = end - nal->p_payload;
    memset (end, 0xff, 64);
    if (h->param.nalu_process)
        h->param.nalu_process (h, nal, h->fenc->opaque);
    h->out.i_nal++;
    return x264_nal_check_buffer (h);
}

int x264_check_encapsulated_buffer (x264_t *h, x264_t *h0, int start, int previous_nal_size, int necessary_size) {
    if (h0->nal_buffer_size < necessary_size) {
        necessary_size *= 2;
        uint8_t *buf = x264_malloc (necessary_size);
        if (!buf)
            return -1;
        if (previous_nal_size)
            memcpy (buf, h0->nal_buffer, previous_nal_size);
        intptr_t delta = buf - h0->nal_buffer;
        for (int i = 0; i < start; i++)
            h->out.nal[i].p_payload += delta;
        x264_free (h0 -> nal_buffer);
        h0->nal_buffer = buf;
        h0->nal_buffer_size = necessary_size;
    }
    return 0;
}

int x264_encoder_encapsulate_nals (x264_t *h, int start) {
    x264_t *h0 = h->thread[0];
    int nal_size = 0, previous_nal_size = 0;
    if (h->param.nalu_process) {
        for (int i = start; i < h->out.i_nal; i++)
            nal_size += h->out.nal[i].i_payload;
        return nal_size;
    }
    for (int i = 0; i < start; i++)
        previous_nal_size += h->out.nal[i].i_payload;
    for (int i = start; i < h->out.i_nal; i++)
        nal_size += h->out.nal[i].i_payload;
    int necessary_size = previous_nal_size + nal_size * 3 / 2 + h->out.i_nal * 4 + 4 + 64;
    for (int i = start; i < h->out.i_nal; i++)
        necessary_size += h->out.nal[i].i_padding;
    if (x264_check_encapsulated_buffer (h, h0, start, previous_nal_size, necessary_size))
        return -1;
    uint8_t *nal_buffer = h0->nal_buffer + previous_nal_size;
    for (int i = start; i < h->out.i_nal; i++) {
        int old_payload_len = h->out.nal[i].i_payload;
        h->out.nal[i].b_long_startcode = !i || h->out.nal[i].i_type == NAL_SPS || h->out.nal[i].i_type == NAL_PPS || h->param.i_avcintra_class;
        x264_nal_encode (h, nal_buffer, & h -> out.nal [i]);
        nal_buffer += h->out.nal[i].i_payload;
        if (h->param.i_avcintra_class) {
            h->out.nal[i].i_padding -= h->out.nal[i].i_payload - (old_payload_len + 5);
            if (h->out.nal[i].i_padding > 0) {
                memset (nal_buffer, 0, h -> out.nal [i].i_padding);
                nal_buffer += h->out.nal[i].i_padding;
                h->out.nal[i].i_payload += h->out.nal[i].i_padding;
            }
            h->out.nal[i].i_padding = ((h->out.nal[i].i_padding) > (0) ? (h->out.nal[i].i_padding) : (0));
        }
    };
    return nal_buffer - (h0->nal_buffer + previous_nal_size);
}

int x264_encoder_headers (x264_t *h, x264_nal_t **pp_nal, int *pi_nal) {
    int frame_size = 0;
    h->out.i_nal = 0;
    bs_init (& h -> out.bs, h -> out.p_bitstream, h -> out.i_bitstream);
    x264_nal_start (h, NAL_SPS, NAL_PRIORITY_HIGHEST);
    x264_sps_write (& h -> out.bs, h -> sps);
    if (x264_nal_end (h))
        return -1;
    x264_nal_start (h, NAL_PPS, NAL_PRIORITY_HIGHEST);
    x264_pps_write (& h -> out.bs, h -> sps, h -> pps);
    if (x264_nal_end (h))
        return -1;
    x264_nal_start (h, NAL_SEI, NAL_PRIORITY_DISPOSABLE);
    if (x264_sei_version_write (h, &h->out.bs))
        return -1;
    if (x264_nal_end (h))
        return -1;
    frame_size = x264_encoder_encapsulate_nals (h, 0);
    if (frame_size < 0)
        return -1;
    *pi_nal = h->out.i_nal;
    *pp_nal = &h->out.nal[0];
    h->out.i_nal = 0;
    return frame_size;
}

void x264_reference_check_reorder (x264_t *h) {
    for (int i = 0; h->frames.reference[i]; i++)
        if (h->frames.reference[i]->b_corrupt) {
            h->b_ref_reorder[0] = 1;
            return;
        }
    for (int list = 0; list <= (h->sh.i_type == SLICE_TYPE_B); list++)
        for (int i = 0; i < h->i_ref[list] - 1; i++) {
            int framenum_diff = h->fref[list][i + 1]->i_frame_num - h->fref[list][i]->i_frame_num;
            int poc_diff = h->fref[list][i + 1]->i_poc - h->fref[list][i]->i_poc;
            if (h->sh.i_type == SLICE_TYPE_P ? framenum_diff > 0 : list == 1 ? poc_diff < 0 : poc_diff > 0) {
                h->b_ref_reorder[list] = 1;
                return;
            }
        }
}

int x264_weighted_reference_duplicate (x264_t *h, int i_ref, const x264_weight_t *w) {
    int i = h->i_ref[0];
    int j = 1;
    x264_frame_t *newframe;
    if (i <= 1)
        return -1;
    if (h->param.analyse.i_weighted_pred != 2)
        return -1;
    if (8 > 8 && w != x264_weight_none)
        return -1;
    newframe = x264_frame_pop_blank_unused (h);
    if (!newframe)
        return -1;
    *newframe = *h->fref[0][i_ref];
    newframe->i_reference_count = 1;
    newframe->orig = h->fref[0][i_ref];
    newframe->b_duplicate = 1;
    memcpy (h -> fenc -> weight [j], w, sizeof (h -> fenc -> weight [i]));
    h->b_ref_reorder[0] = 1;
    if (h->i_ref[0] < 16)
        ++h->i_ref[0];
    h->fref[0][16 - 1] = ((void *) 0);
    x264_frame_unshift (& h -> fref [0] [j], newframe);
    return j;
}

void x264_weighted_pred_init (x264_t *h) {
    for (int i_ref = 0; i_ref < h->i_ref[0]; i_ref++)
        h->fenc->weighted[i_ref] = h->fref[0][i_ref]->filtered[0][0];
    h->fenc->i_lines_weighted = 0;
    for (int i_ref = 0; i_ref < (h->i_ref[0] << 0); i_ref++)
        for (int i = 0; i < 3; i++)
            h->sh.weight[i_ref][i].weightfn = ((void *) 0);
    if (h->sh.i_type != SLICE_TYPE_P || h->param.analyse.i_weighted_pred <= 0)
        return;
    int i_padv = 32 << 0;
    int denom = -1;
    int weightplane [2] = {0, 0};
    int buffer_next = 0;
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < h->i_ref[0]; j++) {
            if (h->fenc->weight[j][i].weightfn) {
                h->sh.weight[j][i] = h->fenc->weight[j][i];
                if (h->sh.weight[j][i].i_scale == 1 << h->sh.weight[j][i].i_denom && h->sh.weight[j][i].i_offset == 0)
                    h->sh.weight[j][i].weightfn = ((void *) 0);
                else {
                    if (!weightplane[!!i]) {
                        weightplane[!!i] = 1;
                        h->sh.weight[0][!!i].i_denom = denom = h->sh.weight[j][i].i_denom;
                        ((x264_clip3 (denom, 0, 7) == denom) ? (void) (0) : __assert_fail ("x264_clip3( denom, 0, 7 ) == denom", "DonorSourceCode/encoder.c", 2097, __PRETTY_FUNCTION__));
                    }
                    ((h->sh.weight[j][i].i_denom == denom) ? (void) (0) : __assert_fail ("h->sh.weight[j][i].i_denom == denom", "DonorSourceCode/encoder.c", 2100, __PRETTY_FUNCTION__));
                    if (!i) {
                        h->fenc->weighted[j] = h->mb.p_weight_buf[buffer_next++] + h->fenc->i_stride[0] * i_padv + 32;
                        if (h->param.i_threads == 1) {
                            pixel *src = h->fref[0][j]->filtered[0][0] - h->fref[0][j]->i_stride[0] * i_padv - 32;
                            pixel *dst = h->fenc->weighted[j] - h->fenc->i_stride[0] * i_padv - 32;
                            int stride = h->fenc->i_stride[0];
                            int width = h->fenc->i_width[0] + 32 * 2;
                            int height = h->fenc->i_lines[0] + i_padv * 2;
                            x264_weight_scale_plane (h, dst, stride, src, stride, width, height, & h -> sh.weight [j] [0]);
                            h->fenc->i_lines_weighted = height;
                        }
                    }
                }
            }
        }
    }
    if (weightplane[1])
        for (int i = 0; i < h->i_ref[0]; i++) {
            if (h->sh.weight[i][1].weightfn && !h->sh.weight[i][2].weightfn) {
                h->sh.weight[i][2].i_scale = 1 << h->sh.weight[0][1].i_denom;
                h->sh.weight[i][2].i_offset = 0;
            }
            else if (h->sh.weight[i][2].weightfn && !h->sh.weight[i][1].weightfn) {
                h->sh.weight[i][1].i_scale = 1 << h->sh.weight[0][1].i_denom;
                h->sh.weight[i][1].i_offset = 0;
            }
        }
    if (!weightplane[0])
        h->sh.weight[0][0].i_denom = 0;
    if (!weightplane[1])
        h->sh.weight[0][1].i_denom = 0;
    h->sh.weight[0][2].i_denom = h->sh.weight[0][1].i_denom;
}

int x264_reference_distance (x264_t *h, x264_frame_t *frame) {
    if (h->param.i_frame_packing == 5)
        return abs ((h->fenc->i_frame & ~1) - (frame->i_frame & ~1)) + ((h->fenc->i_frame & 1) != (frame->i_frame & 1));
    else
        return abs (h->fenc->i_frame - frame->i_frame);
}

void x264_reference_build_list (x264_t *h, int i_poc) {
    int b_ok;
    h->mb.pic.i_fref[0] = h->i_ref[0] = 0;
    h->mb.pic.i_fref[1] = h->i_ref[1] = 0;
    if (h->sh.i_type == SLICE_TYPE_I)
        return;
    for (int i = 0; h->frames.reference[i]; i++) {
        if (h->frames.reference[i]->b_corrupt)
            continue;
        if (h->frames.reference[i]->i_poc < i_poc)
            h->fref[0][h->i_ref[0]++] = h->frames.reference[i];
        else if (h->frames.reference[i]->i_poc > i_poc)
            h->fref[1][h->i_ref[1]++] = h->frames.reference[i];
    }
    if (h->sh.i_mmco_remove_from_end) {
        do {
            b_ok = 1;
            for (int i = 0; i < h->i_ref[0] - 1; i++) {
                if (h->fref[0][i]->i_frame < h->fref[0][i + 1]->i_frame) {
                    do {
                        x264_frame_t *t = h->fref[0][i];
                        h->fref[0][i] = h->fref[0][i + 1];
                        h->fref[0][i + 1] = t;
                    }
                    while (0);
                    b_ok = 0;
                    break;
                }
            }
        }
        while (!b_ok);
        for (int i = h->i_ref[0] - 1; i >= h->i_ref[0] - h->sh.i_mmco_remove_from_end; i--) {
            int diff = h->i_frame_num - h->fref[0][i]->i_frame_num;
            h->sh.mmco[h->sh.i_mmco_command_count].i_poc = h->fref[0][i]->i_poc;
            h->sh.mmco[h->sh.i_mmco_command_count++].i_difference_of_pic_nums = diff;
        }
    }
    for (int list = 0; list < 2; list++) {
        h->fref_nearest[list] = h->fref[list][0];
        do {
            b_ok = 1;
            for (int i = 0; i < h->i_ref[list] - 1; i++) {
                if (list ? h->fref[list][i + 1]->i_poc < h->fref_nearest[list]->i_poc : h->fref[list][i + 1]->i_poc > h->fref_nearest[list]->i_poc)
                    h->fref_nearest[list] = h->fref[list][i + 1];
                if (x264_reference_distance (h, h->fref[list][i]) > x264_reference_distance (h, h->fref[list][i + 1])) {
                    do {
                        x264_frame_t *t = h->fref[list][i];
                        h->fref[list][i] = h->fref[list][i + 1];
                        h->fref[list][i + 1] = t;
                    }
                    while (0);
                    b_ok = 0;
                    break;
                }
            }
        }
        while (!b_ok);
    }
    x264_reference_check_reorder (h);
    h->i_ref[1] = ((h->i_ref[1]) < (h->frames.i_max_ref1) ? (h->i_ref[1]) : (h->frames.i_max_ref1));
    h->i_ref[0] = ((h->i_ref[0]) < (h->frames.i_max_ref0) ? (h->i_ref[0]) : (h->frames.i_max_ref0));
    h->i_ref[0] = ((h->i_ref[0]) < (h->param.i_frame_reference) ? (h->i_ref[0]) : (h->param.i_frame_reference));
    if (((h->fenc->i_type) == 0x0005 || (h->fenc->i_type) == 0x0004) && h->param.b_bluray_compat)
        h->i_ref[0] = ((h->i_ref[0]) < (((h->fref[0][0]->i_type) == 0x0005 || (h->fref[0][0]->i_type) == 0x0004) + 1) ? (h->i_ref[0]) : (((h->fref[0][0]->i_type) == 0x0005 || (h->fref[0][0]->i_type) == 0x0004) + 1));
    if (h->fenc->i_type == 0x0003) {
        int idx = -1;
        if (h->param.analyse.i_weighted_pred >= 1) {
            x264_weight_t w [3];
            w[1].weightfn = w[2].weightfn = ((void *) 0);
            if (h->param.rc.b_stat_read)
                x264_ratecontrol_set_weights (h, h->fenc);
            if (!h->fenc->weight[0][0].weightfn) {
                h->fenc->weight[0][0].i_denom = 0;
                {
                    (w[0]).i_scale = (1);
                    (w[0]).i_denom = (0);
                    (w[0]).i_offset = (-1);
                    if (1)
                        h->mc.weight_cache (h, &w[0]);
                    else
                        w[0].weightfn = ((void *) 0);
                };
                idx = x264_weighted_reference_duplicate (h, 0, w);
            }
            else {
                if (h->fenc->weight[0][0].i_scale == 1 << h->fenc->weight[0][0].i_denom) {
                    {
                        (h->fenc->weight[0][0]).i_scale = (1);
                        (h->fenc->weight[0][0]).i_denom = (0);
                        (h->fenc->weight[0][0]).i_offset = (h->fenc->weight[0][0].i_offset);
                        if (1)
                            h->mc.weight_cache (h, &h->fenc->weight[0][0]);
                        else
                            h->fenc->weight[0][0].weightfn = ((void *) 0);
                    };
                }
                x264_weighted_reference_duplicate (h, 0, x264_weight_none);
                if (h->fenc->weight[0][0].i_offset > -128) {
                    w[0] = h->fenc->weight[0][0];
                    w[0].i_offset--;
                    h->mc.weight_cache (h, &w[0]);
                    idx = x264_weighted_reference_duplicate (h, 0, w);
                }
            }
        }
        h->mb.ref_blind_dupe = idx;
    }
    ((h->i_ref[0] + h->i_ref[1] <= 16) ? (void) (0) : __assert_fail ("h->i_ref[0] + h->i_ref[1] <= 16", "DonorSourceCode/encoder.c", 2265, __PRETTY_FUNCTION__));
    h->mb.pic.i_fref[0] = h->i_ref[0];
    h->mb.pic.i_fref[1] = h->i_ref[1];
}

void x264_fdec_filter_row (x264_t *h, int mb_y, int pass) {
    int b_hpel = h->fdec->b_kept_as_ref;
    int b_deblock = h->sh.i_disable_deblocking_filter_idc != 1;
    int b_end = mb_y == h->i_threadslice_end;
    int b_measure_quality = 1;
    int min_y = mb_y - (1 << 0);
    int b_start = min_y == h->i_threadslice_start;
    int minpix_y = min_y * 16 - 4 * !b_start;
    int maxpix_y = mb_y * 16 - 4 * !b_end;
    b_deblock &= b_hpel || h->param.b_full_recon || h->param.psz_dump_yuv;
    if (h->param.b_sliced_threads) {
        switch (pass) {
        default :
        case 0 :
            b_deblock &= h->param.b_full_recon;
            b_hpel = 0;
            break;
        case 1 :
            b_deblock &= !h->param.b_full_recon;
            b_hpel &= !(b_start && min_y > 0);
            b_measure_quality = 0;
            break;
        case 2 :
            b_deblock = 0;
            b_measure_quality = 0;
            break;
        }
    }
    if (mb_y & 0)
        return;
    if (min_y < h->i_threadslice_start)
        return;
    if (b_deblock)
        for (int y = min_y; y < mb_y; y += (1 << 0))
            x264_frame_deblock_row (h, y);
    if (0 && (!h->param.b_sliced_threads || pass == 1))
        for (int p = 0; p < h->fdec->i_plane; p++)
            for (int i = minpix_y >> (h->mb.chroma_v_shift && p); i < maxpix_y >> (h->mb.chroma_v_shift && p); i++)
                memcpy (h->fdec->plane_fld[p] + i * h->fdec->i_stride[p], h->fdec->plane[p] + i * h->fdec->i_stride[p], h->mb.i_mb_width * 16 * sizeof (pixel));
    if (h->fdec->b_kept_as_ref && (!h->param.b_sliced_threads || pass == 1))
        x264_frame_expand_border (h, h->fdec, min_y);
    if (b_hpel) {
        int end = mb_y == h->mb.i_mb_height;
        if (h->param.analyse.i_subpel_refine) {
            x264_frame_filter (h, h -> fdec, min_y, end);
            x264_frame_expand_border_filtered (h, h -> fdec, min_y, end);
        }
    }
    if (0 && pass == 0)
        for (int i = 0; i < 3; i++) {
            do {
                pixel *t = h->intra_border_backup[0][i];
                h->intra_border_backup[0][i] = h->intra_border_backup[3][i];
                h->intra_border_backup[3][i] = t;
            }
            while (0);
            do {
                pixel *t = h->intra_border_backup[1][i];
                h->intra_border_backup[1][i] = h->intra_border_backup[4][i];
                h->intra_border_backup[4][i] = t;
            }
            while (0);
        }
    if (h->i_thread_frames > 1 && h->fdec->b_kept_as_ref)
        x264_frame_cond_broadcast (h->fdec, mb_y *16 + (b_end ? 10000 : -(24 << 0)));
    if (b_measure_quality) {
        maxpix_y = ((maxpix_y) < (h->param.i_height) ? (maxpix_y) : (h->param.i_height));
        if (h->param.analyse.b_psnr) {
            for (int p = 0; p < ((h->sps->i_chroma_format_idc == CHROMA_444) ? 3 : 1); p++)
                h->stat.frame.i_ssd[p] += x264_pixel_ssd_wxh (&h->pixf, h->fdec->plane[p] + minpix_y * h->fdec->i_stride[p], h->fdec->i_stride[p], h->fenc->plane[p] + minpix_y * h->fenc->i_stride[p], h->fenc->i_stride[p], h->param.i_width, maxpix_y -minpix_y);
            if (!(h->sps->i_chroma_format_idc == CHROMA_444)) {
                uint64_t ssd_u, ssd_v;
                int v_shift = h->mb.chroma_v_shift;
                x264_pixel_ssd_nv12 (& h -> pixf, h -> fdec -> plane [1] + (minpix_y >> v_shift) * h -> fdec -> i_stride [1], h -> fdec -> i_stride [1], h -> fenc -> plane [1] + (minpix_y >> v_shift) * h -> fenc -> i_stride [1], h -> fenc -> i_stride [1], h -> param.i_width >> 1, (maxpix_y - minpix_y) >> v_shift, & ssd_u, & ssd_v);
                h->stat.frame.i_ssd[1] += ssd_u;
                h->stat.frame.i_ssd[2] += ssd_v;
            }
        }
        if (h->param.analyse.b_ssim) {
            int ssim_cnt;
            ;
            minpix_y += b_start ? 2 : -6;
            h->stat.frame.f_ssim += x264_pixel_ssim_wxh (&h->pixf, h->fdec->plane[0] + 2 + minpix_y * h->fdec->i_stride[0], h->fdec->i_stride[0], h->fenc->plane[0] + 2 + minpix_y * h->fenc->i_stride[0], h->fenc->i_stride[0], h->param.i_width - 2, maxpix_y -minpix_y, h->scratch_buffer, &ssim_cnt);
            h->stat.frame.i_ssim_cnt += ssim_cnt;
        }
    }
}

int x264_reference_update (x264_t *h) {
    if (!h->fdec->b_kept_as_ref) {
        if (h->i_thread_frames > 1) {
            x264_frame_push_unused (h, h -> fdec);
            h->fdec = x264_frame_pop_unused (h, 1);
            if (!h->fdec)
                return -1;
        }
        return 0;
    }
    for (int i = 0; i < h->sh.i_mmco_command_count; i++)
        for (int j = 0; h->frames.reference[j]; j++)
            if (h->frames.reference[j]->i_poc == h->sh.mmco[i].i_poc)
                x264_frame_push_unused (h, x264_frame_shift (&h->frames.reference[j]));
    x264_frame_push (h -> frames.reference, h -> fdec);
    if (h->frames.reference[h->sps->i_num_ref_frames])
        x264_frame_push_unused (h, x264_frame_shift (h->frames.reference));
    h->fdec = x264_frame_pop_unused (h, 1);
    if (!h->fdec)
        return -1;
    return 0;
}

void x264_reference_reset (x264_t *h) {
    while (h->frames.reference[0])
        x264_frame_push_unused (h, x264_frame_pop (h->frames.reference));
    h->fdec->i_poc = h->fenc->i_poc = 0;
}

void x264_reference_hierarchy_reset (x264_t *h) {
    int ref;
    int b_hasdelayframe = 0;
    for (int i = 0; h->frames.current[i] && (h->frames.current[i]->i_type == 0x0005); i++)
        b_hasdelayframe |= h->frames.current[i]->i_coded != h->frames.current[i]->i_frame + h->sps->vui.i_num_reorder_frames;
    if (h->param.i_bframe_pyramid != 1 && !b_hasdelayframe && h->frames.i_poc_last_open_gop == -1)
        return;
    for (ref = 0; h->frames.reference[ref]; ref++) {
        if ((h->param.i_bframe_pyramid == 1 && h->frames.reference[ref]->i_type == 0x0004) || (h->frames.reference[ref]->i_poc < h->frames.i_poc_last_open_gop && h->sh.i_type != SLICE_TYPE_B)) {
            int diff = h->i_frame_num - h->frames.reference[ref]->i_frame_num;
            h->sh.mmco[h->sh.i_mmco_command_count].i_difference_of_pic_nums = diff;
            h->sh.mmco[h->sh.i_mmco_command_count++].i_poc = h->frames.reference[ref]->i_poc;
            x264_frame_push_unused (h, x264_frame_shift (& h -> frames.reference [ref]));
            h->b_ref_reorder[0] = 1;
            ref--;
        }
    }
    if (h->param.i_bframe_pyramid)
        h->sh.i_mmco_remove_from_end = ((ref + 2 - h->frames.i_max_dpb) > (0) ? (ref + 2 - h->frames.i_max_dpb) : (0));
}

void x264_slice_init (x264_t *h, int i_nal_type, int i_global_qp) {
    if (i_nal_type == NAL_SLICE_IDR) {
        x264_slice_header_init (h, & h -> sh, h -> sps, h -> pps, h -> i_idr_pic_id, h -> i_frame_num, i_global_qp);
        if (h->param.i_avcintra_class) {
            switch (h->i_idr_pic_id) {
            case 5 :
                h->i_idr_pic_id = 3;
                break;
            case 3 :
                h->i_idr_pic_id = 4;
                break;
            case 4 :
            default :
                h->i_idr_pic_id = 5;
                break;
            }
        }
        else
            h->i_idr_pic_id ^= 1;
    }
    else {
        x264_slice_header_init (h, & h -> sh, h -> sps, h -> pps, - 1, h -> i_frame_num, i_global_qp);
        h->sh.i_num_ref_idx_l0_active = h->i_ref[0] <= 0 ? 1 : h->i_ref[0];
        h->sh.i_num_ref_idx_l1_active = h->i_ref[1] <= 0 ? 1 : h->i_ref[1];
        if (h->sh.i_num_ref_idx_l0_active != h->pps->i_num_ref_idx_l0_default_active || (h->sh.i_type == SLICE_TYPE_B && h->sh.i_num_ref_idx_l1_active != h->pps->i_num_ref_idx_l1_default_active)) {
            h->sh.b_num_ref_idx_override = 1;
        }
    }
    if (h->fenc->i_type == 0x0004 && h->param.b_bluray_compat && h->sh.i_mmco_command_count) {
        h->b_sh_backup = 1;
        h->sh_backup = h->sh;
    }
    h->fdec->i_frame_num = h->sh.i_frame_num;
    if (h->sps->i_poc_type == 0) {
        h->sh.i_poc = h->fdec->i_poc;
        if (0) {
            h->sh.i_delta_poc_bottom = h->param.b_tff ? 1 : -1;
            h->sh.i_poc += h->sh.i_delta_poc_bottom == -1;
        }
        else
            h->sh.i_delta_poc_bottom = 0;
        h->fdec->i_delta_poc[0] = h->sh.i_delta_poc_bottom == -1;
        h->fdec->i_delta_poc[1] = h->sh.i_delta_poc_bottom == 1;
    }
    else {
    }
    x264_macroblock_slice_init (h);
}

struct _x264_bs_bak_t {
    int skip;
    uint8_t cabac_prevbyte;
    bs_t bs;
    x264_cabac_t cabac;
    x264_frame_stat_t stat;
    int last_qp;
    int last_dqp;
    int field_decoding_flag;
};
typedef struct _x264_bs_bak_t x264_bs_bak_t;

void x264_bitstream_backup (x264_t *h, x264_bs_bak_t *bak, int i_skip, int full) {
    if (full) {
        bak->stat = h->stat.frame;
        bak->last_qp = h->mb.i_last_qp;
        bak->last_dqp = h->mb.i_last_dqp;
        bak->field_decoding_flag = h->mb.field_decoding_flag;
    }
    else {
        bak->stat.i_mv_bits = h->stat.frame.i_mv_bits;
        bak->stat.i_tex_bits = h->stat.frame.i_tex_bits;
    }
    if (h->param.b_cabac) {
        if (full)
            memcpy (&bak->cabac, &h->cabac, sizeof (x264_cabac_t));
        else
            memcpy (&bak->cabac, &h->cabac, __builtin_offsetof (x264_cabac_t, f8_bits_encoded));
        bak->cabac_prevbyte = h->cabac.p[-1];
    }
    else {
        bak->bs = h->out.bs;
        bak->skip = i_skip;
    }
}

void x264_bitstream_restore (x264_t *h, x264_bs_bak_t *bak, int *skip, int full) {
    if (full) {
        h->stat.frame = bak->stat;
        h->mb.i_last_qp = bak->last_qp;
        h->mb.i_last_dqp = bak->last_dqp;
        h->mb.field_decoding_flag = bak->field_decoding_flag;
    }
    else {
        h->stat.frame.i_mv_bits = bak->stat.i_mv_bits;
        h->stat.frame.i_tex_bits = bak->stat.i_tex_bits;
    }
    if (h->param.b_cabac) {
        if (full)
            memcpy (&h->cabac, &bak->cabac, sizeof (x264_cabac_t));
        else
            memcpy (&h->cabac, &bak->cabac, __builtin_offsetof (x264_cabac_t, f8_bits_encoded));
        h->cabac.p[-1] = bak->cabac_prevbyte;
    }
    else {
        h->out.bs = bak->bs;
        *skip = bak->skip;
    }
}

intptr_t x264_slice_write (x264_t *h) {
    int i_skip;
    int mb_xy, i_mb_x, i_mb_y;
    int overhead_guess = (5 - (h->param.b_annexb && h->out.i_nal)) + 1 + h->param.b_cabac + 5;
    int slice_max_size = h->param.i_slice_max_size > 0 ? (h->param.i_slice_max_size - overhead_guess) * 8 : 0;
    int back_up_bitstream_cavlc = !h->param.b_cabac && h->sps->i_profile_idc < PROFILE_HIGH;
    int back_up_bitstream = slice_max_size || back_up_bitstream_cavlc;
    int starting_bits = bs_pos (&h->out.bs);
    int b_deblock = h->sh.i_disable_deblocking_filter_idc != 1;
    int b_hpel = h->fdec->b_kept_as_ref;
    int orig_last_mb = h->sh.i_last_mb;
    int thread_last_mb = h->i_threadslice_end * h->mb.i_mb_width - 1;
    uint8_t *last_emu_check;
    x264_bs_bak_t bs_bak [4];
    b_deblock &= b_hpel || h->param.b_full_recon || h->param.psz_dump_yuv;
    bs_realign (& h -> out.bs);
    x264_nal_start (h, h -> i_nal_type, h -> i_nal_ref_idc);
    h->out.nal[h->out.i_nal].i_first_mb = h->sh.i_first_mb;
    x264_macroblock_thread_init (h);
    h->mb.i_mb_xy = h->sh.i_first_mb;
    h->sh.i_qp = x264_ratecontrol_mb_qp (h);
    h->sh.i_qp = (((h->sh.i_qp)) < ((51 + (6 * (8 - 8)))) ? ((h->sh.i_qp)) : ((51 + (6 * (8 - 8)))));
    h->sh.i_qp_delta = h->sh.i_qp - h->pps->i_pic_init_qp;
    x264_slice_header_write (& h -> out.bs, & h -> sh, h -> i_nal_ref_idc);
    if (h->param.b_cabac) {
        bs_align_1 (& h -> out.bs);
        x264_cabac_context_init (h, & h -> cabac, h -> sh.i_type, x264_clip3 (h -> sh.i_qp - (6 * (8 - 8)), 0, 51), h -> sh.i_cabac_init_idc);
        x264_cabac_encode_init (& h -> cabac, h -> out.bs.p, h -> out.bs.p_end);
        last_emu_check = h->cabac.p;
    }
    else
        last_emu_check = h->out.bs.p;
    h->mb.i_last_qp = h->sh.i_qp;
    h->mb.i_last_dqp = 0;
    h->mb.field_decoding_flag = 0;
    i_mb_y = h->sh.i_first_mb / h->mb.i_mb_width;
    i_mb_x = h->sh.i_first_mb % h->mb.i_mb_width;
    i_skip = 0;
    while (1) {
        mb_xy = i_mb_x + i_mb_y * h->mb.i_mb_width;
        int mb_spos = bs_pos (&h->out.bs) + x264_cabac_pos (&h->cabac);
        if (i_mb_x == 0) {
            if (x264_bitstream_check_buffer (h))
                return -1;
            if (!(i_mb_y & 0) && h->param.rc.i_vbv_buffer_size)
                x264_bitstream_backup (h, &bs_bak[3], i_skip, 1);
            if (!h->mb.b_reencode_mb)
                x264_fdec_filter_row (h, i_mb_y, 0);
        }
        if (back_up_bitstream) {
            if (back_up_bitstream_cavlc)
                x264_bitstream_backup (h, &bs_bak[1], i_skip, 0);
            if (slice_max_size && !(i_mb_y & 0)) {
                x264_bitstream_backup (h, & bs_bak [0], i_skip, 0);
                if ((thread_last_mb + 1 - mb_xy) == h->param.i_slice_min_mbs)
                    x264_bitstream_backup (h, &bs_bak[2], i_skip, 0);
            }
        }
        if (0) {
            if (h->mb.b_adaptive_mbaff) {
                if (!(i_mb_y & 1)) {
                    h->mb.b_interlaced = x264_field_vsad (h, i_mb_x, i_mb_y);
                    memcpy (& h -> zigzagf, 0 ? & h -> zigzagf_interlaced : & h -> zigzagf_progressive, sizeof (h -> zigzagf));
                    if (!0 && (i_mb_y + 2) == h->mb.i_mb_height)
                        x264_expand_border_mbpair (h, i_mb_x, i_mb_y);
                }
            }
            h->mb.field[mb_xy] = 0;
        }
        if (0)
            x264_macroblock_cache_load_interlaced (h, i_mb_x, i_mb_y);
        else
            x264_macroblock_cache_load_progressive (h, i_mb_x, i_mb_y);
        x264_macroblock_analyse (h);
    reencode :
        x264_macroblock_encode (h);
        if (h->param.b_cabac) {
            if (mb_xy > h->sh.i_first_mb && !(0 && (i_mb_y & 1)))
                x264_cabac_encode_terminal_c (&h->cabac);
            if (((h->mb.i_type) == P_SKIP || (h->mb.i_type) == B_SKIP))
                x264_cabac_mb_skip (h, 1);
            else {
                if (h->sh.i_type != SLICE_TYPE_I)
                    x264_cabac_mb_skip (h, 0);
                x264_macroblock_write_cabac (h, & h -> cabac);
            }
        }
        else {
            if (((h->mb.i_type) == P_SKIP || (h->mb.i_type) == B_SKIP))
                i_skip++;
            else {
                if (h->sh.i_type != SLICE_TYPE_I) {
                    bs_write_ue_big (& h -> out.bs, i_skip);
                    i_skip = 0;
                }
                if (h->mb.b_overflow) {
                    h->mb.i_chroma_qp = h->chroma_qp_table[++h->mb.i_qp];
                    h->mb.i_skip_intra = 0;
                    h->mb.b_skip_mc = 0;
                    h->mb.b_overflow = 0;
                    x264_bitstream_restore (h, & bs_bak [1], & i_skip, 0);
                    goto reencode;
                }
            }
        }
        int total_bits = bs_pos (&h->out.bs) + x264_cabac_pos (&h->cabac);
        int mb_size = total_bits - mb_spos;
        if (slice_max_size && (!0 || (i_mb_y & 1))) {
            if (!h->param.b_cabac)
                total_bits += bs_size_ue_big (i_skip);
            uint8_t *end = h->param.b_cabac ? h->cabac.p : h->out.bs.p;
            for (; last_emu_check < end - 2; last_emu_check++)
                if (last_emu_check[0] == 0 && last_emu_check[1] == 0 && last_emu_check[2] <= 3) {
                    slice_max_size -= 8;
                    last_emu_check++;
                }
            if (total_bits - starting_bits > slice_max_size && !h->mb.b_reencode_mb) {
                if (!x264_frame_new_slice (h, h->fdec)) {
                    if (mb_xy <= thread_last_mb && (thread_last_mb + 1 - mb_xy) < h->param.i_slice_min_mbs) {
                        if (thread_last_mb - h->param.i_slice_min_mbs < h->sh.i_first_mb + h->param.i_slice_min_mbs) {
                            x264_log (h, 1, "slice-max-size violated (frame %d, cause: slice-min-mbs)\n", h -> i_frame);
                            slice_max_size = 0;
                            goto cont;
                        }
                        x264_bitstream_restore (h, & bs_bak [2], & i_skip, 0);
                        h->mb.b_reencode_mb = 1;
                        h->sh.i_last_mb = thread_last_mb - h->param.i_slice_min_mbs;
                        break;
                    }
                    if (mb_xy - 0 * h->mb.i_mb_stride != h->sh.i_first_mb) {
                        x264_bitstream_restore (h, & bs_bak [0], & i_skip, 0);
                        h->mb.b_reencode_mb = 1;
                        if (0) {
                            if (i_mb_x)
                                h->sh.i_last_mb = mb_xy - 1 + h->mb.i_mb_stride * (!(i_mb_y & 1));
                            else
                                h->sh.i_last_mb = (i_mb_y - 2 + !(i_mb_y & 1)) * h->mb.i_mb_stride + h->mb.i_mb_width - 1;
                        }
                        else
                            h->sh.i_last_mb = mb_xy - 1;
                        break;
                    }
                    else
                        h->sh.i_last_mb = mb_xy;
                }
                else
                    slice_max_size = 0;
            }
        }
    cont :
        h->mb.b_reencode_mb = 0;
        x264_macroblock_cache_save (h);
        if (x264_ratecontrol_mb (h, mb_size) < 0) {
            x264_bitstream_restore (h, & bs_bak [3], & i_skip, 1);
            h->mb.b_reencode_mb = 1;
            i_mb_x = 0;
            i_mb_y = i_mb_y - 0;
            h->mb.i_mb_prev_xy = i_mb_y * h->mb.i_mb_stride - 1;
            h->sh.i_last_mb = orig_last_mb;
            continue;
        }
        h->stat.frame.i_mb_count[h->mb.i_type]++;
        int b_intra = ((h->mb.i_type) == I_4x4 || (h->mb.i_type) == I_8x8 || (h->mb.i_type) == I_16x16 || (h->mb.i_type) == I_PCM);
        int b_skip = ((h->mb.i_type) == P_SKIP || (h->mb.i_type) == B_SKIP);
        if (h->param.i_log_level >= 2 || h->param.rc.b_stat_write) {
            if (!b_intra && !b_skip && !((h->mb.i_type) == B_DIRECT)) {
                if (h->mb.i_partition != D_8x8)
                    h->stat.frame.i_mb_partition[h->mb.i_partition] += 4;
                else
                    for (int i = 0; i < 4; i++)
                        h->stat.frame.i_mb_partition[h->mb.i_sub_partition[i]]++;
                if (h->param.i_frame_reference > 1)
                    for (int i_list = 0; i_list <= (h->sh.i_type == SLICE_TYPE_B); i_list++)
                        for (int i = 0; i < 4; i++) {
                            int i_ref = h->mb.cache.ref[i_list][x264_scan8[4 * i]];
                            if (i_ref >= 0)
                                h->stat.frame.i_mb_count_ref[i_list][i_ref]++;
                        }
            }
        }
        if (h->param.i_log_level >= 2) {
            if (h->mb.i_cbp_luma | h->mb.i_cbp_chroma) {
                if ((h->sps->i_chroma_format_idc == CHROMA_444)) {
                    for (int i = 0; i < 4; i++)
                        if (h->mb.i_cbp_luma & (1 << i))
                            for (int p = 0; p < 3; p++) {
                                int s8 = i * 4 + p * 16;
                                int nnz8x8 = (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[s8] + 0]))->i) | (((x264_union16_t *) (&h->mb.cache.non_zero_count[x264_scan8[s8] + 8]))->i);
                                h->stat.frame.i_mb_cbp[!b_intra + p * 2] += !!nnz8x8;
                            }
                }
                else {
                    int cbpsum = (h->mb.i_cbp_luma & 1) + ((h->mb.i_cbp_luma >> 1) & 1) + ((h->mb.i_cbp_luma >> 2) & 1) + (h->mb.i_cbp_luma >> 3);
                    h->stat.frame.i_mb_cbp[!b_intra + 0] += cbpsum;
                    h->stat.frame.i_mb_cbp[!b_intra + 2] += !!h->mb.i_cbp_chroma;
                    h->stat.frame.i_mb_cbp[!b_intra + 4] += h->mb.i_cbp_chroma >> 1;
                }
            }
            if (h->mb.i_cbp_luma && !b_intra) {
                h->stat.frame.i_mb_count_8x8dct[0]++;
                h->stat.frame.i_mb_count_8x8dct[1] += h->mb.b_transform_8x8;
            }
            if (b_intra && h->mb.i_type != I_PCM) {
                if (h->mb.i_type == I_16x16)
                    h->stat.frame.i_mb_pred_mode[0][h->mb.i_intra16x16_pred_mode]++;
                else if (h->mb.i_type == I_8x8)
                    for (int i = 0; i < 16; i += 4)
                        h->stat.frame.i_mb_pred_mode[1][h->mb.cache.intra4x4_pred_mode[x264_scan8[i]]]++;
                else
                    for (int i = 0; i < 16; i++)
                        h->stat.frame.i_mb_pred_mode[2][h->mb.cache.intra4x4_pred_mode[x264_scan8[i]]]++;
                h->stat.frame.i_mb_pred_mode[3][x264_mb_chroma_pred_mode_fix[h->mb.i_chroma_pred_mode]]++;
            }
            h->stat.frame.i_mb_field[b_intra ? 0 : b_skip ? 2 : 1] += 0;
        }
        if (b_deblock)
            x264_macroblock_deblock_strength (h);
        if (mb_xy == h->sh.i_last_mb)
            break;
        if (0) {
            i_mb_x += i_mb_y & 1;
            i_mb_y ^= i_mb_x < h->mb.i_mb_width;
        }
        else
            i_mb_x++;
        if (i_mb_x == h->mb.i_mb_width) {
            i_mb_y++;
            i_mb_x = 0;
        }
    }
    if (h->sh.i_last_mb < h->sh.i_first_mb)
        return 0;
    h->out.nal[h->out.i_nal].i_last_mb = h->sh.i_last_mb;
    if (h->param.b_cabac) {
        x264_cabac_encode_flush (h, & h -> cabac);
        h->out.bs.p = h->cabac.p;
    }
    else {
        if (i_skip > 0)
            bs_write_ue_big (&h->out.bs, i_skip);
        bs_rbsp_trailing (& h -> out.bs);
        bs_flush (& h -> out.bs);
    }
    if (x264_nal_end (h))
        return -1;
    if (h->sh.i_last_mb == (h->i_threadslice_end * h->mb.i_mb_width - 1)) {
        h->stat.frame.i_misc_bits = bs_pos (&h->out.bs) + (h->out.i_nal * 5 * 8) - h->stat.frame.i_tex_bits - h->stat.frame.i_mv_bits;
        x264_fdec_filter_row (h, h -> i_threadslice_end, 0);
        if (h->param.b_sliced_threads) {
            x264_threadslice_cond_broadcast (h, 1);
            for (int mb_y = h->i_threadslice_start; mb_y <= h->i_threadslice_end; mb_y++)
                x264_fdec_filter_row (h, mb_y, 1);
            x264_threadslice_cond_broadcast (h, 2);
            if (h->i_thread_idx > 0) {
                x264_threadslice_cond_wait (h -> thread [h -> i_thread_idx - 1], 2);
                x264_fdec_filter_row (h, h -> i_threadslice_start + (1 << 0), 2);
            }
        }
        if (h->fdec->mb_info_free && (!h->param.b_sliced_threads || h->i_thread_idx == (h->param.i_threads - 1))) {
            h->fdec->mb_info_free (h->fdec->mb_info);
            h->fdec->mb_info = ((void *) 0);
            h->fdec->mb_info_free = ((void *) 0);
        }
    }
    return 0;
}

void x264_thread_sync_context (x264_t *dst, x264_t *src) {
    if (dst == src)
        return;
    for (x264_frame_t **f = src->frames.reference; *f; f++)
        (*f)->i_reference_count++;
    for (x264_frame_t **f = dst->frames.reference; *f; f++)
        x264_frame_push_unused (src, *f);
    src->fdec->i_reference_count++;
    x264_frame_push_unused (src, dst -> fdec);
    memcpy (& dst -> i_frame, & src -> i_frame, __builtin_offsetof (x264_t, mb.base) - __builtin_offsetof (x264_t, i_frame));
    dst->param = src->param;
    dst->stat = src->stat;
    dst->pixf = src->pixf;
    dst->reconfig = src->reconfig;
}

void x264_thread_sync_stat (x264_t *dst, x264_t *src) {
    if (dst == src)
        return;
    memcpy (& dst -> stat.i_frame_count, & src -> stat.i_frame_count, sizeof (dst -> stat) - sizeof (dst -> stat.frame));
}

void *x264_slices_write (x264_t *h) {
    int i_slice_num = 0;
    int last_thread_mb = h->sh.i_last_mb;
    memset (& h -> stat.frame, 0, sizeof (h -> stat.frame));
    h->mb.b_reencode_mb = 0;
    while (h->sh.i_first_mb + 0 * h->mb.i_mb_stride <= last_thread_mb) {
        h->sh.i_last_mb = last_thread_mb;
        if (!i_slice_num || !x264_frame_new_slice (h, h->fdec)) {
            if (h->param.i_slice_max_mbs) {
                if (0) {
                    int last_mbaff = 2 * (h->sh.i_first_mb % h->mb.i_mb_width) + h->mb.i_mb_width * (h->sh.i_first_mb / h->mb.i_mb_width) + h->param.i_slice_max_mbs - 1;
                    int last_x = (last_mbaff % (2 * h->mb.i_mb_width)) / 2;
                    int last_y = (last_mbaff / (2 * h->mb.i_mb_width)) * 2 + 1;
                    h->sh.i_last_mb = last_x + h->mb.i_mb_stride * last_y;
                }
                else {
                    h->sh.i_last_mb = h->sh.i_first_mb + h->param.i_slice_max_mbs - 1;
                    if (h->sh.i_last_mb < last_thread_mb && last_thread_mb - h->sh.i_last_mb < h->param.i_slice_min_mbs)
                        h->sh.i_last_mb = last_thread_mb - h->param.i_slice_min_mbs;
                }
                i_slice_num++;
            }
            else if (h->param.i_slice_count && !h->param.b_sliced_threads) {
                int height = h->mb.i_mb_height >> 0;
                int width = h->mb.i_mb_width << 0;
                i_slice_num++;
                h->sh.i_last_mb = (height * i_slice_num + h->param.i_slice_count / 2) / h->param.i_slice_count * width - 1;
            }
        }
        h->sh.i_last_mb = ((h->sh.i_last_mb) < (last_thread_mb) ? (h->sh.i_last_mb) : (last_thread_mb));
        if (x264_slice_write (h))
            goto fail;
        h->sh.i_first_mb = h->sh.i_last_mb + 1;
        if (0 && h->sh.i_first_mb % h->mb.i_mb_width)
            h->sh.i_first_mb -= h->mb.i_mb_stride;
    }
    return (void *) 0;
fail :
    if (h->param.b_sliced_threads)
        x264_threadslice_cond_broadcast (h, 2);
    return (void *) -1;
}

int x264_threaded_slices_write (x264_t *h) {
    for (int i = 0; i < h->param.i_threads; i++) {
        x264_t *t = h->thread[i];
        if (i) {
            t->param = h->param;
            memcpy (& t -> i_frame, & h -> i_frame, __builtin_offsetof (x264_t, rc) - __builtin_offsetof (x264_t, i_frame));
        }
        int height = h->mb.i_mb_height >> 0;
        t->i_threadslice_start = ((height * i + h->param.i_slice_count / 2) / h->param.i_threads) << 0;
        t->i_threadslice_end = ((height * (i + 1) + h->param.i_slice_count / 2) / h->param.i_threads) << 0;
        t->sh.i_first_mb = t->i_threadslice_start * h->mb.i_mb_width;
        t->sh.i_last_mb = t->i_threadslice_end * h->mb.i_mb_width - 1;
    }
    x264_analyse_weight_frame (h, h -> mb.i_mb_height * 16 + 16);
    x264_threads_distribute_ratecontrol (h);
    for (int i = 0; i < h->param.i_threads; i++) {
        h->thread[i]->i_thread_idx = i;
        h->thread[i]->b_thread_active = 1;
        x264_threadslice_cond_broadcast (h -> thread [i], 0);
    }
    for (int i = 0; i < h->param.i_threads; i++)
        ;
    for (int i = 0; i < h->param.i_threads; i++)
        x264_threadslice_cond_wait (h->thread[i], 1);
    x264_threads_merge_ratecontrol (h);
    for (int i = 1; i < h->param.i_threads; i++) {
        x264_t *t = h->thread[i];
        for (int j = 0; j < t->out.i_nal; j++) {
            h->out.nal[h->out.i_nal] = t->out.nal[j];
            h->out.i_nal++;
            x264_nal_check_buffer (h);
        }
        for (int j = 0; j < (__builtin_offsetof (x264_t, stat.frame.i_ssd) - __builtin_offsetof (x264_t, stat.frame.i_mv_bits)) / sizeof (int); j++)
            ((int *) &h->stat.frame)[j] += ((int *) &t->stat.frame)[j];
        for (int j = 0; j < 3; j++)
            h->stat.frame.i_ssd[j] += t->stat.frame.i_ssd[j];
        h->stat.frame.f_ssim += t->stat.frame.f_ssim;
        h->stat.frame.i_ssim_cnt += t->stat.frame.i_ssim_cnt;
    }
    return 0;
}

void x264_encoder_intra_refresh (x264_t *h) {
    h = h->thread[h->i_thread_phase];
    h->b_queued_intra_refresh = 1;
}

int x264_encoder_invalidate_reference (x264_t *h, int64_t pts) {
    if (h->param.i_bframe) {
        x264_log (h, 0, "x264_encoder_invalidate_reference is not supported with B-frames enabled\n");
        return -1;
    }
    if (h->param.b_intra_refresh) {
        x264_log (h, 0, "x264_encoder_invalidate_reference is not supported with intra refresh enabled\n");
        return -1;
    }
    h = h->thread[h->i_thread_phase];
    if (pts >= h->i_last_idr_pts) {
        for (int i = 0; h->frames.reference[i]; i++)
            if (pts <= h->frames.reference[i]->i_pts)
                h->frames.reference[i]->b_corrupt = 1;
        if (pts <= h->fdec->i_pts)
            h->fdec->b_corrupt = 1;
    }
    return 0;
}

int x264_encoder_encode (x264_t *h, x264_nal_t **pp_nal, int *pi_nal, x264_picture_t *pic_in, x264_picture_t *pic_out) {
    x264_t *thread_current, *thread_prev, *thread_oldest;
    int i_nal_type, i_nal_ref_idc, i_global_qp;
    int overhead = 5;
    if (h->opencl.b_fatal_error)
        return -1;
    if (h->i_thread_frames > 1) {
        thread_prev = h->thread[h->i_thread_phase];
        h->i_thread_phase = (h->i_thread_phase + 1) % h->i_thread_frames;
        thread_current = h->thread[h->i_thread_phase];
        thread_oldest = h->thread[(h->i_thread_phase + 1) % h->i_thread_frames];
        x264_thread_sync_context (thread_current, thread_prev);
        x264_thread_sync_ratecontrol (thread_current, thread_prev, thread_oldest);
        h = thread_current;
    }
    else {
        thread_current = thread_oldest = h;
    }
    h->i_cpb_delay_pir_offset = h->i_cpb_delay_pir_offset_next;
    *pi_nal = 0;
    *pp_nal = ((void *) 0);
    if (pic_in != ((void *) 0)) {
        x264_frame_t *fenc = x264_frame_pop_unused (h, 0);
        if (!fenc)
            return -1;
        if (x264_frame_copy_picture (h, fenc, pic_in) < 0)
            return -1;
        if (h->param.i_width != 16 * h->mb.i_mb_width || h->param.i_height != 16 * h->mb.i_mb_height)
            x264_frame_expand_border_mod16 (h, fenc);
        fenc->i_frame = h->frames.i_input++;
        if (fenc->i_frame == 0)
            h->frames.i_first_pts = fenc->i_pts;
        if (h->frames.i_bframe_delay && fenc->i_frame == h->frames.i_bframe_delay)
            h->frames.i_bframe_delay_time = fenc->i_pts - h->frames.i_first_pts;
        if (h->param.b_vfr_input && fenc->i_pts <= h->frames.i_largest_pts)
            x264_log (h, 1, "non-strictly-monotonic PTS\n");
        h->frames.i_second_largest_pts = h->frames.i_largest_pts;
        h->frames.i_largest_pts = fenc->i_pts;
        if ((fenc->i_pic_struct < PIC_STRUCT_AUTO) || (fenc->i_pic_struct > PIC_STRUCT_TRIPLE))
            fenc->i_pic_struct = PIC_STRUCT_AUTO;
        if (fenc->i_pic_struct == PIC_STRUCT_AUTO) {
            int b_interlaced = 0;
            if (b_interlaced) {
                int b_tff = fenc->param ? fenc->param->b_tff : h->param.b_tff;
                fenc->i_pic_struct = b_tff ? PIC_STRUCT_TOP_BOTTOM : PIC_STRUCT_BOTTOM_TOP;
            }
            else
                fenc->i_pic_struct = PIC_STRUCT_PROGRESSIVE;
        }
        if (h->param.rc.b_mb_tree && h->param.rc.b_stat_read) {
            if (x264_macroblock_tree_read (h, fenc, pic_in->prop.quant_offsets))
                return -1;
        }
        else
            x264_adaptive_quant_frame (h, fenc, pic_in->prop.quant_offsets);
        if (pic_in->prop.quant_offsets_free)
            pic_in->prop.quant_offsets_free (pic_in->prop.quant_offsets);
        if (h->frames.b_have_lowres)
            x264_frame_init_lowres (h, fenc);
        x264_lookahead_put_frame (h, fenc);
        if (h->frames.i_input <= h->frames.i_delay + 1 - h->i_thread_frames) {
            pic_out->i_type = 0x0000;
            return 0;
        }
    }
    else {
        ;
        h->lookahead->b_exit_thread = 1;
        ;
        ;
    }
    h->i_frame++;
    if (!h->frames.current[0])
        x264_lookahead_get_frames (h);
    if (!h->frames.current[0] && x264_lookahead_is_empty (h))
        return x264_encoder_frame_end (thread_oldest, thread_current, pp_nal, pi_nal, pic_out);
    h->fenc = x264_frame_shift (h->frames.current);
    if (h->param.b_sliced_threads)
        if (x264_threadpool_wait_all (h) < 0)
            return -1;
    if (h->i_frame == h->i_thread_frames - 1)
        h->i_reordered_pts_delay = h->fenc->i_reordered_pts;
    if (h->reconfig) {
        x264_encoder_reconfig_apply (h, & h -> reconfig_h -> param);
        h->reconfig = 0;
    }
    if (h->fenc->param) {
        x264_encoder_reconfig_apply (h, h -> fenc -> param);
        if (h->fenc->param->param_free) {
            h->fenc->param->param_free (h->fenc->param);
            h->fenc->param = ((void *) 0);
        }
    }
    if (x264_reference_update (h))
        return -1;
    h->fdec->i_lines_completed = -1;
    if (!((h->fenc->i_type) == 0x0002 || (h->fenc->i_type) == 0x0001)) {
        int valid_refs_left = 0;
        for (int i = 0; h->frames.reference[i]; i++)
            if (!h->frames.reference[i]->b_corrupt)
                valid_refs_left++;
        if (!valid_refs_left) {
            h->fenc->b_keyframe = 1;
            h->fenc->i_type = 0x0001;
        }
    }
    if (h->fenc->b_keyframe) {
        h->frames.i_last_keyframe = h->fenc->i_frame;
        if (h->fenc->i_type == 0x0001) {
            h->i_frame_num = 0;
            h->frames.i_last_idr = h->fenc->i_frame;
        }
    }
    h->sh.i_mmco_command_count = h->sh.i_mmco_remove_from_end = 0;
    h->b_ref_reorder[0] = h->b_ref_reorder[1] = 0;
    h->fdec->i_poc = h->fenc->i_poc = 2 * (h->fenc->i_frame - ((h->frames.i_last_idr) > (0) ? (h->frames.i_last_idr) : (0)));
    if (h->fenc->i_type == 0x0001) {
        i_nal_type = NAL_SLICE_IDR;
        i_nal_ref_idc = NAL_PRIORITY_HIGHEST;
        h->sh.i_type = SLICE_TYPE_I;
        x264_reference_reset (h);
        h->frames.i_poc_last_open_gop = -1;
    }
    else if (h->fenc->i_type == 0x0002) {
        i_nal_type = NAL_SLICE;
        i_nal_ref_idc = NAL_PRIORITY_HIGH;
        h->sh.i_type = SLICE_TYPE_I;
        x264_reference_hierarchy_reset (h);
        if (h->param.b_open_gop)
            h->frames.i_poc_last_open_gop = h->fenc->b_keyframe ? h->fenc->i_poc : -1;
    }
    else if (h->fenc->i_type == 0x0003) {
        i_nal_type = NAL_SLICE;
        i_nal_ref_idc = NAL_PRIORITY_HIGH;
        h->sh.i_type = SLICE_TYPE_P;
        x264_reference_hierarchy_reset (h);
        h->frames.i_poc_last_open_gop = -1;
    }
    else if (h->fenc->i_type == 0x0004) {
        i_nal_type = NAL_SLICE;
        i_nal_ref_idc = h->param.i_bframe_pyramid == 1 ? NAL_PRIORITY_LOW : NAL_PRIORITY_HIGH;
        h->sh.i_type = SLICE_TYPE_B;
        x264_reference_hierarchy_reset (h);
    }
    else {
        i_nal_type = NAL_SLICE;
        i_nal_ref_idc = NAL_PRIORITY_DISPOSABLE;
        h->sh.i_type = SLICE_TYPE_B;
    }
    h->fdec->i_type = h->fenc->i_type;
    h->fdec->i_frame = h->fenc->i_frame;
    h->fenc->b_kept_as_ref = h->fdec->b_kept_as_ref = i_nal_ref_idc != NAL_PRIORITY_DISPOSABLE && h->param.i_keyint_max > 1;
    h->fdec->mb_info = h->fenc->mb_info;
    h->fdec->mb_info_free = h->fenc->mb_info_free;
    h->fenc->mb_info = ((void *) 0);
    h->fenc->mb_info_free = ((void *) 0);
    h->fdec->i_pts = h->fenc->i_pts;
    if (h->frames.i_bframe_delay) {
        int64_t *prev_reordered_pts = thread_current->frames.i_prev_reordered_pts;
        h->fdec->i_dts = h->i_frame > h->frames.i_bframe_delay ? prev_reordered_pts[(h->i_frame - h->frames.i_bframe_delay) % h->frames.i_bframe_delay] : h->fenc->i_reordered_pts - h->frames.i_bframe_delay_time;
        prev_reordered_pts[h->i_frame % h->frames.i_bframe_delay] = h->fenc->i_reordered_pts;
    }
    else
        h->fdec->i_dts = h->fenc->i_reordered_pts;
    if (h->fenc->i_type == 0x0001)
        h->i_last_idr_pts = h->fdec->i_pts;
    x264_reference_build_list (h, h -> fdec -> i_poc);
    if (h->param.b_sliced_threads) {
        for (int i = 0; i < h->param.i_threads; i++) {
            bs_init (& h -> thread [i] -> out.bs, h -> thread [i] -> out.p_bitstream, h -> thread [i] -> out.i_bitstream);
            h->thread[i]->out.i_nal = 0;
        }
    }
    else {
        bs_init (& h -> out.bs, h -> out.p_bitstream, h -> out.i_bitstream);
        h->out.i_nal = 0;
    }
    if (h->param.b_aud) {
        int pic_type;
        if (h->sh.i_type == SLICE_TYPE_I)
            pic_type = 0;
        else if (h->sh.i_type == SLICE_TYPE_P)
            pic_type = 1;
        else if (h->sh.i_type == SLICE_TYPE_B)
            pic_type = 2;
        else
            pic_type = 7;
        x264_nal_start (h, NAL_AUD, NAL_PRIORITY_DISPOSABLE);
        bs_write (& h -> out.bs, 3, pic_type);
        bs_rbsp_trailing (& h -> out.bs);
        if (x264_nal_end (h))
            return -1;
        overhead += h->out.nal[h->out.i_nal - 1].i_payload + 5;
    }
    h->i_nal_type = i_nal_type;
    h->i_nal_ref_idc = i_nal_ref_idc;
    if (h->param.b_intra_refresh) {
        if (((h->fenc->i_type) == 0x0002 || (h->fenc->i_type) == 0x0001)) {
            h->fdec->i_frames_since_pir = 0;
            h->b_queued_intra_refresh = 0;
            h->fdec->f_pir_position = h->mb.i_mb_width;
        }
        else if (h->fenc->i_type == 0x0003) {
            int pocdiff = (h->fdec->i_poc - h->fref[0][0]->i_poc) / 2;
            float increment = ((((float) h->mb.i_mb_width - 1) / h->param.i_keyint_max) > (1) ? (((float) h->mb.i_mb_width - 1) / h->param.i_keyint_max) : (1));
            h->fdec->f_pir_position = h->fref[0][0]->f_pir_position;
            h->fdec->i_frames_since_pir = h->fref[0][0]->i_frames_since_pir + pocdiff;
            if (h->fdec->i_frames_since_pir >= h->param.i_keyint_max || (h->b_queued_intra_refresh && h->fdec->f_pir_position + 0.5 >= h->mb.i_mb_width)) {
                h->fdec->f_pir_position = 0;
                h->fdec->i_frames_since_pir = 0;
                h->b_queued_intra_refresh = 0;
                h->fenc->b_keyframe = 1;
            }
            h->fdec->i_pir_start_col = h->fdec->f_pir_position + 0.5;
            h->fdec->f_pir_position += increment * pocdiff;
            h->fdec->i_pir_end_col = h->fdec->f_pir_position + 0.5;
            if (h->fdec->i_pir_end_col >= h->mb.i_mb_width - 1) {
                h->fdec->f_pir_position = h->mb.i_mb_width;
                h->fdec->i_pir_end_col = h->mb.i_mb_width - 1;
            }
        }
    }
    if (h->fenc->b_keyframe) {
        if (h->param.b_repeat_headers) {
            x264_nal_start (h, NAL_SPS, NAL_PRIORITY_HIGHEST);
            x264_sps_write (& h -> out.bs, h -> sps);
            if (x264_nal_end (h))
                return -1;
            if (h->param.i_avcintra_class)
                h->out.nal[h->out.i_nal - 1].i_padding = 256 - bs_pos (&h->out.bs) / 8 - 2 * 5;
            overhead += h->out.nal[h->out.i_nal - 1].i_payload + h->out.nal[h->out.i_nal - 1].i_padding + 5;
            x264_nal_start (h, NAL_PPS, NAL_PRIORITY_HIGHEST);
            x264_pps_write (& h -> out.bs, h -> sps, h -> pps);
            if (x264_nal_end (h))
                return -1;
            if (h->param.i_avcintra_class)
                h->out.nal[h->out.i_nal - 1].i_padding = 256 - h->out.nal[h->out.i_nal - 1].i_payload - 5;
            overhead += h->out.nal[h->out.i_nal - 1].i_payload + h->out.nal[h->out.i_nal - 1].i_padding + 5;
        }
        if (h->i_thread_frames == 1 && h->sps->vui.b_nal_hrd_parameters_present) {
            x264_hrd_fullness (h);
            x264_nal_start (h, NAL_SEI, NAL_PRIORITY_DISPOSABLE);
            x264_sei_buffering_period_write (h, & h -> out.bs);
            if (x264_nal_end (h))
                return -1;
            overhead += h->out.nal[h->out.i_nal - 1].i_payload + (5 - (h->param.b_annexb && !h->param.i_avcintra_class && (h->out.i_nal - 1)));
        }
    }
    for (int i = 0; i < h->fenc->extra_sei.num_payloads; i++) {
        x264_nal_start (h, NAL_SEI, NAL_PRIORITY_DISPOSABLE);
        x264_sei_write (& h -> out.bs, h -> fenc -> extra_sei.payloads [i].payload, h -> fenc -> extra_sei.payloads [i].payload_size, h -> fenc -> extra_sei.payloads [i].payload_type);
        if (x264_nal_end (h))
            return -1;
        overhead += h->out.nal[h->out.i_nal - 1].i_payload + (5 - (h->param.b_annexb && !h->param.i_avcintra_class && (h->out.i_nal - 1)));
        if (h->fenc->extra_sei.sei_free) {
            h->fenc->extra_sei.sei_free (h->fenc->extra_sei.payloads[i].payload);
            h->fenc->extra_sei.payloads[i].payload = ((void *) 0);
        }
    }
    if (h->fenc->extra_sei.sei_free) {
        h->fenc->extra_sei.sei_free (h->fenc->extra_sei.payloads);
        h->fenc->extra_sei.payloads = ((void *) 0);
        h->fenc->extra_sei.sei_free = ((void *) 0);
    }
    if (h->fenc->b_keyframe) {
        if (h->param.b_repeat_headers && h->fenc->i_frame == 0 && !h->param.i_avcintra_class) {
            x264_nal_start (h, NAL_SEI, NAL_PRIORITY_DISPOSABLE);
            if (x264_sei_version_write (h, &h->out.bs))
                return -1;
            if (x264_nal_end (h))
                return -1;
            overhead += h->out.nal[h->out.i_nal - 1].i_payload + (5 - (h->param.b_annexb && !h->param.i_avcintra_class && (h->out.i_nal - 1)));
        }
        if (h->fenc->i_type != 0x0001) {
            int time_to_recovery = h->param.b_open_gop ? 0 : ((h->mb.i_mb_width - 1) < (h->param.i_keyint_max) ? (h->mb.i_mb_width - 1) : (h->param.i_keyint_max)) + h->param.i_bframe - 1;
            x264_nal_start (h, NAL_SEI, NAL_PRIORITY_DISPOSABLE);
            x264_sei_recovery_point_write (h, & h -> out.bs, time_to_recovery);
            if (x264_nal_end (h))
                return -1;
            overhead += h->out.nal[h->out.i_nal - 1].i_payload + (5 - (h->param.b_annexb && !h->param.i_avcintra_class && (h->out.i_nal - 1)));
        }
    }
    if (h->param.i_frame_packing >= 0 && (h->fenc->b_keyframe || h->param.i_frame_packing == 5)) {
        x264_nal_start (h, NAL_SEI, NAL_PRIORITY_DISPOSABLE);
        x264_sei_frame_packing_write (h, & h -> out.bs);
        if (x264_nal_end (h))
            return -1;
        overhead += h->out.nal[h->out.i_nal - 1].i_payload + (5 - (h->param.b_annexb && !h->param.i_avcintra_class && (h->out.i_nal - 1)));
    }
    if (h->sps->vui.b_pic_struct_present || h->sps->vui.b_nal_hrd_parameters_present) {
        x264_nal_start (h, NAL_SEI, NAL_PRIORITY_DISPOSABLE);
        x264_sei_pic_timing_write (h, & h -> out.bs);
        if (x264_nal_end (h))
            return -1;
        overhead += h->out.nal[h->out.i_nal - 1].i_payload + (5 - (h->param.b_annexb && !h->param.i_avcintra_class && (h->out.i_nal - 1)));
    }
    if (!((h->fenc->i_type) == 0x0005 || (h->fenc->i_type) == 0x0004) && h->b_sh_backup) {
        h->b_sh_backup = 0;
        x264_nal_start (h, NAL_SEI, NAL_PRIORITY_DISPOSABLE);
        x264_sei_dec_ref_pic_marking_write (h, & h -> out.bs);
        if (x264_nal_end (h))
            return -1;
        overhead += h->out.nal[h->out.i_nal - 1].i_payload + (5 - (h->param.b_annexb && !h->param.i_avcintra_class && (h->out.i_nal - 1)));
    }
    if (h->fenc->b_keyframe && h->param.b_intra_refresh)
        h->i_cpb_delay_pir_offset_next = h->fenc->i_cpb_delay;
    if (h->param.i_avcintra_class) {
        x264_nal_start (h, NAL_FILLER, NAL_PRIORITY_DISPOSABLE);
        x264_filler_write (h, & h -> out.bs, 0);
        if (x264_nal_end (h))
            return -1;
        overhead += h->out.nal[h->out.i_nal - 1].i_payload + 5;
        x264_nal_start (h, NAL_SEI, NAL_PRIORITY_DISPOSABLE);
        if (x264_sei_avcintra_umid_write (h, &h->out.bs) < 0)
            return -1;
        if (x264_nal_end (h))
            return -1;
        overhead += h->out.nal[h->out.i_nal - 1].i_payload + (5 - (h->param.b_annexb && !h->param.i_avcintra_class && (h->out.i_nal - 1)));
        int unpadded_len;
        int total_len;
        if (h->param.i_height == 1080) {
            unpadded_len = 5780;
            total_len = 17 * 512;
        }
        else {
            unpadded_len = 2900;
            total_len = 9 * 512;
        }
        x264_nal_start (h, NAL_SEI, NAL_PRIORITY_DISPOSABLE);
        if (x264_sei_avcintra_vanc_write (h, &h->out.bs, unpadded_len) < 0)
            return -1;
        if (x264_nal_end (h))
            return -1;
        h->out.nal[h->out.i_nal - 1].i_padding = total_len - h->out.nal[h->out.i_nal - 1].i_payload - (5 - (h->param.b_annexb && !h->param.i_avcintra_class && (h->out.i_nal - 1)));
        overhead += h->out.nal[h->out.i_nal - 1].i_payload + h->out.nal[h->out.i_nal - 1].i_padding + (5 - (h->param.b_annexb && !h->param.i_avcintra_class && (h->out.i_nal - 1)));
    }
    x264_ratecontrol_start (h, h -> fenc -> i_qpplus1, overhead * 8);
    i_global_qp = x264_ratecontrol_qp (h);
    pic_out->i_qpplus1 = h->fdec->i_qpplus1 = i_global_qp + 1;
    if (h->param.rc.b_stat_read && h->sh.i_type != SLICE_TYPE_I) {
        x264_reference_build_list_optimal (h);
        x264_reference_check_reorder (h);
    }
    if (h->i_ref[0])
        h->fdec->i_poc_l0ref0 = h->fref[0][0]->i_poc;
    x264_slice_init (h, i_nal_type, i_global_qp);
    if (h->sh.i_type == SLICE_TYPE_B)
        x264_macroblock_bipred_init (h);
    x264_weighted_pred_init (h);
    if (i_nal_ref_idc != NAL_PRIORITY_DISPOSABLE)
        h->i_frame_num++;
    h->i_threadslice_start = 0;
    h->i_threadslice_end = h->mb.i_mb_height;
    if (h->i_thread_frames > 1) {
        ;
        h->b_thread_active = 1;
    }
    else if (h->param.b_sliced_threads) {
        if (x264_threaded_slices_write (h))
            return -1;
    }
    else if ((intptr_t) x264_slices_write (h))
        return -1;
    return x264_encoder_frame_end (thread_oldest, thread_current, pp_nal, pi_nal, pic_out);
}

int x264_encoder_frame_end (x264_t *h, x264_t *thread_current, x264_nal_t **pp_nal, int *pi_nal, x264_picture_t *pic_out) {
    char psz_message [80];
    if (!h->param.b_sliced_threads && h->b_thread_active) {
        h->b_thread_active = 0;
        if ((intptr_t) ((void *) 0))
            return -1;
    }
    if (!h->out.i_nal) {
        pic_out->i_type = 0x0000;
        return 0;
    };
    if (h->i_thread_frames > 1 && h->fenc->b_keyframe && h->sps->vui.b_nal_hrd_parameters_present) {
        x264_hrd_fullness (h);
        x264_nal_start (h, NAL_SEI, NAL_PRIORITY_DISPOSABLE);
        x264_sei_buffering_period_write (h, & h -> out.bs);
        if (x264_nal_end (h))
            return -1;
        int idx = 0;
        while (h->out.nal[idx].i_type == NAL_AUD || h->out.nal[idx].i_type == NAL_SPS || h->out.nal[idx].i_type == NAL_PPS)
            idx++;
        x264_nal_t nal_tmp = h->out.nal[h->out.i_nal - 1];
        memmove (& h -> out.nal [idx + 1], & h -> out.nal [idx], (h -> out.i_nal - idx - 1) * sizeof (x264_nal_t));
        h->out.nal[idx] = nal_tmp;
    }
    int frame_size = x264_encoder_encapsulate_nals (h, 0);
    if (frame_size < 0)
        return -1;
    pic_out->i_type = h->fenc->i_type;
    pic_out->b_keyframe = h->fenc->b_keyframe;
    pic_out->i_pic_struct = h->fenc->i_pic_struct;
    pic_out->i_pts = h->fdec->i_pts;
    pic_out->i_dts = h->fdec->i_dts;
    if (pic_out->i_pts < pic_out->i_dts)
        x264_log (h, 1, "invalid DTS: PTS is less than DTS\n");
    pic_out->opaque = h->fenc->opaque;
    pic_out->img.i_csp = h->fdec->i_csp;
    pic_out->img.i_plane = h->fdec->i_plane;
    for (int i = 0; i < pic_out->img.i_plane; i++) {
        pic_out->img.i_stride[i] = h->fdec->i_stride[i] * sizeof (pixel);
        pic_out->img.plane[i] = (uint8_t *) h->fdec->plane[i];
    }
    x264_frame_push_unused (thread_current, h -> fenc);
    int filler = 0;
    if (x264_ratecontrol_end (h, frame_size *8, &filler) < 0)
        return -1;
    pic_out->hrd_timing = h->fenc->hrd_timing;
    pic_out->prop.f_crf_avg = h->fdec->f_crf_avg;
    if (h->param.i_avcintra_class) {
        x264_t *h0 = h->thread[0];
        int ret = x264_check_encapsulated_buffer (h, h0, h->out.i_nal, frame_size, frame_size +filler);
        if (ret < 0)
            return -1;
        memset (h -> out.nal [0].p_payload + frame_size, 0, filler);
        h->out.nal[h->out.i_nal - 1].i_payload += filler;
        h->out.nal[h->out.i_nal - 1].i_padding = filler;
        frame_size += filler;
    }
    else {
        while (filler > 0) {
            int f, overhead;
            overhead = ((5 + 1) - h->param.b_annexb);
            if (h->param.i_slice_max_size && filler > h->param.i_slice_max_size) {
                int next_size = filler - h->param.i_slice_max_size;
                int overflow = ((overhead - next_size) > (0) ? (overhead - next_size) : (0));
                f = h->param.i_slice_max_size - overhead - overflow;
            }
            else
                f = ((0) > (filler - overhead) ? (0) : (filler - overhead));
            if (x264_bitstream_check_buffer_filler (h, f))
                return -1;
            x264_nal_start (h, NAL_FILLER, NAL_PRIORITY_DISPOSABLE);
            x264_filler_write (h, & h -> out.bs, f);
            if (x264_nal_end (h))
                return -1;
            int total_size = x264_encoder_encapsulate_nals (h, h->out.i_nal - 1);
            if (total_size < 0)
                return -1;
            frame_size += total_size;
            filler -= total_size;
        }
    }
    *pi_nal = h->out.i_nal;
    *pp_nal = h->out.nal;
    h->out.i_nal = 0;
    x264_noise_reduction_update (h);
    x264_thread_sync_stat (h, h -> thread [0]);
    h->stat.i_frame_count[h->sh.i_type]++;
    h->stat.i_frame_size[h->sh.i_type] += frame_size;
    h->stat.f_frame_qp[h->sh.i_type] += h->fdec->f_qp_avg_aq;
    for (int i = 0; i < X264_MBTYPE_MAX; i++)
        h->stat.i_mb_count[h->sh.i_type][i] += h->stat.frame.i_mb_count[i];
    for (int i = 0; i < X264_PARTTYPE_MAX; i++)
        h->stat.i_mb_partition[h->sh.i_type][i] += h->stat.frame.i_mb_partition[i];
    for (int i = 0; i < 2; i++)
        h->stat.i_mb_count_8x8dct[i] += h->stat.frame.i_mb_count_8x8dct[i];
    for (int i = 0; i < 6; i++)
        h->stat.i_mb_cbp[i] += h->stat.frame.i_mb_cbp[i];
    for (int i = 0; i < 4; i++)
        for (int j = 0; j < 13; j++)
            h->stat.i_mb_pred_mode[i][j] += h->stat.frame.i_mb_pred_mode[i][j];
    if (h->sh.i_type != SLICE_TYPE_I)
        for (int i_list = 0; i_list < 2; i_list++)
            for (int i = 0; i < 16 * 2; i++)
                h->stat.i_mb_count_ref[h->sh.i_type][i_list][i] += h->stat.frame.i_mb_count_ref[i_list][i];
    for (int i = 0; i < 3; i++)
        h->stat.i_mb_field[i] += h->stat.frame.i_mb_field[i];
    if (h->sh.i_type == SLICE_TYPE_P && h->param.analyse.i_weighted_pred >= 1) {
        h->stat.i_wpred[0] += !!h->sh.weight[0][0].weightfn;
        h->stat.i_wpred[1] += !!h->sh.weight[0][1].weightfn || !!h->sh.weight[0][2].weightfn;
    }
    if (h->sh.i_type == SLICE_TYPE_B) {
        h->stat.i_direct_frames[h->sh.b_direct_spatial_mv_pred]++;
        if (h->mb.b_direct_auto_write) {
            if (h->stat.i_direct_score[0] + h->stat.i_direct_score[1] > h->mb.i_mb_count)
                for (int i = 0; i < 2; i++)
                    h->stat.i_direct_score[i] = h->stat.i_direct_score[i] * 9 / 10;
            for (int i = 0; i < 2; i++)
                h->stat.i_direct_score[i] += h->stat.frame.i_direct_score[i];
        }
    }
    else
        h->stat.i_consecutive_bframes[h->fenc->i_bframes]++;
    psz_message[0] = '\0';
    double dur = h->fenc->f_duration;
    h->stat.f_frame_duration[h->sh.i_type] += dur;
    if (h->param.analyse.b_psnr) {
        int64_t ssd [3] = {h->stat.frame.i_ssd[0], h->stat.frame.i_ssd[1], h->stat.frame.i_ssd[2],};
        int luma_size = h->param.i_width * h->param.i_height;
        int chroma_size = ((luma_size) >> (h->mb.chroma_h_shift + h->mb.chroma_v_shift));
        pic_out->prop.f_psnr[0] = x264_psnr (ssd[0], luma_size);
        pic_out->prop.f_psnr[1] = x264_psnr (ssd[1], chroma_size);
        pic_out->prop.f_psnr[2] = x264_psnr (ssd[2], chroma_size);
        pic_out->prop.f_psnr_avg = x264_psnr (ssd[0] + ssd[1] + ssd[2], luma_size +chroma_size * 2);
        h->stat.f_ssd_global[h->sh.i_type] += dur * (ssd[0] + ssd[1] + ssd[2]);
        h->stat.f_psnr_average[h->sh.i_type] += dur * pic_out->prop.f_psnr_avg;
        h->stat.f_psnr_mean_y[h->sh.i_type] += dur * pic_out->prop.f_psnr[0];
        h->stat.f_psnr_mean_u[h->sh.i_type] += dur * pic_out->prop.f_psnr[1];
        h->stat.f_psnr_mean_v[h->sh.i_type] += dur * pic_out->prop.f_psnr[2];
        snprintf (psz_message, 80, " PSNR Y:%5.2f U:%5.2f V:%5.2f", pic_out -> prop.f_psnr [0], pic_out -> prop.f_psnr [1], pic_out -> prop.f_psnr [2]);
    }
    if (h->param.analyse.b_ssim) {
        pic_out->prop.f_ssim = h->stat.frame.f_ssim / h->stat.frame.i_ssim_cnt;
        h->stat.f_ssim_mean_y[h->sh.i_type] += pic_out->prop.f_ssim * dur;
        snprintf (psz_message + strlen (psz_message), 80 - strlen (psz_message), " SSIM Y:%.5f", pic_out -> prop.f_ssim);
    }
    psz_message[79] = '\0';
    x264_log (h, 3, "frame=%4d QP=%.2f NAL=%d Slice:%c Poc:%-3d I:%-4d P:%-4d SKIP:%-4d size=%d bytes%s\n", h -> i_frame, h -> fdec -> f_qp_avg_aq, h -> i_nal_ref_idc, h -> sh.i_type == SLICE_TYPE_I ? 'I' : (h -> sh.i_type == SLICE_TYPE_P ? 'P' : 'B'), h -> fdec -> i_poc, h -> stat.frame.i_mb_count_i, h -> stat.frame.i_mb_count_p, h -> stat.frame.i_mb_count_skip, frame_size, psz_message);
    x264_thread_sync_stat (h -> thread [0], h);
    x264_thread_sync_stat (thread_current, h);
    for (int i = 0; i < h->i_ref[0]; i++)
        if (h->fref[0][i] && h->fref[0][i]->b_duplicate) {
            x264_frame_push_blank_unused (h, h -> fref [0] [i]);
            h->fref[0][i] = 0;
        }
    if (h->param.psz_dump_yuv)
        x264_frame_dump (h);
    ;
    return frame_size;
}

void x264_print_intra (int64_t *i_mb_count, double i_count, int b_print_pcm, char *intra) {
    intra += sprintf (intra, "I16..4%s: %4.1f%% %4.1f%% %4.1f%%", b_print_pcm ? "..PCM" : "", i_mb_count[I_16x16] / i_count, i_mb_count[I_8x8] / i_count, i_mb_count[I_4x4] / i_count);
    if (b_print_pcm)
        sprintf (intra, " %4.1f%%", i_mb_count[I_PCM] / i_count);
}

void x264_encoder_close (x264_t *h) {
    int64_t i_yuv_size = ((h->param.i_width * h->param.i_height) + 2 * ((h->param.i_width * h->param.i_height) >> (h->mb.chroma_h_shift + h->mb.chroma_v_shift)));
    int64_t i_mb_count_size [2] [7] = {{0}};
    char buf [200];
    int b_print_pcm = h->stat.i_mb_count[SLICE_TYPE_I][I_PCM] || h->stat.i_mb_count[SLICE_TYPE_P][I_PCM] || h->stat.i_mb_count[SLICE_TYPE_B][I_PCM];
    x264_lookahead_delete (h);
    x264_opencl_lookahead_delete (h);
    x264_opencl_function_t *ocl = h->opencl.ocl;
    if (h->param.b_sliced_threads)
        x264_threadpool_wait_all (h);
    if (h->param.i_threads > 1)
        ;
    if (h->param.i_lookahead_threads > 1)
        ;
    if (h->i_thread_frames > 1) {
        for (int i = 0; i < h->i_thread_frames; i++)
            if (h->thread[i]->b_thread_active) {
                ((h->thread[i]->fenc->i_reference_count == 1) ? (void) (0) : __assert_fail ("h->thread[i]->fenc->i_reference_count == 1", "DonorSourceCode/encoder.c", 4023, __PRETTY_FUNCTION__));
                x264_frame_delete (h -> thread [i] -> fenc);
            }
        x264_t *thread_prev = h->thread[h->i_thread_phase];
        x264_thread_sync_ratecontrol (h, thread_prev, h);
        x264_thread_sync_ratecontrol (thread_prev, thread_prev, h);
        h->i_frame = thread_prev->i_frame + 1 - h->i_thread_frames;
    }
    h->i_frame++;
    for (int i = 0; i < 3; i++) {
        static const uint8_t slice_order [] = {SLICE_TYPE_I, SLICE_TYPE_P, SLICE_TYPE_B};
        int i_slice = slice_order[i];
        if (h->stat.i_frame_count[i_slice] > 0) {
            int i_count = h->stat.i_frame_count[i_slice];
            double dur = h->stat.f_frame_duration[i_slice];
            if (h->param.analyse.b_psnr) {
                x264_log (h, 2, "frame %c:%-5d Avg QP:%5.2f  size:%6.0f  PSNR Mean Y:%5.2f U:%5.2f V:%5.2f Avg:%5.2f Global:%5.2f\n", slice_type_to_char [i_slice], i_count, h -> stat.f_frame_qp [i_slice] / i_count, (double) h -> stat.i_frame_size [i_slice] / i_count, h -> stat.f_psnr_mean_y [i_slice] / dur, h -> stat.f_psnr_mean_u [i_slice] / dur, h -> stat.f_psnr_mean_v [i_slice] / dur, h -> stat.f_psnr_average [i_slice] / dur, x264_psnr (h -> stat.f_ssd_global [i_slice], dur * i_yuv_size));
            }
            else {
                x264_log (h, 2, "frame %c:%-5d Avg QP:%5.2f  size:%6.0f\n", slice_type_to_char [i_slice], i_count, h -> stat.f_frame_qp [i_slice] / i_count, (double) h -> stat.i_frame_size [i_slice] / i_count);
            }
        }
    }
    if (h->param.i_bframe && h->stat.i_frame_count[SLICE_TYPE_B]) {
        char *p = buf;
        int den = 0;
        for (int i = 0; i <= h->param.i_bframe; i++)
            den += (i + 1) * h->stat.i_consecutive_bframes[i];
        for (int i = 0; i <= h->param.i_bframe; i++)
            p += sprintf (p, " %4.1f%%", 100.* (i + 1) * h->stat.i_consecutive_bframes[i] / den);
        x264_log (h, 2, "consecutive B-frames:%s\n", buf);
    }
    for (int i_type = 0; i_type < 2; i_type++)
        for (int i = 0; i < X264_PARTTYPE_MAX; i++) {
            if (i == D_DIRECT_8x8)
                continue;
            i_mb_count_size[i_type][x264_mb_partition_pixel_table[i]] += h->stat.i_mb_partition[i_type][i];
        }
    if (h->stat.i_frame_count[SLICE_TYPE_I] > 0) {
        int64_t *i_mb_count = h->stat.i_mb_count[SLICE_TYPE_I];
        double i_count = h->stat.i_frame_count[SLICE_TYPE_I] * h->mb.i_mb_count / 100.0;
        x264_print_intra (i_mb_count, i_count, b_print_pcm, buf);
        x264_log (h, 2, "mb I  %s\n", buf);
    }
    if (h->stat.i_frame_count[SLICE_TYPE_P] > 0) {
        int64_t *i_mb_count = h->stat.i_mb_count[SLICE_TYPE_P];
        double i_count = h->stat.i_frame_count[SLICE_TYPE_P] * h->mb.i_mb_count / 100.0;
        int64_t *i_mb_size = i_mb_count_size[SLICE_TYPE_P];
        x264_print_intra (i_mb_count, i_count, b_print_pcm, buf);
        x264_log (h, 2, "mb P  %s  P16..4: %4.1f%% %4.1f%% %4.1f%% %4.1f%% %4.1f%%    skip:%4.1f%%\n", buf, i_mb_size [PIXEL_16x16] / (i_count * 4), (i_mb_size [PIXEL_16x8] + i_mb_size [PIXEL_8x16]) / (i_count * 4), i_mb_size [PIXEL_8x8] / (i_count * 4), (i_mb_size [PIXEL_8x4] + i_mb_size [PIXEL_4x8]) / (i_count * 4), i_mb_size [PIXEL_4x4] / (i_count * 4), i_mb_count [P_SKIP] / i_count);
    }
    if (h->stat.i_frame_count[SLICE_TYPE_B] > 0) {
        int64_t *i_mb_count = h->stat.i_mb_count[SLICE_TYPE_B];
        double i_count = h->stat.i_frame_count[SLICE_TYPE_B] * h->mb.i_mb_count / 100.0;
        double i_mb_list_count;
        int64_t *i_mb_size = i_mb_count_size[SLICE_TYPE_B];
        int64_t list_count [3] = {0};
        x264_print_intra (i_mb_count, i_count, b_print_pcm, buf);
        for (int i = 0; i < X264_PARTTYPE_MAX; i++)
            for (int j = 0; j < 2; j++) {
                int l0 = x264_mb_type_list_table[i][0][j];
                int l1 = x264_mb_type_list_table[i][1][j];
                if (l0 || l1)
                    list_count[l1 + l0 * l1] += h->stat.i_mb_count[SLICE_TYPE_B][i] * 2;
            }
        list_count[0] += h->stat.i_mb_partition[SLICE_TYPE_B][D_L0_8x8];
        list_count[1] += h->stat.i_mb_partition[SLICE_TYPE_B][D_L1_8x8];
        list_count[2] += h->stat.i_mb_partition[SLICE_TYPE_B][D_BI_8x8];
        i_mb_count[B_DIRECT] += (h->stat.i_mb_partition[SLICE_TYPE_B][D_DIRECT_8x8] + 2) / 4;
        i_mb_list_count = (list_count[0] + list_count[1] + list_count[2]) / 100.0;
        sprintf (buf + strlen (buf), "  B16..8: %4.1f%% %4.1f%% %4.1f%%  direct:%4.1f%%  skip:%4.1f%%", i_mb_size [PIXEL_16x16] / (i_count * 4), (i_mb_size [PIXEL_16x8] + i_mb_size [PIXEL_8x16]) / (i_count * 4), i_mb_size [PIXEL_8x8] / (i_count * 4), i_mb_count [B_DIRECT] / i_count, i_mb_count [B_SKIP] / i_count);
        if (i_mb_list_count != 0)
            sprintf (buf +strlen (buf), "  L0:%4.1f%% L1:%4.1f%% BI:%4.1f%%", list_count[0] / i_mb_list_count, list_count[1] / i_mb_list_count, list_count[2] / i_mb_list_count);
        x264_log (h, 2, "mb B  %s\n", buf);
    }
    x264_ratecontrol_summary (h);
    if (h->stat.i_frame_count[SLICE_TYPE_I] + h->stat.i_frame_count[SLICE_TYPE_P] + h->stat.i_frame_count[SLICE_TYPE_B] > 0) {
        int64_t i_i8x8 = (h->stat.i_mb_count[SLICE_TYPE_I][I_8x8] + h->stat.i_mb_count[SLICE_TYPE_P][I_8x8] + h->stat.i_mb_count[SLICE_TYPE_B][I_8x8]);
        int64_t i_intra = i_i8x8 + (h->stat.i_mb_count[SLICE_TYPE_I][I_4x4] + h->stat.i_mb_count[SLICE_TYPE_P][I_4x4] + h->stat.i_mb_count[SLICE_TYPE_B][I_4x4]) + (h->stat.i_mb_count[SLICE_TYPE_I][I_16x16] + h->stat.i_mb_count[SLICE_TYPE_P][I_16x16] + h->stat.i_mb_count[SLICE_TYPE_B][I_16x16]);
        int64_t i_all_intra = i_intra + (h->stat.i_mb_count[SLICE_TYPE_I][I_PCM] + h->stat.i_mb_count[SLICE_TYPE_P][I_PCM] + h->stat.i_mb_count[SLICE_TYPE_B][I_PCM]);
        int64_t i_skip = (h->stat.i_mb_count[SLICE_TYPE_I][P_SKIP] + h->stat.i_mb_count[SLICE_TYPE_P][P_SKIP] + h->stat.i_mb_count[SLICE_TYPE_B][P_SKIP]) + (h->stat.i_mb_count[SLICE_TYPE_I][B_SKIP] + h->stat.i_mb_count[SLICE_TYPE_P][B_SKIP] + h->stat.i_mb_count[SLICE_TYPE_B][B_SKIP]);
        const int i_count = h->stat.i_frame_count[SLICE_TYPE_I] + h->stat.i_frame_count[SLICE_TYPE_P] + h->stat.i_frame_count[SLICE_TYPE_B];
        int64_t i_mb_count = (int64_t) i_count * h->mb.i_mb_count;
        int64_t i_inter = i_mb_count - i_skip - i_intra;
        const double duration = h->stat.f_frame_duration[SLICE_TYPE_I] + h->stat.f_frame_duration[SLICE_TYPE_P] + h->stat.f_frame_duration[SLICE_TYPE_B];
        float f_bitrate = (h->stat.i_frame_size[SLICE_TYPE_I] + h->stat.i_frame_size[SLICE_TYPE_P] + h->stat.i_frame_size[SLICE_TYPE_B]) / duration / 125;
        if (0) {
            char *fieldstats = buf;
            fieldstats[0] = 0;
            if (i_inter)
                fieldstats += sprintf (fieldstats, " inter:%.1f%%", h->stat.i_mb_field[1] * 100.0 / i_inter);
            if (i_skip)
                fieldstats += sprintf (fieldstats, " skip:%.1f%%", h->stat.i_mb_field[2] * 100.0 / i_skip);
            x264_log (h, 2, "field mbs: intra: %.1f%%%s\n", h -> stat.i_mb_field [0] * 100.0 / i_intra, buf);
        }
        if (h->pps->b_transform_8x8_mode) {
            buf[0] = 0;
            if (h->stat.i_mb_count_8x8dct[0])
                sprintf (buf, " inter:%.1f%%", 100.* h->stat.i_mb_count_8x8dct[1] / h->stat.i_mb_count_8x8dct[0]);
            x264_log (h, 2, "8x8 transform intra:%.1f%%%s\n", 100.* i_i8x8 / i_intra, buf);
        }
        if ((h->param.analyse.i_direct_mv_pred == 3 || (h->stat.i_direct_frames[0] && h->stat.i_direct_frames[1])) && h->stat.i_frame_count[SLICE_TYPE_B]) {
            x264_log (h, 2, "direct mvs  spatial:%.1f%% temporal:%.1f%%\n", h -> stat.i_direct_frames [1] * 100./ h -> stat.i_frame_count [SLICE_TYPE_B], h -> stat.i_direct_frames [0] * 100./ h -> stat.i_frame_count [SLICE_TYPE_B]);
        }
        buf[0] = 0;
        int csize = (h->sps->i_chroma_format_idc == CHROMA_444) ? 4 : 1;
        if (i_mb_count != i_all_intra)
            sprintf (buf, " inter: %.1f%% %.1f%% %.1f%%", h->stat.i_mb_cbp[1] * 100.0 / ((i_mb_count - i_all_intra) * 4), h->stat.i_mb_cbp[3] * 100.0 / ((i_mb_count - i_all_intra) * csize), h->stat.i_mb_cbp[5] * 100.0 / ((i_mb_count - i_all_intra) * csize));
        x264_log (h, 2, "coded y,%s,%s intra: %.1f%% %.1f%% %.1f%%%s\n", (h -> sps -> i_chroma_format_idc == CHROMA_444) ? "u" : "uvDC", (h -> sps -> i_chroma_format_idc == CHROMA_444) ? "v" : "uvAC", h -> stat.i_mb_cbp [0] * 100.0 / (i_all_intra * 4), h -> stat.i_mb_cbp [2] * 100.0 / (i_all_intra * csize), h -> stat.i_mb_cbp [4] * 100.0 / (i_all_intra * csize), buf);
        int64_t fixed_pred_modes [4] [9] = {{0}};
        int64_t sum_pred_modes [4] = {0};
        for (int i = 0; i <= I_PRED_16x16_DC_128; i++) {
            fixed_pred_modes[0][x264_mb_pred_mode16x16_fix[i]] += h->stat.i_mb_pred_mode[0][i];
            sum_pred_modes[0] += h->stat.i_mb_pred_mode[0][i];
        }
        if (sum_pred_modes[0])
            x264_log (h, 2, "i16 v,h,dc,p: %2.0f%% %2.0f%% %2.0f%% %2.0f%%\n", fixed_pred_modes[0][0] * 100.0 / sum_pred_modes[0], fixed_pred_modes[0][1] * 100.0 / sum_pred_modes[0], fixed_pred_modes[0][2] * 100.0 / sum_pred_modes[0], fixed_pred_modes[0][3] * 100.0 / sum_pred_modes[0]);
        for (int i = 1; i <= 2; i++) {
            for (int j = 0; j <= I_PRED_8x8_DC_128; j++) {
                fixed_pred_modes[i][x264_mb_pred_mode4x4_fix[(j) +1]] += h->stat.i_mb_pred_mode[i][j];
                sum_pred_modes[i] += h->stat.i_mb_pred_mode[i][j];
            }
            if (sum_pred_modes[i])
                x264_log (h, 2, "i%d v,h,dc,ddl,ddr,vr,hd,vl,hu: %2.0f%% %2.0f%% %2.0f%% %2.0f%% %2.0f%% %2.0f%% %2.0f%% %2.0f%% %2.0f%%\n", (3 - i) * 4, fixed_pred_modes[i][0] * 100.0 / sum_pred_modes[i], fixed_pred_modes[i][1] * 100.0 / sum_pred_modes[i], fixed_pred_modes[i][2] * 100.0 / sum_pred_modes[i], fixed_pred_modes[i][3] * 100.0 / sum_pred_modes[i], fixed_pred_modes[i][4] * 100.0 / sum_pred_modes[i], fixed_pred_modes[i][5] * 100.0 / sum_pred_modes[i], fixed_pred_modes[i][6] * 100.0 / sum_pred_modes[i], fixed_pred_modes[i][7] * 100.0 / sum_pred_modes[i], fixed_pred_modes[i][8] * 100.0 / sum_pred_modes[i]);
        }
        for (int i = 0; i <= I_PRED_CHROMA_DC_128; i++) {
            fixed_pred_modes[3][x264_mb_chroma_pred_mode_fix[i]] += h->stat.i_mb_pred_mode[3][i];
            sum_pred_modes[3] += h->stat.i_mb_pred_mode[3][i];
        }
        if (sum_pred_modes[3] && !(h->sps->i_chroma_format_idc == CHROMA_444))
            x264_log (h, 2, "i8c dc,h,v,p: %2.0f%% %2.0f%% %2.0f%% %2.0f%%\n", fixed_pred_modes[3][0] * 100.0 / sum_pred_modes[3], fixed_pred_modes[3][1] * 100.0 / sum_pred_modes[3], fixed_pred_modes[3][2] * 100.0 / sum_pred_modes[3], fixed_pred_modes[3][3] * 100.0 / sum_pred_modes[3]);
        if (h->param.analyse.i_weighted_pred >= 1 && h->stat.i_frame_count[SLICE_TYPE_P] > 0)
            x264_log (h, 2, "Weighted P-Frames: Y:%.1f%% UV:%.1f%%\n", h->stat.i_wpred[0] * 100.0 / h->stat.i_frame_count[SLICE_TYPE_P], h->stat.i_wpred[1] * 100.0 / h->stat.i_frame_count[SLICE_TYPE_P]);
        for (int i_list = 0; i_list < 2; i_list++)
            for (int i_slice = 0; i_slice < 2; i_slice++) {
                char *p = buf;
                int64_t i_den = 0;
                int i_max = 0;
                for (int i = 0; i < 16 * 2; i++)
                    if (h->stat.i_mb_count_ref[i_slice][i_list][i]) {
                        i_den += h->stat.i_mb_count_ref[i_slice][i_list][i];
                        i_max = i;
                    }
                if (i_max == 0)
                    continue;
                for (int i = 0; i <= i_max; i++)
                    p += sprintf (p, " %4.1f%%", 100.* h->stat.i_mb_count_ref[i_slice][i_list][i] / i_den);
                x264_log (h, 2, "ref %c L%d:%s\n", "PB" [i_slice], i_list, buf);
            }
        if (h->param.analyse.b_ssim) {
            float ssim = (h->stat.f_ssim_mean_y[SLICE_TYPE_I] + h->stat.f_ssim_mean_y[SLICE_TYPE_P] + h->stat.f_ssim_mean_y[SLICE_TYPE_B]) / duration;
            x264_log (h, 2, "SSIM Mean Y:%.7f (%6.3fdb)\n", ssim, x264_ssim (ssim));
        }
        if (h->param.analyse.b_psnr) {
            x264_log (h, 2, "PSNR Mean Y:%6.3f U:%6.3f V:%6.3f Avg:%6.3f Global:%6.3f kb/s:%.2f\n", (h -> stat.f_psnr_mean_y [SLICE_TYPE_I] + h -> stat.f_psnr_mean_y [SLICE_TYPE_P] + h -> stat.f_psnr_mean_y [SLICE_TYPE_B]) / duration, (h -> stat.f_psnr_mean_u [SLICE_TYPE_I] + h -> stat.f_psnr_mean_u [SLICE_TYPE_P] + h -> stat.f_psnr_mean_u [SLICE_TYPE_B]) / duration, (h -> stat.f_psnr_mean_v [SLICE_TYPE_I] + h -> stat.f_psnr_mean_v [SLICE_TYPE_P] + h -> stat.f_psnr_mean_v [SLICE_TYPE_B]) / duration, (h -> stat.f_psnr_average [SLICE_TYPE_I] + h -> stat.f_psnr_average [SLICE_TYPE_P] + h -> stat.f_psnr_average [SLICE_TYPE_B]) / duration, x264_psnr ((h -> stat.f_ssd_global [SLICE_TYPE_I] + h -> stat.f_ssd_global [SLICE_TYPE_P] + h -> stat.f_ssd_global [SLICE_TYPE_B]), duration * i_yuv_size), f_bitrate);
        }
        else
            x264_log (h, 2, "kb/s:%.2f\n", f_bitrate);
    }
    x264_ratecontrol_delete (h);
    if (h->param.rc.psz_stat_out)
        free (h->param.rc.psz_stat_out);
    if (h->param.rc.psz_stat_in)
        free (h->param.rc.psz_stat_in);
    x264_cqm_delete (h);
    x264_free (h -> nal_buffer);
    x264_free (h -> reconfig_h);
    x264_analyse_free_costs (h);
    if (h->i_thread_frames > 1)
        h = h->thread[h->i_thread_phase];
    x264_frame_delete_list (h -> frames.unused [0]);
    x264_frame_delete_list (h -> frames.unused [1]);
    x264_frame_delete_list (h -> frames.current);
    x264_frame_delete_list (h -> frames.blank_unused);
    h = h->thread[0];
    for (int i = 0; i < h->i_thread_frames; i++)
        if (h->thread[i]->b_thread_active)
            for (int j = 0; j < h->thread[i]->i_ref[0]; j++)
                if (h->thread[i]->fref[0][j] && h->thread[i]->fref[0][j]->b_duplicate)
                    x264_frame_delete (h->thread[i]->fref[0][j]);
    if (h->param.i_lookahead_threads > 1)
        for (int i = 0; i < h->param.i_lookahead_threads; i++)
            x264_free (h->lookahead_thread[i]);
    for (int i = h->param.i_threads - 1; i >= 0; i--) {
        x264_frame_t **frame;
        if (!h->param.b_sliced_threads || i == 0) {
            for (frame = h->thread[i]->frames.reference; *frame; frame++) {
                (((*frame)->i_reference_count > 0) ? (void) (0) : __assert_fail ("(*frame)->i_reference_count > 0", "DonorSourceCode/encoder.c", 4340, __PRETTY_FUNCTION__));
                (*frame)->i_reference_count--;
                if ((*frame)->i_reference_count == 0)
                    x264_frame_delete (*frame);
            }
            frame = &h->thread[i]->fdec;
            if (*frame) {
                (((*frame)->i_reference_count > 0) ? (void) (0) : __assert_fail ("(*frame)->i_reference_count > 0", "DonorSourceCode/encoder.c", 4348, __PRETTY_FUNCTION__));
                (*frame)->i_reference_count--;
                if ((*frame)->i_reference_count == 0)
                    x264_frame_delete (*frame);
            }
            x264_macroblock_cache_free (h -> thread [i]);
        }
        x264_macroblock_thread_free (h -> thread [i], 0);
        x264_free (h -> thread [i] -> out.p_bitstream);
        x264_free (h -> thread [i] -> out.nal);
        ;
        ;
        x264_free (h -> thread [i]);
    }
    x264_opencl_close_library (ocl);
}

int x264_encoder_delayed_frames (x264_t *h) {
    int delayed_frames = 0;
    if (h->i_thread_frames > 1) {
        for (int i = 0; i < h->i_thread_frames; i++)
            delayed_frames += h->thread[i]->b_thread_active;
        h = h->thread[h->i_thread_phase];
    }
    for (int i = 0; h->frames.current[i]; i++)
        delayed_frames++;
    ;
    ;
    ;
    delayed_frames += h->lookahead->ifbuf.i_size + h->lookahead->next.i_size + h->lookahead->ofbuf.i_size;
    ;
    ;
    ;
    return delayed_frames;
}

int x264_encoder_maximum_delayed_frames (x264_t *h) {
    return h->frames.i_delay;
}

